import os as _os
from pathlib import Path
import bpy 
import mathutils
import addon_utils
import numpy as np
from random import random
from math import pi
from Physics_Lib import Particle as _particle
from miscellaneous import miscellaneous as _MS

class BlenderHandler:

    def __init__(self):
        print('Using Blender version: ' + bpy.app.version_string)
        #Blender addons needed
        addon_utils.enable(module_name='io_import_images_as_planes') 

    def find_3dview_space(self):
        # find the 3D view panel and its screen space
        area = None
        for a in bpy.data.window_managers[0].windows[0].screen.areas:
            if a.type == "VIEW_3D":
                area = a
                break
        return area.spaces[0] if area else bpy.context.space_data

    def setup_scene(self,START_FRAME,END_FRAME):
        # (set a black background)
        bpy.data.worlds["World"].node_tree.nodes["Background"].inputs[0].default_value = (0, 0, 0, 1)
        # (make sure we use the EEVEE render engine + enable bloom effect)
        scene = bpy.context.scene
        scene.render.engine = "BLENDER_EEVEE"
        scene.eevee.use_bloom = True
        # (set the animation start/end/current frames)
        scene.frame_start = START_FRAME
        scene.frame_end = END_FRAME
        scene.frame_current = START_FRAME
        # get the current 3D view (among all visible windows
        # in the workspace)
        space =self. find_3dview_space()
        # apply a "rendered" shading mode + hide all
        # additional markers, grids, cursors...
        space.shading.type = 'RENDERED'
        space.overlay.show_floor = False
        space.overlay.show_axis_x = False
        space.overlay.show_axis_y = False
        space.overlay.show_cursor = False
        space.overlay.show_object_origins = False

    def create_emission_shader(self,color, strength, mat_name):
        # create a new material resource (with its
        # associated shader)
        mat = bpy.data.materials.new(mat_name)
        # enable the node-graph edition mode
        mat.use_nodes = True
    
        # clear all starter nodes
        nodes = mat.node_tree.nodes
        nodes.clear()

        # add the Emission node
        node_emission = nodes.new(type="ShaderNodeEmission")
        # (input[0] is the color)
        node_emission.inputs[0].default_value = color
        # (input[1] is the strength)
        node_emission.inputs[1].default_value = strength
    
        # add the Output node
        node_output = nodes.new(type="ShaderNodeOutputMaterial")
    
        # link the two nodes
        links = mat.node_tree.links
        link = links.new(node_emission.outputs[0], node_output.inputs[0])

        # return the material reference
        return mat

    def create_sphere(self,radius, obj_name, offset=[0,0,0]):
        # instantiate a UV sphere with a given
        # radius, at a given distance from the
        # world origin point
        obj = bpy.ops.mesh.primitive_uv_sphere_add(
            radius=radius,
            location=(offset[0], offset[1], offset[2])    )
        # rename the object
        bpy.context.object.name = obj_name
        # apply smooth shading
        bpy.ops.object.shade_smooth()
        # return the object reference
        return bpy.context.object

    def create_bounding_box(self,size, obj_name, offset=[0,0,0]):
        # instantiate a UV sphere with a given
        # radius, at a given distance from the
        # world origin point
        obj = bpy.ops.mesh.primitive_cube_add(
                size=size, 
                location=(offset[0], offset[1], offset[2]))

        # rename the object
        bpy.context.object.name = obj_name
        # apply smooth shading
        bpy.ops.object.shade_smooth()
        # return the object reference
        return bpy.context.object

    def create_torus(self,radius, obj_name):
        # (same as the create_sphere method)
        obj = bpy.ops.mesh.primitive_torus_add(
            location=(0, 0, 0),
            major_radius=radius,
            minor_radius=0.1,
            major_segments=60
        )
        bpy.context.object.name = obj_name
        # apply smooth shading
        bpy.ops.object.shade_smooth()
        # return the object reference
        return bpy.context.object

    def delete_object(self,name):
        # try to find the object by name
        if name in bpy.data.objects:
            # if it exists, select it and delete it
            obj = bpy.data.objects[name]
            obj.select_set(True)
            bpy.ops.object.delete(use_global=False)

    def delete_all_objects(self):
        # Find all object and select them 
        for index in range(len(bpy.data.objects)):
            # if it exists, select it and delete it
            bpy.data.objects[index].select_set(True)
        #Delete all selected objects
        bpy.ops.object.delete(use_global=False)

        for m in bpy.data.materials:
            bpy.data.materials.remove(m)

        for m in bpy.data.collections:
         bpy.data.collections.remove(m)

    def append_object_to_scene(self,blendfile,section,object):
        """
        Opens a blender file and appends the object to the current scene
        """
        filepath  = blendfile + section + object
        directory = blendfile + section
        filename  = object

        bpy.ops.wm.append(
                filepath=filepath, 
                filename=filename,
                directory=directory
            )

    def open_blender_file(self,filepath):
        """
        Opens a blender file 
        """
        bpy.ops.wm.open_mainfile(filepath=filepath)

    def get_create_object(self, scene, OBJECT_NAME, MESH_NAME, particle_start_pos ):
        ob = scene.objects.get(OBJECT_NAME)
        if ob == None:
            me = bpy.data.meshes.new(MESH_NAME)
            ob = bpy.data.objects.new(OBJECT_NAME, me)
            scene.collection.objects.link(ob)
            ob.location=particle_start_pos
        return ob

    def do_raycast(self, scene, object_pos=(0,0,0),direction_start=(0, 0, 1), MAXIMUM_ITERATIONS=10,EPSILON=0.000001):
        """
        Cast ray in one direction and find nearst surface
        """
        points = [mathutils.Vector(object_pos)]
        direction = mathutils.Vector(direction_start)

        for i in range(MAXIMUM_ITERATIONS):
            origin = points[-1] + direction * EPSILON
            result = scene.ray_cast(scene.view_layers[0], origin=origin, direction = direction)
            hit, location, normal, index, ob, matrix = result
            if not hit:
                break
        
            if normal.dot(direction) < 0:
                normal *= -1
        
            rot_dif = direction.rotation_difference(normal)
            direction.rotate(rot_dif)
            direction.rotate(rot_dif)
            direction *= -1
            points.append(location)
        
        #points.append(points[-1] + direction*10)
        return points

    def get_fold_particle_in_Geometry(self, ScanProject,txyzn_Traj, save_blender_file=1, file_end=''):
        """
        The idea here is to have a trajectory that needs to be folded in to a geometry.
        How well this preserves the trajectory characteristics is not known.
        """
    
        #Creating simple browniana trajectory for testing 
        dtdxdydz=np.diff(txyzn_Traj[:4,:],axis=1)
        delta_t=dtdxdydz[0,0] #s
        dxdydz=dtdxdydz[1:,:].T
        position_xyz=txyzn_Traj[1:4,:]
        time_steps=position_xyz.shape[1]

        if 0: #Debug
            time_steps=100000 
            delta_t=20e-6 #s
            mesh_size=1e-2 #um
            diff_coef=1 #um^2/s per dimension 
            initial_pos_um=(0,0,0)

            time=delta_t*np.arange(time_steps)
    
            dxdydz = np.random.normal(0, np.sqrt(2*diff_coef*delta_t), (int(time_steps-1), 3))
            traj = np.cumsum(np.vstack((initial_pos_um, dxdydz)), axis=0)
            position_xyz=traj.T

        ##### Blender part
        #Clean Scene
        self.delete_all_objects()
        # setup scene settings
        START_FRAME = 1
        END_FRAME = time_steps
        self.setup_scene(START_FRAME,END_FRAME)

        #Import geometry to fold path in
        geometry_pos=(0,0,0)
        ImagePath="C:/Users/elias/Desktop/Shadowfax/BlenderDev/BlenderObjectLib/Blender_Files/"
        blendfile = ImagePath + "BoundingCube_Cell.blend"
        section   = "\\Collection\\"
        object    = "Biology"
        self.append_object_to_scene(blendfile,section,object)
        bpy.data.objects["Cell-01"].location=geometry_pos


        #Add Particle path 
        bpy.ops.object.select_all(action='DESELECT')
        if 0:

            context = bpy.context
            curve = bpy.data.curves.new(name="particle-path", type='CURVE')
            curve.dimensions = '3D'
            spline = curve.splines.new('POLY')
            spline.points.add(position_xyz.shape[1] - 1)
            spline.points.foreach_set("co", np.vstack((position_xyz,np.ones(position_xyz.shape[1]))).T.flatten())
            ob = bpy.data.objects.new("Particle-path", curve)     
            context.collection.objects.link(ob)
       
        #Folding
        position_xyz_new=np.zeros(position_xyz.shape)
        o=bpy.data.objects['Cell-01']       #object defining boundery
        smwi = o.matrix_world.inverted()    #Convert world to local 
        smw  = o.matrix_world               #Convert local to world
        epsilon=0.001                       #[um] particle radius

        pos_index_prev=0
        while pos_index_prev<position_xyz.shape[1]:
            print(pos_index_prev)
            oregin_index=pos_index_prev
            if oregin_index<0:
                oregin_index=0

            #### Debug Test in a cube
            if 0:
                if np.sum(np.abs(mathutils.Vector(position_xyz[:,oregin_index]))>np.array([1,1,1]))>0:
                    print('Crossing detected for new origin')
            #########
            origin=mathutils.Vector(position_xyz[:,oregin_index])

            (hit, loc, norm, face_index) = o.closest_point_on_mesh(smwi @ origin)
            loc=smw@loc #convert to world 
        
            if hit:
                max_radius=( loc - origin).length - epsilon
                if max_radius<=0:
                    print('Wall closer then paricle size')

                all_distance_squared=np.sum((position_xyz[:,pos_index_prev:].T-np.array(origin))**2,axis=1)
                distance_sphere=np.where(all_distance_squared>=max_radius**2)[0]
                if len(distance_sphere)>0: #There are points outside the bounding sphere
                    crossing_index=np.min(distance_sphere)

                    if crossing_index>1: #Add all points with in the bounding sphere and rerun "closest_point_on_mesh" from the last point insided the sphere
                        crossing_index+=pos_index_prev
                        position_xyz_new[:,pos_index_prev:crossing_index] =position_xyz[:,pos_index_prev:crossing_index]

                        if pos_index_prev<crossing_index-1:
                            pos_index_prev=crossing_index-1
                        else:
                            print('Something is wrong')
                    elif crossing_index==1:#Next point in line is close to the wall
                        crossing_index+=pos_index_prev

                        ray_origin_world=mathutils.Vector(position_xyz_new[:,crossing_index-1])
                        ray_origin_local=smwi @ ray_origin_world
                        direction_origin=mathutils.Vector(dxdydz[crossing_index-1,:])
                        next_point_cantidate=mathutils.Vector(position_xyz[:,crossing_index])

                        next_point=None
                        while next_point==None:
                            print("Searching for next point")
                            #Test 1: See if next point is hiting the wall, Obs we can be on the other side of the wall      
                            (hit_next, loc_next, norm_next, _ ) = o.closest_point_on_mesh(smwi @ next_point_cantidate )
                            loc_next=smw@loc_next #convert to world 
                            Test_1= (loc_next-next_point_cantidate).length>=epsilon

                            #Test 2: See if we are on the right side of the wall
                            ray_result = o.ray_cast(origin=ray_origin_local, distance=direction_origin.length, direction = direction_origin)
                    
                            #Evaluating the test results  
                            if (Test_1)*(not ray_result[0]):
                                #case 1: we are larger then epsilon away and no ray hit (thus inside the volume) from previus position to new position. 
                                print('Ray Case 1')
                                next_point=next_point_cantidate

                            elif(not Test_1)*(not ray_result[0]):  
                                #case 2: we are small then epsilon away and no ray hit (thus inside the volume) from previus position to new position. 
                                #        Thus we have hit the wall inside the volume
                                print('Ray Case 2')
                                direction=direction_origin*1
                                direction_norm=direction/direction.length
                                angle=np.arccos(direction_norm@norm_next)
                                if angle>np.pi/2:#Normal pointing in to the volume, oposite diraction as our movement  
                                    angle=angle - np.pi/2
                                else:#Normal pointing out of the volume, same diraction as our movement  
                                    angle=np.pi/2-angle

                                h=epsilon-(loc_next-next_point_cantidate).length
                                length_after_wall=h/np.sin(angle)
                                length_before_wall=direction.length-length_after_wall
                                point_before_wall=ray_origin_world + direction_norm*length_before_wall

                                ## Calculate reflection direction
                                wall_norm=norm_next
                                if wall_norm.dot(direction) < 0:
                                    wall_norm *= -1
        
                                rot_dif = direction.rotation_difference(wall_norm)
                                direction.rotate(rot_dif)
                                direction.rotate(rot_dif)
                                direction *= -(length_after_wall/direction.length)
                        
                                ## New position candidate
                                ray_origin_world=point_before_wall
                                ray_origin_local=smwi @ ray_origin_world
                                direction_origin=direction
                                next_point_cantidate=point_before_wall + direction
                                print('New iteration')

                            elif(not Test_1)*(ray_result[0]):  
                                # 3 we have smaller then epsilon and a ray hit, we have hit the wall and past through and need to do a reflection
                                #Same as case 4
                                print('Ray Case 3')
                                direction=direction_origin*1
                                direction_norm=direction/direction.length
                                angle=np.arccos(direction_norm@ray_result[2])
                                if angle>np.pi/2:#Normal pointing in to the volume, oposite diraction as our movement  
                                    angle=angle-np.pi/2
                                else:#Normal pointing out of the volume, same diraction as our movement  
                                    angle=np.pi/2-angle

                                to_wall_vector=ray_result[1]-ray_origin_world
                                h=epsilon/np.sin(angle)
                                length_before_wall=to_wall_vector.length-h
                                length_after_wall=direction.length-length_before_wall
                                point_before_wall = ray_origin_world+direction_norm*length_before_wall
                            

                                ## Calculate reflection direction
                                wall_norm=ray_result[2]
                                if wall_norm.dot(direction) < 0:
                                    wall_norm *= -1
        
                                rot_dif = direction.rotation_difference(wall_norm)
                                direction.rotate(rot_dif)
                                direction.rotate(rot_dif)
                                direction *= -(length_after_wall/direction.length)
                        
                                ## New position candidate
                                ray_origin_world=point_before_wall
                                ray_origin_local=smwi @ ray_origin_world
                                direction_origin=direction
                                next_point_cantidate=point_before_wall + direction
                                print('New iteration')

                            elif(Test_1)*(ray_result[0]):  
                                # 4 we have bigger  then epsilon and a ray hit, we have hit the wall and past through a large distance and need to do a reflection
                                # using the normal of the face of the ray crossing
                                #Same as case 3
                                print('Ray Case 4')
                                direction=direction_origin*1
                                direction_norm=direction/direction.length
                                angle=np.arccos(direction_norm@ray_result[2])
                                if angle>np.pi/2:#Normal pointing in to the volume, oposite diraction as our movement  
                                    angle=angle-np.pi/2
                                else:#Normal pointing out of the volume, same diraction as our movement  
                                    angle=np.pi/2-angle

                                to_wall_vector=ray_result[1]-ray_origin_world
                                h=epsilon/np.sin(angle)
                                length_before_wall=to_wall_vector.length-h
                                length_after_wall=direction.length-length_before_wall
                                point_before_wall = ray_origin_world+direction_norm*length_before_wall
                            

                                ## Calculate reflection direction
                                wall_norm=ray_result[2]
                                if wall_norm.dot(direction) < 0:
                                    wall_norm *= -1
        
                                rot_dif = direction.rotation_difference(wall_norm)
                                direction.rotate(rot_dif)
                                direction.rotate(rot_dif)
                                direction *= -(length_after_wall/direction.length)
                        
                                ## New position candidate
                                ray_origin_world=point_before_wall
                                ray_origin_local=smwi @ ray_origin_world
                                direction_origin=direction
                                next_point_cantidate=point_before_wall + direction
                                print('New iteration')
                            else:
                                print('Ray case error, somethign wrong in the cases')
                        
                        position_xyz_new[:,crossing_index] =next_point
                        position_xyz[:,crossing_index] =next_point
                        position_xyz[:,crossing_index+1:]=(position_xyz[:,crossing_index]+np.cumsum(dxdydz[crossing_index:,:],axis=0)).T #Re generate the random walk from crossing_index
                        pos_index_prev=crossing_index
                    else:
                        print('Strange crossing_index value: ' + str(crossing_index))

                else: #All point are within the bounding sphere, add and stop the loop
                    position_xyz_new[:,pos_index_prev:] =position_xyz[:,pos_index_prev:]
                    pos_index_prev=position_xyz.shape[1]
            else:
                print('Something is wrong no object found')
                return 
        print('Final Result: ' + str(position_xyz.shape))
        #Overwite old data
        txyzn_Traj[1:4,:]=position_xyz

        if save_blender_file: #DeBug add path to scen and save file for viewing 
            #Add Particle path 
            bpy.ops.object.select_all(action='DESELECT')
        
            context = bpy.context
            curve = bpy.data.curves.new(name="particle-path-folded", type='CURVE')
            curve.dimensions = '3D'
            spline = curve.splines.new('POLY')
            spline.points.add(position_xyz_new.shape[1] - 1)
            spline.points.foreach_set("co", np.vstack((position_xyz_new,np.ones(position_xyz_new.shape[1]))).T.flatten())
            ob = bpy.data.objects.new("Particle-path-folded", curve)     
            context.collection.objects.link(ob)

            # Save file
            bpy.ops.object.select_all(action='DESELECT')
            file='BlenderFoldedPath_' + file_end + '.blend'
            path=ScanProject.Files['AnalysisConfig'][:-18] + 'BlenderGroundTruth/'
            Path(path).mkdir(parents=True, exist_ok=True)
            bpy.ops.wm.save_as_mainfile(filepath=path  + file)
            print('File Saved!')

        return txyzn_Traj

class BlenderTrajPlotHandler:
    def get_traj_to_blender(ScanProject,TrajToGet,TrajDictionary,ImagePlane=True):
        """
        For Blender we will use the scale that in Blender 1m == 1um 

        #Creating test trajectory
        time_steps=1000 
        delta_t=0.6e-3 #s
        mesh_size=1e-2 #um

        p=_particle((0,0,0),time_steps=time_steps,delta_t=delta_t,mesh_size=mesh_size)
        p.make_diffusive_traj(diff_coef=1, reflections=False)
        position_xyz=p.diffusive_traj
        time=delta_t*np.arange(time_steps)

        """

        #Trajectory data
        txyznM=TrajDictionary[TrajToGet]*1.0
        time=txyznM[0,:]   #[s]
        position_xyz=txyznM[1:4,:]
                      
        #Piezo stage position
        TStartEnd=[time[0],time[-1]] 
        StartEndPos=np.where(np.abs(np.diff((ScanProject.ScanMatrix[0,:]>TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<=TStartEnd[1]))))[0]
        StartEndPos=np.expand_dims(np.append(np.array([0]),StartEndPos),axis=1)
            
        T_Raw,XP_Raw,YP_Raw,XE_Raw,YE_Raw,XO_Raw,YO_Raw,ZI_Raw,APD1_Raw,APD2_Raw,J_Raw,MemIndex_Raw,ZI_um_Raw,ZI_OffsetIndex_Raw, zCam_Raw =  _MS.get_traj_raw_data4(StartEndPos, 0, ScanProject.ScanMatrix, ScanProject)
        txyz_Traj_Piezo=np.vstack((T_Raw,XP_Raw,YP_Raw,ZI_um_Raw))
        txyz_Traj_Piezo=_MS.get_Scan_toImagCoordinates(ScanProject, txyz_Traj_Piezo)
        txyz_Traj_Piezo=txyz_Traj_Piezo[:,::int(ScanProject.PosCalibration['TrackingMemsize'])]

        if np.sum(np.abs(txyz_Traj_Piezo[0,:]-time))>0:
            print('Piezo stage time is out of sync with MLE')

        ##### Blender part
        #Clean Scene
        BlenderH=BlenderHandler()
        BlenderH.delete_all_objects()

        # setup scene settings
        START_FRAME = 1
        END_FRAME = len(time)
        BlenderH.setup_scene(START_FRAME,END_FRAME)

        ### Import image plane
        ImagePath=ScanProject.Files['AnalysisConfig'][:-18] + "widefield/"
        ImageFile="widefield.tif"
        ImageMaskFile="Mask.tif"

            #Re-scale image pixels to um 
        Scaling=ScanProject.SetupSpecifics['NM']*(ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2']))*1e6
        if ImagePlane:
            try:
                bpy.ops.import_image.to_plane(
                    files=[{"name":ImageFile}], 
                    directory=ImagePath)

                activOB='widefield'
                bpy.data.objects[activOB].rotation_euler=(np.pi, 0, 0)
                bpy.data.objects[activOB].scale=(Scaling[0], Scaling[1], 1)
                bpy.data.objects[activOB].location=(Scaling[0]/2, Scaling[1]/2, np.min(position_xyz[2,:]))
            except:
                print('Blender: No image plane found')

        #Add Light source 
        bpy.ops.object.light_add(type='SUN', radius=1, location=(Scaling[0]/2, Scaling[1]/2, 10))
        bpy.context.object.data.energy = 5

        #Add Camera 
        CamOffset=np.array([20.965,15.17,5.4951])
        CameranPos=0*np.mean(position_xyz,axis=1)+CamOffset
        CameraRot=np.array([80,-2.13,61.5])*(np.pi/180)
        bpy.ops.object.camera_add(enter_editmode=False, align='VIEW', location=(CameranPos[0],CameranPos[1],CameranPos[2]), rotation=(CameraRot[0],CameraRot[1],CameraRot[2]))

        #Import Bounding Box cage for bounding box
        ImagePath=ScanProject.Files['AnalysisConfig'][:-18] + "Blender_Files/"
        blendfile = ImagePath + "BoundingCube.blend"
        section   = "\\Collection\\"
        object    = "BoundingBox"
        BlenderH.append_object_to_scene(blendfile,section,object)

        bpy.ops.object.select_all(action='DESELECT')
        for obj in bpy.data.collections['BoundingBox'].all_objects: 
            obj.select_set(True)
            
            #Scale the BoundBox
        xyzBounds=ScanProject.PoissonProbXYZMap_NeighborhoodBounds[0][0] #All boxes are the same 
        xyzBoundsScale=np.diff(xyzBounds)
        bpy.ops.transform.resize(value=(xyzBoundsScale[0]/2, xyzBoundsScale[1]/2, xyzBoundsScale[2]/2))
        bpy.ops.object.select_all(action='DESELECT')

        #Create particle 
        n=1
        radius=0.05 #um
        particle = BlenderH.create_sphere(radius, "Particle-{:02d}".format(n))
        particle.data.materials.append(
                    BlenderH.create_emission_shader( (random(), random(), 1, 1),
                                                      2,
                                                        "ParticleMat-{:02d}".format(n)
                                                    )
                                        )

        #Add Particle path 
        bpy.ops.object.select_all(action='DESELECT')
        if 0:
            bpy.ops.curve.primitive_bezier_curve_add(enter_editmode=True)
            bpy.ops.curve.subdivide(number_cuts=END_FRAME-2)
            curve = bpy.context.active_object
            curve.name = 'Particle path'
            bez_points = curve.data.splines[0].bezier_points

            for i in range(END_FRAME):
                bez_points[i].handle_left_type = 'VECTOR'
                bez_points[i].handle_right_type = 'VECTOR'
                bez_points[i].co=(position_xyz[0,i],position_xyz[1,i],position_xyz[2,i])
            bpy.ops.object.mode_set(mode='OBJECT')
        else:
            context = bpy.context
            curve = bpy.data.curves.new(name="particle path", type='CURVE')
            curve.dimensions = '3D'
            spline = curve.splines.new('POLY')
            spline.points.add(position_xyz.shape[1] - 1)
            spline.points.foreach_set("co", np.vstack((position_xyz,np.ones(position_xyz.shape[1]))).T.flatten())
            ob = bpy.data.objects.new("Particle path", curve)     
            context.collection.objects.link(ob)

        

        #Animation of Particle
        bpy.ops.object.select_all(action='DESELECT')

        particle.animation_data_create()
        particle.animation_data.action = bpy.data.actions.new(name="DiffusionPathAction")
        #Define fcurve for xyz position
        fcurve={}
        for i in range(3):
            fcurve[i] = particle.animation_data.action.fcurves.new(
                                                                    data_path="location", index=i
                                                                    )
        for i in range(END_FRAME):
            for j in range(3):
                k = fcurve[j].keyframe_points.insert(
                                                    frame=START_FRAME+i,
                                                    value=position_xyz[j,i]
                                                 )
                k.interpolation = "LINEAR"

        #Animation of Boudning box
        bpy.ops.object.select_all(action='DESELECT')
        n=0
        for obj in bpy.data.collections['BoundingBox'].all_objects: 
            obj.animation_data_create()
            obj.animation_data.action = bpy.data.actions.new(name="PathAction-{:02d}".format(n))
            n+=1

            #Define fcurve for each object xyz position
            fcurve={}
            for i in range(3):
                fcurve[i] = obj.animation_data.action.fcurves.new(
                                                                    data_path="location", index=i
                                                                    )
            for i in range(END_FRAME):
                for j in range(3):
                    k = fcurve[j].keyframe_points.insert(
                                                    frame=START_FRAME+i,
                                                    value=txyz_Traj_Piezo[j+1,i]
                                                 )
                    k.interpolation = "LINEAR"


        # save to file
        file="T_" + str(TrajToGet)  + '.blend'
        path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_Blender/'
        Path(path).mkdir(parents=True, exist_ok=True)
        bpy.ops.wm.save_as_mainfile(filepath=path  + file)
        print('File Saved!')

    def get_all_traj_to_blender_from_existing_blender_files(ScanProject):

        BlenderH=BlenderHandler()
        path=ScanProject.Files['AnalysisConfig'][:-18] 
        filepath= path + 'Traj_Blender/'


        for i in range(len(_os.listdir(filepath))):
            file_name='T_'+ str(i) + '.blend'
            if i==0:
                BlenderH.open_blender_file(filepath+file_name)
            else:
                blendfile=filepath + file_name
                section   = "\\Object\\"
                object='Particle path'
                BlenderH.append_object_to_scene(blendfile,section,object)

                object='Particle-01'
                BlenderH.append_object_to_scene(blendfile,section,object)
                

        # save to file
        file='All_traj_Combined'  + '.blend'
        path=ScanProject.Files['AnalysisConfig'][:-18] 
        Path(path).mkdir(parents=True, exist_ok=True)
        bpy.ops.wm.save_as_mainfile(filepath=path  + file)
        print('File Saved!')

def main_planet_system_test():
    #Clean Scene
    BlenderH=BlenderHandler()
    BlenderH.delete_all_objects()
    # setup scene settings
    START_FRAME = 1
    END_FRAME = 200
    BlenderH.setup_scene(START_FRAME,END_FRAME)

    #Add planets and rings
    ring_mat = BlenderH.create_emission_shader( (1, 1, 1, 1), 1, "RingMat" )
    N_PLANETS = 6
    for n in range(N_PLANETS):
        # get a random radius (a float in [1, 5])
        r = 1 + random() * 4
        # get a random distace to the origin point:
        # - an initial offset of 30 to get out of the sun's sphere
        # - a shift depending on the index of the planet
        # - a little "noise" with a random float
        d = 30 + n * 12 + (random() * 4 - 2)
        # instantiate the planet with these parameters
        # and a custom object name
        planet = BlenderH.create_sphere(r, d, "Planet-{:02d}".format(n))
        planet.data.materials.append(
                BlenderH.create_emission_shader( (random(), random(), 1, 1),
                                                        2,
                                                         "PlanetMat-{:02d}".format(n)
                                                    )
                                    )
        
        # add the radius ring display
        ring = BlenderH.create_torus(d, "Radius-{:02d}".format(n))
        ring.data.materials.append(ring_mat)

        # set planet as active object
        bpy.context.view_layer.objects.active = planet
        planet.select_set(True)
        # set object origin at world origin
        bpy.ops.object.origin_set(type="ORIGIN_CURSOR", center="MEDIAN")

        # setup the planet animation data
        planet.animation_data_create()
        planet.animation_data.action = bpy.data.actions.new(name="RotationAction")
        fcurve = planet.animation_data.action.fcurves.new(
                                                            data_path="rotation_euler", index=2
                                                          )
        k1 = fcurve.keyframe_points.insert(
                                            frame=START_FRAME,
                                            value=0
                                            )
        k1.interpolation = "LINEAR"
        k2 = fcurve.keyframe_points.insert(
                                            frame=END_FRAME,
                                            value=(2 + random() * 2) * pi
                                            )
        k2.interpolation = "LINEAR"


    # add the sun sphere
    sun = BlenderH.create_sphere(12, 0, "Sun")
    sun.data.materials.append(
                            BlenderH.create_emission_shader(  (1, 0.66, 0.08, 1), 10, "SunMat"
                                                  )
                              )

    # deselect all objects
    bpy.ops.object.select_all(action='DESELECT')
    # save to file
    bpy.ops.wm.save_as_mainfile(filepath='C:/Users/elias/Desktop/my.blend')

def main_simple_diffusing_particle():
    """
    Test of importing a diffusion stat to Blender for ploting
    """
    
    #Creating the trajectory
    time_steps=100000 
    delta_t=20e-6 #s
    mesh_size=1e-2 #um
    diff_coef=1 #um^2/s per dimension 

    p=_particle((0,0,0),time_steps=time_steps,delta_t=delta_t,mesh_size=mesh_size)
    p.make_diffusive_traj(diff_coef=diff_coef, reflections=False)
    position_xyz=(p.diffusive_traj).T
    time=delta_t*np.arange(time_steps)


    ##### Blender part
    #Clean Scene
    BlenderH=BlenderHandler()
    BlenderH.delete_all_objects()
    # setup scene settings
    START_FRAME = 1
    END_FRAME = time_steps
    BlenderH.setup_scene(START_FRAME,END_FRAME)


    #Add Particle path 
    bpy.ops.object.select_all(action='DESELECT')

    context = bpy.context
    curve = bpy.data.curves.new(name="particle-path", type='CURVE')
    curve.dimensions = '3D'
    spline = curve.splines.new('POLY')
    spline.points.add(position_xyz.shape[1] - 1)
    spline.points.foreach_set("co", np.vstack((position_xyz,np.ones(position_xyz.shape[1]))).T.flatten())
    ob = bpy.data.objects.new("Particle-path", curve)     
    context.collection.objects.link(ob)
    bpy.ops.object.convert(target='MESH')

    # Save file
    bpy.ops.object.select_all(action='DESELECT')
    file="Test"  + '.blend'
    path="C:/Users/elias/Desktop/Shadowfax/BlenderDev/"
    #Path(path).mkdir(parents=True, exist_ok=True)
    bpy.ops.wm.save_as_mainfile(filepath=path  + file)
    print('File Saved!')

def main_balistic_particle_in_Geometry():
    """
    Test of figuring out how to have reflective boundery condition for a partickle moving whith out forces.
    OBS, this will only give the points in space where the partickle moment is chaning 
    """
    
    ##### Blender part
    geometry_pos=(0,0,0)
    particle_start_pos=(0,0,0)
    particle_direction_start=(1,0,0)

    #Clean Scene
    BlenderH=BlenderHandler()
    BlenderH.delete_all_objects()
    # setup scene settings
    START_FRAME = 1
    END_FRAME = 100
    BlenderH.setup_scene(START_FRAME,END_FRAME)

    #Import Cell geometry 
    ImagePath="C:/Users/elias/Desktop/Shadowfax/BlenderDev/BlenderObjectLib/Blender_Files/"
    blendfile = ImagePath + "BoundingCube_Cell.blend"
    section   = "\\Collection\\"
    object    = "Biology"
    BlenderH.append_object_to_scene(blendfile,section,object)
    bpy.data.objects["Cell-01"].location=geometry_pos

    #Add Particle path 
    bpy.ops.object.select_all(action='DESELECT')

        #  use ray cast to find interfaces for the paritcle trajectory
    EPSILON = 0.00001
    MAXIMUM_ITERATIONS = 50

    context = bpy.context    
    points=BlenderH.do_raycast(context.scene, object_pos=particle_start_pos, direction_start=particle_direction_start, MAXIMUM_ITERATIONS=MAXIMUM_ITERATIONS,EPSILON=EPSILON)
    position_xyz=np.array(points).T

        # Construct the particle path       
    curve = bpy.data.curves.new(name="particle-path", type='CURVE')
    curve.dimensions = '3D'
    spline = curve.splines.new('POLY')
    spline.points.add(position_xyz.shape[1] - 1)
    spline.points.foreach_set("co", np.vstack((position_xyz,np.ones(position_xyz.shape[1]))).T.flatten())
    ob = bpy.data.objects.new("Particle-path", curve)     
    context.collection.objects.link(ob)
    #bpy.ops.object.convert(target='MESH')


    # Save file
    bpy.ops.object.select_all(action='DESELECT')
    file="Test"  + '.blend'
    path="C:/Users/elias/Desktop/Shadowfax/BlenderDev/"
    #Path(path).mkdir(parents=True, exist_ok=True)
    bpy.ops.wm.save_as_mainfile(filepath=path  + file)
    print('File Saved!')

def main_diffusing_particle_in_Geometry_old():
    """
    Test of importing a diffusion stat to Blender for ploting
    """

    #Creating the trajectory
    time_steps=100000 
    delta_t=20e-6 #s
    mesh_size=1e-2 #um
    diff_coef=1 #um^2/s per dimension 
    initial_pos_um=(0,0,0)

    time=delta_t*np.arange(time_steps)
    
    
    dxdydz = np.random.normal(0, np.sqrt(2*diff_coef*delta_t), (int(time_steps-1), 3))
    traj = np.cumsum(np.vstack((initial_pos_um, dxdydz)), axis=0)
    position_xyz=traj.T

    ##### Blender part
    #Clean Scene
    BlenderH=BlenderHandler()
    BlenderH.delete_all_objects()
    # setup scene settings
    START_FRAME = 1
    END_FRAME = time_steps
    BlenderH.setup_scene(START_FRAME,END_FRAME)

    #Import Cell geometry 
    geometry_pos=(0,0,0)
    ImagePath="C:/Users/elias/Desktop/Shadowfax/BlenderDev/BlenderObjectLib/Blender_Files/"
    blendfile = ImagePath + "BoundingCube_Cell.blend"
    section   = "\\Collection\\"
    object    = "BoundingBox"
    BlenderH.append_object_to_scene(blendfile,section,object)
    bpy.data.objects["BoundingWalls"].location=geometry_pos


    #Add Particle path 
    bpy.ops.object.select_all(action='DESELECT')
    if 1:

        context = bpy.context
        curve = bpy.data.curves.new(name="particle-path", type='CURVE')
        curve.dimensions = '3D'
        spline = curve.splines.new('POLY')
        spline.points.add(position_xyz.shape[1] - 1)
        spline.points.foreach_set("co", np.vstack((position_xyz,np.ones(position_xyz.shape[1]))).T.flatten())
        ob = bpy.data.objects.new("Particle-path", curve)     
        context.collection.objects.link(ob)
       

    #Folding
    position_xyz_new=np.zeros(position_xyz.shape)
    o=bpy.data.objects['BoundingWalls']   #object defining boundery
    smwi = o.matrix_world.inverted() #Convert world to local 
    smw  = o.matrix_world            #Convert local to world
    epsilon=0.001                   #[um] particle radius

    pos_index_prev=0
    offset=0
    while pos_index_prev<position_xyz.shape[1]:
        print(pos_index_prev)
        oregin_index=pos_index_prev-offset
        if oregin_index<0:
            oregin_index=0

        if np.sum(np.abs(mathutils.Vector(position_xyz[:,oregin_index]))>np.array([1,1,1]))>0:
            print('Crossing detected for new origin')
        origin=mathutils.Vector(position_xyz[:,oregin_index])

        (hit, loc, norm, face_index) = o.closest_point_on_mesh(smwi @ origin)
        loc=smw@loc #convert to world 
        
        if hit:
            max_radius=( loc - origin).length - epsilon
            if max_radius<=0:
                print('Wall closer then paricle size')

            all_distance_squared=np.sum((position_xyz[:,pos_index_prev:].T-np.array(origin))**2,axis=1)
            distance_sphere=np.where(all_distance_squared>=max_radius**2)[0]
            if len(distance_sphere)>0:
                crossing_index=np.min(distance_sphere)
                crossing_index+=pos_index_prev
                position_xyz_new[:,pos_index_prev:crossing_index] =position_xyz[:,pos_index_prev:crossing_index]

                ## Debug test
                if np.sum(np.abs(position_xyz_new[:,pos_index_prev:crossing_index].T)>np.array([1,1,1]))>0:
                    print('Crossing detected')
                ####

                direction=mathutils.Vector(dxdydz[crossing_index-1,:])
                ray_origin_world=mathutils.Vector(position_xyz_new[:,crossing_index-1])
                ray_origin_local=smwi @ ray_origin_world
                
                wall_hit=1
                while wall_hit>0:
                    ray_result = o.ray_cast(origin=ray_origin_local, distance=direction.length + epsilon, direction = direction)
                    if ray_result[0]:
                        print("Wall hit: " + str(wall_hit))
                        wall_hit+=1
                        wall_location=smw @ ray_result[1]
                        distans_to_wall=(wall_location - ray_origin_world).length
                        distance_after_wall=direction.length-distans_to_wall

                        if distans_to_wall<epsilon: 
                            print('Wall impact closer then then particle radius')
                        if distance_after_wall<epsilon:
                            print('Wall reflection closer then then particle radius')

                        ## Calculate reflection direction
                        wall_norm=ray_result[2]
                        if wall_norm.dot(direction) < 0:
                            wall_norm *= -1
        
                        rot_dif = direction.rotation_difference(wall_norm)
                        direction.rotate(rot_dif)
                        direction.rotate(rot_dif)
                        direction *= -(distance_after_wall/direction.length)
                        #update ray origin for next iteration, placing 1% of the particle radius from the wall  
                        ray_origin_local = ray_result[1] + (direction/direction.length)*epsilon*0.1
                    else:
                        if wall_hit>1:
                            position_xyz[:,crossing_index]=wall_location+direction
                            position_xyz_new[:,crossing_index]=position_xyz[:,crossing_index] #add to new

                            position_xyz[:,crossing_index+1:]=(position_xyz[:,crossing_index]+np.cumsum(dxdydz[crossing_index:,:],axis=0)).T #Re generate the random walk from crossing_index
                            if pos_index_prev<crossing_index:
                                pos_index_prev=crossing_index
                            else:
                                print('Something is wrong')
                        else:
                            if pos_index_prev<crossing_index-1:
                                pos_index_prev=crossing_index-1
                            else:
                                print('Something is wrong')

                        wall_hit=-1


            else:
                position_xyz_new[:,pos_index_prev:] =position_xyz[:,pos_index_prev:]
                pos_index_prev=position_xyz.shape[1]

    print('Final Result: ' + str(position_xyz.shape))
    #Add Particle path 
    bpy.ops.object.select_all(action='DESELECT')
        
    context = bpy.context
    curve = bpy.data.curves.new(name="particle-path-folded", type='CURVE')
    curve.dimensions = '3D'
    spline = curve.splines.new('POLY')
    spline.points.add(position_xyz_new.shape[1] - 1)
    spline.points.foreach_set("co", np.vstack((position_xyz_new,np.ones(position_xyz_new.shape[1]))).T.flatten())
    ob = bpy.data.objects.new("Particle-path-folded", curve)     
    context.collection.objects.link(ob)

    # Save file
    bpy.ops.object.select_all(action='DESELECT')
    file="Test"  + '.blend'
    path="C:/Users/elias/Desktop/Shadowfax/BlenderDev/"
    #Path(path).mkdir(parents=True, exist_ok=True)
    bpy.ops.wm.save_as_mainfile(filepath=path  + file)
    print('File Saved!')

def main_diffusing_particle_in_Geometry(txyzn_Traj):
    """
    The idea here is to have a trajectory that needs to be folded in to a geometry.
    How well this preserves the trajectory characteristics is not known.
    """
    
    #Creating simple browniana trajectory for testing 
    txyzn_Traj

    if 0:
        time_steps=100000 
        delta_t=20e-6 #s
        mesh_size=1e-2 #um
        diff_coef=1 #um^2/s per dimension 
        initial_pos_um=(0,0,0)

        time=delta_t*np.arange(time_steps)
    
        dxdydz = np.random.normal(0, np.sqrt(2*diff_coef*delta_t), (int(time_steps-1), 3))
        traj = np.cumsum(np.vstack((initial_pos_um, dxdydz)), axis=0)
        position_xyz=traj.T

    ##### Blender part
    #Clean Scene
    BlenderH=BlenderHandler()
    BlenderH.delete_all_objects()
    # setup scene settings
    START_FRAME = 1
    END_FRAME = time_steps
    BlenderH.setup_scene(START_FRAME,END_FRAME)

    #Import geometry to fold path in
    geometry_pos=(0,0,0)
    ImagePath="C:/Users/elias/Desktop/Shadowfax/BlenderDev/BlenderObjectLib/Blender_Files/"
    blendfile = ImagePath + "BoundingCube_Cell.blend"
    section   = "\\Collection\\"
    object    = "Biology"
    BlenderH.append_object_to_scene(blendfile,section,object)
    bpy.data.objects["Cell-01"].location=geometry_pos


    #Add Particle path 
    bpy.ops.object.select_all(action='DESELECT')
    if 0:

        context = bpy.context
        curve = bpy.data.curves.new(name="particle-path", type='CURVE')
        curve.dimensions = '3D'
        spline = curve.splines.new('POLY')
        spline.points.add(position_xyz.shape[1] - 1)
        spline.points.foreach_set("co", np.vstack((position_xyz,np.ones(position_xyz.shape[1]))).T.flatten())
        ob = bpy.data.objects.new("Particle-path", curve)     
        context.collection.objects.link(ob)
       
    #Folding
    position_xyz_new=np.zeros(position_xyz.shape)
    o=bpy.data.objects['Cell-01']       #object defining boundery
    smwi = o.matrix_world.inverted()    #Convert world to local 
    smw  = o.matrix_world               #Convert local to world
    epsilon=0.001                       #[um] particle radius

    pos_index_prev=0
    while pos_index_prev<position_xyz.shape[1]:
        print(pos_index_prev)
        oregin_index=pos_index_prev
        if oregin_index<0:
            oregin_index=0

        #### Debug Test in a cube
        if 0:
            if np.sum(np.abs(mathutils.Vector(position_xyz[:,oregin_index]))>np.array([1,1,1]))>0:
                print('Crossing detected for new origin')
        #########
        origin=mathutils.Vector(position_xyz[:,oregin_index])

        (hit, loc, norm, face_index) = o.closest_point_on_mesh(smwi @ origin)
        loc=smw@loc #convert to world 
        
        if hit:
            max_radius=( loc - origin).length - epsilon
            if max_radius<=0:
                print('Wall closer then paricle size')

            all_distance_squared=np.sum((position_xyz[:,pos_index_prev:].T-np.array(origin))**2,axis=1)
            distance_sphere=np.where(all_distance_squared>=max_radius**2)[0]
            if len(distance_sphere)>0: #There are points outside the bounding sphere
                crossing_index=np.min(distance_sphere)

                if crossing_index>1: #Add all points with in the bounding sphere and rerun "closest_point_on_mesh" from the last point insided the sphere
                    crossing_index+=pos_index_prev
                    position_xyz_new[:,pos_index_prev:crossing_index] =position_xyz[:,pos_index_prev:crossing_index]

                    if pos_index_prev<crossing_index-1:
                        pos_index_prev=crossing_index-1
                    else:
                        print('Something is wrong')
                elif crossing_index==1:#Next point in line is close to the wall
                    crossing_index+=pos_index_prev

                    ray_origin_world=mathutils.Vector(position_xyz_new[:,crossing_index-1])
                    ray_origin_local=smwi @ ray_origin_world
                    direction_origin=mathutils.Vector(dxdydz[crossing_index-1,:])
                    next_point_cantidate=mathutils.Vector(position_xyz[:,crossing_index])

                    next_point=None
                    while next_point==None:
                        print("Searching for next point")
                        #Test 1: See if next point is hiting the wall, Obs we can be on the other side of the wall      
                        (hit_next, loc_next, norm_next, _ ) = o.closest_point_on_mesh(smwi @ next_point_cantidate )
                        loc_next=smw@loc_next #convert to world 
                        Test_1= (loc_next-next_point_cantidate).length>=epsilon

                        #Test 2: See if we are on the right side of the wall
                        ray_result = o.ray_cast(origin=ray_origin_local, distance=direction_origin.length, direction = direction_origin)
                    
                        #Evaluating the test results  
                        if (Test_1)*(not ray_result[0]):
                            #case 1: we are larger then epsilon away and no ray hit (thus inside the volume) from previus position to new position. 
                            print('Ray Case 1')
                            next_point=next_point_cantidate

                        elif(not Test_1)*(not ray_result[0]):  
                            #case 2: we are small then epsilon away and no ray hit (thus inside the volume) from previus position to new position. 
                            #        Thus we have hit the wall inside the volume
                            print('Ray Case 2')
                            direction=direction_origin*1
                            direction_norm=direction/direction.length
                            angle=np.arccos(direction_norm@norm_next)
                            if angle>np.pi/2:#Normal pointing in to the volume, oposite diraction as our movement  
                                angle=angle - np.pi/2
                            else:#Normal pointing out of the volume, same diraction as our movement  
                                angle=np.pi/2-angle

                            h=epsilon-(loc_next-next_point_cantidate).length
                            length_after_wall=h/np.sin(angle)
                            length_before_wall=direction.length-length_after_wall
                            point_before_wall=ray_origin_world + direction_norm*length_before_wall

                            ## Calculate reflection direction
                            wall_norm=norm_next
                            if wall_norm.dot(direction) < 0:
                                wall_norm *= -1
        
                            rot_dif = direction.rotation_difference(wall_norm)
                            direction.rotate(rot_dif)
                            direction.rotate(rot_dif)
                            direction *= -(length_after_wall/direction.length)
                        
                            ## New position candidate
                            ray_origin_world=point_before_wall
                            ray_origin_local=smwi @ ray_origin_world
                            direction_origin=direction
                            next_point_cantidate=point_before_wall + direction
                            print('New iteration')

                        elif(not Test_1)*(ray_result[0]):  
                            # 3 we have smaller then epsilon and a ray hit, we have hit the wall and past through and need to do a reflection
                            #Same as case 4
                            print('Ray Case 3')
                            direction=direction_origin*1
                            direction_norm=direction/direction.length
                            angle=np.arccos(direction_norm@ray_result[2])
                            if angle>np.pi/2:#Normal pointing in to the volume, oposite diraction as our movement  
                                angle=angle-np.pi/2
                            else:#Normal pointing out of the volume, same diraction as our movement  
                                angle=np.pi/2-angle

                            to_wall_vector=ray_result[1]-ray_origin_world
                            h=epsilon/np.sin(angle)
                            length_before_wall=to_wall_vector.length-h
                            length_after_wall=direction.length-length_before_wall
                            point_before_wall = ray_origin_world+direction_norm*length_before_wall
                            

                            ## Calculate reflection direction
                            wall_norm=ray_result[2]
                            if wall_norm.dot(direction) < 0:
                                wall_norm *= -1
        
                            rot_dif = direction.rotation_difference(wall_norm)
                            direction.rotate(rot_dif)
                            direction.rotate(rot_dif)
                            direction *= -(length_after_wall/direction.length)
                        
                            ## New position candidate
                            ray_origin_world=point_before_wall
                            ray_origin_local=smwi @ ray_origin_world
                            direction_origin=direction
                            next_point_cantidate=point_before_wall + direction
                            print('New iteration')

                        elif(Test_1)*(ray_result[0]):  
                            # 4 we have bigger  then epsilon and a ray hit, we have hit the wall and past through a large distance and need to do a reflection
                            # using the normal of the face of the ray crossing
                            #Same as case 3
                            print('Ray Case 4')
                            direction=direction_origin*1
                            direction_norm=direction/direction.length
                            angle=np.arccos(direction_norm@ray_result[2])
                            if angle>np.pi/2:#Normal pointing in to the volume, oposite diraction as our movement  
                                angle=angle-np.pi/2
                            else:#Normal pointing out of the volume, same diraction as our movement  
                                angle=np.pi/2-angle

                            to_wall_vector=ray_result[1]-ray_origin_world
                            h=epsilon/np.sin(angle)
                            length_before_wall=to_wall_vector.length-h
                            length_after_wall=direction.length-length_before_wall
                            point_before_wall = ray_origin_world+direction_norm*length_before_wall
                            

                            ## Calculate reflection direction
                            wall_norm=ray_result[2]
                            if wall_norm.dot(direction) < 0:
                                wall_norm *= -1
        
                            rot_dif = direction.rotation_difference(wall_norm)
                            direction.rotate(rot_dif)
                            direction.rotate(rot_dif)
                            direction *= -(length_after_wall/direction.length)
                        
                            ## New position candidate
                            ray_origin_world=point_before_wall
                            ray_origin_local=smwi @ ray_origin_world
                            direction_origin=direction
                            next_point_cantidate=point_before_wall + direction
                            print('New iteration')
                        else:
                            print('Ray case error, somethign wrong in the cases')
                        

                        if 0:
                            wall_hit=1
                            while wall_hit>0:
                                ray_result = o.ray_cast(origin=ray_origin_local, distance=direction.length + epsilon, direction = direction)
                                if ray_result[0]:
                                    print("Wall hit: " + str(wall_hit))
                                    wall_hit+=1
                                    wall_location=smw @ ray_result[1]
                                    distans_to_wall=(wall_location - ray_origin_world).length
                                    distance_after_wall=direction.length-distans_to_wall

                                    if distans_to_wall<epsilon: 
                                        print('Wall impact closer then then particle radius')
                                    if distance_after_wall<epsilon:
                                        print('Wall reflection closer then then particle radius')

                                    ## Calculate reflection direction
                                    wall_norm=ray_result[2]
                                    if wall_norm.dot(direction) < 0:
                                        wall_norm *= -1
        
                                    rot_dif = direction.rotation_difference(wall_norm)
                                    direction.rotate(rot_dif)
                                    direction.rotate(rot_dif)
                                    direction *= -(distance_after_wall/direction.length)
                                    #update ray origin for next iteration, placing 1% of the particle radius from the wall  
                                    ray_origin_local = ray_result[1] + (direction/direction.length)*epsilon*0.1
                                else:
                                    if wall_hit>1:
                                        position_xyz[:,crossing_index]=wall_location+direction
                                        position_xyz_new[:,crossing_index]=position_xyz[:,crossing_index] #add to new

                                        position_xyz[:,crossing_index+1:]=(position_xyz[:,crossing_index]+np.cumsum(dxdydz[crossing_index:,:],axis=0)).T #Re generate the random walk from crossing_index
                                        if pos_index_prev<crossing_index:
                                            pos_index_prev=crossing_index
                                        else:
                                            print('Something is wrong')
                                    wall_hit=-1
                    position_xyz_new[:,crossing_index] =next_point
                    position_xyz[:,crossing_index] =next_point
                    position_xyz[:,crossing_index+1:]=(position_xyz[:,crossing_index]+np.cumsum(dxdydz[crossing_index:,:],axis=0)).T #Re generate the random walk from crossing_index
                    pos_index_prev=crossing_index
                else:
                    print('Strange crossing_index value: ' + str(crossing_index))

            else: #All point are within the bounding sphere, add and stop the loop
                position_xyz_new[:,pos_index_prev:] =position_xyz[:,pos_index_prev:]
                pos_index_prev=position_xyz.shape[1]
        else:
            print('Something is wrong no object found')
            return 
    print('Final Result: ' + str(position_xyz.shape))


    if 1: #DeBug add path to scen and save file for viewing 
        #Add Particle path 
        bpy.ops.object.select_all(action='DESELECT')
        
        context = bpy.context
        curve = bpy.data.curves.new(name="particle-path-folded", type='CURVE')
        curve.dimensions = '3D'
        spline = curve.splines.new('POLY')
        spline.points.add(position_xyz_new.shape[1] - 1)
        spline.points.foreach_set("co", np.vstack((position_xyz_new,np.ones(position_xyz_new.shape[1]))).T.flatten())
        ob = bpy.data.objects.new("Particle-path-folded", curve)     
        context.collection.objects.link(ob)

        # Save file
        bpy.ops.object.select_all(action='DESELECT')
        file="Test"  + '.blend'
        path="C:/Users/elias/Desktop/Shadowfax/BlenderDev/"
        #Path(path).mkdir(parents=True, exist_ok=True)
        bpy.ops.wm.save_as_mainfile(filepath=path  + file)
        print('File Saved!')

if __name__ == '__main__':
    """
    Main
    """
    main_diffusing_particle_in_Geometry()


    """
    import numpy as np 
    import bpy 
    import mathutils



    def add_path(name, position_xyz):
        context = bpy.context
        curve = bpy.data.curves.new(name=name, type='CURVE')
        curve.dimensions = '3D'
        spline = curve.splines.new('POLY')
        spline.points.add(position_xyz.shape[1] - 1)
        spline.points.foreach_set("co", np.vstack((position_xyz,np.ones(position_xyz.shape[1]))).T.flatten())
        ob = bpy.data.objects.new(name + '-path', curve)     
        context.collection.objects.link(ob)
        #bpy.ops.object.convert(target='MESH')
    


    #Creating the trajectory
    time_steps=100000 
    delta_t=20e-6 #s
    mesh_size=1e-2 #um
    diff_coef=1 #um^2/s per dimension 
    initial_pos_um=(0,0,0)
    epsilon=0.00001
    time=delta_t*np.arange(time_steps)


    dxdydz = np.random.normal(0, np.sqrt(2*diff_coef*delta_t), (int(time_steps-1), 3))
    traj = np.cumsum(np.vstack((initial_pos_um, dxdydz)), axis=0)
    position_xyz=traj.T


    #Add Particle path 
    bpy.ops.object.select_all(action='DESELECT')
    add_path('original', position_xyz)


    #Folding
    position_xyz_new=np.zeros(position_xyz.shape)
    o=bpy.data.objects['Cube']   #object defining boundery
    smwi = o.matrix_world.inverted() #Convert world to local 
    smw  = o.matrix_world            #Convert local to world

    pos_index_prev=0
    S=1
    while pos_index_prev<position_xyz.shape[1]:
        print(pos_index_prev)
        origin=mathutils.Vector(position_xyz[:,pos_index_prev])

        (hit, loc, norm, face_index) = o.closest_point_on_mesh(smwi @ origin)
        loc=smw@loc #convert to world 

        if hit:
            max_radius_squared=(loc-origin).length_squared
            all_distance_squared=np.sum((position_xyz[:,pos_index_prev:].T-np.array(origin))**2,axis=1)
            distance_sphere=np.where(all_distance_squared>=max_radius_squared)[0]
            if len(distance_sphere)>0:
                crossing_index=distance_sphere[0]
                crossing_index+=pos_index_prev
                position_xyz_new[:,pos_index_prev:crossing_index] =position_xyz[:,pos_index_prev:crossing_index]
            
                direction=mathutils.Vector(dxdydz[crossing_index-1,:])
                ray_result = o.ray_cast(origin=smwi @ mathutils.Vector(position_xyz_new[:,crossing_index-1]),  distance=direction.length ,direction = direction)

                if ray_result[0]:
                    print("Wall hit")
                    wall_location=smw@ray_result[1]
                    distans_to_wall=(wall_location-mathutils.Vector(position_xyz_new[:,crossing_index-1])).length
                    distance_after_wall=direction.length-distans_to_wall
                    add_path('Wall_hit', position_xyz[:,crossing_index-1:crossing_index+1])
                
                    wall_norm=ray_result[2]
                    if wall_norm.dot(direction) < 0:
                        wall_norm *= -1
    
                    rot_dif = direction.rotation_difference(wall_norm)
                    direction.rotate(rot_dif)
                    direction.rotate(rot_dif)
                    direction *= -(distance_after_wall/direction.length)
                    position_xyz[:,crossing_index]=wall_location+direction
                    position_xyz_new[:,crossing_index]=position_xyz[:,crossing_index]
                    position_xyz[:,crossing_index+1:]=(position_xyz[:,crossing_index]+np.cumsum(dxdydz[crossing_index:,:],axis=0)).T

                pos_index_prev=crossing_index
            else:
                position_xyz_new[:,pos_index_prev:] =position_xyz[:,pos_index_prev:]
                pos_index_prev=position_xyz.shape[1]


    add_path('Folded', position_xyz_new)
    """

    print("Done!")