import configparser # for .ini files
import numpy as np
import re
import ast


class ConfigObject():

    def __init__(self, configfile='./AnalysisConfig_default.ini'):
        self.config = configparser.ConfigParser()
        self.config.optionxform = lambda option: option  # preserves case in key names
        self.config.read(configfile, encoding='utf-8-sig')

        self.get_section_PosCalibration()
        self.get_section_DataSelectionRules()
        self.get_section_TrajectorySelectionRules()
        self.get_section_Ph_Retrieval()
        self.get_section_SetupSpecifics()
        self.get_section_zCam()
        self.get_section_RegistrationCamToScan()

        if self.SetupSpecifics['is_simulation']:
            self.get_section_Simulation()
        print("Done!")

    def get_section_RegistrationCamToScan(self):
        section_title= 'RegistrationCamToScan'
        self.RegistrationCamToScan={}
        try:
            self.RegistrationCamToScan['Dest']          =self.str2array(self.config[section_title]['Dest'])
            self.RegistrationCamToScan['Traget']        =self.str2array(self.config[section_title]['Traget'])
            self.RegistrationCamToScan['RotAngle']      =self.str2array(self.config[section_title]['RotAngle'])
            self.RegistrationCamToScan['Trans']         =self.str2array(self.config[section_title]['Trans'])
        except:
            self.RegistrationCamToScan['Dest']          =np.array([[0,0],[0,0],[0,0]])
            self.RegistrationCamToScan['Traget']        =np.array([[0,0],[0,0],[0,0]])
            self.RegistrationCamToScan['RotAngle']      =0.0
            self.RegistrationCamToScan['Trans']         =np.array([0,0])
            
    def get_section_PosCalibration(self):
        section_title= 'PosCalibration'
        self.PosCalibration={}
        self.PosCalibration['Max']            =self.str2array(self.config[section_title]['Max'])
        self.PosCalibration['Offset']         =self.str2array(self.config[section_title]['Offset'])
        self.PosCalibration['ScaleFactor']    =self.str2array(self.config[section_title]['ScaleFactor'])
        self.PosCalibration['PositionLength'] =self.str2array(self.config[section_title]['PositionLength'])
        self.PosCalibration['APDShift']       =np.double(self.str2array(self.config[section_title]['APDShift']))
        self.PosCalibration['EODoffset']      =self.str2array(self.config[section_title]['EODoffset'])*(-1)
        self.PosCalibration['EODScale']       =self.str2array(self.config[section_title]['EODScale'])
        self.PosCalibration['EODPiezofilter'] =np.double(self.str2array(self.config[section_title]['EODPiezofilter']))
        self.PosCalibration['EODPiezofilterScaling']   =self.topower(self.str2array(self.config[section_title]['EODPiezofilterScaling']))
        self.PosCalibration['EODUnitconversion']       =self.EODUnitconversion(self.str2array(self.config[section_title]['EODUnitconversion']))
        self.PosCalibration['SLMSwitchHight']   =self.str2array(self.config[section_title]['SLMSwitchHight'])
        try:
            self.PosCalibration['zLayerOffset']        =self.str2array(self.config[section_title]['zLayerOffset'])
        except:
            print('SLMSwitch is no longer in use, please rename it to zLayerOffset in the ini file')
            self.PosCalibration['zLayerOffset']        =self.str2array(self.config[section_title]['SLMSwitch'])
        try:
            self.PosCalibration['LaserShotPattern']        =self.str2array(self.config[section_title]['LaserShotPattern'])
        except:
            
            self.PosCalibration['LaserShotPattern']        =1.0

        self.PosCalibration['ScanMemsize']      =self.str2array(self.config[section_title]['ScanMemsize'])
        self.PosCalibration['TrackingMemsize']  =self.str2array(self.config[section_title]['TrackingMemsize'])
        self.PosCalibration['ScanSpeed']        =self.str2array(self.config[section_title]['ScanSpeed'])
        self.PosCalibration['ScanSpeedTracking']=self.PeriodToSpeed(self.str2array(self.config[section_title]['ScanPeriodTracking']))
        self.PosCalibration['LaserOnTime']      =self.str2array(self.config[section_title]['LaserOnTime'])
        self.PosCalibration['MeasurementOnTime']=self.str2array(self.config[section_title]['MeasurementOnTime'])
        self.PosCalibration['zScaleUpp']        =self.str2array(self.config[section_title]['zScaleUpp'])
        self.PosCalibration['zThresholdHigh']   =self.str2array(self.config[section_title]['zThresholdHigh'])
        self.PosCalibration['zThresholdLow']    =self.str2array(self.config[section_title]['zThresholdLow'])
        self.PosCalibration['zLevels']          =self.str2array(self.config[section_title]['zLevels'])
        self.PosCalibration['zOffsetIndex']     =self.str2array(self.config[section_title]['zOffsetIndex'])
        self.PosCalibration['run_wndw']         =self.str2array(self.config[section_title]['run_wndw'])
        self.PosCalibration['run_cycles']       =self.str2array(self.config[section_title]['run_cycles'])
        self.PosCalibration['skip_cycles']      =self.str2array(self.config[section_title]['skip_cycles'])
        self.PosCalibration['zScale']           =self.str2array(self.config[section_title]['zScale'])
        try:
            self.PosCalibration['TrackingMemSlots1']        =self.str2array(self.config[section_title]['TrackingMemSlots1'])
        except:
            self.PosCalibration['TrackingMemSlots1']        =[0,200]
        try:
            self.PosCalibration['TrackingMemSlots2']        =self.str2array(self.config[section_title]['TrackingMemSlots2'])
        except:
            self.PosCalibration['TrackingMemSlots2']        =[0,200]

        
        
       
    def get_section_DataSelectionRules(self): 
        section_title= 'DataSelectionRules'
        self.DataSelectionRules={}
        self.DataSelectionRules['APDThreshold']             =self.str2array(self.config[section_title]['APDThreshold'])
        self.DataSelectionRules['APDCountRange']            =self.str2array(self.config[section_title]['APDCountRange']) 
        self.DataSelectionRules['APDCountRange_ErrorBound'] =self.str2array(self.config[section_title]['APDCountRange_ErrorBound']) 
        self.DataSelectionRules['xBounds'] =self.str2array(self.config[section_title]['xBounds']) 
        self.DataSelectionRules['yBounds'] =self.str2array(self.config[section_title]['yBounds']) 
        self.DataSelectionRules['TBounds'] =self.str2array(self.config[section_title]['TBounds']) 
        
    def get_section_TrajectorySelectionRules(self): 
        section_title= 'TrajectorySelectionRules'
        self.TrajectorySelectionRules={}
        self.TrajectorySelectionRules['PosStepDiviation']        =self.str2array(self.config[section_title]['PosStepDiviation'])
        self.TrajectorySelectionRules['CutDisplacment']          =self.str2array(self.config[section_title]['CutDisplacment'])
        self.TrajectorySelectionRules['MinTrajectoryTimeLength'] =self.str2array(self.config[section_title]['MinTrajectoryTimeLength'])
        self.TrajectorySelectionRules['BackgroundThreshold']     =self.str2array(self.config[section_title]['BackgroundThreshold'])
        self.TrajectorySelectionRules['CutTraj']                 =self.str2array(self.config[section_title]['CutTraj'])
        self.TrajectorySelectionRules['CutTrajWhenLaserHigh']    =self.str2array(self.config[section_title]['CutTrajWhenLaserHigh'])
        self.TrajectorySelectionRules['Mode']                    =self.config[section_title]['Mode']
        self.TrajectorySelectionRules['wndw']                    =self.str2array(self.config[section_title]['wndw'])
        self.TrajectorySelectionRules['Sigma']                   =self.str2array(self.config[section_title]['Sigma'])
        self.PosCalibration['beta']                              =self.str2array(self.config[section_title]['beta'])
        self.TrajectorySelectionRules['verbose']                 =self.str2array(self.config[section_title]['verbose'])
        self.TrajectorySelectionRules['Wprio']                   =self.str2array(self.config[section_title]['Wprio'])
        self.TrajectorySelectionRules['background']              =self.str2array(self.config[section_title]['background'])              
        self.TrajectorySelectionRules['TopHatoffset']            =self.str2array(self.config[section_title]['zoffset'])  
        self.TrajectorySelectionRules['SGxyz']                   =self.str2array(self.config[section_title]['SGxyz'])  
        self.TrajectorySelectionRules['STxyz']                   =self.str2array(self.config[section_title]['STxyz'])  
        try:
            self.TrajectorySelectionRules['TrimParam']           =self.str2array(self.config[section_title]['TrimParam'])
        except:
            self.TrajectorySelectionRules['TrimParam']=np.array([0.0,0.0,-1.0])
        try:
            self.TrajectorySelectionRules['CrossSection']        =self.str2array(self.config[section_title]['CrossSection'])
        except:
            self.TrajectorySelectionRules['CrossSection']=np.array([0.0,1000])
        try:
            self.TrajectorySelectionRules['MLETailFactor']        =self.str2array(self.config[section_title]['MLETailFactor'])
        except:
            self.TrajectorySelectionRules['MLETailFactor']=1.0
        try:
            self.TrajectorySelectionRules['CTailFactor']        =self.str2array(self.config[section_title]['CTailFactor'])
        except:
            self.TrajectorySelectionRules['CTailFactor']=[1,10000]
        try:
            self.TrajectorySelectionRules['nBackgroundBias']    =self.str2array(self.config[section_title]['nBackgroundBias'])
        except:
            self.TrajectorySelectionRules['nBackgroundBias']=0.0

    def get_section_SetupSpecifics(self):
        section_title= 'SetupSpecifics'
        self.SetupSpecifics={}
        self.SetupSpecifics['Lambda']=self.str2array(self.config[section_title]['Lambda'])
        self.SetupSpecifics['Px']    =self.str2array(self.config[section_title]['Px'])
        self.SetupSpecifics['TLens'] =self.str2array(self.config[section_title]['TLens'])
        self.SetupSpecifics['Na']    =self.str2array(self.config[section_title]['Na'])
        self.SetupSpecifics['n']     =self.str2array(self.config[section_title]['n'])
        self.SetupSpecifics['Mag']   =self.str2array(self.config[section_title]['Mag'])
        self.SetupSpecifics['Mag2']  =self.str2array(self.config[section_title]['Mag2'])
        self.SetupSpecifics['NM']    =self.str2array(self.config[section_title]['NM']).astype('int')
        self.SetupSpecifics['PholeDia']     =self.str2array(self.config[section_title]['PholeDia'])
        self.SetupSpecifics['flens']        =self.SetupSpecifics['TLens']/self.SetupSpecifics['Mag']   
        self.SetupSpecifics['is_simulation']=int(self.str2array(self.config[section_title]['is_simulation']))              
        
    def get_section_Ph_Retrieval(self):
        section_title= 'Ph_Retrieval'
        self.Ph_Retrieval={}
        self.Ph_Retrieval['zssize']         =self.str2array(self.config[section_title]['zssize'])              
        self.Ph_Retrieval['path_PSF']       =self.config[section_title]['path_PSF']              
        self.Ph_Retrieval['fname_PSF']      =self.config[section_title]['fname_PSF']             
        self.Ph_Retrieval['IntencityC']     =self.str2array(self.config[section_title]['IntencityC'])              
        self.Ph_Retrieval['XYSWScalefactor']=int(self.str2array(self.config[section_title]['XYSWScalefactor']))
        self.Ph_Retrieval['ZSWScalefactor'] =int(self.str2array(self.config[section_title]['ZSWScalefactor']))
        self.Ph_Retrieval['CropS']          =int(self.str2array(self.config[section_title]['CropS'])) 
        try:
            self.Ph_Retrieval['load_IntencityC']=int(self.str2array(self.config[section_title]['load_IntencityC']))
        except:
            self.Ph_Retrieval['load_IntencityC']=int(0)

    def get_section_Simulation(self):
        section_title= 'Simulation'
        self.Sim_par={}
        self.Sim_par['n_cross']         =self.str2array(self.config[section_title]['n_cross'])              
        self.Sim_par['points_to_use']   =int(self.str2array(self.config[section_title]['points_to_use']))              
        self.Sim_par['TrajType']        =self.config[section_title]['TrajType']              
        self.Sim_par['radius_xyz']      =self.str2array(self.config[section_title]['radius_xyz'])              
        self.Sim_par['offset_xyz']      =self.str2array(self.config[section_title]['offset_xyz'])              
        self.Sim_par['freq_xyz']        =self.str2array(self.config[section_title]['freq_xyz']) 
        self.Sim_par['Save_groundtruth']=int(self.str2array(self.config[section_title]['Save_groundtruth']))              
        self.Sim_par['PhotonStatistics']=self.config[section_title]['PhotonStatistics']              
        self.Sim_par['diff_coef']       =self.str2array(self.config[section_title]['diff_coef']) 
        self.Sim_par['background']      =self.str2array(self.config[section_title]['background'])              
        self.Sim_par['Bouders_xyz']     =self.str2array(self.config[section_title]['Bouders_xyz'])
        self.Sim_par['Npoints_xyz']     =self.str2array(self.config[section_title]['Npoints_xyz']).astype('int')
        try:
            self.Sim_par['PiezoFollow']     =self.str2array(self.config[section_title]['PiezoFollow'])              
        except:
            self.Sim_par['PiezoFollow']     = 0       
        try:
            self.Sim_par['Subsampling']     =self.str2array(self.config[section_title]['Subsampling'])    
        except:
            self.Sim_par['Subsampling']     = 1       
            


    def get_section_zCam(self):
        section_title= 'zCam'
        try: 
            self.zCam={}
            self.zCam['path_ref']          =self.config[section_title]['path_ref']   
            self.zCam['file_prefix_ref']   =self.config[section_title]['file_prefix_ref']   
            self.zCam['N']                 =int(self.str2array(self.config[section_title]['N']))              
            self.zCam['zssize']            =self.str2array(self.config[section_title]['zssize'])         
            self.zCam['Scaling']           =self.str2array(self.config[section_title]['Scaling'])              
            self.zCam['path_tracking']     =self.config[section_title]['path_tracking']              
            self.zCam['file_prefix']       =self.config[section_title]['file_prefix']   
            self.zCam['MemIndex']          =int(self.str2array(self.config[section_title]['MemIndex']))
            self.zCam['Cycles']            =int(self.str2array(self.config[section_title]['Cycles']))              
        except:
            pass
        
    def PeriodToSpeed(self,s):
        r=1/s
        return r
    
    def EODUnitconversion(self,s):
        """
        The parameter are explained using original parameter values:
        
         s[0     1    2   3  4       5     6     7  8  9]
          [10, 113, 114, 23, 2, -16384, -110, -120, 2,-8] 
         C_cam= mean(10./[113 114])= 0.088um/pix is the camera resolution, this is measured using a grid of 10um and counting the pixels for the grid spacing 
         D_cam= 23 pixels is the circle diameter traced out by the EOD "star", this is set in camera view as a target for the scale calbration S_EOD
         D_EOD= -16384*2 = -2**15  which the diameter of the cricule given in EOD values without scaling (the top point of the "star" in the FPGA memory is at (-16384=-2**14), this value is to be scaled down, it is done in this way to not loss resolution due to integer percision)
         S_EOD=(mean([-110 -120]))*2^-8 is the scaling of D_EOD, measured by running the tracking and looking at reflection with the camera and optimizing the parameters to fitt the circle D_cam
        
        Dimansion analysis: [um/pix]*[pix]/[EOD steps] = [um/step]
        """   
        C_cam=np.mean(s[0]/np.array([s[1], s[2]])) 
        D_cam= s[3]
        D_EOD=s[4]*s[5]
        S_EOD= np.mean(np.array([s[6], s[7]]))*np.power(s[8],float(s[9]))
        
        r=C_cam*D_cam/(D_EOD*S_EOD) #[um/step] 
        return r

    def topower(self,s):
        r=s[0]*2**s[1]
        return r
    
    def str2array(self, s):
        # Remove space after [
        s=re.sub('\[ +', '[', s.strip())
        # Replace commas and spaces
        s=re.sub('[,\s]+', ', ', s)
        # Handle inf by using a place holder that is a big number 2**32=4294967296
        s=re.sub('inf', '4294967296', s)

        #r=np.array(ast.literal_eval(s))  # literal_eval is a safe way to convert a string to a python data type, it will not evaluate commands.
        r=np.array(eval(s)) #To be able to use numpy in eval

        # Extchange the large place holder with propper inf
        r=np.where(r==2**32, np.inf, r)
        r=np.where(r==-2**32, -np.inf, r)
        
        #
        if r.size==1:
            r=np.double(r)

        return r

    def GetSection(self, section_title):
        Section = {}
        for key in self.config[section_title]:
            Section[key] = eval(self.config[section_title][key])
        
        return Section






##### Class test code ##################
def main():
    # Path prefix for local data storage
    configfile = "./AnalysisConfig_default.ini"
    config=ConfigObject(configfile)
    config.get_section_PosCalibration()    
    config.get_section_DataSelectionRules()   
    config.get_section_TrajectorySelectionRules()

if __name__ == '__main__':
    main()