import numpy as np
from scipy.optimize import minimize
from scipy.optimize import Bounds
from Physics_Lib import Particle as _particle
import matplotlib.pyplot as plt

class DiffusioMLE():

    def get_sigle_D_trajectory_sample(traj_settings):
        time_steps      =traj_settings['time_steps']
        delta_t         =traj_settings['delta_t']        
        diff_coef       =traj_settings['D_coef']      
        initial_pos_um  =traj_settings['initial_pos_um'] 

        time=delta_t*np.arange(time_steps)
    
        dxdydz = np.random.normal(0, np.sqrt(2*diff_coef*delta_t), (int(time_steps-1), 3))
        traj = np.cumsum(np.vstack((initial_pos_um, dxdydz)), axis=0)
        traj_txyz=np.vstack((time,traj.T))
        return traj_txyz

    def get_multi_D_trajectory_sample(traj_settings):
        """
        #Simulation track
        steps=1000
        substeps=1        #Simulate blur this will give R=1/6.2
        dt=0.84e-3        #[s]
        sigma_e= [40e-3,40e-3,40e-3] #[um] xyz

        D=[0.1,1]        # [um^2/s]
        Dsteps=[80,20]   # steps for each D
        Pos_init=[0,0,0] # Start position
        """
        steps   =traj_settings['time_steps']    
        substeps=traj_settings['sub_steps']     
        dt      =traj_settings['delta_t']       
        D       =traj_settings['D_coef']        
        Dsteps  =traj_settings['D_steps']       
        sigma_e =traj_settings['P_error']       
        Pos_init=traj_settings['initial_pos_um']

        #############################################################################
        # Create 3D Brownian motion 
        k=np.arange(steps)
        DTrue=np.array([])
        dxdydz=np.array([Pos_init])
        for i in range(np.ceil(steps/(np.sum(Dsteps))).astype('int')):
            for j in range(len(D)):
                dxdydz1 = np.random.normal(0, np.sqrt(2*D[j]*(dt/substeps)), (int(Dsteps[j]*substeps), 3))
                dxdydz=np.concatenate((dxdydz,dxdydz1))
                DTrue=np.concatenate((DTrue,np.ones(Dsteps[j])*D[j]))
            
        trajT = np.cumsum(dxdydz, axis=0)
        traj=np.zeros((steps,3))
        for i in range(3):
            if substeps>1:
                traj[:,i]= np.convolve(np.ones(substeps)/substeps,trajT[:,i])[substeps-1:-substeps+1:substeps]
            else:
                traj[:,i]=trajT[:-1,i]
    
        
        for i in range(3):
            traj[:,i]=traj[:,i] + np.random.normal(0, sigma_e[i], steps)      #[um]
        T=k*dt 
        
        txyzn_Map=np.vstack((T,traj.T,DTrue))

        TrajDictionary={}
        TrajToGet=0
        TrajDictionary[TrajToGet]=txyzn_Map
        return TrajDictionary

    def get_covariance_matrix(D=1,S=0.05,n=4,j=[1,2],Dt=0.84e-3,R=1/12):
        """
        Creates the covariant matrix
        D= diffusion constant [um^2/s]
        S= position  error [um]
        Dt= time step size [s]
        R= blur coefficient 
        n= matix size will out put an nxn matrix
        j= step size between position elements, accumilate matrices for j[0] to j[1] 

        """
        CovMat=np.zeros((n,n,j[1]-j[0]))
        CovMat_inv=np.zeros((n,n,j[1]-j[0]))
        CovMat_det=np.zeros(j[1]-j[0])
        for jj in range(j[0],j[1]):
            for nn in range(n):
                if nn==0:
                    M=(2*D*(jj-2*R)*Dt+2*S**2)/2
                elif nn<jj:
                    M=2*D*(jj-nn)*Dt
                elif nn==jj:
                    M=2*D*R*Dt-S**2
                elif nn>jj:
                    M=0
                CovMat[:,:,jj-j[0]]+=np.diag(np.diag(CovMat[:,:,jj-j[0]],nn)+M,nn)
            CovMat[:,:,jj-j[0]]+=CovMat[:,:,jj-j[0]].T
        
        return CovMat

    def get_log_likelihood(X,MLE_settings):
        """
        Log likelihood function to be minimized
        """
        D=np.abs(X[0])
        S=X[1]
        n=MLE_settings['n']
        R=MLE_settings['R']
        Dt=MLE_settings['delta_t']
        xyz=MLE_settings['xyz']

        CovMat = DiffusioMLE.get_covariance_matrix(D=D,S=S,n=n,j=[1,n+1],Dt=Dt,R=R)
        interval_run= range(n)
        for nn in interval_run:
            CM=CovMat[:n-nn,:n-nn,nn]
            CM_inv=np.linalg.inv(CM)

            A=(3/2)*np.log(2*np.pi*np.linalg.det(CM))
            Dxyz=xyz[:,(nn+1):]-xyz[:,:-(nn+1)]
            F1=np.matmul(Dxyz,CM_inv)
            B=(1/2)*np.sum(np.diag(np.matmul(F1,Dxyz.T)))
            if nn==interval_run[0]:
                res=A+B
            else:
                res+=A+B

        return res
        
        

def Main():
    
    traj_settings={}
    traj_settings['time_steps']     =1000 
    traj_settings['sub_steps']      =1 
    traj_settings['delta_t']        =840e-6 #s
    traj_settings['D_coef']         =[1, 1] #um^2/s per dimension
    traj_settings['D_steps']        =[80, 20] #um^2/s per dimension
    traj_settings['P_error']        =[0e-3,0e-3,0e-3] #[um] xyz
    traj_settings['initial_pos_um'] =[0,0,0]

    TrajDictionary=DiffusioMLE.get_multi_D_trajectory_sample(traj_settings)

    # Analysis 
    Dest=()
    for k in range(1000): #range(np.floor(DMLE.traj_txyz[1,:].shape[0]/n).astype('int')):
        MLE_settings={}
        MLE_settings['n']       =6
        MLE_settings['R']       =0
        MLE_settings['delta_t'] =840e-6 #s
        MLE_settings['xyz']     =TrajDictionary[0][1:-1,k:k+MLE_settings['n']+1]        
        X=[1,0]#[D, P_error]
        bounds = Bounds([0, 50], [0, 100e-3])
        #res = minimize(DiffusioMLE.get_log_likelihood, X, args=MLE_settings, method='SLSQP', bounds=bounds)
        res = minimize(DiffusioMLE.get_log_likelihood, X, args=MLE_settings, method='Nelder-Mead')
        print(res['x'])




    print('asdf')

if __name__ == '__main__':
    Main()