import sys
from ConfigObject import ConfigObject # for .ini files
from TrajObject import TrajObject # Object to find trajectory data according to parameters from config file
from Physics_Lib import physics_Lib as _Ph
from DataFileIO_Lib import DataFileIO_Lib as _DS
from zCamHandler import ZCamHandler as _zCam
from miscellaneous import miscellaneous as _MS

import numpy as np
from scipy.signal import lfilter
import FilterLibrary

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib.gridspec import GridSpec

from PIL import Image
from skimage import transform as _ski_tran
from skimage.feature import (blob_log, match_template, ORB, plot_matches, match_descriptors) 
from skimage.measure import ransac
import cv2 as _cv2


class DataHandler():

    def __init__(self):
        # Create hanlder for libraies used in
        self._DS=_DS()
        self._Ph=_Ph()
    
    def GetScanData(self, ScanProject):
        """
            Opens files and returns the content in the right format
        """
        Data=self._DS.GetScanData(ScanProject)
        return Data
    
    def PickleData(self, ScanProject, override=0,folder='',FileEnd='_ScanProject.pkl'):    
        self._DS.PickleData(ScanProject,override=0,folder=folder,FileEnd=FileEnd)    

    def UnPickleData(self, ScanProject=None, folder_='', file_='',loadCustom=0):
        LoadFlag=0
        if ScanProject!=None:
            LoadFlag=ScanProject.loadSaveData

        if (LoadFlag + loadCustom)>0:
            ScanProject_res=self._DS.UnPickleData(ScanProject=ScanProject, folder=folder_, file=file_)    
            return ScanProject_res
        else:
            print('Load data flag is not sett!')

    def savetoMat(self, ScanProject):
        self._DS.savetoMat(ScanProject)

    def CheckDataIntegrity(self, RawData):
        """
        Looks through the Raw data to see if the over all pattern between the 3 data sets are reasonable. The data is formatted according to:
        _____________________________________
        % Extract data from 64 bit number coming from the bit file.
       This can be visualized as  
       F= Flag;  63-48        47-32              31-16           15-0
       |FFF001| APD_1   ||MemInd|APD_2    ||  x-PIEZO_RAW   ||  y-PIEZO_RAW   |  
        xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx
       |0000000000000000||     x-EOD      ||      y-EOD     ||0000000000000000|
        xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx
       |0000100000000000||     x-Piezo    ||      y-Piezo   ||     z-signal   |
        xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx
            ab
        "ab" indicate the first and second possition that is of intrest, the flags from the three numbers are suppose to follow the pattern 
        For a. 001 (FFT)
        For b. 100 (TFF)
        We are currently only using b, as test 

        INPUT: Vector with 64 bit numbers, following the pattern above
        OUTPUT: 
        """
        
        PosErrorFlag = np.array(((RawData >> 48) & 2**10)!=0) # TFF TFF ... Pattern we are after between the numbers
        PosErrorFlag = PosErrorFlag.astype('int')               # convert TFF to 100
        WhereTrueFlagsAre = np.where(PosErrorFlag==1)[0]
        
        # Check for error at the beginning of the data file.  
        # If the first value is not True then find the first True and supress everything before.
        # If the distance to the first True value is larger than 4 then something bad has happend to the data and there might be parts that is corrupted
        if PosErrorFlag[0] != 1:
            FirstOne = WhereTrueFlagsAre[0]
            PosErrorFlag[:FirstOne-1] = -128      # Displace these values
            if FirstOne > 4:
                print('Error in data pattern, data sets are not aligned')
        
        # Checking the end of the data file, by displacing the last data set of (100) to guarantee that nothing has been missed
        LastTrueFlag = WhereTrueFlagsAre[-1]
        #print(PosErrorFlag[LastTrueFlag:])
        PosErrorFlag[LastTrueFlag:] = -128 
        
        ValidPosErrorFlag = np.where(PosErrorFlag>-1)[0]
        PosErrorFlag_Valid = PosErrorFlag[ValidPosErrorFlag]
        
        # Check if the data set is divisible by 3
        if np.remainder(len(PosErrorFlag_Valid),3) != 0:
            print('Error in data pattern, data sets after correcting start values are not aligned')
             
        # Check if the data set follows the 100 trend
        pattern = [1, 0, 0]
        PosErrorFlag_Reshaped = PosErrorFlag_Valid.reshape(int(len(PosErrorFlag[ValidPosErrorFlag])/3),3)
        for i in range(3):
            if np.where(PosErrorFlag_Reshaped[:,i] != pattern[i])[0].size != 0:
                print(['Error in data pattern "1 0 0", data sets is not aligned'])
            
        return PosErrorFlag

    def AddFalseFlagToFront(self,FirstPostions):
        # Adds a False to the front of an numpy 1D array
        FirstPostionsOnePuch = np.insert(FirstPostions,0, False)[:-1]
        return FirstPostionsOnePuch
        
    def UnpackRawData(self, ScanProject ,RawData, PosErrorFlag):        
        """
        _____________________________________
        % Extract data from 64 bit number coming from the bit file.
        Data content is divide in three 64 bit numbers. When describing the content we will read the number from right to left
        exampel:  43 in binary is 101011= 1*2^5 + 0*2^4 + 1*2^3 + 0*2^2 + 1*2^1 + 1*2^0 thus bit 0 is furthes to the right
        The three numbers are:
        1:  
            bit(0-15):  y piezo values sampeled in the 200Mhz loop, these are raw closed loop values from the piezo unit. These are sampled usually 15us (5*3000) after that the FPGA has requested a piezo move 
            bit(16-31): x piezo values sampeled in the 200Mhz loop, these are raw closed loop values from the piezo unit. These are sampled usually 15us (5*3000) after that the FPGA has requested a piezo move 
            bit(32-41): APD signal Obs since it is 10 bits it can only holed up to 1023 photon counts
            bit(42-47): FPGA Memory index indicating which memory slot is used both for scaning and tracking, OBS this can only hold index 0 to 63 thus out side this range will not be included (ex. scaning usally uses index 0-199)
            bit(48-57): APD signal Obs since it is 10 bits it can only holed up to 1023 photon counts
            bit(58-63): Flags: 
                               58:  1  Data identifier  
                               59:  0
                               60:  0
                               61:  1/0 Release threshold (0 not reached / 1 reached)
                               62:  1/0 Laser Switch during tracking (0 normal laser power level, 1 the other laser power level)
                               63:  1/0 Tracking (0 trackin off / 1 tracking on)
        2:
            bit(0-15):  0  
            bit(16-31): y signal,  Scanning: 0, Tracking: current EOD center value  + full PID signal.  
            bit(32-47): x signal,  Scanning: 0, Tracking: current EOD center value  + full PID signal.  
            bit(48-63): 0  
        3. 
            bit(0-15):  z signal,  Scanning: 0, Tracking: current center value  + FPGA PID signal low frequency part. OBS! this signal is operating scaled in the FPGA (usally by 1000)  
            bit(16-31): y signal,  Scanning: 0, Tracking: PID signal (same as 2-y-signal) after low frequency filter + current PID Piezo center value.  
            bit(32-47): x signal,  Scanning: 0, Tracking: PID signal (same as 2-x-signal) after low frequency filter + current PID Piezo center value.  
            bit(48-63): 0000 1000 0000 0000    The 1 is in place 59        

       This can be visualized as  
       F= Flag
       |FFF001| APD_1   ||MemInd|APD_2    ||  x-PIEZO_RAW   ||  y-PIEZO_RAW   |  
        xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx
       |0000000000000000||     x-EOD      ||      y-EOD     ||0000000000000000|
        xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx
       |0000100000000000||     x-Piezo    ||      y-Piezo   ||     z-signal   |
        xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx
        
        """   
        # Extract each data set
        FirstPostions = PosErrorFlag == 1
        
        Data = RawData[FirstPostions]       # First output from FPGA:      |FFF001| APD_1   ||MemInd|APD_2    ||  x-PIEZO_RAW   ||  y-PIEZO_RAW   |  
                                                                         #  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx
        SecondPostions = self.AddFalseFlagToFront(FirstPostions)
        EODData = RawData[SecondPostions]   # Second output from FPGA:     |0000000000000000||     x-EOD      ||      y-EOD     ||0000000000000000| 
                                                                         #  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx
        ThirdPostions = self.AddFalseFlagToFront(SecondPostions)
        PIEZOData = RawData[ThirdPostions]  # Third output from FPGA       |0000100000000000||     x-Piezo    ||      y-Piezo   ||     z-signal   |
                                                                         #  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx  xxxxxxxxxxxxxxxx
                                                                                    
        # Check that all data is of the same length
        #DFlag = (len(EODData) != len(Data) & len(PIEZOData) != len(Data) & len(EODData) !=len (PIEZOData));
        if (len(EODData) != len(Data) & len(PIEZOData) != len(Data) & len(EODData) != len(PIEZOData)):
            print('Data Loading error, data are not of the same size')
        
        ##############################################################################################################################            
        ###########   Unpacking data from the data set
        ##############################################################################################################################
        
        ############# Unpacking first data set 
        # Data flag 1
        ySG = np.array([Data & 2**16-1])                                # Y piezo position data direct from FPGA Sampling, Used when scanning   
        xSG = np.array([(Data >> 16) & 2**16-1])                        # X piezo position data direct from FPGA Sampling, Used when scanning   
        APD_2 = np.array([(Data >> 32) & 2**10-1])                      # V polarization  
        Memindex = np.array([(Data >> 42) & 2**6-1])                    # Tracking Memory index
        APD_1 = np.array([(Data >> 48) & 2**10-1])                      # H polarization  
        
        ##################################
        ##### Check ySG - APD_1
        ##################################
        # Error data flag 0
        TrackingFlag = np.array([((Data >> 48) & 2**15) != 0])                          # Tracking on
        TrackingFlag = np.append(TrackingFlag,[((Data >> 48) & 2**13) != 0], axis=0)    # Release Threshold not reached 
        #TrackingFlag = np.append(TrackingFlag,[((Data >> 48) & 2**10) != 0], axis=0)   # REMOVE and use "Data_integrity" Is data coming from position data set? Supposed to be 1
        TrackingFlag = np.append(TrackingFlag,[((Data >> 48) & 2**14) != 0], axis=0)    # Laser Switch during tracking (OR if FPGA after 2021-12-28, addaprive tacking state switch)
        
        Data_integrity= np.array([((Data >> 48) & 2**11) != 0])                         # Is data coming from position data set? Supposed to be 0
        Data_integrity= np.append(Data_integrity,[((Data >> 48) & 2**10) != 0], axis=0) # Is data coming from position data set? Supposed to be 1

        ############## Unpacking second data set, Data used when tracking 
        zEODSG = np.array([EODData & 2**16-1]).astype(np.int16)                   # Supposed to be 0
        yEODSG = np.array([(EODData >> 16) & 2**16-1]).astype(np.int16)           # Y EOD position data
        xEODSG = np.array([(EODData >> 32) & 2**16-1]).astype(np.int16)           # X EOD position data
        TrackingFlagEODData = [((EODData >> 48) & 2**10) != 0]                    # Is data coming from EODData data set? Supposed to be 0

        Data_integrityEOD= np.array([((EODData >> 48) & 2**11) != 0])                             # Is data coming from position data set? Supposed to be 0
        Data_integrityEOD= np.append(Data_integrityEOD,[((EODData >> 48) & 2**10) != 0], axis=0)  # Is data coming from position data set? Supposed to be 0

        
        ############## Unpacking third data set, Data used when tracking
        zSLMIndexSG = np.array([PIEZOData & 2**16-1]).astype(np.int16)            # Z FPGA z error signal which is related to the SLM hologram index 
        yPIEZOSG = np.array([(PIEZOData >> 16) & 2**16-1]).astype(np.int16)       # Y piezo position data
        xPIEZOSG = np.array([(PIEZOData >> 32) & 2**16-1]).astype(np.int16)       # X piezo position data
        TrackingFlagPIEZOData = [((PIEZOData >> 48) & 2**10) != 0]                # Is data coming from PIEZOData data set? Supposed to be 0    

        Data_integrityPIEZO= np.array([((PIEZOData >> 48) & 2**11) != 0])                               # Is data coming from position data set? Supposed to be 1
        Data_integrityPIEZO= np.append(Data_integrityPIEZO,[((PIEZOData >> 48) & 2**10) != 0], axis=0)  # Is data coming from position data set? Supposed to be 0


        # Test the tracking flag integrity,
        # a. Combine all relevant tracking data
        #   0. Tracking on
        #   1. Release Threshold not reached 
        #   2. Laser power Switch during tracking (OR if FPGA after 2021-12-28, addaprive tacking state switch)
        #   3. Memory index obs! not complete during scanning (using 6 bits 0-63, but the scanning is 0-199), only relaible for the tracking data  
        # b. Test if memory index slip has occured 
        # c. Correct for the events
        TrackingFlags=TrackingFlag
        TrackingFlags = np.int16(np.append(TrackingFlags, Memindex, axis=0))

        if 0: # Debug look at the memory index pattern, and the photon counts
            index=200
            intervall=np.arange(500000-14,500000-14+27*index)
            n=np.sum(np.reshape((APD_1+APD_2)[0,intervall],(index,27)),axis=0)
            Mem=np.sum(np.reshape(Memindex[0,intervall],(index,27)),axis=0)/index
            
            #plt.plot(n/10,'-*b');plt.plot(Mem,'-*g');
            #plt.grid();plt.show()

            nr=np.roll(n,0)
            NP=np.zeros((3,3,3))
            NP[1,1,1]=nr[0]
            NP[1,0,1]=nr[1]
            NP[0,0,1]=nr[2]
            NP[0,1,1]=nr[3]
            NP[0,2,1]=nr[4]
            NP[1,2,1]=nr[5]
            NP[2,2,1]=nr[6]
            NP[2,1,1]=nr[7]
            NP[2,0,1]=nr[8]

            NP[1,1,0]=nr[9]
            NP[1,0,0]=nr[10]
            NP[0,0,0]=nr[11]
            NP[0,1,0]=nr[12]
            NP[0,2,0]=nr[13]
            NP[1,2,0]=nr[14]
            NP[2,2,0]=nr[15]
            NP[2,1,0]=nr[16]
            NP[2,0,0]=nr[17]

            NP[1,1,2]=nr[18]
            NP[1,0,2]=nr[19]
            NP[0,0,2]=nr[20]
            NP[0,1,2]=nr[21]
            NP[0,2,2]=nr[22]
            NP[1,2,2]=nr[23]
            NP[2,2,2]=nr[24]
            NP[2,1,2]=nr[25]
            NP[2,0,2]=nr[26]
            fig, ax = plt.subplots(1, 3)
            for i in range(3):
                ax[i].imshow(NP[:,:,i]);
            plt.show()

        _, MemIndexScaled = self.GetIndexPosTracking(ScanProject,TrackingFlags, verbose=True)

        TrackingFlags[0,:]=np.int16(MemIndexScaled>=0) 


        ##############################################################################
        # Relable data for clarity and simplicity
        ##############################################################################
        # Photon counts 
        PhotonAPDCounts = APD_1    
        PhotonAPDCounts = np.append(PhotonAPDCounts, APD_2, axis=0)
        
        # Piezo signals sampled in the 200Mhz loop 
        PiezoSampled = xSG
        PiezoSampled = np.append(PiezoSampled, ySG, axis=0)
        
        # Tracking flags:
        # 0. Tracking on
        # 1. Release Threshold not reached 
        # 2. Laser power Switch during tracking
        # 3. Memory index obs! not complete during scanning (using 6 bits 0-63, but the scanning is 0-199), only relaible for the tracking data  
        TrackingFlags=TrackingFlags
        
        # Signal that in the FPGA will be send to the EOD (high frequency signal from the PID estimation)
        # Also known as the EOD center position estimate pluse the newest PID value 
        # Obs!! The EODSignal is the total EOD movement thus it is both the current center plus the total PID error (no filter).
        # Scanning: 0, Tracking: current EOD center value  + full FPGA PID signal.  
        EODSignal = xEODSG
        EODSignal = np.append(EODSignal, yEODSG, axis=0)
        EODSignal = np.append(EODSignal, zEODSG, axis=0)# Not in use, is 0 currently 
        
        # Signal that in the FPGA will be send to the Piezo stage (low frequency signal from the PID estimation)
        # For x an y it is also known as the piezo center position estimate. 
        # For z it is the signal that is converted to an index that moves between levels using EOD and the hologram staircase  
        # x,y Scanning: 0, Tracking: PID signal after low frequency filter + current PID Piezo center value.  
        # z   Scanning: 0, Tracking: current center value    
        # OLD remove:  z   Scanning: 0, Tracking: current center value  + FPGA PID signal low frequency part. OBS! this signal is operating scaled in the FPGA (usally by 1000)  
        
        PiezoSignal = xPIEZOSG
        PiezoSignal = np.append(PiezoSignal, yPIEZOSG, axis=0)
        PiezoSignal = np.append(PiezoSignal, zSLMIndexSG, axis=0)
        
        return PhotonAPDCounts, PiezoSampled, TrackingFlags, EODSignal, PiezoSignal
        
    def ConvertPiezoToMicroMeters(self,ScanProject, PiezoRawData):
        """
        This function only takes data comming from the FPGA that corespondes to the piezo tip/tilt mirror

        Converting Piezo signal of I16 to micrometers, we use S*L(x-O)/(M-O) where: 
        x: data, sampled from the piezo unit in the 200Mhz loop (current value) or estimated value from the PID centroid calculation (next value)
        O: Offest 
        M: Max value
        L: Length Range in micrometers corresponding to the pizo value range M-O
        S: scale factor if needed (Is usally 1)

        Output: converted data from I16 to [um]
        """
        s=ScanProject.PosCalibration['ScaleFactor']
        L=ScanProject.PosCalibration['PositionLength']*1.0
        O=ScanProject.PosCalibration['Offset']
        M=ScanProject.PosCalibration['Max']

        numerator = np.array(s).reshape(2,1) * L * (PiezoRawData-np.array(O).reshape(2,1))
        denominator = (np.array(M).reshape(2,1) - np.array(O).reshape(2,1))
        
        PiezoSampledInMicroMeters = numerator/denominator
        return PiezoSampledInMicroMeters
   
    def ConvertZFPGAToSLMIndex(self, zData):
        """
        Convert the FPGA z error signal to SLM index 
        OBS! Need to move the coefficients to the initialization part
        """
        zScaleUpp    = self.PosCalibration['zScaleUpp']
        zThresholdHigh = self.PosCalibration['zThresholdHigh']/zScaleUpp
        zThresholdLow = self.PosCalibration['zThresholdLow']/zScaleUpp

        zLevels      = self.PosCalibration['zLevels']
        zOffsetIndex = self.PosCalibration['zOffsetIndex']
        
        zData = zOffsetIndex + (zData/zScaleUpp) # this is the error offset to get a positive value from 0 to levle-1, normaly it is:  D = 2 + E/1024
        
        # First step to get the index used in the FPGA is to get the center index which is calcuated as:
        # 1. take the integer value of zData 
        # 2. get the reminder left 
        # 3. to index add 0 if reminder is below the threshold else add 1
        zIndex_int_part = np.floor(zData)
        zIndex_remander = zData - zIndex_int_part
        zIndex = zIndex_int_part + (zIndex_remander > zThresholdHigh)


        ######### DeBugging    
        debug=False
        if debug:
            import matplotlib.pyplot as plt
            plt.plot(zData,'b')
            plt.plot(zIndex_int_part+zThresholdHigh,'r')
            plt.plot(zIndex_int_part+zThresholdLow,'g')
            plt.plot(zIndex_int_part,'b')
            #plt.plot(current_level,'m')
            plt.plot(b,'m')
            plt.grid()
            plt.show()
        ##################################################


        zIndex=np.array([zIndex])

        return zIndex
    
    def ConvertData(self,ScanProject, Data):
        """
        This function only handels conversion of I16 to um originating from the piezo til/tilt mirror
        other data like the z index is flushed through
        Handles the conversion of the raw data from the Piezo and z FPGA error signal
        There are two situations: 
        1. one when the input is x,y (which is the Second output from FPGA)
        2. the other is when the input has x,y,z   (which is the Third output from FPGA)

        Input: Data 2xN or 3xN matrix including in the x,y I16 values from the FPGA sampling of the piezo tip/tilt mirror 
        output: Converted data which is a 2xN or 3xN matrix depending on the input, the piezo tip/tilt mirror is now in [um] 
        """
        if Data.shape[0] == 3:
            zData = Data[2,:]
            Data = Data[0:2,:]
            
            zIndex = np.array([zData])
            ConvertedData = self.ConvertPiezoToMicroMeters(ScanProject, Data)
            ConvertedData = np.append(ConvertedData,zIndex, axis=0)
            
        elif Data.shape[0] == 2:
            ConvertedData = self.ConvertPiezoToMicroMeters(ScanProject, Data)
            
        else: 
            print("There is something wrong with the matrix size, expected dimension 2 (x,y) or 3 (x,y,z) but got something else")
        
        return ConvertedData
       
    def GetScanTime(self, ScanProject, TrackingFlags):
        """
        1. The time increment is changing depending if the system is running in scanning mode or tracking mode
           If Tracking is ON then use tracking speed, else, use Scan speed

        2. The time evolusion during tracking depends on when we are measuring: 
            a. The system starts tracking and runs #"run_cycles" where one cycle is defined by "run_wndw" event (in fast tracking this is usaly 12 or 24 but this is related to the running window used during the meaurement)
            b. When a. is done then the system is still for #"skip_cycles" where one cycle is defined by "skip_wndw" ("skip_wndw" is usaly 6 but it dependes on the tracking patterns)
        """     
        TrackingONFlag    = TrackingFlags[0,:].copy()
        Memindex          = TrackingFlags[3,:].copy()
        ScanSpeedTracking = ScanProject.PosCalibration['ScanSpeedTracking']
        ScanSpeed         = ScanProject.PosCalibration['ScanSpeed']
        run_wndw          = np.int(ScanProject.PosCalibration['run_wndw'])
        skip_wndw         = np.int(ScanProject.PosCalibration['TrackingMemsize'])
        run_cycles        = np.int(ScanProject.PosCalibration['run_cycles'])  
        skip_cycles       = np.int(ScanProject.PosCalibration['skip_cycles']) 
        
        ### During tracking we add after each #run iterations a dealy that is #skip amount 
        if (run_cycles!=0) & (skip_cycles!=0):
            run=run_wndw*run_cycles
            skip=skip_wndw*skip_cycles
            TrackingStart = np.where(np.diff(TrackingONFlag)==1)[0]+1
            TrackingStop = np.where(np.diff(TrackingONFlag)==-1)[0]
            DTracking=TrackingStop-TrackingStart            
            NTracking=np.int64(np.floor((DTracking+1)/run)) #Amount of run cycles available for each tracking event

            for m in range(NTracking.size):
                for i in range(NTracking[m]):
                    if TrackingONFlag[TrackingStart[m]+(i+1)*run]==1:
                        # Move along the tracking trajectory in steps of run 
                        # and inserting skip steps, these skips are supposed to coinside with memory index 0 
                        # and supposed to be handled under "GetIndexPosTracking"
                        TrackingONFlag[TrackingStart[m]+(i+1)*run]=skip
                        if Memindex[TrackingStart[m]+(i+1)*run]!=0:
                            print('OBS! Skip steps is not alighed to memindex')
        
        
        SkipIndexVector=TrackingONFlag>1 # Vector that indicates where skip where inserted
        # Add the different speeds for tracking and scaning 
        # And last make a cumelative sum to get each time point       
        ScanSpeedIncrement = TrackingONFlag/ScanSpeedTracking + np.invert(TrackingONFlag.astype('bool'))/ScanSpeed
        ScanTime = np.cumsum(ScanSpeedIncrement)
        
        return ScanTime, SkipIndexVector
        
    def GetEODOffsetValueInMicroMeters(self, ScanProject):
        """
        Function that converts the EOD offsett values to [um], this is applied to the star pattern 

        Input: takes values from the ini file for position calibration
        output: 
        """
        # Inputs
        zLayerOffset= ScanProject.PosCalibration['zLayerOffset'] #used as template of indexis for scale factors
        xy_Offset_I16=ScanProject.PosCalibration['EODoffset']
        xy_scalevalues=np.array(ScanProject.PosCalibration['EODScale'])
        I16toum=ScanProject.PosCalibration['EODUnitconversion']
        MemSize=ScanProject.PosCalibration['TrackingMemsize']
        ####

        #IndexToExtract = ((zLayerOffset == np.array([0]))+1) #Convert the SLM +-1 and 0 to the index used when extracting scale factors in PosCalibration['EODScale']
        IndexToExtract = (zLayerOffset>=-100)*1 #Convert the SLM +-1 and 0 to the index used when extracting scale factors in PosCalibration['EODScale']

        #The conversion is as follows:
        # O*2^B *A*C 
        # O: the offset values in the star pattern in the FPGA memory
        # A: Multiplicative scale in the FPGA that depends on the memory index (related to the index in the SLM vector)
        # B: exponetial scale 
        # C: convertion unit from FPGA output value to [um]
        
        OffsetValue = xy_Offset_I16*2**xy_scalevalues[:,0].reshape(2,1)* \
                      xy_scalevalues[:,IndexToExtract]*I16toum  #[um] 
        
        R=np.sqrt(np.sum(OffsetValue**2,axis=0))[0:int(MemSize)]
        print("Tracking pattern radius: " + str(R))
       
        return OffsetValue 
        
    def RemoveShortTraj(self,MemIndexScaled, MaxmMemIndexScaledAllowed):
        """
        Function used to remove very short trajectores by puting there index 
        values as negative numbers in the variable: MemIndexScaled
        This function is used in the function: GetIndexPosTracking
        """
        StartPos = np.where(np.diff(np.int32(MemIndexScaled>=0))==1)[0]+1
        EndPos = np.where(np.diff(np.int32(MemIndexScaled>=0))==-1)[0]
        StartPos=StartPos[:len(EndPos)] #Trim down Start to not include last traking if tracking was stopped with out an end 
        Toshort=np.where((EndPos-StartPos)<2*(MaxmMemIndexScaledAllowed+1))[0]
        if Toshort.size!=0:
            print("Found " + str(Toshort.size) + " super short trajectories, these are removed")
            for i in Toshort:
                MemIndexScaled[StartPos[i]:EndPos[i]+1]=-MemIndexScaled[StartPos[i]:EndPos[i]+1]-1 # -1 to remove zeros if any 

        return MemIndexScaled
        
    def GetIndexPosTracking(self, ScanProject, TrackingFlags, verbose=True):   
        """
            From the tracking flag we extract:
            1. MemIndexScaled which gives the memory index with tracking events as positives (>=0) and scanning events as negatives (<0).
               OBS! this vector is corrected for events that hade a tracking ON flag but the memory index was outside the tracking memory size. These events are converted to negative values to "remove" them as tracking events. 
            2. IndexPosTracking are the positions in MemIndexScaled where the valus are positive (Tracking)
        """
        # Get the index values which are usually 0-5 for the tracking events and 0-63 for the scanning (OBS! scanning events normally 0-199 and thus this value is not reliable)
        MemIndex = TrackingFlags[3,:]
        # Get the flag that shows if the tracking is ON
        TrackingONFlag = TrackingFlags[0,:].astype('int16')

        if TrackingONFlag[0]==1: #system is tracking from start 
            print("Sytem started with tracking, removing these tracking events")
            StartPos = np.where(np.diff(TrackingONFlag)==1)[0]
            #EndPos = np.where(np.diff(TrackingONFlag)==-1)[0]
            #CrossPoint=EndPos[np.where(StartPos[0]>EndPos)[0][-1]]
            TrackingONFlag[:StartPos[0]+1]=0

        # Move the memindex value such that tracking events (1) are positive and scanning events (-1) are negative
        # We first offset the index values by 1 and then multiply by the Tracking flag (No =1, Off=-1), we then remove the offset of 1.
        # thus all tracking events are now >=0 and all scanning events are now <0 
        MemIndexScaled = (MemIndex+1)*(TrackingONFlag*2-1)-1
        MaxmMemIndexScaledAllowed = np.max(np.where(ScanProject.PosCalibration['LaserShotPattern']>0))# Largest value in the memory index that can occur during tracking
                                                                                                      # OLD ScanProject.PosCalibration['TrackingMemsize']-1 

        # There are ocations when the tracking starts when the Tracking flag is ON 
        # but 
        # 1. index of memory is larger than tracking memory size
        # 2. tracking index sequense that are very short, two times the memory index cycle 
        # 3. a few tracking events at the beginging do not follow the pattern 0,1,2,3,4,... MaxMemsize
        # 4. the tracking first postion is not 0

        #1. First negate index of memory that are larger than tracking memory size during Tracking ON
        Memslips = np.array(np.where(MemIndexScaled > MaxmMemIndexScaledAllowed))
        if Memslips.size != 0: 
            # Removing memslip events, if any, from the tracking events. 
            MemIndexScaled[Memslips] = -MemIndexScaled[Memslips]
            if verbose:
                print('Mem slip occured ', Memslips.size, ' times')
        
        #2. Remove trajectories that are too short: 2 memory cycles
        MemIndexScaled=self.RemoveShortTraj(MemIndexScaled,MaxmMemIndexScaledAllowed)
        print('RemDone')
        #3. and 4. Remove trajectories do not have preiodicity in there zero mem index
        #          and trim trajectories, to nerest zero index, if they do not start with a zero men index 

        StartPos = np.where(np.diff(np.int32(MemIndexScaled>=0))==1)[0]+1
        EndPos = np.where(np.diff(np.int32(MemIndexScaled>=0))==-1)[0]
        StartPos=StartPos[:len(EndPos)] #Trim down Start to not include last traking if tracking was stopped without an end 
        for i in range(len(StartPos)):
            T=MemIndexScaled[StartPos[i]:EndPos[i]+1] # Extract trajectory

            if np.sum(ScanProject.PosCalibration['LaserShotPattern']==0)>0:
                print('Tracking pattern includes blanking, find first non blanking position') 
                TestIndex=np.where(ScanProject.PosCalibration['LaserShotPattern']>0)[0][0]
                TestSpacing=len(np.where(ScanProject.PosCalibration['LaserShotPattern']>0)[0])
            else:
                TestIndex=0
                TestSpacing=MaxmMemIndexScaledAllowed+1

            DT=np.where(T==TestIndex)[0]                    # Find all memory index zero or first not blanked index
            MemError=np.where(np.diff(DT)!=TestSpacing)[0] # Check if the spaising of the zeros are correct 
            if MemError.size!=0: #Remove trajectory if spacing is strange 
                if np.sum(np.abs(np.diff(TrackingFlags[2,:])))==0:
                    print('Trajectory has strange memory pattern, removing it')
                    print("found ex.: " + str(T[DT[MemError[0]]-2:DT[MemError[0]]+10]))
                    MemIndexScaled[StartPos[i]:EndPos[i]+1]=-MemIndexScaled[StartPos[i]:EndPos[i]+1] - 1 # the -1 is to remove zeros 
                else: 
                    print('Mem index switching seems to be used, skipping pattern memory test')

            elif DT[0]!=0: #Remove point in the begining that do not follow the tracking memory pattern
                print('Trajectory begining is trimed due to memory pattern')
                print("found ex.: " + str(MemIndexScaled[StartPos[i]-2:StartPos[i]+10]))
                MemIndexScaled[StartPos[i]:StartPos[i]+DT[0]]=-MemIndexScaled[StartPos[i]:StartPos[i]+DT[0]]-1
                
        # 2. run this one more time
        MemIndexScaled=self.RemoveShortTraj(MemIndexScaled,MaxmMemIndexScaledAllowed)
        
        # Find all tracking events
        IndexPosTracking = np.where(MemIndexScaled >= 0)  
        
        return IndexPosTracking, MemIndexScaled
    
    def GetBeamOffset(self, ScanProject, TrackingFlags):    
        """
        Add the EOD offset
        """
        IndexPosTracking, MemIndexScaled = self.GetIndexPosTracking(ScanProject,TrackingFlags)
        MemIndexTracking = MemIndexScaled[IndexPosTracking]
        
        # Get offset values for the EOD pattern in micro meters
        EODOffsetValuesXY = self.GetEODOffsetValueInMicroMeters(ScanProject) #[um]
        OffsetValueZ = np.array(ScanProject.PosCalibration['zLayerOffset']) #[index]
        
        # Create offset event vector for only the tracking trajectories 
        OffsetXYValueMicroMeter = EODOffsetValuesXY[:,MemIndexTracking.astype('int')] #[um]        
        OffsetZValueIndex = OffsetValueZ[MemIndexTracking.astype('int')] #[index]
        
        # Create offset event vector of the full scan
        ScanLength = TrackingFlags[3,:].shape[0]
        XYoffsetFullScan = np.zeros([2,ScanLength])
        XYoffsetFullScan[:,IndexPosTracking[0]] = OffsetXYValueMicroMeter  #[um]
        
        ZoffsetFullScan = np.zeros(ScanLength)
        ZoffsetFullScan[IndexPosTracking[0]] = OffsetZValueIndex  #[index]
        
        return XYoffsetFullScan, ZoffsetFullScan  
        
    def GetEODXYPIDOutputInMicroMeters(self, ScanProject, EODSignal):
        """
        For the EOD x and y signal we multiply with the scale factor that is implemented in the FPGA before sending the signal to the EOD
        We then multipy with the int 16 to micrometer conversion to get everything in micro meters.
        Obs! this is only done to the x and y data 
             z EOD signal is flushed through since it is not used. 

        Obs!! The EODSignal is the total EOD movement thus it is both the current center plus the total PID error (no filter).
        
        """
        FilterScaling = ScanProject.PosCalibration['EODPiezofilterScaling']*ScanProject.PosCalibration['EODUnitconversion'] #[um]
        EODXYSignalScaled = EODSignal[0:2,:]*FilterScaling

        zpart=EODSignal[2,:].reshape((1,len(EODSignal[2,:])))
        EODXYSignalScaled = np.append(EODXYSignalScaled, zpart, axis=0)

        return EODXYSignalScaled     

    def GetHighLowFilteredEODSignal(self, ScanProject, EODPIDCenterXYZ, TrackingFlags):
        """
        Create the filter (exponential moving average)
        Obs xyESG is the total EOD movement thus it is both the current center plus the total PID error (no filter). 
        Let y0=0 and then y_i=W*x_i+(1-W)*y_{i-1}   =>  y_i - (1-W)*y_{i-1}=W*x_i which is form that the filter function is written in.
        We define the low frequency component of the signal (LoFilterTracksXY) with the filter.
        Using, Lo + Hi = Total => Hi = Total - Lo, we calculate the high frequency component of the signal (HiFilterTracksXY).
        """

        IndexPosTracking, _ = self.GetIndexPosTracking(ScanProject,TrackingFlags, False)

        # Get the flag that shows if the tracking is ON
        TrackingStart = np.where(np.diff(IndexPosTracking)[0] != 1)
        TrackingStart = np.concatenate(([0],TrackingStart[0],[IndexPosTracking[0].size]))


        EODPIDCenterTracksXY = np.squeeze(EODPIDCenterXYZ[0:2,IndexPosTracking])
        W = ScanProject.PosCalibration['EODPiezofilter']

        
        FilteredTracksX = np.zeros(IndexPosTracking[0].shape)
        FilteredTracksY = np.zeros(IndexPosTracking[0].shape)
        for TrackIndex in range(TrackingStart.size-1):
            TrackInterval = np.arange(TrackingStart[TrackIndex],TrackingStart[TrackIndex+1])
            # ***** difference with MATLAB code --> Filtered values have a difference of ~10**-3. Could be due to rounding error in the filter, but I don't know for sure
            FilteredTracksX[TrackingStart[TrackIndex]:TrackingStart[TrackIndex+1]] = lfilter([W],[1, -(1-W)],EODPIDCenterTracksXY[0,TrackInterval])
            FilteredTracksY[TrackingStart[TrackIndex]:TrackingStart[TrackIndex+1]] = lfilter([W],[1, -(1-W)],EODPIDCenterTracksXY[1,TrackInterval])

        LoFilterTracksXY = np.zeros(np.squeeze(EODPIDCenterXYZ[0:2]).shape)
        LoFilterTracksXY[:,IndexPosTracking[0]] = np.array((FilteredTracksX, FilteredTracksY))

        HiFilterTracksXY = np.zeros(LoFilterTracksXY.shape)
        HiFilterTracksXY[:,IndexPosTracking[0]] = EODPIDCenterTracksXY - LoFilterTracksXY[:,IndexPosTracking[0]]

        return LoFilterTracksXY, HiFilterTracksXY
        
    def AddStarPattern(self, PiezoSampledInMicroMeters, HiFilterTracksXY, XYoffsetFullScan):
        """
        Total center postion Obs! Using the piezo raw since this is the true position of the pinhole
        On top off that we add the star offset

        Input: 
            PiezoSampledInMicroMeters: Piezo center slow movment 
            HiFilterTracksXY:          EOD center fast movment 
            XYoffsetFullScan:          Tracking pattern 

        OutPut:
            x:          x value with all three parts added 
            y:          y value with all three parts added
            XYTotal:   x and y values without the tracking pattern

        """
        XYTotal = PiezoSampledInMicroMeters - HiFilterTracksXY #[um]
        xy = XYTotal + XYoffsetFullScan #[um]

        return xy[0,:], xy[1,:], XYTotal  

    def BuildScanMatrix(self,
                        ScanTime,
                        x,
                        y,
                        PhotonAPDCounts,
                        TrackingFlags,
                        PiezoSampledInMicroMeters,
                        XYoffsetFullScan,
                        EODPIDCenterXYZ,
                        XYTotal,
                        PiezoSignalInMicroMetersAndSLMIndex,
                        ZoffsetFullScan,
                        PiezoSignal,
                        HiFilterTracksXY,
                        x2, y2,
                        Ztotal,
                        SkipIndexVector):
        """
        Package all data into a single 2D numpy array, ScanMatrix. 
        ScanInfo is a list containing descriptions of each column of data in ScanMatrix
        """
        ScanInfo = [
            '0.  Time[s]',
            '1.  x value including all contributions: piezo center (see 9) + EOD center (see 23) + EOD tracking pattern [um] (see 11)', # ***** not the same as MATLAB code
            '2.  y value including all contributions: piezo center (see 10)+ EOD center (see 24) + EOD tracking pattern [um] (see 12)', # ***** not the same as MATLAB code
            '3.  APD1[counts], H polarization',
            '4.  APD2[counts], V polarization',
            '5.  Tracking on',
            '6.  Release Threshold not reached',
            '7.  Laser power switch during tracking',
            '8.  Memory index',
            '9.  x (Current piezo value) Piezo values are sampled in the 200Mhz loop, first part in x (see 1)', 
            '10. y (Current piezo value) Piezo values are sampled in the 200Mhz loop, first part in y (see 2)', 
            '11. x EOD offset vector which gives the tracking pattern [um], thired part in x (see 1)',
            '12. y EOD offset vector which gives the tracking pattern [um], thired part in y (see 2)',
            '13. x EOD position signal from FPGA estimation [um], with out filtering', 
            '14. y EOD position signal from FPGA estimation [um], with out filtering', 
            '15. x value including contributions from piezo center + EOD center [um], first and second part in x (see 1)',
            '16. y value including contributions from piezo center + EOD center [um], first and second part in y (see 2)',
            '17. x value Scanning: (Current piezo value) Piezo values are sampled in the 200Mhz loopx piezo (same as 9), Tracking: (Next piezo setting) Signal that in the FPGA is sending to the Piezo stage (low frequency signal from the PID estimation)',
            '18. y value Scanning: (Current piezo value) Piezo values are sampled in the 200Mhz loopx piezo (same as 9), Tracking: (Next piezo setting) Signal that in the FPGA is sending to the Piezo stage (low frequency signal from the PID estimation)',
            '19. z Offset index of 1/-1 during tracking corresponding to upper and lower layer',
            '20. z SLM center index',
            '21. z Scanning: 0, Tracking: current center value  + FPGA PID signal low frequency part',
            '22. z total offset is the sum if 19 and 20 with filter to restrict the values acording to the FPGA logic times the staircase hight [um]',
            '23. x EOD high frequency part, second part in x (see 1)',
            '24. y EOD high frequency part, second part in x (see 2)',
            '25. Where the skip steps are inserted when using skip cycles HW setting'
        ]


        ScanMatrix = np.vstack((
                ScanTime,                                   #0
                x,                                          #1
                y,                                          #2
                PhotonAPDCounts[0,:],                       #3
                PhotonAPDCounts[1,:],                       #4
                TrackingFlags[0,:],                         #5
                TrackingFlags[1,:],                         #6
                TrackingFlags[2,:],                         #7
                TrackingFlags[3,:],                         #8
                PiezoSampledInMicroMeters[0,:],             #9 
                PiezoSampledInMicroMeters[1,:],             #10 
                XYoffsetFullScan[0,:],                      #11
                XYoffsetFullScan[1,:],                      #12
                EODPIDCenterXYZ[0,:],                       #13
                EODPIDCenterXYZ[1,:],                       #14
                XYTotal[0,:],                               #15 
                XYTotal[1,:],                               #16
                x2,                                         #17 
                y2,                                         #18 
                ZoffsetFullScan,                            #19 
                PiezoSignalInMicroMetersAndSLMIndex[2,:],   #20 
                PiezoSignal[2,:],                           #21 
                Ztotal,                                     #22
                HiFilterTracksXY[0,:],                      #23
                HiFilterTracksXY[1,:],                      #24
                SkipIndexVector                             #25
        ))

        return ScanMatrix, ScanInfo

    def getScanMatrix(self,ScanProject):
        """
        Swich board for how to get data
        """
        self.getNewScanMatrix(ScanProject)

    def getNewScanMatrix(self, ScanProject):
        """
        RawData --> ScanMatrix
        Principal method for this class. Almost all other DataObject methods are called here.
        """
        FileNumberToOpen=ScanProject.FILE_TO_OPEN
        RawData=self.GetScanData(ScanProject)
        
        # Check integrity of raw data 
        PosErrorFlag = self.CheckDataIntegrity(RawData)
        
        # Unpack data
        PhotonAPDCounts, PiezoSampled, TrackingFlags, EODSignal, PiezoSignal = self.UnpackRawData(ScanProject,RawData, PosErrorFlag)
        
        ###################################################################################
        # The pipline of the data assambly is:
        #  
        # 1. Reconstruct time
        #
        # 2. Reconstruct position "Slow tracking", 
        #    get the piezo tip/tilt mirror position in [um], there are two versions of this signal sampled in the FPGA
        #      a. (Current piezo value) Piezo values are sampled in the 200Mhz loop 
        #      b. (Next piezo setting + next SLM index) Signal that in the FPGA is sending to the Piezo stage 
        #         , low frequency signal from the PID estimation. 
        #         z center index from the FPGA is flushed through. 
        #
        # 3. "Fast tracking": Get the EOD xy in [um], 
        #    a. Get the EOD xy, this is the total EOD movement thus it is both the current center plus the total PID error (no filter).
        #    b. Filter out the high and low frequensy part of the dignal in aa. 
        #
        # 4. Reconstruct the tracking pattern for each point in the tacking for xy [um] and z [index(+-1)]  
        #
        # 5. Put all parts together for the xy signal. We are adding together the points above:
        #   a. 2a. the sampled Piezo signal (current true postion), corresponding to the center slow movment of the piezo stage in [um]
        #   b. 3b. OED high frequency part, corresponding the fast centre movment of the EOD   [um]
        #   c. 4.  OED tracking pattern                                [um]
        #
        # 6. Organize collected data from the entire scan into a matrix. ScanInfo is a list for labeling the ScanMatrix
        #
        # 7. If the tracking pattern includes blanking steps add these time steps and memory indexis to the scan matrix 
        ###################################################################################

        #1. 
        #Get the scantime vector in [sec], and vector of where (if any) skips where inserted
        ScanTime, SkipIndexVector = self.GetScanTime(ScanProject,TrackingFlags)

        #2. 
        #Convert piezo signals of tip/tilt mirror to micro meter [um], z center from the FPGA is flushed through
        # a. (Current piezo value) Piezo values are sampled in the 200Mhz loop 
        # b. (Next piezo setting + next SLM index) Signal that in the FPGA is sending to the Piezo stage 
        #     (low frequency signal from the PID estimation)
        #     z center from the FPGA is flushed through 
        PiezoSampledInMicroMeters = self.ConvertData(ScanProject, PiezoSampled)           
        PiezoSignalInMicroMetersAndSLMIndex = self.ConvertData(ScanProject, PiezoSignal)  
        
        #3.        
        # a. Get EOD beam center position [um]
        #     OBS! only EOD XY is used, EOD Z is not in use and only flushed through, see "Unpacking second data set, Data used when tracking"
        #     Obs!! The EODSignal is the total EOD movement thus it is both the current center plus the total PID error (no filter).
        #
        # b. Get the high and low filtered signals of the PID signal,
        EODPIDCenterXYZ = self.GetEODXYPIDOutputInMicroMeters(ScanProject, EODSignal)
        _, HiFilterTracksXY = self.GetHighLowFilteredEODSignal(ScanProject,EODPIDCenterXYZ, TrackingFlags)        
        
        #4.
        # Get EOD offset vector which gives the tracking pattern
        # XYoffsetFullScan is in [um]
        # ZoffsetFullScan are index of 1/-1 during tracking corresponding to upper and lower layer
        XYoffsetFullScan, ZoffsetFullScan = self.GetBeamOffset(ScanProject, TrackingFlags)

        #5.       
        # a. Put all parts together for the xy signal. We are adding together the points above:
        #    2a. the sampled Piezo signal (current true postion), corresponding to the center slow movment of the piezo stage in [um]
        #    3b. OED high frequency part, corresponding the fast centre movment of the EOD   [um]
        #    4.  xy OED tracking pattern                                [um]
        # 
        # b. Get z total in [um] by combining:
        #    2b. next SLM index, z center from the FPGA
        #    4.  z tracking pattern                                    [um]
        x, y, XYTotal = self.AddStarPattern(PiezoSampledInMicroMeters, HiFilterTracksXY, XYoffsetFullScan)
        Ztotal = self.GetZinMicrometers(ScanProject,PiezoSignalInMicroMetersAndSLMIndex,ZoffsetFullScan)

        #6. Combining the scanning position with the tracking PID piezo estimation  
        # HERE 
        x2, y2 = self.GetPiezoScanWithFPGATrackingPID(ScanProject, PiezoSampledInMicroMeters, PiezoSignalInMicroMetersAndSLMIndex,TrackingFlags)

        #7.Organize collected data from the entire scan into a matrix. ScanInfo is a list for labeling the ScanMatrix
        ScanProject.ScanMatrix, ScanProject.ScanInfo = self.BuildScanMatrix(
                                ScanTime,
                                x,
                                y,
                                PhotonAPDCounts,
                                TrackingFlags,
                                PiezoSampledInMicroMeters,
                                XYoffsetFullScan,
                                EODPIDCenterXYZ,
                                XYTotal,
                                PiezoSignalInMicroMetersAndSLMIndex,
                                ZoffsetFullScan,
                                PiezoSignal,
                                HiFilterTracksXY,
                                x2, y2,
                                Ztotal,
                                SkipIndexVector)

        #8. Add blanking steps if tracking includes blanking 
        ScanProject = self.AddBlankingToScanMatrix(ScanProject)

    def AddBlankingToScanMatrix(self,ScanProject):
        """
        Inserts the missings blanking events in time and memory indecis into the ScanMatrix 
        """
        MemTrackSize=np.int8(ScanProject.PosCalibration['TrackingMemsize'])
        BlankingIndex=np.where(ScanProject.PosCalibration['LaserShotPattern']==0)[0]
        BlankingIndexP=np.mod(BlankingIndex+1,MemTrackSize)
        if len(BlankingIndex)>0: 
            for i in range(len(BlankingIndex)):
                TrackingOn=(2*np.int8(ScanProject.ScanMatrix[5,:])-1)
                MemIndexTracking= (np.int8(ScanProject.ScanMatrix[8,:])+1)*TrackingOn -1 #Tracking index are >=0
                P=np.where(MemIndexTracking==BlankingIndexP[i])
                ToInsert=ScanProject.ScanMatrix[:,P[0]];#+ScanProject.ScanMatrix[:,P[0]-1])/2
                ToInsert[8,:]=BlankingIndex[i]; ToInsert[5,:]=1
                ScanProject.ScanMatrix=np.insert(ScanProject.ScanMatrix,P[0],ToInsert,axis=1)

            #Time compensation 
            TrackingOn=(2*np.int8(ScanProject.ScanMatrix[5,:])-1)
            MemIndexTracking= (np.int8(ScanProject.ScanMatrix[8,:])+1)*TrackingOn -1
            TimeSteps=np.zeros(ScanProject.ScanMatrix[0,:].shape)
            for i in range(len(BlankingIndex)):
                WhereTo=(MemIndexTracking==BlankingIndex[i])
                TimeSteps[np.where(WhereTo)]=1/ScanProject.PosCalibration['ScanSpeedTracking']

            ScanProject.ScanMatrix[0,:]=ScanProject.ScanMatrix[0,:]+np.cumsum(TimeSteps)
            print('Blanking compensation done!')
        return ScanProject

    def AddBlankingToScanMatrix2(self,ScanProject):
        """
        Inserts the missings blanking events in time and memory indecis into the ScanMatrix 
        """
        MemTrackSize=np.int8(ScanProject.PosCalibration['TrackingMemsize'])
        if (np.diff(ScanProject.PosCalibration['TrackingMemSlots1'])[0]<MemTrackSize)&(np.diff(ScanProject.PosCalibration['TrackingMemSlots2'])[0]<MemTrackSize):
            print("Found missmetch in tracking total memory and the slot size, pattern is assumed to switch!")

            IndexMemSlots=(np.int8(ScanProject.PosCalibration['TrackingMemSlots1']),np.int8(ScanProject.PosCalibration['TrackingMemSlots2']))

            for j in range(len(IndexMemSlots)):
                IndexMemSlot=IndexMemSlots[j]
                BlankingIndex=np.where(ScanProject.PosCalibration['LaserShotPattern']==0)[0]
                BlankingIndex=BlankingIndex[np.where((BlankingIndex>=IndexMemSlots[j][0])*(BlankingIndex<=IndexMemSlots[j][1]))][::-1]
                BlankingIndexP=BlankingIndex+1
                BlankingIndexP=np.invert(BlankingIndexP>IndexMemSlots[j][1])*BlankingIndexP + (BlankingIndexP>IndexMemSlots[j][1])*IndexMemSlots[j][0]

                if len(BlankingIndex)>0: 
                    for i in range(len(BlankingIndex)):
                        if j==0:
                            PatternType=ScanProject.ScanMatrix[7,:].astype(bool)
                        elif j==1:
                            PatternType=np.invert(ScanProject.ScanMatrix[7,:].astype(bool))

                        TrackingOn=(2*np.int8(ScanProject.ScanMatrix[5,:]*PatternType)-1)
                        MemIndexTracking= (np.int8(ScanProject.ScanMatrix[8,:])+1)*TrackingOn -1 #Tracking index are >=0
                    
                        P=np.where(MemIndexTracking==BlankingIndexP[i])
                        ToInsert=ScanProject.ScanMatrix[:,P[0]];#+ScanProject.ScanMatrix[:,P[0]-1])/2
                        ToInsert[8,:]=BlankingIndex[i]; ToInsert[5,:]=1
                        ScanProject.ScanMatrix=np.insert(ScanProject.ScanMatrix,P[0],ToInsert,axis=1)

            #Time compensation 
            TrackingOn=(2*np.int8(ScanProject.ScanMatrix[5,:])-1)
            MemIndexTracking= (np.int8(ScanProject.ScanMatrix[8,:])+1)*TrackingOn -1
            TimeSteps=np.zeros(ScanProject.ScanMatrix[0,:].shape)
            for i in range(len(BlankingIndex)):
                WhereTo=(MemIndexTracking==BlankingIndex[i])
                TimeSteps[np.where(WhereTo)]=1/ScanProject.PosCalibration['ScanSpeedTracking']

            ScanProject.ScanMatrix[0,:]=ScanProject.ScanMatrix[0,:]+np.cumsum(TimeSteps)
            print('Blanking compensation done!')

        else:
            ScanProject=self.AddBlankingToScanMatrix(ScanProject)
        return ScanProject
                   
    def GetZinMicrometers(self,ScanProject, PiezoSignalInMicroMetersAndSLMIndex,ZoffsetFullScan):
        # Gets the full z tracking in [um]
        # 1. Add the tracikng offset [+-1] to the center index value
        # 2. Truncate above and below to the amount of z levels used in the staircase
        # 3. convert index to [um] 
        Z_index_total = PiezoSignalInMicroMetersAndSLMIndex[2,:] + ZoffsetFullScan 
        Z_index_total[Z_index_total<0]=0
        Z_index_total[Z_index_total > ScanProject.PosCalibration['zLevels']-1 ] = ScanProject.PosCalibration['zLevels']-1
        Z_total=Z_index_total*ScanProject.PosCalibration['SLMSwitchHight'] #[um]
        return Z_total

    def GetPiezoScanWithFPGATrackingPID(self, ScanProject, PiezoSampledInMicroMeters, PiezoSignalInMicroMetersAndSLMIndex,TrackingFlags):
        # Combining the scanning position with the tracking PID piezo estimation  
        # 1. find where we have tracking.
        # 2. Making a new copy of the 200 Mhz piezo sampling (consider correct for the scanning) 
        # 3. Extgange the parts related to tracking with the FPGA PID values in [um]
        
        IndexPosTracking, _ = self.GetIndexPosTracking(ScanProject,TrackingFlags, False)
        x2 = PiezoSampledInMicroMeters[0,:].copy()
        x2[IndexPosTracking] = PiezoSignalInMicroMetersAndSLMIndex[0,IndexPosTracking]

        y2 = PiezoSampledInMicroMeters[1,:].copy()
        y2[IndexPosTracking] = PiezoSignalInMicroMetersAndSLMIndex[1,IndexPosTracking]

        return x2, y2

    def CalcBasicPhotonStats(self, ScanProject, verbose=True):
        """
        Assumes that "ProcessData" has been run beforehand.
        
        Calculates 5 basic statistics from APD photon counts.
        1) AvgTotalPhotons. Mean of total (sum of both APDs) photons in ScanMatrix
        2) AvgTotalPhotonRate. Total photon count / total measurement time
        3) ThresholdPhotonRate. Photon rate for counts below background threshold
        4) AvgPolTotalPhotons. Mean polarization of all photons (H - V)/Total
        5) AvgPolAboveBkg. Mean polarization for counts above background threshold
        
        INPUT: within self it will use:  .ScanMatrix, .DataSelectionRules,  and ini file settings 
        OUTPUT: within self it will created:  .BasicStat, and print to stout 
        """
        HPhotons = ScanProject.ScanMatrix[3,:]
        VPhotons = ScanProject.ScanMatrix[4,:]
        HplusVPhotons = HPhotons + VPhotons 
        AvgTotalPhotons = np.mean(HplusVPhotons)
        AvgTotalPhotonRate = AvgTotalPhotons / ScanProject.PosCalibration['MeasurementOnTime']

        # Find counts and calculate rate for counts below background threshold (specified in .ini).
        BelowBackground = np.where(HplusVPhotons < ScanProject.TrajectorySelectionRules['BackgroundThreshold'])
        ThresholdPhotonRate = np.mean(HplusVPhotons[BelowBackground]) / ScanProject.PosCalibration['MeasurementOnTime']
        AvgPolTotalPhotons = (np.sum(HPhotons) - np.sum(VPhotons)) / np.sum(HplusVPhotons)

        # Mean photon count over data with more than the background threchold
        AboveBackground = np.where(HplusVPhotons >= ScanProject.TrajectorySelectionRules['BackgroundThreshold'])
        AvgPolAboveBkg = (np.sum(HPhotons[AboveBackground]) - np.sum(VPhotons[AboveBackground])) / np.sum(HplusVPhotons[AboveBackground])

        if verbose:
            print('Mean photon count over all data:', AvgTotalPhotons)
            print('Mean photon rate over all data:', AvgTotalPhotonRate, ' Hz')
            print('Mean photon rate over data with less than', ScanProject.TrajectorySelectionRules['BackgroundThreshold'], 'counts:', ThresholdPhotonRate, 'Hz')
            print('Mean polarization over all data:', AvgPolTotalPhotons)
            print('Mean polarization over data with more than', ScanProject.TrajectorySelectionRules['BackgroundThreshold'], 'counts:', AvgPolAboveBkg) 

        ScanProject.BasicStat['AvgTotalPhotons'] = AvgTotalPhotons
        ScanProject.BasicStat['AvgTotalPhotonRate'] = AvgTotalPhotonRate
        ScanProject.BasicStat['ThresholdPhotonRate'] = ThresholdPhotonRate
        ScanProject.BasicStat['AvgPolTotalPhotons'] = AvgPolTotalPhotons
        ScanProject.BasicStat['AvgPolAboveBkg'] = AvgPolAboveBkg

        # Adjust trajectory selection criteria based on background photon rate
        cut_traj=ScanProject.TrajectorySelectionRules['CutTraj']
        cut_traj_laser_high=ScanProject.TrajectorySelectionRules['CutTrajWhenLaserHigh']
        
        if cut_traj[2] == 1:
            cut_traj[0] = cut_traj[0] + ThresholdPhotonRate
            cut_traj_laser_high[0] = cut_traj_laser_high[3] + ThresholdPhotonRate

    def get_raw_data_filter(self,ScanProject):
        """
        Generate filter vector for the ScanMatrix
        INPUT: Assumes that the folowing exists:
               1.      ScanProject.DataSelectionRules
               2.      ScanProject.ScanMatrix
        
        OUTPUT: ScanProject.filter_vector with is a Nx1 vector with 0/1 (not pass/ pass) values
        """
        wndw=ScanProject.TrajectorySelectionRules['wndw'].astype('int')
        ScanProject.filter_vector=FilterLibrary.Raw_data_filters.raw_data_filter(ScanProject.DataSelectionRules, ScanProject.ScanMatrix,wndw=wndw)

    def get_TrajObject(self):
        self.TrajObject = TrajObject(self)
    
    def get_Trajectories(self, ScanProject,Filter=1):
        self.TrajObject.GetTrajs(ScanProject,Filter=Filter)

    def do_TrimTrajs(self, ScanProject, StartTrim, EndTrim, ForceMaxTrajLength=-1):
        self.TrajObject.TrimTrajs(ScanProject, StartTrim, EndTrim, ForceMaxTrajLength)
        
    def get_Gaussian_traj_estimation(self,ScanProject): 
        mode=ScanProject.TrajectorySelectionRules['Mode']
        wndw=ScanProject.TrajectorySelectionRules['wndw']
        APDfilter=[0, np.inf]
        MeasurementOnTime=ScanProject.PosCalibration['MeasurementOnTime']
        beta =ScanProject.PosCalibration['beta']
        
        run_cycles        = np.int(ScanProject.PosCalibration['run_cycles'])  
        skip_cycles       = np.int(ScanProject.PosCalibration['skip_cycles']) 
        if (run_cycles!=0) & (skip_cycles!=0) & (mode=='RollingAv'):
            print("OBS! None linear timing, result might be strange, consider to use mode: FixWindowAv")
        
        self.TrajObject.get_Gaussian_traj_estimation(ScanProject, wndw, beta, mode, ScanProject.Trajs, ScanProject.ScanMatrix, APDfilter, MeasurementOnTime)
    
    def get_traj_Poisson_ProbMap_AnaPSF(self, ScanProject, TrajToGet, MaxTrajTime=-1, verbose=1):
        self.TrajObject.get_traj_Poisson_ProbMap_AnaPSF(ScanProject, TrajToGet, MaxTrajTime, verbose)

    def get_All_traj_Poisson_ProbMap_AnaPSF(self, ScanProject, verbose=1):
        """
        Go through all trajectories in ScanProject.Trajs
        and calculates the Poisson_ProbMap using Gaussian analytical PSF, 
        Data is stored in: ScanProject.PoissonProbXYZMap_AnaPSF
        """
        for TrajToGet in range(ScanProject.Trajs.shape[1]):
            print("Procesing Traj:" + str(TrajToGet))
            self.get_traj_Poisson_ProbMap_AnaPSF(ScanProject,TrajToGet,verbose=0) # Calculations are stored in, ScanProject.TrajObject.PoissonProbXYMap

    def get_traj_Poisson_ProbMap_NumPFS(self, ScanProject, TrajToGet, verbose=1, do_onlinesave=0,ResamplePhotons=0,EstBackground=0):
        #self.TrajObject.get_traj_Poisson_ProbMap_NumPFS(ScanProject, TrajToGet, verbose, do_onlinesave)
        if EstBackground==0:
            self.TrajObject.get_traj_Poisson_ProbMap_fix_NumPFS(ScanProject, TrajToGet, verbose=verbose, do_onlinesave=do_onlinesave, ResamplePhotons=ResamplePhotons)
        elif EstBackground==1:
            self.TrajObject.get_traj_Poisson_ProbMap_fix_NumPFS2(ScanProject, TrajToGet, verbose=verbose, do_onlinesave=do_onlinesave, ResamplePhotons=ResamplePhotons)

    def get_All_traj_Poisson_ProbMap_NumPFS(self, ScanProject, verbose=1, do_onlinesave=1,ResamplePhotons=0,ContinueFrom=0,EstBackground=0):
        """
        Go through all trajectories in ScanProject.Trajs
        and calculates the Poisson_ProbMap, 
        Data is stored in: ScanProject.PoissonProbXYZMap_NumPSF
        """
        for TrajToGet in range(ContinueFrom,ScanProject.Trajs.shape[1]):
            print("Procesing Traj:" + str(TrajToGet) + "/" + str(ScanProject.Trajs.shape[1]))
            self.get_traj_Poisson_ProbMap_NumPFS(ScanProject,TrajToGet,verbose, do_onlinesave=0, ResamplePhotons=ResamplePhotons,EstBackground=EstBackground) # Calculations are stored in, ScanProject.TrajObject.PoissonProbXYMap
            if do_onlinesave:
                if ResamplePhotons>0:
                    self.PickleData(ScanProject,override=0,FileEnd='_ScanProject_Resampled.pkl')
                else:
                    self.PickleData(ScanProject,override=1)

    def get_TrajsDic(self,ScanProject, FilterdTrajDic):
        """
        Make new TrajDic which includes all trajectoris, 

        Input: 
            FilterdTrajDic: is a dictionary with trajectories each in the t,x,y,z,n format 

        Output: 
            ScanProject.TrajsDic: 
                The new TrajDic which includes all trajectoris are anotated as:
                ('Original','-') is the original 
                ('Filtered','0') then comes the new one which are the filterd ones, to them we also attach the new photon counts

                Trajs format is (N x 3) numpy array which contains:
                0. Trajectory time length [s]
                1. Start index in the ScanMatrix
                2. End index   in the ScanMatrix

        """
        ScanProject.TrajsDic={}
        ScanProject.TrajsDic['Original']=ScanProject.Trajs
        NewTrajs=np.zeros((3,len(FilterdTrajDic)))
        for index in FilterdTrajDic: 
            FilterdTrajDic[index]
            length=FilterdTrajDic[index][0,:][-1]-FilterdTrajDic[index][0,:][0]
            Time=ScanProject.ScanMatrix[0,:] #Get time from ScanMatrix
            StartStopIndex=np.where((Time>=FilterdTrajDic[index][0,:][0]) * (Time<=FilterdTrajDic[index][0,:][-1]))[0][[0,-1]]
            NewTrajs[:,index]=[length,StartStopIndex[0],StartStopIndex[1]]
                
        ScanProject.TrajsDic['Filtred']=NewTrajs

    def get_PhotonResampledTraj(self,ScanProject, wndw=1,runs=1):
        """
        Re run the MLE estimation with resampled photoncounts
        Input: 
            ScanProject.TrajsDic['Filtred'] : the new filterd start and stop positions, this will overrite the ScanProject.Trajs used in the estimation function 
            wndw: amount measurements used to estimate the mean photon count used in resampling 
            runs: amount of times to run a trajectory with resampled photon counts

        """
        ScanProject.TrajsDic['wndw']=wndw
        ScanProject.Trajs=ScanProject.TrajsDic['Filtred'] #override old Traj to run new reconstruction 

        #Run re-sampling MLE estimation
        self.get_TrajObject()
        for index in range(runs):
            self.get_All_traj_Poisson_ProbMap_NumPFS(ScanProject,do_onlinesave=1, ResamplePhotons=1) #New reconstruction
        
    def get_All_traj_MSD(self, ScanProject,TrajDictionary):
        """
        Go through all poisson trajectory estimations and calculates the MSD
        Note: This might be good to make generic, such that on can input what to estimate. 
        """
        ScanProject.txyzMSD_Poisson_ProbMap={}
        for TrajToGet in range(len(TrajDictionary)):        
            txyzn_Map  = TrajDictionary[TrajToGet]
            txyzMSD = self._Ph.get_MSD(txyzn_Map[:4,:])
            ScanProject.txyzMSD_Poisson_ProbMap[TrajToGet]=txyzMSD

    def get_All_traj_CVE(self, ScanProject,TrajDictionary):
        """
        Go through all poisson trajectory estimations and calculates the MSD
        Note: This might be good to make generic, such that on can input what to estimate. 
        """
        txyzCVE_Poisson_ProbMap=self._Ph.get_All_traj_CVE(TrajDictionary)
        ScanProject.txyzCVE_Poisson_ProbMap=txyzCVE_Poisson_ProbMap

    def get_All_traj_DMSD_Dt(self, ScanProject):
        """
        Calculates the derivative of the data that is orginized as: 
        INPUT:
            ScanProject.txyzMSD_Poisson_ProbMap is a dictionary with  
        """
        ScanProjectMSD = ScanProject.txyzMSD_Poisson_ProbMap
        ScanProject.txyzDMSD_Dt_Poisson_ProbMap={}
        for TrajToGet in range(len(ScanProjectMSD)):        
            xyz_MSD  = ScanProjectMSD[TrajToGet][1:,:] 
            t=ScanProjectMSD[TrajToGet][0,:]

            Dt=np.diff(t)
            Dxyz_MSD = np.diff(xyz_MSD,axis=1)
            Dxyz_Dt  = Dxyz_MSD/Dt
            t=t[:-1]
            t_Dxyz_Dt = np.vstack((t,Dxyz_Dt))
            ScanProject.txyzDMSD_Dt_Poisson_ProbMap[TrajToGet]=t_Dxyz_Dt

    def get_Diffusion(self, ScanProject):
            """
            Calculates the derivative of the data that is orginized as: 
            INPUT:
                ScanProject need to contain ScanProject.txyzMSD_Poisson_ProbMap

            The MSD is assumed to follow the model M[t]=2*d*(s^2 - 2*R*D*dt) + 2*d*D*t
            where:  M is the mean squeare displacment at time lag t
                    t is the time lag
                    dt time step
                    d dimension will be asummed to be 1 since we calculate for each dimension separatly 
                    s localization error 
                    R blur parameter, usign the one givnen by Minflux R= 1/6.2
                    D difussion constant
                    
                    
            """
            ScanProjectMSD = ScanProject.txyzMSD_Poisson_ProbMap
            ScanProject.txyzDMSD_Dt_Poisson_ProbMap={}            
            for TrajToGet in range(len(ScanProjectMSD)):        
                xyz_MSD  = ScanProjectMSD[TrajToGet][1:,:] 
                t=ScanProjectMSD[TrajToGet][0,:]
                dt=np.diff(t)[0]
                R= 1/6.2
                d=1

                fig = plt.figure(constrained_layout=True)
                fig.set_size_inches(12, 3*3, forward=True)
                gs = GridSpec(3, 1, figure=fig)
                ax={}
                ax[0]=fig.add_subplot(gs[0, 0])
                ax[1]=fig.add_subplot(gs[1, 0])
                ax[2]=fig.add_subplot(gs[2, 0])
                #linear regration
                StringAxis=['x','y','z']
                for xyz in range(3):
                
                    L=int(np.floor(np.log2(len(t))))
                    B=np.zeros((2,L))
                    M=np.zeros((2,2,L))

                    color=cm.rainbow(np.linspace(0,1,L))
                    for i in range(L):
                        l=2**(i+1)
                        W=np.diag(np.ones(l))
                        XT=np.ones((2,l))
                        XT[1,:]=t[:l]
                        Y=xyz_MSD[xyz,:l]
                        M[:,:,i]=np.linalg.inv(np.dot(XT,np.dot(W,XT.transpose())))
                        B[:,i]=np.dot(M[:,:,i],np.dot(XT,np.dot(W,Y)))
                    
                        D=B[1,i]/(2*d)
                        s=np.sqrt((B[0,i]/2) + 2*R*D*dt)
                        legendString= 'l:' + str(l) + ' s:' + str(np.round(s*1000, decimals=0)) + 'nm' + ' D:' + str(np.round(D, decimals=3)) + 'um2/s' 
                        ax[xyz].plot(t*1e3,B[0,i]+B[1,i]*t, label=legendString)

                    D=B[1,:]/(2*d)
                    s=np.sqrt((B[0,:]/2) + 2*R*D*dt)

                    ax[xyz].plot(t*1e3,xyz_MSD[xyz,:],'.',label='data')
                    ax[xyz].set_ylim((0,np.max(xyz_MSD[xyz,:])))
                
                    #box = ax[xyz].get_position()
                    #ax[xyz].set_position([box.x0, box.y0, box.width * 0.8, box.height])
                    ax[xyz].legend(loc='center left', bbox_to_anchor=(1, 0.5))
                    ax[xyz].set_ylabel("MSD " + StringAxis[xyz])

                ax[xyz].set_xlabel("Time lag in [ms]")
                do_save_plot=1
                if do_save_plot:
                    file="T_" + str(TrajToGet) +"_MSD_" + '.png'
                    path=ScanProject.Files['AnalysisConfig'][:-18]
                    fig.savefig(path + file)
                    plt.close()
                    print('Done!')
                else:
                    plt.show()

    def get_Diffusion_SumAll(self, ScanProject):
            """
            Calculates the derivative of the data that is orginized as: 
            INPUT:
                ScanProject need to contain ScanProject.txyzMSD_Poisson_ProbMap

            The MSD is assumed to follow the model M[t]=2*d*(s^2 - 2*R*D*dt) + 2*d*D*t
            where:  M is the mean squeare displacment at time lag t
                    t is the time lag
                    dt time step
                    d dimension will be asummed to be 1 since we calculate for each dimension separatly 
                    s localization error 
                    R blur parameter, usign the one givnen by Minflux R= 1/6.2
                    D difussion constant
                    
                    
            """
            ScanProjectMSD = ScanProject.txyzMSD_Poisson_ProbMap
            #ScanProject.StxyzMSD            
            first=1
            for TrajToGet in range(len(ScanProjectMSD)):   
                if first==1:
                    Sxyz_MSD = ScanProjectMSD[TrajToGet][1:,:] 
                    t=ScanProjectMSD[TrajToGet][0,:]
                    N=np.ones(t.shape)
                    first=0
                else:
                    xyz_MSD=ScanProjectMSD[TrajToGet][1:,:]
                    if xyz_MSD.shape[1]>Sxyz_MSD.shape[1]:
                        xyz_MSD[:,:Sxyz_MSD.shape[1]] =Sxyz_MSD + xyz_MSD[:,:Sxyz_MSD.shape[1]] 
                        Sxyz_MSD = xyz_MSD
                        t=ScanProjectMSD[TrajToGet][0,:]
                        N_t=np.ones(t.shape)
                        N_t[:N.shape[0]]=N + N_t[:N.shape[0]] 
                        N=N_t
                    elif xyz_MSD.shape[1]<Sxyz_MSD.shape[1]:
                        Sxyz_MSD[:,:xyz_MSD.shape[1]]  = Sxyz_MSD[:,:xyz_MSD.shape[1]]  + xyz_MSD 
                        N[:xyz_MSD.shape[1]]=N[:xyz_MSD.shape[1]]+1
                    else: 
                        Sxyz_MSD  = Sxyz_MSD  + xyz_MSD 
                        N=N+1

            Sxyz_MSD=Sxyz_MSD/N
            dt=np.diff(t)[0]
            R= 1/6.2
            d=1

            fig = plt.figure(constrained_layout=True)
            fig.set_size_inches(12, 3*3, forward=True)
            gs = GridSpec(3, 1, figure=fig)
            ax={}
            ax[0]=fig.add_subplot(gs[0, 0])
            ax[1]=fig.add_subplot(gs[1, 0])
            ax[2]=fig.add_subplot(gs[2, 0])
            #linear regration
            StringAxis=['x','y','z']
            for xyz in range(3):
                
                L=int(np.floor(np.log2(len(t))))
                B=np.zeros((2,L))
                M=np.zeros((2,2,L))

                color=cm.rainbow(np.linspace(0,1,L))
                for i in range(5,12): #range(L):
                    l=2**(i+1)
                    W=np.diag(np.ones(l))
                    XT=np.ones((2,l))
                    XT[1,:]=t[:l]
                    Y=Sxyz_MSD[xyz,:l]
                    M[:,:,i]=np.linalg.inv(np.dot(XT,np.dot(W,XT.transpose())))
                    B[:,i]=np.dot(M[:,:,i],np.dot(XT,np.dot(W,Y)))
                    
                    D=B[1,i]/(2*d)
                    s=np.sqrt((B[0,i]/2) + 2*R*D*dt)
                    legendString= 'l:' + str(l) + ' s:' + str(np.round(s*1000, decimals=0)) + 'nm' + ' D:' + str(np.round(D, decimals=3)) + 'um2/s' 
                    ax[xyz].plot(t*1e3,B[0,i]+B[1,i]*t, label=legendString)

                D=B[1,:]/(2*d)
                s=np.sqrt((B[0,:]/2) + 2*R*D*dt)

                ax[xyz].plot(t*1e3,Sxyz_MSD[xyz,:],'.',label='data')
                ax[xyz].set_ylim((0,np.max(Sxyz_MSD[xyz,:])))
                
                #box = ax[xyz].get_position()
                #ax[xyz].set_position([box.x0, box.y0, box.width * 0.8, box.height])
                ax[xyz].legend(loc='center left', bbox_to_anchor=(1, 0.5))
                ax[xyz].set_ylabel("MSD " + StringAxis[xyz])

            ax[xyz].set_xlabel("Time lag in [ms]")
            do_save_plot=1
            if do_save_plot:
                file="T_MSD_SumAll" + '.png'
                path=ScanProject.Files['AnalysisConfig'][:-18]
                fig.savefig(path + file)
                plt.close()
                print('Done!')
            else:
                plt.show()

    def get_test_MSD(self, ScanProject,TrajDictionary):
        """
        Go through all poisson trajectory estimations and calculates the MSD
        Note: This might be good to make generic, such that on can input what to estimate. 
        """
        ScanProject.txyzMSD_Poisson_ProbMap={}
        for TrajToGet in [4]:#range(len(TrajDictionary)):        
            txyzn_Map  = TrajDictionary[TrajToGet]
            traj=txyzn_Map[:4,:]
            L=5
            color=cm.rainbow(np.linspace(0,1,L))
            color[:,3]=0.5

            fig = plt.figure(constrained_layout=True)
            fig.set_size_inches(12, 3*3, forward=True)
            gs = GridSpec(3, 1, figure=fig)
            ax={}
            ax[0]=fig.add_subplot(gs[0, 0])
            ax[1]=fig.add_subplot(gs[1, 0])
            ax[2]=fig.add_subplot(gs[2, 0])
            #linear regration
            StringAxis=['x','y','z']
            for xyz in range(3):
                for i in range(2,L):
                    j=5*i
                    t = (traj[0,j:]-traj[0,0])*1e3 #[ms]
                    dt  = traj[0,j:] - traj[0,:-j]
                    MSD = ((traj[1:,j:] - traj[1:,:-j])**2)/(2*dt)

                    wndw=20
                    wndwS=1 #if 1 it is a moving average if equal to wndw it is a fix window average
                    T= np.convolve(np.ones(wndw)/wndw,t)
                    T=T[wndw-1:-wndw+1:wndwS]
                    X = np.convolve(np.ones(wndw)/wndw,MSD[xyz,:])
                    X=X[wndw-1:-wndw+1:wndwS]

                    ax[xyz].plot(T,X,label=str(j))
                    #plt.hist(X,bins=20, color=color[i,:])
                ax[xyz].set_ylabel("D [um2/s]" + StringAxis[xyz])
            plt.legend()
            ax[2].set_xlabel("t [ms]")

            do_save_plot=1
            if do_save_plot:
                file="T_" + str(TrajToGet) +"_MSD_traj_" + '.png'
                path=ScanProject.Files['AnalysisConfig'][:-18]
                fig.savefig(path + file)
                plt.close()
                print('Done!')
            else:
                plt.show()
            print("adsf")

    def get_MSD_traj(self, ScanProject,TrajDictionary,wndw=5,j=4,n=1,Flag=2,MFlag=1):

        R=self._Ph.get_R(ScanProject=ScanProject)

        if 0: #CVE and ECVE
            resCVE=self._Ph.get_CVE_MSD_traj(TrajDictionary,wndw=wndw,R=R,j=j,n=n,Flag=Flag)#wndw=5,R=1/6.2,j=4,n=1,Flag=2; wndw=wndw,R=R,j=4,n=1,Flag=2
            ScanProject.txyzCVEMSD_traj=resCVE.copy()
        else: #MSD 
            resMSD=self._Ph.get_MSD_traj2(ScanProject=ScanProject,TrajDictionary=TrajDictionary,window=wndw,MSDStart=1,MSDSteps=wndw-1,PointsToUse=wndw-1,R=R,Flag=0)
            ScanProject.txyzCVEMSD_traj=resMSD.copy()
        ## Do a linear regreation over the first points of the ECVE method when using the first offset diagonal (n=1)
        if 0:
            j=wndw;n=1;Flag=2
            result_txyz=self._Ph.get_CVE_traj_Special(TrajDictionary,wndw=1,R=R,j=j,n=n,Flag=Flag)
            ScanProject.txyzCVEMSD_traj=result_txyz.copy()

        Dwndw=wndw
        if MFlag==0:
            j=1;n=1;Flag=0
            result_txyz=self._Ph.get_CVE_MSD_traj(TrajDictionary,wndw=Dwndw,R=R,j=j,n=n,Flag=Flag)
            Name_prefix='CVE'
        elif MFlag==1:
            j=int(Dwndw/2);n=1;Flag=2
            result_txyz=self._Ph.get_CVE_MSD_traj(TrajDictionary,wndw=Dwndw,R=R,j=j,n=n,Flag=Flag)
            Name_prefix='ECVE'
        elif MFlag==2:
            j=Dwndw;n=1;Flag=2; jjStart=2
            result_txyz=self._Ph.get_CVE_traj_Special(TrajDictionary,wndw=1,R=R,j=j,n=n,jjStart=jjStart,Flag=Flag)
            Name_prefix='ECVE_lineFitt'
        elif MFlag==3:
            j=Dwndw;n=1;Flag=2
            result_txyz=self._Ph.get_CVE_traj_Specialv0(TrajDictionary,wndw=1,R=R,j=j,n=n,Flag=Flag)
            Name_prefix='ECVE_lineFitt0'
        elif MFlag==4:
            result_txyz=self._Ph.get_MSD_traj2(ScanProject=None,TrajDictionary=TrajDictionary,window=Dwndw,MSDStart=1,MSDSteps=Dwndw-1,PointsToUse=Dwndw-1,R=R,Flag=0)
            Name_prefix='MSD'
        ScanProject.txyzCVEMSD_traj=result_txyz.copy()

    def get_TrajAveraging(self, ScanProject, TrajDictionary,wndw=10,wndwS=10):

        #wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
        TrajDictionaryNew={}
        for i in range(len(TrajDictionary)):
            txyznM=TrajDictionary[i]
            for j in range(txyznM.shape[0]):
                T=np.convolve(np.ones(wndw)/wndw,txyznM[j,:])[wndw-1:-wndw+1:wndwS]
                if j==0:
                    txyznMNew=T
                else:
                    txyznMNew=np.vstack((txyznMNew,T))
            TrajDictionaryNew[i]=txyznMNew
        return TrajDictionaryNew
        
    def get_traj_filter(self, ScanProject, TrajDictionary, Flag=True ,PhotonThreachhold=10,wndw=10,MaxLength=np.inf):
        """
        Filter the trajectory data according to:
            1. Photon count on moving average of size given by wndw
            2. Flag prameter that indicate 0 for not include point and 1 to include, of one a tuple with filters the we multiply each dictionary 
            3. Trajectory length, it longer then MaxLength then the trajectory will be splited into several parts of max length given by: MaxLength
               The rest part is always at the end of the trajectory
        """

        
        TrajDictionary_New={}
        newtraj=0
        for i in range(len(TrajDictionary)):
            try:
                if type(Flag) is tuple:
                    for j in range(len(Flag)):
                        if j==0:
                            F=Flag[j][i]
                        else:
                            F=F*Flag[j][i]
                else:
                    F=Flag[i]
            except:
                print('No flag found for filter')
                F=Flag

            txyznM=TrajDictionary[i]*1.0
            n=txyznM[4,:]
            t=txyznM[0,:]
            Pass=np.ones(n.shape)
            
            #Photon count threshold over a moving average
            wndwS=1 #if 1 it is a moving average, if equal to wndw it is a fix window average
            N_Mean= np.convolve(np.ones(wndw)/wndw,n)
            N_Mean=N_Mean[wndw-1:-wndw+1:wndwS]
            Pass[wndw-1:]=N_Mean>PhotonThreachhold

            #Filter flag threchhold
            Pass=Pass*F

            Pass=np.concatenate(([0],Pass))
            if Pass[-1]==1:
                Pass=np.concatenate((Pass,[0]))

            Starts=np.where(np.diff(Pass)>0)[0]
            Stops=np.where(np.diff(Pass)<0)[0]-1

            if len(Starts)!=len(Stops):
                print('Traj Start Stop are of different length in get_traj_filter')
            
            TrajDuration=(t[Stops]-t[Starts])
            Segments=np.floor(TrajDuration/MaxLength)
            StartsN=np.array([])
            StopsN=np.array([])
            for subTraj in range(len(TrajDuration)):
                if Segments[subTraj]>0:
                    SegStep=np.floor((Stops[subTraj]-Starts[subTraj])/Segments[subTraj])
                    StartsN= np.append(StartsN, Starts[subTraj]+(np.arange(Segments[subTraj]))*SegStep)
                    StopsN= np.append(StopsN,np.append((Starts[subTraj]+(np.arange(Segments[subTraj]))*SegStep-1)[1:],[Stops[subTraj]]))
                else:
                    StartsN=np.append(StartsN,Starts[subTraj])
                    StopsN= np.append(StopsN,Stops[subTraj])

            Starts=StartsN.astype('int')
            Stops=StopsN.astype('int')
            
            Intervall=np.where((t[Stops]-t[Starts])>ScanProject.TrajectorySelectionRules['MinTrajectoryTimeLength'])[0]
            for inter in Intervall:
                TrajDictionary_New[newtraj]=txyznM[:,Starts[inter]:Stops[inter]]
                newtraj+=1
                
        return TrajDictionary_New

    def get_traj_HMM_PhotonFilter(self, ScanProject, TrajDictionary, ScanMatrixTypFilterVector):
        """
        """
        FilterPass={}
        for i in range(len(TrajDictionary)):
            wndw= int(ScanProject.TrajectorySelectionRules['wndw'])
            txyznM=TrajDictionary[i]*1.0
            t=txyznM[0,:]
            StartStop=np.where((ScanProject.ScanMatrix[0,:]>=t[0])*(ScanProject.ScanMatrix[0,:]<=t[-1]))[0][[0,-1]]
            StartStop[0]=StartStop[0]-wndw #time is defined as the last time point in each "wndw" time slot
            Filter=ScanMatrixTypFilterVector[StartStop[0]:StartStop[1]]
            FilterPass[i]=np.round(np.convolve(np.ones(wndw)/wndw,Filter)[wndw-1:-wndw+1:wndw])

        return FilterPass

    def get_traj_MLE_Filter(self, ScanProject, TrajDictionary):
        """
        Creates a filter vector for each trajectory,
        we sheck that the mean is smaler then 500nm and that the std is bigger then 15nm for each trajectory
        if not it is very likely that the trajectory is at the boundery of the MLE bounding box and the trajectory estimation is compremized 
        """
        FilterPass={}
        for i in range(len(TrajDictionary)):
            wndw= int(ScanProject.TrajectorySelectionRules['wndw'])
            txyznM=TrajDictionary[i]*1.0
            t=txyznM[0,:]
            StartEndPos=np.expand_dims(np.append([0],np.where((ScanProject.ScanMatrix[0,:]>=t[0])*(ScanProject.ScanMatrix[0,:]<=t[-1]))[0][[0,-1]]),axis=1)

            T_Raw,XP_Raw,YP_Raw,XE_Raw,YE_Raw,XO_Raw,YO_Raw,ZI_Raw,APD1_Raw,APD2_Raw,J_Raw,MemIndex_Raw,ZI_um_Raw,ZI_OffsetIndex_Raw, zCam_Raw =  _MS.get_traj_raw_data4(StartEndPos, 0, ScanProject.ScanMatrix, ScanProject)
            txyz_Traj_Piezo=np.vstack((T_Raw,XP_Raw,YP_Raw,ZI_um_Raw))
            txyz_Traj_Piezo=_MS.get_Scan_toImagCoordinates(ScanProject, txyz_Traj_Piezo)

            Xtest=(np.mean(np.abs(txyz_Traj_Piezo[1,::wndw]-txyznM[1,:]))<0.5)*(np.std(txyz_Traj_Piezo[1,::wndw]-txyznM[1,:])>0.015)
            Ytest=(np.mean(np.abs(txyz_Traj_Piezo[2,::wndw]-txyznM[2,:]))<0.5)*(np.std(txyz_Traj_Piezo[2,::wndw]-txyznM[2,:])>0.015)
            Ztest=(np.mean(np.abs(txyz_Traj_Piezo[3,::wndw]-txyznM[3,:]))<0.5)*(np.std(txyz_Traj_Piezo[3,::wndw]-txyznM[3,:])>0.015)
            if (Xtest+Ytest+Ztest)>0:
                FilterPass[i]=np.ones(t.shape)
            else:
                FilterPass[i]=np.zeros(t.shape)

        return FilterPass

    def get_traj_toImagCoordinates(self, ScanProject, TrajDictionary):
        """
        Convert the trajectories to image coordinates
        This is using the affine transformation in: ScanProject.ImgOverlayTransform
        If not found it will use the Identity matrix
        Some info on Affine transfomration can be found here: https://en.wikipedia.org/wiki/Affine_transformation 
        
        TrajDictionary is assumed to be a dictionary with each element formated as a numpy array with  t,x,y,z,n  
        """
        CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]
        
        
        try:
            Tran=ScanProject.ImgOverlayTransform.copy()
        except:
            print('No Scan to Cam transformation found, identity transform will be used')
            Tran=np.eye(3)

        #Get the transformation and get the Affine transform invers
        Tran[0:2,0:2]=np.linalg.inv(Tran[0:2,0:2])
        Tran[0:2,2]=-np.matmul(Tran[0:2,0:2],Tran[0:2,2])
        
        Traj_txyzn_CamRefFrame_um={}
        for i in range(len(TrajDictionary)):
            txyznM=TrajDictionary[i]*1.0
            Traj_xyz=txyznM[1:4,:]/CamRes #Extract x,y,z and move to pixel coordinates 
            Traj_xyz[2,:]=1               #Remove z and change it to a 1 for the augmented Affine transromation matrix
            Traj_xyz= np.matmul(Tran,Traj_xyz) #Apply transformation 
            Traj_xyz= Traj_xyz*CamRes  #[um] go back to micrometers

            txyznM[1:3,:]=Traj_xyz[:2,:]  #Combine with time and z and photon counts
            Traj_txyzn_CamRefFrame_um[i]=txyznM   
            
            
        return Traj_txyzn_CamRefFrame_um

    def get_traj_inCells(self, ScanProject, TrajDictionary,mask='IncludeInMask'):

        ScanProject=_DS.get_cellMask(ScanProject)
        CellMask=ScanProject.CamCellMask[:,:,0]>0

        CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]

        CellMaskTrajFlag={}
        for i in range(len(TrajDictionary)):
            txyznM=TrajDictionary[i]
            TrajCam_xy=txyznM[1:3,:]/CamRes
            Traj_xypixCoord=np.round(TrajCam_xy).astype('int')
            Traj_yxpixCoord=np.flip(Traj_xypixCoord,axis=0)
            if mask=='IncludeInMask':
                CellMaskTrajFlag[i]=CellMask[tuple(Traj_yxpixCoord)]
            elif mask=='NotIncludeInMask':
                CellMaskTrajFlag[i]=(CellMask<1)[tuple(Traj_yxpixCoord)]
            elif mask=='IncludeAll':
                CellMaskTrajFlag[i]=CellMask[tuple(Traj_yxpixCoord)]
                CellMaskTrajFlag[i][:]=True
            elif mask=='IncludePartialInMask':
                P_bigger=0 #procentage that can be outside of mask
                CellMaskTrajFlag[i]=CellMask[tuple(Traj_yxpixCoord)]
                if np.sum(CellMaskTrajFlag[i]>0)/len(CellMaskTrajFlag[i])>P_bigger:
                    CellMaskTrajFlag[i][:]=True
                else:
                    CellMaskTrajFlag[i][:]=False

            if 0:
                plt.imshow(CellMask);
                plt.plot(Traj_xypixCoord[0,:],Traj_xypixCoord[1,:],'-');
                color=(CellMaskTrajFlag[i]<1)*1.0 #0=blue(include), 1=red(reject)
                plt.scatter(np.append(Traj_xypixCoord[0,0],Traj_xypixCoord[0,:]),np.append(Traj_xypixCoord[1,0],Traj_xypixCoord[1,:]),c=np.append(np.array([0.5]),color),cmap='bwr')
                plt.show()
                print('!')

        return CellMaskTrajFlag

    def get_MSD_Distribution(self, ScanProject,TrajDictionary):
        res=self._Ph.get_MSD_Distribution(TrajDictionary)
        #ScanProject.txyzMSD_traj=res
        return res
          
    def get_ini_file(self, ScanProject):
        ########################################################################################################################
        ############# Default Parameters the Shadowfax system    ###############################################################
        ########################################################################################################################               
        
        Project_config = ConfigObject(ScanProject.Files['AnalysisConfig'])
        ScanProject.PosCalibration = Project_config.PosCalibration
        
        ########################################################################################################################
        #############   Default Parameter for trajectory building and plotting    ###############################################
        ########################################################################################################################
        ScanProject.DataSelectionRules = Project_config.DataSelectionRules   
        ScanProject.TrajectorySelectionRules = Project_config.TrajectorySelectionRules  
        ScanProject.RegistrationCamToScan    = Project_config.RegistrationCamToScan  

        ########################################################################################################################
        #############   Default Parameter for microscopy Setup            ######################################################
        ########################################################################################################################

        # This should be moved either to the ini file for the phase retrieval or to normal ini file
        # May be better to have it in the normal ini file since these are setup specific 
        ScanProject.SetupSpecifics = Project_config.SetupSpecifics  
        
        # Remove these when cleaning out PFS_Parameters dependencies
        ScanProject.PFS_Parameters={}
        s=(0.822/2)*(ScanProject.SetupSpecifics['Lambda']/ScanProject.SetupSpecifics['Na'])*1e6  #[um] theoretical PSF width, the radius of the Airy disk is scaled to 1.22/2 and for the e-2 intensity drop that is std value one use 0.822/2
        ScanProject.PFS_Parameters['sigma'] =[s,s,2*s]             #[um] [x y z]
        
        ########################################################################################################################
        #############   Default Parameter for numerical PSF               ######################################################
        ########################################################################################################################
        ScanProject.Ph_Retrieval = Project_config.Ph_Retrieval
        ########################################################################################################################
        #############   Default Parameter for zCam                        ######################################################
        ########################################################################################################################
        ScanProject.zCam = Project_config.zCam
        ########################################################################################################################
        #############   Default Parameter for simulation                  ######################################################
        ########################################################################################################################
        is_simulation= ScanProject.SetupSpecifics['is_simulation']
        if is_simulation:
            ScanProject.loadSaveData=1
            ScanProject.Sim_par = Project_config.Sim_par

    def get_ovelayTrajOnCells(self, ScanProject,iterations_to_do=5,do_save_plot=1):
        """
            Do a registration of the EMCCD and the Scan imgae,

        """
        #Get EMCCD data
        file="widefield_cells_MMStack_Default.ome.tif"
        path=ScanProject.Files['AnalysisConfig'][:-18] + "widefield/"
        dataset = Image.open(path + file)
        h,w = np.shape(dataset)
        if dataset.n_frames>2:
            print("Widefiled image stack contains more then 2 images, will only use the 2 fist")
            print("Assuming that first is widefiled and second is EPI phluoresent image")
            
        tiffarray = np.zeros((h,w,2))
        for i in range(2):
            dataset.seek(i)
            tiffarray[:,:,i] = np.array(dataset)
            tiffarray[:,:,i]=tiffarray[:,:,i]/np.max(tiffarray[:,:,i])
        
        #tiffarray[:,:,0][tiffarray[:,:,0]>0.87]=1
        #Get Scan data
        y=ScanProject.ScanMatrix[1,:]
        x=ScanProject.ScanMatrix[2,:]
        I=ScanProject.ScanMatrix[3,:]+ScanProject.ScanMatrix[4,:]

            #Convert scan to image
        range_um=ScanProject.PosCalibration['PositionLength']  #[um]
        res=range_um/ScanProject.PosCalibration['ScanMemsize'] #[um/pix]
        
        points=np.array(range(int(range_um/res)+1))*res
        ScanImg,xedges,yedges=np.histogram2d(x, y,bins=points, weights=I)
            #Reshape the scan to the same size as the EMCCD image resolution
        CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]
        ScaleFactor=res/CamRes
        ScanImg=_ski_tran.rescale(ScanImg, ScaleFactor)
            #Cap the intensity, 
        ScanImg[ScanImg>300]=300       
            #Pad the image to be of the same format as the EMCCD image
        ScanImg_big=np.zeros(tiffarray[:,:,0].shape)
        ScanImg_big[:ScanImg.shape[0],:ScanImg.shape[1]]=ScanImg

        # First coarse estimation  
        # The used version is one where we have done a fix estimation 
        # Tested is also a templet matching version but it needs some works
        tform = _ski_tran.AffineTransform()
                 
        angle=ScanProject.RegistrationCamToScan['RotAngle']*(2*np.pi/360)
        rot=[[np.cos(angle), -np.sin(angle)],[np.sin(angle), np.cos(angle)]]

        translate=ScanProject.RegistrationCamToScan['Trans']

        dst=ScanProject.RegistrationCamToScan['Dest']    #scan image
        src=ScanProject.RegistrationCamToScan['Traget']  #camera EMCCD image
        dst=np.matmul(dst,rot)+translate

        tform.estimate(src, dst)
        ScanImg_tranFirst=_ski_tran.warp(ScanImg_big, tform) 
        #EstimationTemp = match_template(tiffarray[:,:,1],ScanImgTemp,pad_input=True) 
        #ij = np.unravel_index(np.argmax(EstimationTemp), EstimationTemp.shape)
        #tform = _ski_tran.AffineTransform(translation=(-xFirstEstimation+ScanImg.shape[0]/2, -yFirstEstimation+ScanImg.shape[1]/2))
        #ScanImg_tran_iteration=_ski_tran.warp(ScanImg_big, tform)

        # iteration for fin tuning
        # After the coarse adjustment we do a iterative fine adjustment where:
        # in the fluorecent channel and in the scan channel we detect blobs 
        # The two list of blob koordinates are matched by shortest eukliden distance
        # The short list of matching points are then used in the affine transformation
        # The procedure is repited a few times to fine tune the transformation   
        ScanImg_tran_iteration=ScanImg_tranFirst
        MTransform_iteration=tform.params

        if 0:# Debuging plot showing the bright field image, fluorecent, and the new scan image 
            amount=3
            fig = plt.figure(constrained_layout=True)
            fig.set_size_inches(amount*5, 5, forward=True)
            gs = GridSpec(1, amount, figure=fig)
            ax={}
            for i in range(amount):
                ax[i]=fig.add_subplot(gs[0,i])
        
                #Combined image
            ax[0].imshow(tiffarray[:,:,0],cmap='gray',alpha=1.0)
            ax[0].imshow(tiffarray[:,:,1],cmap='jet',alpha=0.3)
            ax[0].plot(src[:,0],src[:,1],'r*')
                #Scan image
            ax[1].imshow(ScanImg_big,cmap='jet',alpha=1.0)
            ax[1].plot(dst[:,0],dst[:,1],'g*')

                #Combined image
            ax[2].imshow(tiffarray[:,:,0],cmap='gray',alpha=1.0)
            #ax[2].imshow(tiffarray[:,:,1],cmap='jet',alpha=0.3)
            ax[2].imshow(ScanImg_tran_iteration,cmap='Reds',alpha=0.3)
            plt.show()

        if 0:# plot with interactive 
            global TRANSLATE
            TRANSLATE=translate
            def on_press(event,TRANSLATE):
                print('press', event.key)
                sys.stdout.flush()
                if event.key == 'x':
                    TRANSLATE+= np.array([1,0])
                    dst_new=np.matmul(dst,rot)+TRANSLATE
                    print(str(TRANSLATE))
                    tform.estimate(src, dst_new)
                    ScanImg_tran_iteration=_ski_tran.warp(ScanImg_big, tform) 
                    ax[2].clear()
                    ax[2].imshow(tiffarray[:,:,0],cmap='gray',alpha=1.0)
                    ax[2].imshow(ScanImg_tran_iteration,cmap='Reds',alpha=0.3)
                    fig.canvas.draw()
                if event.key == 'c':
                    TRANSLATE+= np.array([-1,0])
                    dst_new=np.matmul(dst,rot)+TRANSLATE
                    print(str(TRANSLATE))
                    tform.estimate(src, dst_new)
                    ScanImg_tran_iteration=_ski_tran.warp(ScanImg_big, tform) 
                    ax[2].clear()
                    ax[2].imshow(tiffarray[:,:,0],cmap='gray',alpha=1.0)
                    ax[2].imshow(ScanImg_tran_iteration,cmap='Reds',alpha=0.3)
                    fig.canvas.draw()
                if event.key == 'a':
                    TRANSLATE+= np.array([0,1])
                    dst_new=np.matmul(dst,rot)+TRANSLATE
                    print(str(TRANSLATE))
                    tform.estimate(src, dst_new)
                    ScanImg_tran_iteration=_ski_tran.warp(ScanImg_big, tform) 
                    ax[2].clear()
                    ax[2].imshow(tiffarray[:,:,0],cmap='gray',alpha=1.0)
                    ax[2].imshow(ScanImg_tran_iteration,cmap='Reds',alpha=0.3)
                    fig.canvas.draw()
                if event.key == 'z':
                    TRANSLATE+= np.array([0,-1])
                    dst_new=np.matmul(dst,rot)+TRANSLATE
                    print(str(TRANSLATE))
                    tform.estimate(src, dst_new)
                    ScanImg_tran_iteration=_ski_tran.warp(ScanImg_big, tform) 
                    ax[2].clear()
                    ax[2].imshow(tiffarray[:,:,0],cmap='gray',alpha=1.0)
                    ax[2].imshow(ScanImg_tran_iteration,cmap='Reds',alpha=0.3)
                    fig.canvas.draw()

            amount=3
            fig = plt.figure(constrained_layout=True)
            fig.set_size_inches(amount*5, 5, forward=True)
            fig.canvas.mpl_connect('key_press_event', lambda event: on_press(event,translate))

            gs = GridSpec(1, amount, figure=fig)
            ax={}
            for i in range(amount):
                ax[i]=fig.add_subplot(gs[0,i])
        
                #Combined image
            ax[0].imshow(tiffarray[:,:,0],cmap='gray',alpha=1.0)
            ax[0].imshow(tiffarray[:,:,1],cmap='jet',alpha=0.3)
            ax[0].plot(src[:,0],src[:,1],'r*')
                #Scan image
            ax[1].imshow(ScanImg_big,cmap='jet',alpha=1.0)
            ax[1].plot(dst[:,0],dst[:,1],'g*')

                #Combined image
            ax[2].imshow(tiffarray[:,:,0],cmap='gray',alpha=1.0)
            #ax[2].imshow(tiffarray[:,:,1],cmap='jet',alpha=0.3)
            ax[2].imshow(ScanImg_tran_iteration,cmap='Reds',alpha=0.3)
            plt.show()

        #iterations_to_do=5
        if iterations_to_do>0:
            for i in range(iterations_to_do):
          
                #Feature detector
                    # find blobs in fluorecent EMCCD image and in scan
                blobs_EMCCD = blob_log(tiffarray[:,:,1], max_sigma=4, min_sigma=2,threshold=.05)
                blobs_EMCCD[:, 2] = blobs_EMCCD[:, 2] * np.sqrt(2)
                keypoints1=blobs_EMCCD[:, 0:2].astype('int')

                blobs_Scan  = blob_log(ScanImg_tran_iteration, max_sigma=3, min_sigma=2, threshold=20)
                blobs_Scan[:, 2] = blobs_Scan[:, 2] * np.sqrt(2)
                keypoints2=blobs_Scan[:, 0:2].astype('int')

                #Look at distances between the EMCCD and Scan blobs and find which of them are closest  
                WhichisSmalest=np.argmin((keypoints1.shape[0], keypoints2.shape[0]))
                if WhichisSmalest==0:
                    matches12=np.zeros((keypoints1.shape[0],3))
                    L=keypoints1.shape[0]
                    C1=keypoints2
                    C2=keypoints1
                else:
                    matches12=np.zeros((keypoints2.shape[0],3))
                    L=keypoints2.shape[0]
                    C2=keypoints2
                    C1=keypoints1

                for i in range(L): #Range test bewtween EMCCD and Scan
                    dis= (C1[:,0]-C2[i,0])**2 + (C1[:,1]-C2[i,1])**2
                    if WhichisSmalest==0:
                        matches12[i,:]=[i,np.argmin(dis),np.min(dis)]
                    else:
                        matches12[i,:]=[np.argmin(dis),i,np.min(dis)]

                    #sort according to distance
                matches12=matches12[np.argsort(matches12[:,2]),:]
                    #Select only the ones closest
                maxDis=10**2
                matches12=matches12[: np.where(matches12[:,2]<maxDis)[0][-1],:]

                if 0: #Show which are selected 
                    fig, ax = plt.subplots(nrows=1, ncols=1)
                    plot_matches(ax, tiffarray[:,:,1]*100, ScanImg_tran_iteration, keypoints1, keypoints2, matches12[:,0:2].astype('int'))
                    ax.axis('off')
                    plt.show()


                # Fine tune the affine transformation between Scan and EMCCD 
                # Use the blob matches 
                print('Matched points found: ' + str(matches12.shape[0]))
                tformFine = _ski_tran.AffineTransform()
                src=np.flip(keypoints1[matches12[:,0].astype('int'),:],axis=1) #camera image
                dst=np.flip(keypoints2[matches12[:,1].astype('int'),:],axis=1) #scan image
                tformFine.estimate(src,dst)

                MTransform_iteration=np.matmul(MTransform_iteration,tformFine.params) #It is here the order Old*New since it is the invers transform (Y=A2*A1*X but one get the invers of A (A') from the transform thus X=A1'*A2'*Y is what we have)
                ScanImg_tran_iteration_old=ScanImg_tran_iteration
                ScanImg_tran_iteration = _ski_tran.warp(ScanImg_big, MTransform_iteration) #Do full transform, obs the transformation need to be the invers 
                #ScanImg_tran_iteration = _ski_tran.warp(ScanImg_tran_iteration, tformFine.params) #Do increment
        if 0:
            plt.imshow(tiffarray[:,:,0],cmap='gray',alpha=1.0);
            plt.imshow(tiffarray[:,:,1],cmap='jet',alpha=0.3);
            plt.imshow(ScanImg_tran_iteration,cmap='Reds',alpha=0.3);
            plt.show()


        ##################################################################################

        ScanImg_tran = ScanImg_tran_iteration
        MTransform = MTransform_iteration

        ImTot=np.append(tiffarray,np.expand_dims(ScanImg_tran,2),2)
        ScanProject.ImgOverlay=ImTot
        ScanProject.ImgOverlayTransform=MTransform

        # Show/save result 
        if do_save_plot:
            amount=4
            fig = plt.figure(constrained_layout=True)
            fig.set_size_inches(amount*5, 5, forward=True)
            gs = GridSpec(1, amount, figure=fig)
            ax={}
            for i in range(amount):
                ax[i]=fig.add_subplot(gs[0,i])
        
                #EMCCD image
            ax[0].imshow(ImTot[:,:,0],cmap='gray',alpha=1.0)
            ax[0].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)
            if iterations_to_do>0:
                for blobi in range(blobs_EMCCD.shape[0]):
                    yb,xb,rb=blobs_EMCCD[blobi,:]
                    if len(np.where(matches12[:,0]==blobi)[0])>0:
                        Color='g'
                        ax[0].text(xb, yb,str(np.where(matches12[:,0]==blobi)[0][0]))
                    else: 
                        Color='r'
                    c=plt.Circle((xb, yb), rb,color=Color, linewidth=2, fill=False)
                    ax[0].add_patch(c)
                ax[0].set_title('Last iteration, bright field')

                        #Scan image
                ax[1].imshow(ScanImg_tran_iteration_old,cmap='jet',alpha=1.0)
                for blobi in range(blobs_Scan.shape[0]):
                    yb,xb,rb=blobs_Scan[blobi,:]
                    if len(np.where(matches12[:,1]==blobi)[0])>0:
                        Color='g'
                        ax[1].text(xb, yb,str(np.where(matches12[:,1]==blobi)[0][0]))
                    else: 
                        Color='r'
                    c=plt.Circle((xb, yb), rb,color=Color, linewidth=2, fill=False)
                    ax[1].add_patch(c)
                    #ax[1].plot(xb, yb,'r*')
                ax[1].set_title('Last iteration, scan')
            else: 
                ax[1].imshow(ScanImg_big,cmap='jet',alpha=1.0)
                ax[1].set_title('Scan image, scan')

                #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
            ax[2].imshow(ImTot[:,:,0],cmap='gray',alpha=1.0)
            ax[2].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)
            ax[2].imshow(ScanImg_tranFirst,cmap='Blues',alpha=0.3)
            ax[2].set_title('First coarse registration')

            ax[3].imshow(ImTot[:,:,0],cmap='gray',alpha=1.0) #Brighfield
            ax[3].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)  #Fluorecent
            ax[3].imshow(ImTot[:,:,2],cmap='Blues',alpha=0.3)#Scan
            #ax[3].imshow(ScanImg_tests,cmap='Blues',alpha=0.3)
            ax[3].set_title('Registration result')

            do_save=1
            if do_save:
                file="Overlay_" + '.png'
                path=ScanProject.Files['AnalysisConfig'][:-18]
                fig.savefig(path + file)
                plt.close()
                print('Done!')
            else:
                plt.show()


        return ScanProject

    def GetzCam_data(self, ScanProject):
        print("Getting zCam data")
        try:
            path_ref=ScanProject.zCam['path_ref']          
            file_prefix=ScanProject.zCam['file_prefix_ref']  
            N=int(ScanProject.zCam['N'])
            StepSize=ScanProject.zCam['zssize'] 
            scale=ScanProject.zCam['Scaling']
            
            zpix2um=(StepSize*1e6 /scale) #um/y-pix
            Zmap=_zCam.getZref(path_ref,file_prefix,N,scale=scale)
            
            path=ScanProject.Files['AnalysisConfig'][:-19]+ScanProject.zCam['path_tracking']    
            file_prefix=ScanProject.zCam['file_prefix']      
            ZCam_um=_zCam.getZum(Zmap,zpix2um,path,file_prefix)
            ScanProject=_zCam.addToScanProject(ScanProject,ZCam_um)
        except:
            print('zCam settings or files are not exicting')
            print('No zCam data will be added to data')
        

            

        

