import numpy as np
import pomegranate as _PG
import scipy as _sc
import scipy.stats as stats    

import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import seaborn as sn

class HMM_Traj():
    """
    Library containing HMM state estimations 
    """
    def init_hmm_model(D,M,trans_mat,starts=None):
        """
        Create an HMM, using invers-Gamma distribution for simulating the observed diffusion 
        The HMM works on the Gamma distribution thus we convert the data X to 1/X to get a gamma distribution.
        To define the iverse-Gamma function we use the mean (beta/(alpha-1)) and mode (beta/(alpha+1)) and solve for alpha and beta 

        input:
            D:          kx1 array of mean diffusion constants, where k is the amount of states
            M:          kx1 array of the mode (where the distribution peak) diffusion constants is, where k is the amount of states
            trans_mat:  kxk transition matrix with elements a_ij describing probability to move between state i to j
                            rows in the matrix must sum to 1 (total probability to move to any state is 1)
            start:      kx1 array describing the probability to start at a specific state, this need to sum to 1.
                            if not specified it is set to eaqual probability

        output:
            model:      pomegrante object describing the model
        """
        ###Create a model
        #D=np.array([0.05,2]) #mean in an inverse gamma distribution
        #M=np.array([0.01,1]) #Mode in an inverse gamma distribution
        #trans_mat = numpy.array([[0.3, 0.7],
        #                         [0.2, 0.8]])
        #starts = np.array([0.5, 0.5])

        if starts==None:
            starts = np.ones(D.shape)
            starts = starts/np.sum(starts)

        alphas=(D+M)/(D-M) #Using the relations Mean=beta/(alpha-1), and Mode=beta/(alpha+1) for a invers-Gamma
        betas=(2*D*M)/(D-M) 

        if len(np.where(alphas<=1)[0])!=0:
            print('Alpha value out of range: ' + str(alphas))

        #alpha and beta are now the two parameters defining our invers-Gamma distribution, 
        #but in the HMM we will use Gamma distribution and relate the invers-Gamma through 1/x
        #One can show that if X~Gamma(a,b) then 1/X~inv-Gamma(a,b) given the PDF from in wiki
        #see:  https://en.wikipedia.org/wiki/Inverse-gamma_distribution
        dists=()
        for alpha,beta in zip(alphas,betas):
            dists += (_PG.distributions.GammaDistribution(alpha, beta),)
    
        model = _PG.HiddenMarkovModel.from_matrix(trans_mat, dists, starts)
        model.bake()
        #print(model.dense_transition_matrix())
        return model

    def get_sample_sequences(model, n_sequences=1, length_limits=[100,1000]):
        """
        Given a model generate sequences of diffusion constants
        The length of each sequence is randomly selected
        input:
            model:          pomegrante hmm model
            n_trajectories: amount of sequences to produce 
            length_limits:  the length of the sequence will be randomly selected with in the bounds      
        """
                #Generate Samples
        samples_all=()
        SP_all=()
        for i in range(n_sequences):
            length=(length_limits[0] + np.random.rand(1)*(length_limits[1]-length_limits[0])).astype('int')[0]
            SP=model.sample(length=length,path=True) 
            SP_all+=(SP,)
            samples_all +=(np.array(SP[0]),)

            #Extract States Ground truth
        StateGroundTruth=()
        for i in range(n_sequences):
            value=np.array([])
            for state in SP_all[i][1]:
                try:
                    value=np.append(value,np.array([state.name[1:]]).astype(int))
                except:
                    pass
            StateGroundTruth+=(value,)

        return samples_all, StateGroundTruth

    def train_HMM(X, n_components=2, BootStrap=0, distribution='GammaDistribution',weights=None):
        """
        Train a HMM given a data set and amount of states
        X:              a tuple of arrays of data describing 1/D, assuming that they are Gamma distribution
        n_components:   amount of componets that the model is to fit to.
        BootStrap:      if >0 then the first 
        """
        if distribution=='GammaDistribution':
            use_distribution= _PG.distributions.GammaDistribution
        elif distribution=='PoissonDistribution':
            use_distribution= _PG.distributions.PoissonDistribution
        elif distribution=='NormalDistribution':
            use_distribution= _PG.distributions.NormalDistribution
        else:
            print('Distribution not supported:' + use_distribution)

        FoundModel={}
        index=0
        FoundModel[index] = _PG.HiddenMarkovModel.from_samples(use_distribution, n_components=n_components, X=X, weights=weights, n_init=100, init='random',min_iterations=1000,stop_threshold=1e-11)
        
        if BootStrap>0:
            index=1
            while index<BootStrap+1:
                print('Running bootstrap number: ' + str(index) + '/' + str(BootStrap))
                Samples=np.random.uniform(low=0,high=len(X),size=len(X)).astype('int')
                XBootStrap=()
                for S in Samples:
                    XBootStrap+=(X[S],)
                try:
                    FoundModel[index] = _PG.HiddenMarkovModel.from_samples(use_distribution, n_components=n_components, X=XBootStrap, n_init=100, init='random',min_iterations=1000,stop_threshold=1e-11)
                    
                    #Test Modle
                    new_trans_mat, dist_params = HMM_Traj.get_model_parm(FoundModel[index]) 
                    TestPass=0
                    for state in dist_params:
                        alpha_dis=dist_params[state][0] #is the Gamma dist fitt (shape) but also the displaced inv Gamma alpha (shape)
                        beta_dis=dist_params[state][1]  #is the Gamma dist fitt (rate)  but also the displaced inv Gamma beta  (scale)

                        #D_estimated =beta_dis/(alpha_dis-1)-DBias #mean inv gamma = beta/(alpha-1) invers-Gamma mean and removed bias shift
                        VarinvGamma = beta_dis**2/((alpha_dis-2)*(alpha_dis-1)**2)
                        if VarinvGamma<0:
                            print('Something whent wrong, re-doing index: ' + str(index))
                        else:
                            TestPass=1
                    if TestPass==1:
                        index += 1
                except:
                    print('Something whent wrong, re-doing index: ' + str(index))
                
        return FoundModel

    def get_model_parm(model):
        """
        Extract model information
        input:
            model: pomegranate model 

        output:
            M: tansition matrix
        """

        M=model.dense_transition_matrix()[:-2,:-2]
        dist_param={}
        for state in model.states:
            try:
                dist_param[state.name]=state.distribution.parameters
            except:
                pass

        if M.shape[0]!=len(dist_param):
            print("Something is wrong with the distribution size and matrix size")

        return M, dist_param

    def plot_hist_sequence_with_fit(samples_all,dist_params,xoffset=0):
        """
        Plot hitogram of the 1/data and the invers-Gamma distributions defiend in dist_params
        """
        bins=200

        #Flatten sampel array for visualization
        samples_all_flatten=np.array([])
        for sample in samples_all:
            samples_all_flatten=np.append(samples_all_flatten,sample)

        D=(1/samples_all_flatten)-xoffset

        plotRange=[np.min(D),20]
        # Generate the inverse-Gamma distributions to plot
        DFittParam=np.linspace(plotRange[0],plotRange[1],num=bins*10)
        invGamma={}
        for state in dist_params:
            param=dist_params[state]
            invGamma[state]=stats.invgamma.pdf(DFittParam+xoffset, a=param[0],scale=param[1])

        # Figure construction
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(5*3, 3*5, forward=True)
        gs = GridSpec(1, 1, figure=fig)
        ax0=fig.add_subplot(gs[0,0]) #Hist 

        ax0.hist(D,bins=np.linspace(plotRange[0],plotRange[1],num=bins),density=True)
        for state in invGamma:
            ax0.plot(DFittParam,invGamma[state],label=state)
            ax0.axvline(dist_params[state][1]/(dist_params[state][0]-1) -xoffset, color='g',linestyle='--',label=state + ' mean:' + str(np.round(dist_params[state][1]/(dist_params[state][0]-1)-xoffset,decimals=2)))
            ax0.axvline(dist_params[state][1]/(dist_params[state][0]+1) -xoffset, color='r',linestyle='--',label=state + ' mode:' + str(np.round(dist_params[state][1]/(dist_params[state][0]+1)-xoffset,decimals=2)))

        ax0.legend()
        plt.show()

    def plot_hist_sequence_with_fit2(samples_all,HMM_Model, bins=200 ,do_save_plot=1 ,Save_path='',FileEnd=''):
        """
        Plot hitogram of the data and the invers-Gamma distributions defiend in dist_params
        """
        dist_params=HMM_Model['Distributions']
        Occupancy_estimated=HMM_Model['Occupancy']
        alpha_beta=HMM_Model['alpha_beta']
        DBias=HMM_Model['DBias']

        D_estimated=HMM_Model['D_dic']
        M_estimated=HMM_Model['M_dic']    
        VarinvGamma=HMM_Model['VarinvGamma_dic']


        #Flatten sample array for visualization
        samples_all_flatten=np.array([])
        for sample in samples_all:
            samples_all_flatten=np.append(samples_all_flatten,sample)

        D=samples_all_flatten

        #plotRange=[np.min(D),20]
        plotRange=[-1,20]
        # Generate the inverse-Gamma distributions to plot
        DFittParam=np.linspace(plotRange[0],plotRange[1],num=bins)
        if  HMM_Model['DistributionType']=='GammaDistribution':
            invGammaB={}
            invGamma={}
            invGammaBAll=np.zeros(DFittParam.shape)
            invGammaAll=np.zeros(DFittParam.shape)
            for state,Occ in zip(dist_params,Occupancy_estimated):
                param=dist_params[state]
                invGammaB[state]=stats.invgamma.pdf(DFittParam+DBias, a=param[0],scale=param[1])
                invGammaBAll=invGammaBAll + Occ*invGammaB[state]

                invGamma[state]=stats.invgamma.pdf(DFittParam, a=alpha_beta[state][0],scale=alpha_beta[state][1])
                invGammaAll=invGammaAll + Occ*invGamma[state]
                
        
            NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
            invGammaAll=invGammaAll/NormFactor
            NormFactorB=np.sum(invGammaBAll*np.diff(DFittParam+DBias)[0])
            invGammaBAll=invGammaBAll/NormFactorB

        elif  HMM_Model['DistributionType']=='NormalDistribution':
            invGammaB={}
            invGamma={}
            invGammaBAll=np.zeros(DFittParam.shape)
            invGammaAll=np.zeros(DFittParam.shape)
            for state,Occ in zip(dist_params,Occupancy_estimated):
                param=dist_params[state]
                invGammaB[state]=stats.norm.pdf(DFittParam+DBias, loc=param[0],scale=param[1])
                invGamma[state]=stats.norm.pdf(DFittParam, loc=alpha_beta[state][0],scale=alpha_beta[state][1])
                invGammaAll=invGammaAll + Occ*invGamma[state]
                invGammaBAll=invGammaBAll + Occ*invGammaB[state]
        
            NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
            invGammaAll=invGammaAll/NormFactor
            NormFactorB=np.sum(invGammaBAll*np.diff(DFittParam+DBias)[0])
            invGammaBAll=invGammaBAll/NormFactorB

        # Figure construction
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(15, 8, forward=True)
        gs = GridSpec(1, 1, figure=fig)
        ax0=fig.add_subplot(gs[0,0]) #Hist 

        ax0.hist(D,bins=DFittParam,density=True)
        if 1: #Biased but ploted over displaced range 
            for state,Occ in zip(invGammaB,Occupancy_estimated):
                ax0.plot(DFittParam,Occ*invGammaB[state]/NormFactorB,label=state)
                if  HMM_Model['DistributionType']=='GammaDistribution':
                    
                    #StateMeanD=dist_params[state][1]/(dist_params[state][0]-1) -DBias
                    StateModeD=dist_params[state][1]/(dist_params[state][0]+1) -DBias
                    StateMeanD=D_estimated[state]
                    #StateModeD=M_estimated[state] # This might not be correct since it is the biased version we see
                    StateSTDD=np.sqrt(VarinvGamma[state])
                    text1=' mean:' + str(np.round(StateMeanD,decimals=2)) + '+-' + str(np.round(StateSTDD,decimals=2))
                    text2=' mode:' + str(np.round(StateModeD,decimals=2))

                elif  HMM_Model['DistributionType']=='NormalDistribution':
                    StateMeanD=dist_params[state][0] -DBias
                    StateModeD=dist_params[state][1]
                    text1=' mean:' + str(np.round(StateMeanD,decimals=2))
                    text2=' std:' + str(np.round(StateModeD,decimals=2))

                ax0.axvline(StateMeanD, color='g',linestyle='--',label=state + text1)
                ax0.axvline(StateModeD, color='r',linestyle='--',label=state + text2)

                print('Hist info, state: ' + state)
                print('Hist info, MaxValue: ' + str(np.max(Occ*invGammaB[state]/NormFactorB)))
                print('Hist info, Occ: ' + str(Occ))

            ax0.plot(DFittParam,invGammaBAll,'-k',label='envelope')
            print('Hist info, state: envelope')
            print('Hist info, MaxValue: ' + str(np.max(invGammaBAll)))
            
        if 0: #Not biased Some problems here since the inverse Gamma can not be negative which reshapes everything
            for state,Occ in zip(invGamma,Occupancy_estimated):
                ax0.plot(DFittParam,Occ*invGamma[state]/NormFactor,label=state)
                StateMeanD=alpha_beta[state][1]/(alpha_beta[state][0]-1) 
                StateModeD=alpha_beta[state][1]/(alpha_beta[state][0]+1)
                ax0.axvline(StateMeanD, color='b',linestyle='--',label=state + ' mean:' + str(np.round(StateMeanD,decimals=2)))
                ax0.axvline(StateModeD, color='m',linestyle='--',label=state + ' mode:' + str(np.round(StateModeD,decimals=2)))
            ax0.plot(DFittParam,invGammaAll,'--k',label='envelope')
        ax0.legend()
        print('Hist info, Y_limits: ' + str(ax0.get_ylim()))

        #########################################################
        if do_save_plot:
            file="HistAll_HMM_" + str(HMM_Model['TransMat'].shape[0]) + FileEnd + '.svg'
            if Save_path!='':
                path= Save_path +'/'
                fig.savefig(path + file, bbox_inches='tight')
            else: print('No path was given to save CovMatrix image')
            plt.close()
            print('Image saved Done!')
        else:
            plt.show()
            print('Done!')

    def plot_hist_sequence_with_fit3(samples_all,HMM_Model, bins=200 ,do_save_plot=1 ,Save_path='',FileEnd='',NormValue=1, ExpOcc=1,plotYToplim=0,PDF=1):
        """
        Plot hitogram of the data and the invers-Gamma distributions defiend in dist_params
        This will normalize to value "NormValue" and scale by "ExpOcc"
        The logic here is to compeare this calculated distributions to an existing distribution which 
        has a known overall max value "NormValue" and subdistributions with occupancise "ExpOcc"
        """
        dist_params=HMM_Model['Distributions']
        Occupancy_estimated=HMM_Model['Occupancy']
        alpha_beta=HMM_Model['alpha_beta']
        DBias=HMM_Model['DBias']

        D_estimated=HMM_Model['D_dic']
        M_estimated=HMM_Model['M_dic']    
        VarinvGamma=HMM_Model['VarinvGamma_dic']


        #Flatten sample array for visualization
        samples_all_flatten=np.array([])
        for sample in samples_all:
            samples_all_flatten=np.append(samples_all_flatten,sample)

        D=samples_all_flatten

        #plotRange=[np.min(D),20]
        plotRange=[-1,20]
        # Generate the inverse-Gamma distributions to plot
        DFittParam=np.linspace(plotRange[0],plotRange[1],num=bins)
        if  HMM_Model['DistributionType']=='GammaDistribution':
            invGammaB={}
            invGamma={}
            invGammaBAll=np.zeros(DFittParam.shape)
            invGammaAll=np.zeros(DFittParam.shape)
            for state,Occ in zip(dist_params,Occupancy_estimated):
                param=dist_params[state]
                invGammaB[state]=stats.invgamma.pdf(DFittParam+DBias, a=param[0],scale=param[1])
                invGammaBAll=invGammaBAll + Occ*invGammaB[state]

                invGamma[state]=stats.invgamma.pdf(DFittParam, a=alpha_beta[state][0],scale=alpha_beta[state][1])
                invGammaAll=invGammaAll + Occ*invGamma[state]
                
        
            NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
            invGammaAll=invGammaAll/NormFactor
            if PDF==1:
                NormFactorB=np.sum(invGammaBAll*np.diff(DFittParam+DBias)[0])/ExpOcc
            else:
                NormFactorB=np.max(invGammaBAll)/(NormValue*ExpOcc)
            invGammaBAll=invGammaBAll/NormFactorB

        elif  HMM_Model['DistributionType']=='NormalDistribution':
            invGammaB={}
            invGamma={}
            invGammaBAll=np.zeros(DFittParam.shape)
            invGammaAll=np.zeros(DFittParam.shape)
            for state,Occ in zip(dist_params,Occupancy_estimated):
                param=dist_params[state]
                invGammaB[state]=stats.norm.pdf(DFittParam+DBias, loc=param[0],scale=param[1])
                invGamma[state]=stats.norm.pdf(DFittParam, loc=alpha_beta[state][0],scale=alpha_beta[state][1])
                invGammaAll=invGammaAll + Occ*invGamma[state]
                invGammaBAll=invGammaBAll + Occ*invGammaB[state]
        
            NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
            invGammaAll=invGammaAll/NormFactor
            NormFactorB=np.sum(invGammaBAll*np.diff(DFittParam+DBias)[0])
            invGammaBAll=invGammaBAll/NormFactorB



        #Simple fit to a single invers-gamma
        invfit_alpha, invfit_loc, invfit_scale=stats.invgamma.fit(D+DBias)
        PDF_invGamma=stats.invgamma.pdf(DFittParam+DBias, a=invfit_alpha,loc=invfit_loc, scale=invfit_scale)
        if PDF==1:
            PDF_invGamma=PDF_invGamma*ExpOcc
        else:
            PDF_invGamma=NormValue*ExpOcc*PDF_invGamma/np.max(PDF_invGamma)

        # Figure construction
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(15, 8, forward=True)
        gs = GridSpec(1, 1, figure=fig)
        ax0=fig.add_subplot(gs[0,0]) #Hist 

        if PDF==1:
            n=np.histogram(D,bins=DFittParam,density=True)
            ax0.bar(n[1][:-1], ExpOcc*n[0],width=np.diff(DFittParam)[0], align='edge')
        else:
            n=np.histogram(D,bins=DFittParam,density=True)
            ax0.bar(n[1][:-1], (NormValue*ExpOcc)*n[0]/np.max(n[0]),width=np.diff(DFittParam)[0], align='edge')

        if 1: #Biased but ploted over displaced range 
            for state,Occ in zip(invGammaB,Occupancy_estimated):
                ax0.plot(DFittParam,Occ*invGammaB[state]/NormFactorB,label=state)
                if  HMM_Model['DistributionType']=='GammaDistribution':
                    
                    #StateMeanD=dist_params[state][1]/(dist_params[state][0]-1) -DBias
                    StateModeD=dist_params[state][1]/(dist_params[state][0]+1) -DBias
                    StateMeanD=D_estimated[state]
                    #StateModeD=M_estimated[state] # This might not be correct since it is the biased version we see
                    StateSTDD=np.sqrt(VarinvGamma[state])
                    text1=' mean:' + str(np.round(StateMeanD,decimals=2)) + '+-' + str(np.round(StateSTDD,decimals=2))
                    text2=' mode:' + str(np.round(StateModeD,decimals=2))

                elif  HMM_Model['DistributionType']=='NormalDistribution':
                    StateMeanD=dist_params[state][0] -DBias
                    StateModeD=dist_params[state][1]
                    text1=' mean:' + str(np.round(StateMeanD,decimals=2))
                    text2=' std:' + str(np.round(StateModeD,decimals=2))

                ax0.axvline(StateMeanD, color='g',linestyle='--',label=state + text1)
                ax0.axvline(StateModeD, color='r',linestyle='--',label=state + text2)

            ax0.plot(DFittParam,PDF_invGamma,'-m',label='single invGam' + str(np.round(np.mean(D),decimals=2)) + '+-' + str(np.round(np.std(D),decimals=2)))
            ax0.plot(DFittParam,invGammaBAll,'-k',label='envelope')

        if 0: #Not biased Some problems here since the inverse Gamma can not be negative which reshapes everything
            for state,Occ in zip(invGamma,Occupancy_estimated):
                ax0.plot(DFittParam,Occ*invGamma[state]/NormFactor,label=state)
                StateMeanD=alpha_beta[state][1]/(alpha_beta[state][0]-1) 
                StateModeD=alpha_beta[state][1]/(alpha_beta[state][0]+1)
                ax0.axvline(StateMeanD, color='b',linestyle='--',label=state + ' mean:' + str(np.round(StateMeanD,decimals=2)))
                ax0.axvline(StateModeD, color='m',linestyle='--',label=state + ' mode:' + str(np.round(StateModeD,decimals=2)))
            ax0.plot(DFittParam,invGammaAll,'--k',label='envelope')
        ax0.legend()
        if plotYToplim>0:
            ax0.set_ylim(bottom=0,top=plotYToplim)
        

        #########################################################
        if do_save_plot:
            if PDF==1:
                file="HistAll_scaled_PDF_HMM_" + str(HMM_Model['TransMat'].shape[0]) + FileEnd + '.svg'
            else:
                file="HistAll_scaled_HMM_" + str(HMM_Model['TransMat'].shape[0]) + FileEnd + '.svg'
            if Save_path!='':
                path= Save_path +'/'
                fig.savefig(path + file, bbox_inches='tight')
            else: print('No path was given to save CovMatrix image')
            plt.close()
            print('Image saved Done!')
        else:
            plt.show()
            print('Done!')

    def predict_sequence(model,observed_sequence):
        """
        Predict a state sequence given a observed sequence
        input:
            model:              pomegranagte hmm model
            observed_sequence:  iterable set of sample sequences
        output:
            predicted_state_sequence: tuple of state sequences 
        """
        predicted_state_sequence=()
        for sequence in observed_sequence:
            predicted_state_sequence+=(model.predict(sequence, algorithm='viterbi')[1:],)

        return predicted_state_sequence

    def OrgenizeHMM_parm(FoundModelDic,dt_est,DBias,distribution='GammaDistribution'):
        """
        The relation between D "mean", M "Mode" of the inv-Gamma and alpha and beta of the distribution is
        alpha=(M+D)/(D-M) or D=beta/(alpha-1) 
        beta=2*D*M/(D-M)  or M=beta/(alpha+1)
        """
        HMM_Models={}
        for ModelIndex in  FoundModelDic:
            FoundModel= FoundModelDic[ModelIndex]
            new_trans_mat, dist_params =HMM_Traj.get_model_parm(FoundModel) 

            D_estimated=()
            M_estimated=()
            VarinvGamma=()
            D_estimated_dic={}
            M_estimated_dic={}
            VarinvGamma_dic={}
            alpha_beta={}
            for state in dist_params:
                #alpha,beta= HMM_Traj.convert_displaced_ab_GammaTo_ab_invGamma(dist_params[state],DinvBias)
                alpha_dis=dist_params[state][0] #is the Gamma dist fitt (shape) but also the displaced inv Gamma alpha (shape)
                beta_dis=dist_params[state][1]  #is the Gamma dist fitt (rate)  but also the displaced inv Gamma beta  (scale)
                
                if distribution=='GammaDistribution':

                    D_estimated+=(beta_dis/(alpha_dis-1)-DBias,) #mean inv gamma = beta/(alpha-1) invers-Gamma mean and removed bias shift
                    VarinvGamma+=(beta_dis**2/((alpha_dis-2)*(alpha_dis-1)**2),) #variance inv gamma = beta^2/(alpha-1)^2(alpha-2) 
                    alpha = (D_estimated[-1]**2 / VarinvGamma[-1]**2) + 2
                    beta  = D_estimated[-1] * (alpha-1)
                    M_estimated+=(beta/(alpha+1),) #beta/(alpha+1) invers-Gamma mode

                    if 0: #Old and might be wrong 
                        D_estimated+=(beta_dis/(alpha_dis-1)-DBias,) #beta/(alpha-1) invers-Gamma mean
                        M_estimated+=(beta_dis/(alpha_dis+1)-DBias,) #beta/(alpha+1) invers-Gamma mode, Think this is an assumption that might not be valied
                        VarinvGamma+=(beta_dis**2/((alpha_dis-2)*(alpha_dis-1)**2),)
                        alpha=(D_estimated[-1]+M_estimated[-1])/(D_estimated[-1]-M_estimated[-1]) #alpha without bias
                        beta=2*D_estimated[-1]*M_estimated[-1]/(D_estimated[-1]-M_estimated[-1])  #beta without bias

                elif distribution=='NormalDistribution':
                    D_estimated+=(alpha_dis-DBias,) #alpha is the Gausian mean
                    M_estimated+=(alpha_dis-DBias,) #alpha is the Gausian mode
                    VarinvGamma+=(beta_dis**2,)
                    alpha=alpha_dis #alpha without bias
                    beta=beta_dis  #beta without bias

                alpha_beta[state]=[alpha,beta]
                D_estimated_dic[state]=D_estimated[-1]
                M_estimated_dic[state]=M_estimated[-1]
                VarinvGamma_dic[state]=VarinvGamma[-1]

            Dwelltime_estimated=()
            for i in range(new_trans_mat.shape[0]):
                #Dwelltime_estimated+=(dt_est/(1-new_trans_mat[i,i]),) #[ms]
                Dwelltime_estimated+=(-dt_est/np.log(new_trans_mat[i,i]),) #[ms]
            Dwelltime_estimated=np.array(Dwelltime_estimated)
    
            #Occupancy estimate 
            #Occupancy_estimated=()
            #for Dwellt in Dwelltime_estimated: 
            #    Occupancy_estimated+=(Dwellt/np.sum(Dwelltime_estimated),)
            #Occupancy_estimated=np.array(Occupancy_estimated)
            
            eigValues,eigVectors,_=_sc.linalg.eig(new_trans_mat,left=True)
            eigValueArg=np.argmin(np.abs(np.abs(eigValues)-1)).astype('int') #find the one closets to 1
            Occupancy_estimated=eigVectors[:,eigValueArg]/np.sum(eigVectors[:,eigValueArg])
            
            HMM_Model={}
            HMM_Model['DistributionType']=distribution
            HMM_Model['TransMat']=new_trans_mat
            HMM_Model['Distributions']=dist_params
            HMM_Model['D_dic']=D_estimated_dic
            HMM_Model['M_dic']=M_estimated_dic
            HMM_Model['VarinvGamma_dic']=VarinvGamma_dic
            HMM_Model['alpha_beta']=alpha_beta
            HMM_Model['D']=D_estimated
            HMM_Model['M']=M_estimated
            HMM_Model['VarinvGamma']=VarinvGamma
            HMM_Model['Dwelltime']=Dwelltime_estimated
            HMM_Model['Occupancy']=Occupancy_estimated
            HMM_Model['DBias']=DBias

            HMM_Models[ModelIndex]=HMM_Model
    
        return HMM_Models

    def convert_displaced_ab_GammaTo_ab_invGamma(dist_params,DinvBias):
        alpha1=dist_params[0]
        beta1=dist_params[1]
        alpha=(alpha1-beta1*DinvBias)**2 /alpha1   #Converting alpha of Gamma to invers Gamma 
        beta=beta1*(alpha1-beta1*DinvBias) /alpha1 #Converting beta of Gamma to invers Gamma 
        if alpha<=1:
            print('OBS! alpha is under threshhold 1: ' + str(alpha))
        elif alpha<=2:
            print('OBS! alpha is under threshhold 2: ' + str(alpha))
        return alpha,beta

    def get_HMMCovMatrix(HMM_Models_In,do_save_plot=1,Save_path=''):
        """
        This produces the covariance matrix of the transition matrix together with the diffusion estimation and mode estimation 
        the NxN transition matrix is reduced by removing the diagonal (since each row is normolized to 1) and each row is extended 
        to include D and M thus the new matrix is (N-1+2)xN.
        this is then flattend per row thus the element order for a 3 state system is 
        Xi = [a12 a13 D1 M1 a21 a23 D2 M2 a31 a32 D3 M3]
    
        this vektor Xi contains the diffent estiamtions from original data (i=0) and bootstrap (i>0)
        from these vectors we construct the cocvariance matrix elemets which are 
        C(k,l)=E[( Xi(k)-E[Xi(k)]_i )( Xi(l)-E(Xi(l))_i )]_i
        Where the mean is taken over the index i which is the bootstrap data
    
        The HMM fit do not have an order in its parameters, thus the bootstrap HMM models are not alined to each other. The order we will use here is
        States will be orderd according to the rules:
            1. Diffuision constantes are orderd according to the first HMM which in the run with out bootstrap 


        """
    
        print('Getting covariance matrix')
        HMM_Models=HMM_Models_In.copy()

        #Order all models to each other 
        DStateRank0=np.argsort(np.argsort(HMM_Models[0]['D']))
        TStateRank0=np.argsort(np.argsort(HMM_Models[0]['Dwelltime']))
        #print(HMM_Models[0]['D'])
        #print(DStateRank0)
        #print('##################')
        if len(HMM_Models)>1:
            for index in range(1,len(HMM_Models)):
                #print(HMM_Models[index]['D'])
                HMM_Model=HMM_Models[index].copy()
                DStateRank=np.argsort(np.argsort(np.array(HMM_Model['D'])))
                TStateRank=np.argsort(np.argsort(HMM_Model['Dwelltime']))
                Dmap=()
                Tmap=()
                for i in range(len(DStateRank0)):
                    Dmap+=(np.where(DStateRank0[i]==DStateRank)[0][0],)
                    #Tmap+=(np.where(TStateRank0[i]==TStateRank)[0][0],)
                Smap=Dmap
                #print('rank:' + str(DStateRank))
                #print('map  :' + str(Smap))

                new_trans_mat=np.zeros(HMM_Model['TransMat'].shape)
                for i in range(len(DStateRank0)):
                    for j in range(len(DStateRank0)):
                        new_trans_mat[Smap[i],Smap[j]]=HMM_Model['TransMat'][i,j]
                HMM_Model['TransMat']=new_trans_mat

                new_Distributions={}
                for i in range(len(HMM_Model['Distributions'])):
                    new_Distributions['s' + str(i)]=HMM_Model['Distributions']['s' + str(Smap[i])]
                HMM_Model['Distributions']=new_Distributions

                HMM_Model['D']=tuple(np.array(HMM_Model['D'])[np.array(Smap)])
                HMM_Model['M']=tuple(np.array(HMM_Model['M'])[np.array(Smap)])
                HMM_Model['VarinvGamma']=tuple(np.array(HMM_Model['VarinvGamma'])[np.array(Smap)])
                HMM_Model['Dwelltime']=tuple(np.array(HMM_Model['Dwelltime'])[np.array(Smap)])
                HMM_Model['Occupancy']=tuple(np.array(HMM_Model['Occupancy'])[np.array(Smap)])

                HMM_Models[index]=HMM_Model
                #print(HMM_Models[index]['D'])
                #print('-------next------------')

            #X=np.zeros((len(HMM_Models),len(HMM_Models[0]['D'])**2 + 2*len(HMM_Models[0]['D'])-len(HMM_Models[0]['D'])))
            X=np.zeros((len(HMM_Models),3*len(HMM_Models[0]['D'])))
            for index in range(len(HMM_Models)):
                HMM_Model=HMM_Models[index].copy()
                #Creating the relevent vector that we will construct the covariance
                #ReducedTransMat=HMM_Model['TransMat'][~np.eye(HMM_Model['TransMat'].shape[0],dtype=bool)].reshape(HMM_Model['TransMat'].shape[0],-1)
                #X[index,:]=(np.vstack((ReducedTransMat.T,HMM_Model['D'],HMM_Model['M'])).T).flatten()
                X[index,:]=(np.vstack((HMM_Model['Occupancy'],HMM_Model['D'],np.sqrt(HMM_Model['VarinvGamma']))).T).flatten()
            CovMatrix=np.cov(X.T,bias=False)
            
            #Ceate lables
            if 0:
                x_axis_labels=()
                for i in range(HMM_Model['TransMat'].shape[0]):
                    for j in range(HMM_Model['TransMat'].shape[0]):
                        if i!=j:
                            x_axis_labels+=('a'+str(i)+str(j),)
                        if (j==HMM_Model['TransMat'].shape[0]-1):
                            x_axis_labels+=('D'+str(i),)
                            x_axis_labels+=('M'+str(i),)
                CovMatrix_labels=x_axis_labels
            else:
                x_axis_labels=()
                for i in range(len(HMM_Models[0]['D'])):
                    for j in ['O','D','S']:
                            x_axis_labels+=(j+str(i),)
                CovMatrix_labels=x_axis_labels

            # Figure construction
            fig = plt.figure(constrained_layout=False)
            fig.set_size_inches(20, 15, forward=True)
            gs = fig.add_gridspec(nrows=6, ncols=6, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.2, hspace=1.0)
            ax=fig.add_subplot(gs[2:5, 0:3]) #covmatrix
            ax12=fig.add_subplot(gs[0:2, 4:5]) #Info

            #ax covmatrix
            sn.heatmap(np.round(CovMatrix,decimals=4), annot=True, fmt='g',xticklabels=x_axis_labels, yticklabels=x_axis_labels,ax=ax)

            #ax12 Info box 
            HMM_Model=HMM_Models[0]           
            Values=np.zeros(HMM_Model['TransMat'].shape+np.array([3,0])) 
            Values[:HMM_Model['TransMat'].shape[0],:HMM_Model['TransMat'].shape[1]]=HMM_Model['TransMat']
            Values[HMM_Model['TransMat'].shape[0],:]=HMM_Model['D']
            Values[HMM_Model['TransMat'].shape[0]+1,:]=HMM_Model['Dwelltime']
            Values[HMM_Model['TransMat'].shape[0]+2,:]=HMM_Model['Occupancy']
            Values=np.round(Values,decimals=2)

            S=()
            for state in HMM_Model['Distributions']:
                S+=(state,)
            StatesLables=list(S)
            S+=('D:', 'dwelltime:','POcc',)
            LablesAll=list(S)

            ax12.pcolor(Values*0, cmap=plt.cm.Greys)
            #ax12.imshow(Values)
            ax12.set_xticks(np.arange(len(StatesLables))+0.5, minor=False)
            ax12.set_yticks(np.arange(len(LablesAll))+0.5, minor=False)
            ax12.set_xticklabels(StatesLables, minor=False)
            ax12.set_yticklabels(LablesAll, minor=False)
            ax12.invert_yaxis()

            for i in range(len(LablesAll)):
                for j in range(len(StatesLables)):
                    text = ax12.text(j+0.5, i+0.5, Values[i, j],
                                    ha="center", va="center", color="k")                
            
            #########################################################
            if do_save_plot:
                file="CovMatrix_HMM_" + str(HMM_Model['TransMat'].shape[0]) + '.svg'
                if Save_path!='':
                    path= Save_path +'/'
                    fig.savefig(path + file,bbox_inches='tight')
                else: print('No path was given to save CovMatrix image')
                plt.close()
                print('Covariance image saved Done!')
            else:
                plt.show()
                print('Done!')
        else:
            CovMatrix=None
            CovMatrix_labels=None

        return CovMatrix, CovMatrix_labels

    def get_photon_state_HMM(ScanProject_Dic,wndw_h=1000,n_components=4,Target_mean_photoncount=1, UpperLower_offsetIndex=[1,0], UpperLower_offsetValue=[-0.05,0], do_save_plot=0):
        """
        This function is:
            1. takes all photon counts and finds the traking parts 
            2. makes an averageing defined by the window size, and makes a list of all these photon trajectories
            3. an HMM with n_components is trained on the photon list
            4. using the HMM a prediction is made on each photon trajectory
            5. some further filtering is made on the prediction like removing single peaks 
            6. the prediction is then re sampled to the resolution of the ScanMatrix 

        Inputs
        ScanProject_Dic: a dictionary of ScanProjects
        wndw_h:          averageing window size to use 
        n_components:    HMM components

        Output: dictionary of predicted photon states related to each ScanProject of the input
        photon_count_HMM_dic['Model']:               HMM model found
        photon_count_HMM_dic['HMM_parameters']:      HMM parameters found divided by window size 'wndw_h'
        photon_count_HMM_dic[num]['StartStop']:      Vector with all start and stop positions in the ScanMatrix that are considered
        photon_count_HMM_dic[num]['HMM_prediction']: Parameter prediction (given by 'HMM_parameters') of the photon count list, thus this is the predicted average photon count per shot when tracking  
            
        """
        photon_count_HMM_dic={}
        photon_trace_list=()
        for Project_Number in ScanProject_Dic:
            photon_count_HMM_dic[Project_Number]={}    
            ScanProject=ScanProject_Dic[Project_Number]    
            
            #Find the tracking on/off parts and remove those shorter then the window size
            starts=np.where(np.diff(ScanProject.ScanMatrix[5,:])==1)[0]+1
            stops=np.where(np.diff(ScanProject.ScanMatrix[5,:])==-1)[0]+1            
            ToUse=np.where((stops-starts)>wndw_h)
            starts=starts[ToUse]
            stops=stops[ToUse]
            photon_count_HMM_dic[Project_Number]['StartStop']=np.vstack((starts,stops))    

            for start,stop in zip(starts,stops):
                trace=np.convolve(np.ones(wndw_h),np.sum(ScanProject.ScanMatrix[3:5,start:stop],axis=0))[wndw_h-1:-wndw_h+1:wndw_h]
                photon_trace_list+=(trace,)
                

        #Run HMM on the photon trajectoris and make a prediction for each trajectory    
        FoundModelDic=HMM_Traj.train_HMM(X=photon_trace_list, n_components=n_components,BootStrap=0,distribution='PoissonDistribution')
        photon_count_HMM_dic['Model']=FoundModelDic
        HMM_predict_states=HMM_Traj.predict_sequence(model=FoundModelDic[0],observed_sequence=photon_trace_list)

        HMM_param=np.zeros(n_components)
        for i in range(n_components):
            HMM_param[i]=FoundModelDic[0].states[i].distribution.parameters[0]/wndw_h
        photon_count_HMM_dic['HMM_parameters']=HMM_param

        HMM_predict=()
        for i in range(len(photon_trace_list)):
            HMM_predict+=(HMM_param[HMM_predict_states[i]],)
        
        if 0:
            i=2;
            plt.plot(photon_trace_list[i]/wndw_h);
            plt.plot(HMM_predict[i],'-r');
            plt.plot(np.convolve(np.ones(4)/4,HMM_predict[i])[4-1:-4+1:1],'-g');
            
            plt.show()

        #Create a vector of the ScanMatrix size for each ScanProject that gives the predicted average photon count  
        HMM_predict=list(HMM_predict)
        for Project_Number in ScanProject_Dic:
            ScanProject=ScanProject_Dic[Project_Number]    
            average_photon = np.zeros(ScanProject.ScanMatrix[0,:].shape)

            StartStop=photon_count_HMM_dic[Project_Number]['StartStop']
            for start,stop in zip(StartStop[0,:],StartStop[1,:]):
                predicted=HMM_predict.pop(0)
                P=np.repeat(np.expand_dims(predicted,axis=1),wndw_h,axis=1).flatten()
                if len(P)<(stop-start):
                    diff=(stop-start)-len(P)
                    P=np.append(P,P[-1]*np.ones(diff))
                average_photon[start:stop]=P
            photon_count_HMM_dic[Project_Number]['HMM_prediction']=average_photon

            #Construct filter
            lower=np.sort(photon_count_HMM_dic['HMM_parameters'][np.where((photon_count_HMM_dic['HMM_parameters']-Target_mean_photoncount)<0)])[-1::-1][UpperLower_offsetIndex[1]]+UpperLower_offsetValue[1]
            upper=np.sort(photon_count_HMM_dic['HMM_parameters'][np.where((photon_count_HMM_dic['HMM_parameters']-Target_mean_photoncount)>0)])[UpperLower_offsetIndex[0]]+UpperLower_offsetValue[0]
            smooth_average_photon=np.convolve(np.ones(5*wndw_h)/(5*wndw_h),average_photon)[5*wndw_h-1:-5*wndw_h+1:1]
            smooth_average_photon[np.where(smooth_average_photon<=lower)]=0
            smooth_average_photon[np.where(smooth_average_photon>=upper)]=0
            smooth_average_photon[np.where(smooth_average_photon>0)]=1
            photon_count_HMM_dic[Project_Number]['HMM_predictionFilter']=smooth_average_photon

            if do_save_plot: #Debug
                plt.plot(np.convolve(np.ones(wndw_h)/wndw_h,np.sum(ScanProject.ScanMatrix[3:5,:],axis=0))[wndw_h-1:-wndw_h+1:1]);
                plt.plot(np.convolve(np.ones(5*wndw_h)/(5*wndw_h),average_photon)[5*wndw_h-1:-5*wndw_h+1:1],'-g',label='HMM par: ' + str(np.round(photon_count_HMM_dic['HMM_parameters'], decimals=2)));
                plt.plot(lower*np.ones(average_photon.shape),'-m')
                plt.plot(upper*np.ones(average_photon.shape),'-m')
                plt.plot(photon_count_HMM_dic[Project_Number]['HMM_predictionFilter'],'-r');
                plt.legend()
                file="PhotonHMMFilter" + '.svg'
                path=ScanProject.Files['AnalysisConfig'][:-18]                 
                plt.savefig(path + file)
                plt.close()
                #plt.show()

        return photon_count_HMM_dic

def Test_DiffusionState_pomegranate_hmm():
    n_sequences=50

    D=np.array([2,0.05]) #mean in an inverse gamma distribution
    M=np.array([1,0.01]) #Mode in an inverse gamma distribution
    trans_mat = np.array([[0.8, 0.2],
                             [0.7, 0.3]])

    model=HMM_Traj.init_hmm_model(D,M,trans_mat)
    samples_all, StateGroundTruth= HMM_Traj.get_sample_sequences(model, n_sequences=n_sequences, max_length=1000)
    
    ####### Train the HMM
    FoundModel=HMM_Traj.train_HMM(X=samples_all, n_components=2)
    newM, dist_params =HMM_Traj.get_model_parm(FoundModel)    

    ### Predict 
    predicted_state_sequence=HMM_Traj.predict_sequence(model=FoundModel,observed_sequence=samples_all)

    ####Plot
    print(newM)
    print(dist_params)
    HMM_Traj.plot_hist_sequence_with_fit(samples_all,dist_params)
    
    for i in range(len(predicted_state_sequence)):
        plt.plot(StateGroundTruth[i],'-g')
        plt.plot(predicted_state_sequence[i],'-r')
        plt.show()
    
def Test_AmountOfMoleculsState_pomegranate_hmm():
    n_sequences=50

    lvalue=np.array([0.01,0.05]) #mean in an inverse gamma distribution
    M=np.array([1,0.01]) #Mode in an inverse gamma distribution
    trans_mat = np.array([[0.8, 0.2],
                             [0.7, 0.3]])

    model=HMM_Traj.init_hmm_model(D,M,trans_mat)
    samples_all, StateGroundTruth= HMM_Traj.get_sample_sequences(model, n_sequences=n_sequences, max_length=1000)
    
    ####### Train the HMM
    FoundModel=HMM_Traj.train_HMM(X=samples_all, n_components=2)
    newM, dist_params =HMM_Traj.get_model_parm(FoundModel)    

    ### Predict 
    predicted_state_sequence=HMM_Traj.predict_sequence(model=FoundModel,observed_sequence=samples_all)

    ####Plot
    print(newM)
    print(dist_params)
    HMM_Traj.plot_hist_sequence_with_fit(samples_all,dist_params)
    
    for i in range(len(predicted_state_sequence)):
        plt.plot(StateGroundTruth[i],'-g')
        plt.plot(predicted_state_sequence[i],'-r')
        plt.show()

def Test_Main():
    a=np.array([[0.61,0.3,0.06,0.04],[0.14,0.65,0.01,0.21],[0.32,0.0,0.65,0.04],[0.04,0.33,0.0,0.63]])
    eigValues,eigVectors,_=_sc.linalg.eig(a,left=True)
    eigValueArg=np.argmin(np.abs(np.abs(eigValues)-1)).astype('int') #find the one closets to 1
    Occupancy_estimated=eigVectors[:,eigValueArg]/np.sum(eigVectors[:,eigValueArg])


if __name__ == '__main__':
    Test_Main()
    #test_pomegranate_hmm()



