import concurrent.futures
import numpy as np
from PIL import Image
from skimage import io as _skim
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import scipy.io as io
import concurrent.futures
import scipy as sp
import scipy.interpolate
from scipy.interpolate import RegularGridInterpolator
from datetime import date
import cv2
import time

from ConfigObject_PhRetrieval import ConfigObject_PhRetrieval as _CPhR
from miscellaneous import miscellaneous as _MI
from DataFileIO_Lib import DataFileIO_Lib as _DS



class PSF_Lib:
    def Cal_LogL_Poisson(xi,yi,zi,ni,sx,sy,sz,x0,y0,z0,gsx,gsy,gsz,lx,ly,lz): 
        """
        Cal_LogL_Poisson(xi,yi,ni,sx,x0,y0,gsx,gsy,lx,ly):

        Calculates the log likelihood of the Poisson model, when the fluorophore brightnes is set as: Ns=sqrt{2*pi}^3 *sx*sy*sz* sum(ni)/sum(exp( Dx^2 ...   See notes 
        Parameters not included: 
        1. the laser power, could be nice to include if we can get a measurement per laser shot or at least per z-level  
        2. A constant team that depends on factorial, if used we get overflow in the calculations  

        Description: 
        A x_p, y_p grid is created, the output is the log likelihood over that grid 
        
        xi,yi,ni are vectors containing (x,y) laser coordinates and photon counts 
        sx,sy defines amount of point used for the grid 
        lx,ly defined the area of the grid centerd around x0,y0
        """

        # Function that calculates the log likelihood 
        # Note that one can not calcuate the likelihood directly since ine get 0*inf=nan 
        # The log likelihood is do not handel very large numbers thus 0*inf=nan do not ocure
        # If the likelihood is needed take the result to exp 

        #Create the x_p, y_p grid 
        xp = np.linspace(-lx, lx, endpoint=True, num=gsx) + x0       #grid centered around the mid point of the evaluation 
        yp = np.linspace(-ly, ly, endpoint=True, num=gsy) + y0 
        zp = np.linspace(-lz, lz, endpoint=True, num=gsz) + z0 
        uu, vv, ww = np.meshgrid(xp, yp, zp)
        xx = uu.flatten() 
        yy = vv.flatten()
        zz = ww.flatten()

        #Rechape data 
        xt= np.outer(xi,np.ones(len(xx))) #laser pos x list
        yt= np.outer(yi,np.ones(len(yy))) #laser pos y list
        zt= np.outer(zi,np.ones(len(zz))) #laser pos z list

        xxt=np.outer(np.ones(len(xi)),xx) #xp position 
        yyt=np.outer(np.ones(len(yi)),yy) #yp position 
        zzt=np.outer(np.ones(len(zi)),zz) #zp position
        
        nt= np.outer(ni,np.ones(len(xx))) #photon counts
        # 
        M=np.sum(ni)
        Ax=1/(2*sx**2)
        Ay=1/(2*sy**2)
        Az=1/(2*sz**2)

        #
        Dx2=np.power(xt-xxt,2)
        Dy2=np.power(yt-yyt,2)
        Dz2=np.power(zt-zzt,2)
        
        X1log = np.sum(nt*(Ax*Dx2+Ay*Dy2+Az*Dz2),axis=0)
        X2log = M*np.log(np.sum(np.exp(-(Ax*Dx2+Ay*Dy2+Az*Dz2)),axis=0))
        X3log=0 #np.sum(self.vfactorial(ni)) we get overflow when calculating this
        LogL=X3log-M*(1-np.log(M))-X2log-X1log
        # reshape to a matrix 
        LogL=np.reshape(LogL,(len(xp),len(yp),len(zp)))
        return LogL, uu, vv, ww

    def PSF_Gaussian(r,rp,s,NL,NS,Norm):
        """
        r = laser shot position x,y,z can be in a list of r[:,i] where i is each 
        rp= fluorophor postion  x,y,z can be in a list of rp[:,i] where i is each 
        s = PFS std             x,y,z can be in a list of rs[:,i] where i is each 
             
        NL= Laser strength      Single value over the list 
        NS= fluorophor stregnth Singule value over the list 

        r,rp,s need to be of the same dimensions
        """
        if r.shape[0]!=3 and rp.shape[0]!=3 and  s.shape[0]!=3:
            print("Input wrong")
        
        a= (r-rp)/s
        a= np.sum(a**2, axis=0)
        b=1
        if Norm:
            b=1/(np.power(np.sqrt(2*np.pi),3)*np.prod(s,axis=0))
         
        res=NS*NL*b*np.exp(-a/2)
        return res 

    def Get_tif_stack(file_path):
        #load tif stack data and convert to a numpy array
        dataset = Image.open(file_path)
        h,w = np.shape(dataset)
        tiffarray = np.zeros((h,w,dataset.n_frames))

        for i in range(dataset.n_frames):
            dataset.seek(i)
            tiffarray[:,:,i] = np.array(dataset)
        
        expim = tiffarray.astype(np.double)
        return expim
        
    def PS_Retrieval(I_IP, center_index, parm, iter=100, verbose=0):
        """
        The itration model is the here the standard gerchberg saxton  
        """
        zssize=parm['zssize'] #[m] z- step size between images
        #Start data position 0
        I_IP_N=I_IP[:,:,center_index].copy()

        for iteration in range(iter):
            #Move 0 to +1 
            Dxyz=[0,0,zssize] #[m]
            _, I_IP_N, _,_, _, _, _, _, _, _ = PSF_Lib.LightPropagator(I_IP_N,parm,Dxyz)
                #Inforce measured image ampletud of +1 but keeping the phase 
            I_IP_N=I_IP[:,:,center_index+1]*np.exp(1.0j*np.angle(I_IP_N))
    
            #Move +1 to 0 
            Dxyz=[0,0,-zssize] #[m]
            _, I_IP_N, _,_, _, _, _, _, _, _ = PSF_Lib.LightPropagator(I_IP_N,parm,Dxyz)
                #Inforce measured image ampletud of 0 but keeping the phase 
            I_IP_N=I_IP[:,:,center_index]*np.exp(1.0j*np.angle(I_IP_N))
    
            #Check quality
            if verbose:
                I_IP_NF=I_IP[:,:,center_index]*np.exp(1.0j*np.angle(I_IP_N))


                diff_img_0=np.absolute(I_IP_NF)**2-np.absolute(I_IP[:,:,center_index])**2

                Dxyz=[0,0,zssize] #[m]
                _, I_IP_N_P1, xx,yy, _, _, _, _, _, _ = PSF_Lib.LightPropagator(I_IP_NF,parm,Dxyz)
                diff_img_P1=np.absolute(I_IP_N_P1)**2-np.absolute(I_IP[:,:,center_index+1])**2
                          
                MSE=np.sum(diff_img_0**2)/np.prod(diff_img_0.shape)
                MSE_P=np.sum(diff_img_P1**2)/np.prod(diff_img_0.shape)

                print(str(iteration) + ". Sum square of diff M/P: " + str(np.round(MSE, decimals=3)) + '/' + str(np.round(MSE_P, decimals=3)))    
        
        #Last step replace is to replace image with the raw data and keeping phase
        #I_IP_N=I_IP[:,:,center_index]*np.exp(1.0j*np.angle(I_IP_N))
        return I_IP_N,xx,yy

    def PS_RetrievalDW(I_IP, center_index, parm, iter=100, verbose=0):
        """
        The itration model is the here the double weighted gerchberg saxton, and are described in the two articles

        Article 1: A fast-converging iterative method based on weighted feedback for multi-distance phase retrieval 
                   DOI:10.1038/s41598-018-24666-8 

        Article 2: Lensfree on-chip microscopy based on dual-plane phase retrieval 
                   https://doi.org/10.1364/OE.27.035216


        """
        
        zssize=parm['zssize'] #[m] z- step size between images
        #Start data position 0
        I_IP_N_M1=I_IP[:,:,center_index].copy()
        I_IP_N_P1=I_IP[:,:,center_index+1].copy()

        O_k= (I_IP_N_M1 + I_IP_N_P1)/2
        O_kM1=O_k
        for iteration in range(iter):
            #Move 0 to -1/2 
            Dxyz=[0,0,-zssize/2] #[m]
            _, C_kM, _,_, _, _, _, _, _, _ = PSF_Lib.LightPropagator(O_k,parm,Dxyz)
                #Inforce measured image ampletud of +1 but keeping the phase 
            C_k=I_IP_N_M1*np.exp(1.0j*np.angle(C_kM))
                #Move back the middel
            Dxyz=[0,0,zssize/2] #[m]
            _, g_M_k, _,_, _, _, _, _, _, _ = PSF_Lib.LightPropagator(C_k,parm,Dxyz)
    
            #Move 0 to 1/2 
            Dxyz=[0,0,zssize/2] #[m]
            _, C_kP, _,_, _, _, _, _, _, _ = PSF_Lib.LightPropagator(O_k,parm,Dxyz)
                #Inforce measured image ampletud of 0 but keeping the phase 
            C_k=I_IP_N_P1*np.exp(1.0j*np.angle(C_kP))
                #Move back the middel
            Dxyz=[0,0,-zssize/2] #[m]
            _, g_P_k, _,_, _, _, _, _, _, _ = PSF_Lib.LightPropagator(C_k,parm,Dxyz)

            #mean of the two propagations
            O_k_new=(g_P_k+g_M_k)/2
            
            #weighting step with feedback
            a=0.8; b=-0.13*a+0.6        # this step is an optimized step for convergence in Article 1 
            if iteration>0:
                O_k1=(1+a+b)*O_k_new -a*O_k -b*O_kM1 
                O_kM1=O_k.copy()
                O_k=O_k1.copy()
            else:
                O_kM1=O_k.copy()
                O_k=O_k_new.copy()

            #Check quality
            if verbose:
                ImgkM=np.abs(C_kM)**2
                ImgGM=I_IP_N_M1**2

                ImgkP=np.abs(C_kP)**2
                ImgGP=I_IP_N_P1**2

                MSE_M=np.sum( (ImgkM-ImgGM)**2 )/np.prod(C_kM.shape)
                MSE_P=np.sum( (ImgkP-ImgGP)**2 )/np.prod(C_kP.shape)
                print(str(iteration) + ". Sum square of diff M/P: " + str(np.round(MSE_M, decimals=3)) + '/' + str(np.round(MSE_P, decimals=3)))    
               
        #Last step, move back the first image
        Dxyz=[0,0,-zssize/2] #[m]
        _, I_IP_N, xx,yy, _, _, _, _, _, _ = PSF_Lib.LightPropagator(O_k,parm,Dxyz)
            #Inforce measured image ampletud but keeping the phase 
        I_IP_N=I_IP_N_M1*np.exp(1.0j*np.angle(I_IP_N))

        return I_IP_N,xx,yy

    def Ph_RetrievalStack(data, parm, iter=100,verbose=0):
        # data is a image stack of camear images 
        I_IP=np.sqrt(data)    
        # Do phase retrieval
        I_IP_final=I_IP.copy().astype(complex)    
        for center_index in range(0,I_IP.shape[2]-1):
            if verbose:
                print('Retrieval: ' + str(center_index) + '/'+ str(I_IP.shape[2]-2))
            I_IP_N,xx,yy=PSF_Lib.PS_Retrieval(I_IP,center_index, parm, iter,verbose)
            I_IP_final[:,:,center_index]=I_IP_N  #Tryed to feed back result into next itarasion but this seems to give bad results, strange?
        return I_IP_final, xx, yy
    
    def Ph_RetrievalDWStack(data, parm, iter=100,verbose=0):
        # data is a image stack of camear images 
        I_IP=np.sqrt(data)    
        # Do phase retrieval
        I_IP_final=I_IP.copy().astype(complex)    
        for center_index in range(0,I_IP.shape[2]-1):
            if verbose:
                print('Retrieval: ' + str(center_index) + '/'+ str(I_IP.shape[2]-2))
            I_IP_N,xx,yy=PSF_Lib.PS_RetrievalDW(I_IP,center_index, parm, iter,verbose)
            I_IP_final[:,:,center_index]=I_IP_N  #Tryed to feed back result into next itarasion but this seems to give bad results, strange?
        return I_IP_final, xx, yy

    def Ph_RetrievalStackContinueOnOld(data, parm, iter=100,verbose=0):
        # data is a image stack of camear images 
        I_IP=data
        # Do phase retrieval
        I_IP_final=I_IP.copy().astype(complex)    
        for center_index in range(0,I_IP.shape[2]-1):
            if verbose:
                print('Retrieval: ' + str(center_index) + '/'+ str(I_IP.shape[2]-2))
            I_IP_N,xx,yy=PSF_Lib.PS_Retrieval(I_IP,center_index, parm, iter,verbose)
            I_IP_final[:,:,center_index]=I_IP_N  #Tryed to feed back result into next itarasion but this seems to give bad results, strange?
        return I_IP_final, xx, yy

    def getXYGrid(parm):
        #get pixel Coordinates in [um]
        Px    =parm['Px']*1e6 #[um] Andor ixon 897(512x512) pixel size is 16um, for ixon 888 (1024x1024) it is 13um
        Mag   =parm['Mag']    # Magnification M=f_tube/f_objective
        Mag2  =parm['Mag2']   # Secondery magnification to EMCCD
        NM    =parm['NM']     # Camera pixel dimentions 
        SWS   =parm['SWS']    # Software scaling factor

        N=NM[0]; M=NM[1]
        
        #
        x = np.linspace(0,(Px/(Mag*Mag2))*N, N*SWS)
        y = np.linspace(0,(Px/(Mag*Mag2))*M, M*SWS)
        xx, yy=np.meshgrid(x,y)
        return xx,yy

    def getXYZGrid(parm):
        #get pixel Coordinates in [um]
        Px    =parm['Px']*1e6 #[um] Andor ixon 897(512x512) pixel size is 16um, for ixon 888 (1024x1024) it is 13um
        Mag   =parm['Mag']    # Magnification M=f_tube/f_objective
        Mag2  =parm['Mag2']   # Secondery magnification to EMCCD
        NM    =parm['NM']     # Image pixel dimentions 
        SWS   =parm['SWS']    # Software scaling factor
        zssize=parm['zssize'] #[m] z- step size between images

        N=NM[0]; M=NM[1]; O=NM[2]
        
        #
        x = np.linspace(0,(Px/(Mag*Mag2))*N, N*SWS)
        y = np.linspace(0,(Px/(Mag*Mag2))*M, M*SWS)
        z = np.linspace(0, zssize*O, O)
        xx, yy=np.meshgrid(x,y)
        xxz, zzx=np.meshgrid(x,z)
        yyz, zzy=np.meshgrid(y,z)
        return xx, yy, xxz, zzx, yyz, zzy

    def LightPropagator(Data,parm,Dxyz):
        #Coordinates convention    
        #1. We will consider the image in the sample plan thus scaling by the magnification 
        #2. The back foval plan is the one directly after objective, thus the lens used in the calculations if fLens=TLens/Mag
        #   and the pixle size is the one of the sample plan: Px/(Mag*Mag2)
        Lambda=parm['Lambda'] #[m] wave length 
        Px    =parm['Px'] #[m] Andor ixon 897(512x512) pixel size is 16um, for ixon 888 (1024x1024) it is 13um
        TLens =parm['TLens'] #[m] tube lens
        Na    =parm['Na'] # Numerical apature of objective 
        n     =parm['n'] # Refractive index of glass/oli
        Mag   =parm['Mag'] # Magnification M=f_tube/f_objective
        Mag2  =parm['Mag2'] # Secondery magnification to EMCCD
        flens =parm['flens'] # Objective focal lengh M=f_tube/f_objective    
        NM    =parm['NM'] # Camera pixel dimentions 
        zssize=parm['zssize'] #[m] z- step size between images


        N=NM[0]; M=NM[1]
        Dx=Dxyz[0]; Dy=Dxyz[1]; Dz=Dxyz[2]

        #Handeling Point 1.
        # Consider that the origo is at the center pixel an each step is a pixel size big
        x = np.linspace(-(Px/(Mag*Mag2))*N/2,(Px/(Mag*Mag2))*N/2, N)
        y = np.linspace(-(Px/(Mag*Mag2))*M/2,(Px/(Mag*Mag2))*M/2, M)
        xx, yy=np.meshgrid(x,y)
    
        #Handeling point 2:
        #Coordinates in the back focal plan of the objective
        # The discrete frequensies are:
        #  kp=k/M where M is number of pixel and k is the pixel index in [-M/2, -(M-1)/2,...,-M/2 + x/2 ,... ,-(M-(2*M-1))/2 ,-(M-2*M)/2]=[-256, ..., 256]
        #  Thus kp is in [-1/2, ... ,1/2], with steps of x/(M*2) where x in [0,..., 2*M]
        # 1/kp will then be the period of the wave in the image in pixels which will have values 2M/(x-M) which includ inf (plan illuminatio), the 2M pixels with is the lowes frequensy, and 2 pix which is the highes frequensy

        # The relation to the spatial frequency of the image plan is kx=kp/p where p is the pixel size thus
        # kx=kp/p where:
        # kx in [-1/2p, ... ,1/2p], with steps of x/(p*M*2) where x in [0,..., 512], the unit for kx is [1/m] 
        # kx are thus the frequencis in the image plan created by incoming monocromatic light at different angles before the imaging lens

        # The relation of the image plan frequencis "fx" to the "x" position in the back focal plan is given by
        # "kx*L=x/f" where kx a image spatial frequency [1/m], L wavelength [m], x position from center at the backfocal plane [m], f focal length of the lens used [m] 
        # this formula can be derived by considering the folowing:
        # 1. Consider a plan wave incoming with an angel "alpha" towards a lens of focal length "f"
        # 2. This plan wave will be focused on a spot "x" away from the center.  
        #    a. The relation between the angle "alpha" and "x" is: tan(alpha)=x/f 
        # 3. Looking at the plan wave side and consider that this plan wave originate form an image somewhere before the lens we have
        #   a. A plan wave with an angle "alpha" at the image plan coresponeds to a image plan frequency "kx"
        #   b. The preiodisity "px=1/kx" on the image plan due to the equal phace lines is given by: sin(alpha)=L/px=L*kx
        # 4. If we now assume the small angle approximation we have x/f = L*kx  
        # 5. thus x=f*L*kx where kx varying, the NA of the objective will restrict the possible values of kx
        #   a. NA=sin(alpha_max)=kx_max*L thus kx_max=NA/L, for the x,y plan and round apature we get 
        #      kx^2 + ky^2 <= (NA/L)^2    
        # 
        # Thus x=f*L*kx where f is focal length [m], L is wavelength [m], kx is spatial frequency [1/m]
        kx= np.linspace(-1,1, N)/(2*Px/(Mag*Mag2)) #[1/m]
        ky= np.linspace(-1,1, M)/(2*Px/(Mag*Mag2)) #[1/m]
        kkx, kky=np.meshgrid(kx,ky)

        u = Lambda*flens*kx
        v = Lambda*flens*ky
        uu, vv=np.meshgrid(u,v)

        # do the fourier transform
        I_IP=Data
        T_BFP=np.fft.ifft2(I_IP)
        T_BFP=np.fft.fftshift(T_BFP)
    
        # There are several restrictions on the values of k=(kx,ky,kz)
        # 1. Vecktor length: |k|^2=(n/L)^2 where n is the refractiv index end L is the wave length
        #   a. BK7, Float Glass has n=1.51-1.52 which is the same as the oil (1.515)
        # 2. Objective numerical apature restriction: kx**2 + ky**2 < (Na/L)**2 <= (n/L)^2

        #Point 2. NA BFP restriction 
        # Since our pixeled image allows for high spatial frequencis then suported by the optics
        # Calculate a mask and then remove k values outside the alowed region
        R_BFP=np.ones((T_BFP.shape))
        R_BFP[np.where(kkx**2 + kky**2 >(Na/Lambda)**2)]=0
        kkx=R_BFP*kkx
        kky=R_BFP*kky

        # Point 1. Wave vector length 
        kkz=np.sqrt( (n/Lambda)**2 - kkx**2 - kky**2 )
    
        #Do the xyz move and restrict the back focal plan 
        T_BFP=R_BFP*T_BFP*np.exp(2.0j*np.pi*(kkx*Dx + kky*Dy + kkz*Dz))

        #Going back to the image plan
        I_IP_N=np.fft.fft2(np.fft.fftshift(T_BFP))
        
        return I_IP, I_IP_N, xx,yy, T_BFP, R_BFP, uu, vv, kkx, kky

    def Plot_IP_BFP(xx,yy,I_IP,xxBFP,yyBFP, I_BFP,P_BFP ):
        #Plot
        fig, ax= plt.subplots(1,3)
        fig.set_size_inches(18, 6, forward=True)
        ax[0].contourf(xx, yy, I_IP,cmap=cm.coolwarm) #cm.gray, cm.coolwarm,
        ax[0].ticklabel_format(style='sci',scilimits=(0,0),axis='both')
        ax[0].set_ylabel("[m]")
        ax[0].set_xlabel("[m]")
        ax[0].set_title("IP, Amplitude")

        ax[1].contourf(xxBFP, yyBFP, I_BFP,cmap=cm.coolwarm) #cm.gray, cm.coolwarm
        ax[1].ticklabel_format(style='sci',scilimits=(0,0),axis='both')
        ax[1].set_ylabel("[m]")
        ax[1].set_xlabel("[m]")
        ax[1].set_title("BFP, Amplitude")

        ax[2].contourf(xxBFP, yyBFP, P_BFP,cmap=cm.coolwarm) #cm.gray, cm.coolwarm,
        ax[2].ticklabel_format(style='sci',scilimits=(0,0),axis='both')
        ax[2].set_ylabel("[m]")
        ax[2].set_xlabel("[m]")
        ax[2].set_title("BFP Phase")
        plt.show()

    def Plot_IP_comparison(xx,yy,I_IP, I_IP_N,diff_imag):
        #Plot
        fig, ax= plt.subplots(1,3)
        fig.set_size_inches(18, 6, forward=True)

        ax[0].contourf(xx, yy, I_IP,cmap=cm.coolwarm) #cm.gray, cm.coolwarm,
        ax[0].ticklabel_format(style='sci',scilimits=(0,0),axis='both')
        ax[0].set_ylabel("[m]")
        ax[0].set_xlabel("[m]")
        ax[0].set_title("IP original, Amplitude")

        ax[1].contourf(xx, yy, I_IP_N,cmap=cm.coolwarm) #cm.gray, cm.coolwarm
        ax[1].ticklabel_format(style='sci',scilimits=(0,0),axis='both')
        ax[1].set_ylabel("[m]")
        ax[1].set_xlabel("[m]")
        ax[1].set_title("IP propagated, Amplitude")

        cs=ax[2].contourf(xx, yy, diff_imag,cmap=cm.coolwarm) #cm.gray, cm.coolwarm,
        ax[2].ticklabel_format(style='sci',scilimits=(0,0),axis='both')
        ax[2].set_ylabel("[m]")
        ax[2].set_xlabel("[m]")
        ax[2].set_title("IP diff, Amplitude")
        fig.colorbar(cs,ax=ax[2])

        plt.show()

    def savetoMat(I_IP_final,parm,path,name):
        Ph_stack={}
        Ph_stack['Lambda']=parm['Lambda'] #[m] wave length 
        Ph_stack['Px']    =parm['Px']     #[m] Andor ixon 897(512x512) pixel size is 16um, for ixon 888 (1024x1024) it is 13um
        Ph_stack['TLens'] =parm['TLens']  #[m] tube lens
        Ph_stack['Na']    =parm['Na']     # Numerical apature of objective
        Ph_stack['n']     =parm['n']      # Refractive index of glass/oli
        Ph_stack['Mag']   =parm['Mag']    # Magnification M=f_tube/f_objective
        Ph_stack['Mag2']  =parm['Mag2']   # Secondery magnification to EMCCD
        Ph_stack['flens'] =parm['flens']  # Objective focal lengh M=f_tube/f_objective    
        Ph_stack['NM']    =parm['NM']     # Camera pixel dimentions 
        Ph_stack['zssize']=parm['zssize'] #[m] z-step size between images
        Ph_stack['IntencityC']=parm['IntencityC'] #Spot positions in the stack
        Ph_stack['PH_IP']=I_IP_final
        io.savemat(path + name, Ph_stack)
        print('Data saved!')

    def LoadMat(path,fname):
        Ph_stack=io.loadmat(path + fname)
        
        parm={}
        parm['Lambda'] =Ph_stack['Lambda'][0,0]     #[m] wave length 
        parm['Px']     =Ph_stack['Px'][0,0]         #[m] Andor ixon 897(512x512) pixel size is 16um, for ixon 888 (1024x1024) it is 13um
        parm['TLens']  =Ph_stack['TLens'][0,0]      #[m] tube lens
        parm['Na']     =Ph_stack['Na'][0,0]         # Numerical apature of objective
        parm['n']      =Ph_stack['n'][0,0]          # Refractive index of glass/oli
        parm['Mag']    =Ph_stack['Mag'][0,0]        # Magnification M=f_tube/f_objective
        parm['Mag2']   =Ph_stack['Mag2'][0,0]       # Secondery magnification to EMCCD
        parm['flens']  =Ph_stack['flens'][0,0]      # Objective focal lengh M=f_tube/f_objective    
        parm['NM']     =Ph_stack['NM'][0,:]         # Camera pixel dimentions 
        parm['zssize'] =Ph_stack['zssize']          #[m] z-step size between images
        try:
            parm['IntencityC'] =Ph_stack['IntencityC']  #Spot positions in the stack
        except:
            parm['IntencityC'] =np.zeros((3,3))  #Spot positions in the stack

        I_IP_final=Ph_stack['PH_IP']
        
        print('Data loaded!')

        return I_IP_final, parm

    def resizeImage(ParmSubBundle):
        """
        Function to resize image 
        Maybe one should try: cv2.INTER_CUBIC  insted of cv2.INTER_LANCZOS4
        """
        res=cv2.resize(ParmSubBundle[0],ParmSubBundle[1], interpolation = cv2.INTER_CUBIC)[ParmSubBundle[2][1]:ParmSubBundle[2][1]+2*ParmSubBundle[3],ParmSubBundle[2][0]:ParmSubBundle[2][0]+2*ParmSubBundle[3]]

        return res                    

    def zInterpolator(LF, SWSZ, parm_PR,p='Forward',debug=0):
        """
        Input
        LF:         light filed, complex matrix of the form (ny,nx,nz) where one which to expand nz to Nz=nz*SWSZ 
        SWSZ:       amount of times to increase the z resolution by interpolation through light propagation 
        parm_PR:    Phase retrival parameters
            'Lambda': [m] Wavelength
            'Mag'   : [#] Magnification of objective with tublens  
            'Mag2'  : [#] Extra magnification after pinhole 
            'NM'    : [#] LF image size (nx,ny)
            'Na'    : [#] Numerical apature of the opbjective  
            'Px'    : [m] Pixel size of the LF image (usaly the same as the camera that was used)
            'TLens' : [m] Tublens focal length  
            'flens' : [m] Objective focal length (usaly: TLens/Mag)
            'n'     : [#] Refractive index  
            'zssize': [m] Step size between z images in LF
        p:          which direction to evaluate the image: Forward, Backward, WeightedTowWay
        """
        # Add z resolution

        zssize=parm_PR['zssize']
        Dz=zssize/SWSZ #[m]
        IP=np.zeros((LF.shape[0],LF.shape[1],LF.shape[2]*SWSZ))
        for kk in range(LF.shape[2]):
            for kkk in range(SWSZ):
                if p=='Forward':
                    Dxyz=[0,0,kkk*Dz] #[m]
                    _, LF_N, _,_, _, _, _, _, _, _ = PSF_Lib.LightPropagator(LF[:,:,kk],parm_PR,Dxyz)
                    IP[:,:,kk*SWSZ+kkk]=np.absolute(LF_N)**2  
                elif p=='Backward':                
                    Dxyz=[0,0,(kkk-SWSZ)*Dz] #[m]
                    _, LF_N, _,_, _, _, _, _, _, _ = PSF_Lib.LightPropagator(LF[:,:,kk+1],parm_PR,Dxyz)
                    IP[:,:,kk*SWSZ+kkk]=np.absolute(LF_N)**2
                elif p=='WeightedTwoWay':
                    #Forward
                    Dxyz=[0,0,kkk*Dz] #[m]
                    _, LF_N, _,_, _, _, _, _, _, _ = PSF_Lib.LightPropagator(LF[:,:,kk],parm_PR,Dxyz)
                    IP[:,:,kk*SWSZ+kkk]=np.absolute(LF_N)**2  
                    if kk<LF.shape[2]-1:
                        #Backward
                        Dxyz=[0,0,(kkk-SWSZ)*Dz] #[m]
                        _, LF_N, _,_, _, _, _, _, _, _ = PSF_Lib.LightPropagator(LF[:,:,kk+1],parm_PR,Dxyz)
                        #Weighted 
                        IP[:,:,kk*SWSZ+kkk]=((SWSZ-kkk)*IP[:,:,kk*SWSZ+kkk] + kkk*np.absolute(LF_N)**2)/SWSZ  
                else: 
                    print("wrong input to zInterpolator")

        if debug: #Debug
            for i in range(IP.shape[2]):
                fig = plt.figure(constrained_layout=True)
                ax0=fig.add_subplot()
                cs0=ax0.imshow(IP[:,:,i],cmap='gray')
        
                save=1
                if save:
                    file="T" + "_pos_" + str(i) + '.tif'
                    path= "C:/Users/elias/Desktop/Shadowfax/Data/"
                    fig.savefig(path + file)
                    plt.close()
                else:
                    plt.show()

        return IP

    def get_Smatrix(ScanProject,IP,xiP,yiP,xiE,yiE,xiO,yiO,xPinOffset,yPinOffset,SaveSmatrix=''):
        """
        Creates the S matrix 
        ScanProject:            Used for getting parameters and save folder         
        IP:                     PSF image stack with z interpolation already implemented
        xiP,yiP:                [um] x and y piezo stage position 
        xiE,yiE:                [um] x and y EOD position 
        xiO,yiO:                [um] x and y EOD Offset positon for the tracking pattern in x,y plan
        xPinOffset, yPinOffset: [um] PSF image offset to center each part of the PSFs

        SaveSmatrix: If string '' then not used, else the string will be appended to the seved file
                     Data will be saved in the ScanProject.Files['AnalysisConfig'] folder             
        """

        parm={}
        parm['Px']    = ScanProject.SetupSpecifics['Px']            
        parm['Mag']   = ScanProject.SetupSpecifics['Mag']           
        parm['Mag2']  = ScanProject.SetupSpecifics['Mag2']                     
        parm['SWS']   = ScanProject.Ph_Retrieval['XYSWScalefactor'] # Software scaling factor
        parm['zssize']= ScanProject.Ph_Retrieval['zssize']        #[m] z- step size between images in PSF stack
        parm['SWSZ']  = ScanProject.Ph_Retrieval['ZSWScalefactor'] # Software scaling factor for z


        #Pixel convertion factors  
        px2um=parm['Px']*1e6/(parm['Mag']*parm['Mag2'])  # from camera pixel to um
        spx2um=px2um/parm['SWS']                         # from super resolution to um 
        
        zInFocusIndex=ScanProject.Ph_Retrieval['IntencityC'][:,2]
        
        m =int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use

        #Get pinhole size in [um] 
        PholeDia = ScanProject.SetupSpecifics['PholeDia']/ScanProject.SetupSpecifics['Mag'] #size of confocal pinhole in the sample plan

        #Get total offset of where the beam is in relation to the pinhole 
        #We start by placing PSF image corner as the origo and 
        #1. move to the center of the beam 
        #2. add offsets for the OED and tracking pattern 
        #The pinhole is now at the origo in our coordinate system 
        #Thus  xOffset_um is the corner coordinate  and xOffset_spx will be amount of pixels one need to move to get from the oregin to the corner 
        print ("Max EOD   move: x=" + str(np.round((np.max(xiE)-np.min(xiE))*1e3, decimals=3)) + "nm, y=" + str(np.round((np.max(yiE)-np.min(yiE))*1e3, decimals=3)) + "nm")
        print ("Max Piezo move: x=" + str(np.round((np.max(xiP)-np.min(xiP))*1e3, decimals=3)) + "nm, y=" + str(np.round((np.max(yiP)-np.min(yiP))*1e3, decimals=3)) + "nm")

        xOffset_um= -xPinOffset -xiE + xiO
        yOffset_um= -yPinOffset -yiE + yiO
        xOffset_spx= xOffset_um/spx2um
        yOffset_spx= yOffset_um/spx2um

        # Add the tip/tilt mirror offset and remove the first in the sequens which will be at the center of the gird 
        #Thus the center of the grid is at (xiP[0],yiP[0]) but the pinhole position migh move (but not for the first)
        #Question Elias: is this really correct since we demodulate the piezo mirror, thus the IP cut our should not depend on the piezo mirror or?
        #                This will not affect the trajecotry estimation since it is a small movement per estimation point, but might if long averageis are taken 
        xPMaskOffset= xiP - xiP[0]
        yPMaskOffset= yiP - yiP[0]
        xOffsetTotal_um = xOffset_um + xPMaskOffset
        yOffsetTotal_um = yOffset_um + yPMaskOffset

        xOffsetTotal_px= np.floor(np.abs(xOffsetTotal_um/px2um)).astype('int')
        yOffsetTotal_px= np.floor(np.abs(yOffsetTotal_um/px2um)).astype('int')

        xOffsetTotal_spx= np.round(np.abs(xOffsetTotal_um/spx2um) - xOffsetTotal_px*parm['SWS']).astype('int')
        yOffsetTotal_spx= np.round(np.abs(yOffsetTotal_um/spx2um) - yOffsetTotal_px*parm['SWS']).astype('int')

        #Create image um xyz grid
        PS_px=ScanProject.Ph_Retrieval['CropS']  
        PS_spx=ScanProject.Ph_Retrieval['CropS']*parm['SWS']

        x=xiP[0] + (np.array(range(2*PS_spx))-PS_spx)*spx2um
        y=yiP[0] + (np.array(range(2*PS_spx))-PS_spx)*spx2um
        z=(np.array(range(IP.shape[2])) -zInFocusIndex[0]*parm['SWSZ'] ) *parm['zssize']*1e6/parm['SWSZ']
        xx,yy,zz=np.meshgrid(x,y,z)

        PMask=np.ones((2*PS_spx,2*PS_spx,m))
        S=np.zeros((2*PS_spx,2*PS_spx,IP.shape[2],m))    
        
        kk_range=range(IP.shape[2])
        for k in range(m):
            #print('Start constructing sensing matrix: ' + str(k))
            IPpxCut=IP[yOffsetTotal_px[k]-PS_px:yOffsetTotal_px[k]+PS_px+1,
                        xOffsetTotal_px[k]-PS_px:xOffsetTotal_px[k]+PS_px+1,:] #adding 1px to be able to fine adjust subpixle in next step
            PMask[:,:,k][np.where(((xx[:,:,0]-xiP[k])**2 + (yy[:,:,0]-yiP[k])**2 ) > (PholeDia/2)**2)]=np.nan
                
            #Resize image and make a cut out 
            #This part is prepered for parallelization
            #1. We prepera all arguments that are neessesery for the resize acoriding to the function PSF.resizeImage
            # The original implementation is:
            #     S[:,:,kk,k]= cv2.resize(IPpxCut[:,:,kk],((2*PS_px+1)*parm['SWS'],(2*PS_px+1)*parm['SWS']), 
            #                             interpolation = cv2.INTER_LANCZOS4)[yOffsetTotal_spx[k]:yOffsetTotal_spx[k]+2*PS_spx, xOffsetTotal_spx[k]:xOffsetTotal_spx[k]+2*PS_spx]
            #     
            #The ParmBundle which is the argument bundel is a tuple of tuples since the map function needs tuples 
            #
            #2. We make the resize 
            #   a. ether direct with the function  PSF.resizeImage
            #   b. or with the  ProcessPoolExecutor which will parallelize the exicution 
            #      On my (Elias) computor this did slowdown the exicution from 0.15s to 0.5s. I do not know why.
            #      I have tryed to move out the "with" statment outside the "k" loop, but this did not help. 

            ParmBundle=()
            for kk in kk_range:
                ParmBundle =(*ParmBundle, (IPpxCut[:,:,kk],
                                ((2*PS_px+1)*parm['SWS'],(2*PS_px+1)*parm['SWS']),
                                (xOffsetTotal_spx[k], yOffsetTotal_spx[k]),
                                PS_spx))
            for kk in kk_range:
                S[:,:,kk,k]= PSF_Lib.resizeImage(ParmBundle[kk])

            if 0:#Debug plot 
                    # Where we cut out is marked by blue crosses 
                    fig, ax= plt.subplots(1,4)
                    fig.set_size_inches(4*4, 4, forward=True)
                    zi=2 #Look in the ini file for the infocus plan t use 
                    ax[0].contourf(IP[:,:,int(zInFocusIndex[zi])],alpha=1.0)
                    ax[0].plot(xOffsetTotal_px[k],yOffsetTotal_px[k],'r*')
                    ax[0].plot(xOffsetTotal_px[k]-PS_px,yOffsetTotal_px[k]-PS_px,'b*')
                    ax[0].plot(xOffsetTotal_px[k]+PS_px,yOffsetTotal_px[k]-PS_px,'b*')
                    ax[0].plot(xOffsetTotal_px[k]-PS_px,yOffsetTotal_px[k]+PS_px,'b*')
                    ax[0].plot(xOffsetTotal_px[k]+PS_px,yOffsetTotal_px[k]+PS_px,'b*')

                    ax[1].contourf(IPpxCut[:,:,int(zInFocusIndex[zi])],alpha=1.0)
                    ax[2].contourf(S[:,:,int(zInFocusIndex[zi]),k],alpha=1.0)
                    ax[3].contourf(PMask[:,:,k]*S[:,:,int(zInFocusIndex[zi]),k],alpha=1.0)
                    plt.show()

        if 0:
            # View of each S projected to the xy, xz and yz plan
            fig, ax= plt.subplots(1,3)
            fig.set_size_inches(3*4, 4, forward=True)
            ax[0].contourf(np.sum(np.sum(S,axis=3),axis=2),alpha=1.0)
            ax[1].contourf(np.sum(np.sum(S,axis=3),axis=1),alpha=1.0)
            ax[2].contourf(np.sum(np.sum(S,axis=3),axis=0),alpha=1.0)
            plt.show()
        
        PMaskc=np.moveaxis(np.tile(PMask,(S.shape[2],1,1,1)),0,-2)   
        S = PMaskc*S

        if np.sum(np.where(S<0))>0:
           print("S negativa")
           S[np.where(S<0)]=0


        if SaveSmatrix!='':# Save Smatrix if needed
            SData={}
            SData[0]=S
            _DS.saveTrajtoMat(ScanProject,SData,'SMatrix_' + SaveSmatrix)

        return S,xx,yy,zz

    def get_Smatrix2(ScanProject,IP,xiP,yiP,ziP,xiE,yiE,xiO,yiO,xPinOffset,yPinOffset,SaveSmatrix=''):
        """
        This version assumes that there is a piezo stage for z

        Creates the S matrix 
        ScanProject:            Used for getting parameters and save folder         
        IP:                     PSF image stack with z interpolation already implemented
        xiP,yiP,ziP:            [um] x, y and z piezo stage position 
        xiE,yiE:                [um] x and y EOD position 
        xiO,yiO:                [um] x and y EOD Offset positon for the tracking pattern in x,y plan
        xPinOffset, yPinOffset: [um] PSF image offset to center each part of the PSFs

        SaveSmatrix: If string '' then not used, else the string will be appended to the seved file
                     Data will be saved in the ScanProject.Files['AnalysisConfig'] folder             
        """

        parm={}
        parm['Px']    = ScanProject.SetupSpecifics['Px']            
        parm['Mag']   = ScanProject.SetupSpecifics['Mag']           
        parm['Mag2']  = ScanProject.SetupSpecifics['Mag2']                     
        parm['SWS']   = ScanProject.Ph_Retrieval['XYSWScalefactor'] # Software scaling factor
        parm['zssize']= ScanProject.Ph_Retrieval['zssize']        #[m] z- step size between images in PSF stack
        parm['SWSZ']  = ScanProject.Ph_Retrieval['ZSWScalefactor'] # Software scaling factor for z


        #Pixel convertion factors  
        px2um=parm['Px']*1e6/(parm['Mag']*parm['Mag2'])  # from camera pixel to um
        spx2um=px2um/parm['SWS']                         # from super resolution to um 
        
        zInFocusIndex=ScanProject.Ph_Retrieval['IntencityC'][:,2]
        
        m =int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use

        #Get pinhole size in [um] 
        PholeDia = ScanProject.SetupSpecifics['PholeDia']/ScanProject.SetupSpecifics['Mag'] #size of confocal pinhole in the sample plan

        #Get total offset of where the beam is in relation to the pinhole 
        #We start by placing PSF image corner as the origo and 
        #1. move to the center of the beam 
        #2. add offsets for the OED and tracking pattern 
        #The pinhole is now at the origo in our coordinate system 
        #Thus  xOffset_um is the corner coordinate  and xOffset_spx will be amount of pixels one need to move to get from the oregin to the corner 
        print ("Max EOD   move: x=" + str(np.round((np.max(xiE)-np.min(xiE))*1e3, decimals=3)) + "nm, y=" + str(np.round((np.max(yiE)-np.min(yiE))*1e3, decimals=3)) + "nm")
        print ("Max Piezo move: x=" + str(np.round((np.max(xiP)-np.min(xiP))*1e3, decimals=3)) + "nm, y=" + str(np.round((np.max(yiP)-np.min(yiP))*1e3, decimals=3)) + "nm")

        xOffset_um= -xPinOffset -xiE + xiO
        yOffset_um= -yPinOffset -yiE + yiO
        xOffset_spx= xOffset_um/spx2um
        yOffset_spx= yOffset_um/spx2um

        # Add the tip/tilt mirror offset and remove the first in the sequens which will be at the center of the gird 
        #Thus the center of the grid is at (xiP[0],yiP[0]) but the pinhole position migh move (but not for the first)
        #Question Elias: is this really correct since we demodulate the piezo mirror, thus the IP cut our should not depend on the piezo mirror or?
        #                This will not affect the trajecotry estimation since it is a small movement per estimation point, but might if long averageis are taken 
        xPMaskOffset= xiP - xiP[0]
        yPMaskOffset= yiP - yiP[0]
        xOffsetTotal_um = xOffset_um + xPMaskOffset
        yOffsetTotal_um = yOffset_um + yPMaskOffset

        xOffsetTotal_px= np.floor(np.abs(xOffsetTotal_um/px2um)).astype('int')
        yOffsetTotal_px= np.floor(np.abs(yOffsetTotal_um/px2um)).astype('int')

        xOffsetTotal_spx= np.round(np.abs(xOffsetTotal_um/spx2um) - xOffsetTotal_px*parm['SWS']).astype('int')
        yOffsetTotal_spx= np.round(np.abs(yOffsetTotal_um/spx2um) - yOffsetTotal_px*parm['SWS']).astype('int')

        #Create image um xyz grid
        PS_px=ScanProject.Ph_Retrieval['CropS']  
        PS_spx=ScanProject.Ph_Retrieval['CropS']*parm['SWS']

        x=xiP[0] + (np.array(range(2*PS_spx))-PS_spx)*spx2um
        y=yiP[0] + (np.array(range(2*PS_spx))-PS_spx)*spx2um
        z=ziP[0] + (np.array(range(IP.shape[2])) -zInFocusIndex[1]*parm['SWSZ'] ) *parm['zssize']*1e6/parm['SWSZ']
        xx,yy,zz=np.meshgrid(x,y,z)

        PMask=np.ones((2*PS_spx,2*PS_spx,m))
        S=np.zeros((2*PS_spx,2*PS_spx,IP.shape[2],m))    
        
        kk_range=range(IP.shape[2])
        for k in range(m):
            #print('Start constructing sensing matrix: ' + str(k))
            IPpxCut=IP[yOffsetTotal_px[k]-PS_px:yOffsetTotal_px[k]+PS_px+1,
                        xOffsetTotal_px[k]-PS_px:xOffsetTotal_px[k]+PS_px+1,:] #adding 1px to be able to fine adjust subpixle in next step
            PMask[:,:,k][np.where(((xx[:,:,0]-xiP[k])**2 + (yy[:,:,0]-yiP[k])**2 ) > (PholeDia/2)**2)]=np.nan
                
            #Resize image and make a cut out 
            #This part is prepered for parallelization
            #1. We prepera all arguments that are neessesery for the resize acoriding to the function PSF.resizeImage
            # The original implementation is:
            #     S[:,:,kk,k]= cv2.resize(IPpxCut[:,:,kk],((2*PS_px+1)*parm['SWS'],(2*PS_px+1)*parm['SWS']), 
            #                             interpolation = cv2.INTER_LANCZOS4)[yOffsetTotal_spx[k]:yOffsetTotal_spx[k]+2*PS_spx, xOffsetTotal_spx[k]:xOffsetTotal_spx[k]+2*PS_spx]
            #     
            #The ParmBundle which is the argument bundel is a tuple of tuples since the map function needs tuples 
            #
            #2. We make the resize 
            #   a. ether direct with the function  PSF.resizeImage
            #   b. or with the  ProcessPoolExecutor which will parallelize the exicution 
            #      On my (Elias) computor this did slowdown the exicution from 0.15s to 0.5s. I do not know why.
            #      I have tryed to move out the "with" statment outside the "k" loop, but this did not help. 

            ParmBundle=()
            for kk in kk_range:
                ParmBundle =(*ParmBundle, (IPpxCut[:,:,kk],
                                ((2*PS_px+1)*parm['SWS'],(2*PS_px+1)*parm['SWS']),
                                (xOffsetTotal_spx[k], yOffsetTotal_spx[k]),
                                PS_spx))
            
            if 0: #paralelization of resizeImage if true, use ThreadPoolExecutor on standard computer with threading, use ProcessPoolExecutor on uppmax
                with concurrent.futures.ThreadPoolExecutor() as executor: #ThreadPoolExecutor / ProcessPoolExecutor
                    for kk, res in zip(kk_range, executor.map(PSF_Lib.resizeImage, ParmBundle)):
                        S[:,:,kk,k]=res
            else:
                for kk in kk_range:
                    S[:,:,kk,k]= PSF_Lib.resizeImage(ParmBundle[kk])

            if 0:#Debug plot 
                    # Where we cut out is marked by blue crosses 
                    fig, ax= plt.subplots(1,4)
                    fig.set_size_inches(4*4, 4, forward=True)
                    zi=2 #Look in the ini file for the infocus plan t use 
                    ax[0].contourf(IP[:,:,int(zInFocusIndex[zi])],alpha=1.0)
                    ax[0].plot(xOffsetTotal_px[k],yOffsetTotal_px[k],'r*')
                    ax[0].plot(xOffsetTotal_px[k]-PS_px,yOffsetTotal_px[k]-PS_px,'b*')
                    ax[0].plot(xOffsetTotal_px[k]+PS_px,yOffsetTotal_px[k]-PS_px,'b*')
                    ax[0].plot(xOffsetTotal_px[k]-PS_px,yOffsetTotal_px[k]+PS_px,'b*')
                    ax[0].plot(xOffsetTotal_px[k]+PS_px,yOffsetTotal_px[k]+PS_px,'b*')

                    ax[1].contourf(IPpxCut[:,:,int(zInFocusIndex[zi])],alpha=1.0)
                    ax[2].contourf(S[:,:,int(zInFocusIndex[zi]),k],alpha=1.0)
                    ax[3].contourf(PMask[:,:,k]*S[:,:,int(zInFocusIndex[zi]),k],alpha=1.0)
                    plt.show()

        if 0:#Debug plot 
            # View of each S projected to the xy, xz and yz plan
            fig, ax= plt.subplots(1,3)
            fig.set_size_inches(3*4, 4, forward=True)
            ax[0].contourf(np.sum(np.sum(S,axis=3),axis=2),alpha=1.0)
            ax[1].contourf(np.sum(np.sum(S,axis=3),axis=1),alpha=1.0)
            ax[2].contourf(np.sum(np.sum(S,axis=3),axis=0),alpha=1.0)
            plt.show()
        
        PMaskc=np.moveaxis(np.tile(PMask,(S.shape[2],1,1,1)),0,-2)   
        S = PMaskc*S

        if np.sum(np.where(S<0))>0:
           print("S negativa")
           S[np.where(S<0)]=0


        if SaveSmatrix!='':# Save Smatrix if needed
            SData={}
            SData[0]=S
            _DS.saveTrajtoMat(ScanProject,SData,'SMatrix_' + SaveSmatrix)

        return S,xx,yy,zz

    def get_Smatrix3(ScanProject,IP,xiO,yiO,xPinOffset,yPinOffset,SaveSmatrix=''):
        """
        This version assumes:
            0. There is no movement in the EOD and Piezo stage, thus over the estimation shoots we have a fix laser pattern
            1. That there is a piezo stage for z
            2. No pinhole is added since the tophat prior has the same function             
        

        Creates the S matrix 
        ScanProject:            Used for getting parameters and save folder         
        IP:                     PSF image stack with z interpolation already implemented
        xiP,yiP,ziP:            [um] x, y and z piezo stage position 
        xiE,yiE:                [um] x and y EOD position 
        xiO,yiO:                [um] x and y EOD Offset positon for the tracking pattern in x,y plan
        xPinOffset, yPinOffset: [um] PSF image offset to center each part of the PSFs

        SaveSmatrix: If string '' then not used, else the string will be appended to the seved file
                     Data will be saved in the ScanProject.Files['AnalysisConfig'] folder             
        """

        parm={}
        parm['Px']    = ScanProject.SetupSpecifics['Px']            
        parm['Mag']   = ScanProject.SetupSpecifics['Mag']           
        parm['Mag2']  = ScanProject.SetupSpecifics['Mag2']                     
        parm['SWS']   = ScanProject.Ph_Retrieval['XYSWScalefactor'] # Software scaling factor
        parm['zssize']= ScanProject.Ph_Retrieval['zssize']        #[m] z- step size between images in PSF stack
        parm['SWSZ']  = ScanProject.Ph_Retrieval['ZSWScalefactor'] # Software scaling factor for z


        #Pixel convertion factors  
        px2um=parm['Px']*1e6/(parm['Mag']*parm['Mag2'])  # from camera pixel to um
        spx2um=px2um/parm['SWS']                         # from super resolution to um 
        
        zInFocusIndex=ScanProject.Ph_Retrieval['IntencityC'][:,2]
        
        m =int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use

        #Get total offset of where the beam is in relation to the pinhole 
        #We start by placing PSF image corner as the origo and 
        #1. move to the center of the beam 
        #2. add offsets for the OED and tracking pattern 
        #The pinhole is now at the origo in our coordinate system 
        #Thus  xOffset_um is the corner coordinate  and xOffset_spx will be amount of pixels one need to move to get from the oregin to the corner 

        xOffset_um= -xPinOffset + xiO 
        yOffset_um= -yPinOffset + yiO 
        xOffset_spx= xOffset_um/spx2um
        yOffset_spx= yOffset_um/spx2um

        # Add the tip/tilt mirror offset and remove the first in the sequens which will be at the center of the gird 
        #Thus the center of the grid is at (xiP[0],yiP[0]) but the pinhole position migh move (but not for the first)
        #Question Elias: is this really correct since we demodulate the piezo mirror, thus the IP cut our should not depend on the piezo mirror or?
        #                This will not affect the trajecotry estimation since it is a small movement per estimation point, but might if long averageis are taken 
        #DEL#xPMaskOffset= xiP - xiP[0]
        #DEL#yPMaskOffset= yiP - yiP[0]
        xOffsetTotal_um = xOffset_um #DEL#+ xPMaskOffset
        yOffsetTotal_um = yOffset_um #DEL#+ yPMaskOffset

        xOffsetTotal_px= np.floor(np.abs(xOffsetTotal_um/px2um)).astype('int')
        yOffsetTotal_px= np.floor(np.abs(yOffsetTotal_um/px2um)).astype('int')

        xOffsetTotal_spx= np.round(np.abs(xOffsetTotal_um/spx2um) - xOffsetTotal_px*parm['SWS']).astype('int')
        yOffsetTotal_spx= np.round(np.abs(yOffsetTotal_um/spx2um) - yOffsetTotal_px*parm['SWS']).astype('int')

        #Create image um xyz grid
        PS_px=ScanProject.Ph_Retrieval['CropS']  
        PS_spx=ScanProject.Ph_Retrieval['CropS']*parm['SWS']

        x=(np.array(range(2*PS_spx))-PS_spx)*spx2um #DEL# +xiP[0]
        y=(np.array(range(2*PS_spx))-PS_spx)*spx2um #DEL# +yiP[0]
        z=(np.array(range(IP.shape[2])) -zInFocusIndex[1]*parm['SWSZ'] ) *parm['zssize']*1e6/parm['SWSZ'] #DEL#+ ziP[0]
        xx,yy,zz=np.meshgrid(x,y,z)

        S=np.zeros((2*PS_spx,2*PS_spx,IP.shape[2],m))    
        
        priorOffset=[0,0]
        kk_range=range(IP.shape[2])
        for k in range(m):
            #print('Start constructing sensing matrix: ' + str(k))
            IPpxCut=IP[yOffsetTotal_px[k]-PS_px:yOffsetTotal_px[k]+PS_px+1,
                        xOffsetTotal_px[k]-PS_px:xOffsetTotal_px[k]+PS_px+1,:] #adding 1px to be able to fine adjust subpixle in next step
                
            #Resize image and make a cut out 
            #This part is prepered for parallelization
            #1. We prepera all arguments that are neessesery for the resize acoriding to the function PSF.resizeImage
            # The original implementation is:
            #     S[:,:,kk,k]= cv2.resize(IPpxCut[:,:,kk],((2*PS_px+1)*parm['SWS'],(2*PS_px+1)*parm['SWS']), 
            #                             interpolation = cv2.INTER_LANCZOS4)[yOffsetTotal_spx[k]:yOffsetTotal_spx[k]+2*PS_spx, xOffsetTotal_spx[k]:xOffsetTotal_spx[k]+2*PS_spx]
            #     
            #The ParmBundle which is the argument bundel is a tuple of tuples since the map function needs tuples 
            #
            #2. We make the resize 
            #   a. ether direct with the function  PSF.resizeImage
            #   b. or with the  ProcessPoolExecutor which will parallelize the exicution 
            #      On my (Elias) computor this did slowdown the exicution from 0.15s to 0.5s. I do not know why.
            #      I have tryed to move out the "with" statment outside the "k" loop, but this did not help. 

            ParmBundle=()
            for kk in kk_range:
                ParmBundle =(*ParmBundle, (IPpxCut[:,:,kk],
                                ((2*PS_px+1)*parm['SWS'],(2*PS_px+1)*parm['SWS']),
                                (xOffsetTotal_spx[k], yOffsetTotal_spx[k]),
                                PS_spx))
            
            if 0: #paralelization of resizeImage if true, use ThreadPoolExecutor on standard computer with threading, use ProcessPoolExecutor on uppmax
                with concurrent.futures.ThreadPoolExecutor() as executor: #ThreadPoolExecutor / ProcessPoolExecutor
                    for kk, res in zip(kk_range, executor.map(PSF_Lib.resizeImage, ParmBundle)):
                        S[:,:,kk,k]=res
            else:
                for kk in kk_range:
                    S[:,:,kk,k]= PSF_Lib.resizeImage(ParmBundle[kk])

            if 0:#Debug plot 
                    # Where we cut out is marked by blue crosses 
                    fig, ax= plt.subplots(1,4)
                    fig.set_size_inches(4*4, 4, forward=True)
                    zi=2 #Look in the ini file for the infocus plan t use 
                    ax[0].contourf(IP[:,:,int(zInFocusIndex[zi])],alpha=1.0)
                    ax[0].plot(xOffsetTotal_px[k],yOffsetTotal_px[k],'r*')
                    ax[0].plot(xOffsetTotal_px[k]-PS_px,yOffsetTotal_px[k]-PS_px,'b*')
                    ax[0].plot(xOffsetTotal_px[k]+PS_px,yOffsetTotal_px[k]-PS_px,'b*')
                    ax[0].plot(xOffsetTotal_px[k]-PS_px,yOffsetTotal_px[k]+PS_px,'b*')
                    ax[0].plot(xOffsetTotal_px[k]+PS_px,yOffsetTotal_px[k]+PS_px,'b*')

                    ax[1].contourf(IPpxCut[:,:,int(zInFocusIndex[zi])],alpha=1.0)
                    ax[2].contourf(S[:,:,int(zInFocusIndex[zi]),k],alpha=1.0)
                    ax[3].contourf(S[:,:,int(zInFocusIndex[zi]),k],alpha=1.0)
                    plt.show()

        if 0:#Debug plot 
            # View of each S projected to the xy, xz and yz plan
            fig, ax= plt.subplots(1,3)
            fig.set_size_inches(3*4, 4, forward=True)
            ax[0].contourf(np.sum(np.sum(S,axis=3),axis=2),alpha=1.0)
            ax[1].contourf(np.sum(np.sum(S,axis=3),axis=1),alpha=1.0)
            ax[2].contourf(np.sum(np.sum(S,axis=3),axis=0),alpha=1.0)
            plt.show()

        if np.sum(np.where(S<0))>0:
           print("S negativa")
           S[np.where(S<0)]=0


        if SaveSmatrix!='':# Save Smatrix if needed
            SData={}
            SData[0]=S
            _DS.saveTrajtoMat(ScanProject,SData,'SMatrix_' + SaveSmatrix)

        return S,xx,yy,zz

    def get_Smatrix_parallel2(ScanProject,IP,xiP,yiP,xiE,yiE,xiO,yiO,xPinOffset,yPinOffset,SaveSmatrix='',InterpolMethod="RegularGridInterpolator"):
        """
        OBS! This is a test version to see if one can do a 3D interpolation
             Currently 202021117 it is not used since it was slow and gives bad results
        Creates the S matrix 
        ScanProject:            Used for getting parameters and save folder         
        IP:                     PSF image stack with no z interpolation, all interpolations are done here with Scipy 3D RegularGridInterpolator (works) or spline (still need some work)
        xiP,yiP:                [um] x and y piezo stage position 
        xiE,yiE:                [um] x and y EOD position 
        xiO,yiO:                [um] x and y EOD Offset positon for the tracking pattern in x,y plan
        xPinOffset, yPinOffset: [um] PSF image offset to center each part of the PSFs

        SaveSmatrix: If string '' then not used, else the string will be appended to the seved file
                     Data will be saved in the ScanProject.Files['AnalysisConfig'] folder             
        """

        parm={}
        parm['Px']    = ScanProject.SetupSpecifics['Px']            
        parm['Mag']   = ScanProject.SetupSpecifics['Mag']           
        parm['Mag2']  = ScanProject.SetupSpecifics['Mag2']                     
        parm['SWS']   = ScanProject.Ph_Retrieval['XYSWScalefactor'] # Software scaling factor
        parm['zssize']= ScanProject.Ph_Retrieval['zssize']        #[m] z- step size between images in PSF stack
        parm['SWSZ']  = ScanProject.Ph_Retrieval['ZSWScalefactor'] # Software scaling factor for z

        #Pixel convertion factors  
        px2um=parm['Px']*1e6/(parm['Mag']*parm['Mag2'])  # from camera pixel to um
        spx2um=px2um/parm['SWS']                         # from super resolution to um 
        
        zInFocusIndex=ScanProject.Ph_Retrieval['IntencityC'][:,2]
        
        m =int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use

        #Get pinhole size in [um] 
        PholeDia = ScanProject.SetupSpecifics['PholeDia']/ScanProject.SetupSpecifics['Mag'] #size of confocal pinhole in the sample plan

        #Get total offset of where the beam is in relation to the pinhole 
        #We start by placing PSF image corner as the origo and 
        #1. move to the center of the beam 
        #2. add offsets for the OED and tracking pattern 
        #The pinhole is now at the origo in our coordinate system 
        #Thus  xOffset_um is the corner coordinate  and xOffset_spx will be amount of pixels one need to move to get form the oregin to the corner 
        xOffset_um= -xPinOffset -xiE + xiO
        yOffset_um= -yPinOffset -yiE + yiO
        xOffset_spx= xOffset_um/spx2um
        yOffset_spx= yOffset_um/spx2um

        # Add the tip/tilt mirror offset and remove the first in the sequens which will be at the center of the gird 
        #Thus the center of the grid is at (xiP[0],yiP[0]) but the pinhole position migh move (but not for the first)
        xPMaskOffset= xiP - xiP[0]
        yPMaskOffset= yiP - yiP[0]
        xOffsetTotal_um = xOffset_um + xPMaskOffset
        yOffsetTotal_um = yOffset_um + yPMaskOffset

        xOffsetTotal_px= np.floor(np.abs(xOffsetTotal_um/px2um)).astype('int')
        yOffsetTotal_px= np.floor(np.abs(yOffsetTotal_um/px2um)).astype('int')

        xOffsetTotal_spx= np.round(np.abs(xOffsetTotal_um/spx2um) - xOffsetTotal_px*parm['SWS']).astype('int')
        yOffsetTotal_spx= np.round(np.abs(yOffsetTotal_um/spx2um) - yOffsetTotal_px*parm['SWS']).astype('int')

        #Create image um xyz grid
        PS_px=ScanProject.Ph_Retrieval['CropS']  
        PS_spx=ScanProject.Ph_Retrieval['CropS']*parm['SWS']

        x=xiP[0] + (np.array(range(2*PS_spx))-PS_spx)*spx2um
        y=yiP[0] + (np.array(range(2*PS_spx))-PS_spx)*spx2um
        z=(np.array(range(IP.shape[2]*parm['SWSZ'])) -zInFocusIndex[0] ) *parm['zssize']*1e6/parm['SWSZ']  #Stepsize 
        xx,yy,zz=np.meshgrid(x,y,z)

        PMask=np.ones((2*PS_spx,2*PS_spx,m))
        S=np.zeros((2*PS_spx,2*PS_spx,IP.shape[2]*parm['SWSZ'],m))    
        
            #interpolaition index space before and after interpolation, x and y are the same the we only use x
        xi = np.linspace(0,2*PS_px,2*PS_px+1)
        zi = np.linspace(0,IP.shape[2]-1,IP.shape[2])
        xxi,yyi,zzi=np.meshgrid(xi,xi,zi)

        xn = np.linspace(0,2*PS_px,(2*PS_px+1)*parm['SWS'])
        zn = np.linspace(0,IP.shape[2]-1,(IP.shape[2])*parm['SWSZ'])
        xxn,yyn,zzn=np.meshgrid(xn,xn,zn)
        xyz_points=np.transpose(np.array([yyn.flatten(),xxn.flatten(),zzn.flatten()]))

        for k in range(m):
            print('Start constructing sensing matrix: ' + str(k))
            IPpxCut=IP[yOffsetTotal_px[k]-PS_px:yOffsetTotal_px[k]+PS_px+1,
                        xOffsetTotal_px[k]-PS_px:xOffsetTotal_px[k]+PS_px+1,:] #adding 1px to be able to fine adjust subpixle in next step
            PMask[:,:,k][np.where(((xx[:,:,0]-xiP[k])**2 + (yy[:,:,0]-yiP[k])**2 ) > (PholeDia/2)**2)]=np.nan
                           
            #Resize image and make a cut out using 3D linear interpolation
            # Spline interpolation migh also be something to test 
            # See: https://stackoverflow.com/questions/46040382/spline-interpolation-in-3d-in-python
            
            if InterpolMethod=="RegularGridInterpolator":
                fn = RegularGridInterpolator((xi,xi,zi), IPpxCut)
                S[:,:,:,k]=np.reshape(fn(xyz_points), xxn.shape)[yOffsetTotal_spx[k]:yOffsetTotal_spx[k]+2*PS_spx, xOffsetTotal_spx[k]:xOffsetTotal_spx[k]+2*PS_spx,:]
            elif InterpolMethod=="SlineInterpolator":
                #This need some more work
                xyzd=np.transpose(np.array([xxi.flatten(),yyi.flatten(),zzi.flatten(),IPpxCut.flatten()]))
                fn = sp.interpolate.Rbf(xyzd[:,0],xyzd[:,1],xyzd[:,2],xyzd[:,3],function='gaussian')
                S[:,:,:,k]=np.reshape(fn(xyz_points[:,0],xyz_points[:,1],xyz_points[:,2]), xxn.shape)[yOffsetTotal_spx[k]:yOffsetTotal_spx[k]+2*PS_spx, xOffsetTotal_spx[k]:xOffsetTotal_spx[k]+2*PS_spx,:]

                        
            if 0:#Debug plot 
                # Where wo we cut out is marked by the blue corsses 
                for zindex in range(S.shape[2]):
                    fig, ax= plt.subplots(1,4)
                    fig.set_size_inches(4*4, 4, forward=True)
                    ax[0].contourf(IP[:,:,int(np.floor(zindex/parm['SWSZ']))],alpha=1.0,levels=100)
                    ax[0].plot(xOffsetTotal_px[k],yOffsetTotal_px[k],'r*')
                    ax[0].plot(xOffsetTotal_px[k]-PS_px,yOffsetTotal_px[k]-PS_px,'b*')
                    ax[0].plot(xOffsetTotal_px[k]+PS_px,yOffsetTotal_px[k]-PS_px,'b*')
                    ax[0].plot(xOffsetTotal_px[k]-PS_px,yOffsetTotal_px[k]+PS_px,'b*')
                    ax[0].plot(xOffsetTotal_px[k]+PS_px,yOffsetTotal_px[k]+PS_px,'b*')

                    ax[1].contourf(IPpxCut[:,:,int(np.floor(zindex/parm['SWSZ']))],alpha=1.0,levels=100)
                    ax[2].contourf(S[:,:,zindex,k],alpha=1.0,levels=100)
                    ax[3].contourf(PMask[:,:,k]*S[:,:,zindex,k],alpha=1.0,levels=100)
                    save=1
                    if save:
                        file="T" + "_pos_" + str(zindex) + '.png'
                        path="C:/Users/elias/Desktop/Shadowfax/Data/20200909/S_beam_Pos_" + str(k) + "_lineraInterpol/"
                        fig.savefig(path + file)
                        plt.close()
                    else:
                        plt.show()

            

        if 0:
            # View of each S prjected to the xy, xz and yz plan
            fig, ax= plt.subplots(1,3)
            fig.set_size_inches(3*4, 4, forward=True)
            ax[0].contourf(np.sum(np.sum(S,axis=3),axis=2),alpha=1.0)
            ax[1].contourf(np.sum(np.sum(S,axis=3),axis=1),alpha=1.0)
            ax[2].contourf(np.sum(np.sum(S,axis=3),axis=0),alpha=1.0)
            save=1
            if save:
                file="T" + "_S_projections_" + '.png'
                path="C:/Users/elias/Desktop/Shadowfax/Data/20200909/"
                fig.savefig(path + file)
                plt.close()
            else:
                plt.show()

        
        PMaskc=np.moveaxis(np.tile(PMask,(S.shape[2],1,1,1)),0,-2)   
        S = PMaskc*S

        if np.sum(np.where(S<0))>0:
           print("S negativa")
           S[np.where(S<0)]=0


        if SaveSmatrix!='':# Save Smatrix if needed
            SData={}
            SData[0]=S
            _DS.saveTrajtoMat(ScanProject,SData,'SMatrix_' + SaveSmatrix)

        return S,xx,yy,zz

    def get_ProbFromSmatrix(S):
        P   =np.zeros(S.shape)
        Ssum=np.sum(S,axis=3)
        with np.errstate(divide='ignore', invalid='ignore'):
            for i in range(S.shape[3]):  
                P[:,:,:,i]=S[:,:,:,i]/Ssum

        P[np.isnan(P)]=0
        P[np.isinf(P)]=0

        return P

    def get_InvFromSmatrix(S):
        Ssum=np.sum(S,axis=3)
        with np.errstate(divide='ignore', invalid='ignore'):
            C=1/Ssum

        C[np.isnan(C)]=0
        C[np.isinf(C)]=0

        return C


    def get_FullProbFromSmatrix(S):
        SC=S.copy()
        P   =np.zeros(SC.shape)
        SC[np.isnan(SC)]=0
        Ssum=np.sum(SC)
        with np.errstate(divide='ignore', invalid='ignore'):
            for i in range(SC.shape[3]):  
                P[:,:,:,i]=SC[:,:,:,i]/Ssum

        P[np.isnan(P)]=0
        P[np.isinf(P)]=0

        return P

    def get_logProb(P):
        with np.errstate(divide='ignore', invalid='ignore'):
            LogP=np.log(P)
        LogP[np.isnan(LogP)]=-np.inf
        LogP[np.isinf(LogP)]=-1000
        return LogP

    def get_logPrior(P,xx,yy,zz,n,SGxyz,STxyz,xyzCenterGuess,Wprio=[1,1,2]):
            #Gaussian 
        GPrior=-((xx-xyzCenterGuess[0])**2 / (2*SGxyz[0]**2) + (yy-xyzCenterGuess[1])**2 / (2*SGxyz[1]**2) + (zz-xyzCenterGuess[2])**2 / (2*SGxyz[2]**2))
            #Tophat
        TopHatPrior=np.zeros(xx.shape) 
        TopHatPrior[np.where(((xx-xyzCenterGuess[0])**2 / (STxyz[0]**2) + (yy-xyzCenterGuess[1])**2 / (STxyz[1]**2) + (zz-xyzCenterGuess[2])**2 / (STxyz[2]**2))>1.0)]=-1e10
            #Low Photon count 
        LPPrior=np.zeros(xx.shape) 
        if Wprio[2]>0:
            for i in range(P.shape[3]):
                LPPrior=LPPrior + Wprio[2]*PSF_Lib.get_logProb(1-P[:,:,:,i])/(1+n[i])

        logPrior=Wprio[0]*TopHatPrior + Wprio[1]*GPrior + LPPrior 
        return logPrior

    def get_logPrior_V2(P,PP,xx,yy,zz,n,SGxyz,STxyz,xyzCenterGuess,xyzCenterGuessG,Wprio=[1,0,1,1]):
        #Tophat  
        TopHatPrior=np.zeros(xx.shape)
        if Wprio[0]>0:
            TopHatPrior[np.where(((xx-xyzCenterGuess[0])**2 / (STxyz[0]**2) + (yy-xyzCenterGuess[1])**2 / (STxyz[1]**2) + (zz-xyzCenterGuess[2])**2 / (STxyz[2]**2))>1.0)]=-1e10  
            
        #Gaussian 
        GPrior=0
        if Wprio[1]>0:
            GPrior=np.zeros(xx.shape)
            if xyzCenterGuessG[0]!=-100:
                GPrior=-((xx-xyzCenterGuessG[0])**2 / (2*SGxyz[0]**2) + (yy-xyzCenterGuessG[1])**2 / (2*SGxyz[1]**2) + (zz-xyzCenterGuessG[2])**2 / (2*SGxyz[2]**2))

        #Low Photon count  
        LPPrior=0
        if Wprio[2]>0:
            LPPrior=np.zeros(xx.shape)
            for i in range(P.shape[3]):
                LPPrior=LPPrior + PSF_Lib.get_logProb(1-PP[:,:,:,i])/(1+n[i])

        #Distance prior
        DPPrior=0
        if Wprio[3]>0:
            DPPrior=np.zeros(xx.shape)
            for i in range(PP.shape[3]):
                DPPrior=DPPrior + n[i]*PSF_Lib.get_logProb(PP[:,:,:,i])

        logPrior= Wprio[0]*TopHatPrior + Wprio[1]*GPrior + Wprio[2]*LPPrior + Wprio[3]*DPPrior

        return logPrior,TopHatPrior, GPrior, LPPrior, DPPrior

    def get_logPrior_V3(xx,yy,zz,n,SGxyz,STxyz,xyzCenterGuess,xyzCenterGuessG,Wprio=[1,0,1,1]):
        #Tophat  
        TopHatPrior=0
        if Wprio[0]>0:
            TopHatPrior=np.zeros(xx.shape)
            TopHatPrior[np.where(((xx-xyzCenterGuess[0])**2 / (STxyz[0]**2) + (yy-xyzCenterGuess[1])**2 / (STxyz[1]**2) + (zz-xyzCenterGuess[2])**2 / (STxyz[2]**2))>1.0)]=-1e10  

        #Gaussian 
        GPrior=0
        if Wprio[1]>0:
            GPrior=np.zeros(xx.shape)
            if xyzCenterGuessG[0]!=-100:
                #GPrior=-((xx-xyzCenterGuessG[0])**2 / (2*SGxyz[0]**2) + (yy-xyzCenterGuessG[1])**2 / (2*SGxyz[1]**2) + (zz-xyzCenterGuessG[2])**2 / (2*SGxyz[2]**2))
                GPrior=-( xx2/(2*SGxyz[0]**2) + yy2/(2*SGxyz[1]**2) + zz2/(2*SGxyz[2]**2) -
                          xx*(xyzCenterGuessG[0]/(2*SGxyz[0]**2)) - yy*(xyzCenterGuessG[1]/(2*SGxyz[1]**2)) - zz*(xyzCenterGuessG[2]/(2*SGxyz[2]**2)) +
                          xyzCenterGuessG[0]**2 /(2*SGxyz[0]**2)  + xyzCenterGuessG[1]**2 /(2*SGxyz[1]**2)  +   xyzCenterGuessG[2]**2 /(2*SGxyz[2]**2)  )

        logPrior= Wprio[0]*TopHatPrior + Wprio[1]*GPrior

        return logPrior,TopHatPrior, GPrior, LPPrior, DPPrior

    def get_PoissonLoglikelihood(ScanProject,S,xx,yy,zz,ni,xyzCenterGuess):
        print("Running Poisson estimation")
        m = int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use
        MemSize=ScanProject.PosCalibration['TrackingMemsize']   #Amount of points that is a pattern 
        P=PSF_Lib.get_ProbFromSmatrix(S)
        LogP=PSF_Lib.get_logProb(P)

        MeanPhotonfactor=1
        if len(ScanProject.TrajectorySelectionRules['Wprio'])>3:
            if ScanProject.TrajectorySelectionRules['Wprio'][3]>0:
                print("Using mean photon")
                MeanPhotonfactor=ScanProject.TrajectorySelectionRules['Wprio'][3]/(m/MemSize)
           
        nn=np.broadcast_to(MeanPhotonfactor*ni,LogP.shape)

        SGxyz=[0.04, 0.04, 0.04] #[um] Gaussian prior sigma
        STxyz=[0.08, 0.08, 0.08] #[um] Tophat prior sigma
        LogPrior=PSF_Lib.get_logPrior(P,xx,yy,zz,ni,SGxyz,STxyz,xyzCenterGuess,Wprio=ScanProject.TrajectorySelectionRules['Wprio'])

        #Construct the log likelihood 
        LogPTot =np.sum(LogP*nn,axis=3) + LogPrior
        LogPTot=LogPTot-np.max(LogPTot) 
        return LogPTot

    def get_PoissonLoglikelihood_V2(ScanProject,S,xx,yy,zz,ni,xyzCenterGuess,xyzCenterGuessG):
        print("Running Poisson estimation")
        m = int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use
        MemSize=ScanProject.PosCalibration['TrackingMemsize']   #Amount of points that is a pattern 

        ## Slow ~0.07sec (without PP ~0.04sec )
        P=PSF_Lib.get_ProbFromSmatrix(S)
        LogP=PSF_Lib.get_logProb(P)

        if ScanProject.TrajectorySelectionRules['Wprio'][2]>0 or ScanProject.TrajectorySelectionRules['Wprio'][3]>0:
            PP=PSF_Lib.get_FullProbFromSmatrix(S)
        else:
            PP=0
        ####  

        if 0: #Debug
            #Obs the order of the stack might not be that of the FPGA memmory, depends on which is the first index when tracking sarts. 
            DebugPath="C:/Users/elias/Desktop/Shadowfax/Data/Debug/"
            #DebugPath="C:/Users/elias/Desktop/Shadowfax/Data/20201008_(20201007_PSF_zSpacing_135-185)_(noXYMovement_zmovement)/20nm-beads_NO-xy_hand-moved-Z_135nm-staircase/PScan/"
            from PIL import Image
            for num in range(P.shape[3]):
                for kk in range(P.shape[2]):
                    Image.fromarray(P[:,:,kk,num]).save(DebugPath + str(num) + "/P_PSF" + str(kk) + "_" + str(num) + '.tif')
                
        ## Slow ~0.012sec
        nn=np.broadcast_to(ni,LogP.shape)
        LogP=np.sum(LogP*nn,axis=3)
        ###

        SGxyz= ScanProject.TrajectorySelectionRules['SGxyz'] #[um] Gaussian prior sigma
        STxyz= ScanProject.TrajectorySelectionRules['STxyz'] #[um] Gaussian prior sigma
        
        ## Slow ~0.026sec
        LogPrior,TopHatPrior, GPrior, LPPrior, DPPrior=PSF_Lib.get_logPrior_V2(P,PP,xx,yy,zz,ni,SGxyz,STxyz,xyzCenterGuess,xyzCenterGuessG,Wprio=ScanProject.TrajectorySelectionRules['Wprio'])
        ####

        #Construct the log likelihood 
        LogPTot = LogP + LogPrior
        LogPTot=LogPTot-np.max(LogPTot) 
        return LogPTot, LogP, TopHatPrior, GPrior, LPPrior, DPPrior

    def get_PoissonLoglikelihood_V3(ScanProject,P,PP,LogPLin,TopHatPrior,xx,yy,zz,ni,xyzCenterGuessG):
        print("Running Poisson estimation")

        ###
        Pshape=PP.shape
        LogP=np.reshape(np.matmul(LogPLin,ni),Pshape[:-1])

        ## Slow ~0.012sec
        #nn=np.broadcast_to(ni,LogP.shape)
        #LogP=np.sum(LogP*nn,axis=3)
        ###


        SGxyz= ScanProject.TrajectorySelectionRules['SGxyz'] #[um] Gaussian prior sigma
        STxyz= ScanProject.TrajectorySelectionRules['STxyz'] #[um] Tophat prior sigma
        
        ## Do not evaluate tophat since it is alredy evaluated
        Wprio=ScanProject.TrajectorySelectionRules['Wprio']
        Wprio[0]=0; xyzCenterGuess=[0,0,0]
        LogPrior,_, GPrior, LPPrior, DPPrior=PSF_Lib.get_logPrior_V2(P,PP,xx,yy,zz,ni,SGxyz,STxyz,xyzCenterGuess,xyzCenterGuessG,Wprio=Wprio)
        ####

        #Construct the log likelihood 
        LogPTot = LogP + LogPrior + TopHatPrior
        LogPTot=LogPTot-np.max(LogPTot) 
        return LogPTot, LogP, TopHatPrior, GPrior, LPPrior, DPPrior

    def get_BinomialLoglikelihood(ScanProject,S,xx,yy,zz,ni,xyzCenterGuess):
        print("Running Binomial estimation")
        m = int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use
        MemSize=ScanProject.PosCalibration['TrackingMemsize']   #Amount of points that is a pattern 
        FPGA_RecSpeed=5e-9
        N=ScanProject.PosCalibration['MeasurementOnTime']/FPGA_RecSpeed

        P=PSF_Lib.get_ProbFromSmatrix(S)*np.sum(ni)/N

        LogP=PSF_Lib.get_logProb(P)
        Log1MP=PSF_Lib.get_logProb(1-P)
           
        nn=np.broadcast_to(ni,LogP.shape)
        
        SGxyz=[0.04, 0.04, 0.04] #[um] Gaussian prior sigma
        STxyz=[0.08, 0.08, 0.08] #[um] Tophat prior sigma
        LogPrior=PSF_Lib.get_logPrior(P,xx,yy,zz,ni,SGxyz,STxyz,xyzCenterGuess,Wprio=ScanProject.TrajectorySelectionRules['Wprio'])

        #Construct the log likelihood 
        LogPTot =np.sum(LogP*nn,axis=3) + np.sum(Log1MP*(N-nn),axis=3) + LogPrior
        LogPTot=LogPTot-np.max(LogPTot) 
        return LogPTot

    def get_PHretrival(filepath,configFile,im_yxz, Spot_coordinates_xyz=np.zeros((3,3)), iter=5000):
        ##### Definitions #######################################################################################
        #filepath="C:/Users/elias/Desktop/Shadowfax/Data/20210114_PSF_Bead_rotation_with_weights/alignment/Processed/"
        #file=[ "staircase_MagellanStack.tif" ]
        #configFile='PhaseRetrival_AnalysisConfig.ini'
        #iter=5000  #Iterations to run the GS-algorithm
        ########################################################################################################

        # Get configuration settings
        SettingHandler=_CPhR(configfile= filepath+configFile)
        SettingHandler.get_section_Ph_Retrieval()
        parm=SettingHandler.Ph_Retrieval
        parm['IntencityC']=Spot_coordinates_xyz

        # Get Data
        data=im_yxz# PSF_Lib.Get_tif_stack(filepath + file[0])

        #Correct for image offset, and remove small fluctuations
        backgroundfloor=np.min(data)
        data=data-backgroundfloor
        data[data<0]=0
        
        runPh_Retrieval=1
        if runPh_Retrieval:    
            # Run Phase retrieval
            #I_IP_final, _, _=PSF_Lib.Ph_RetrievalStack(data, parm, iter=iter,verbose=1)
            I_IP_final, _, _=PSF_Lib.Ph_RetrievalDWStack(data, parm, iter=iter,verbose=1)

            today=date.today()
            fname= today.strftime("%Y%m%d") +'_Estimated_phase_stack_iter_' + str(iter) + '.mat'
            PSF_Lib.savetoMat(I_IP_final,parm,filepath,fname)
        else: #Continue iteration on an already existing 
            fname='Estimated_phase_stack__100x100x30crop_iter_10000.mat'
            data, parm_L= PSF_Lib.LoadMat(filepath,fname)

             # Run more Phase retrieval
            iter=10000
            I_IP_final, _, _=PSF_Lib.Ph_RetrievalStackContinueOnOld(data, parm_L, iter=iter,verbose=1)

            fname= today.strftime("%Y%m%d") +'_Estimated_phase_stack_iter_' + str(iter) + '.mat'
            PSF_Lib.savetoMat(I_IP_final,parm,filepath,fname)        
    
        #PSF_Lib.zInterpolator(LF=I_IP_final, SWSZ=4, parm_PR=parm,p='Forward')
        print("GS recovery is finished and saved under:" )
        print(str(filepath+fname))

    def get_zProfiles(filepath,file,configFile, time_points=15,crop_margin=3, spot_num = 3,do_show=0):
        # Get configuration settings
        SettingHandler=_CPhR(configfile= filepath+configFile)
        SettingHandler.get_section_Ph_Retrieval()
        parm=SettingHandler.Ph_Retrieval
        delta_z=parm['zssize']*10**6 # um

        # load, reshape, and average (uncomment check_stack to see if loaded properly)
        im = _skim.imread(filepath + file + '.tif' )
        # check_stack(im)
        im_tzyx = im.reshape((time_points,int(im.shape[0]/time_points),im.shape[1],im.shape[2]), order='F')
        # print(im_tzyx.shape)
        im_zyx = np.mean(im_tzyx, axis=0)
        im_yxz=np.moveaxis(im_zyx,0,2) #move axis to get it in the format used in phase retrival
        im_yxz_original=im_yxz.copy()
        # check_stack(im_zyx)
    
        #Get spot positions
        Spot_coordinates_xyz_zsorted_comp, Spot_coordinates_xyz_zsorted, z_profiles, z = _MI.get_Spot_coordinates(im_yxz, BlankingRadius=6,spot_num = spot_num,crop_margin=crop_margin,delta_z=delta_z)

        #Crop z-stack
        im_yxz=im_yxz_original[:,:,Spot_coordinates_xyz_zsorted[0,2].astype('int')-crop_margin:Spot_coordinates_xyz_zsorted[-1,2].astype('int')+1+crop_margin]

        for i in range(spot_num):
            plt.plot(z,z_profiles[:,i],'-*',label=str(Spot_coordinates_xyz_zsorted_comp[i,:].astype('int'))) 
            First_in_crop=Spot_coordinates_xyz_zsorted[0,2].astype('int')-crop_margin
            Last_in_crop= Spot_coordinates_xyz_zsorted[-1,2].astype('int')+crop_margin
            plt.plot(z[First_in_crop],z_profiles[First_in_crop,i],'r*') 
            plt.plot(z[Last_in_crop],z_profiles[Last_in_crop,i],'r*') 

        plt.legend()
        plt.savefig(filepath + file + '_plot.tif')
        if do_show:
            plt.show()
        else:
            plt.close()

        print("z-Stack is croped and spots are identified")
        return im_yxz, Spot_coordinates_xyz_zsorted_comp

    def get_zProfiles2(filepath,file,configFile, time_points=15,crop_margin=3, spot_num = 3,do_show=0):
        # Get configuration settings
        SettingHandler=_CPhR(configfile= filepath+configFile)
        SettingHandler.get_section_Ph_Retrieval()
        parm=SettingHandler.Ph_Retrieval
        delta_z=parm['zssize']*10**6 # um

        # load, reshape, and average (uncomment check_stack to see if loaded properly)
        im = _skim.imread(filepath + file + '.tif' )
        # check_stack(im)
        im_tzyx = im.reshape((time_points,int(im.shape[0]/time_points),im.shape[1],im.shape[2]), order='F')
        # print(im_tzyx.shape)
        im_zyx = np.mean(im_tzyx, axis=0)
        im_yxz=np.moveaxis(im_zyx,0,2) #move axis to get it in the format used in phase retrival
        im_yxz_original=im_yxz.copy()
        # check_stack(im_zyx)
    
        #Get spot positions
        Spot_coordinates_xyz_zsorted_comp, Spot_coordinates_xyz_zsorted, z_profiles, z = _MI.get_Spot_coordinates2(im_yxz, BlankingRadius=12,spot_num = spot_num,crop_margin=crop_margin,delta_z=delta_z)

        #Crop z-stack
        im_yxz=im_yxz_original[:,:,Spot_coordinates_xyz_zsorted[0,2].astype('int')-crop_margin:Spot_coordinates_xyz_zsorted[-1,2].astype('int')+1+crop_margin]

        for i in range(spot_num):
            plt.plot(z,z_profiles[:,i],label=str(Spot_coordinates_xyz_zsorted_comp[i,:].astype('int'))) 
        plt.legend()
        plt.savefig(filepath + file + '_plot.tif')
        if do_show:
            plt.show()
        else:
            plt.close()

        print("z-Stack is croped and spots are identified")
        return im_yxz, Spot_coordinates_xyz_zsorted_comp

class optimizers():
    def min_target_Poisson(rp, rL, nL, s, NL):
        """
        rp = is the fluorophor position xp,yp,zp and is the unknown  
        rL = args[0] is the list of laser positions x,y,z each one column 
        nL = args[1] photon counts for each row in rL
        """
        #rL=args[0]
        #nL=args[1]
        #s=args[2]
        #NL=args[3]

        m=rL.shape[1]
        
        #Reshape to fitt the dimenstions of rL
        rp=np.outer(rp,np.ones(m))
        s=np.outer(s,np.ones(m))

        #calculate the brightness 
        NS_i=PSF_Lib.PSF_Gaussian(rL,rp,s,NL,1,Norm=0)
        NS=np.sum(nL)/np.sum(NS_i)

        N=PSF_Lib.PSF_Gaussian(rL,rp,s,NL,NS,Norm=1)
        
        Dr = (rL-rp) 
        DN = (nL-N)
        res=np.sum(Dr[0,:]*DN)**2 + np.sum(Dr[1,:]*DN)**2 + np.sum(Dr[2,:]*DN)**2
        return res

def main_PHretrival():
    
    path="C:/Users/elias/Desktop/Shadowfax/Data/20210114_PSF_Bead_rotation_with_weights/alignment/Processed/"
    file=[ "AVG_staircase_100x100x24.tif",  #0
         ]
    configFile='AnalysisConfig_100x100x24.ini'

    # Get configuration settings
    SettingHandler=_CPhR(configfile= path+configFile)
    SettingHandler.get_section_Ph_Retrieval()
    parm=SettingHandler.Ph_Retrieval
    # Get Data
    data=PSF_Lib.Get_tif_stack(path + file[0])

    #Correct for image offset, and remove small fluctuations
    backgroundfloor=np.min(data)
    data=data-backgroundfloor
    data[data<0]=0

    runPh_Retrieval=1
    if runPh_Retrieval:    
        # Run Phase retrieval
        iter=5000
        #I_IP_final, _, _=PSF_Lib.Ph_RetrievalStack(data, parm, iter=iter,verbose=1)
        I_IP_final, _, _=PSF_Lib.Ph_RetrievalDWStack(data, parm, iter=iter,verbose=1)

        fname='Estimated_phase_stack_iter_5000.mat'
        PSF_Lib.savetoMat(I_IP_final,parm,path,fname)
    else:
        fname='Estimated_phase_stack__100x100x30crop_iter_10000.mat'
        data, parm_L= PSF_Lib.LoadMat(path,fname)

         # Run more Phase retrieval
        iter=10000
        I_IP_final, _, _=PSF_Lib.Ph_RetrievalStackContinueOnOld(data, parm_L, iter=iter,verbose=1)

        fname='Estimated_phase_stack__100x100x15crop_iter_20000.mat'
        PSF_Lib.savetoMat(I_IP_final,parm,path,fname)        
    
    
    PSF_Lib.zInterpolator(LF=I_IP_final, SWSZ=4, parm_PR=parm,p='Forward')

if __name__ == '__main__':
    main_PHretrival()