from pathlib import Path
import numpy as np
import sys
from scipy.io import savemat
import matplotlib.pyplot as plt

class physics_Lib():
    """
    This library is for physics quantity calculations like
    Diffusion constants
    """
    @staticmethod
    def get_diffusion(traj):
        """
        Calculates diff coeff from trajectories according to Vestegaard et al (CVE estimator)
        Input:
            trajs: numpy array with position t,x,y,z,... data, the size of the array will be the dimension of the data
            the first column must be time  in [s]
            dimension of the problem will be defined by column 2-end (for a matrix as txy dimension is 2 for x and y data)

        output:
            coeffDiffs in [um]**2 [s]**-1
        """

        time=traj[0,:]
        traj=traj[1:,:]

        timeStep=np.diff(time)
        thediffs=np.diff(traj, axis=1) #[um]
        coeffDiffComponents=np.mean(thediffs**2 /(2*timeStep), axis=1) + np.mean(thediffs[:,0:-1]*thediffs[:,1:]/timeStep[0:-1], axis=1)
        coeffDiff=np.mean(coeffDiffComponents)

        return coeffDiff, coeffDiffComponents

    @staticmethod
    def get_MSD(traj):
        """
        Calculates mean square displacement defined as M(dt)=<|x(t+dt)-x(t)|**2>
        Input:
            trajs: numpy array with position t,x,y,z,... data
            the first column must be time in [s]

        output:
            M vector with each column for each dimension
            dt vector with time lags
        """
        MSD=np.zeros((traj.shape))
        for i in range(1,len(traj[0,:])):
            MSD[0,i]  = np.mean((traj[0,i:] - traj[0,:-i]))
            MSD[1:,i] = np.mean((traj[1:,i:] - traj[1:,:-i])**2, axis=1)

        return MSD

    @staticmethod
    def get_MSD_traj(TrajDictionary):
        """
        Go through all poisson trajectory estimations and calculates the MSD
        """
        result_txyz ={}
        for TrajToGet in range(len(TrajDictionary)):        
            txyzn_Map  = TrajDictionary[TrajToGet]
            traj=txyzn_Map[:4,:]
            
            R=1/6.2 #blur factor
            for i in [2,3,4]: 
                j   =5*i  #[10,15,20] points 
                dtj  = traj[0,j:] - traj[0,:-j]
                t   = traj[0,:-j]+ dtj/2 #[s]
                MSD = ((traj[1:,j:] - traj[1:,:-j])**2)/(2*dtj*(1-2*(R/j)))

                wndw=20
                wndwS=1 #if 1 it is a moving average, if equal to wndw it is a fix window average
                T= np.convolve(np.ones(wndw)/wndw,t)
                T=T[wndw-1:-wndw+1:wndwS]
                D=np.ones((len(T),4))
                D[:,0]=T
                for xyz in range(1,3+1):
                    Dtemp = np.convolve(np.ones(wndw)/wndw,MSD[xyz-1,:])
                    D[:,xyz]=Dtemp[wndw-1:-wndw+1:wndwS]
                result_txyz[str(TrajToGet),str(j)]=D

        return result_txyz

    @staticmethod
    def get_R(ScanProject=None):
        """
        R=(1/Dt)*integral_0^Dt (S(t')*(1-S(t')))dt'
        Where Dt is the frame time and  S(t)=integral_0^t(s(t'))dt'
        and s(t') is the frame function; a non negative fundtion with S(Dt) = 1 (equals to 1 when integrated over the full frame)
        Most of the time we will model s(t') as a squear wave with high value when laser is on and 0 when it is off, 
        the high value is then normolized with the requierment S(Dt) = 1
        """
        LaserOnTime    =ScanProject.PosCalibration['MeasurementOnTime'] #[s]
        LaserPeriodTime=1/ScanProject.PosCalibration['ScanSpeedTracking'] #[s]
        LaserOffTime=LaserPeriodTime-LaserOnTime

        LaserPattern=ScanProject.PosCalibration['LaserShotPattern']
        ShotsPerFrame= ScanProject.TrajectorySelectionRules['wndw']
        FrameTime=ShotsPerFrame*LaserPeriodTime

        LaserOnOffPattern=np.vstack((LaserPattern,LaserPattern)).T
        LaserOnOffPattern[:,0]=LaserOnOffPattern[:,0]*0

        t=np.ones(LaserOnOffPattern.shape)
        t[:,0]=t[:,0]*LaserOffTime
        t[:,1]=t[:,1]*LaserOnTime
        t=t.flatten()
        t=np.append([0],t)
        t=np.cumsum(t)
        dt=np.diff(t)
        s=LaserOnOffPattern.flatten()
        s=s/np.sum(s*dt) 
        S=np.cumsum(s*dt)
        R=np.sum(S*(1-S)*dt)/FrameTime
                    
        if 0:# Look at the 
            plt.step(t[1:],s);plt.show()        

        print('R found is:' + str(R))
        return R

    @staticmethod
    def get_MSD_traj2(ScanProject=None,TrajDictionary=None,window=10,MSDStart=2,MSDSteps=8,PointsToUse=7,R=1/6.2,Flag=0):
        """
        Go through all poisson trajectory estimations and calculates the diffution constant and localization error
        Here we use the relation of the slope and intersect of the MSD plots
        """
        #window=10       #segiment size of the trajectory to be used for MSD calcultaions 
        #MSDSteps=8      #Amount of MSD steps to use over the window defined above, OBS! we skip the first point thus we will always start from step 2 and move MSDSteps from that point 
        #MSDStart=2      #Start the MDS from this step and move MSDSteps-1 forward 
        #PointsToUse=7   #Amount of points to use in the linear regration, use only the begining of the window 
        #R=1/6.2         #blur factor
        #Flag            #0: estiamte slope and intersept, 1: force estimation through origo 

        if 0: #test trajectory 
            do_save_plot=0
            for TrajToGet in range(len(TrajDictionary)):        
                MSD=physics_Lib.get_MSD(TrajDictionary[TrajToGet][0:4,:])

                fig = plt.figure()
                ax= fig.add_subplot()
                for i in range(10):
                    ax.plot(MSD[0,:],MSD[0,:]*2*0.2*(i+1),'gray',alpha=0.3)
                ax.plot(MSD[0,:],MSD[1,:],'r',label='x');ax.plot(MSD[0,:],MSD[2,:],'g',label='y');ax.plot(MSD[0,:],MSD[3,:],'b',label='z');
                ax.plot(MSD[0,:],np.sum(MSD[1:4,:],axis=0)/3,'m',label='xyz');
                ax.grid();ax.legend()
                #ax.set_xlim(0,0.4);ax.set_ylim(0,1)

                ax.set_title('Traj:' + str(TrajToGet))
                ax.set_xlabel('time offset [s]')

                if do_save_plot:
                    file="T_" + str(TrajToGet) + '_D_MDSCurve' + '.png'
                    path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_MSDCurve' + '/'
                    Path(path).mkdir(parents=True, exist_ok=True)
                    fig.savefig(path + file)
                    plt.close()
                    print('Done!')
                else:
                    plt.show()
            print('Plot Done')

        if 0: #histogram test trajectory 
            do_save_plot=0
            for TrajToGet in range(len(TrajDictionary)):        
                MSD=physics_Lib.get_MSD(TrajDictionary[TrajToGet][0:4,:])

                fig = plt.figure()
                ax= fig.add_subplot()
                ax.plot(MSD[0,:-1],np.diff(MSD[1,:])/np.diff(MSD[0,:]),'r',label='x');
                ax.plot(MSD[0,:-1],np.diff(MSD[2,:])/np.diff(MSD[0,:]),'g',label='y');
                ax.plot(MSD[0,:-1],np.diff(MSD[3,:])/np.diff(MSD[0,:]),'b',label='z');
                ax.plot(MSD[0,:-1],np.diff(np.sum(MSD[1:4,:],axis=0)/3)/np.diff(MSD[0,:]),'m',label='xyz');
                ax.grid();ax.legend()
                #ax.set_xlim(0,0.4);ax.set_ylim(0,1)

                ax.set_title('Traj:' + str(TrajToGet))
                ax.set_xlabel('time offset [s]')

                if do_save_plot:
                    file="T_" + str(TrajToGet) + '_D_MDSCurve' + '.png'
                    path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_MSDCurve' + '/'
                    Path(path).mkdir(parents=True, exist_ok=True)
                    fig.savefig(path + file)
                    plt.close()
                    print('Done!')
                else:
                    plt.show()
            print('Plot Done')

        result_txyz ={}
        for TrajToGet in range(len(TrajDictionary)):        
            txyzn_Map  = TrajDictionary[TrajToGet]
            traj=txyzn_Map[:4,:]                          #Remove the n term left are t,x,y,z
            #print(str(int(np.floor(traj.shape[1]/window))-1))
            if (int(np.floor(traj.shape[1]/window))-1)>1: #Only run if there is more then 2 window intervals 
                Time=traj[0,:][0:(int(np.floor(traj.shape[1]/window))-1)*window:window]
                dt=(traj[0,1] - traj[0,0])
                MSD=np.zeros((3,int(np.floor(traj.shape[1]/window))-1,MSDSteps))
                T=np.zeros(MSDSteps)
                for j in range(MSDStart,MSDSteps+MSDStart): 
                    dtj= (traj[0,j:] - traj[0,:-j])
                    T[j-MSDStart]  = dtj[0] 
                    Dj = (traj[1:,j:] - traj[1:,:-j])**2

                    wndw=window-j
                    wndwS=wndw+j #if 1 it is a moving average, if equal to wndw it is a fix window average
                    for xyz in range(3):
                        MSDTemp = np.convolve(np.ones(wndw)/wndw,Dj[xyz,:])
                        pos=(np.arange(MSD.shape[1])+1)*window-j
                        MSD[xyz,:,j-MSDStart]=MSDTemp[pos]

                #linera regreation
                t=T[:PointsToUse]/dt
                Sy=np.sum(MSD[:,:,:PointsToUse],axis=2)
                St2=np.sum(t**2)
                St=np.sum(t)

                MSDt=np.zeros(MSD[:,:,:PointsToUse].shape)
                for i in range(len(t)):
                    MSDt[:,:,i]=MSD[:,:,i]*t[i]
                Sty=np.sum(MSDt,axis=2)
                
                if Flag==0: #Get both the slope and the intersept 
                    nominator_m=St2*Sy-St*Sty
                    nominator_g=Sty-St*Sy
                    denominator=St2- St**2

                    D=(nominator_g/denominator)/(2*dt)
                    sigma2=(nominator_m/denominator)/2 + 2*R*D*dt

                elif Flag==1: #Evaluate the slope assuming that we are going throught origo
                    #this mean that 0=St2*Sy-St*Sty => St2*Sy = St*Sty
                    #and g = (Sty-St*Sy)/(St2- St**2) => with Sty= (St2*Sy)/ St
                    ##   g = ((St2*Sy/St)-(St**2)*Sy/St)/(St2- St**2)=(Sy/St)(St2-St**2)/(St2- St**2)= Sy/St
                    D=(Sy/St)/(2*dt)
                    sigma2=np.zeros(D.shape)

                #Results
                #result_txyz[str(TrajToGet),'TDS2']=(Time,D,sigma2)
                result_txyz[str(TrajToGet),'Sigma2']=np.vstack((Time,sigma2)).T #[mu^2]   vector with t,x,y,z
                result_txyz[str(TrajToGet),'D']=np.vstack((Time,D)).T           #[mu^2/s] vector with t,x,y,z
                result_txyz[str(TrajToGet),'VarD']=np.vstack((Time,sigma2)).T     #[mu^2/s] vector with t,x,y,z
            else:
                print('MDS missed')


        return result_txyz

    @staticmethod
    def get_CVE_MSD_traj(TrajDictionary,wndw=20,R=1/6.2,j=10,n=3,Flag=0):
        """
        Go through all poisson trajectory estimations and calculates the MSD
        """
        #wndw=20     #Window size for each estimation  
        #R=1/6.2     #Blur factor
        #j=10        #Offset between elements to use, can be >0 
        #n=3         #Offset diagonal to use in the covariant matrix, can be 1 to (j-1), only used for method Flag>0
        #Flag=0      #Method to use, 0: using main diagonla and the last, j, diagonal 
                     #               1: using offset diagonal, n, and last, j, diagonal 
                     #               2: using offset diagonal, n, and main diagonal 

        result_txyz ={}
        for TrajToGet in range(len(TrajDictionary)):        
            txyzn_Map  = TrajDictionary[TrajToGet]
            traj=txyzn_Map[:4,:]
            
            dt  = traj[0,1] - traj[0,0]
            t   = traj[0,:]+ dt*j/2                 #[s]                                                 Full length N 
            DTraj_xyz=traj[1:,j:] - traj[1:,:-j]   #                                                    N-j elements
            DTraj_xyz2   = (DTraj_xyz[:,:])**2     #Diagonal terms in cov-matrix                        N-j elements
            CrossTraj_xyz_n= DTraj_xyz[:,:-n]*DTraj_xyz[:,n:] #Off diagonal with n steps where 0<n<j    N-j-n elements
            CrossTraj_xyz_j= DTraj_xyz[:,:-j]*DTraj_xyz[:,j:] #Last diagonal with j step                N-j-j elements
            
            if Flag==0:
                #Diagonal and last diagonal version 
                D_all = DTraj_xyz2[:,:-j]/(2*dt*j) + CrossTraj_xyz_j/(dt*j)      #[mu^2/s]
                Sigma2_all=(R/j)*DTraj_xyz2[:,:-j]+(2*(R/j)-1)*CrossTraj_xyz_j       #[mu^2]
                t=t[:-2*j]
            elif Flag==1:
                #Off Diagonal version for D and last diagonal for sigma
                D_all = (CrossTraj_xyz_n/(2*dt*(j-n)))[:,:-(j-n)]      #[mu^2/s]
                Sigma2_all=(R/(j-n))*CrossTraj_xyz_n[:,:-(j-n)] - CrossTraj_xyz_j       #[mu^2]
                t=t[:-2*j]
            elif Flag==2:
                #Off Diagonal version for D and main diagonal for sigma
                D_all = ((CrossTraj_xyz_n)/(2*dt*(j-n)))      #[mu^2/s]
                Sigma2_all=(DTraj_xyz2[:,:-n]/2) - ((j-2*R)/(2*(j-n)))*CrossTraj_xyz_n       #[mu^2]
                t=t[:-(j+n)]

                #D_all = (np.abs(CrossTraj_xyz_n)/(2*dt*(j-n)))[:,:-(j-n)]      #[mu^2/s]
                #D_all = ((CrossTraj_xyz_n)/(2*dt*(j-n)))[:,:-(j-n)]      #[mu^2/s]
                #Sigma2_all=(DTraj_xyz2[:,:-j]/2) - ((j-2*R)/(2*(j-n)))*CrossTraj_xyz_n[:,:-(j-n)]       #[mu^2]
                #t=t[:-2*j]


            wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            if wndw>1:
                T= np.convolve(np.ones(wndw)/wndw,t)
                T=T[wndw-1:-wndw+1:wndwS]
            else:
                T=t
            
            
            D=np.ones((len(T),4));      D[:,0]=T
            Sigma2=np.ones((len(T),4)); Sigma2[:,0]=T
            VarD=np.ones((len(T),4));   VarD[:,0]=T
            
            for xyz in range(1,3+1):
                if wndw>1:
                    Dtemp = np.convolve(np.ones(wndw)/wndw,D_all[xyz-1,:])
                    D[:,xyz]=Dtemp[wndw-1:-wndw+1:wndwS]
                
                    D2temp = np.convolve(np.ones(wndw)/wndw,(D_all[xyz-1,:])**2)
                    D2temp=D2temp[wndw-1:-wndw+1:wndwS]
                    VarD[:,xyz]=D2temp-(D[:,xyz])**2

                    Sigma2temp = np.convolve(np.ones(wndw)/wndw,Sigma2_all[xyz-1,:])
                    Sigma2[:,xyz]=Sigma2temp[wndw-1:-wndw+1:wndwS]
                else:
                    D[:,xyz] =D_all[xyz-1,:]
                    VarD[:,xyz]=(D_all[xyz-1,:])**2 - (D[:,xyz])**2
                    Sigma2[:,xyz] = Sigma2_all[xyz-1,:]
            
            
            #print(str([traj.shape[1], D_all.shape[1]+j+n, D_all.shape[1]/wndw, D.shape[0]]))
            result_txyz[str(TrajToGet),'Sigma2']=Sigma2 #[mu^2]    vector with t,x,y,z
            result_txyz[str(TrajToGet),'D']=D           #[mu^2/s]  vector with t,x,y,z
            result_txyz[str(TrajToGet),'VarD']=VarD     #[mu^2/s]  vector with t,x,y,z

        return result_txyz


    @staticmethod
    def get_CVE_traj_Special(TrajDictionary,wndw=1,R=1/6.2,j=10,n=1,jjStart=2,Flag=2):
                        
            if n==1 and Flag==2 and wndw==1 and j>2:
                N=j+n
                y={}; x=0
                yy={}
                for jj in range(jjStart,j+1):
                    #Maximumlikelihood linear regrestion for the slope k for a line passing origo is: sum(yi)/sum(xi)=sum(k*xi)/sum(xi)=k*sum(xi)/sum(xi)=k
                    resCVE_t=physics_Lib.get_CVE_MSD_traj(TrajDictionary,wndw=wndw,R=R,j=jj,n=n,Flag=Flag)#wndw=5,R=1/6.2,j=4,n=1,Flag=2; wndw=wndw,R=R,j=4,n=1,Flag=2
                    x+= jj-n  #sum(xi)
                    
                    for TrajToGet in range(len(TrajDictionary)):
                        if TrajDictionary[TrajToGet].shape[1]!= resCVE_t[str(TrajToGet),'D'].shape[0]+jj+n: 
                            print(str([TrajDictionary[TrajToGet].shape[1], resCVE_t[str(TrajToGet),'D'].shape[0]+jj+n]))

                        PointsToGet=np.floor(TrajDictionary[TrajToGet].shape[1]/N).astype('int')-1
                        if jj==jjStart:
                            for i in range(3):
                                if i==0:
                                    y[TrajToGet]=np.convolve(np.ones(N-jj)/(N-jj),resCVE_t[str(TrajToGet),'D'][:,i+1]*(jj-n))[N-jj-1::N][:PointsToGet] #y0
                                else:
                                    y[TrajToGet]=np.vstack((y[TrajToGet],np.convolve(np.ones(N-jj)/(N-jj),resCVE_t[str(TrajToGet),'D'][:,i+1]*(jj-n))[N-jj-1::N][:PointsToGet]))
                            y[TrajToGet]=y[TrajToGet].T
                        else:
                            for i in range(3):
                                y[TrajToGet][:,i]=y[TrajToGet][:,i] + np.convolve(np.ones(N-jj)/(N-jj),resCVE_t[str(TrajToGet),'D'][:,i+1]*(jj-n))[N-jj-1::N][:PointsToGet] #sum(yi)

                        #print(str([TrajDictionary[TrajToGet].shape[1]/N, y[TrajToGet].shape[0], resCVE_t[str(TrajToGet),'D'].shape[0]/N]))
                                

                for TrajToGet in range(len(TrajDictionary)):
                    T=resCVE_t[str(TrajToGet),'D'][::N,0][:y[TrajToGet].shape[0]]
                    resCVE_t[str(TrajToGet),'D']=np.vstack((T, (y[TrajToGet]/x).T )).T
      
                return resCVE_t
            else: 
                print('Can not run get_CVE_traj_Special with these parameters')

    @staticmethod
    def get_CVE_traj_Specialv0(TrajDictionary,wndw=1,R=1/6.2,j=10,n=1,Flag=2):
        
            if n==1 and Flag==2 and wndw==1 and j>2:
                y={}; x=0
                yy={}
                for jj in range(2,j+1):
                    #Maximumlikelihood linear regrestion for the slope k for a line passing origo is: sum(yi)/sum(xi)=sum(k*xi)/sum(xi)=k*sum(xi)/sum(xi)=k
                    wndw=j
                    resCVE_t=physics_Lib.get_CVE_MSD_traj(TrajDictionary,wndw=wndw,R=R,j=jj,n=n,Flag=Flag)#wndw=5,R=1/6.2,j=4,n=1,Flag=2; wndw=wndw,R=R,j=4,n=1,Flag=2
                    x+= jj-n  #sum(xi)
                    for TrajToGet in range(len(TrajDictionary)): 
                        if jj==2:
                            y[TrajToGet]=resCVE_t[str(TrajToGet),'D'][:,1:]*(jj-n) # y0
                            #yy[TrajToGet,jj]=resCVE_t[str(TrajToGet),'D'][:,1:]
                        else:
                            y[TrajToGet]=y[TrajToGet][:resCVE_t[str(TrajToGet),'D'].shape[0],:] + resCVE_t[str(TrajToGet),'D'][:,1:]*(jj-n)   #sum(yi)
                            #yy[TrajToGet,jj]=resCVE_t[str(TrajToGet),'D'][:,1:]

                for TrajToGet in range(len(TrajDictionary)): 
                    #print(str([TrajDictionary[TrajToGet].shape[1], y[TrajToGet].shape[0], np.floor(TrajDictionary[TrajToGet].shape[1]/(j+n))]))
                    resCVE_t[str(TrajToGet),'D'][:,1:]=(y[TrajToGet]/x)

                return resCVE_t
            else: 
                print('Can not run get_CVE_traj_Special with these parameters')

    @staticmethod
    def get_single_CVE_traj(TrajDictionary,TrajToGet,wndw=20,R=1/6.2,j=10,n=3,Flag=0):
        """
        Get from one poisson trajectory estimations and calculates the CVE D sigma parameters 
        """
        #wndw=20     #Window size for each estimation  
        #R=1/6.2     #Blur factor
        #j=10        #Offset between elements to use, can be >0 
        #n=3         #Offset diagonal to use in the covariant matrix, can be 1 to (j-1), only used for method Flag=1
        #Flag=0      #Method to use, 0: using main diagonla and the last, j, diagonal 
                    #                1: using offset diagonal, n, and last, j, diagonal 

        txyzn_Map  = TrajDictionary[TrajToGet]
        traj=txyzn_Map[:4,:]
            
        dt  = traj[0,1] - traj[0,0]
        t   = traj[0,:]+ dt*j/2                #[s]                                                 Full length N 
        DTraj_xyz=traj[1:,j:] - traj[1:,:-j]   #                                                    N-j elements
        DTraj_xyz2   = (DTraj_xyz[:,:])**2     #Diagonal terms in cov-matrix                        N-j elements
        CrossTraj_xyz_n= DTraj_xyz[:,:-n]*DTraj_xyz[:,n:] #Off diagonal with n steps where 0<n<j    N-j-n elements
        CrossTraj_xyz_j= DTraj_xyz[:,:-j]*DTraj_xyz[:,j:] #Last diagonal with j step                N-j-j elements
            
        if Flag==0:
            #Diagonal and last diagonal version 
            D_all = DTraj_xyz2[:,:-j]/(2*dt*j) + CrossTraj_xyz_j/(dt*j)      #[mu^2/s]
            Sigma2_all=(R/j)*DTraj_xyz2[:,:-j]+(2*(R/j)-1)*CrossTraj_xyz_j       #[mu^2]
            t=t[:-2*j]
        elif Flag==1:
            #Off Diagonal version for D and last diagonal for sigma
            D_all = (CrossTraj_xyz_n/(2*dt*(j-n)))[:,:-(j-n)]      #[mu^2/s]
            Sigma2_all=(R/(j-n))*CrossTraj_xyz_n[:,:-(j-n)] - CrossTraj_xyz_j       #[mu^2]
            t=t[:-2*j]
        elif Flag==2:
            #Off Diagonal version for D and main diagonal for sigma
            D_all = (CrossTraj_xyz_n/(2*dt*(j-n)))[:,:-(j-n)]      #[mu^2/s]
            Sigma2_all=(DTraj_xyz2[:,:-j]/2) - ((j-2*R)/(2*(j-n)))*CrossTraj_xyz_n[:,:-(j-n)]       #[mu^2]
            t=t[:-2*j]

        wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
        T= np.convolve(np.ones(wndw)/wndw,t)
        T=T[wndw-1:-wndw+1:wndwS]
        D=np.ones((len(T),4));      D[:,0]=T
        Sigma2=np.ones((len(T),4)); Sigma2[:,0]=T
        VarD=np.ones((len(T),4));   VarD[:,0]=T
        for xyz in range(1,3+1):
            Dtemp = np.convolve(np.ones(wndw)/wndw,D_all[xyz-1,:])
            D[:,xyz]=Dtemp[wndw-1:-wndw+1:wndwS]
                
            D2temp = np.convolve(np.ones(wndw)/wndw,(D_all[xyz-1,:])**2)
            D2temp=D2temp[wndw-1:-wndw+1:wndwS]
            VarD[:,xyz]=D2temp-(D[:,xyz])**2

            Sigma2temp = np.convolve(np.ones(wndw)/wndw,Sigma2_all[xyz-1,:])
            Sigma2[:,xyz]=Sigma2temp[wndw-1:-wndw+1:wndwS]
            
            
        return D, Sigma2,VarD

    @staticmethod
    def get_All_traj_CVE(TrajDictionary):
        """
        Go through all poisson trajectory estimations and calculates the MSD
        Note: This might be good to make generic, such that on can input what to estimate. 
        """
        nmax=10
        txyzCVE_Poisson_ProbMap={}
        for TrajToGet in range(len(TrajDictionary)):
            D_final=np.transpose(np.array([[],[],[],[]]))
            for j in range(2,np.min((np.floor(TrajDictionary[TrajToGet].shape[1]/2).astype('int'),200))):
                for n in range(1,np.min((j-1,nmax))):                
                    wndw=TrajDictionary[TrajToGet].shape[1]-2*j
                    #D, Sigma2,VarD=self._Ph.get_single_CVE_traj(TrajDictionary,TrajToGet,wndw=wndw,R=1/6.2,j=j,n=n,Flag=2)
                    txyzn_Map  = TrajDictionary[TrajToGet]
                    traj=txyzn_Map[:4,:]
            
                    dt  = traj[0,1] - traj[0,0]
                    DTraj_xyz=traj[1:,j:] - traj[1:,:-j]                              #                                                    N-j elements
                    DTraj_xyz2   = np.mean((DTraj_xyz[:,:])**2,axis=0)[:-j]           #Diagonal terms in cov-matrix                        N-j elements   => N-2j
                    CrossTraj_xyz_n= np.mean(DTraj_xyz[:,:-n]*DTraj_xyz[:,n:],axis=0)[:-(j-n)]  #Off diagonal with n steps where 0<n<j    N-j-n elements => N-2j
                    CrossTraj_xyz_j= np.mean(DTraj_xyz[:,:-j]*DTraj_xyz[:,j:],axis=0)            #Last diagonal with j step                N-j-j elements

                    MDTraj_xyz2=np.mean(DTraj_xyz2)
                    MCrossTraj_xyz_n=np.mean(CrossTraj_xyz_n)

                    D_final=np.concatenate((D_final,np.array([[j,n,MDTraj_xyz2,MCrossTraj_xyz_n]])))
            
            for n in range(1,nmax):
                getPoints=np.where(D_final[:,1]==n)
                j=D_final[getPoints][:,0]
                D_CVE=D_final[getPoints][:,3]
                D=D_final[getPoints][:,3]/(2*dt*(j-n))
                D_MSD=D_final[getPoints][:,2]
                if n==1: MD=D[nmax:]
                else:MD=MD+D[nmax-(n-1):]
                
                if n==1:
                    plt.plot(j,D_MSD,'-',label='MSDn: ' + str(n))
                plt.plot(j,D_CVE,'-',label='CVEn: ' + str(n))
            MD=MD/(nmax-1)
            txyzCVE_Poisson_ProbMap[TrajToGet]=(MD,j[2:])
            print(str(TrajToGet) + '/' + str(len(TrajDictionary)))
            #plt.plot(j[2:],MD,'--',label='Mean')
            plt.legend()
            plt.grid()
            plt.show()

        #for TrajToGet in range(len(txyzCVE_Poisson_ProbMap)): 
        #    plt.plot(txyzCVE_Poisson_ProbMap[TrajToGet][1],txyzCVE_Poisson_ProbMap[TrajToGet][0],label=str(TrajToGet))
        #plt.legend();plt.grid();plt.show()

        return txyzCVE_Poisson_ProbMap
    
    def get_MSD_Distribution(self,TrajDictionary):
        """
        Go through all poisson trajectory estimations and calculates the MSD
        """
        result_txyz ={}
        for TrajToGet in range(len(TrajDictionary)):        
            txyzn_Map  = TrajDictionary[TrajToGet]
            traj_xyz=txyzn_Map[1:4,:]
            t=txyzn_Map[0,:]

            fig, ax= plt.subplots(4,2)
            for step in [1,4,10]: #range(1,6):
                R=np.sqrt(np.sum(traj_xyz**2,axis=0))
                DR= R[step:] -R[:-step]
                Dx= traj_xyz[0,step:] - traj_xyz[0,:-step] 
                Dy= traj_xyz[1,step:] - traj_xyz[1,:-step] 
                Dz= traj_xyz[2,step:] - traj_xyz[2,:-step] 
                tt=t[step:]
                #### Plot 
                ax[0,0].plot(tt,Dx,alpha=0.3,label=str(step))
                ax[1,0].plot(tt,Dy,alpha=0.3,label=str(step))
                ax[2,0].plot(tt,Dz,alpha=0.3,label=str(step))
                ax[3,0].plot(tt,DR,alpha=0.3,label=str(step))
                ax[0,1].hist(Dx,bins=50,alpha=0.3,label=str(step))
                ax[1,1].hist(Dy,bins=50,alpha=0.3,label=str(step))
                ax[2,1].hist(Dz,bins=50,alpha=0.3,label=str(step))                
                ax[3,1].hist(DR,bins=50,alpha=0.3,label=str(step))

            ax[0,0].set_ylabel('Dx')
            ax[1,0].set_ylabel('Dy')
            ax[2,0].set_ylabel('Dz')
            ax[3,0].set_ylabel('DR')
            ax[3,0].set_xlabel('Time [s]')
            ax[3,1].set_xlabel('step size [um]')
            ax[0,0].set_title('Traj:' + str(TrajToGet))
            plt.legend()
            plt.show()

            """
            R=0.1 #blur factor R= LaserOnTime/ScanPeriodTracking/6
            for i in [2,3,4]: 
                j   =5*i  #[10,15,20] points 
                dtj  = traj[0,j:] - traj[0,:-j]
                t   = traj[0,:-j]+ dtj/2 #[s]
                MSD = ((traj[1:,j:] - traj[1:,:-j])**2)/(2*dtj*(1-2*(R/j)))

                wndw=20
                wndwS=1 #if 1 it is a moving average, if equal to wndw it is a fix window average
                T= np.convolve(np.ones(wndw)/wndw,t)
                T=T[wndw-1:-wndw+1:wndwS]
                D=np.ones((len(T),4))
                D[:,0]=T
                for xyz in range(1,3+1):
                    Dtemp = np.convolve(np.ones(wndw)/wndw,MSD[xyz-1,:])
                    D[:,xyz]=Dtemp[wndw-1:-wndw+1:wndwS]
                result_txyz[str(TrajToGet),str(j)]=D
               """
        return result_txyz

class Particle:
    """
    This class creates a particle with a specific trajectory. 5 trajectory
    methods are available:
    1) Ballistic
    2) Helical
    3) Circular
    4) Diffusive
    5) Multi-state diffusion (HMM)

    Units:
    space - [um]
    time  - [s]
    """
    def __init__(self, initial_pos, imagesz_xyz=(200, 200, 150),
                 time_steps=200*6*4, delta_t=25e-6, mesh_size=8e-3):
        """
        Input:
        initial_pos - array-like variable for initial position in pixels
        imagesz_xyz - array-like variable for pixel dimensions of ROI
        time_steps  - length of trajectory array
        delta_t     - scale for time
        mesh_size   - scale for space
        """
        self.initial_pos_um = mesh_size*np.array(initial_pos) #[um]
        self.imagesz_xyz = np.array(imagesz_xyz)
        self.time_steps = time_steps
        self.delta_t = delta_t #[s]
        self.time = delta_t*np.arange(time_steps) #[s]
        self.mesh_size = mesh_size #[um]

    def make_ballistic_traj(self, direction_vec, speed=1, reflections=True):
        """
        Creates and returns a ballistic trajectory
        Input:
        direction_vec - array-like variable for initial direction (unitless)
        speed - scale factor (units) for direction_vec

        Output:
        traj - numpy array of dimension (time_steps)x3
        """
        vel = speed*np.array(direction_vec)
        traj = np.outer(self.time, vel) + self.initial_pos_um
        if reflections:
            traj = self.reflect_boundaries(traj)
        self.ballistic_traj = traj

    def make_helical_traj(self, radius_xy, freq_xy, z_speed):
        """
        Creates and returns a helical trajectory
        Input:
        radius_xy - array-like variable for radii in x,y [um]
        freq_xy - array-like variable for frequency (inverse period) in x,y [s]**-1
        z_speed - [um][s]**-1

        Output:
        traj - numpy array of dimension (time_steps)x3
        """
        theta_xy = 2*np.pi*np.outer(self.time, np.array(freq_xy))
        traj_xy = np.column_stack((radius_xy[0]*np.cos(theta_xy[0]),
                                   radius_xy[1]*np.sin(theta_xy[1])))
        traj_z = z_speed*self.time + self.initial_pos_um[2]
        traj = np.column_stack((traj_xy, traj_z))

        self.helical_traj = traj

    def make_circular_traj(self, radius_xyz, freq_xyz, offset_xyz):
        """
        Creates and returns a circular trajectory (sinusoidal in x,y,z)
        Input:
        radius_xyz - array-like variable for radii in x,y,z
        freq_xyz - array-like variable for frequency (inverse period) in x,y,z [s]**-1
        offset_xyz - array-like variable for offset of the center in x,y,z
        Output:
        traj - numpy array of dimension (time_steps)x3
        """
        theta_xyz = 2*np.pi*np.outer(np.array(freq_xyz), self.time)
        traj = np.column_stack((offset_xyz[0]+radius_xyz[0]*np.cos(theta_xyz[0]),
                                offset_xyz[1]+radius_xyz[1]*np.sin(theta_xyz[1]),
                                offset_xyz[2]+radius_xyz[2]*np.sin(theta_xyz[2])))
        self.circular_traj = traj

    def make_Grid_cloud(self, Bouders_xyz, Npoints_xyz, num, offset_xyz):
        """
        Create a grid the evaluate all point in 
        Input:
        Bouders_xyz - array-like variable for radii in x,y,z
        Npoints_xyz - array-like variable for frequency (inverse period) in x,y,z [s]**-1
        num         - number of laser shots to make duplicates 
        offset_xyz  - array-like variable for offset of the center in x,y,z
        Output:
        traj - numpy array of dimension (time_steps)x3
        """
        if self.time.shape[0]!=(np.prod(Npoints_xyz)*num).astype('int'):
            print('Shape of product of Npoints and amount of time points do not match!')
            print('Total Npoints:' + str(np.prod(Npoints_xyz)) + ' != Total time points:' + str(self.time.shape[0]/num))
            print('Please change settings in ini file and rerun')
            input("Press Enter to continue...")
            sys.exit()

        x=np.linspace(Bouders_xyz[0,0],Bouders_xyz[0,1],Npoints_xyz[0])+offset_xyz[0]
        y=np.linspace(Bouders_xyz[1,0],Bouders_xyz[1,1],Npoints_xyz[1])+offset_xyz[1]
        z=np.linspace(Bouders_xyz[2,0],Bouders_xyz[2,1],Npoints_xyz[2])+offset_xyz[2]
        
        xx,yy,zz=np.meshgrid(x,y,z)
        #Change the order to move between points with minimal distance
        zz[:,::2,:]=np.flip(zz[:,::2,:],axis=2)
        zz[::2,:,:]=np.flip(zz[::2,:,:],axis=1)
        zz[::2,:,:]=np.flip(zz[::2,:,:],axis=2)
        
        yy[:,::2,:]=np.flip(yy[:,::2,:],axis=2)
        yy[::2,:,:]=np.flip(yy[::2,:,:],axis=1)
        yy[::2,:,:]=np.flip(yy[::2,:,:],axis=2)
        
        xx[:,::2,:]=np.flip(xx[:,::2,:],axis=2)
        xx[::2,:,:]=np.flip(xx[::2,:,:],axis=1)
        xx[::2,:,:]=np.flip(xx[::2,:,:],axis=2)

        if 0: #Debug
            plt.plot(xx.flatten(),'r-');
            plt.plot(yy.flatten(),'g-');
            plt.plot(zz.flatten(),'b-');
            plt.show()
        if 0: #Debug
            plt.plot(np.diff(xx.flatten()),'r-');
            plt.plot(np.diff(yy.flatten()),'g-');
            plt.plot(np.diff(zz.flatten()),'b-');
            plt.show()

        #Repeat each point for the the full laser pattern
        xx=np.repeat(np.expand_dims(xx.flatten(),1),num,1).flatten()
        yy=np.repeat(np.expand_dims(yy.flatten(),1),num,1).flatten()
        zz=np.repeat(np.expand_dims(zz.flatten(),1),num,1).flatten()

        Grid_cloud= np.column_stack(( xx,yy,zz ))

        self.grid_cloud = Grid_cloud

    def make_diffusive_traj(self, diff_coef, offset_xyz=[0,0,0], reflections=True):
        """
        Input:
        diff_coef  - diffusion coefficient [um]**2 [s]**-1
        offset_xyz - move the trajectory to an other start position, [m]

        Output:
        traj - numpy array of dimension (time_steps)x3 in [um]
        """
        # np.random.seed(0)
        print("Diffusion length per step: " + str(np.sqrt(2*diff_coef*self.delta_t)*1e3) + " nm")
        dxdydz = np.random.normal(0, np.sqrt(2*diff_coef*self.delta_t), (int(self.time_steps-1), 3))
        traj = np.cumsum(np.vstack((self.initial_pos_um, dxdydz)), axis=0)
        if reflections:
            print("Calculating reflections")
            traj = self.reflect_boundaries(traj)
            print("Done")
        
        traj[:,0]=traj[:,0]+offset_xyz[0]*1e6
        traj[:,1]=traj[:,1]+offset_xyz[1]*1e6
        traj[:,2]=traj[:,2]+offset_xyz[2]*1e6

        self.diffusive_traj = traj
        
    def make_hmm_diffusion(self, diff_coefs, trans_mat):
        """
        Creates a diffusive trajectory for a particle with N states of diffusion.
        Initial probabilities are chosen from equilibrium, the eigenvector of
        trans_mat with eigenvalue 1.

        Input:

        diff_coefs - array-like param of length N that contains the diffusion
            coefficent of each state. [um]**2 [s]**-1

        trans_mat - (NxN) array-like param that contains the transition
            probabilities (p_ij) between the N diffusion states. p_ij is the
            probability of state i decaying in to state j. Rows must sum to 1.

        Output:

        traj - numpy array of dimensions (time_steps x 3) [um]
        """
        diff_coefs = np.array(diff_coefs)
        trans_mat = np.array(trans_mat)

        # Sanity checks for input parameters
        if trans_mat.size is not len(diff_coefs)**2:
            print(('The number of transition rates and diffusion states do not'
                   'correspond to the same model size.'))
            return
        np.testing.assert_allclose(np.sum(trans_mat, axis=1),
                                   np.ones(len(diff_coefs)),
                                   err_msg='Transition matrix rows do not sum to 1')

        # initial probabilities are stationary (equilibrium) probabilities
        stat_probs = np.linalg.eig(trans_mat.T)[1][:, 0]
        stat_probs /= np.sum(stat_probs)

        # create the Markov chain by iteration
        monte_carlo = np.random.random(self.time_steps-1)
        latent_states = np.zeros(self.time_steps-1, dtype=np.int16)
        latent_states[0] = self.find_next_state(monte_carlo[0], stat_probs)
        for i in range(1, self.time_steps-1):
            latent_states[i] = self.find_next_state(monte_carlo[i],
                                                    trans_mat[latent_states[i-1], :])

        # calculate the displacements in 3D
        diff_states = np.outer(diff_coefs[latent_states], np.ones(3))
        dxdydz = np.random.normal(0, np.sqrt(2*diff_states*self.delta_t))
        init_pos = (0, 0, 0)
        traj = np.cumsum(np.vstack((init_pos, dxdydz)), axis=0)
        self.hmm_traj = traj
        self.latent_states = latent_states

    @staticmethod
    def find_next_state(rand_val, prob_array):
        """
        Generates the next state for a Markov chain

        1) prepend the random value to the cumulative probability
        2) sort the resulting vector
        3) next state is the index position of the random value in the sorted array
        """
        next_state = np.sort(np.insert(np.cumsum(prob_array), 0, rand_val))
        return np.where(next_state == rand_val)[0]

    def reflect_boundaries(self, traj):
        """
        Confines a trajectory within the region defined by imagesz_xyz
        Input:
        traj - numpy array of dimension (time_steps)x3 in [um]
        Output:
        traj - numpy array of dimension (time_steps)x3 in [um]

        Finds first point out of bounds for traj and performs a reflection at
        the appropriate plane. The trajectory is reflected by replacing the
        coordinate normal to the reflecting plane for all steps after the
        reflection such that:

        x_i_refl = sgn(x_i) * (2*x_bound - |x_i|)

        where the index i runs from the boundary crossing to the end of traj.

        Checks by brute force until all boundary crossings are reflected.
        """
        max_xyz = (np.abs(self.imagesz_xyz) * self.mesh_size / 2.0)*1e6
        bound_xings = np.where(np.abs(traj) > max_xyz)
        while bound_xings[0].size:
            first_refl = (slice(bound_xings[0][0],None), bound_xings[1][0])
            traj[first_refl] = np.sign(traj[first_refl])*(2*max_xyz[bound_xings[1][0]]
                                                          -np.abs(traj[first_refl]))
            bound_xings = np.where(np.abs(traj) > max_xyz)

        return traj

    def save_trajs(self, save_path):
        traj_dict = {}
        if hasattr(self, 'ballistic_traj'):
            traj_dict['ballistic_traj'] = self.ballistic_traj
        if hasattr(self, 'helical_traj'):
            traj_dict['helical_traj'] = self.helical_traj
        if hasattr(self, 'circular_traj'):
            traj_dict['circular_traj'] = self.circular_traj
        if hasattr(self, 'diffusive_traj'):
            traj_dict['diffusive_traj'] = self.diffusive_traj
        if hasattr(self, 'grid_cloud'):
            traj_dict['Grid_cloud'] = self.grid_cloud
        if hasattr(self, 'time'):
            traj_dict['time'] = self.time

        savemat(save_path, traj_dict)

    @staticmethod
    def plot_cuboid(ax, max_xyz):
        points_cube = np.zeros((8,3))
        for i, _ in enumerate(points_cube):
            coords = 2*np.float32(np.unpackbits((np.uint8(i)))) - 1
            points_cube[i] = coords[-3:]*max_xyz
        xyz = [[-r, r] for r in max_xyz]
        XZ, YZ = np.meshgrid(*xyz[:-1])
        XY, ZY = np.meshgrid(*xyz[::2])
        YX, ZX = np.meshgrid(*xyz[1:])
        XX, YY, ZZ = np.reshape(np.outer(max_xyz, np.ones(XZ.shape)), (3,2,2))

        ax.plot_surface(XZ,YZ,ZZ, alpha=0.2)
        ax.plot_surface(XZ,YZ,-ZZ, alpha=0.2)
        ax.plot_surface(XZ,YY,ZY, alpha=0.2)
        ax.plot_surface(XZ,-YY,ZY, alpha=0.2)
        ax.plot_surface(XX,YX,ZX, alpha=0.2)
        ax.plot_surface(-XX,YX,ZX, alpha=0.2)
        ax.scatter3D(*points_cube.T)
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')

        return ax




def main():
    # a=np.array([1,2,3,4,5])
    # b=np.array([2,3,4,5,6])
    # c=np.array([20,30,40,50,60])
    # d=np.array([200,300,400,500,600])
    # args=[b,c,d]

    # Dimension=2
    # t_raw, out=physics_Lib.get_diffusion(trajs,timeStep,pixelSize,Dimension)

    from mpl_toolkits.mplot3d import Axes3D # plot bounding cuboid
    from matplotlib import pyplot as plt # plot trajectory
    # p = Particle((-90,-90,-70))
    p = Particle((0,0,0))
    p.make_ballistic_traj((1,2,-3), 10, reflections=False)
    p.make_diffusive_traj(5, reflections=False) # first create traj w/o reflections
    # p.save_trajs('test_trajectories.mat')

    _, ax1 = plt.subplots()
    ax1 = plt.axes(projection='3d')
    ax1.plot3D(*p.ballistic_traj.T)
    ax1 = Particle.plot_cuboid(ax1, p.imagesz_xyz*p.mesh_size/2)

    traj = p.ballistic_traj.copy()
    traj = p.reflect_boundaries(traj)
    _, ax2 = plt.subplots()
    ax2 = plt.axes(projection='3d')
    ax2.plot3D(*traj.T)
    ax2 = Particle.plot_cuboid(ax2, p.imagesz_xyz*p.mesh_size/2)
    plt.show()

    print("Done!")

if __name__ == '__main__':
    main()
