from DataObject import DataObject # Object that will hold all relevent data
from PipelineOrganizer import PipelineOrganizer as _PO
from DataHandler import DataHandler as _DH
from DataPathHandler import DataPathHandler
from PSF_Lib import PSF_Lib as PSF
from Physics_Lib import Particle as _Particle
from PlotLibrary import TrajPlots as _TP

import scipy.io as io
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import mpl_toolkits.mplot3d.axes3d as p3
from matplotlib.gridspec import GridSpec

class Simulator:  
    def __init__(self, ScanProject,txyzn_Traj=None):
        """
        Simularor initialization 
        n_cross:        parameter propotional to the fluorophor absorbtion cross section 
        points_to_use:  amount of full trajectory evaluations that will be used 
                        the trajectory length (amount of tie steps) is then given by:
                       "points_to_use" times the pattern window size "wndw" in the TrajectorySelectionRules section of the ini file 
        """
        if ScanProject.SetupSpecifics['is_simulation']:           
            #Pixel convertion factors  
            self.px2um=ScanProject.SetupSpecifics['Px']*1e6/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])  # from camera pixel to um
            self.spx2um=self.px2um/ScanProject.Ph_Retrieval['XYSWScalefactor']          # from super resolution to um 

            #PSF image offset values 
            if ScanProject.Ph_Retrieval['load_IntencityC']:
                #Get Numerical PSF light field, to get intensity take |LF|^2
                path=ScanProject.Ph_Retrieval['path_PSF']
                fname=ScanProject.Ph_Retrieval['fname_PSF']
                _, parm_PR= PSF.LoadMat(path,fname)
                ScanProject.Ph_Retrieval['IntencityC']=parm_PR['IntencityC']
            
            self.PinholeOffset=ScanProject.Ph_Retrieval['IntencityC'][:,0:2]*self.px2um #[um]  

            #amount of steps rounded to the closest full 
            self.time_steps=ScanProject.TrajectorySelectionRules['wndw']*ScanProject.Sim_par['points_to_use']

            if txyzn_Traj is not None:
               print('Using trajectory from initialisation input, rounding time steps to closest multiple of tracking pattern') 
               self.time_steps=int(int(txyzn_Traj.shape[0]/(ScanProject.TrajectorySelectionRules['wndw']*ScanProject.Sim_par['Subsampling']))*ScanProject.TrajectorySelectionRules['wndw']) #amount of full pattern iterations without subsampling 
               self.ParticleGroundtruth_xyz=txyzn_Traj[:self.time_steps,1:4,:].copy() #Data dimension format is samples, type(xyz extracted from index 1-3), subsamples 
        else:
            raise ValueError('is_simulation is 0 for this Project, can not run simulation')

    def Load_NumPSF():
        SData=io.loadmat("C:/Users/elias.amselem/Desktop/Shadowfax/20200508/beads-xy1umCircle_switch-limit-2000_MAF-24_xyPID_-0.03906/SMatrix.mat")
        S=SData['T0']
        return S

    def generate_Gaussian_NumPFS(xyxO,pixImsizexyz):
        #Assuming 8nm per pixel in x,y,z directions
        # xyxO is in nm 

        Lambda=540e-9 #[m]
        Na    =1.45
        n     =1.515
        sxy   =1.6*(0.822/2)*(Lambda/Na)*1e6  #[um] theoretical PSF width, the radius of the aridisk is scaled to 1.22/2 and for the e-2 intensity drop that is std value one use 0.822/2
        sz    =2*(Lambda*n/Na**2)*1e6         #[um]

        pix2um=8e-3 #[um/pix]
        xpix=np.array(range(pixImsizexyz[0]))*pix2um
        ypix=np.array(range(pixImsizexyz[1]))*pix2um
        zpix=np.array(range(pixImsizexyz[2]))*pix2um
        xx,yy,zz=np.meshgrid(xpix,ypix,zpix)

        xyzs=[sxy,sxy,sz]

        argExp=(xx-xyxO[0])**2 /(2*xyzs[0]**2) + (yy-xyxO[1])**2 /(2*xyzs[1]**2) + (zz-xyxO[2])**2 /(2*xyzs[2]**2)
        Sr=np.exp(-argExp)
        return Sr

    def generate_S(xyxRefOrigo,pixImsizexyz):
        S=np.zeros(tuple(pixImsizexyz + [xyxRefOrigo.shape[0]]))
        for i in range(xyxRefOrigo.shape[0]):
            S[:,:,:,i]=generate_Gaussian_NumPFS(xyxRefOrigo[i,:],pixImsizexyz)
        return S

    def get_LightField(self,ScanProject):
        #Get Numerical PSF light field, to get intensity take |LF|^2
        LF, parm_PR= PSF.LoadMat(ScanProject.Ph_Retrieval['path_PSF'],ScanProject.Ph_Retrieval['fname_PSF'])
        return LF, parm_PR 

    def get_zInterpolation(self, ScanProject,p='Forward'):
        IP=PSF.zInterpolator(self.LF, ScanProject.Ph_Retrieval['ZSWScalefactor'], self.parm_PR,p)
        return IP

    def get_groundtruth(self, ScanProject,do_save=0,imagesz_xyz=(200, 200, 150),subsampling=1):
        #Generate groundtruth data

        SimTrackSizeXYZ = (2*ScanProject.Ph_Retrieval['XYSWScalefactor']*ScanProject.Ph_Retrieval['CropS'], 
                           2*ScanProject.Ph_Retrieval['XYSWScalefactor']*ScanProject.Ph_Retrieval['CropS'],
                           ScanProject.Ph_Retrieval['ZSWScalefactor']*self.LF.shape[2])

        delta_t=1/ScanProject.PosCalibration['ScanSpeedTracking']/subsampling #[s]
        mesh_size =ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2']*ScanProject.Ph_Retrieval['XYSWScalefactor']) #[m]
        particle = _Particle((0,0,0), time_steps=self.time_steps*subsampling, delta_t=delta_t, mesh_size=mesh_size, imagesz_xyz=imagesz_xyz)

        if ScanProject.Sim_par['TrajType']=='circular_traj':
            particle.make_circular_traj(ScanProject.Sim_par['radius_xyz'], ScanProject.Sim_par['freq_xyz'],ScanProject.Sim_par['offset_xyz']) #[m] data saved under particle.circular_traj
            if subsampling>1:
                particle.circular_traj=np.moveaxis(np.reshape(particle.circular_traj,(int(self.time_steps),subsampling,3)),1,-1)
            else:
                particle.circular_traj=np.expand_dims(particle.circular_traj,2)
            self.ParticleGroundtruth_xyz=particle.circular_traj  #Saved data

        elif ScanProject.Sim_par['TrajType']=='diffusive_traj':
            particle.make_diffusive_traj(ScanProject.Sim_par['diff_coef'],offset_xyz=ScanProject.Sim_par['offset_xyz']) #[um] data saved under particle.circular_traj
            #particle.diffusive_traj=(particle.diffusive_traj+[10,10,0]) #add 10um to x,y to bring trajectory to the center of scanning window, this is saved 
            particle.diffusive_traj=particle.diffusive_traj #add 10um to x,y to bring trajectory to the center of scanning window, this is saved 
            if subsampling>1:
                particle.diffusive_traj=np.moveaxis(np.reshape(particle.diffusive_traj,(int(self.time_steps),subsampling,3)),1,-1)
            else:
                particle.diffusive_traj=np.expand_dims(particle.diffusive_traj,2)
            self.ParticleGroundtruth_xyz=particle.diffusive_traj*1e-6  #[m] Make data avilable

        elif ScanProject.Sim_par['TrajType']=='Internal_traj':
            pass

        elif ScanProject.Sim_par['TrajType']=='Grid_cloud':
            particle.make_Grid_cloud(ScanProject.Sim_par['Bouders_xyz'], ScanProject.Sim_par['Npoints_xyz'],ScanProject.PosCalibration['TrackingMemsize'],ScanProject.Sim_par['offset_xyz']) #[m] data saved under particle.Grid_cloud
            if subsampling>1:
                particle.grid_cloud=np.moveaxis(np.reshape(particle.grid_cloud,(int(self.time_steps),subsampling,3)),1,-1)
            else:
                particle.grid_cloud=np.expand_dims(particle.grid_cloud,2)
            self.ParticleGroundtruth_xyz=particle.grid_cloud  #[m] Make data avilable

        #Save data to file if not from Internal_traj
        if ScanProject.Sim_par['TrajType']!='Internal_traj':
            if do_save: #save all trajectories in particle class 
                folder=ScanProject.Files['AnalysisConfig'][:-18]
                file='GroundTruth.mat'
                particle.save_trajs(folder+file)

    def generate_Piezo_EOD_vector(self,ScanProject,time_steps):
        # we will assume here that: 
        #   1. no movement is induced in  Piezo mirror position 
        #   2. no movement is induced in  EOD movement 
        #   3. no scaning is performed only tracking 
        # T,XP,YP,XE,YE,XO,YO,ZI,APD1,APD2,J = TrajObject.get_traj_raw_data2(Trajs, TrajToGet, ScanMatrix,ScanProject)
        
            #Get offset EOD pattern from ini file
        DataHandler=_DH()
        #time_steps=self.time_steps
        TrackingMemsize=ScanProject.PosCalibration['TrackingMemsize']
        EODOffsetValuesXY = DataHandler.GetEODOffsetValueInMicroMeters(ScanProject)[:,:int(TrackingMemsize)] #[um]
        OffsetValueZ = np.array(ScanProject.PosCalibration['zLayerOffset'])[:int(TrackingMemsize)] #[index]
        
            #generate MemIndexTracking vektor and z offset for the full scan 
        usable_time_steps=int(time_steps/TrackingMemsize)
        MemIndexTracking=np.array(tuple(range(int(TrackingMemsize)))*usable_time_steps)
        ZoffsetFullScan=np.array(tuple(OffsetValueZ)*usable_time_steps)

        XO=EODOffsetValuesXY[0,(MemIndexTracking).astype('int')]; YO=EODOffsetValuesXY[1,(MemIndexTracking).astype('int')]
        XP=YP=ZP=XE=YE=np.zeros(int(usable_time_steps*TrackingMemsize))

        zindex=np.zeros(int(usable_time_steps*TrackingMemsize))+ScanProject.PosCalibration['zOffsetIndex']
        ZI =ZoffsetFullScan+zindex #[index] z-value 

        return XP,YP,ZP,XE,YE,XO,YO,ZI,MemIndexTracking

    def generate_SMatrix(self,ScanProject,IP,XP,YP,ZP,XE,YE,XO,YO,ZI,V="standard"):    
        #Generate xiO, index i parameters are reserved for single parttern data sets (size is defined as "TrackingMemsize") 
        #we do this since the Smatrix is assumed to repeat itself each "TrackingMemsize".
        #When reconstructing we can use as many points as we like this is thus only to generate all possible shots and that generates photons
        TrackingMemsize=ScanProject.PosCalibration['TrackingMemsize']

        xiP=XP[:int(TrackingMemsize)]; yiP=YP[:int(TrackingMemsize)]; ziP=ZP[:int(TrackingMemsize)]
        xiE=XE[:int(TrackingMemsize)]; yiE=YE[:int(TrackingMemsize)]
        xiO=XO[:int(TrackingMemsize)]; yiO=YO[:int(TrackingMemsize)]
        zi=ZI[:int(TrackingMemsize)]

        xPinOffset=self.PinholeOffset[(zi).astype('int'),0]; yPinOffset=self.PinholeOffset[(zi).astype('int'),1]
        
        #Generate Smatrix xx,yy,zz are in [um]
        if V=="standard":
            #S,xx,yy,zz=PSF.get_Smatrix2(ScanProject,IP,xiP,yiP,ziP,xiE,yiE,xiO,yiO,xPinOffset,yPinOffset,SaveSmatrix='Static')
            S,xx,yy,zz=PSF.get_Smatrix3(ScanProject,IP,xiO,yiO,xPinOffset,yPinOffset,SaveSmatrix='Static')
            #Setting signal to noise ratio according to the background parameter
            Background=ScanProject.TrajectorySelectionRules['background']
            if isinstance(Background,(np.ndarray)):
                S=S/np.max(S)
                #Background= Background[MemOrderToKeep.astype('int')]

                for j in range(S.shape[-1]):
                    a=np.max(S[:,:,:,j])
                    B=a/(Background[j]-1)
                    S[:,:,:,j]=a*(S[:,:,:,j]+B)/(a+B)
        else:
           print("Wrong input into function: generate_SMatrix")

        ###### Debug plot 
        if 0: 
            for i in range(6):
                plt.contourf(np.sum(S[:,:,:,i],axis=2),alpha=0.3)
            plt.show()

        return S,xx,yy,zz

    def get_ScanMatrix(self,ScanProject,XP,YP,ZP,XE,YE,XO,YO,ZI,n,MemIndexTracking):
        #Construct Scan matrix
        delta_t=1/ScanProject.PosCalibration['ScanSpeedTracking'] #[s]
        
        NoneTrackingBA=np.array([50,50]) #amount of points before and after that is scan data
        ScanMatrix=np.zeros((26,int(self.time_steps)+np.sum(NoneTrackingBA)))
        ScanMatrix[0, :]=np.cumsum(delta_t*np.ones(int(self.time_steps)+np.sum(NoneTrackingBA))) #[s] Time
        ScanMatrix[1, NoneTrackingBA[0]:-NoneTrackingBA[1]]=XP+XE+XO #[um] X value all contributions 
        ScanMatrix[2, NoneTrackingBA[0]:-NoneTrackingBA[1]]=YP+YE+YO #[um] X value all contributions 
        ScanMatrix[3,NoneTrackingBA[0]:-NoneTrackingBA[1]]=n         #Photon counts H all photons are here 

        ScanMatrix[8,NoneTrackingBA[0]:-NoneTrackingBA[1]]=MemIndexTracking   #Memory index 
        ScanMatrix[9, NoneTrackingBA[0]:-NoneTrackingBA[1]]=XP       #[um] X value piezo center 
        ScanMatrix[10,NoneTrackingBA[0]:-NoneTrackingBA[1]]=YP       #[um] Y value piezo center 
        ScanMatrix[23,NoneTrackingBA[0]:-NoneTrackingBA[1]]=XE       #[um] X value EOD center 
        ScanMatrix[24,NoneTrackingBA[0]:-NoneTrackingBA[1]]=YE       #[um] Y value EOD center
        ScanMatrix[11,NoneTrackingBA[0]:-NoneTrackingBA[1]]=XO       #[um] X value EOD tracking pattern
        ScanMatrix[12,NoneTrackingBA[0]:-NoneTrackingBA[1]]=YO       #[um] Y value EOD tracking pattern
        ScanMatrix[19,NoneTrackingBA[0]:-NoneTrackingBA[1]]=ZI-ScanProject.PosCalibration['zOffsetIndex']     #offset +-1 
        ScanMatrix[20,NoneTrackingBA[0]:-NoneTrackingBA[1]]=ZP/(ScanProject.PosCalibration['PositionLength']/(ScanProject.PosCalibration['zScale'] * 2**14 )) #raw signal 
        ScanMatrix[21,NoneTrackingBA[0]:-NoneTrackingBA[1]]=ZP/(ScanProject.PosCalibration['PositionLength']/(ScanProject.PosCalibration['zScale'] * 2**14 )) #raw signal 
        ScanMatrix[22,NoneTrackingBA[0]:-NoneTrackingBA[1]]=ZI*ScanProject.PosCalibration['SLMSwitchHight'] #[index] z-value 
        
        ScanMatrix[5,NoneTrackingBA[0]:-NoneTrackingBA[1]]=np.ones(int(self.time_steps)) #tracking on/off
        if len(ScanProject.ScanMatrix)==0:
            ScanProject.ScanMatrix=ScanMatrix
        else:
            print('Appending data to ScanMatrix')
            ScanMatrix[0,:]=ScanMatrix[0,:]+ScanProject.ScanMatrix[0,:][-1]
            ScanProject.ScanMatrix=np.hstack((ScanProject.ScanMatrix,ScanMatrix))
        print('Data generation is Done!')

    def get_PhotonMeanValue(self,ScanProject,S,xx,yy,zz,MemIndexTracking,subsampling=1):
        #Create measured photons for full trajectory assuming that it is only the particle moving,   
        ParticleGroundtruth_xyz=np.reshape(np.moveaxis(self.ParticleGroundtruth_xyz,1,-1),(int(self.time_steps*subsampling),3)) #[um] Converted in get_SMatrix
        
        #find closest pixel to the particle and get probability 
        p=np.zeros(int(self.time_steps*subsampling))
        for i in range(int(self.time_steps)): #this loop can be done in parallel/threding 
            for j in range(subsampling):
                print('Generating photon probabilities: ' + str(i*subsampling+j) + "/" + str(int(self.time_steps*subsampling)) + ', using subsampling=' + str(subsampling))
                r2=(xx-ParticleGroundtruth_xyz[i*subsampling+j,0])**2 + (yy-ParticleGroundtruth_xyz[i*subsampling+j,1])**2 + (zz-ParticleGroundtruth_xyz[i*subsampling+j,2])**2
                Groundtruth_yxz_index=np.unravel_index(np.argmin(r2),xx.shape)
                p[i*subsampling+j]=S[Groundtruth_yxz_index[0],Groundtruth_yxz_index[1],Groundtruth_yxz_index[2],MemIndexTracking[i]]

                if 1:
                    print('Groundtruth: ' + '\n' + 
                          'in=' + str(Groundtruth_yxz_index) + '\n' + 
                          'pos=' + str((xx[Groundtruth_yxz_index], yy[Groundtruth_yxz_index],zz[Groundtruth_yxz_index])) )
                    print("lambda=" + str(ScanProject.Sim_par['n_cross']*p[i]))
        

        n_mean=(ScanProject.Sim_par['n_cross']*p + ScanProject.Sim_par['background'])/subsampling
        if subsampling>1:
            n_mean=np.reshape(n_mean,(int(self.time_steps),subsampling))

        return n_mean 

    def get_PhotonStatistics(self,ScanProject,n_mean,subsampling=1):
        if ScanProject.Sim_par['PhotonStatistics']=='mean':
            if subsampling>1:
                n=np.sum(n_mean,axis=1)
            else:
                n=n_mean
        elif ScanProject.Sim_par['PhotonStatistics']=='poisson':
            n=np.random.poisson(n_mean,n_mean.shape)
            if subsampling>1:
                n=np.sum(n,axis=1)
        return n

    def get_SMatrix(self,ScanProject,p='Forward',imagesz_xyz=(200, 200, 150),PiezoFollow=0,subsampling=1):
        """
        Takes a ScanProject with minimal content and generates all parameters needed to genarat photon counts and ScanMatrix
         
        """
        self.LF, self.parm_PR = self.get_LightField(ScanProject)
        self.get_groundtruth(ScanProject,do_save=ScanProject.Sim_par['Save_groundtruth'],imagesz_xyz=imagesz_xyz,subsampling=subsampling)
        IP=self.get_zInterpolation(ScanProject,p)
        # Add background determined by the signal to noise ratio 
        Background=ScanProject.TrajectorySelectionRules['background']
        if not isinstance(Background,(np.ndarray)):
            IP=IP+Background
            if Background<0:
                IP[np.where(IP<0)]=0

        if ScanProject.Ph_Retrieval['load_IntencityC']:
            ScanProject.Ph_Retrieval['IntencityC']=self.parm_PR['IntencityC']
            self.PinholeOffset=ScanProject.Ph_Retrieval['IntencityC'][:,0:2]*self.px2um

        XP,YP,ZP,XE,YE,XO,YO,ZI,MemIndexTracking=self.generate_Piezo_EOD_vector(ScanProject,time_steps=self.time_steps)
        S,xx,yy,zz = self.generate_SMatrix(ScanProject,IP,XP,YP,ZP,XE,YE,XO,YO,ZI)
        zz=-zz #Flip z 

        if np.sum(PiezoFollow)>0:
            # Create a slow moving for pinhole, within the pinhole the beams are fixed, 
            # We will subtract this from the ground truth to simulate movement of the pinhole but not the EOD and add 10 to get it in the center of the scaning area 
            # This will be re added when the trajectory is rebuild, note that this is a very crued simulation since it uses the ground truth
            if np.isscalar(PiezoFollow):
                window=[PiezoFollow,PiezoFollow,PiezoFollow]
            else:
                window=PiezoFollow

            XP=(np.convolve(np.ones(int(window[0]))/window[0],np.concatenate((self.ParticleGroundtruth_xyz[0,0,0]*np.ones(int(window[0])),self.ParticleGroundtruth_xyz[:,0,0])))[int(window[0]):-int(window[0])+1])*1e6 #[um]
            YP=(np.convolve(np.ones(int(window[1]))/window[1],np.concatenate((self.ParticleGroundtruth_xyz[0,1,0]*np.ones(int(window[1])),self.ParticleGroundtruth_xyz[:,1,0])))[int(window[1]):-int(window[1])+1])*1e6 #[um]
            ZP=(np.convolve(np.ones(int(window[2]))/window[2],np.concatenate((self.ParticleGroundtruth_xyz[0,2,0]*np.ones(int(window[2])),self.ParticleGroundtruth_xyz[:,2,0])))[int(window[2]):-int(window[2])+1])*1e6 #[um]
            
        if 0: #Debug
            plt.plot(self.ParticleGroundtruth_xyz[:,0,0]*1e6,'-b')
            plt.plot(XP,'--b')
            plt.plot(self.ParticleGroundtruth_xyz[:,0,0]*1e6-XP,'-b')

            plt.plot(self.ParticleGroundtruth_xyz[:,1,0]*1e6,'-r')
            plt.plot(YP,'--r')
            plt.plot(self.ParticleGroundtruth_xyz[:,1,0]*1e6-YP,'r')

            plt.plot(self.ParticleGroundtruth_xyz[:,2,0]*1e6,'-g')
            plt.plot(ZP,'--g')
            plt.plot(self.ParticleGroundtruth_xyz[:,2,0]*1e6-ZP,'g')

            plt.grid(True)
            plt.show()
        
        #Correct particle ground truth and convert to um
        self.ParticleGroundtruth_xyz[:,0,:]=self.ParticleGroundtruth_xyz[:,0,:]*1e6-np.reshape(np.repeat(XP,subsampling),(XP.shape[0],subsampling))  #[um]
        self.ParticleGroundtruth_xyz[:,1,:]=self.ParticleGroundtruth_xyz[:,1,:]*1e6-np.reshape(np.repeat(YP,subsampling),(YP.shape[0],subsampling))  #[um]
        self.ParticleGroundtruth_xyz[:,2,:]=self.ParticleGroundtruth_xyz[:,2,:]*1e6-np.reshape(np.repeat(ZP,subsampling),(ZP.shape[0],subsampling))  #[um]

        return S,xx,yy,zz,XP,YP,ZP,XE,YE,XO,YO,ZI,MemIndexTracking
    
    def generate_ScanMatrix(self,ScanProject,imagesz_xyz=(200, 200, 150),PiezoFollow=0,subsampling=1):
        """
        Takes a ScanProject with minimal content and generates:
        1. ScanMatrix with synthetick data defined by the ini file in the project folder
        """
        S,xx,yy,zz,XP,YP,ZP,XE,YE,XO,YO,ZI,MemIndexTracking=self.get_SMatrix(ScanProject,p='WeightedTwoWay',imagesz_xyz=imagesz_xyz,PiezoFollow=PiezoFollow,subsampling=subsampling)

        n_mean     = self.get_PhotonMeanValue(ScanProject,S,xx,yy,zz,MemIndexTracking,subsampling=subsampling)
        n          = self.get_PhotonStatistics(ScanProject,n_mean,subsampling=subsampling)
        print("Mean photon number: " + str(np.mean(n)))
        self.get_ScanMatrix(ScanProject,XP,YP,ZP,XE,YE,XO,YO,ZI,n,MemIndexTracking)
        return n_mean
        

def main():
    
    #1. Generate data:
    #   Laser shot posisions (Need S matrix for this)
    #   Generat mean Photon  
    #   Generate Poisson photon statistics
    #   Generate offset vektor 
    #   Generate zlevel used   
    #
    #2. Generate Scan matrix 
    #   See: get_traj_raw_data2 for all scan data needed to be generated
    #   Everything else is mostlikly from the ini file 
    

    #Get ini file
    
    prefix = "C:/Users/elias/Desktop/Shadowfax/Data"
    pathnames = [
                 '/20201117_(20201113_PSF)_SIM/Simulation_1_R16383_48pointSlow/',             #0
                 ]
    
    ProjectToRun=0
    for Project_Folder in [ProjectToRun]:
        #Create a ScanProject
        data_path_handler= DataPathHandler(prefix,pathnames)
        data_path_handler.get_data_paths() # pathes are stored in, data_path_handler.Files

        DataHandler=_DH() #Class object that can process data 
        ScanProject = DataObject(data_path_handler, Project_Folder,loadSaveData=0) #Class object that stores all data for a given scan

        #Start a Simulator        
        Sim=Simulator(ScanProject)
        imagesz_xyz=(100,100,100) #Cell shape 10nm per pix 
        subsampling=10
        Sim.generate_ScanMatrix(ScanProject,imagesz_xyz=imagesz_xyz,PiezoFollow=0,subsampling=subsampling)
        DataHandler.PickleData(ScanProject,override=0)


if __name__ == '__main__':
    main()