from Physics_Lib import physics_Lib as _PhLib

import matplotlib.pyplot as plt
import matplotlib.cm as cm
import mpl_toolkits.mplot3d.axes3d as p3
from matplotlib.gridspec import GridSpec

import numpy as np
import scipy.io as io
import scipy.stats as stats    
import scipy as _sp
import scipy.interpolate

import tensorflow.compat.v2 as tf
tf.enable_v2_behavior()
import tensorflow_probability as tfp
from tensorflow_probability import distributions as tfd

import numpy as np
import tensorflow as tf
 


def main_HMMTensorFlow():
    """
    This is a test of the tensorflow HMM with a poisson emission model
    https://www.tensorflow.org/probability/examples/Multiple_changepoint_detection_and_Bayesian_model_selection
    See also:
    https://colab.research.google.com/drive/1IUe9lfoIiQsL49atSOgxnCmMR_zJazKI
    """
    #Create data
    true_rates = [40, 3, 20, 50]
    true_durations = [10, 20, 5, 35]

    observed_counts = np.concatenate([
                                       scipy.stats.poisson(rate).rvs(num_steps)
                                       for (rate, num_steps) in zip(true_rates, true_durations)
                                    ]).astype(np.float32)

    #plt.plot(observed_counts)
    
    #Construct model
    num_states = 4
    initial_state_logits = np.zeros([num_states], dtype=np.float32) # uniform distribution

    daily_change_prob = 0.05
    transition_probs = daily_change_prob / (num_states-1) * np.ones([num_states, num_states], dtype=np.float32)
    np.fill_diagonal(transition_probs,1-daily_change_prob)

    print("Initial state logits:\n{}".format(initial_state_logits))
    print("Transition matrix:\n{}".format(transition_probs))
    
    trainable_log_rates = tf.Variable(np.log(np.mean(observed_counts)) + tf.random.normal([num_states]), name='log_rates')

    hmm = tfd.HiddenMarkovModel(
                                initial_distribution=tfd.Categorical(logits=initial_state_logits),
                                transition_distribution=tfd.Categorical(probs=transition_probs),
                                observation_distribution=tfd.Poisson(log_rate=trainable_log_rates),
                                num_steps=len(observed_counts))

    #function to optimize
    rate_prior = tfd.LogNormal(5, 5)
    def log_prob():
        return (tf.reduce_sum(rate_prior.log_prob(tf.math.exp(trainable_log_rates))) + hmm.log_prob(observed_counts))

    #How to optimize
    optimizer = tf.keras.optimizers.Adam(learning_rate=0.1)

    @tf.function(autograph=False)
    def train_op():
        with tf.GradientTape() as tape:
            neg_log_prob = -log_prob()
        grads = tape.gradient(neg_log_prob, [trainable_log_rates])[0]
        optimizer.apply_gradients([(grads, trainable_log_rates)])
        return neg_log_prob, tf.math.exp(trainable_log_rates)


    for step in range(501):
        loss, rates = [t.numpy() for t in train_op()]
        if step % 20 == 0:
            print("step {}: log prob {} rates {}".format(step, -loss, rates))

    print("Inferred rates: {}".format(rates))
    print("True rates: {}".format(true_rates))

    #Prediction
    #Most probable state at each time point
    posterior_dists = hmm.posterior_marginals(observed_counts)
    posterior_probs = posterior_dists.probs_parameter().numpy()
    most_probable_states = np.argmax(posterior_probs, axis=1)
    most_probable_rates = rates[most_probable_states]
    
    #Most probable trajectory
    most_probable_traj=hmm.posterior_mode(observed_counts).numpy()
    most_probable_traj_rates = rates[most_probable_traj]

    #plot
    def plot_state_posterior(ax, state_posterior_probs, title):
        ln1 = ax.plot(state_posterior_probs, c='blue', lw=3, label='p(state | counts)')
        ax.set_ylim(0., 1.1)
        ax.set_ylabel('posterior probability')
        ax2 = ax.twinx()
        ln2 = ax2.plot(observed_counts, c='black', alpha=0.3, label='observed counts')
        ax2.set_title(title)
        ax2.set_xlabel("time")
        lns = ln1+ln2
        labs = [l.get_label() for l in lns]
        ax.legend(lns, labs, loc=4)
        ax.grid(True, color='white')
        ax2.grid(False)

    fig = plt.figure(figsize=(10, 10))
    plot_state_posterior(fig.add_subplot(2, 2, 1),
                         posterior_probs[:, 0],
                         title="state 0 (rate {:.2f})".format(rates[0]))
    plot_state_posterior(fig.add_subplot(2, 2, 2),
                         posterior_probs[:, 1],
                         title="state 1 (rate {:.2f})".format(rates[1]))
    plot_state_posterior(fig.add_subplot(2, 2, 3),
                         posterior_probs[:, 2],
                         title="state 2 (rate {:.2f})".format(rates[2]))
    plot_state_posterior(fig.add_subplot(2, 2, 4),
                         posterior_probs[:, 3],
                         title="state 3 (rate {:.2f})".format(rates[3]))
    plt.tight_layout()
    plt.show()


    fig = plt.figure(figsize=(10, 4))
    ax = fig.add_subplot(1, 1, 1)
    ax.plot(most_probable_rates, c='green', lw=3, label='inferred rate')
    ax.plot(most_probable_traj_rates, c='red', lw=3, label='inferred rate t')    
    ax.plot(observed_counts, c='black', alpha=0.3, label='observed counts')
    ax.set_ylabel("latent rate")
    ax.set_xlabel("time")
    ax.set_title("Inferred latent rate over time")
    ax.legend(loc=4)
    plt.show()

def main_HMMTensorFlow_Simulation():
    """
    Test of MSD and CVE for a simulated Brownian motion
    """
    do_save_plot=0

    ##### Generate HMM data 
    #Simulation track
    substeps=10                           #Simulate blur this will give R=1/6.2
    dt= 0.21e-3                           #[s]
    sigma_exyz= [30e-3,30e-3,40e-3]       #[um]
    D_true    =np.array([3,1,0.01])           #[um^2/s] mean of the D inverse gamma distribution 
    state_lifetime=np.array([5e-3,2e-3,100e-3])  #[s]
    Mode=np.array([0.45*D_true[0],0.55*D_true[1],0.55*D_true[2]]) #Mode of the D inverse gamma distribution

        #HMM Model 
    aii=1-(dt/state_lifetime)  #Probability to stay in the position is given by the life time 
    a23=a32=0                  #Moving from Test state (state 2) to a Bound state (state 3) can not be done 
    a21=(1-aii[1])             #a21+a22+a23=1
    a31=(1-aii[2])             #a31+a32+a33=1
    a12=(1-aii[0])*0.90        #Moving from Free state (state 1) to a Test state (state 2) is more probabule  
    a13=(1-aii[0])*0.10        #then moving from a Free state (state 1) to a Bound state (state 3)
    
    trans_mat = np.array([[aii[0],a12,a13],
                          [a21,aii[1],a23],
                          [a31,a32,aii[2]]])
    n_sequences=50                  #Amount of trajectories to build 
    length_limits=[300,4000]        #trajectory length will be random between these valuse
    DBias=100                       #Analysis Bias to avoide negative values 


    num_states = trans_mat.shape[0]
    initial_state_logits = np.zeros([num_states], dtype=np.float32) # uniform distribution

    transition_probs = trans_mat.astype(np.float32)
    alpha_true=(Mode+D_true)/(D_true -Mode)
    beta_true=D_true*(alpha_true-1)

    trainable_alpha = tf.Variable(alpha_true, name='alpha')
    trainable_beta = tf.Variable(beta_true, name='beta')
    
    hmm_state_predictor = tfd.HiddenMarkovModel(initial_distribution=tfd.Categorical(logits=initial_state_logits),
                                                transition_distribution=tfd.Categorical(probs=transition_probs),
                                                observation_distribution=tfd.Normal(loc=np.arange(num_states).astype(np.float32),scale=0.001*np.ones(num_states).astype(np.float32)),
                                                num_steps=500)

    samples_all = np.round(hmm_state_predictor.sample(sample_shape=n_sequences).numpy())

    StateGroundTruth=()
    for i in range(n_sequences):
        StateGroundTruth+=(samples_all[i,:],)

    #############################################################################
    # Create 3D Brownian motion using StateGroundTruth
    traj_all=()
    for sequence in StateGroundTruth:
        var=np.sqrt(2*D_true[sequence.astype('int')]*(dt/substeps))
        var_substeps=np.repeat(np.expand_dims(var,axis=1),substeps,axis=1)
        var3D_substeps=np.repeat(np.expand_dims(var_substeps,axis=2),3,axis=2)
        dxdydz = np.random.normal(0, var3D_substeps)
        dxdydz=dxdydz.reshape((dxdydz.shape[0]*dxdydz.shape[1],dxdydz.shape[2]),order='C')
        dxdydz[0,:]=0
            
        trajT = np.cumsum(dxdydz, axis=0)
        traj=np.zeros((len(sequence),3))
        for i in range(3):
            if substeps>1:
                traj[:,i]= np.convolve(np.ones(substeps)/substeps,trajT[:,i])[substeps-1:-substeps+1:substeps]
            else:
                traj[:,i]=trajT[:,i]
        traj_all += (traj,)
    
    TrajDictionary={}
    for TrajToGet in range(len(traj_all)):
        traj=traj_all[TrajToGet]
        k=StateGroundTruth[TrajToGet]

        traj= traj +  np.random.normal(0, np.ones(traj.shape)*np.array([sigma_exyz])) #[um]
        t_track=dt*np.arange(traj.shape[0])
        txyzn_Map=np.vstack((t_track,traj.T,k))

        TrajDictionary[TrajToGet]=txyzn_Map
    

    
    ################################################################################################
    ### Estimate D and Sigma
    ###############################################################################################
    if 1: #CVE estimation of D and sigma
        j=4;n=1;window=5; Flag=2
        result_txyz=_PhLib.get_CVE_MSD_traj(TrajDictionary,wndw=window,R=1/6.2,j=j,n=n,Flag=Flag)
        Name_prefix='CVE_2stateHMM'
        Name_Mid='_j'+ str(j) + '_n' + str(n) + '_F' + str(Flag) + '_'

    else: #MSD estimation of D and sigma 
        PointsToUse=4;  MSDStart=1; window=5; Flag=0
        MSDSteps=window-MSDStart
        result_txyz_temp =_PhLib.get_MSD_traj2(TrajDictionary,window=window,MSDStart=MSDStart,MSDSteps=MSDSteps,PointsToUse=PointsToUse,R=1/6.2,Flag=Flag)

        result_txyz={}
        for TrajToGet in range(len(result_txyz_temp)):
            t=result_txyz_temp[str(TrajToGet),'TDS2'][0]   #[s]
            D=result_txyz_temp[str(TrajToGet),'TDS2'][1]   #[um^2/s]
            S2= result_txyz_temp[str(TrajToGet),'TDS2'][2] #[um^2]
            result_txyz[str(TrajToGet),'D']=np.append(np.expand_dims(t,axis=1).T,D,axis=0).T
            result_txyz[str(TrajToGet),'Sigma2']=np.append(np.expand_dims(t,axis=1).T,S2,axis=0).T           

        Name_prefix='MSD'
        Name_Mid= '_n' + str(PointsToUse) + '_MS' + str(MSDStart) +'_F' + str(Flag) + '_'
   
    #Find HMM 
    ##################################
    est_num_states = 3
    observed_counts=result_txyz['0','D']

    est_initial_state_logits = np.zeros([num_states], dtype=np.float32) # uniform distribution

    est_transition_probs = np.random.rand(est_num_states,est_num_states).astype(np.float32)
    est_transition_probs=est_transition_probs/np.repeat(np.expand_dims(np.sum(est_transition_probs,axis=1),axis=1),est_num_states,1)

    trainable_alpha = tf.Variable((np.random.rand(est_num_states)*10).astype(np.float32), name='alpha')
    trainable_beta = tf.Variable((np.random.rand(est_num_states)*10).astype(np.float32), name='beta')
    
    est_hmm_state_predictor = tfd.HiddenMarkovModel(initial_distribution=tfd.Categorical(logits=initial_state_logits),
                                                transition_distribution=tfd.Categorical(probs=transition_probs),
                                                observation_distribution=tfd.InverseGamma(concentration=trainable_alpha, scale=trainable_beta),
                                                num_steps=500)


    #function to optimize
    def log_prob():
        return (hmm.log_prob(observed_counts))

    #How to optimize
    optimizer = tf.keras.optimizers.Adam(learning_rate=0.1)

    @tf.function(autograph=False)
    def train_op():
        with tf.GradientTape() as tape:
            neg_log_prob = -log_prob()
        grads = tape.gradient(neg_log_prob, [trainable_log_rates])[0]
        optimizer.apply_gradients([(grads, trainable_log_rates)])
        return neg_log_prob, tf.math.exp(trainable_log_rates)


    for step in range(501):
        loss, rates = [t.numpy() for t in train_op()]
        if step % 20 == 0:
            print("step {}: log prob {} rates {}".format(step, -loss, rates))

    print("Inferred rates: {}".format(rates))
    print("True rates: {}".format(true_rates))



    #####################################################################
    HMM_n_components=3
    Dinv_all=()
    for i in range(int(len(result_txyz)/3)):
        Dinv=1/(np.mean(result_txyz[str(i),'D'][:,1:],axis=1)+DBias)
        Dinv_all+=(Dinv,)
    dt_est=np.mean(np.diff(result_txyz[str(0),'D'][:,0]))
    
    FoundModel=_HMM.train_HMM(X=Dinv_all, n_components=HMM_n_components)
    new_trans_mat, dist_params =_HMM.get_model_parm(FoundModel) 
    
    new_lifetime=()
    for i in range(new_trans_mat.shape[0]):
        new_lifetime+=(dt_est/(1-new_trans_mat[i,i]),) #[s]
    new_lifetime=np.array(new_lifetime)

    print(trans_mat)
    print(new_trans_mat)

    D_estimated=()
    for state in dist_params:
        D_estimated+=(dist_params[state][1]/(dist_params[state][0]-1) -DBias,)
    predicted_state_sequence=_HMM.predict_sequence(model=FoundModel,observed_sequence=Dinv_all)
    
    _HMM.plot_hist_sequence_with_fit(Dinv_all,dist_params,xoffset=DBias)

    # Generate the inverse-Gamma distributions to plot
    plotRange=[-5,20]
    bins=200
    DFittParam=np.linspace(plotRange[0],plotRange[1],num=bins*10)
    invGamma={}
    for state in dist_params:
        param=dist_params[state]
        invGamma[state]=stats.invgamma.pdf(DFittParam+DBias, a=param[0],scale=param[1])


    ###############################################################
    ## Ploting
    for TrajToGet in range(10):#range(len(TrajDictionary)):
        txyzn_Map=TrajDictionary[TrajToGet]

        t=result_txyz[str(TrajToGet),'D'][:,0]
        D=np.transpose(result_txyz[str(TrajToGet),'D'][:,1:])
        S2=np.transpose(result_txyz[str(TrajToGet),'Sigma2'][:,1:])
 
        ts=()
        for tsearch in t:
            ts+=(np.argmin((txyzn_Map[0,:]-tsearch)**2),)
        
        ST  =D_true[txyzn_Map[4,:].astype('int')][np.array(ts)]
        SEst=np.array(D_estimated)[predicted_state_sequence[TrajToGet]]
        StateTransError=ST-SEst       

        ###############################################################
        # Figure construction CVE
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(2*6, 2*5, forward=True)
        gs = GridSpec(3,4, figure=fig)
        ax0=fig.add_subplot(gs[0, 0:3]) #Trajectory 
        ax1=fig.add_subplot(gs[1, 0:3]) #Diffusion
        ax2=fig.add_subplot(gs[2, 0:3]) #Localization error 
        ax3=fig.add_subplot(gs[1, 3])   #Diffusion histogram 
        ax4=fig.add_subplot(gs[2, 3])   #Localization error  histogram
        ax5=fig.add_subplot(gs[0, 3])   #HMM error distribution 

        ax0.set_title(Name_prefix + ' method for D and sigma')
        colors=['gray','bisque','white']

        StateSection=np.where((np.diff(txyzn_Map[4,:])!=0)*1)[0]
        StartSection=np.append(np.array([0]),StateSection+1)
        EndSection=np.append(StateSection+1,np.array(len(txyzn_Map[4,:])-1))
        #D_labelsTracker=len(D_true)

        for i in range(len(StartSection)):
            ax0.axvspan(txyzn_Map[0,StartSection[i]], txyzn_Map[0,EndSection[i]], alpha=0.5, color=colors[txyzn_Map[4,StartSection[i]].astype('int')],lw=0)#label=
            #print(str(D_true[txyzn_Map[4,StartSection[i]].astype('int')]))
        ax0.plot(txyzn_Map[0,:],txyzn_Map[1,:],'r'); ax0.plot(txyzn_Map[0,:],txyzn_Map[2,:],'g'); ax0.plot(txyzn_Map[0,:],txyzn_Map[3,:],'b')
        ax0.set_ylabel('um');ax0.grid();#ax0.legend()    
        
        # Diffusion constant estimation 
        for i in range(len(StartSection)):
            ax1.axvspan(txyzn_Map[0,StartSection[i]], txyzn_Map[0,EndSection[i]], alpha=0.5, color=colors[txyzn_Map[4,StartSection[i]].astype('int')],lw=0)
        ax1.plot(t,np.mean(D,axis=0),'k')
        ax1.plot(txyzn_Map[0,:],D_true[txyzn_Map[4,:].astype('int')],'g',label='D-true:' + str(np.round(D_true,decimals=2)) + ' Ltime:' + str(np.round(state_lifetime*1e3,decimals=2)) + 'ms') 
        ax1.plot(t,np.array(D_estimated)[predicted_state_sequence[TrajToGet]],'r',label='D-est:'+str(np.round(np.array(D_estimated),decimals=2)) + ' Ltime:' + str(np.round(np.array(new_lifetime)*1e3,decimals=2)) + 'ms')
        ax1.set_ylabel('um^2/s');ax1.grid()
        ax1.legend()
        
        #Localization estimation 
        for i in range(len(StartSection)):
            ax2.axvspan(txyzn_Map[0,StartSection[i]], txyzn_Map[0,EndSection[i]], alpha=0.5, color=colors[txyzn_Map[4,StartSection[i]].astype('int')],lw=0)
        ax2.plot(t,np.mean(S2,axis=0)*1e6,'k')
        ax2.plot(t,(np.ones(len(t))*np.mean(sigma_exyz)**2)*1e6,'r',label='sigma-true: ' + str(np.round(np.mean(sigma_exyz)*1e3,decimals=2)) + 'nm')
        ax2.set_ylabel('nm^2');ax2.grid()
        ax2.set_xlabel('s')
        ax2.legend()
        
        # Histograms D
        for state in invGamma:
            ax3.plot(DFittParam,invGamma[state],label=state)
            ax3.axvline(dist_params[state][1]/(dist_params[state][0]-1) -DBias, color='g',linestyle='--',label=state + ' mean:' + str(np.round(dist_params[state][1]/(dist_params[state][0]-1)-DBias,decimals=2)))
            ax3.axvline(dist_params[state][1]/(dist_params[state][0]+1) -DBias, color='r',linestyle='--',label=state + ' mode:' + str(np.round(dist_params[state][1]/(dist_params[state][0]+1)-DBias,decimals=2)))

        ax3.hist(np.mean(D,axis=0),bins=10,density=True,facecolor='b',label='M:'+ str(np.round(np.mean(np.mean(D,axis=0)),decimals=2)) +  ' STD: '+ str(np.round(np.std(np.mean(D,axis=0)),decimals=2)))
        ax3.set_ylim(0,1)
        ax3.legend()

        # Histograms S2
        ax4.hist(np.mean(S2,axis=0)*1e6,bins=100,label='sqrt(|M|):'+ str(np.round(np.sqrt(np.abs(np.mean(np.mean(S2,axis=0))))*1e3,decimals=0).astype('int')) + 'nm \n' + ' sqrt(STD): '+ str(np.round(np.sqrt(np.std(np.mean(S2,axis=0)))*1e3,decimals=0).astype('int')) + 'nm')
        ax4.legend();ax4.set_xlabel('nm^2')
        
        #Histogram State estimation
        ax5.hist(StateTransError,bins=20,label='StateTrue-StateEst')
        ax5.legend()

        if do_save_plot:
            file= Name_prefix + Name_Mid + str(TrajToGet) +'.png'
            path="C:/Users/elias/Desktop/Shadowfax/Data/Debug/"
            fig.savefig(path + file)
            plt.close()
        else:
            plt.show()

        print('Done!')

if __name__ == '__main__':
    main_HMMTensorFlow()
    #main_HMMTensorFlow_Simulation()



    
   
