from pathlib import Path
import os
import numpy as np
import scipy.special as _scs
import scipy.io as io
import cv2
import time
from PSF_Lib import PSF_Lib as PSF
from PSF_Lib import optimizers as opt
from scipy.optimize import fmin
from miscellaneous import miscellaneous as _mis

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
import mpl_toolkits.mplot3d.axes3d as p3
from itertools import product, combinations

from DataFileIO_Lib import DataFileIO_Lib as _DS
from PlotLibrary import TrajPlots as _TP

#from mayavi import mlab

class TrajObject:
    def __init__(self, ScanObject):
        """
        Class for finding Trajectories that fit the selection criteria 
        """
        if not hasattr(ScanObject,'Trajs'):
            ScanObject.Trajs = np.array([])
            ScanObject.TrajInfo = []
        print("TrajObject initialized")

    def FindTrajs(self, ScanProject,Filter=1):
        """
        Finds the start and end indices of all trajectories
        Also finds the length (in seconds) of all trajectories
        This do not include any filtering, it  looks at
        1. The flags TrackingOn and ReleaseOff logic and vector, call it R
        2. Start is defined as diff(R)=1, thus a rising edge
        3. End   is defined as diff(R)=-1,thus a falling edge
        4. If there are more starts then ends remove the last End
        5. Trajectory time is defined as time of End - Start  
        """
        # Tracking on and Release Threshold not reached 
        TrackingOn = ScanProject.ScanMatrix[5,:]
        ReleaseOff = ScanProject.ScanMatrix[6,:]
        ScanTime   = ScanProject.ScanMatrix[0,:]

        #Note: ReleaseOff has a strange behavior, removing it until we know more
        #TrajThreshold = np.logical_and(TrackingOn, ReleaseOff)
        TrajThreshold = np.int32(TrackingOn*Filter)
        StartTraj = np.array(np.where(np.diff(TrajThreshold.astype(int)) == 1))+1
        EndTraj = np.array(np.where(np.diff(TrajThreshold.astype(int)) == -1))
        StartTraj = StartTraj[0:np.size(EndTraj)] # Trim down to equal length
        TrajTimeLength = ScanTime[EndTraj] - ScanTime[StartTraj]

        return StartTraj, EndTraj, TrajTimeLength

    def WeedTrajs(self, ScanProject, StartTraj, EndTraj, TrajTimeLength):
        """
        Weeds out unwanted trajectories according to the selection criteria from the .ini file
        """
        verbose = ScanProject.TrajectorySelectionRules['verbose']
        MinTrajTimeLen = ScanProject.TrajectorySelectionRules['MinTrajectoryTimeLength']
        GreaterThanMinTrajLen = np.where(TrajTimeLength > MinTrajTimeLen)

        StartTrajWeeded = StartTraj[GreaterThanMinTrajLen]
        EndTrajWeeded = EndTraj[GreaterThanMinTrajLen]
        TrajTimeLengthWeeded = TrajTimeLength[GreaterThanMinTrajLen]

        if verbose:
            print('Found', StartTrajWeeded.size, 'trajectories')
        
        TrajInfo = [
            'Found ' + str(StartTrajWeeded.size) + ' trajectories',
            '0. Trajectory time length [s]',
            '1. Start index',
            '2. End index'
        ]
        
        Trajs = np.vstack((
            TrajTimeLengthWeeded,
            StartTrajWeeded,
            EndTrajWeeded
        ))
        return Trajs, TrajInfo

    def TrimTrajs(self, ScanProject, StartTrim, EndTrim, ForceMaxTrajLength=-1):
        """
        Assumse that getTraj has alrady been run
        Trim trajectorise:
        StartTrim [s]:   shave of from the begining, single number or vector of the same length as self.Trajs
        ENDTrim   [s]:   shave of from the end, single number or vector of the same length as self.Trajs
        ForceLength [s]: Cut traj to a max length pices

        This ends by running weedtraj thus removing any traj that is to short 
        OBS! This function overrides the old ScanProject.Trajs, ScanProject.TrajInfo
        """
        if ScanProject.Trajs.size!=0:
            StartTrajs     = np.int64(ScanProject.Trajs[1,:]) 
            EndTrajs       = np.int64(ScanProject.Trajs[2,:])            

            ScanTime       = ScanProject.ScanMatrix[0,:]
            J              = np.int32(ScanProject.ScanMatrix[25,:])
            MinTrajTimeLen = ScanProject.TrajectorySelectionRules['MinTrajectoryTimeLength']
            run_cycles     = np.int(ScanProject.PosCalibration['run_cycles']) 
            run_wndw       = np.int(ScanProject.PosCalibration['run_wndw'])
            xndw=run_cycles*run_wndw

            #Find Trim time of trajectories 
            StartTime = ScanTime[StartTrajs] + StartTrim
            EndTime = ScanTime[EndTrajs] - EndTrim
            validTrajs=np.where(EndTime-StartTime>0)


            #Remove Trajs that are negative or zero short
            StartTrajs=StartTrajs[validTrajs]
            StartTime=StartTime[validTrajs]
            EndTrajs =EndTrajs[validTrajs]
            EndTime  =EndTime[validTrajs]

            # Trim 
            for i in range(len(StartTrajs)):
                i_valid=np.where(np.logical_and(ScanTime>=StartTime[i], ScanTime<=EndTime[i]))[0]
                StartTrajs[i]=i_valid[0]
                EndTrajs[i]=i_valid[-1]

                # Check cycle behavior
                Jump=np.where(J[StartTrajs[i]:EndTrajs[i]]==1)[0]
                if (Jump.size!=0):
                    if (Jump[0]==0) or (Jump[0]==xndw):
                        pass #all good 
                    else:
                        print("Traj starts in midd cycle, adjusting start to previos cycle")
                        
                

            #Infoce max length
            if ForceMaxTrajLength<= MinTrajTimeLen and ForceMaxTrajLength>0:
                print("All trajectorise are removed since: ForceMaxTrajLength <= MinTrajTimeLen")
            
            if ForceMaxTrajLength>0:
                TrajTimeLength = EndTime - StartTime
                NminiTrajs=np.floor(TrajTimeLength/ForceMaxTrajLength)
                NminiTrajs[NminiTrajs>0]=NminiTrajs[NminiTrajs>0]+1
            
                NewStartTrajs=np.array([])
                NewEndTrajs=np.array([])
                for i in range(len(StartTrajs)): 
                    if NminiTrajs[i]!=0:
                        Split=np.int64(np.linspace(StartTrajs[i], EndTrajs[i], endpoint=False, num=int(NminiTrajs[i])))
                        SplitStarts=Split[0:].copy() 
                        SplitStarts[1:]=SplitStarts[1:]+1
                        SplitEnd=Split[1:].copy()
                        SplitEnd=np.concatenate((SplitEnd, [EndTrajs[i]]))
                        
                        NewStartTrajs=np.concatenate((NewStartTrajs,SplitStarts)) 
                        NewEndTrajs=np.concatenate((NewEndTrajs,SplitEnd))
                    else:
                        NewStartTrajs=np.concatenate((NewStartTrajs,[StartTrajs[i]])) 
                        NewEndTrajs=np.concatenate((NewEndTrajs,[EndTrajs[i]]))
                StartTrajs=NewStartTrajs
                EndTrajs  =NewEndTrajs  
            
            StartTrajs=np.int64(StartTrajs)
            EndTrajs  =np.int64(EndTrajs)
            TrajTimeLength= ScanTime[EndTrajs] - ScanTime[StartTrajs]
            ScanProject.Trajs, ScanProject.TrajInfo = self.WeedTrajs(ScanProject, StartTrajs, EndTrajs, TrajTimeLength) 

    def GetTrajs(self, ScanProject,Filter=1):
        """
        Returns Trajs (N x 3) numpy array which contains:
        0. Trajectory time length [s]
        1. Start index
        2. End index 
        where N is the number of trajectories that fit the selection criteria

        TrajInfo is a list that explains the contains the number of trajectories
        and labels the columns of Trajs
        """
        StartTraj, EndTraj, TrajTimeLength = self.FindTrajs(ScanProject,Filter=Filter)
        ScanProject.Trajs, ScanProject.TrajInfo = self.WeedTrajs(ScanProject, StartTraj, EndTraj, TrajTimeLength)

    def get_Gaussian_traj_estimation(self, ScanProject, wndw, beta, mode, Trajs, ScanMatrix, APDfilter, MeasurementOnTime):
        """
        Function to calculate trajectories from raw data.
        It supports two modes 
            FixWindowAv: fix window where the window moves in steps equal to the window size. With in each window a photon weighted centre of mass calculation is done for each dimension 
            RollingAv:   fix window where the window moves in one step to as in a moving average. With in each window a photon weighted centre of mass calculation is done for each dimension 
        
        INPUT:
            wndw:          Window size
            mode:          FixWindowAv, or RollingAv
            Trajs:         Matrix with [total time, start, end], start and end are refering to rows in ScanMatrix
            ScanMatrix:    Matrix containing all raw data 
            APDfilter:     Total Photon count [min, max]
            MeasurementOnTime: Time preiod between each laser shoot

        UOTPUT: 
            TrajPosEstimation[n]: for each trajectory there is a matrix defining that trajectory, the content of the matrix is: 
                        [TTRM, XXRM, YYRM, Pol, DPol, RMCAPD1, RMCAPD2, LaserFlag, WindowTime, AcumilatedMeasurementTime, ZZRM]
                TTRM: Time estimated acording to the mode  
                XXRM: x estimated acording to the mode  
                YYRM: y estimated acording to the mode  
                Pol:
                DPol:
                RMCAPD1:
                RMCAPD2:
                LaserFlag:
                WindowTimev:
                AcumilatedMeasurementTime:
                ZZRM:

        Note: Need to handle time jumps in when we have skip cycles
        """

        # Create a dictionary for the trajectories
        ScanProject.TrajPosEstimation={}

        wndw=int(wndw) #Need to force the window to int since this is used as an index 
        # Chose step size depending of the mode
        if mode =='FixWindowAv': 
            wndwStep=wndw
        elif mode =='RollingAv':
            wndwStep=1;
        else:
            print('Wrong Mode value used in getTrajAvraging, use: FixWindowAv or RollingAv')
    
        debug=False
        #Going through each trajectory in Traj 
        for n in range(Trajs.shape[1]):
            if debug:
                n=2

            interval=range(int(Trajs[1,n]), int(Trajs[2,n]+1))
            
            Z=ScanMatrix[0, interval] #[s] Time
            J=ScanMatrix[25, interval]# logic vector indicating where time jumps  
            X=ScanMatrix[1, interval] #[um] X value including all contributions piezo center + EOD center + EOD tracking pattern
            Y=ScanMatrix[2, interval] #[um] Y value including all contributions piezo center + EOD center + EOD tracking pattern
            
            ZI_OffsetIndex=ScanMatrix[19, interval]*ScanProject.PosCalibration['SLMSwitchHight'] #offset in um
            ZI_um=ScanMatrix[20, interval]*20/(2**14) #z raw data from FPGA same as 21
            ZZindex=ZI_um+ZI_OffsetIndex
            #ZZindex=ScanMatrix[22, interval]  #[um] z-value 
            MemIndex=ScanMatrix[8, interval]  #0-5 memory index  
            APD1 = ScanMatrix[3, interval] # Photon counts H
            APD2 = ScanMatrix[4, interval] # Photon counts V
            TotCounts=APD1+APD2
         
            LaserFlag=ScanMatrix[8, interval]
            
            W = (TotCounts >= APDfilter[0]) & (TotCounts <= APDfilter[1])
            TotCounts=TotCounts*W

            #Trajectory total time
            TrajTime=Z[-1]-Z[0]
            
            if wndw < len(Z):
                WindowTime=Z[wndw-1]-Z[0]
            else:
                print('Fewer elemets compared to requested time window')
                WindowTime=Z[-1]-Z(0)
  
            # Calculating averiage weighted by the photon counts
            # if wndwStep=1 then we have a rolling mean, 
            # if wndwStep=wndw then we have the mean ofver fix steps 
            # For x and y the calculation is x_c= sum x_i*p_i/(sum p_i) where the sum over i are over the window used, x_i is the position of the laser shoot and p_i is the amount of photons collected at that position.
            # One can interpreat P_i = p_i/(sum p_i) as the probability to be in x_i 
            # For z we take the staircase index and multiply by the hight, and then multiply by the staircase hight
            XXW=X*TotCounts
            YYW=Y*TotCounts
            ZZW= TotCounts*ZZindex



            #plt.hist(TotCounts[np.where(MemIndex>3)],100)
            #plt.show()

            #####
            with np.errstate(divide='ignore', invalid='ignore'):
                SAPD = np.convolve(np.ones(wndw),TotCounts)
                SAPD=SAPD[wndw-1:-wndw+1:wndwStep]
                #photon_wight = 1 - beta/SAPD
                #photon_wight[photon_wight<0]=0
                
                XXRM = np.convolve(np.ones(wndw),XXW)
                XXRM=XXRM[wndw-1:-wndw+1:wndwStep]/SAPD
 
                YYRM = np.convolve(np.ones(wndw),YYW)
                YYRM=YYRM[wndw-1:-wndw+1:wndwStep]/SAPD

                ZZRM = np.convolve(np.ones(wndw),ZZW)
                ZZRM=ZZRM[wndw-1:-wndw+1:wndwStep]/SAPD
               
                TTRM=Z[wndw-1::wndwStep]          
            
                #### APD signals
            RMCAPD1=np.convolve(np.ones(wndw),APD1)
            RMCAPD1=RMCAPD1[wndw-1:-wndw+1:wndwStep]

            RMCAPD2=np.convolve(np.ones(wndw),APD2)
            RMCAPD2=RMCAPD2[wndw-1:-wndw+1:wndwStep]
        
            A=np.zeros(wndw)
            A[0]=1
            A[-1]=-1
            WindowTime=np.convolve(A, Z)
            WindowTime=WindowTime[wndw-1:-wndw+1:wndwStep]
        
            AcumilatedMeasurementTime=np.convolve(np.ones(len(Z)), np.ones(wndw))[wndw-1:-wndw+1:wndwStep]  *MeasurementOnTime;
        
            with np.errstate(divide='ignore', invalid='ignore'):
                Pol=(RMCAPD1-RMCAPD2)/(RMCAPD1+RMCAPD2)
        
            ##Error assuming poisson statistics 
            DAPD_1=np.sqrt(RMCAPD1)
            DAPD_1[DAPD_1==0]=1

            DAPD_2=np.sqrt(RMCAPD2)
            DAPD_2[DAPD_2==0]=1;
            with np.errstate(divide='ignore', invalid='ignore'):
                DPDAPD_1=2*RMCAPD2 / ((RMCAPD1+RMCAPD2)**2)
                DPDAPD_2=-2*RMCAPD1/ ((RMCAPD1+RMCAPD2)**2)
  
            DPol=np.sqrt((DAPD_1*DPDAPD_1)**2 + (DAPD_2*DPDAPD_2)**2)
  
            ## Laser LaserFlag
            LaserFlag=LaserFlag[wndw-1::wndwStep]

            ######### DeBugging 
            if debug:
                import matplotlib.pyplot as plt
                offset=0
                interval=range(int(Trajs[1,n])+offset, int(Trajs[2,n]+1)+offset)
                ZIn=ScanMatrix[19, interval]*TotCounts
                with np.errstate(divide='ignore', invalid='ignore'):
                    ZInT = np.convolve(np.ones(wndw),ZIn)
                    ZInT=ZInT[wndw-1:-wndw+1:wndwStep]/SAPD
            
                #plt.plot(Z, ScanMatrix[22, interval],'b')
                plt.plot(Z, ScanMatrix[20, interval],'r')
                plt.plot(Z, 2+(ScanMatrix[21, interval]/1000),'g')
                plt.plot(TTRM, ZZRM,'b')
                plt.grid()
                #plt.plot(TTRM, ZInT)
                #plt.show()
                #plt.close()
            
            
            ScanProject.TrajPosEstimation[n]=np.array([TTRM, XXRM, YYRM, Pol, DPol, RMCAPD1, RMCAPD2, LaserFlag, WindowTime, AcumilatedMeasurementTime, ZZRM])
            
    def get_traj_raw_data(self, Trajs, TrajToGet, ScanMatrix):
        """
        Function to extract the raw data of a trajectory
        
        INPUT:
            Trajs:         Matrix with [total time, start, end], start and end are refering to rows in ScanMatrix    
            TrajToGet:     which trajectory to get
            ScanMatrix:    Matrix containing all raw data 

        OUTPUT:       
                T: Time estimated acording to the mode  
                X:      x [um] data
                Y:      y [um] data
                Z:      z [um] data 
                APD1:   Photon counts H
                APD2:   Photon counts V
        """

        n=TrajToGet
        
        interval=range(int(Trajs[1,n]), int(Trajs[2,n]+1))
            
        T=ScanMatrix[0, interval]      #[s] Time
        X=ScanMatrix[1, interval]      #[um] X value including all contributions piezo center + EOD center + EOD tracking pattern
        Y=ScanMatrix[2, interval]      #[um] Y value including all contributions piezo center + EOD center + EOD tracking pattern
        Z=ScanMatrix[22, interval]     #[um] z-value 
        J=ScanMatrix[25, interval]     #Time Jump indicator 
        APD1 = ScanMatrix[3, interval] #Photon counts H
        APD2 = ScanMatrix[4, interval] #Photon counts V
        
        return T,X,Y,Z,APD1,APD2,J

    def get_traj_raw_data2(Trajs, TrajToGet, ScanMatrix, ScanProject):
        """
        Function to extract the raw data of a trajectory
        
        INPUT:
            Trajs:         Matrix with [total time, start, end], start and end are refering to rows in ScanMatrix    
            TrajToGet:     which trajectory to get
            ScanMatrix:    Matrix containing all raw data 

        OUTPUT:       
                T: Time estimated acording to the mode  
                XP: [um] X value piezo center 
                YP: [um] Y value piezo center 
                XE: [um] X value EOD center 
                YE: [um] Y value EOD center 
                XO: [um] X value EOD tracking pattern
                YO: [um] Y value EOD tracking pattern
                ZI: [index] z-value 
                APD1: Photon counts H
                APD2: Photon counts V
                J:  Time Jump indicator 
                MemIndex: Memory index
        """

        n=TrajToGet
        
        interval=range(int(Trajs[1,n]), int(Trajs[2,n]+1))
            
        T=ScanMatrix[0, interval]        #[s] Time
        XP=ScanMatrix[9, interval]       #[um] X value piezo center 
        YP=ScanMatrix[10, interval]      #[um] Y value piezo center 
        XE=ScanMatrix[23, interval]      #[um] X value EOD center 
        YE=ScanMatrix[24, interval]      #[um] Y value EOD center
        XO=ScanMatrix[11, interval]      #[um] X value EOD tracking pattern
        YO=ScanMatrix[12, interval]      #[um] Y value EOD tracking pattern
        
        ZI=np.round(ScanMatrix[22, interval]/ScanProject.PosCalibration['SLMSwitchHight']).astype('int')     #[index] z-value 
        J=ScanMatrix[25, interval]     #Time Jump indicator 
        APD1 = ScanMatrix[3, interval] #Photon counts H
        APD2 = ScanMatrix[4, interval] #Photon counts V
        MemIndex= ScanMatrix[8, interval] #Memory index
        
        return T,XP,YP,XE,YE,XO,YO,ZI,APD1,APD2,J,MemIndex

    def get_traj_Poisson_ProbMap_AnaPSF(self, ScanProject, TrajToGet, MaxTrajTime=-1, verbose=1):
        """
        Generates a Poisson likelihood over x,y 
        Out put: 
            xyM = list of x,y values over where likelihood is max
            LT  = likelihood as a function of x,y given by the grid outputs
            xx  = x coordinate grid for LT
            yy  = y coordinate grid for LT
            verbose = boolian, showe iterations 
        """
        if verbose:
            print("Evaluating trajectory #: " + str(TrajToGet))
        
        if not hasattr(ScanProject,'PoissonProbXYZMap'):
            ScanProject.PoissonProbXYZMap={}
        
        #Def parameters (Need to be moved out)
        sx=ScanProject.PFS_Parameters['sigma'][0]
        sy=ScanProject.PFS_Parameters['sigma'][1]
        sz=ScanProject.PFS_Parameters['sigma'][2]
        m =int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use
        lx=ly=lz=0.05                                     #[um] half length of the x,y plan to use, uesd if -lx to lx and -ly to ly 
        gsx=gsy=gsz=10                                    #amount of points in each dimension for the total length 
        

        ######## Get data
        ScanMatrix=ScanProject.ScanMatrix
        Trajs=ScanProject.Trajs
        run_cycles     = np.int(ScanProject.PosCalibration['run_cycles']) 
        run_wndw       = np.int(ScanProject.PosCalibration['run_wndw'])
        xndw=run_cycles*run_wndw
        #time [s], x_i Position [um], y_i Position [um], [um] z-value
        t_raw, x_EOD_off, y_EOD_off, Z_level,APD1,APD2, J=self.get_traj_raw_data(Trajs, TrajToGet, ScanMatrix)
        n=APD1+APD2
                
        #Define variabale to fill 
        NT=int(len(x_EOD_off)/m)
        LT=np.zeros((gsx,gsy,gsz,NT))
        xx=np.zeros((gsx,gsy,gsz,NT))
        yy=np.zeros((gsx,gsy,gsz,NT))
        zz=np.zeros((gsz,gsz,gsz,NT))
        txyznM=np.zeros((5,NT))
        xyzMind=np.int32(np.zeros((3,NT)))
        
        #Run itaration
        for i in range(NT):
            if verbose:
                print(str(i) + " of " + str(int(len(x_EOD_off)/m)))
            
            #Points to be used in evaluation
            xi = x_EOD_off[i*m:(i+1)*m]
            yi = y_EOD_off[i*m:(i+1)*m]
            zi = Z_level[i*m:(i+1)*m]
            ni=n[i*m:(i+1)*m]
            ti=t_raw[i*m:(i+1)*m]
            Ji=np.int32(J[i*m:(i+1)*m])
            Jump=np.where(Ji==1)[0]
            if (Jump.size!=0): 
                if (Jump[0]==0) or (Jump[0]==xndw):
                    pass
                else:
                    print("Varning: Possible estimation over a time jumpe")
                
            if i ==0:
                x0=np.mean(xi)
                y0=np.mean(yi)
                z0=np.mean(zi)
            else: 
                x0=txyznM[1,i-1]
                y0=txyznM[2,i-1]
                z0=txyznM[3,i-1]

            t=np.mean(ti)
            ntot=np.sum(ni)

            #Evaluate log likelihood and then likelihood                      
            LogLT,xx[:,:,:,i] ,yy[:,:,:,i] ,zz[:,:,:,i]=PSF.Cal_LogL_Poisson(xi,yi,zi,ni,sx,sy,sz,x0,y0,z0,gsx,gsy,gsz,lx,ly,lz)
            LT[:,:,:,i]=np.exp(LogLT)

            #Get the maximum value over the area
            xyzMind[:,i]=np.unravel_index(np.argmax(LT[:,:,:,i], axis=None), LT[:,:,:,i].shape)
            xtemp = xx[xyzMind[:,i][0],xyzMind[:,i][1],xyzMind[:,i][2],i]
            ytemp = yy[xyzMind[:,i][0],xyzMind[:,i][1],xyzMind[:,i][2],i]
            ztemp = zz[xyzMind[:,i][0],xyzMind[:,i][1],xyzMind[:,i][2],i]

            txyznM[:,i]=[t, xtemp, ytemp, ztemp, ntot]
            
            ### Debug plot 
            if 0:    
                #plt.contourf(LT[:,:,i],cmap=cm.gray_r) #cm.coolwarm, 
                #plt.show()
                fig = plt.figure()
                ax = fig.add_subplot(111)    
                cf=ax.contourf(xx[:,:,i], yy[:,:,i], LT[:,:,i],cmap=cm.gray_r) #cm.coolwarm, 
                ax.plot(txyznM[1,:],txyznM[2,:],'-*b')
                ax.plot(txyznM[1,i],txyznM[2,i],'*r')
                ax.set_title('Min/Max=' + str("{:.2e}".format(np.min(LT[:,:,i]))) + '/' + str("{:.2e}".format(np.max(LT[:,:,i]))) )
                fig.colorbar(cf, ax=ax)
                plt.show()
                print("Done: " + str(i))

        #ScanProject.PoissonProbXYZMap[TrajToGet]=(txyznM, LT, xx, yy, zz)     
        ScanProject.PoissonProbXYZMap[TrajToGet]=txyznM     

    def get_traj_Poisson_ProbMap_NumPFS_old(self, ScanProject, TrajToGet, MaxTrajTime=-1, verbose=1):
        """
        Generates a Poisson likelihood over x,y using Numerical PFS from the Phase retrival module
        Out put: 
            xyM = list of x,y values over where likelihood is max
            LT  = likelihood as a function of x,y given by the grid outputs
            xx  = x coordinate grid for LT
            yy  = y coordinate grid for LT
            verbose = boolian, showe iterations 
        """
        if verbose:
            print("Evaluating trajectory #: " + str(TrajToGet))
        
        if not hasattr(ScanProject,'PoissonProbXYZMap_NumPFS'):
            ScanProject.PoissonProbXYZMap_NumPFS={}

        now1 = time.time()
        ######## Get data
        ScanMatrix=ScanProject.ScanMatrix
        Trajs=ScanProject.Trajs
        run_cycles     = np.int(ScanProject.PosCalibration['run_cycles']) 
        run_wndw       = np.int(ScanProject.PosCalibration['run_wndw'])
        xndw=run_cycles*run_wndw
        m =int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use

        T,XP,YP,XE,YE,XO,YO,ZI,APD1,APD2,J,_ = self.get_traj_raw_data2(Trajs, TrajToGet, ScanMatrix,ScanProject)
        n=APD1+APD2

        #Get Numerical PSF light field, to get intensity take |LF|^2
        path=ScanProject.Ph_Retrieval['path_PSF']
        fname=ScanProject.Ph_Retrieval['fname_PSF']
        LF, parm_PR= PSF.LoadMat(path,fname)

        # Add zx,y resolution during corping to minimiz the data load
        Dz=ScanProject.Ph_Retrieval['zssize']/ScanProject.Ph_Retrieval['ZSWScalefactor'] #[m]
        IP=np.zeros((LF.shape[0],LF.shape[1],LF.shape[2]*ScanProject.Ph_Retrieval['ZSWScalefactor']))
        for kk in range(LF.shape[2]):
            for kkk in range(ScanProject.Ph_Retrieval['ZSWScalefactor']):
                Dxyz=[0,0,kkk*Dz] #[m]
                _, LF_N, _,_, _, _, _, _, _, _ = PSF.LightPropagator(LF[:,:,kk],parm_PR,Dxyz)
                IP[:,:,kk*10+kkk]=np.absolute(LF_N)**2                

        # Rescale the image to get higher pixel density and get the xy grid
        parm={}
        parm['Px']    = ScanProject.SetupSpecifics['Px']            
        parm['Mag']   = ScanProject.SetupSpecifics['Mag']           
        parm['Mag2']  = ScanProject.SetupSpecifics['Mag2']          
        parm['NM']    = LF.shape            
        parm['SWS']   = ScanProject.Ph_Retrieval['XYSWScalefactor'] # Software scaling factor
        parm['zssize']= ScanProject.Ph_Retrieval['zssize']        #[m] z- step size between images

        #IP=np.absolute(LF)**2 # Get the intensity 
        #IP=cv2.resize(IP, (IP.shape[0]*parm['SWS'],IP.shape[1]*parm['SWS']), interpolation = cv2.INTER_LANCZOS4) # cv2.INTER_NEAREST, cv2.INTER_AREA, cv2.INTER_LINEAR, cv2.INTER_CUBIC, cv2.INTER_LANCZOS4
        xxIP, yyIP = PSF.getXYGrid(parm) #[um]
        
        #Get pinhole size in [um] 
        PholeDia = ScanProject.SetupSpecifics['PholeDia']/ScanProject.SetupSpecifics['Mag'] #size of confocal pinhole in the sample plan

        #Offset for each levles [x0,y0],[x1,y1],[x2,y2],[x2,y2],... in [um] 
        # and is suppose to be the PSF max intencity in the center of the pinhole
        PinholeOffset=ScanProject.Ph_Retrieval['IntencityC'][:,0:2]*parm['Px']*1e6/(parm['Mag']*parm['Mag2']) #[um]
        zInFocusIndex=ScanProject.Ph_Retrieval['IntencityC'][:,2]

        if 0: #Debug Plot 
            fig, ax= plt.subplots(3,3)
            fig.set_size_inches(8, 8, forward=True)
            leveltosee=2
            for k in range(9):
                zindex=ScanProject.Ph_Retrieval['IntencityC'][leveltosee,2]
                if 0: # See in pixel values 
                    ax[np.unravel_index(k,ax.shape)].contourf(IP[:,:,k],cmap=cm.coolwarm,alpha=1.0) #cm.gray, cm.coolwarm,
                    ax[np.unravel_index(k,ax.shape)].plot(parm['SWS']*ScanProject.Ph_Retrieval['IntencityC'][leveltosee,0],parm['SWS']*ScanProject.Ph_Retrieval['IntencityC'][leveltosee,1],'r*')
                    ax[np.unravel_index(k,ax.shape)].set_title("k= " + str(k) + '/' + str(zindex))
                    ax[np.unravel_index(k,ax.shape)].ticklabel_format(style='sci',scilimits=(0,0),axis='both')
                    ax[np.unravel_index(k,ax.shape)].set_ylabel("[px]")
                    ax[np.unravel_index(k,ax.shape)].set_xlabel("[px]")
                if 1: # See in um values 
                    ax[np.unravel_index(k,ax.shape)].contourf(xxIP,yyIP,IP[:,:,k],cmap=cm.coolwarm,alpha=1.0) #cm.gray, cm.coolwarm,
                    ax[np.unravel_index(k,ax.shape)].plot(PinholeOffset[leveltosee,0],PinholeOffset[leveltosee,1],'r*')
                    ax[np.unravel_index(k,ax.shape)].set_title("k= " + str(k) + '/' + str(zindex))
                    ax[np.unravel_index(k,ax.shape)].ticklabel_format(style='sci',scilimits=(0,0),axis='both')
                    ax[np.unravel_index(k,ax.shape)].set_ylabel("[um]")
                    ax[np.unravel_index(k,ax.shape)].set_xlabel("[um]")
            plt.show()   
                
        now2 = time.time()
        print('Time 1 is: ' + str(now2-now1) )
        #Run itaration
        NT=int(len(T)/m) #Amount of evaluations over the trajectory
        txyznM=np.zeros((5,NT))
        for i in range(NT):            
            if verbose:
                print(str(i) + " of " + str(NT))

            now3 = time.time()
            ni=n[i*m:(i+1)*m]
            ti=T[i*m:(i+1)*m]
            Ji=np.int32(J[i*m:(i+1)*m])
            Jump=np.where(Ji==1)[0]

            t=np.mean(ti)
            ntot=np.sum(ni)

            if (Jump.size!=0): 
                if (Jump[0]==0) or (Jump[0]==xndw):
                    pass
                else:
                    print("Varning: Possible estimation over a time jumpe")
            
            zi=ZI[i*m:(i+1)*m] 
            xiP=XP[i*m:(i+1)*m]
            yiP=YP[i*m:(i+1)*m]

            xx=np.moveaxis(np.tile(xxIP,(m,1,1)),0,-1)
            yy=np.moveaxis(np.tile(yyIP,(m,1,1)),0,-1)
            #Place beam to use (depends on zindex) at the origo
            xx=xx - np.broadcast_to(PinholeOffset[(zi).astype('int'),0],(xxIP.shape[0],xxIP.shape[1],m))
            yy=yy - np.broadcast_to(PinholeOffset[(zi).astype('int'),1],(xxIP.shape[0],xxIP.shape[1],m)) 
        
            # Move origo to the piezo + EOD postions 
            # This is done in two step since we like to make a pinhole mask at the piezo position
            # 1. add only the EOD part 
            # 2. Make the mask around the origo which is the current piezo position  
            # 3. Move to the ture piezo position
            # 4. Make a cut out mask around the fist pinhole position, this defines the area that is the same for all frames 
            
                #Do point 1.
            xx = xx + np.broadcast_to((-XE+XO)[i*m:(i+1)*m],(xxIP.shape[0],xxIP.shape[1],m))
            yy = yy + np.broadcast_to((-YE+YO)[i*m:(i+1)*m],(xxIP.shape[0],xxIP.shape[1],m))
                #Do point 2.
            PMask=np.ones((xx.shape))              
            PMask[np.where(xx**2 + yy**2 >(PholeDia/2)**2)]=np.nan          
                #Do point 3.
            xx = xx + np.broadcast_to(xiP,(xxIP.shape[0],xxIP.shape[1],m))
            yy = yy + np.broadcast_to(yiP,(xxIP.shape[0],xxIP.shape[1],m))
                #Do point 4. We will state the problem assuming that the first pinhole i the center and all other will be displesed to it 
                # Fitst we find the pixel coordinates closest to the point of the piezo stage / first pinehole
                # these are in "Carg"
                # We then cut out for xx,yy,PMask and the PSF around these coordiantes
                # We will then assume that all PFS images are aligned to each other upt to one pixel width
                # Thus we can sum all PFS and they will represent the pattern of the shots, (we do keep the true positions in xxc,yyc if needed).
                # Thus our sensing matrix is now S which is the resut of the cut out
            carg=np.zeros(m)  
            for k in range(m): 
                carg[k]=np.argmin( (xx[:,:,k]-xiP[0])**2 + (yy[:,:,k]-yiP[0])**2 )
                if 0: #Debug plot
                    print(np.unravel_index(carg[k].astype('int'), xx[:,:,0].shape))
                    plt.contourf((xx[:,:,k]-xiP[0])**2 + (yy[:,:,k]-yiP[0])**2)
                    plt.contourf(PMask[:,:,k]*10,alpha=0.5)
                    plt.plot(np.unravel_index(carg[0:k+1].astype('int'), xx[:,:,0].shape)[1],np.unravel_index(carg[0:k+1].astype('int'), xx[:,:,0].shape)[0],'r*')
                    plt.show()
            Carg=np.array(np.unravel_index(carg.astype('int'), xx[:,:,0].shape)) # index coordinates 
            now4 = time.time()
            print('Time 2 is: ' + str(now4-now3) )

            PS=ScanProject.Ph_Retrieval['CropS']*ScanProject.Ph_Retrieval['XYSWScalefactor'] #amount of pixles to use in the cutout  
            PMaskc=np.zeros((2*PS,2*PS,m))  
            S=np.zeros((2*PS,2*PS,LF.shape[2]*ScanProject.Ph_Retrieval['ZSWScalefactor'],m))  
            for k in range(m): 
                PMaskc[:,:,k]=PMask[Carg[0,k]-PS:Carg[0,k]+PS,Carg[1,k]-PS:Carg[1,k]+PS,k]
                # Add zx,y resolution during corping to minimiz the data load
                #with concurrent.futures.ProcessPoolExecutor() as executor:
                #    for   res  in executor.map(cv2.resize(IP[:,:,kk],(IP.shape[0]*parm['SWS'],IP.shape[1]*parm['SWS']),interpolation = cv2.INTER_LANCZOS4), range(IP.shape[2])):
                #        S[:,:,kk,k]=res[Carg[0,k]-PS:Carg[0,k]+PS,Carg[1,k]-PS:Carg[1,k]+PS]

                for kk in range(IP.shape[2]):
                    S[:,:,kk,k]= cv2.resize(IP[:,:,kk],
                                                  (IP.shape[0]*parm['SWS'],IP.shape[1]*parm['SWS']), 
                                                  interpolation = cv2.INTER_LANCZOS4)[Carg[0,k]-PS:Carg[0,k]+PS,Carg[1,k]-PS:Carg[1,k]+PS]
                if 0:#Debug plot 
                    plt.contourf(S[:,:,int(zInFocusIndex[zi[k]]),k],alpha=1.0)
                    plt.contourf(PMaskc[:,:,k]*10,alpha=0.5)
                    plt.show()
            
            now5 = time.time()
            print('Time 3 is: ' + str(now5-now4) )

            #Create photon map and pinhole mask of the same size as S and restate S as the full sensing matrix 
            xxE,yyE,zzE=np.meshgrid(xx[0,Carg[1,0]-PS:Carg[1,0]+PS,0],
                                    yy[Carg[0,0]-PS:Carg[0,0]+PS,0,0],
                                    np.array(range(S.shape[2]))*ScanProject.Ph_Retrieval['zssize']*1e6/ScanProject.Ph_Retrieval['ZSWScalefactor']-zInFocusIndex[0]*ScanProject.Ph_Retrieval['zssize']*1e6)
            nn=np.broadcast_to(ni,(2*PS,2*PS,S.shape[2],m))
            PMaskc=np.moveaxis(np.tile(PMaskc,(S.shape[2],1,1,1)),0,-2)
            S = PMaskc*S
            
            #Construct prior based on the gaussain estimate
            txyzG=[ScanProject.TrajPosEstimation[TrajToGet][0,i*m], ScanProject.TrajPosEstimation[TrajToGet][1,i*m], ScanProject.TrajPosEstimation[TrajToGet][2,i*m], ScanProject.TrajPosEstimation[TrajToGet][-1,i*m]]
            XYPsigma=0.1
            ZPsigma=0.2
            XYZprior=-((xxE-txyzG[1])**2+(yyE-txyzG[2])**2)/(2*XYPsigma**2) - (zzE-txyzG[3])**2/(2*ZPsigma**2)
            
            #Construct the log likelihood 
            now6 = time.time()
            print('Time 4 is: ' + str(now6-now5) )

            G1=np.sum(nn*np.log(S), axis=3) #might be posible to parallelize the log part 

            now7 = time.time()
            print('Time 5 is: ' + str(now7-now6) )

            G2=np.log(np.sum(S, axis=3))*np.sum(ni)

            now8 = time.time()
            print('Time 6 is: ' + str(now8-now7) )

            G=(G1-G2)+XYZprior

            now9 = time.time()
            print('Time 7 is: ' + str(now9-now8) )
            
            MG=np.max(np.ma.array(G, mask=np.isnan(G))) 
            indextosee=np.array(np.unravel_index(np.argmax(np.ma.array(G, mask=np.isnan(G))) ,G.shape))
            txyznM[:,i]=[t, xxE[indextosee[0],indextosee[1],indextosee[2]],yyE[indextosee[0],indextosee[1],indextosee[2]], zzE[indextosee[0],indextosee[1],indextosee[2]], ntot]

            if 0:
                fig, ax= plt.subplots(10,3)
                fig.set_size_inches(12, 4, forward=True)
                for kk in range(10):
                    #ax[kk,0].contourf(xxc[:,:,0],yyc[:,:,0],XYprior[:,:,kk],alpha=1.0) #indextosee[2]
                    ax[kk,0].contourf(xxc[:,:,0],yyc[:,:,0],G1[:,:,kk],alpha=1.0) #indextosee[2]
                    ax[kk,1].contourf(xxc[:,:,0],yyc[:,:,0],G2[:,:,kk],alpha=1.0)
                    ax[kk,1].set_title("M=" + str(np.around(MG, decimals=4)) + " z-fond=" + str(indextosee[2]))
                    ax[kk,2].contourf(xxc[:,:,0],yyc[:,:,0],np.exp(G[:,:,kk]),alpha=1.0)
                    ax[kk,2].plot(txyznM[1,i],txyznM[2,i],'r*')
                    ax[kk,2].plot(txyzG[1],txyzG[2],'b*')
                    ax[kk,2].set_title("M=" + str(np.around(np.max(G[:,:,kk]), decimals=4)) + " z =" + str(kk))
                    for k in range(3):
                        ax[kk,k].ticklabel_format(style='sci',scilimits=(0,0),axis='both')
                        ax[kk,k].set_ylabel("[um]")
                        ax[kk,k].set_xlabel("[um]")
                plt.show()

            if 0:
                #3D density plot
                validData=np.where(np.logical_not(np.isnan(G)))
                xyz_pos=np.transpose(np.vstack((xxE[validData], yyE[validData], zzE[validData])))
                rgba_colors=cm.jet(G[validData]/np.max(G[validData]))

                GP=np.exp(G[validData])
                rgba_colors[:,3]=(GP - np.min(GP))/np.max(GP)
                QTplt.DensityPlot(xyz_pos,rgba_colors)
                
            if 0:
                fig, ax= plt.subplots(1,1)
                fig.set_size_inches(5, 5, forward=True)
                ax.contourf(xxc[:,:,0],yyc[:,:,0],np.sum(np.exp(G),axis=2),alpha=1.0)
                ax.plot(txyznM[1,i],txyznM[2,i],'r*')
                ax.plot(txyzG[1],txyzG[2],'b*')
                ax.ticklabel_format(style='sci',scilimits=(0,0),axis='both')
                ax.set_ylabel("[um]")
                ax.set_xlabel("[um]")
                plt.show()
            
            print(str(indextosee) + '  ' + str(txyznM[:,i]))
            ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]=txyznM[:,i]
            now10 = time.time()
            print('Time 8 is: ' + str(now10-now9) )

    def get_traj_Poisson_ProbMap_NumPFS(self, ScanProject, TrajToGet, verbose=1, do_onlinesave=0, SaveSmatrix=''):
        """
        Generates a Poisson likelihood over x,y using Numerical PFS
        Input:
            ScanProject         All raw data possible to be used in the estimation  
            TrajToGet           Identifier for which data to be used 
            verbose=1           Showing steps of estimation in stdout
            do_onlinesave=0     save to mat file on each iteration 
            SaveSmatrix=''      save the S matrix for each iteration if the string is not '', 
                                if the string is 'something' it will be appended to the name of the saved file 
        Output: 
            xyM = list of x,y values over where likelihood is max
            LT  = likelihood as a function of x,y given by the grid outputs
            xx  = x coordinate grid for LT
            yy  = y coordinate grid for LT
            verbose = boolian, showe iterations 
        """
        if verbose:
            print("Evaluating trajectory #: " + str(TrajToGet))
        
        if not hasattr(ScanProject,'PoissonProbXYZMap_NumPFS'):
            ScanProject.PoissonProbXYZMap_NumPFS={}

        if ScanProject.SetupSpecifics['is_simulation']:
            file='GroundTruth.mat'
            folder=ScanProject.Files['AnalysisConfig'][:-18]
            GroundTruth=io.loadmat(folder+file)
            GroundTruth=GroundTruth[ScanProject.Sim_par['TrajType']]
        else:
            GroundTruth=np.nan

        ######## Get data
        ScanMatrix=ScanProject.ScanMatrix
        Trajs=ScanProject.Trajs
        run_cycles     = np.int(ScanProject.PosCalibration['run_cycles']) 
        run_wndw       = np.int(ScanProject.PosCalibration['run_wndw'])
        xndw=run_cycles*run_wndw
        m = int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use
        MemSize=ScanProject.PosCalibration['TrackingMemsize']   #Amount of points that is a pattern 

        T,XP,YP,XE,YE,XO,YO,ZI,APD1,APD2,J, MemIndex, ZI_um, ZI_OffsetIndex = _mis.get_traj_raw_data3(Trajs, TrajToGet, ScanMatrix,ScanProject)
        n=APD1+APD2
        #override old ZI to the new ZI_OffsetIndex
        ZI=ZI_OffsetIndex

            #Get Numerical PSF light field, to get intensity take |LF|^2
        path=ScanProject.Ph_Retrieval['path_PSF']
        fname=ScanProject.Ph_Retrieval['fname_PSF']
        LF, parm_PR= PSF.LoadMat(path,fname)
            
            #bundle some parameters
        parm={}
        parm['Px']    = ScanProject.SetupSpecifics['Px']            
        parm['Mag']   = ScanProject.SetupSpecifics['Mag']           
        parm['Mag2']  = ScanProject.SetupSpecifics['Mag2']          
        parm['NM']    = LF.shape            
        parm['SWS']   = ScanProject.Ph_Retrieval['XYSWScalefactor'] # Software scaling factor
        parm['zssize']= ScanProject.Ph_Retrieval['zssize']        #[m] z- step size between images in PSF stack
        parm['SWSZ']   = ScanProject.Ph_Retrieval['ZSWScalefactor'] # Software scaling factor for z

            #Pixel convertion factors  
        px2um=parm['Px']*1e6/(parm['Mag']*parm['Mag2'])  # from camera pixel to um
        spx2um=px2um/parm['SWS']                         # from super resolution to um 

        # Add z resolution            
        IP=PSF.zInterpolator(LF, parm['SWSZ'],parm_PR,p='WeightedTwoWay')
        # Add background determined by the signal to noise ratio 
        Background=ScanProject.TrajectorySelectionRules['background']
        IP=IP+Background
        
        #Get pinhole size in [um] 
        PholeDia = ScanProject.SetupSpecifics['PholeDia']/ScanProject.SetupSpecifics['Mag'] #size of confocal pinhole in the sample plan

        #Offset for each levles [x0,y0],[x1,y1],[x2,y2],[x2,y2],... in [um] 
        # and is suppose to be the PSF max intencity in the center of the pinhole
        PinholeOffset=ScanProject.Ph_Retrieval['IntencityC'][:,0:2]*px2um #[um]
        zInFocusIndex=ScanProject.Ph_Retrieval['IntencityC'][:,2]*parm['SWSZ']

        if 0: #Debug photon count plot  
            wndw=24
            wndwS=wndw #if 1 it is a moving average if equal to wndw it is like a fix window average
            Elements=np.unique(MemIndex)
            color=cm.rainbow(np.linspace(0,1,len(Elements)))
            legendString=[]
            for MemUnique,c in zip(Elements,color):
                OfInterest=np.where(MemIndex==MemUnique)
                TT=T[OfInterest]
                nT=n[OfInterest]
                TAV = np.convolve(np.ones(wndw),TT)
                TAV=TAV[wndw-1:-wndw+1:wndwS]/np.sum(np.ones(wndw))
                nAV = np.convolve(np.ones(wndw),nT)
                nAV=nAV[wndw-1:-wndw+1:wndwS]
                plt.plot(TAV,((-1)**MemUnique)*nAV,'-',c=c) 
                legendString.append('M' + str(int(MemUnique)) + ' ZI' + str(ZI[OfInterest][0]) )
            plt.xlabel('Time [ms]')
            plt.ylabel('Counts over ' + str(wndw) + ' measurements')
            plt.legend(tuple(legendString))
            plt.grid()
            plt.show()            

        #Run itaration
        NT=int(len(T)/m) #Amount of evaluations over the trajectory
        txyznM=np.zeros((5,NT))
        for i in range(NT):   
            starttime=time.time()
            if verbose:
                print(str(i) + " of " + str(NT))

            ni=n[i*m:(i+1)*m]
            ti=T[i*m:(i+1)*m]
            Ji=np.int32(J[i*m:(i+1)*m])
            Jump=np.where(Ji==1)[0]

            t=ti[-1] #time is set as the last slot, this is to conform to the fast gaussian calculation
            ntot=np.sum(ni)

            if (Jump.size!=0): 
                if (Jump[0]==0) or (Jump[0]==xndw):
                    pass
                else:
                    print("Varning: Possible estimation over a time jumpe")

            zi =ZI[i*m:(i+1)*m]  
            xiP=XP[i*m:(i+1)*m]; yiP=YP[i*m:(i+1)*m]; ziP=ZI_um[i*m:(i+1)*m]
            xiE=XE[i*m:(i+1)*m]; yiE=YE[i*m:(i+1)*m]
            xiO=XO[i*m:(i+1)*m]; yiO=YO[i*m:(i+1)*m]
            xPinOffset=PinholeOffset[(zi).astype('int'),0]; yPinOffset=PinholeOffset[(zi).astype('int'),1]
            
            SaveSmatrix_i=SaveSmatrix
            if SaveSmatrix!='':
                SaveSmatrix_i=SaveSmatrix + '_' + str(i)
            
            N1=time.time()
            S,xx,yy,zz = PSF.get_Smatrix2(ScanProject,IP,xiP,yiP,ziP,xiE,yiE,xiO,yiO,xPinOffset,yPinOffset, SaveSmatrix_i)
            print('Smatrix time: ' + str(time.time()-N1))

            
            #Get Loglikelihood 
            env="Poisson"
            if env=="Poisson":
                # Create prior start estimation 
                Myxz=np.mean(self.get_SmaxArg(S, xx, yy, zz), axis=0)
                #xyzCenterGuess=np.array([xiP[0],yiP[0],0.36])
                if i==0:
                    xyzCenterGuessG=np.array([-100,-100,-100])
                else:
                    if i==1:
                        xyzCenterGuessG=txyznM[1:,i-1]
                    else:
                        RValue=1#0.6
                        xyzCenterGuessG=(1-RValue)*xyzCenterGuessG + RValue*txyznM[1:,i-1]
                try:
                    zoffset=ScanProject.TrajectorySelectionRules['TopHatoffset'][2]
                except:
                    zoffset=ScanProject.TrajectorySelectionRules['TopHatoffset']
                xyzCenterGuess=np.array([Myxz[1],Myxz[0], zoffset+Myxz[2]])
                LogPTot, _, TopHatPrior, _, _, _= PSF.get_PoissonLoglikelihood_V2(ScanProject,S,xx,yy,zz,ni,xyzCenterGuess,xyzCenterGuessG)
            elif env=="Binomial":
                LogPTot = PSF.get_BinomialLoglikelihood(ScanProject,S,xx,yy,zz,ni,xyzCenterGuess)

            indextosee=np.unravel_index(np.argmax(LogPTot) ,LogPTot.shape)
            txyznM[:,i]=[t, xx[indextosee],yy[indextosee], zz[indextosee], ntot]
            
            if 0: #Debug plot find best weights
                WTemp=ScanProject.TrajectorySelectionRules['Wprio']
                ScanProject.TrajectorySelectionRules['Wprio']=[1,1,1,1]
                _, LogP, TopHatPrior, GPrior, LPPrior, DPPrior= PSF.get_PoissonLoglikelihood_V2(ScanProject,S,xx,yy,zz,ni,xyzCenterGuess)
                ScanProject.TrajectorySelectionRules['Wprio']=WTemp

                Wprio=[0,0,0,0.01]
                #Ground truth 
                xyzpix=(GroundTruth[:m*i+1:m,:]*1e6)[-1,:] #[um] using the last point to show

                #Optimizations 
                for k in range(50):
                    Wprio[3]= k/200
                    LogPTotW= LogP + Wprio[0]*TopHatPrior + Wprio[1]*GPrior + Wprio[2]*LPPrior + Wprio[3]*DPPrior
                    LogPTotW=LogPTotW-np.max(LogPTotW)
                    indextosee=np.unravel_index(np.argmax(LogPTotW) ,LogPTotW.shape)
                    PTMxyz=[xx[indextosee],yy[indextosee], zz[indextosee]]                

                    #Distance 
                    DMLE=np.sqrt(np.sum((xyzpix-PTMxyz)**2))*1e3 #nm
                    print("Opt " + "a=" + str(Wprio[3]) + " D=" + str(DMLE) )
                
                print('Done!')

            if 0: #Debug plot
                _TP.plotProbMapPointCloud(LogPTot,xx,yy,zz)
            if 0: #Debug plot
                _TP.plotProbMapProjections(ScanProject,LogPTot,xx,yy,zz,txyznM,txyzG,i,do_save_plot=1)
            if 0: #Debug plot
                if 1: #np.mod(i,10)==0:
                    _TP.plotProbMap_Hist_Traj(ScanProject,LogPTot,S[:,:,:,:int(MemSize)],xx,yy,zz,txyznM,n,TrajToGet,i,TopHatPrior,MemIndex,ZI,GroundTruth)               
            
            ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]=txyznM
            if do_onlinesave:
                _DS.saveTrajtoMat(ScanProject, ScanProject.PoissonProbXYZMap_NumPFS,'PoissonProbXYZMap_NumPFS')

            endtime=time.time()
            print('Looptime: ' + str(endtime-starttime))

    def get_traj_Poisson_ProbMap_fix_NumPFS(self, ScanProject, TrajToGet, verbose=1, do_onlinesave=0, SaveSmatrix='', ResamplePhotons=0):
        """
        Generates a Poisson likelihood over x,y using Numerical PFS
        Input:
            ScanProject         All raw data possible to be used in the estimation  
            TrajToGet           Identifier for which data to be used 
            verbose=1           Showing steps of estimation in stdout
            do_onlinesave=0     save to mat file on each iteration 
            SaveSmatrix=''      save the S matrix for each iteration if the string is not '', 
                                if the string is 'something' it will be appended to the name of the saved file 
        Output: 
            xyM = list of x,y values over where likelihood is max
            LT  = likelihood as a function of x,y given by the grid outputs
            xx  = x coordinate grid for LT
            yy  = y coordinate grid for LT
            verbose = boolian, showe iterations 
        """
        if verbose:
            print("Evaluating trajectory #: " + str(TrajToGet))
        
        if not hasattr(ScanProject,'PoissonProbXYZMap_NumPFS'):
            ScanProject.PoissonProbXYZMap_NumPFS={}
            ScanProject.PoissonProbXYZMapS_NumPFS={}
            ScanProject.PoissonProbXYZMapC_NumPFS={}

        if ResamplePhotons:
            if not hasattr(ScanProject,'PoissonProbXYZMap_NumPFS_PhotoReSampled'):
                ScanProject.PoissonProbXYZMap_NumPFS_PhotoReSampled={}
                ScanProject.PoissonProbXYZMapS_NumPFS_PhotoReSampled={}
                ScanProject.PoissonProbXYZMapC_NumPFS_PhotoReSampled={}

        if ScanProject.SetupSpecifics['is_simulation']:
            if ScanProject.Sim_par['TrajType']!='Internal_traj':
                file='GroundTruth.mat'
                folder=ScanProject.Files['AnalysisConfig'][:-18]
                GroundTruth=io.loadmat(folder+file)
                GroundTruth=GroundTruth[ScanProject.Sim_par['TrajType']]
            elif ScanProject.Sim_par['TrajType']=='Internal_traj':
                GroundTruth=ScanProject.GroundTruthTrajs[TrajToGet]
            else:
                print('Groundtruth data can not be found')
        else:
            GroundTruth=np.nan

        ######## Get data
        ScanMatrix=ScanProject.ScanMatrix
        Trajs=ScanProject.Trajs
        run_cycles     = np.int(ScanProject.PosCalibration['run_cycles']) 
        run_wndw       = np.int(ScanProject.PosCalibration['run_wndw'])
        xndw=run_cycles*run_wndw
        m = int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use
        MemSize=ScanProject.PosCalibration['TrackingMemsize']   #Amount of points that is a pattern 
        nBackgroundBias=ScanProject.TrajectorySelectionRules['nBackgroundBias']   #Amount of points that is a pattern 

        T,XP,YP,XE,YE,XO,YO,ZI,APD1,APD2,J, MemIndex, ZI_um, ZI_OffsetIndex = _mis.get_traj_raw_data3(Trajs, TrajToGet, ScanMatrix,ScanProject)
        n=APD1+APD2+nBackgroundBias

        BlankingIndex=np.where(ScanProject.PosCalibration['LaserShotPattern']==0)[0]
        if len(BlankingIndex)>0: 
            for BlankIndex in range(len(BlankingIndex)):
                ToBlank=np.where(MemIndex==BlankingIndex[BlankIndex])
                n[ToBlank]=0
        
        if ScanProject.PlotShotPattern: #Debug
            import os
            from mpl_toolkits.mplot3d import Axes3D
            fig = plt.figure();
            ax = fig.add_subplot(111, projection='3d');
            intervall=range(0,43);
            ax.plot(XO[intervall],YO[intervall],ZI_OffsetIndex[intervall],'-*');

            file="Shot_pattern.svg" 
            path='/'.join(ScanProject.Files['AnalysisConfig'].split('/')[:-3]) + '/'
            plt.savefig(path + file, bbox_inches='tight')
            plt.close()
            print('Shot patten part of run is done, the program will exit')
            os._exit(1)
            

        if 0: #View photon map
            def plotScatter(R):
                for ii in range(R):
                    intervall=np.arange(30*5*ii,30*5*(ii+1))
                    MenUnique=np.unique(MemIndex[intervall]).astype('int')
                    Ntot=np.zeros(len(MenUnique));xUnique=np.zeros(len(MenUnique));yUnique=np.zeros(len(MenUnique))
                    for i in MenUnique:
                        Ntot[i]=np.sum(n[intervall][np.where(MemIndex[intervall]==i)])
                        xUnique[i]=XO[intervall][np.where(MemIndex[intervall]==i)[0][0]]
                        yUnique[i]=YO[intervall][np.where(MemIndex[intervall]==i)[0][0]]

                    for i in range(len(MenUnique)):
                        plt.text(xUnique[i],yUnique[i],str(MenUnique[i]))
                    #plt.plot(xUnique,yUnique,'-g')
                    plt.plot(np.sum(xUnique*Ntot/np.sum(Ntot)),np.sum(yUnique*Ntot/np.sum(Ntot)),'r*')
                    #plt.scatter(xUnique,yUnique,c=Ntot,s=(2*Ntot)**2,alpha=0.5);plt.clim(0,50)
                    plt.scatter(xUnique,yUnique,c=Ntot/np.sum(Ntot),s=(200*Ntot/np.sum(Ntot))**2,alpha=0.5);plt.clim(0,0.2)
                    plt.colorbar();
                    
                    if 1:
                        file='T_' + str(ii)
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'imgMap/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                        plt.savefig(path + file)
                        plt.close()
                    else:
                        plt.show()

            
            plotScatter(R=int(n.shape[0]/(5*30)))
            print('Image Done')

        if ResamplePhotons:
            #Resample photons according a Poission statistics
            #Overright n and save them within TrajsDic
            Photon_wndw=ScanProject.TrajsDic['wndw']
            Photon_wndwS=1 #Photon_wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            n_Mean=n.copy()
            if Photon_wndw>1:
                MenUnique=np.unique(MemIndex)
                for M in MenUnique:
                    MWhere=np.where(MemIndex==M)[0]
                    n_Mean[MWhere]= np.convolve(np.ones(Photon_wndw)/Photon_wndw,n[MWhere])[0:-Photon_wndw+1:Photon_wndwS]
            else:
                n_Mean=n
            n=np.random.poisson(lam=n_Mean)
            try:
                ScanProject.TrajsDic['NewPhotonCounts',str(TrajToGet)]+=(n,)
            except:
                ScanProject.TrajsDic['NewPhotonCounts',str(TrajToGet)]=()
                ScanProject.TrajsDic['NewPhotonCounts',str(TrajToGet)]+=(n,)

        #override old ZI to the new ZI_OffsetIndex
        ZI=ZI_OffsetIndex

        ###############################
        if 0: #Debug look at photon distribution and see if they are poissonien
            TestInt=np.where((T-T[0]<0.4)*(T-T[0]>0.3))
            n0=n[TestInt][np.where(MemIndex[TestInt]==0)]
            n1=n[TestInt][np.where(MemIndex[TestInt]==1)]
            n2=n[TestInt][np.where(MemIndex[TestInt]==2)]
            n3=n[TestInt][np.where(MemIndex[TestInt]==3)]
            n4=n[TestInt][np.where(MemIndex[TestInt]==4)]
            n5=n[TestInt][np.where(MemIndex[TestInt]==5)]
       
            ###Hists
            bins=np.linspace(0,50,50)
            Size=500
            fig, ax = plt.subplots(); ax.hist(np.random.poisson(lam=np.mean(n0),size=Size),bins=bins,alpha=0.6,label=str(np.mean(n0)));ax.hist(np.random.poisson(lam=np.mean(n1),size=Size),bins=bins,alpha=0.6,label=str(np.mean(n1)));ax.legend();
            fig, ax = plt.subplots(); ax.hist(n0,bins=bins,alpha=0.6,label=str(np.mean(n0)));ax.hist(n1,bins=bins,alpha=0.6,label=str(np.mean(n1)));ax.legend();
            plt.show()
            fig, ax = plt.subplots(); ax.hist(np.random.poisson(lam=np.mean(n2),size=Size),bins=bins,alpha=0.6,label=str(np.mean(n2)));ax.hist(np.random.poisson(lam=np.mean(n3),size=Size),bins=bins,alpha=0.6,label=str(np.mean(n3)));ax.legend();
            fig, ax = plt.subplots(); ax.hist(n2,bins=bins,alpha=0.6,label=str(np.mean(n2)));ax.hist(n3,bins=bins,alpha=0.6,label=str(np.mean(n3)));ax.legend();
            plt.show()
            fig, ax = plt.subplots(); ax.hist(np.random.poisson(lam=np.mean(n4),size=Size),bins=bins,alpha=0.6,label=str(np.mean(n4)));ax.hist(np.random.poisson(lam=np.mean(n5),size=Size),bins=bins,alpha=0.6,label=str(np.mean(n5)));ax.legend();
            fig, ax = plt.subplots(); ax.hist(n4,bins=bins,alpha=0.6,label=str(np.mean(n4)));ax.hist(n5,bins=bins,alpha=0.6,label=str(np.mean(n5)));ax.legend();
            plt.show()

            #### Hist 2
            fig, ax = plt.subplots(); ax.hist(150*(n0-n1)/(n0+n1),bins=50,alpha=1.0,label=str(np.mean(150*(n0-n1)/(n0+n1))));ax.legend();plt.show()
            fig, ax = plt.subplots(); ax.hist(150*(n2-n3)/(n2+n3),bins=50,alpha=1.0,label=str(np.mean(150*(n2-n3)/(n2+n3))));ax.legend();plt.show()
            fig, ax = plt.subplots(); ax.hist(325*(n4-n5)/(n4+n5),bins=50,alpha=1.0,label=str(np.mean(325*(n4-n5)/(n4+n5))));ax.legend();plt.show()
        ############################################

            #Get Numerical PSF light field, to get intensity take |LF|^2
        path=ScanProject.Ph_Retrieval['path_PSF']
        fname=ScanProject.Ph_Retrieval['fname_PSF']
        LF, parm_PR= PSF.LoadMat(path,fname)
        if ScanProject.Ph_Retrieval['load_IntencityC']:
            ScanProject.Ph_Retrieval['IntencityC']=parm_PR['IntencityC']

        #bundle some parameters
        parm={}
        parm['Px']    = ScanProject.SetupSpecifics['Px']            
        parm['Mag']   = ScanProject.SetupSpecifics['Mag']           
        parm['Mag2']  = ScanProject.SetupSpecifics['Mag2']          
        parm['NM']    = LF.shape            
        parm['SWS']   = ScanProject.Ph_Retrieval['XYSWScalefactor'] # Software scaling factor
        parm['zssize']= ScanProject.Ph_Retrieval['zssize']        #[m] z- step size between images in PSF stack
        parm['SWSZ']   = ScanProject.Ph_Retrieval['ZSWScalefactor'] # Software scaling factor for z

            #Pixel convertion factors  
        px2um=parm['Px']*1e6/(parm['Mag']*parm['Mag2'])  # from camera pixel to um
        spx2um=px2um/parm['SWS']                         # from super resolution to um 

        # Add z resolution            
        IP=PSF.zInterpolator(LF, parm['SWSZ'],parm_PR,p='WeightedTwoWay')
        
        # Add background determined by the signal to noise ratio 
        # Only if the background is not a vector, a vector is added after cropping out and interpolation, this is for backword compatibility
        Background=ScanProject.TrajectorySelectionRules['background']
        if not isinstance(Background,(np.ndarray)):
            IP=IP+Background
            if Background<0:
                IP[np.where(IP<0)]=0
       
        
        
        #Get pinhole size in [um] 
        PholeDia = ScanProject.SetupSpecifics['PholeDia']/ScanProject.SetupSpecifics['Mag'] #size of confocal pinhole in the sample plan

        #Offset for each levles [x0,y0],[x1,y1],[x2,y2],[x2,y2],... in [um] 
        # and is suppose to be the PSF max intencity in the center of the pinhole
        PinholeOffset=ScanProject.Ph_Retrieval['IntencityC'][:,0:2]*px2um #[um]
        zInFocusIndex=ScanProject.Ph_Retrieval['IntencityC'][:,2]*parm['SWSZ']

        if 0: #Debug photon count plot  
            wndw=24
            wndwS=wndw #if 1 it is a moving average if equal to wndw it is like a fix window average
            Elements=np.unique(MemIndex)
            color=cm.rainbow(np.linspace(0,1,len(Elements)))
            legendString=[]
            for MemUnique,c in zip(Elements,color):
                OfInterest=np.where(MemIndex==MemUnique)
                TT=T[OfInterest]
                nT=n[OfInterest]
                TAV = np.convolve(np.ones(wndw),TT)
                TAV=TAV[wndw-1:-wndw+1:wndwS]/np.sum(np.ones(wndw))
                nAV = np.convolve(np.ones(wndw),nT)
                nAV=nAV[wndw-1:-wndw+1:wndwS]
                plt.plot(TAV,((-1)**MemUnique)*nAV,'-',c=c) 
                legendString.append('M' + str(int(MemUnique)) + ' ZI' + str(ZI[OfInterest][0]) )
            plt.xlabel('Time [ms]')
            plt.ylabel('Counts over ' + str(wndw) + ' measurements')
            plt.legend(tuple(legendString))
            plt.grid()
            plt.show()            

        #Run itaration
        NT=int(len(T)/m) #Amount of evaluations over the trajectory
        txyznM=np.zeros((5,NT))
        StxyznM=np.zeros((3,NT))
        Cvalue=np.zeros(NT)
        for i in range(NT):   
            starttime=time.time()
            if verbose:
                print(str(i) + " of " + str(NT))

            ni=n[i*m:(i+1)*m]
            ti=T[i*m:(i+1)*m]
            Ji=np.int32(J[i*m:(i+1)*m])
            Jump=np.where(Ji==1)[0]

            t=ti[-1] #time is set as the last slot, this is to conform to the fast gaussian calculation
            ntot=np.sum(ni-nBackgroundBias) #Remove the background bias 

            if (Jump.size!=0): 
                if (Jump[0]==0) or (Jump[0]==xndw):
                    pass
                else:
                    print("Varning: Possible estimation over a time jumpe")

            zi =ZI[i*m:(i+1)*m]  
            xiP=XP[i*m:(i+1)*m]; yiP=YP[i*m:(i+1)*m]; ziP=ZI_um[i*m:(i+1)*m]
            xiE=XE[i*m:(i+1)*m]; yiE=YE[i*m:(i+1)*m]
            xiO=XO[i*m:(i+1)*m]; yiO=YO[i*m:(i+1)*m]
            xPinOffset=PinholeOffset[(zi).astype('int'),0]; yPinOffset=PinholeOffset[(zi).astype('int'),1]
            
            SaveSmatrix_i=SaveSmatrix
            if SaveSmatrix!='':
                SaveSmatrix_i=SaveSmatrix + '_' + str(i)

            print ("Max EOD   move: x=" + str(np.round((np.max(xiE)-np.min(xiE))*1e3, decimals=3)) + "nm, y=" + str(np.round((np.max(yiE)-np.min(yiE))*1e3, decimals=3)) + "nm")
            print ("Max Piezo move: x=" + str(np.round((np.max(xiP)-np.min(xiP))*1e3, decimals=3)) + "nm, y=" + str(np.round((np.max(yiP)-np.min(yiP))*1e3, decimals=3)) + "nm")

            ### Time: up to here 0ms
            if i==0:
                MemOrderToKeep=MemIndex[i*m:(i+1)*m]

            if (i==0) or len(np.where((MemOrderToKeep==MemIndex[i*m:(i+1)*m])==False)[0]): #only calculate S, P logP, PP and tophat matrix one time
                #Verification that the order of the pattern is preserved
                print('Mem index is in wrong order!! Shot position to photon counts relation is compromised')

                MemOrderToKeep=MemIndex[i*m:(i+1)*m]
                N1=time.time()
                S_original,xx_original,yy_original,zz_original = PSF.get_Smatrix3(ScanProject,IP,xiO,yiO,xPinOffset,yPinOffset, SaveSmatrix_i)
                print('Smatrix time: ' + str(time.time()-N1))

                if 0: #plot beam configuration
                    fig = plt.figure()
                    ax = plt.axes(projection='3d')
                    ax.plot3D(xiO,yiO, zInFocusIndex[zi.astype('int')],'-*', linewidth=2, markersize=12);
                    ax.axes.xaxis.set_ticklabels([])
                    ax.axes.yaxis.set_ticklabels([])
                    ax.axes.zaxis.set_ticklabels([])
                    ax.set_xlabel('x')
                    ax.set_ylabel('y')
                    ax.set_zlabel('z')
                    plt.show()
                    print('done!')
                    

                #Setting signal to noise ratio according to the background parameter
                if isinstance(Background,(np.ndarray)):
                    S_original=S_original/np.max(S_original)
                    Background= Background[MemOrderToKeep.astype('int')]

                    for j in range(S_original.shape[-1]):
                        a=np.max(S_original[:,:,:,j])
                        B=a/(Background[j]-1)
                        S_original[:,:,:,j]=a*(S_original[:,:,:,j]+B)/(a+B)
                    

                # Do a check on the S matrix that is it valid
                #if np.sum(np.where(S_original<0))>0:
                #    print("S has negative enteris, puting them to zero")
                #    S_original[np.where(S_original<0)]=0

                BlankingIndex=np.where(ScanProject.PosCalibration['LaserShotPattern']==0)[0]
                if len(BlankingIndex)>0: 
                    for BlankIndex in BlankingIndex:
                        #S_original[:,:,:,BlankIndex]=0
                        S_original[:,:,:,np.where(MemOrderToKeep==BlankIndex)[0][0]]=0
                        
                        
                if 0: 
                    a="C:\\Users\\elias\\Desktop\\Shadowfax\\Data\\20210929_EXP-21-BL9411_(SimulationMultibeam)\\Simulation_7\\"
                    groundtruth=_DS.load_MatFile(a,'GroundTruth.mat')
                    groundtruth=np.squeeze(groundtruth['Grid_cloud'])
                        
                P_original=PSF.get_ProbFromSmatrix(S_original)
                LogP_original=PSF.get_logProb(P_original)
                C_original=PSF.get_InvFromSmatrix(S_original)

                if ScanProject.TrajectorySelectionRules['Wprio'][2]>0 or ScanProject.TrajectorySelectionRules['Wprio'][3]>0:
                    PP_original=PSF.get_FullProbFromSmatrix(S_original)
                else:
                    PP_original=np.zeros(S_original.shape)
                


                #Get only tophat prior if used
                SGxyz=[0, 0, 0]; xyzCenterGuessG=[0, 0, 0]
                
                TopHatOffset=ScanProject.TrajectorySelectionRules['TopHatoffset']
                if np.isscalar(TopHatOffset):
                    TopHatOffset=np.array([0,0,TopHatOffset])
                Myxz=np.mean(self.get_SmaxArg(S_original,xx_original,yy_original,zz_original), axis=0)*0 #Disabled 
                xyzCenterGuess=np.array([TopHatOffset[0]+Myxz[1],TopHatOffset[1]+Myxz[0], TopHatOffset[2]+Myxz[2]])
                STxyz= ScanProject.TrajectorySelectionRules['STxyz'] #[um] Gaussian prior sigma
                _, TopHatPrior_original, _, _, _=PSF.get_logPrior_V2(P_original,PP_original,xx_original,yy_original,zz_original,ni,SGxyz,STxyz,xyzCenterGuess,xyzCenterGuessG,Wprio=ScanProject.TrajectorySelectionRules['Wprio'])

                #Debug plots for S, P and C matrix
                if 0: #Debug Plot S function 
                    DebugPath=ScanProject.Files['AnalysisConfig'][:-18]
                    DebugPath= DebugPath + "S_matrix_Visualization/"
                    from PIL import Image
                    for num in range(m):
                        for kk in range(S_original.shape[2]):
                            #Image.fromarray(P[:,:,kk,num]).save(DebugPath + str(num) + "/P_PSF" + str(kk) + "_" + str(num) + '.tif')
                            #TopHat=np.ones(S_original.shape)
                            TopHat=np.ones(S_original.shape)#np.zeros(S_original.shape)
                            #TopHat[np.where(TopHatPrior_original==0)]=1
                            path_SVisualization=DebugPath + "Memindex_"+ str(int(MemOrderToKeep[num])) + "/"
                            if not os.path.exists(path_SVisualization):
                                os.makedirs(path_SVisualization)
                            Image.fromarray((TopHat*S_original)[:,:,kk,num]).save(path_SVisualization + "S_PSF" + str(kk) + "_" + str(int(MemOrderToKeep[num])) + '.tif')
                if 0: #Debug Plot P function 
                    #Obs the order of the stack might not be that of the FPGA memmory, depends on which is the first index when tracking sarts. 
                    #DebugPath="C:/Users/elias/Desktop/Shadowfax/Data/Debug/"
                    DebugPath=ScanProject.Files['AnalysisConfig'][:-18]
                    DebugPath= DebugPath + "P_matrix_Visualization/"
                    from PIL import Image
                    for num in range(m):
                        for kk in range(P_original.shape[2]):
                            #Image.fromarray(P[:,:,kk,num]).save(DebugPath + str(num) + "/P_PSF" + str(kk) + "_" + str(num) + '.tif')
                            #TopHat=np.ones(P_original.shape)
                            TopHat=np.ones(S_original.shape)#np.zeros(S_original.shape)
                            #TopHat[np.where(TopHatPrior_original==0)]=1
                            path_PVisualization=DebugPath + "Memindex_"+ str(int(MemOrderToKeep[num])) + "/"
                            if not os.path.exists(path_PVisualization):
                                os.makedirs(path_PVisualization)
                            Image.fromarray((TopHat*P_original)[:,:,kk,num]).save(path_PVisualization + "P_PSF" + str(kk) + "_" + str(int(MemOrderToKeep[num])) + '.tif')
                if 0: #Debug Plot C function 
                    DebugPath=ScanProject.Files['AnalysisConfig'][:-18]
                    DebugPath= DebugPath + "C_matrix_Visualization/"
                    from PIL import Image
                    for kk in range(C_original.shape[2]):
                        #Image.fromarray(P[:,:,kk,num]).save(DebugPath + str(num) + "/P_PSF" + str(kk) + "_" + str(num) + '.tif')
                        #TopHat=np.ones(S_original.shape)
                        #TopHat=np.zeros(C_original.shape)
                        #TopHat[np.where(TopHatPrior_original==0)]=1
                        path_CVisualization=DebugPath #+ "Memindex_"+ str(int(MemOrderToKeep[num])) + "/"
                        if not os.path.exists(path_CVisualization):
                            os.makedirs(path_CVisualization)
                        Image.fromarray((C_original)[:,:,kk]).save(path_CVisualization + "S_PSF_" + str(kk) + '.tif')


                #Find the bigest non zero volume defined by tophat and redefine all parameters 
                xedges=(np.sum(np.sum(TopHatPrior_original==0,axis=2),axis=1)>0)*1.0
                xL,xR=_mis.get_LR_pos(xedges)
                yedges=(np.sum(np.sum(TopHatPrior_original==0,axis=2),axis=0)>0)*1.0
                yL,yR=_mis.get_LR_pos(yedges)
                zedges=(np.sum(np.sum(TopHatPrior_original==0,axis=0),axis=0)>0)*1.0
                zL,zR=_mis.get_LR_pos(zedges)

                #xx_original,yy_original,zz_original
                TopHatPrior=TopHatPrior_original[xL:xR,yL:yR,zL:zR]
                S=S_original[xL:xR,yL:yR,zL:zR]
                P=P_original[xL:xR,yL:yR,zL:zR]
                C=C_original[xL:xR,yL:yR,zL:zR]
                PP=PP_original[xL:xR,yL:yR,zL:zR]
                LogP=LogP_original[xL:xR,yL:yR,zL:zR]
                xx_original=xx_original[xL:xR,yL:yR,zL:zR]
                yy_original=yy_original[xL:xR,yL:yR,zL:zR]
                zz_original=zz_original[xL:xR,yL:yR,zL:zR]

                xx2_original=xx_original**2
                yy2_original=yy_original**2
                zz2_original=zz_original**2

                zz_original=-1*zz_original #Reversing z 


                    #PreCalculations for gaussian prior
                SGxyz= ScanProject.TrajectorySelectionRules['SGxyz'] #[um] Gaussian prior sigma
                Wprio=ScanProject.TrajectorySelectionRules['Wprio']
                xyz2      = xx2_original/(2*SGxyz[0]**2) + yy2_original/(2*SGxyz[1]**2) + zz2_original/(2*SGxyz[2]**2)
                    #PreCalculation for photon log probability
                LogPLin=np.reshape(LogP,(np.prod(LogP.shape[:-1]),LogP.shape[-1]))

                
                #np.reshape(n[:NT*m],[NT,m])
                if 0: #Debug plot photon counts
                    for index in range(m):
                        #plt.hist(np.reshape(n[:NT*m],[NT,m])[:,index],bins=100,label=str(MemOrderToKeep[index]))
                        plt.plot(np.reshape(n[:NT*m],[NT,m])[:,index],label=str(MemOrderToKeep[index]))
                    plt.legend()
                    plt.show()

                #Get CRB for current beam profiles
                if 0: 
                    _TP.PlotCRB(ScanProject,P,xx_original,yy_original,zz_original,TopHatPrior,do_save_plot=1)
 
                
            #Create the offset (EOD and Piezo positions) to be added to the xx,yy,zz grid 
            xOffset=np.mean(xiP)-np.mean(xiE)
            yOffset=np.mean(yiP)-np.mean(yiE)
            zOffset=np.mean(ziP)

            xx=xOffset+xx_original
            yy=yOffset+yy_original
            zz=zOffset+zz_original

            ##### Get Loglikelihood 
            # Create prior start estimation 
            #xyzCenterGuess=np.array([xiP[0],yiP[0],0.36])
            if i==0:
                xyzCenterGuessG=np.array([-100,-100,-100])
                SGxyz=ScanProject.TrajectorySelectionRules['SGxyz'] #set according to the ini file
            else:
                if i==1:
                    xyzCenterGuessG=txyznM[1:-1,i-1]
                    SGxyz=ScanProject.TrajectorySelectionRules['SGxyz'] #set according to the ini file
                else:
                    if 1: #addaptive guess 
                        RValue=0.5#*np.exp(-100*StxyznM[:,i-1]**2) #select weight according the the stability of previoes points up to 95%
                        xyzCenterGuessG=(1-RValue)*xyzCenterGuessG + RValue*txyznM[1:-1,i-1]

                        #RValue=0.95*np.exp(-1000*StxyznM[:,i-1]**2) #select weight according the the stability of previoes points up to 95%
                        #SGxyz=(1-0.8)*SGxyz + 0.8*((1-RValue)*1 + RValue*0.68)*ScanProject.TrajectorySelectionRules['SGxyz']
                    else: #fixed guess
                        RValue=1#0.6
                        xyzCenterGuessG=(1-RValue)*xyzCenterGuessG + RValue*txyznM[1:-1,i-1]
                    
            # Create cross section prior
            
            if  np.sum(ScanProject.TrajectorySelectionRules['CrossSection'])!=0:
                TC=np.sum(ni)*C
                cminmax=ScanProject.TrajectorySelectionRules['CrossSection']
                TCPrior=(-1e10)*np.logical_not((cminmax[0]<TC)*(TC<cminmax[1]))
                if i==0:
                    CvalueRunningMean=0
                    TCPrior2=0
                else:
                    CTailFactor=ScanProject.TrajectorySelectionRules['CTailFactor']
                    CvalueRunningMean=(1-CTailFactor[0])*CvalueRunningMean + CTailFactor[0]*Cvalue[i-1]
                    TCPrior2=-(TC-CvalueRunningMean)**2/(2*CTailFactor[1]**2) #Gaussian C-prior around a running mean
            else:
                TCPrior=0
                TCPrior2=0
            #    
            ### Time: up to here 2ms

            if 0:
                LogPTot, _, _, _, _, _= PSF.get_PoissonLoglikelihood_V3(ScanProject,P,PP,LogPLin,TopHatPrior,xx,yy,zz,ni,xyzCenterGuessG)
            else:
                print("Running Poisson estimation")

                ### 1ms
                Pshape=PP.shape
                LogPN=np.reshape(np.matmul(LogPLin,ni),Pshape[:-1])
                ####
                
                #Gaussian Prior 5ms
                GPrior=0
                if Wprio[1]>0:
                    if xyzCenterGuessG[0]==-100:
                        GPrior=np.zeros(xx.shape)
                    else:
                        #GPrior=-((xx+(xoffset-xyzCenterGuessG[0]))**2 / (2*SGxyz[0]**2) + (yy+(yoffset-xyzCenterGuessG[1]))**2 / (2*SGxyz[1]**2) + (zz+(zoffset-xyzCenterGuessG[2]))**2 / (2*SGxyz[2]**2))                        
                        #GPrior=-((xx-xyzCenterGuessG[0])**2 / (2*SGxyz[0]**2) + (yy-xyzCenterGuessG[1])**2 / (2*SGxyz[1]**2) + (zz-xyzCenterGuessG[2])**2 / (2*SGxyz[2]**2))                        
                        #xyz2      = xx2_original/(2*SGxyz[0]**2) + yy2_original/(2*SGxyz[1]**2) + zz2_original/(2*SGxyz[2]**2)
                        xyz_Offset_CrossTerms = xx_original*((xOffset - xyzCenterGuessG[0])/(2*SGxyz[0]**2)) + yy_original*((yOffset - xyzCenterGuessG[1])/(2*SGxyz[1]**2)) + zz_original*((zOffset - xyzCenterGuessG[2])/(2*SGxyz[2]**2))
                        Offset_Terms   = (xOffset - xyzCenterGuessG[0])**2 /(2*SGxyz[0]**2)  + (yOffset - xyzCenterGuessG[1])**2 /(2*SGxyz[1]**2)  +   (zOffset - xyzCenterGuessG[2])**2 /(2*SGxyz[2]**2)
                        GPrior=-( xyz2 + 2*xyz_Offset_CrossTerms + Offset_Terms )

                ####

                #Construct the log likelihood 
                ScaleFactor=np.sum(ni)*(np.log(np.sum(ni))-1)-np.sum(_scs.gammaln(ni+1))
                LogPFinal=ScaleFactor + LogPN
                if i==0:
                    LogPFinalTail=LogPFinal
                else:
                    MLETailFactor=ScanProject.TrajectorySelectionRules['MLETailFactor']
                    LogPFinalTail=(1-MLETailFactor)*LogPFinalTail+MLETailFactor*LogPFinal

                LogPTot = LogPFinalTail + GPrior + TopHatPrior + TCPrior + TCPrior2
                LogPTot=LogPTot-np.max(LogPTot) 
                
            
            ### Time: up to here 8ms (between time 6ms)

            indextosee=np.unravel_index(np.argmax(LogPTot) ,LogPTot.shape)
            txyznM[:,i]=[t, xx[indextosee],yy[indextosee], zz[indextosee], ntot]
            Cvalue[i]=TC[indextosee]
            ## Look at the 98% Neighborhood and extract value: std, box bounds, Piezo_offset
            if 1:
                GP=np.exp(LogPTot)
                validData=np.where(GP>np.max(GP)*0.95)
                GP=GP[validData]
                xv=xx[validData] 
                yv=yy[validData]
                zv=zz[validData]
                StxyznM[:,i]=[np.std(xv), np.std(yv), np.std(zv)]
                xyzBounds=np.array([[np.min(xx),np.max(xx)],[np.min(yy),np.max(yy)],[np.min(zz),np.max(zz)]])
                Piezo_Offset=np.array([xOffset,yOffset,zOffset])
            
                try:
                    ScanProject.PoissonProbXYZMap_NeighborhoodBounds[TrajToGet]+=(xyzBounds,)
                    ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet]=np.vstack((ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet],Piezo_Offset))
                except:
                    if not hasattr(ScanProject,'PoissonProbXYZMap_NeighborhoodBounds'):
                        ScanProject.PoissonProbXYZMap_Piezo_Offset={}
                        ScanProject.PoissonProbXYZMap_NeighborhoodBounds={}
                    ScanProject.PoissonProbXYZMap_NeighborhoodBounds[TrajToGet]=()
                    ScanProject.PoissonProbXYZMap_NeighborhoodBounds[TrajToGet]+=(xyzBounds,)
                    ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet]=Piezo_Offset

                if 1: #Save Neighborhood, to look at the 3D plot use:  _TP.Neighborhood(ScanProject,TrajToGet=0,intervalEnd=200,intervalLength=100)
                    Gxyz_pos=np.transpose(np.vstack((GP,xv, yv, zv)))
                    try: 
                        ScanProject.PoissonProbXYZMap_Neighborhood[TrajToGet]+=(Gxyz_pos,)
                    except:
                        if not hasattr(ScanProject,'PoissonProbXYZMap_Neighborhood'):
                            ScanProject.PoissonProbXYZMap_Neighborhood={}
                        ScanProject.PoissonProbXYZMap_Neighborhood[TrajToGet]=()
                        ScanProject.PoissonProbXYZMap_Neighborhood[TrajToGet]+=(Gxyz_pos,)

            if 0:
                def do_plot_1():
                    plt.plot(txyznM[3,0:i+1],'-*');plt.plot(ZI_um[0:(i+1)*m:m]);plt.show()

                def do_plot_2():
                    plt.plot(MemIndex[i*m:(i+1)*m],ni,'*');plt.show()

                def do_plot_3():
                    rgba_colors=cm.jet((ni-np.min(ni))/(np.max(ni)-np.min(ni))); 
                    rgba_colors[:,3]=(ni-np.min(ni))/(np.max(ni)-np.min(ni)); 
                    fig = plt.figure(); 
                    ax = fig.add_subplot(projection='3d'); 
                    cax=ax.scatter(xiO,yiO,zi,c=rgba_colors); 
                    plt.show()

                def do_plot_4():
                    GP=np.exp(LogPTot); validData=np.where(GP>np.max(GP)*0.5); GP=GP[validData]; xv=xx[validData]; yv=yy[validData]; zv=zz[validData]; rgba_colors=cm.jet((GP.flatten()-np.min(GP.flatten()))/(np.max(GP.flatten())-np.min(GP.flatten()))); rgba_colors[:,3]=(GP.flatten()-np.min(GP.flatten()))/(np.max(GP.flatten())-np.min(GP.flatten())); fig = plt.figure(); ax = fig.add_subplot(projection='3d'); ax.scatter(xv.flatten(),yv.flatten(),zv.flatten(),c=rgba_colors);ax.scatter(xOffset,yOffset,zOffset); plt.show()

                def do_plot_5():
                    fig = plt.figure(); 
                    ax = fig.add_subplot(projection='3d'); 

                    rgba_colors=cm.Reds((ni-np.min(ni))/(np.max(ni)-np.min(ni))); 
                    rgba_colors[:,3]=(ni-np.min(ni))/(np.max(ni)-np.min(ni)); 
                    PHotonScatterPlot=ax.scatter(xiO+xOffset,yiO+yOffset,zz_original[0,0,zInFocusIndex.astype('int')][zi.astype('int')]+zOffset,c=rgba_colors,cmap=cm.Reds); 

                    GP=np.exp(LogPTot); 
                    validData=np.where(GP>np.max(GP)*0.8); 
                    GP=GP[validData]; xv=xx[validData]; yv=yy[validData]; zv=zz[validData]; 
                    rgba_colors=cm.plasma((GP.flatten()-np.min(GP.flatten()))/(np.max(GP.flatten())-np.min(GP.flatten()))); 
                    rgba_colors[:,3]=(GP.flatten()-np.min(GP.flatten()))/(np.max(GP.flatten())-np.min(GP.flatten())); 
                    ax.scatter(xv.flatten(),yv.flatten(),zv.flatten(),c=rgba_colors,cmap=cm.plasma);

                    ax.plot(XP[0:(i+1)*m:m],YP[0:(i+1)*m:m],ZI_um[0:(i+1)*m:m],'b-')
                    ax.plot(txyznM[1,0:i+1],txyznM[2,0:i+1],txyznM[3,0:i+1],'g-')
                    ax.scatter(xyzCenterGuessG[0],xyzCenterGuessG[1],xyzCenterGuessG[2], c=[1,0,0])
                    
                    #Create bounding box
                    xB=xyzBounds[0,:];yB=xyzBounds[1,:];zB=xyzBounds[2,:]
                    for s, e in combinations(np.array(list(product(xB, yB, zB))), 2):
                        if np.sum((s-e)==0)==2:
                            ax.plot3D(*zip(s, e), color="b")

                    ax.set_title('Tot Phtons: ' + str(np.sum(ni)))
                    #fig.colorbar(PHotonScatterPlot)

                    do_save_plot=1
                    if do_save_plot:
                        file="T_" + str(TrajToGet) + '_' + str(i) + '.png'
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Hikelihood_Traj' + '/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                        fig.savefig(path + file)
                        plt.close()
                        print('Done!')
                    else:
                        plt.show()

                do_plot_5()

            if 0: #Debug plot
                _TP.plotProbMapPointCloud(LogPTot,xx,yy,zz)
            if 0: #Debug plot
                _TP.plotProbMapProjections(ScanProject,LogPTot,xx,yy,zz,txyznM,txyzG,i,do_save_plot=0)
            if 0: #Debug plot
                if 1: #np.mod(i,10)==0:
                    _TP.plotProbMap_Hist_Traj(ScanProject,LogPTot,S[:,:,:,:int(MemSize)],xx,yy,zz,txyznM,n,TrajToGet,i,TopHatPrior,MemIndex,ZI,GroundTruth)               
            
            if ResamplePhotons==0:
                ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]=txyznM
                ScanProject.PoissonProbXYZMapS_NumPFS[TrajToGet]=StxyznM
                ScanProject.PoissonProbXYZMapC_NumPFS[TrajToGet]=Cvalue
            else:
                try:
                    ScanProject.PoissonProbXYZMap_NumPFS_PhotoReSampled[TrajToGet]+=(txyznM,)
                    ScanProject.PoissonProbXYZMapS_NumPFS_PhotoReSampled[TrajToGet]+=(StxyznM,)
                    ScanProject.PoissonProbXYZMapC_NumPFS_PhotoReSampled[TrajToGet]+=(Cvalue,)
                except:
                    ScanProject.PoissonProbXYZMap_NumPFS_PhotoReSampled[TrajToGet]=()
                    ScanProject.PoissonProbXYZMap_NumPFS_PhotoReSampled[TrajToGet]+=(txyznM,)
                    ScanProject.PoissonProbXYZMapS_NumPFS_PhotoReSampled[TrajToGet]=()
                    ScanProject.PoissonProbXYZMapS_NumPFS_PhotoReSampled[TrajToGet]+=(StxyznM,)
                    ScanProject.PoissonProbXYZMapC_NumPFS_PhotoReSampled[TrajToGet]=()
                    ScanProject.PoissonProbXYZMapC_NumPFS_PhotoReSampled[TrajToGet]+=(Cvalue,)
                

            if do_onlinesave:
                _DS.saveTrajtoMat(ScanProject, ScanProject.PoissonProbXYZMap_NumPFS,'PoissonProbXYZMap_NumPFS')

            endtime=time.time()
            print('Looptime: ' + str(endtime-starttime))

    def get_traj_Poisson_ProbMap_fix_NumPFS2(self, ScanProject, TrajToGet, verbose=1, do_onlinesave=0, SaveSmatrix='', ResamplePhotons=0):
        """
        Generates a Poisson likelihood over x,y using Numerical PFS
        In this version the background is also estimated

        Input:
            ScanProject         All raw data possible to be used in the estimation  
            TrajToGet           Identifier for which data to be used 
            verbose=1           Showing steps of estimation in stdout
            do_onlinesave=0     save to mat file on each iteration 
            SaveSmatrix=''      save the S matrix for each iteration if the string is not '', 
                                if the string is 'something' it will be appended to the name of the saved file 
        Output: 
            xyM = list of x,y values over where likelihood is max
            LT  = likelihood as a function of x,y given by the grid outputs
            xx  = x coordinate grid for LT
            yy  = y coordinate grid for LT
            verbose = boolian, showe iterations 
        """
        if verbose:
            print("Evaluating trajectory #: " + str(TrajToGet))
        
        if not hasattr(ScanProject,'PoissonProbXYZMap_NumPFS'):
            ScanProject.PoissonProbXYZMap_NumPFS={}
            ScanProject.PoissonProbXYZMapS_NumPFS={}
            ScanProject.PoissonProbXYZMapC_NumPFS={}

        if ResamplePhotons:
            if not hasattr(ScanProject,'PoissonProbXYZMap_NumPFS_PhotoReSampled'):
                ScanProject.PoissonProbXYZMap_NumPFS_PhotoReSampled={}
                ScanProject.PoissonProbXYZMapS_NumPFS_PhotoReSampled={}
                ScanProject.PoissonProbXYZMapC_NumPFS_PhotoReSampled={}

        if ScanProject.SetupSpecifics['is_simulation']:
            file='GroundTruth.mat'
            folder=ScanProject.Files['AnalysisConfig'][:-18]
            GroundTruth=io.loadmat(folder+file)
            GroundTruth=GroundTruth[ScanProject.Sim_par['TrajType']]
        else:
            GroundTruth=np.nan

        ######## Get data
        ScanMatrix=ScanProject.ScanMatrix
        Trajs=ScanProject.Trajs
        run_cycles     = np.int(ScanProject.PosCalibration['run_cycles']) 
        run_wndw       = np.int(ScanProject.PosCalibration['run_wndw'])
        xndw=run_cycles*run_wndw
        m = int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use
        MemSize=ScanProject.PosCalibration['TrackingMemsize']   #Amount of points that is a pattern 
        nBackgroundBias=ScanProject.TrajectorySelectionRules['nBackgroundBias']   #Amount of points that is a pattern 

        T,XP,YP,XE,YE,XO,YO,ZI,APD1,APD2,J, MemIndex, ZI_um, ZI_OffsetIndex = _mis.get_traj_raw_data3(Trajs, TrajToGet, ScanMatrix,ScanProject)
        n=APD1+APD2+nBackgroundBias

        BlankingIndex=np.where(ScanProject.PosCalibration['LaserShotPattern']==0)[0]
        if len(BlankingIndex)>0: 
            for BlankIndex in range(len(BlankingIndex)):
                ToBlank=np.where(MemIndex==BlankingIndex[BlankIndex])
                n[ToBlank]=0
        

        if 0: #View photon map
            def plotScatter(R):
                for ii in range(R):
                    intervall=np.arange(30*5*ii,30*5*(ii+1))
                    MenUnique=np.unique(MemIndex[intervall]).astype('int')
                    Ntot=np.zeros(len(MenUnique));xUnique=np.zeros(len(MenUnique));yUnique=np.zeros(len(MenUnique))
                    for i in MenUnique:
                        Ntot[i]=np.sum(n[intervall][np.where(MemIndex[intervall]==i)])
                        xUnique[i]=XO[intervall][np.where(MemIndex[intervall]==i)[0][0]]
                        yUnique[i]=YO[intervall][np.where(MemIndex[intervall]==i)[0][0]]

                    for i in range(len(MenUnique)):
                        plt.text(xUnique[i],yUnique[i],str(MenUnique[i]))
                    #plt.plot(xUnique,yUnique,'-g')
                    plt.plot(np.sum(xUnique*Ntot/np.sum(Ntot)),np.sum(yUnique*Ntot/np.sum(Ntot)),'r*')
                    #plt.scatter(xUnique,yUnique,c=Ntot,s=(2*Ntot)**2,alpha=0.5);plt.clim(0,50)
                    plt.scatter(xUnique,yUnique,c=Ntot/np.sum(Ntot),s=(200*Ntot/np.sum(Ntot))**2,alpha=0.5);plt.clim(0,0.2)
                    plt.colorbar();
                    
                    if 1:
                        file='T_' + str(ii)
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'imgMap/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                        plt.savefig(path + file)
                        plt.close()
                    else:
                        plt.show()

            
            plotScatter(R=int(n.shape[0]/(5*30)))
            print('Image Done')

        if ResamplePhotons:
            #Resample photons according a Poission statistics
            #Overright n and save them within TrajsDic
            Photon_wndw=ScanProject.TrajsDic['wndw']
            Photon_wndwS=1 #Photon_wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            n_Mean=n.copy()
            if Photon_wndw>1:
                MenUnique=np.unique(MemIndex)
                for M in MenUnique:
                    MWhere=np.where(MemIndex==M)[0]
                    n_Mean[MWhere]= np.convolve(np.ones(Photon_wndw)/Photon_wndw,n[MWhere])[0:-Photon_wndw+1:Photon_wndwS]
            else:
                n_Mean=n
            n=np.random.poisson(lam=n_Mean)
            try:
                ScanProject.TrajsDic['NewPhotonCounts',str(TrajToGet)]+=(n,)
            except:
                ScanProject.TrajsDic['NewPhotonCounts',str(TrajToGet)]=()
                ScanProject.TrajsDic['NewPhotonCounts',str(TrajToGet)]+=(n,)

        #override old ZI to the new ZI_OffsetIndex
        ZI=ZI_OffsetIndex

        ###############################
        if 0: #Debug look at photon distribution and see if they are poissonien
            TestInt=np.where((T-T[0]<0.4)*(T-T[0]>0.3))
            n0=n[TestInt][np.where(MemIndex[TestInt]==0)]
            n1=n[TestInt][np.where(MemIndex[TestInt]==1)]
            n2=n[TestInt][np.where(MemIndex[TestInt]==2)]
            n3=n[TestInt][np.where(MemIndex[TestInt]==3)]
            n4=n[TestInt][np.where(MemIndex[TestInt]==4)]
            n5=n[TestInt][np.where(MemIndex[TestInt]==5)]
       
            ###Hists
            bins=np.linspace(0,50,50)
            Size=500
            fig, ax = plt.subplots(); ax.hist(np.random.poisson(lam=np.mean(n0),size=Size),bins=bins,alpha=0.6,label=str(np.mean(n0)));ax.hist(np.random.poisson(lam=np.mean(n1),size=Size),bins=bins,alpha=0.6,label=str(np.mean(n1)));ax.legend();
            fig, ax = plt.subplots(); ax.hist(n0,bins=bins,alpha=0.6,label=str(np.mean(n0)));ax.hist(n1,bins=bins,alpha=0.6,label=str(np.mean(n1)));ax.legend();
            plt.show()
            fig, ax = plt.subplots(); ax.hist(np.random.poisson(lam=np.mean(n2),size=Size),bins=bins,alpha=0.6,label=str(np.mean(n2)));ax.hist(np.random.poisson(lam=np.mean(n3),size=Size),bins=bins,alpha=0.6,label=str(np.mean(n3)));ax.legend();
            fig, ax = plt.subplots(); ax.hist(n2,bins=bins,alpha=0.6,label=str(np.mean(n2)));ax.hist(n3,bins=bins,alpha=0.6,label=str(np.mean(n3)));ax.legend();
            plt.show()
            fig, ax = plt.subplots(); ax.hist(np.random.poisson(lam=np.mean(n4),size=Size),bins=bins,alpha=0.6,label=str(np.mean(n4)));ax.hist(np.random.poisson(lam=np.mean(n5),size=Size),bins=bins,alpha=0.6,label=str(np.mean(n5)));ax.legend();
            fig, ax = plt.subplots(); ax.hist(n4,bins=bins,alpha=0.6,label=str(np.mean(n4)));ax.hist(n5,bins=bins,alpha=0.6,label=str(np.mean(n5)));ax.legend();
            plt.show()

            #### Hist 2
            fig, ax = plt.subplots(); ax.hist(150*(n0-n1)/(n0+n1),bins=50,alpha=1.0,label=str(np.mean(150*(n0-n1)/(n0+n1))));ax.legend();plt.show()
            fig, ax = plt.subplots(); ax.hist(150*(n2-n3)/(n2+n3),bins=50,alpha=1.0,label=str(np.mean(150*(n2-n3)/(n2+n3))));ax.legend();plt.show()
            fig, ax = plt.subplots(); ax.hist(325*(n4-n5)/(n4+n5),bins=50,alpha=1.0,label=str(np.mean(325*(n4-n5)/(n4+n5))));ax.legend();plt.show()
        ############################################

            #Get Numerical PSF light field, to get intensity take |LF|^2
        path=ScanProject.Ph_Retrieval['path_PSF']
        fname=ScanProject.Ph_Retrieval['fname_PSF']
        LF, parm_PR= PSF.LoadMat(path,fname)
        if ScanProject.Ph_Retrieval['load_IntencityC']:
            ScanProject.Ph_Retrieval['IntencityC']=parm_PR['IntencityC']

        #bundle some parameters
        parm={}
        parm['Px']    = ScanProject.SetupSpecifics['Px']            
        parm['Mag']   = ScanProject.SetupSpecifics['Mag']           
        parm['Mag2']  = ScanProject.SetupSpecifics['Mag2']          
        parm['NM']    = LF.shape            
        parm['SWS']   = ScanProject.Ph_Retrieval['XYSWScalefactor'] # Software scaling factor
        parm['zssize']= ScanProject.Ph_Retrieval['zssize']        #[m] z- step size between images in PSF stack
        parm['SWSZ']   = ScanProject.Ph_Retrieval['ZSWScalefactor'] # Software scaling factor for z

            #Pixel convertion factors  
        px2um=parm['Px']*1e6/(parm['Mag']*parm['Mag2'])  # from camera pixel to um
        spx2um=px2um/parm['SWS']                         # from super resolution to um 

        # Add z resolution            
        IP=PSF.zInterpolator(LF, parm['SWSZ'],parm_PR,p='WeightedTwoWay')
        # Add background determined by the signal to noise ratio 
        Background=ScanProject.TrajectorySelectionRules['background']
        #IP=IP+Background
        if Background<0:
            IP[np.where(IP<0)]=0
        
        #Get pinhole size in [um] 
        PholeDia = ScanProject.SetupSpecifics['PholeDia']/ScanProject.SetupSpecifics['Mag'] #size of confocal pinhole in the sample plan

        #Offset for each levles [x0,y0],[x1,y1],[x2,y2],[x2,y2],... in [um] 
        # and is suppose to be the PSF max intencity in the center of the pinhole
        PinholeOffset=ScanProject.Ph_Retrieval['IntencityC'][:,0:2]*px2um #[um]
        zInFocusIndex=ScanProject.Ph_Retrieval['IntencityC'][:,2]*parm['SWSZ']

        if 0: #Debug photon count plot  
            wndw=24
            wndwS=wndw #if 1 it is a moving average if equal to wndw it is like a fix window average
            Elements=np.unique(MemIndex)
            color=cm.rainbow(np.linspace(0,1,len(Elements)))
            legendString=[]
            for MemUnique,c in zip(Elements,color):
                OfInterest=np.where(MemIndex==MemUnique)
                TT=T[OfInterest]
                nT=n[OfInterest]
                TAV = np.convolve(np.ones(wndw),TT)
                TAV=TAV[wndw-1:-wndw+1:wndwS]/np.sum(np.ones(wndw))
                nAV = np.convolve(np.ones(wndw),nT)
                nAV=nAV[wndw-1:-wndw+1:wndwS]
                plt.plot(TAV,((-1)**MemUnique)*nAV,'-',c=c) 
                legendString.append('M' + str(int(MemUnique)) + ' ZI' + str(ZI[OfInterest][0]) )
            plt.xlabel('Time [ms]')
            plt.ylabel('Counts over ' + str(wndw) + ' measurements')
            plt.legend(tuple(legendString))
            plt.grid()
            plt.show()            

        #Run itaration
        NT=int(len(T)/m) #Amount of evaluations over the trajectory
        txyznM=np.zeros((5,NT))
        StxyznM=np.zeros((3,NT))
        Cvalue=np.zeros(NT)
        for i in range(NT):   
            starttime=time.time()
            if verbose:
                print(str(i) + " of " + str(NT))

            ni=n[i*m:(i+1)*m]
            ti=T[i*m:(i+1)*m]
            Ji=np.int32(J[i*m:(i+1)*m])
            Jump=np.where(Ji==1)[0]

            t=ti[-1] #time is set as the last slot, this is to conform to the fast gaussian calculation
            ntot=np.sum(ni-nBackgroundBias) #Remove the background bias 

            if (Jump.size!=0): 
                if (Jump[0]==0) or (Jump[0]==xndw):
                    pass
                else:
                    print("Varning: Possible estimation over a time jumpe")

            zi =ZI[i*m:(i+1)*m]  
            xiP=XP[i*m:(i+1)*m]; yiP=YP[i*m:(i+1)*m]; ziP=ZI_um[i*m:(i+1)*m]
            xiE=XE[i*m:(i+1)*m]; yiE=YE[i*m:(i+1)*m]
            xiO=XO[i*m:(i+1)*m]; yiO=YO[i*m:(i+1)*m]
            xPinOffset=PinholeOffset[(zi).astype('int'),0]; yPinOffset=PinholeOffset[(zi).astype('int'),1]
            
            SaveSmatrix_i=SaveSmatrix
            if SaveSmatrix!='':
                SaveSmatrix_i=SaveSmatrix + '_' + str(i)

            print ("Max EOD   move: x=" + str(np.round((np.max(xiE)-np.min(xiE))*1e3, decimals=3)) + "nm, y=" + str(np.round((np.max(yiE)-np.min(yiE))*1e3, decimals=3)) + "nm")
            print ("Max Piezo move: x=" + str(np.round((np.max(xiP)-np.min(xiP))*1e3, decimals=3)) + "nm, y=" + str(np.round((np.max(yiP)-np.min(yiP))*1e3, decimals=3)) + "nm")

            ### Time: up to here 0ms

            if i==0: #only calculate S, P logP, PP and tophat matrix one time
                MemOrderToKeep=MemIndex[i*m:(i+1)*m]
                N1=time.time()
                S_original,xx_original,yy_original,zz_original = PSF.get_Smatrix3(ScanProject,IP,xiO,yiO,xPinOffset,yPinOffset, SaveSmatrix_i)
                print('Smatrix time: ' + str(time.time()-N1))
                # Do a check on the S matrix that is it valid
                if np.sum(np.where(S_original<0))>0:
                    print("S has negative enteris, puting them to zero")
                    S_original[np.where(S_original<0)]=0

                        
                P_original=PSF.get_ProbFromSmatrix(S_original)
                LogP_original=PSF.get_logProb(P_original)
                C_original=PSF.get_InvFromSmatrix(S_original)
                PP_original=np.zeros(S_original.shape)
                


                #Get only tophat prior if used
                SGxyz=[0, 0, 0]; xyzCenterGuessG=[0, 0, 0]
                
                TopHatOffset=ScanProject.TrajectorySelectionRules['TopHatoffset']
                if np.isscalar(TopHatOffset):
                    TopHatOffset=np.array([0,0,TopHatOffset])
                Myxz=np.mean(self.get_SmaxArg(S_original,xx_original,yy_original,zz_original), axis=0)*0 #Disabled 
                xyzCenterGuess=np.array([TopHatOffset[0]+Myxz[1],TopHatOffset[1]+Myxz[0], TopHatOffset[2]+Myxz[2]])
                STxyz= ScanProject.TrajectorySelectionRules['STxyz'] #[um] Gaussian prior sigma
                _, TopHatPrior_original, _, _, _=PSF.get_logPrior_V2(P_original,PP_original,xx_original,yy_original,zz_original,ni,SGxyz,STxyz,xyzCenterGuess,xyzCenterGuessG,Wprio=ScanProject.TrajectorySelectionRules['Wprio'])

                #Debug plots for S, P and C matrix
                if 0: #Debug Plot S function 
                    DebugPath=ScanProject.Files['AnalysisConfig'][:-18]
                    DebugPath= DebugPath + "S_matrix_Visualization/"
                    from PIL import Image
                    for num in range(S_original.shape[3]):
                        for kk in range(S_original.shape[2]):
                            #Image.fromarray(P[:,:,kk,num]).save(DebugPath + str(num) + "/P_PSF" + str(kk) + "_" + str(num) + '.tif')
                            #TopHat=np.ones(S_original.shape)
                            TopHat=np.zeros(S_original.shape)
                            TopHat[np.where(TopHatPrior_original==0)]=1
                            path_SVisualization=DebugPath + "Memindex_"+ str(int(MemOrderToKeep[num])) + "/"
                            if not os.path.exists(path_SVisualization):
                                os.makedirs(path_SVisualization)
                            Image.fromarray((TopHat*S_original)[:,:,kk,num]).save(path_SVisualization + "S_PSF" + str(kk) + "_" + str(num) + '.tif')
                if 0: #Debug Plot P function 
                    #Obs the order of the stack might not be that of the FPGA memmory, depends on which is the first index when tracking sarts. 
                    #DebugPath="C:/Users/elias/Desktop/Shadowfax/Data/Debug/"
                    DebugPath=ScanProject.Files['AnalysisConfig'][:-18]
                    DebugPath= DebugPath + "P_matrix_Visualization/"
                    from PIL import Image
                    for num in range(P_original.shape[3]):
                        for kk in range(P_original.shape[2]):
                            #Image.fromarray(P[:,:,kk,num]).save(DebugPath + str(num) + "/P_PSF" + str(kk) + "_" + str(num) + '.tif')
                            #TopHat=np.ones(P_original.shape)
                            TopHat=np.zeros(P_original.shape)
                            TopHat[np.where(TopHatPrior_original==0)]=1
                            path_PVisualization=DebugPath + "Memindex_"+ str(int(MemOrderToKeep[num])) + "/"
                            if not os.path.exists(path_PVisualization):
                                os.makedirs(path_PVisualization)
                            Image.fromarray((TopHat*P_original)[:,:,kk,num]).save(path_PVisualization + "P_PSF" + str(kk) + "_" + str(num) + '.tif')
                if 0: #Debug Plot C function 
                    DebugPath=ScanProject.Files['AnalysisConfig'][:-18]
                    DebugPath= DebugPath + "C_matrix_Visualization/"
                    from PIL import Image
                    for kk in range(C_original.shape[2]):
                        #Image.fromarray(P[:,:,kk,num]).save(DebugPath + str(num) + "/P_PSF" + str(kk) + "_" + str(num) + '.tif')
                        #TopHat=np.ones(S_original.shape)
                        #TopHat=np.zeros(C_original.shape)
                        #TopHat[np.where(TopHatPrior_original==0)]=1
                        path_CVisualization=DebugPath #+ "Memindex_"+ str(int(MemOrderToKeep[num])) + "/"
                        if not os.path.exists(path_CVisualization):
                            os.makedirs(path_CVisualization)
                        Image.fromarray((C_original)[:,:,kk]).save(path_CVisualization + "S_PSF_" + str(kk) + '.tif')


                #Find the bigest non zero volume defined by tophat and redefine all parameters 
                xedges=(np.sum(np.sum(TopHatPrior_original==0,axis=2),axis=1)>0)*1.0
                xL,xR=_mis.get_LR_pos(xedges)
                yedges=(np.sum(np.sum(TopHatPrior_original==0,axis=2),axis=0)>0)*1.0
                yL,yR=_mis.get_LR_pos(yedges)
                zedges=(np.sum(np.sum(TopHatPrior_original==0,axis=0),axis=0)>0)*1.0
                zL,zR=_mis.get_LR_pos(zedges)

                #xx_original,yy_original,zz_original
                TopHatPrior=TopHatPrior_original[xL:xR,yL:yR,zL:zR]
                S=S_original[xL:xR,yL:yR,zL:zR]
                P=P_original[xL:xR,yL:yR,zL:zR]
                C=C_original[xL:xR,yL:yR,zL:zR]
                PP=PP_original[xL:xR,yL:yR,zL:zR]
                LogP=LogP_original[xL:xR,yL:yR,zL:zR]
                xx_original=xx_original[xL:xR,yL:yR,zL:zR]
                yy_original=yy_original[xL:xR,yL:yR,zL:zR]
                zz_original=zz_original[xL:xR,yL:yR,zL:zR]

                xx2_original=xx_original**2
                yy2_original=yy_original**2
                zz2_original=zz_original**2

                zz_original=-1*zz_original #Reversing z 


                    #PreCalculations for gaussian prior
                SGxyz= ScanProject.TrajectorySelectionRules['SGxyz'] #[um] Gaussian prior sigma
                Wprio=ScanProject.TrajectorySelectionRules['Wprio']
                xyz2      = xx2_original/(2*SGxyz[0]**2) + yy2_original/(2*SGxyz[1]**2) + zz2_original/(2*SGxyz[2]**2)
                    #PreCalculation for photon log probability
                LogPLin=np.reshape(LogP,(np.prod(LogP.shape[:-1]),LogP.shape[-1]))
                SLin=np.reshape(S,(np.prod(S.shape[:-1]),S.shape[-1]))

                
                #np.reshape(n[:NT*m],[NT,m])
                if 0: #Debug plot photon counts
                    for index in range(m):
                        #plt.hist(np.reshape(n[:NT*m],[NT,m])[:,index],bins=100,label=str(MemOrderToKeep[index]))
                        plt.plot(np.reshape(n[:NT*m],[NT,m])[:,index],label=str(MemOrderToKeep[index]))
                    plt.legend()
                    plt.show()

                #Get CRB for current beam profiles
                if 0: 
                    _TP.PlotCRB(ScanProject,P,xx_original,yy_original,zz_original,TopHatPrior,do_save_plot=1)
 
                
            #Verification that the order of the pattern is preserved
            if len(np.where((MemOrderToKeep==MemIndex[i*m:(i+1)*m])==False)[0]):
                print('Mem index is in wrong order!! Shot position to photon counts relation is compromised')
                input("Press Enter to continue...")

            #Create the offset (EOD and Piezo positions) to be added to the xx,yy,zz grid 
            xOffset=np.mean(xiP)-np.mean(xiE)
            yOffset=np.mean(yiP)-np.mean(yiE)
            zOffset=np.mean(ziP)

            xx=xOffset+xx_original
            yy=yOffset+yy_original
            zz=zOffset+zz_original

            ##### Get Loglikelihood 
            # Create prior start estimation 
            #xyzCenterGuess=np.array([xiP[0],yiP[0],0.36])
            if i==0:
                xyzCenterGuessG=np.array([-100,-100,-100])
                SGxyz=ScanProject.TrajectorySelectionRules['SGxyz'] #set according to the ini file
            else:
                if i==1:
                    xyzCenterGuessG=txyznM[1:-1,i-1]
                    SGxyz=ScanProject.TrajectorySelectionRules['SGxyz'] #set according to the ini file
                else:
                    if 1: #addaptive guess 
                        RValue=0.5#*np.exp(-100*StxyznM[:,i-1]**2) #select weight according the the stability of previoes points up to 95%
                        xyzCenterGuessG=(1-RValue)*xyzCenterGuessG + RValue*txyznM[1:-1,i-1]

                        #RValue=0.95*np.exp(-1000*StxyznM[:,i-1]**2) #select weight according the the stability of previoes points up to 95%
                        #SGxyz=(1-0.8)*SGxyz + 0.8*((1-RValue)*1 + RValue*0.68)*ScanProject.TrajectorySelectionRules['SGxyz']
                    else: #fixed guess
                        RValue=1#0.6
                        xyzCenterGuessG=(1-RValue)*xyzCenterGuessG + RValue*txyznM[1:-1,i-1]
            
            #Background estimation 
            def bTest(b):
                MemSkip=np.where(ScanProject.PosCalibration['LaserShotPattern']==0)[0]
                ToUse=np.ones(MemOrderToKeep.shape)
                for ToUseIndex in MemSkip:
                    ToUse[np.where(MemOrderToKeep==ToUseIndex)[0]]=0
                nValid=ni[np.where(ToUse>0)[0]]/np.sum(ni)

                SLinValid=(SLin+b)[:,np.where(ToUse>0)[0]]
                PSFValid = ((np.sum(SLinValid,axis=1)/SLinValid.T).T)/np.sum(ScanProject.PosCalibration['LaserShotPattern']>0)

                FinalVector = np.matmul(PSFValid,nValid)-1
                FinalVector[np.where(np.isinf(FinalVector))[0]]=0
                return FinalVector

            FinalVector=bTest(b=2000); print(np.matmul(FinalVector,FinalVector));print(np.std(FinalVector))
            plt.plot(FinalVector[:100000]);plt.show()
            #LogPN=np.reshape(np.matmul(LogPLin,ni),Pshape[:-1])


            # Create cross section prior
            TC=np.sum(ni)*C
            cminmax=ScanProject.TrajectorySelectionRules['CrossSection']
            TCPrior=(-1e10)*np.logical_not((cminmax[0]<TC)*(TC<cminmax[1]))
            if i==0:
                CvalueRunningMean=0
                TCPrior2=0
            else:
                CTailFactor=ScanProject.TrajectorySelectionRules['CTailFactor']
                CvalueRunningMean=(1-CTailFactor[0])*CvalueRunningMean + CTailFactor[0]*Cvalue[i-1]
                TCPrior2=-(TC-CvalueRunningMean)**2/(2*CTailFactor[1]**2) #Gaussian C-prior around a running mean
            #    
            ### Time: up to here 2ms

            if 0:
                LogPTot, _, _, _, _, _= PSF.get_PoissonLoglikelihood_V3(ScanProject,P,PP,LogPLin,TopHatPrior,xx,yy,zz,ni,xyzCenterGuessG)
            else:
                print("Running Poisson estimation")

                ### 1ms
                Pshape=PP.shape
                LogPN=np.reshape(np.matmul(LogPLin,ni),Pshape[:-1])
                ####
                
                #Gaussian Prior 5ms
                GPrior=0
                if Wprio[1]>0:
                    if xyzCenterGuessG[0]==-100:
                        GPrior=np.zeros(xx.shape)
                    else:
                        #GPrior=-((xx+(xoffset-xyzCenterGuessG[0]))**2 / (2*SGxyz[0]**2) + (yy+(yoffset-xyzCenterGuessG[1]))**2 / (2*SGxyz[1]**2) + (zz+(zoffset-xyzCenterGuessG[2]))**2 / (2*SGxyz[2]**2))                        
                        #GPrior=-((xx-xyzCenterGuessG[0])**2 / (2*SGxyz[0]**2) + (yy-xyzCenterGuessG[1])**2 / (2*SGxyz[1]**2) + (zz-xyzCenterGuessG[2])**2 / (2*SGxyz[2]**2))                        
                        #xyz2      = xx2_original/(2*SGxyz[0]**2) + yy2_original/(2*SGxyz[1]**2) + zz2_original/(2*SGxyz[2]**2)
                        xyz_Offset_CrossTerms = xx_original*((xOffset - xyzCenterGuessG[0])/(2*SGxyz[0]**2)) + yy_original*((yOffset - xyzCenterGuessG[1])/(2*SGxyz[1]**2)) + zz_original*((zOffset - xyzCenterGuessG[2])/(2*SGxyz[2]**2))
                        Offset_Terms   = (xOffset - xyzCenterGuessG[0])**2 /(2*SGxyz[0]**2)  + (yOffset - xyzCenterGuessG[1])**2 /(2*SGxyz[1]**2)  +   (zOffset - xyzCenterGuessG[2])**2 /(2*SGxyz[2]**2)
                        GPrior=-( xyz2 + 2*xyz_Offset_CrossTerms + Offset_Terms )

                ####

                #Construct the log likelihood 
                ScaleFactor=np.sum(ni)*(np.log(np.sum(ni))-1)-np.sum(_scs.gammaln(ni+1))
                LogPFinal=ScaleFactor + LogPN 
                if i==0:
                    LogPFinalTail=LogPFinal
                else:
                    MLETailFactor=ScanProject.TrajectorySelectionRules['MLETailFactor']
                    LogPFinalTail=(1-MLETailFactor)*LogPFinalTail+MLETailFactor*LogPFinal

                LogPTot = LogPFinalTail + GPrior + TopHatPrior + TCPrior + TCPrior2
                LogPTot=LogPTot-np.max(LogPTot) 
                
                
            
            ### Time: up to here 8ms (between time 6ms)

            indextosee=np.unravel_index(np.argmax(LogPTot) ,LogPTot.shape)
            txyznM[:,i]=[t, xx[indextosee],yy[indextosee], zz[indextosee], ntot]
            Cvalue[i]=TC[indextosee]
            ## Look at the 98% Neighborhood and extract value: std, box bounds, Piezo_offset
            if 1:
                GP=np.exp(LogPTot)
                validData=np.where(GP>np.max(GP)*0.95)
                GP=GP[validData]
                xv=xx[validData] 
                yv=yy[validData]
                zv=zz[validData]
                StxyznM[:,i]=[np.std(xv), np.std(yv), np.std(zv)]
                xyzBounds=np.array([[np.min(xx),np.max(xx)],[np.min(yy),np.max(yy)],[np.min(zz),np.max(zz)]])
                Piezo_Offset=np.array([xOffset,yOffset,zOffset])
            
                try:
                    ScanProject.PoissonProbXYZMap_NeighborhoodBounds[TrajToGet]+=(xyzBounds,)
                    ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet]=np.vstack((ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet],Piezo_Offset))
                except:
                    if not hasattr(ScanProject,'PoissonProbXYZMap_NeighborhoodBounds'):
                        ScanProject.PoissonProbXYZMap_Piezo_Offset={}
                        ScanProject.PoissonProbXYZMap_NeighborhoodBounds={}
                    ScanProject.PoissonProbXYZMap_NeighborhoodBounds[TrajToGet]=()
                    ScanProject.PoissonProbXYZMap_NeighborhoodBounds[TrajToGet]+=(xyzBounds,)
                    ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet]=Piezo_Offset

                if 1: #Save Neighborhood, to look at the 3D plot use:  _TP.Neighborhood(ScanProject,TrajToGet=0,intervalEnd=200,intervalLength=100)
                    Gxyz_pos=np.transpose(np.vstack((GP,xv, yv, zv)))
                    try: 
                        ScanProject.PoissonProbXYZMap_Neighborhood[TrajToGet]+=(Gxyz_pos,)
                    except:
                        if not hasattr(ScanProject,'PoissonProbXYZMap_Neighborhood'):
                            ScanProject.PoissonProbXYZMap_Neighborhood={}
                        ScanProject.PoissonProbXYZMap_Neighborhood[TrajToGet]=()
                        ScanProject.PoissonProbXYZMap_Neighborhood[TrajToGet]+=(Gxyz_pos,)

            if 0: #Debug plot
                _TP.plotProbMapPointCloud(LogPTot,xx,yy,zz)
            if 0: #Debug plot
                _TP.plotProbMapProjections(ScanProject,LogPTot,xx,yy,zz,txyznM,txyzG,i,do_save_plot=1)
            if 0: #Debug plot
                if 1: #np.mod(i,10)==0:
                    _TP.plotProbMap_Hist_Traj(ScanProject,LogPTot,S[:,:,:,:int(MemSize)],xx,yy,zz,txyznM,n,TrajToGet,i,TopHatPrior,MemIndex,ZI,GroundTruth)               
            
            if ResamplePhotons==0:
                ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]=txyznM
                ScanProject.PoissonProbXYZMapS_NumPFS[TrajToGet]=StxyznM
                ScanProject.PoissonProbXYZMapC_NumPFS[TrajToGet]=Cvalue
            else:
                try:
                    ScanProject.PoissonProbXYZMap_NumPFS_PhotoReSampled[TrajToGet]+=(txyznM,)
                    ScanProject.PoissonProbXYZMapS_NumPFS_PhotoReSampled[TrajToGet]+=(StxyznM,)
                    ScanProject.PoissonProbXYZMapC_NumPFS_PhotoReSampled[TrajToGet]+=(Cvalue,)
                except:
                    ScanProject.PoissonProbXYZMap_NumPFS_PhotoReSampled[TrajToGet]=()
                    ScanProject.PoissonProbXYZMap_NumPFS_PhotoReSampled[TrajToGet]+=(txyznM,)
                    ScanProject.PoissonProbXYZMapS_NumPFS_PhotoReSampled[TrajToGet]=()
                    ScanProject.PoissonProbXYZMapS_NumPFS_PhotoReSampled[TrajToGet]+=(StxyznM,)
                    ScanProject.PoissonProbXYZMapC_NumPFS_PhotoReSampled[TrajToGet]=()
                    ScanProject.PoissonProbXYZMapC_NumPFS_PhotoReSampled[TrajToGet]+=(Cvalue,)
                

            if do_onlinesave:
                _DS.saveTrajtoMat(ScanProject, ScanProject.PoissonProbXYZMap_NumPFS,'PoissonProbXYZMap_NumPFS')

            endtime=time.time()
            print('Looptime: ' + str(endtime-starttime))

    def get_SmaxArg(self, S, xx, yy, zz):
        SC=S.copy()
        SC[np.isnan(SC)]=0
        SC[np.where(SC<np.max(SC)*0.4)]=0
        Myxz=np.zeros((SC.shape[3],3))
        for ii in range(SC.shape[3]):
            indextosee=np.unravel_index(np.argmax(SC[:,:,:,ii]),SC[:,:,:,ii].shape)
            Myxz[ii,:]=[yy[indextosee],xx[indextosee],zz[indextosee]]  

        return Myxz

    def get_Poisson_trajectory(self, ScanProject, t_interval_low, t_interval_high, xyt_Map, verbose=1):
        """
        Finds the maximum of the likelihood function
        Note: need to move out the NL parameter 
        """
        #Def parameters
        sx=ScanProject.PFS_Parameters['sigma'][0]
        sy=ScanProject.PFS_Parameters['sigma'][1]
        sz=ScanProject.PFS_Parameters['sigma'][2]
        m =int(ScanProject.TrajectorySelectionRules['wndw'])   #Sample size to use
        NL=1                                              #Laser intensity

        # Dictionery where all data will be stored
        self.Poisson_traj={}        
        
        ## Get raw data        
        t_raw=ScanProject.ScanMatrix[0,:]
        t_valid=np.where(np.logical_and(t_raw>=t_interval_low, t_raw<=t_interval_high))[0]
        
        t_raw=ScanProject.ScanMatrix[0,t_valid]                               #time [s]
        x_EOD_off=ScanProject.ScanMatrix[1,t_valid]                           #x_i Position [um]
        y_EOD_off=ScanProject.ScanMatrix[2,t_valid]                           #y_i Position [um]
        n=ScanProject.ScanMatrix[3,t_valid]+ScanProject.ScanMatrix[4,t_valid] #n_i photon counts 


        NT=int(len(x_EOD_off)/m)
        xyztM=np.zeros((4,NT))

        for i in range(NT):
            print(str(i) + " of " + str(NT))
            
            #Points to be used in evaluation
            xi = x_EOD_off[i*m:(i+1)*m]
            yi = y_EOD_off[i*m:(i+1)*m]
            zi = y_EOD_off[i*m:(i+1)*m]*0 + 1
            ti = np.mean(t_raw[i*m:(i+1)*m])
            ni = n[i*m:(i+1)*m]

            s  = np.array([sx,sy,sz])
            rL = np.vstack((xi,yi,zi))
            nL = ni
            
            if i==0:
                #rp0=np.array([xi[0],yi[0],zi[0]])
                rp0=[xyt_Map[0,i], xyt_Map[1,i], zi[0]]
            else:
                #rp0=xyztM[0:3,i-1]
                rp0=[xyt_Map[0,i], xyt_Map[1,i], xyztM[2,i-1]]
            
            
            args=(rL, nL, s, NL)
            rp_min=fmin(opt.min_target_Poisson,rp0,args,xtol=1e-6,ftol=1e-5, disp=1)
            xyztM[:,i]=np.concatenate((rp_min,[ti]))
           
        self.Poisson_traj[0]=xyztM
        
        