import torch
from torch.utils.data import Dataset
import numpy as np
from skimage import io, transform, filters
from PIL import Image
import matplotlib.pyplot as plt

class mmDataSetTrain(Dataset):
 
    """Dataset loader for training images from Elf lab pipeline.
    args: 
        phase_ims = list of phase contrast directories as full path: path/to/image/im.tiff
        mask_ims = list of masks to train segmentation as full path: path/to/mask/mask.tiff
        important! the list must be in the same order."""
    
    def __init__(self, phase_ims, mask_ims, transform=None):
        self.phase_ims = phase_ims
        self.mask_ims = mask_ims
        self.transform = transform
            
    def __len__(self):
        return len(self.phase_ims)
    
    def __getitem__(self, idx):
        
        #i dont know what it does but it was there in tutorial
        if torch.is_tensor(idx):
            idx = idx.tolist()
            
        #send to float32 type, PyTorch supposedly likes it
        phase_im = io.imread(self.phase_ims[idx],as_gray=True).astype('float32')
        
        mask_im = io.imread(self.mask_ims[idx],as_gray=True).astype('bool')
        
        sample = {'im' : phase_im, 'mask' : mask_im}
        
        if self.transform:
            sample = self.transform(sample)
            
        return sample
        
class mmDataSetSegment(Dataset):
    
    """Dataset loader for segmentation images from Elf lab pipeline.
    args: 
        phase_ims = list of phase contrast directories as full path: path/to/image/im.tiff
    returns normalised image and a path to save image to"""
    
    def __init__(self, phase_ims, output_folder, transform=None):

        # assert 'PreprocessedPhase' in phase_ims[0] #check if the phase images have been preprocessed before analysis

        self.phase_ims = phase_ims
        self.transform = transform
        self.output_folder = output_folder
            
    def __len__(self):
        return len(self.phase_ims)
    
    def __getitem__(self, idx):
        #i dont know what it does but it was there in tutorial
        if torch.is_tensor(idx):
            idx = idx.tolist()
            
        #send to float32 type, PyTorch supposedly likes it
        #phase_im = io.imread(self.phase_ims[idx], as_gray=True).astype('float32')
        
        dataset = Image.open(self.phase_ims[idx])
        h,w = np.shape(dataset)
        tiffarray = np.zeros((h,w,dataset.n_frames))
        for i in range(dataset.n_frames):
            dataset.seek(i)
            tiffarray[:,:,i] = np.array(dataset)
            tiffarray[:,:,i]=(tiffarray[:,:,i]-np.min(tiffarray[:,:,i]))/np.max(tiffarray[:,:,i]-np.min(tiffarray[:,:,i]))
        
        phase_im = tiffarray[:,:,0].astype('float32')
        path_to_save = self.output_folder + str(idx) + '.tiff'
        sz=np.array([h,w])
        if self.transform:
            phase_im = self.transform(phase_im)
            
        return {'im' : phase_im, 'imsize': sz, 'path_to_save' : path_to_save}