"""This module is used for analysis of FCS measurements. Currently, only camera-
based FCS is implemented.
"""
import os
import numpy as np
from tqdm import tqdm
from tifffile import imread
from matplotlib import pyplot as plt
from scipy.optimize import least_squares

class FcsCam:
    """Extracts FCS camera data from tiff stacks. Plotting and fitting tools too.
    """
    def __init__(self, data_path, save_name,
                 binning=12, delta_t=0.001, tot_time=10, normalize_fit=False):
        """Requires a data path containing replicates of a given experiment 
        (e.g. ~/data/exp1/).

        binning. Centered NxN binning can be specified by a single positive 
        integer. Custom ROI binning must be specified by 4 positive integers
        [row, col, height, width] where row and col is the index of the upper
        left (inclusive) pixel.
        """
        self.data_path = data_path
        self.delta_t = delta_t # seconds
        self.tot_time = tot_time # seconds
        self.normalize_fit = normalize_fit

        if np.isscalar(binning):
            self.custom_roi = False
            self.binning = np.array([-binning//2, binning-binning//2])
        else:
            self.custom_roi = True
            self.r_slice = np.s_[binning[0] : binning[0]+binning[2]]
            self.c_slice = np.s_[binning[1] : binning[1]+binning[3]]

        if os.path.exists(self.data_path + save_name + '.npz'):
            print("Save file already exists. Data loaded from there.")
            data = np.load(self.data_path + save_name + '.npz')
            self.acorrs = data['acorrs']
            self.tau = self.delta_t * np.arange(1, self.acorrs.shape[1]+1) #seconds
            self.fit_params = data['fit_params']
        else:
            self.acorrs, self.traces = self.tiff_to_acorr()
            self.tau = self.delta_t * np.arange(1, self.acorrs.shape[1]+1) #seconds
            self.keepers_ind = self.select_traces()
            self.acoors = self.acorrs[self.keepers_ind]
            self.traces = self.traces[self.keepers_ind]
            self.fit_params = self.fit_acorr()
            np.savez(self.data_path + save_name + '.npz',
                     acorrs=self.acorrs,
                     fit_params=self.fit_params)


    @staticmethod
    def autocorr(I_t):
        """Calculates the autocorrelation of time series data.

        G(τ) = <δI(t)δI(t+τ)> / <I(t)>**2

        where δI(t) = I(t) - <I(t)>
        """
        g2 = []
        for Δt in range(1,I_t.size-1):
            g2.append(np.cov(I_t[:-Δt], I_t[Δt:])[0,1])
        return g2 / np.mean(I_t)**2

    @staticmethod
    def autocorr2(I_t):
        """Identical to autocorr, but faster
        """
        g2 = []
        for Δt in range(1,I_t.size-1):
            g2.append(np.mean(I_t[:-Δt] * I_t[Δt:])
                      - np.mean(I_t[:-Δt]) * np.mean(I_t[Δt:]))
        return g2 / np.mean(I_t)**2

    def tiff_to_acorr(self):
        """Calculates the autocorrelation for a collection of tiff stacks.

        Saves all autocorrs to a npy file and returns a 2D numpy array where
        each replicate is a single row.
        """
        im_paths = [x[0]+'/'+files for x in os.walk(self.data_path) 
                for folders in x for files in folders if '.tif' in files]
        print(f"{len(im_paths)} replicates found")

        acorrs = np.empty((0, int(self.tot_time/self.delta_t)-2))
        traces = np.empty((0, int(self.tot_time/self.delta_t)))
        for im_path in tqdm(im_paths):
            image = imread(im_path)
            if not self.custom_roi:
                im_sh = np.array(image.shape[1:])
                self.r_slice = np.s_[self.binning[0]+im_sh[0]//2 :
                                     self.binning[1]+im_sh[0]//2]
                self.c_slice = np.s_[self.binning[0]+im_sh[1]//2 :
                                     self.binning[1]+im_sh[1]//2]
            roi = np.sum(image[:, self.r_slice, self.c_slice], axis=(1,2))
            traces = np.vstack((traces, roi))
            acorrs = np.vstack((acorrs, self.autocorr2(roi)))
        return acorrs, traces

    @staticmethod
    def fcs_model(theta, x):
        """Theoretical model of FCS for normal 3D diffusion

        G(τ) = 1/<N> * 1/(1 + τ/τ_D) * 1/√(1 + τ/(K**2 * τ_D))  +  G(∞)

        where

        <N> = the mean number of diffusers in the volume
        τ_D = characteristic residence time of the diffuser
        K = ω_z/ω_xy is the axial to radial ratio of (e**(-2)) radii of the confocal volume
        G(∞) = autocorrelation at long time scales
        """
        return 1/theta[0] * (1 + x/theta[1])**(-1) * (1 + x/(theta[2]**2*theta[1]))**(-1/2) + theta[3]

    def fit_acorr(self):
        """Performs least-squares fit. Data is averaged so that 10 data points
        are in each decade which weights data at small time scales more heavily.
        """
        mean_acorr = np.mean(self.acorrs, axis=0)
        if self.normalize_fit:
            mean_acorr /= mean_acorr[0]

        log_filter = 10**np.arange(1, np.log10(self.tot_time/self.delta_t).astype(int))
        start = 1
        mean_filtered = mean_acorr[:log_filter[start]]
        tau_filtered = self.tau[:log_filter[start]]
        for i in log_filter[start:]:
            for j in range(i,10*i,i):
                mean_filtered = np.append(mean_filtered, sum(mean_filtered[j:j+i])/i)
                tau_filtered = np.append(tau_filtered, j*self.delta_t)

        def fit_diff(theta):
            out = self.fcs_model(theta, tau_filtered[1:]) - mean_filtered[1:]
            return out[(np.isnan(out) == False) & (np.isinf(out) == False)]

        res = least_squares(fit_diff, x0=[0.9, 0.004, 2., 0.],
                            bounds=([0.5,0.001,0.9,-0.1],[5,0.5,3,0.2]))
        return res.x

    def plot_acorr(self, acorrs=None, mode=None, label=None, fig=plt.figure()):
        """General plotting function. Mean is always plotted. Modes can be 
        stacked, (e.g. "fit std").

        all - plots every acorr
        std - plots standard deviation as fill_between
        minmax - plots entire spread fill_between
        fit - plots fit
        """
        if acorrs is None:
            acorrs = self.acorrs
        ax = fig.gca()
        mean_acorr = np.mean(acorrs, axis=0)
        if self.normalize_fit:
            scale = mean_acorr[0]
            mean_acorr /= scale
            acorrs /= scale
        p = ax.plot(self.tau, mean_acorr, label=label)
        if 'all' in mode:
            for acorr in self.acorrs:
                ax.plot(self.tau, acorr, color=p[0].get_color(), alpha=0.5)
        if 'std' in mode:
            std_acorr = np.std(acorrs, axis=0)
            ax.fill_between(self.tau, mean_acorr+std_acorr, mean_acorr-std_acorr,
                            color=p[0].get_color(), alpha=0.25)
        if 'minmax' in mode:
            max_acorr = np.max(acorrs, axis=0)
            min_acorr = np.min(acorrs, axis=0)
            ax.fill_between(self.tau, max_acorr, min_acorr, color=p[0].get_color(),
                            alpha=0.25)
        if 'fit' in mode:
            ax.plot(self.tau, self.fcs_model(self.fit_params, self.tau), 'k--')

        ax.set_xscale('log')
        ax.set_ylabel(r'G($\tau$)')
        ax.set_xlabel(r'$\tau$ (seconds)')
        if label is not None:
            ax.legend()
        return fig

    def select_traces(self):
        """Wrapper function to choose traces
        """
        fig, axs = plt.subplots(1, 2)
        _TP = TracePicker(fig, axs, self.acorrs, self.traces, self.delta_t)
        fig.canvas.mpl_connect('key_press_event', _TP.on_press)
        plt.show()
        return _TP.keepers_ind

class TracePicker:
    """Plots autocorrelations and intensity traces for each replicate experiment.
    By default, all traces are 'selected'. Here are the 4 key presses:

    right/left arrow - proceed through replicate data
    space - 'unselect' replicate data
    k - undo 'unselect', keeps replicate data

    Repeat (un)selections are ignored
    """
    def __init__(self, fig, axs, acorrs, traces, delta_t):
        self.fig = fig
        self.ax1 = axs[0]
        self.ax2 = axs[1]
        self.acorrs = acorrs
        self.traces = traces
        self.ind = 0
        self.num_acorrs = self.acorrs.shape[0]
        self.tau = delta_t * np.arange(1, self.acorrs.shape[1]+1)
        self.time = delta_t * np.arange(self.traces.shape[1])
        self.keepers_ind = [i for i in range(self.num_acorrs)]
        self.acorr, = self.ax1.plot(self.tau, self.acorrs[self.ind,:])
        self.ax1.plot(self.tau, np.mean(self.acorrs, axis=0), 'k--', label='mean')
        self.ax1.set_title('Autocorrelation')
        self.ax1.set_xscale('log')
        self.ax1.legend()
        self.trace, = self.ax2.plot(self.time, self.traces[self.ind,:])
        self.ax2.set_title('Intensity Trace')
        self.ax2.set_ylim(np.min(traces), np.max(traces))
        self.update()

    def on_press(self, event):
        if event.key == 'left':
            if self.ind > 0:
                self.ind -= 1
        elif event.key == 'right':
            if self.ind < self.num_acorrs-1:
                self.ind += 1
        elif event.key == 'k':
            if self.ind not in self.keepers_ind:
                self.keepers_ind.append(self.ind)
                print(f"kept trace #{self.ind+1}")
        elif event.key == ' ':
            if self.ind in self.keepers_ind:
                self.keepers_ind.remove(self.ind)
                print(f"removed trace #{self.ind+1}")
        self.update()

    def update(self):
        self.acorr.set_data(self.tau, self.acorrs[self.ind,:])
        self.trace.set_data(self.time, self.traces[self.ind,:])
        self.fig.suptitle(f"trace # {self.ind+1} / {self.num_acorrs}")
        self.fig.canvas.draw()

def main():
    data_path = 'c:/Users/bobro/Documents/ShadowDev/Data/20210226/'
    fcs80 = FcsCam(data_path+'80percent/focused-spot-5ms/', save_name='acorrs_80percent', delta_t=0.005)#, normalize_fit=True)
    fcs90 = FcsCam(data_path+'90percent/focused-spot-5ms/', save_name='acorrs_90percent', delta_t=0.005)#, normalize_fit=True)

    fig1 = fcs80.plot_acorr(mode='all', label='80%')
    fig1 = fcs90.plot_acorr(mode='all', fig=fig1, label='90%')

    fig2 = plt.figure()
    fig2 = fcs80.plot_acorr(mode='fit', fig=fig2)
    fig2 = fcs90.plot_acorr(mode='fit', fig=fig2)

    plt.show()

if __name__ == '__main__':
    main()