"""
This module helps to analyze hidden Markov Models (HMMs).
"""
import numpy as np

class DiffusionModel:
    """
    Defines an HMM for a particle with multiple diffusion states.

    Attributes:
    num_states - number of discrete diffusion states, M
    diff_coefs - array-like param of length M that contains the diffusion
        coefficent of each state. [um]**2 [s]**-1
    trans_mat - (MxM) array-like param that contains the transition
        probabilities (p_ij) between the M diffusion states. p_ij is the
        probability of state i decaying in to state j. Rows must sum to 1.
    delta_t - timescale [s]

    stat_probs - the equilibrium probabilities of the transition matrix. These
        are taken to be the initial probabilities.
    """
    def __init__(self, num_states, diff_coefs=None, trans_mat=None, delta_t=25e-6):
        self.num_states = num_states
        self.diff_coefs = np.array(diff_coefs)
        self.trans_mat = np.array(trans_mat)
        self.delta_t = delta_t

        if trans_mat is not None:
            self.stat_probs = np.linalg.eig(self.trans_mat.T)[1][:, 0]
            self.stat_probs /= np.sum(self.stat_probs)
        else:
            self.stat_probs = None

    def viterbi(self, traj):
        """
        Returns the most probable sequence of latent variables, Zi.

        Iteratively computes the diffusion state that maximizes the probability
        beginning with state 1.

        mu_1 = max( p(Z1) * p(X1|Z1) ) for all M states of Z1
        mu_2 = max( mu_1 * p(Z2|Z1) * p(X2|Z2)) for all M states of Z2
        .
        .
        .
        mu_T = max( mu_1 * mu_2 *...* p(ZT|Z(T-1)) * p(XT|ZT) ) for all M states of ZT

        where

        p(Z1) are chosen from equilibrium (stationary probabilities)
        p(Xi|Zi) are the emission probabilities defined by diffusion,
            p(Xi|Zi) ~ N(0, 6*diff_coef*delta_t)
        p(Z(i+1)|Zi) are the transition probabilities, i.e. the ith row of the
            transition matrix, trans_mat[i, :]

        The log probabilities are calculated to avoid under/overflow issues

        Input:
        traj - (T+1,3) array-like that is a sequence of xyz positions [um]

        Output:
        z_star - (T,1) array-like that is the most probable latent variable seq.
        """
        trans_mat = self.trans_mat
        stat_probs = self.stat_probs
        steps_r = np.sqrt(np.sum(np.diff(traj, axis=0)**2, axis=1))

        mu_z = np.log(stat_probs) + self.log_emission(steps_r[0])
        lg_mu = np.atleast_1d(np.max(mu_z))
        z_star = np.atleast_1d(np.argmax(mu_z))
        for i in range(1, len(steps_r)):
            mu_z = (lg_mu[i-1] + np.log(trans_mat[z_star[i-1], :])
                    + self.log_emission(steps_r[i]))
            lg_mu = np.append(lg_mu, np.max(mu_z))
            z_star = np.append(z_star, np.argmax(mu_z))
        return z_star

    def forward_backward(self, traj):
        """
        Calculates 4 helpful probabilities:

        P(X1:t|Zt) - prob of seeing the observed sequence (X1, X2, ..., Xt) for
            a given latent state Zt.
            alpha[t] = alpha[t-1].trans_mat * emission
        P(Xt+1:T|Zt) - prob of seeing the observed sequence (X(t+1), X(t+2),
            ..., XT) for a given latent state Zt
            beta[t] = trans_mat.(beta[t+1] * emission)
        P(Zt|X1:T) - prob of being in the latent state Zt given the entire
            observed sequence of Xi's (X1, X2, ..., XT)
        P(Zt & Zt+1|X1:T) - prob of seeing latent state Zi followed by Zt+1
            given the entire observed sequence of Xi's (X1, X2, ..., XT)

        The first is known as the forward parameter, alpha, and the second as
        the backward parameter, beta. The third and fourth are gamma and xi.

        We use logs again to avoid under/overflow. The log-sum-exp trick is used
        because the log and sum operators cannot be exchanged as the log and max
        operators were exchanged in the Viterbi algorithm.

        A = log(a) = log(sum(a_i)) = log(sum(exp(log(a_i))))
          = max(a_i) - log(sum(exp(log(a_i - max(a_i)))))

        Input:
        traj - (T+1,3) array-like that is a sequence of xyz positions [um]

        Output:
        lg_alpha - (T,M) array of log(P(X1:t|Zt))
        lg_beta  - (T,M) array of log(P(Xt+1:T|Zt))
        lg_gamma - (T,M) array of log(P(Zt|X1:T))
        lg_xi    - (T,M,M) array of log(P(Zt & Zt+1|X1:T))
        """
        lg_tm = np.log(self.trans_mat)
        lg_sp = np.log(self.stat_probs)
        steps_r = np.sqrt(np.sum(np.diff(traj, axis=0)**2, axis=1))

        M = lg_tm.shape[0]
        T = len(steps_r)

        lg_alpha = np.zeros((T, M))
        lg_alpha[0, :] = lg_sp + self.log_emission(steps_r[0])
        for t in range(1, T):
            lg_em = self.log_emission(steps_r[t])
            for j in range(M):
                a = np.zeros(lg_em.shape)
                # alpha[t, j] = alpha[t-1].dot(trans_mat[:,j]) * emission[j]
                for i in range(M):
                    a[i] = lg_alpha[t-1, i] + lg_tm[i, j] + lg_em[j]
                amax = np.max(a)
                lg_alpha[t, j] = amax - np.log(np.sum(np.exp(a - amax)))

        lg_beta = np.zeros(lg_alpha.size)
        for t in range(T-2, -1, -1):
            lg_em = self.log_emission(steps_r[t+1])
            for j in range(M):
                b = np.zeros(lg_em.shape)
                # beta[t, j] = (beta[t+1] * emission).dot(trans_mat[j, :])
                for i in range(M):
                    b[i] = lg_beta[t+1, j] + lg_tm[j, i] + lg_em[j]
                bmax = np.max(b)
                lg_beta[t, j] = bmax - np.log(np.sum(np.exp(b - bmax)))

        # TODO calculate logs of gamma and xi
        alpha = np.exp(lg_alpha)
        beta = np.exp(lg_beta)
        xi = np.zeros((M, M, T - 1))
        for t in range(T - 1):
            emission = np.exp(self.log_emission(steps_r[t+1]))
            denominator = np.dot(np.dot(alpha[t, :].T, self.trans_mat) * emission.T, beta[t + 1, :])
            for i in range(M):
                numerator = alpha[t, i] * a[i, :] * emission.T * beta[t + 1, :].T
                xi[i, :, t] = numerator / denominator

        gamma = np.sum(xi, axis=1)

        return lg_alpha, lg_beta, gamma, xi


    def baum_welch(self, traj, n_iter=100):
        steps_r = np.sqrt(np.sum(np.diff(traj, axis=0)**2, axis=1))
        M = self.trans_mat.shape[0]
        T = len(steps_r)
        for _ in range(n_iter):
            _, _, gamma, xi = self.forward_backward(traj)
            self.trans_mat = np.sum(xi, axis=2) / np.sum(gamma, axis=1).reshape((-1, 1))
            self.stat_probs = np.linalg.eig(self.trans_mat.T)[1][:, 0]
            self.stat_probs /= np.sum(self.stat_probs)

            gamma = np.hstack((gamma, np.sum(xi[:, :, T-2], axis=0).reshape((-1, 1))))
            # TODO replace loop with vectorized version; calculate means
            val = np.zeros(gamma.shape[0])
            var = []
            for i in range(gamma.shape[1]):
                val = gamma[:, i]*steps_r**2 # force means = 0
                var.append(np.sum(val)/np.sum(gamma[:, i]))

            self.diff_coefs = var/(6*self.delta_t)


    def log_emission(self, X_i):
        """
        Calculates the log emission probabilities, log(p(X_i|Z_i)), of an observed
        displacement X_i from all possible latent states Z_i as defined by
        diff_coefs

        Input:
        X_i - observed displacement at ith step

        Output:
        lg_em - (M,) array of emission log probabilities
        """
        var = 6*self.diff_coefs*self.delta_t
        return -(X_i)**2/(2*var) - 0.5*np.log(2*np.pi*var)

def main():
    from Physics_Lib import Particle

    diff_coeffs = (1, 0.2)
    trans_matrix = ((0.99, 0.01),
                    (0.02, 0.98))
    # diff_coeffs = (1, 0.2, 0.01)
    # trans_matrix = ((0.98, 0.01, 0.01),
    #                 (0.02, 0.97, 0.01),
    #                 (0.03, 0.03, 0.94))

    p1 = Particle((0, 0, 0), time_steps=100)
    p1.make_hmm_diffusion(diff_coeffs, trans_matrix)

    hmm = DiffusionModel(2, diff_coeffs, trans_matrix)
    hmm.viterbi(p1.hmm_traj)

    # import numpy as np
    # from matplotlib import pyplot as plt
    # from hmmlearn import hmm
    # from time import perf_counter as get_time
    # from Physics_Lib import Particle

    # t0 = get_time()
    # diff_coeffs = (1, 0.01)
    # trans_matrix = ((0.99, 0.01),
    #                 (0.02, 0.98))

    # n_trajs = 1000 # takes about 4 minutes to run for n_trajs=1000
    # p = Particle((0, 0, 0))
    # X = np.atleast_1d()
    # Z = np.zeros((n_trajs, len(p.time)-1))
    # for i in range(n_trajs):
    #     p.make_hmm_diffusion(diff_coeffs, trans_matrix)
    #     X = np.append(X, np.sqrt(np.sum(np.diff(p.hmm_traj, axis=0)**2, axis=1)))
    #     Z[i] = p.latent_states

    # model = hmm.GaussianHMM(n_components=2, covariance_type="spherical", n_iter=100)
    # # model.covars_ = np.array(diff_coeffs)*6*p.delta_t
    # # model.transmat_ = trans_matrix
    # stat_probs = np.linalg.eig(np.array(trans_matrix).T)[1][:, 0]
    # # model.startprob_ =  stat_probs/np.sum(stat_probs)
    # # model.means_ = np.zeros(2)
    # model.monitor_.tol = 0.01

    # X = np.expand_dims(X, axis=1)
    # Xlens = np.array(Z.shape[1]*np.ones(n_trajs), dtype=int)
    # model.fit(X, lengths=Xlens)
    # print("Converged?", model.monitor_.converged)
    # print("Total time:", get_time()-t0)
    # print("actual diffusion coeffs:\n", diff_coeffs)
    # print("predicted diffusion coeffs:\n", model.covars_.squeeze()/p.delta_t/6) # bad prediction
    # print("actual transition matrix:\n", trans_matrix)
    # print("predicted transition matrix:\n", model.transmat_)
    # print("actual initial probs:\n", stat_probs/np.sum(stat_probs))
    # print("predicted initial probs:\n", model.startprob_)

    # Z2 = model.predict(X[:Z.shape[1]])
    # _, (ax1, ax2) = plt.subplots(1, 2)
    # ax1.plot(range(len(Z[0])), Z[0])
    # ax1.set_title('generated latent state')
    # ax2.plot(range(len(Z2)), Z2)
    # ax2.set_title('modeled latent state')
    # plt.show()

if __name__ == '__main__':
    main()
