import numpy as np
from scipy import signal

class miscellaneous:
    
    def vfactorial(v):
        """
        Calculates v! factorial
        """
        vv=np.zeros(len(v))
        for i in range(len(v)):
            if v[i]!=0:
                vv[i]=np.prod(np.uint64(np.array(range(int(v[i]))))+1)
            else:
                vv[i]=1
        return vv

    def DataCrop(MaxTrajTime,t_raw, args): 
        # Option to truncate and only analyse shorter parts of a tajectory 
        # Input:
        #   MaxTrajTime: max length in seconds
        #   t_raw      : time points 
        #   args       : a list with vectors that needs to be trimed, length need to be equal to t_raw
        #
        #Output:
        #   list with the data and time  

        if MaxTrajTime>0:
            t_interval_low=t_raw[0]
            t_interval_high=t_raw[0]+MaxTrajTime
            t_valid=np.where(np.logical_and(t_raw>=t_interval_low, t_raw<=t_interval_high))[0]
            
            t_raw=t_raw[t_valid]
            
            out=[]
            for i in range(len(args)):
                out=out+[args[i][t_valid]]

        return t_raw, out 

    def fft_processing(D, t):
        fft_D = np.fft.fft(D)
        T = t[1] - t[0]  # sampling interval 
        N = D.shape[0]

        # 1/T = frequency
        f = np.linspace(0, 1 / T, N)

        # Amplitude spectra
        fft_D_abs = np.abs(fft_D)[:N // 2] * 1 / N
        f=f[:N // 2]

        return fft_D_abs, f

    def autocorrelation(x,t):
        xp = (x - np.mean(x))/np.std(x)
        D_corr=np.correlate(xp,xp,mode='full')/xp.size
        
        N = xp.shape[0]
        T = t[1]-t[0]        #step size
        M=D_corr.shape[0]

        tau = np.linspace(0, T*M, M)
        tau=tau-tau[N-1]

        D_corr= D_corr[N-1:] 
        tau   = tau[N-1:]

        return D_corr, tau

    def autocorrelation2(x,t):
        xp = np.fft.ifftshift((x - np.average(x))/np.std(x))
        n, = xp.shape
        xp = np.r_[xp[:n//2], np.zeros_like(xp), xp[n//2:]]
        f = np.fft.fft(xp)
        p = np.absolute(f)**2
        pi = np.fft.ifft(p)
        Corr=np.real(pi)[:n//2]/(np.arange(n//2)[::-1]+n//2)

        Dtau=t[1]-t[0]
        tau = np.linspace(0, Dtau*Corr.size, Corr.size)
        
        return Corr, tau

    def conv_3D(IM3D, kernel):
        """
        s1 = np.array(IM3D.shape)
        s2 = np.array(kernel.shape)
        size = s1 
        fsize = 2 ** np.ceil(np.log2(size)).astype(int)
        fslice = tuple([slice(0, int(sz)) for sz in size])
        fft_IM3D = np.fft.fftn(IM3D , fsize)
        fft_kernel = np.fft.fftn(kernel , fsize)
        conv2=np.abs(np.fft.ifftn(fft_IM3D*fft_kernel)[fslice].copy())
        """
        conv=signal.fftconvolve(IM3D, kernel, mode='same')        
        return conv

    def SobelFeldman3DKernal():
        Hx=np.zeros((3,3,3))
        Hx[:,:,0]=[[1,0,-1],[2,0,-2],[1,0,-1]]
        Hx[:,:,1]=2*Hx[:,:,0]
        Hx[:,:,2]=Hx[:,:,0]

        Hy=np.zeros((3,3,3))
        Hy[:,:,0]=[[1,2,1],[0,0,0],[-1,-2,-1]]
        Hy[:,:,1]=2*Hy[:,:,0]
        Hy[:,:,2]=Hy[:,:,0]

        Hz=np.zeros((3,3,3))
        Hz[:,:,0]=[[1,2,1],[2,4,2],[1,2,1]]
        Hz[:,:,2]=-Hz[:,:,0]
        return Hx,Hy,Hz

    def SobelFeldman2DKernal():
        Hx=[[1,0,-1],[2,0,-2],[1,0,-1]]    
        Hy=[[1,2,1],[0,0,0],[-1,-2,-1]]

        return Hx,Hy

    def get_traj_raw_data3(Trajs, TrajToGet, ScanMatrix, ScanProject):
        """
        Function to extract the raw data of a trajectory
        
        INPUT:
            Trajs:         Matrix with [total time, start, end], start and end are refering to rows in ScanMatrix    
            TrajToGet:     which trajectory to get
            ScanMatrix:    Matrix containing all raw data 

        OUTPUT:       
                T: Time estimated acording to the mode  
                XP: [um] X value piezo center 
                YP: [um] Y value piezo center 
                XE: [um] X value EOD center 
                YE: [um] Y value EOD center 
                XO: [um] X value EOD tracking pattern
                YO: [um] Y value EOD tracking pattern
                ZI: [index] z-value 
                APD1: Photon counts H
                APD2: Photon counts V
                J:  Time Jump indicator 
                MemIndex: Memory index
        """

        n=TrajToGet
        
        interval=range(int(Trajs[1,n]), int(Trajs[2,n]+1))
            
        T=ScanMatrix[0, interval]        #[s] Time
        XP=ScanMatrix[9, interval]       #[um] X value piezo center 
        YP=ScanMatrix[10, interval]      #[um] Y value piezo center 
        XE=ScanMatrix[23, interval]      #[um] X value EOD center 
        YE=ScanMatrix[24, interval]      #[um] Y value EOD center
        XO=ScanMatrix[11, interval]      #[um] X value EOD tracking pattern
        YO=ScanMatrix[12, interval]      #[um] Y value EOD tracking pattern
        
        ZI_OffsetIndex=ScanMatrix[19, interval] + ScanProject.PosCalibration['zOffsetIndex'] #+- offset index, now from 0 to offset (suposse to be 1 if usign 3 levels)
        ZI_Raw=ScanMatrix[20, interval] #z raw data from FPGA same as 21
        ZI_um =ZI_Raw*ScanProject.PosCalibration['PositionLength']/(ScanProject.PosCalibration['zScale'] * 2**14 )    #Converted to um, it is z raw data from FPGA times comvertion factor, if this works then we should move this to a better place  
        ZI=np.round(ScanMatrix[22, interval]/ScanProject.PosCalibration['SLMSwitchHight']).astype('int')     #[index] z-value, it is 19+20 but passed through filters  

        J=ScanMatrix[25, interval]     #Time Jump indicator 
        APD1 = ScanMatrix[3, interval] #Photon counts H
        APD2 = ScanMatrix[4, interval] #Photon counts V
        MemIndex= ScanMatrix[8, interval] #Memory index


        if 0:
            import matplotlib.pyplot as plt

            plt.plot(ScanMatrix[0,:], np.sum(ScanMatrix[3:5,:],axis=0))
            plt.show()

            plt.plot(ScanMatrix[0,:], ScanMatrix[9,:],'r')
            plt.plot(ScanMatrix[0,:], ScanMatrix[10,:],'b')
            plt.show()
        
        return T,XP,YP,XE,YE,XO,YO,ZI,APD1,APD2,J,MemIndex,ZI_um,ZI_OffsetIndex

    def get_traj_raw_data4(Trajs, TrajToGet, ScanMatrix, ScanProject):
        """
        Function to extract the raw data of a trajectory
        New from 3, this includes extraction of zcam data
        
        INPUT:
            Trajs:         Matrix with [total time, start, end], start and end are refering to rows in ScanMatrix    
            TrajToGet:     which trajectory to get
            ScanMatrix:    Matrix containing all raw data 

        OUTPUT:       
                T: Time estimated acording to the mode  
                XP: [um] X value piezo center 
                YP: [um] Y value piezo center 
                XE: [um] X value EOD center 
                YE: [um] Y value EOD center 
                XO: [um] X value EOD tracking pattern
                YO: [um] Y value EOD tracking pattern
                ZI: [index] z-value 
                APD1: Photon counts H
                APD2: Photon counts V
                J:  Time Jump indicator 
                MemIndex: Memory index
                zCam: [um] z value from z cam 
        """

        n=TrajToGet
        
        interval=range(int(Trajs[1,n]), int(Trajs[2,n]+1))
            
        T=ScanMatrix[0, interval]        #[s] Time
        XP=ScanMatrix[9, interval]       #[um] X value piezo center 
        YP=ScanMatrix[10, interval]      #[um] Y value piezo center 
        XE=ScanMatrix[23, interval]      #[um] X value EOD center 
        YE=ScanMatrix[24, interval]      #[um] Y value EOD center
        XO=ScanMatrix[11, interval]      #[um] X value EOD tracking pattern
        YO=ScanMatrix[12, interval]      #[um] Y value EOD tracking pattern
        try: 
            zCam=ScanMatrix[26, interval]    #[um] zCam value 
        except:
            zCam=np.zeros(len(interval))
            print('No zCam data found, replacing with: zeros')

        ZI_OffsetIndex=ScanMatrix[19, interval] + ScanProject.PosCalibration['zOffsetIndex'] #+- offset index, now from 0 to offset (suposse to be 1 if usign 3 levels)
        ZI_Raw=ScanMatrix[20, interval] #z raw data from FPGA same as 21
        ZI_um =ZI_Raw*ScanProject.PosCalibration['PositionLength']/(ScanProject.PosCalibration['zScale'] * 2**14 )    #Converted to um, it is z raw data from FPGA times comvertion factor, if this works then we should move this to a better place  
        ZI=np.round(ScanMatrix[22, interval]/ScanProject.PosCalibration['SLMSwitchHight']).astype('int')     #[index] z-value, it is 19+20 but passed through filters  

        J=ScanMatrix[25, interval]     #Time Jump indicator 
        APD1 = ScanMatrix[3, interval] #Photon counts H
        APD2 = ScanMatrix[4, interval] #Photon counts V
        MemIndex= ScanMatrix[8, interval] #Memory index
        
        return T,XP,YP,XE,YE,XO,YO,ZI,APD1,APD2,J,MemIndex,ZI_um,ZI_OffsetIndex,zCam

    def get_LR_pos(xedges):
            #x- L side
        if  len(np.where(np.diff(xedges)==1)[0])>0:
            xL=np.where(np.diff(xedges)==1)[0][0]+1
        else:
            xL=0
            #x- R side
        if len(np.where(np.diff(xedges)==-1)[0])>0:
            xR=np.where(np.diff(xedges)==-1)[0][0]+1
        else:
            xR=len(xedges)
        return xL,xR

    def Get_CRB(P,xx,Hx,n):
        # See notes: "On Balzarotti Science 2017 ", section: "Numerical apporach for 3D analysis" , regarding the derivation of the formulas used, 
        PHx=np.zeros(P.shape)
        PHxx=np.zeros(P.shape)
        Dxx=miscellaneous.conv_3D(xx*1e-6, Hx) #[m]
        for i in range(P.shape[3]):
            PHx[:,:,:,i]=miscellaneous.conv_3D(P[:,:,:,i], Hx)/Dxx
            PHxx[:,:,:,i]=miscellaneous.conv_3D(PHx[:,:,:,i], Hx)/Dxx
        
        CRBx=np.zeros(P.shape[:-1])
        for i in range(P.shape[3]):
            CRBx=CRBx + n[i]*(P[:,:,:,i]**-2)*(PHx[:,:,:,i]**2-P[:,:,:,i]*PHxx[:,:,:,i])

        CRBx=1/CRBx #[m^2]
        #Put negative numbers to 0, we interpret negative number as convex curvature
        CRBx[np.where(CRBx<0)]=np.nan
        SqCRBx=np.sqrt(CRBx)*1e9 #[nm]
        return SqCRBx 

    def get_Spot_coordinates(im_yxz,BlankingRadius=6,spot_num = 3,crop_margin=0,delta_z=0.05):
        # find the brightest voxels and plot z-axis profile
        radius = BlankingRadius # pixels
        z = np.arange(im_yxz.shape[2])*delta_z # convert to um
        z_profiles = np.zeros((im_yxz.shape[2],spot_num))
    
        Spot_coordinates_xyz=np.zeros((spot_num,3))
        for i in range(spot_num):
            ymax, xmax, zmax = np.unravel_index(np.argmax(im_yxz), im_yxz.shape)
            Spot_coordinates_xyz[i,:]=[xmax, ymax, zmax]
            z_profiles[:,i] = im_yxz[ymax,xmax,:]
            im_yxz[ymax-radius:ymax+radius, xmax-radius:xmax+radius,:] = 0

        # Rearange spots according to z axis and remove frames not needed
        Spot_coordinates_xyz_zsorted=Spot_coordinates_xyz[np.argsort(Spot_coordinates_xyz[:,2]),:]
        z_profiles=z_profiles[:,np.argsort(Spot_coordinates_xyz[:,2])]
        Spot_coordinates_xyz[:,2]=Spot_coordinates_xyz[:,2]-(Spot_coordinates_xyz_zsorted[0,2]-crop_margin)
    
        Spot_coordinates_xyz_zsorted_comp=Spot_coordinates_xyz_zsorted.copy()
        Spot_coordinates_xyz_zsorted_comp[:,2]= Spot_coordinates_xyz_zsorted_comp[:,2]-(Spot_coordinates_xyz_zsorted[0,2]-crop_margin)
        return Spot_coordinates_xyz_zsorted_comp, Spot_coordinates_xyz_zsorted, z_profiles,z

    def get_Spot_coordinates2(im_yxz,BlankingRadius=6,spot_num = 3,crop_margin=0,delta_z=0.05):
        # find the brightest voxels and plot z-axis profile
        radius = BlankingRadius # pixels
        z = np.arange(im_yxz.shape[2])*delta_z # convert to um
        z_profiles = np.zeros((im_yxz.shape[2],spot_num))
    
        Spot_coordinates_xyz=np.zeros((spot_num,3))
        for i in range(spot_num):
            ymax, xmax, zmax = np.unravel_index(np.argmax(im_yxz), im_yxz.shape)
            Spot_coordinates_xyz[i,:]=[xmax, ymax, zmax]
            z_profiles[:,i] = im_yxz[ymax,xmax,:]

            yxzBoundLow=np.array([ymax, xmax, zmax])-radius
            yxzBoundHigh=np.array([ymax, xmax, zmax])+radius
            yxzBoundLow=yxzBoundLow*(yxzBoundLow>=0)
            yxzBoundHigh=yxzBoundHigh*(yxzBoundHigh<=im_yxz.shape) + im_yxz.shape*(yxzBoundHigh>im_yxz.shape)
            
            im_yxz[yxzBoundLow[0]:yxzBoundHigh[0], yxzBoundLow[1]:yxzBoundHigh[1],yxzBoundLow[2]:yxzBoundHigh[2]] = 0

        # Rearange spots according to z axis and remove frames not needed
        Spot_coordinates_xyz_zsorted=Spot_coordinates_xyz[np.argsort(Spot_coordinates_xyz[:,2]),:]
        z_profiles=z_profiles[:,np.argsort(Spot_coordinates_xyz[:,2])]
        Spot_coordinates_xyz[:,2]=Spot_coordinates_xyz[:,2]-(Spot_coordinates_xyz_zsorted[0,2]-crop_margin)
    
        Spot_coordinates_xyz_zsorted_comp=Spot_coordinates_xyz_zsorted.copy()
        Spot_coordinates_xyz_zsorted_comp[:,2]= Spot_coordinates_xyz_zsorted_comp[:,2]-(Spot_coordinates_xyz_zsorted[0,2]-crop_margin)
        return Spot_coordinates_xyz_zsorted_comp, Spot_coordinates_xyz_zsorted, z_profiles,z

    def get_Scan_toImagCoordinates(ScanProject, txyz_Traj):
        """
        Convert a data set matrix with rows t,x,y,z from scan coordinated to camera image coordinates
        """
        CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]

        # Get the transformation and get the Affine transform invers see: https://en.wikipedia.org/wiki/Affine_transformation
        try:
            Tran=ScanProject.ImgOverlayTransform.copy()
        except:
            print('No Scan to Cam transformation found, identity transform will be used')
            Tran=np.eye(3)

        Tran[0:2,0:2]=np.linalg.inv(Tran[0:2,0:2])
        Tran[0:2,2]=-np.matmul(Tran[0:2,0:2],Tran[0:2,2])
        
        Traj_xyz=txyz_Traj[1:4,:]
        Traj_xyz=Traj_xyz/CamRes #move to pixel coordinates 
        Traj_xyz[2,:]=1          #Remove z and change it to a 1 for the Affine transromation matrix
        Traj_xyz= np.matmul(Tran,Traj_xyz) #Apply transformation 
        Traj_xyz= Traj_xyz*CamRes  #[um] go back to micrometers
        txyz_Traj[1:3,:]=Traj_xyz[:2]  #Combine with time and z and photon counts
        
        return txyz_Traj



def main():
    a=np.array([1,2,3,4,5])
    b=np.array([2,3,4,5,6])
    c=np.array([20,30,40,50,60])
    d=np.array([200,300,400,500,600])
    args=[b,c,d]

    t_raw, out=miscellaneous.DataCrop(3,a, args)
    print("Done!")


if __name__ == '__main__':
    main()