from os import walk
import fnmatch
import numpy as np
from skimage import io as _skiIO
from skimage import transform as _skiTran
from matplotlib import pyplot as plt


class ZCamHandler:
    def getZref(path_ref, file_prefix, N, scale=10):
        """
        Load ref data and sum over the rows, add all of the ref profiles into one image
        Interpolate between rows to get higher spatial resolution
        output the rescaled zmap image 
        """

        for i in range(N): 
            im = _skiIO.imread(path_ref + file_prefix + str(i) + '.tiff')
            if i==0:
                Zmap=np.zeros((N,im.shape[1]))            
            Zmap[i,:]=np.sum(im,axis=0)
        Zmap_Scale= _skiTran.resize(Zmap,(Zmap.shape[0]*scale,Zmap.shape[1]),order=3) #0: Nearest-neighbor 1: Bi-linear (default) 2: Bi-quadratic 3: Bi-cubic 4: Bi-quartic 5: Bi-quintic
    
        if 0: #Debug plot
            plt.imshow(Zmap_Scale)
            plt.show()
    
        return Zmap_Scale

    def getZum(Zmap,zpix2um,path,file_prefix):
        """
        get the z position in um from a profile compared to the Zmap
        The method used is to find the minima of the square distance: min(Z-Zref)^2

        """
        filenames = next(walk(path), (None, None, []))[2]  # [] if no file
        filenames = fnmatch.filter(filenames, '*.tiff')
        N=len(filenames)
        Zum=np.zeros(N)
        ZProfiles=np.zeros((N,Zmap.shape[0]))
        for i in range(N):
            print(str(i) + '/' + str(N))
            im = _skiIO.imread(path + file_prefix + str(i) + '.tiff')
            P=np.repeat(np.expand_dims(np.sum(im,axis=0),axis=0),Zmap.shape[0],axis=0)
            Profile= np.sum((Zmap-P)**2,axis=1)
            ZProfiles[i,:]=Profile
            PMin=np.min(Profile)
            PMinArg=np.argmin(Profile)
            Zum[i]=PMinArg
            if 0:
                Profile=Profile-PMin
                plt.plot(Profile)
                plt.plot(PMinArg,0,'r*')
                plt.show() 
        if 1:
            plt.imshow(ZProfiles)
            plt.plot(Zum,np.arange(N),'-r')
            plt.savefig(path + 'zProfile.png')
            plt.close()
            print('zProfile image saved!')
            #plt.show()

        return Zum*zpix2um

    def addToScanProject(ScanProject,ZCam_um):
        """
        Integrates the zCam data into the ScanMatrix

        """
        MemIndexRef=ScanProject.zCam['MemIndex']
        Cycles=ScanProject.zCam['Cycles']
        Tracking=ScanProject.ScanMatrix[5,:]
        MemIndex=ScanProject.ScanMatrix[8,:]
        
        Starts=np.where(np.diff(Tracking)>0)[0]+1
        Stops=np.where(np.diff(Tracking)<0)[0]+1
        
        ZList=np.zeros(Tracking.shape)
        index=0
        indexMax=len(ZCam_um)-1
        for i in range(len(Starts)):
            TM=(MemIndex[Starts[i]:Stops[i]]==MemIndexRef)
            TMFind=np.where(TM)[0]
            #TM_Index0=np.where(TM)[0][0]
            CC=np.sum(TM)//Cycles
            for C in range(CC+1):
                if C==0:
                    if CC>0:
                        ZList[Starts[i]:Starts[i]+TMFind[(C+1)*Cycles]]=ZCam_um[index]
                    elif CC==0:
                        ZList[Starts[i]:Stops[i]]=ZCam_um[index]
                elif (C<CC-1)*(C>0):
                    ZList[Starts[i]+TMFind[C*Cycles]:Starts[i]+TMFind[(C+1)*Cycles]]=ZCam_um[index]
                elif C==CC-1:
                    ZList[Starts[i]+TMFind[C*Cycles]:Stops[i]]=ZCam_um[index]
                if index<indexMax:
                    index+=1

        if len(ScanProject.ScanInfo)==26:
            ScanProject.ScanMatrix=np.vstack((ScanProject.ScanMatrix,ZList))
            ScanProject.ScanInfo.append('26. ZCam data in [um]')
        else:
            print("Trying to add zCam data. ScanMatrix size is not correct. Data will not be added")

        return ScanProject
  
def MainTester():
    path_ref="C:\\Users\\elias\\Desktop\\Shadowfax\\Data\\20210707_EXP-21-BL9409_(tRNA)\\zCam\\ref\\"
    file_prefix="frame_"
    N=80
    StepSize=50 #nm
    scale=10 
    zpix2um=(StepSize*1e-3 /scale) #um/y-pix

    Zmap=ZCamHandler.getZref(path_ref,file_prefix,N,scale=scale)
    path="C:\\Users\\elias\\Desktop\\Shadowfax\\Data\\20210707_EXP-21-BL9409_(tRNA)\\zCam\\Sample3\\1\\"
    file_prefix="frame_"

    ZCam_um=ZCamHandler.getZum(Zmap,zpix2um,path,file_prefix)
    print("Done!")
    
if __name__=='__main__':
    MainTester()