from pathlib import Path
import matplotlib
#matplotlib.use('Agg')
#matplotlib.use('Qt4Agg')
#matplotlib.interactive(True)
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import matplotlib.animation as animation
import matplotlib.cm as cm
from matplotlib.collections import LineCollection
from matplotlib.ticker import (MultipleLocator, FormatStrFormatter, AutoMinorLocator)
from mpl_toolkits.mplot3d.art3d import Line3DCollection

from mpl_toolkits.mplot3d import Axes3D
import mpl_toolkits.mplot3d.axes3d as p3

import numpy as np
from itertools import product, combinations
from PIL import Image
from skimage import transform as _ski_tran
from skimage.feature import (blob_log, match_template, ORB, plot_matches, match_descriptors) 
from skimage.measure import ransac
import cv2 as _cv2
import scipy.stats as stats    

from miscellaneous import miscellaneous as _MS
from QTgraphPlotLib import QTgraphPlotLib as QTplt

from scipy.io.wavfile import write as _waveWrite


class RawDataPlot:
    """ 
        This class is used for ploting raw scan data. 
        It functions as follows: 
        1. Initiate the class, this creates matplotlib figure with corresponding axis
        2. Run the filter fnction on a ScanObject, this will create an internal pass/fail vector related to the scanmatrix
        3. Run the plot function that will use the internal pass/fail vektor on the ScanObject to plot the correct data
    """

    def __init__(self):
        """
        Creats the matplot lib figure and axis that will be the end result
        """

        self.raw_fig= plt.figure()
        self.raw_ax = self.raw_fig.add_subplot(111, projection='3d')

    def scatter(self, x_vals, y_vals, z_vals, color, filter):
        """
        Keyword arguments:
        x_vals: 1-D vector containing scan x-values  
        y_vals: 1-D vector containing scan y-values
        z_vals: 1-D vector containing scan z-values or time values 
        color: 1-D vector containing information incoded in color 
        """
        x_vals=x_vals[filter==1]
        y_vals=y_vals[filter==1]
        z_vals=z_vals[filter==1]
        color=color[filter==1]
        
        
        #Configure plot 
        #colors = np.log(1 + (10*color/np.max(color)) )
        colors = color/np.max(color)

        colors = colors/np.max(colors)
        colors = np.array([colors])

        rgba_colors = np.transpose(colors)
        rgba_colors = np.concatenate((rgba_colors,rgba_colors,rgba_colors,rgba_colors),axis=1)
        
        #cmap = plt.cm.get_cmap("inferno") # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
        size=5

        # Plot the values
        #self.raw_ax.scatter(x_vals, y_vals, z_vals, s=size, c=colors, cmap=cmap ,  marker='.', alpha=alpha)
        self.raw_ax.scatter(x_vals, y_vals, z_vals, s=size, c=rgba_colors, marker='o')

    def configure_labels(self):
        self.raw_ax.set_xlabel('X-axis')
        self.raw_ax.set_ylabel('Y-axis')
        self.raw_ax.set_zlabel('Z-axis')
        self.raw_ax.set_title('Raw data')

    def show(self):
        plt.show()

    def plot_raw_data(self, ScanProject):
        ######### Raw data plot    

        ### New filter settings for ploting
        ScanProject.DataSelectionRules['APDThreshold']= [0, 0]   # First threshold for each APD, 
        ScanProject.DataSelectionRules['APDCountRange']  = [5, np.inf] # Second threshold for the sum APD_1+APD_2
        ScanProject.DataSelectionRules['APDCountRange_ErrorBound'] =[0, np.inf] # Select data within the error bound  (LowerErrorBound < UpperErrorBound)
        ScanProject.DataSelectionRules['xBounds'] =[-1, 21]       #[um] 
        ScanProject.DataSelectionRules['yBounds'] =[-1, 21]       #[um] 
        ScanProject.DataSelectionRules['TBounds'] =[0, np.inf]    #[s]
        ScanProject.get_raw_data_filter()  # update filter vector, Calculations are stored in: ScanProject.filter_vector


        APD=np.sum(ScanProject.ScanMatrix[3:4,:], axis=0)

        self.scatter(ScanProject.ScanMatrix[1,:], ScanProject.ScanMatrix[2,:], ScanProject.ScanMatrix[0,:], APD, ScanProject.filter_vector)
        self.show()

class TrajPlots:
    def __init__(self):
        """
        Creats the matplot lib figure and axis that will be the end result
        """
        #self.traj_fig= plt.figure()
        #self.traj_ax = self.traj_fig.add_subplot(111)
        pass

    def plot_trajectory_Old(self, Traj, n, path, file, do_save, do_show, M, ScanProject):
        """
        [TTRM, XXRM, YYRM, Pol, DPol, RMCAPD1, RMCAPD2, LaserFlag, WindowTime, AcumilatedMeasurementTime, ZZRM]
        """
                
        # Set data
        if M==-1:
            t=Traj[0,:]
            x=Traj[1,:]
            y=Traj[2,:]
            z=Traj[10,:]
            APD=Traj[5,:]+Traj[6,:]
        else: 
            t=Traj[0,:M]
            x=Traj[1,:M]
            y=Traj[2,:M]
            z=Traj[10,:M]
            APD=Traj[5,:M]+Traj[6,:M]
        

        #### Debug
        t_interval_low=t[0]
        t_interval_high=t[-1]

        #### Normal data
        t_valid=np.where(np.logical_and(t>=t_interval_low, t<=t_interval_high))
        t=t[t_valid]
        x=x[t_valid]
        y=y[t_valid]
        z=z[t_valid]
        APD=APD[t_valid]
        ###############
        
        t_raw=ScanProject.ScanMatrix[0,:]
        t_valid=np.where(np.logical_and(t_raw>=t_interval_low, t_raw<=t_interval_high))[0]
        
        t_raw=ScanProject.ScanMatrix[0,t_valid]

        x_EOD_off=ScanProject.ScanMatrix[11,t_valid] #EOD offset vector [um]
        y_EOD_off=ScanProject.ScanMatrix[12,t_valid]

        x_P_EOD=ScanProject.ScanMatrix[15,t_valid]   #piezo center + EOD center [um] 15 
        y_P_EOD=ScanProject.ScanMatrix[16,t_valid]
        
        
        I_z_raw=ScanProject.ScanMatrix[20,t_valid]     #center index
        I_M_raw=ScanProject.ScanMatrix[8,t_valid]    #Mem index 0-5
        z_Off=ScanProject.ScanMatrix[19,t_valid]     #1/-1 
        APDH_raw=ScanProject.ScanMatrix[3,t_valid]
        APDV_raw=ScanProject.ScanMatrix[4,t_valid]
        
        N=12
        wndwStep=N

        A_con=np.convolve((APDH_raw+APDV_raw), np.ones((N,)))[N-1:-N+1:wndwStep]
        B_con=np.convolve((APDH_raw+APDV_raw)*I_M_raw, np.ones((N,)))[N-1:-N+1:wndwStep]
        z_con=B_con/A_con
        x_con_temp=np.convolve(x_EOD_off*(APDH_raw+APDV_raw), np.ones((N,)))[N-1:-N+1:wndwStep]
        x_con=x_con_temp/A_con

        t_con=np.convolve(t_raw, np.ones((N,))/N)[N-1:-N+1:wndwStep]
        #################################################################

        #Clean out Nan
        x_Nan=np.where(np.isnan(x))[0]
        y_Nan=np.where(np.isnan(y))[0]
        z_Nan=np.where(np.isnan(z))[0]
        APD_Nan=np.where(np.isnan(APD))[0]
        
        P_Nan = np.unique(np.array(list(filter(None,[x_Nan.tolist(),y_Nan.tolist(),z_Nan.tolist(),APD_Nan.tolist()]))).flatten())

        t=np.delete(t,P_Nan)
        x=np.delete(x,P_Nan)
        y=np.delete(y,P_Nan)
        z=np.delete(z,P_Nan)
        APD=np.delete(APD,P_Nan)

        #Move to around origin
        x_m=x-np.mean(x)
        y_m=y-np.mean(y)
        z_m=z-np.mean(z)
        
        alphaV=0.4
        
        ################## Ploting ############################################################
        # Initialize plot
        self.traj_fig, self.traj_ax= plt.subplots(4,1)

        # set ticks 
        axisTicksMax=np.ceil(np.max(np.append(np.append(x_m,y_m),z_m)))
        axisTicksMin=np.floor(np.min(np.append(np.append(x_m,y_m),z_m)))
        spacing = 0.1

        # Title of the plot 
        self.traj_fig.suptitle(file[:-4] + ', time: ' + str(np.around(t[-1]-t[0], decimals=3)))

        # xyz traj plot
        self.traj_ax[0].plot(t,x_m,
                             t,y_m,
                             t,z_m,
                             alpha=alphaV)

        self.traj_ax[0].grid()
        self.traj_ax[0].legend(('x','y','z'))
        self.traj_ax[0].set_ylabel('position [um]')
        self.traj_ax[0].yaxis.set_major_locator(plt.FixedLocator(np.arange(axisTicksMin, axisTicksMax, step=spacing)))
        

        # Count rate
        self.traj_ax[1].plot(t_raw, x_EOD_off*(APDH_raw+APDV_raw),
                             t_con, x_con, 
                             t_raw, (APDH_raw+APDV_raw),
                             alpha=alphaV)

        #self.traj_ax[1].plot(t, APD)  #/(t[1]-t[0]))
        self.traj_ax[1].grid()
        self.traj_ax[1].set_ylabel('counts rate [#/sec]')

        # set x axis lable which is the same in both plots
        self.traj_ax[1].set_xlabel('time [s]')
        
        # plot the fourier spectra of the signals
        fft_x, fx = self.fft_processing(x,t)
        fft_y, fy = self.fft_processing(y,t)
        fft_z, fz = self.fft_processing(z,t)

        self.traj_ax[2].semilogy(fx, fft_x, fy, fft_y, fz, fft_z, alpha=alphaV)
        self.traj_ax[2].grid()
        self.traj_ax[2].legend(('x','y','z'))
        self.traj_ax[2].set_ylabel("Amplitude")
        self.traj_ax[2].set_xlabel("Frequency [Hz]")

        # Plot the aoutocorrelation 
        x_corr, taux =self.autocorrelation(x,t)
        y_corr, tauy =self.autocorrelation(y,t)
        z_corr, tauz =self.autocorrelation(z,t)
        
        self.traj_ax[3].semilogx(taux, x_corr, tauy, y_corr, tauz, z_corr, alpha=alphaV)
        self.traj_ax[3].grid()
        self.traj_ax[3].legend(('x','y','z'))
        self.traj_ax[3].set_ylabel("Corr")
        self.traj_ax[3].set_xlabel("Tau")


        if do_save:
            self.traj_fig.savefig(path + file)
            plt.close()

        if do_show:
           plt.show()

    def plot_trajectory(ScanProject, trajs=[], SavePlot=0):
        """
        [TTRM, XXRM, YYRM, Pol, DPol, RMCAPD1, RMCAPD2, LaserFlag, WindowTime, AcumilatedMeasurementTime, ZZRM]
        """
        if trajs==[]:
            trajs=range(len(ScanProject.TrajPosEstimation))
            
        for TrajToGet in trajs:
            print('Plot traj: ' + str(TrajToGet))
            #txyzn_Map  = ScanProject.PoissonProbXYZMap[TrajToGet]
            txyzn_Map  = ScanProject.TrajPosEstimation[TrajToGet][(0,1,2,10,5),:]
            txyzn_Map[4,:]  = txyzn_Map[4,:] + ScanProject.TrajPosEstimation[TrajToGet][6,:]
            file="TrajPlot_" + str(TrajToGet) + '.pdf'
            path=ScanProject.Files['AnalysisConfig'][:-18]                

            # Set data
            t=txyzn_Map[0,:]
            x=txyzn_Map[1,:]
            y=txyzn_Map[2,:]
            z=txyzn_Map[3,:]
            APD=txyzn_Map[4,:]

            #### Debug
            t_interval_low=t[0]
            t_interval_high=t[-1]

            #### Normal data
            t_valid=np.where(np.logical_and(t>=t_interval_low, t<=t_interval_high))
            t=t[t_valid]
            x=x[t_valid]
            y=y[t_valid]
            z=z[t_valid]
            APD=APD[t_valid]
            ###############
        
            ### Raw Data
            t_raw=ScanProject.ScanMatrix[0,:]
            t_valid=np.where(np.logical_and(t_raw>=t_interval_low, t_raw<=t_interval_high))[0]
        
            t_raw=ScanProject.ScanMatrix[0,t_valid]

            x_P=ScanProject.ScanMatrix[9,t_valid]   #piezo center [um] 
            y_P=ScanProject.ScanMatrix[10,t_valid]

            x_EOD =ScanProject.ScanMatrix[13,t_valid] #EOD signal from FPGA [um]
            y_EOD =ScanProject.ScanMatrix[14,t_valid]

            x_EOD_HF =ScanProject.ScanMatrix[23,t_valid] #EOD signal from FPGA [um]
            y_EOD_HF =ScanProject.ScanMatrix[24,t_valid]

            x_P_EOD=ScanProject.ScanMatrix[15,t_valid]   #piezo center + EOD center [um]
            y_P_EOD=ScanProject.ScanMatrix[16,t_valid]
            
            I_z_raw=ScanProject.ScanMatrix[20,t_valid]     #center index

            I_M_raw=ScanProject.ScanMatrix[8,t_valid]    #Mem index 0-5
            z_Off=ScanProject.ScanMatrix[19,t_valid]     #1/-1 

            APDH_raw=ScanProject.ScanMatrix[3,t_valid]
            APDV_raw=ScanProject.ScanMatrix[4,t_valid]
            #################################################################

            #Clean out Nan
            x_Nan=np.where(np.isnan(x))[0]
            y_Nan=np.where(np.isnan(y))[0]
            z_Nan=np.where(np.isnan(z))[0]
            APD_Nan=np.where(np.isnan(APD))[0]
        
            P_Nan = np.unique(np.array(list(filter(None,[x_Nan.tolist(),y_Nan.tolist(),z_Nan.tolist(),APD_Nan.tolist()]))).flatten())
            if P_Nan.size!=0:
                t=np.delete(t,P_Nan)
                x=np.delete(x,P_Nan)
                y=np.delete(y,P_Nan)
                z=np.delete(z,P_Nan)
                APD=np.delete(APD,P_Nan)

            #Move to around origin
            x_m=x-np.mean(x)
            y_m=y-np.mean(y)
            z_m=z-np.mean(z)
        
            ################## Ploting ############################################################
            # Initialize plot
            traj_fig, traj_ax= plt.subplots(4,1)

            # set ticks 
            axisTicksMax=np.ceil(np.max(np.append(np.append(x_m,y_m),z_m)))
            axisTicksMin=np.floor(np.min(np.append(np.append(x_m,y_m),z_m)))
            spacing = 0.1
            
            alphaV=1.0
            linewidthV=0.2

            # Title of the plot 
            traj_fig.suptitle(file[:-4] + ', time: ' + str(np.around(t[-1]-t[0], decimals=3)))

            # xyz traj plot
            traj_ax[0].plot(t,x_m,
                                 t,y_m,
                                 t,z_m,
                                 alpha=alphaV, linewidth=linewidthV)

            traj_ax[0].grid()
            traj_ax[0].legend(('x','y','z'))
            traj_ax[0].set_ylabel('position [um]')
            traj_ax[0].yaxis.set_major_locator(plt.FixedLocator(np.arange(axisTicksMin, axisTicksMax, step=spacing)))

            # Count rate
            traj_ax[1].plot(t, APD,
                                 alpha=alphaV, linewidth=linewidthV)

            traj_ax[1].grid()
            traj_ax[1].set_ylabel('counts rate [#/sec]')

            # set x axis lable which is the same in both plots
            traj_ax[1].set_xlabel('time [s]')
        
            # plot the fourier spectra of the signals
            fft_x, fx = _MS.fft_processing(x,t)
            fft_y, fy = _MS.fft_processing(y,t)
            fft_z, fz = _MS.fft_processing(z,t)
            
            fft_x2, fx2 = _MS.fft_processing(x_P,t_raw)
            fft_y2, fy2 = _MS.fft_processing(y_P,t_raw)
            fft_xEOD, fxEOD = _MS.fft_processing(x_EOD,t_raw)
            fft_yEOD, fyEOD = _MS.fft_processing(y_EOD,t_raw)
            fft_xEOD_HF, fxEOD_HF = _MS.fft_processing(x_EOD_HF,t_raw)
            fft_yEOD_HF, fyEOD_HF = _MS.fft_processing(y_EOD_HF,t_raw)            

            #traj_ax[2].semilogy(fx, fft_x, fy, fft_y, fz, fft_z, alpha=alphaV)
            traj_ax[2].loglog(fx, fft_x, 
                                fy, fft_y, 
                                fz, fft_z, 
                                fx2, fft_x2, 
                                fy2, fft_y2, 
                                fxEOD, fft_xEOD, 
                                fyEOD, fft_yEOD, 
                                fxEOD_HF, fft_xEOD_HF, 
                                fyEOD_HF, fft_yEOD_HF, alpha=alphaV, linewidth=linewidthV)
            traj_ax[2].grid()
            traj_ax[2].legend(('x','y','z','xp','yp','xEOD','yEOD','xEODHF','yEODHF'))
            traj_ax[2].set_ylabel("Amplitude")
            traj_ax[2].set_xlabel("Frequency [Hz]")

            # Plot the aoutocorrelation 
            x_corr, taux =_MS.autocorrelation(x,t)
            y_corr, tauy =_MS.autocorrelation(y,t)
            z_corr, tauz =_MS.autocorrelation(z,t)
            
            y_corr2, tauy2 =_MS.autocorrelation(y_P,t_raw)
            x_corr2, taux2 =_MS.autocorrelation(x_P,t_raw)
            
            traj_ax[3].semilogx(taux, x_corr, tauy, y_corr, tauz, z_corr,tauy2, y_corr2, taux2, x_corr2, alpha=alphaV, linewidth=linewidthV)
            traj_ax[3].grid()
            traj_ax[3].legend(('x','y','z','y_p','x_p'))
            traj_ax[3].set_ylabel("Corr")
            traj_ax[3].set_xlabel("Tau")

            #Save or show
            if SavePlot:
                traj_fig.savefig(path + file)
                plt.close()
            else:
                plt.show()

    def plot_trajectory_MSD(self, ScanProject, trajs=[]):
        if trajs==[]:
            trajs=range(len(ScanProject.PoissonProbXYZMap))

        for TrajToGet in trajs:
            print('Plot MSD traj' + str(TrajToGet))
            txyzn_Map  = ScanProject.PoissonProbXYZMap[TrajToGet]
            txyzMSD    = ScanProject.txyzMSD_Poisson_ProbMap[TrajToGet]
            t_Dxyz_dt_MSD = ScanProject.txyzDMSD_Dt_Poisson_ProbMap[TrajToGet]
            
            path=ScanProject.Files['AnalysisConfig'][:-18]
            file="MSD" + str(TrajToGet) + '.pdf'

            #### Plot 
            fig, ax= plt.subplots(4,1)
            linewidthV=0.1
            alphaV=0.4
            
            # Title of the plot 
            fig.suptitle(file[:-4] + ', time: ' + str(np.around(txyzn_Map[0,-1]-txyzn_Map[0,0], decimals=3)))

            ax[0].plot(txyzn_Map[0,:],txyzn_Map[1,:]-txyzn_Map[1,0], alpha=alphaV,linewidth=linewidthV)
            ax[0].plot(txyzn_Map[0,:],txyzn_Map[2,:]-txyzn_Map[2,0], alpha=alphaV,linewidth=linewidthV)
            ax[0].plot(txyzn_Map[0,:],txyzn_Map[3,:]-txyzn_Map[3,0], alpha=alphaV,linewidth=linewidthV)
            ax[0].legend(('x','y','z'))
            ax[0].set_ylabel("[um]")

            ax[1].plot(txyzn_Map[0,:],txyzn_Map[4,:],'-b', alpha=alphaV, linewidth=linewidthV)
            ax[1].set_ylabel("[#]")

            ax[2].plot(txyzMSD[0,:],txyzMSD[1,:],txyzMSD[0,:],txyzMSD[2,:],txyzMSD[0,:],txyzMSD[3,:], alpha=alphaV, linewidth=linewidthV)
            ax[2].legend(('x','y','z'))
            ax[2].set_ylabel("[um^2]")

            ax[3].plot(t_Dxyz_dt_MSD[0,:],t_Dxyz_dt_MSD[1,:],t_Dxyz_dt_MSD[0,:],t_Dxyz_dt_MSD[2,:],t_Dxyz_dt_MSD[0,:],t_Dxyz_dt_MSD[3,:],alpha=alphaV ,linewidth=linewidthV)
            ax[3].set_yscale('log') #'linear'
            ax[3].legend(('x','y','z'))
            ax[3].set_ylabel("[um^2/s]")

            #Save or show
            if ScanProject.SavePlot:
                fig.savefig(path + file)
                plt.close()
            else:
                plt.show()

    def plot_Poisson_trajectory(self, Traj, n, path, file, do_save, do_show, M, ScanProject):
        
        """
        plot_trajectory(self, Traj,n, path, file, do_save, do_show):
        [TTRM, XXRM, YYRM, Pol, DPol, RMCAPD1, RMCAPD2, LaserFlag, WindowTime, AcumilatedMeasurementTime, ZZRM]
        """
                
        # Set data
        if M==-1:
            t=Traj[0,:]
            x=Traj[1,:]
            y=Traj[2,:]
            z=Traj[10,:]
            APD=Traj[5,:]+Traj[6,:]
        else: 
            t=Traj[0,:M]
            x=Traj[1,:M]
            y=Traj[2,:M]
            z=Traj[10,:M]
            APD=Traj[5,:M]+Traj[6,:M]
        

        #### Debug
        t_interval_low=12.2  #t[0]
        t_interval_high=12.5 #t[-1]

        #### Gaussian data
        t_valid=np.where(np.logical_and(t>=t_interval_low, t<=t_interval_high))
        t=t[t_valid]
        x=x[t_valid]
        y=y[t_valid]
        z=z[t_valid]
        APD=APD[t_valid]
        ###############
        
        t_raw=ScanProject.ScanMatrix[0,:]
        t_valid=np.where(np.logical_and(t_raw>=t_interval_low, t_raw<=t_interval_high))[0]
        
        t_raw=ScanProject.ScanMatrix[0,t_valid]                               #time [s]
        x_EOD_off=ScanProject.ScanMatrix[1,t_valid]                           #x_i Position [um]
        y_EOD_off=ScanProject.ScanMatrix[2,t_valid]                           #y_i Position [um]
        n=ScanProject.ScanMatrix[3,t_valid]+ScanProject.ScanMatrix[4,t_valid] #n_i photon counts 
        m=24                                                                  #Sample size to use
        Lambda=540e-9                                                         #[m]
        Na=1.45                                                               #Numerical aperture
        s=(0.822/2)*(Lambda/Na)*1e6                                           #[um] theoretical PSF width, the radius of the aridisk is scaled to 1.22/2 and for the e-2 intensity drop that is std value one use 0.822/2
        sx=sy=s
        sz=s*2
        NL=1

        NT=int(len(x_EOD_off)/m)
        xyztM=np.zeros((4,NT))

        for i in range(NT):
            print(str(i) + " of " + str(NT))
            
            #Points to be used in evaluation
            xi = x_EOD_off[i*m:(i+1)*m]
            yi = y_EOD_off[i*m:(i+1)*m]
            zi = y_EOD_off[i*m:(i+1)*m]*0 + 1
            ti = np.mean(t_raw[i*m:(i+1)*m])
            ni = n[i*m:(i+1)*m]

            s  = np.array([sx,sy,sz])
            rL = np.vstack((xi,yi,zi))
            nL = ni
            
            if i==0:
                rp0=np.array([xi[0],yi[0],zi[0]])
            else:
                rp0=xyztM[0:3,i-1]
            
            args=(rL, nL, s, NL)
            rp_min=fmin(opt.min_target_Poisson,rp0,args, disp=0)
            xyztM[:,i]=np.concatenate((rp_min,[ti]))
            

        xyM, LT, xx, yy = Tr.ProbMap_trajectory(ScanProject, t_interval_low, t_interval_high, verbose=1)

        print("Done")

    def Plot_2D_L(self, LT, xx, yy, i):        
        """
        xx,yy,LT are NxNxT matrix describing the surface as X,Y,Z 
        """
        xlim=[0,20]
        ylim=[0,20]

        fig = plt.figure()
        ax = fig.add_subplot(111)   
          
        cf=ax.contourf(xx[:,:,i], yy[:,:,i], LT[:,:,i],cmap=cm.gray) #cm.coolwarm, 
        ax.set_title('Min/Max=' + str("{:.2e}".format(np.min(LT[:,:,i]))) + '/' + str("{:.2e}".format(np.max(LT[:,:,i]))) )
        ax.set_facecolor('k')
        fig.colorbar(cf, ax=ax)
        plt.xlim(xlim[0],xlim[1])
        plt.ylim(ylim[0],ylim[1])
        
        plt.show()
        
    def Animate_L(self, LT, xx, yy):        
        """
        This is an atempt to make an 2D animation over time. 
        The out put is a movie of the function LT color projected to the 2D plane over NT frames 
        One thing to note is that the output resolusion is not set thus it will depend on xlim and ylim (see code)
        xx,yy,LT are NxNxT matrix describing the surface as X,Y,Z 
        """
        NT=LT.shape[2]

        fig = plt.figure()
        ax = fig.add_subplot(111)   
        
        def Frame(i):   
            ax.clear()
            ax.contourf(xx[:,:,i], yy[:,:,i], LT[:,:,i],cmap=cm.gray) #cm.coolwarm, 
            ax.set_title('Min/Max=' + str("{:.2e}".format(np.min(LT[:,:,i]))) + '/' + str("{:.2e}".format(np.max(LT[:,:,i]))) )
            ax.set_facecolor('k')
            #fig.colorbar(cf, ax=ax)
            plt.xlim(0,20)
            plt.ylim(0,20)
            return ax

        def init():
            ax=Frame(0)
            return ax

        ax=Frame(0)

        # Creating the Animation object
        ani = animation.FuncAnimation(fig, Frame, init_func=init, 
                                      frames=NT, interval=10, blit=False)
        
        ani.save('Animate_L.gif', writer='imagemagick')

    def plot_3D_trajectory(self, Traj, n, path, file, do_save, do_show, do_ani,M):
        
        """
        plot_trajectory(self, Traj,n, path, file, do_save, do_show):
        [TTRM, XXRM, YYRM, Pol, DPol, RMCAPD1, RMCAPD2, LaserFlag, WindowTime, AcumilatedMeasurementTime, ZZRM]
        """
        fig = plt.figure()
        ax = p3.Axes3D(fig)
        
        # Set data
        if M==-1:
            t=Traj[0,:]
            x=Traj[1,:]
            y=Traj[2,:]
            z=Traj[10,:]
            APD=Traj[5,:]+Traj[6,:]
        else: 
            t=Traj[0,:M]
            x=Traj[1,:M]
            y=Traj[2,:M]
            z=Traj[10,:M]
            APD=Traj[5,:M]+Traj[6,:M]
       
        x_m=x-np.mean(x)
        y_m=y-np.mean(y)
        z_m=z-np.mean(z)

        #lines=ax.plot(x, y, z)  
        N=len(x)
        color=z/np.max(z)
        for i in range(N):
            ax.plot(x_m[i:i+2], y_m[i:i+2], z_m[i:i+2], color=plt.cm.jet(color[i]), alpha=0.3)
        ax.grid()
        ax.view_init(90, 90)
        #ax.colorbar()

        # Setting the axes properties
        ax.set_xlim3d([-1.0, 1.0])
        ax.set_xlabel('X [um]')

        ax.set_ylim3d([-1.0, 1.0])
        ax.set_ylabel('Y [um]')

        ax.set_zlim3d([-2.0, 2.0])
        ax.set_zlabel('Z [um]')

        ax.set_title(file[:-4] + ' time: ' + str(t[-1]-t[0]))
        
        if do_ani:
            lines=ax.plot(x_m[0:1], y_m[0:1], z_m[0:1], color=plt.cm.jet(color[0]), alpha=0.4)  
            ax.grid()
            data=[np.vstack([x_m,y_m,z_m, color])]

            # Setting the axes properties
            ax.set_xlim3d([-2.0, 2.0])
            ax.set_xlabel('X [um]')

            ax.set_ylim3d([-2.0, 2.0])
            ax.set_ylabel('Y [um]')

            ax.set_zlim3d([-2.0, 2.0])
            ax.set_zlabel('Z [um]')

            ax.set_title(file[:-4] + ' time: ' + str(t[-1]-t[0]))

            # Creating the Animation object
            ani = animation.FuncAnimation(fig, self.update_lines, 25, fargs=(data, lines),
                                       interval=50, blit=False)
        
            ani.save('sine_wave.gif', writer='imagemagick')
        

        if do_save:
            fig.savefig(path + file)
            plt.close()

        if do_show:
            plt.show()

    def update_lines(self, num, dataLines, lines):
        for line, data in zip(lines, dataLines):
            # NOTE: there is no .set_data() for 3 dim data...
            # thus first update xy and then z 
            line.set_data(data[0:2, :num])        #xy update 
            line.set_3d_properties(data[2, :num]) #z update
        return lines

    def plotProbMapPointCloud(LogPTot,xx,yy,zz):
        #3D density plot
        GP=np.exp(LogPTot)
        validData=np.where(GP>np.max(GP)*0.1)
        GP=GP[validData]
        xv=xx[validData] 
        yv=yy[validData]
        zv=zz[validData]
                
        xyz_pos=np.transpose(np.vstack((xv, yv, zv)))
        rgba_colors=cm.jet((GP-np.min(GP))/(np.max(GP)-np.min(GP)))            
        #rgba_colors[:,3]=(GP-np.min(GP))/(np.max(GP)-np.min(GP))
        QTplt.DensityPlot(xyz_pos,rgba_colors)
   
    def Neighborhood(ScanProject,TrajToGet,intervalEnd,intervalLength):

        txyznM=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]
        SxyzM=ScanProject.PoissonProbXYZMapS_NumPFS[TrajToGet]
        xyzBounds=ScanProject.PoissonProbXYZMap_NeighborhoodBounds[TrajToGet][0]
        fig = plt.figure();
        ax = fig.add_subplot(projection='3d');
        for j in range(intervalEnd-intervalLength,intervalEnd):
            GP=ScanProject.PoissonProbXYZMap_Neighborhood[TrajToGet][j][:,0]
            xyz_pos=ScanProject.PoissonProbXYZMap_Neighborhood[TrajToGet][j][:,1:]
            rgba_colors=cm.jet((GP-np.min(GP))/(np.max(GP)-np.min(GP)))   
            rgba_colors[:,3]=(GP-np.min(GP))/(np.max(GP)-np.min(GP))
            ax.scatter(xyz_pos[:,0], xyz_pos[:,1], xyz_pos[:,2], c=rgba_colors);    
        ax.plot3D(txyznM[1,intervalEnd-intervalLength:intervalEnd], txyznM[2,intervalEnd-intervalLength:intervalEnd], txyznM[3,intervalEnd-intervalLength:intervalEnd],'gray');  
        ax.plot3D((ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet].T)[0,intervalEnd-intervalLength:intervalEnd], (ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet].T)[1,intervalEnd-intervalLength:intervalEnd], (ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet].T)[2,intervalEnd-intervalLength:intervalEnd],'b');  
        
        #Create bounding box
        xB=xyzBounds[0,:];yB=xyzBounds[1,:];zB=xyzBounds[2,:]
        for s, e in combinations(np.array(list(product(xB, yB, zB))), 2):
            if np.sum((s-e)==0)==2:
                ax.plot3D(*zip(s, e), color="b")

        ax.set_xlim(xyzBounds[0,:])
        ax.set_ylim(xyzBounds[1,:])
        ax.set_zlim(xyzBounds[2,:])
        plt.show()

    def NeighborhoodSphere(ScanProject,TrajToGet,do_save_plot=0, i=0, Ssize=10**3, Nmin=-1 ,FolderEnd='',FileEnd=''):

        T=(ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet][0,:]-ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet][0,0])
        N=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet][4,:]
        ofInterest= np.where((T>=i*Ssize)*(T<(i+1)*Ssize)*(N>=Nmin))[0]
        
        txyznM=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet][:,ofInterest]
        SxyzM=ScanProject.PoissonProbXYZMapS_NumPFS[TrajToGet][:,ofInterest]
        if np.sum(np.isnan(SxyzM))>0:
            SxyzM=np.zeros(SxyzM.shape)
        Cvalue=ScanProject.PoissonProbXYZMapC_NumPFS[TrajToGet][ofInterest]
        #plt.plot(ScanProject.PoissonProbXYZMapC_NumPFS[TrajToGet]);plt.show()
        
        xyzBounds=ScanProject.PoissonProbXYZMap_NeighborhoodBounds[TrajToGet][0]
        t=(txyznM[0,:]-txyznM[0,0])*1e3; x=txyznM[1,:]; y=txyznM[2,:]; z=txyznM[3,:]; n=txyznM[4,:]
        Sx=SxyzM[0,:]; Sy=SxyzM[1,:]; Sz=SxyzM[2,:]

        Sxyz=np.sqrt(((Sx-np.min(Sx))/(np.max(Sx)-np.min(Sx)))**2+((Sy-np.min(Sy))/(np.max(Sy)-np.min(Sy)))**2 + ((Sz-np.min(Sz))/(np.max(Sz)-np.min(Sz)))**2)
        rgba_colors=cm.jet(Sxyz)
        #rgba_colors[:,3]=Sxyz


        ########## Plot
        # Figure construction
        fig = plt.figure(constrained_layout=False)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = fig.add_gridspec(nrows=8, ncols=6, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.2, hspace=1.0)

        ax0=fig.add_subplot(gs[0:2, 0:2]) #pos
        ax1=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

        ax2=fig.add_subplot(gs[2:4, 0:2]) #Sigma
        ax3=fig.add_subplot(gs[2:4, 2])  #hist Sigma      

        ax4=fig.add_subplot(gs[4:6, 0:2]) #Photon count 
        ax5=fig.add_subplot(gs[4:6, 2])  #Photon hist

        ax6=fig.add_subplot(gs[6:8, 0:2]) #Cross section  
        ax7=fig.add_subplot(gs[6:8, 2])  #Cross section hist

        #ax6=fig.add_subplot(gs[4:5, 0:2]) #Sigma
        #ax5=fig.add_subplot(gs[4:5, 2])  #hist sigma2     
        #ax15=fig.add_subplot(gs[5:7, 3:5])  #fft     

        ax11=fig.add_subplot(gs[0:4, 3:6],projection='3d') #Image 
        ax12=fig.add_subplot(gs[4:8, 3:6],projection='3d') #Image 2
                
        #ax13=fig.add_subplot(gs[4, 3:5]) #Pol count 
        #ax14=fig.add_subplot(gs[4, 5])  #Pol hist

        #######
        ## ax0 postion 
        ax0.plot(t,x,'r',label='x [um]',alpha=0.4)
        ax0.plot(t,y,'g',label='y',alpha=0.4)
        ax0.plot(t,z,'b',label='z',alpha=0.4)
        ax0.legend()
        ax0.grid()

        ## ax1 Histogram over positions
        ax1.hist(x-np.mean(x), 20, facecolor='r', alpha=0.4); 
        ax1.hist(y-np.mean(y), 20, facecolor='g', alpha=0.4); 
        ax1.hist(z-np.mean(z), 20, facecolor='b', alpha=0.4)          
        ax1.set_xlabel('x-mean(x) [um]')

        ## ax2 plot 95% MLE std
        ax2.plot(t,Sx,'r',label='x std 95% MLE',alpha=0.4)
        ax2.plot(t,Sy,'g',label='y',alpha=0.4)
        ax2.plot(t,Sz,'b',label='z',alpha=0.4)
        ax2.grid(); ax2.legend()

        ## ax3 hist of 95% MLE std
        ax3.hist(Sx, 20, facecolor='r', alpha=0.4); 
        ax3.hist(Sy, 20, facecolor='g', alpha=0.4); 
        ax3.hist(Sz, 20, facecolor='b', alpha=0.4); 

        ## ax4 photon counts
        ax4.plot(t,n,'b',label='Photon counts')
        ax4.grid(); ax4.legend()
        ## ax5 photon hist
        ax5.hist(n,20, facecolor='b')

        ## ax6 cross section 
        ax6.plot(t,Cvalue,'b',label='Cvalue')
        ax6.grid(); ax6.legend()
        ## ax7 Cross section hist
        ax7.hist(Cvalue,20, facecolor='b')
        ax7.set_xlabel('[ms]')

        ### ax11
        ax11.plot3D(x, y, z,'gray');
        ## Piezo stage
        ax11.plot3D((ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet].T)[0,:], (ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet].T)[1,:], (ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet].T)[2,:],'b',alpha=0.4);  
        scatterPlot=ax11.scatter(x, y, z,s=np.pi*(1+Sxyz)**2, c=Sxyz,cmap=cm.jet);   
        ax11.set_xlabel('x');ax11.set_ylabel('y');ax11.set_zlabel('z')
        ax11.view_init(elev=90., azim=-90)

        #Create bounding box
        xB=xyzBounds[0,:];yB=xyzBounds[1,:];zB=xyzBounds[2,:]
        for s, e in combinations(np.array(list(product(xB, yB, zB))), 2):
            if np.sum((s-e)==0)==2:
                ax11.plot3D(*zip(s, e), color="b")
        
        ax11.set_xlim(xyzBounds[0,:])
        ax11.set_ylim(xyzBounds[1,:])
        ax11.set_zlim(xyzBounds[2,:])
        fig.colorbar(scatterPlot,ax=ax11)


        ### ax12
        ax12.plot3D(x, y, z,'gray');
        ## Piezo stage
        ax12.plot3D((ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet].T)[0,:], (ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet].T)[1,:], (ScanProject.PoissonProbXYZMap_Piezo_Offset[TrajToGet].T)[2,:],'b',alpha=0.4);  
        scatterPlot=ax12.scatter(x, y, z,s=np.pi*(1+Sxyz)**2, c=Sxyz,cmap=cm.jet);   
        ax12.set_xlabel('x');ax12.set_ylabel('y');ax12.set_zlabel('z')
        ax12.view_init(elev=0., azim=-90)

        #Create bounding box
        xB=xyzBounds[0,:];yB=xyzBounds[1,:];zB=xyzBounds[2,:]
        for s, e in combinations(np.array(list(product(xB, yB, zB))), 2):
            if np.sum((s-e)==0)==2:
                ax12.plot3D(*zip(s, e), color="b")
        
        ax12.set_xlim(xyzBounds[0,:])
        ax12.set_ylim(xyzBounds[1,:])
        ax12.set_zlim(xyzBounds[2,:])
        fig.colorbar(scatterPlot,ax=ax12)

        #########################################################
        if do_save_plot:
            file="T_" + str(TrajToGet) + '_' + FileEnd + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_' + FolderEnd +'/'
            Path(path).mkdir(parents=True, exist_ok=True)
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

    def NeighborhoodSphere_v2(ScanProject,TrajToGet,do_save_plot=0, i=0, Ssize=10**3, Nmin=-1 ,FolderEnd='',FileEnd=''):

        T=(ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet][0,:]-ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet][0,0])
        N=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet][4,:]
        ofInterest= np.where((T>=i*Ssize)*(T<(i+1)*Ssize)*(N>=Nmin))[0]
        
        txyznM=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet][:,ofInterest]
        SxyzM=ScanProject.PoissonProbXYZMapS_NumPFS[TrajToGet][:,ofInterest]
        if np.sum(np.isnan(SxyzM))>0:
            SxyzM=np.zeros(SxyzM.shape)
        Cvalue=ScanProject.PoissonProbXYZMapC_NumPFS[TrajToGet][ofInterest]
        #plt.plot(ScanProject.PoissonProbXYZMapC_NumPFS[TrajToGet]);plt.show()
        
        xyzBounds=ScanProject.PoissonProbXYZMap_NeighborhoodBounds[TrajToGet][0]
        t=(txyznM[0,:]-txyznM[0,0])*1e3; x=txyznM[1,:]; y=txyznM[2,:]; z=txyznM[3,:]; n=txyznM[4,:]
        Sx=SxyzM[0,:]; Sy=SxyzM[1,:]; Sz=SxyzM[2,:]

        Sxyz=np.sqrt(((Sx-np.min(Sx))/(np.max(Sx)-np.min(Sx)))**2+((Sy-np.min(Sy))/(np.max(Sy)-np.min(Sy)))**2 + ((Sz-np.min(Sz))/(np.max(Sz)-np.min(Sz)))**2)
        rgba_colors=cm.jet(Sxyz)
        #rgba_colors[:,3]=Sxyz


        ########## Plot
        # Figure construction
        fig = plt.figure(constrained_layout=False)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = fig.add_gridspec(nrows=8, ncols=6, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.2, hspace=1.0)

        ax0=fig.add_subplot(gs[0:2, 0:2]) #pos
        ax1=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

        ax2=fig.add_subplot(gs[2:4, 0:2]) #Sigma
        ax3=fig.add_subplot(gs[2:4, 2])  #hist Sigma      

        ax4=fig.add_subplot(gs[4:6, 0:2]) #Photon count 
        ax5=fig.add_subplot(gs[4:6, 2])  #Photon hist

        ax6=fig.add_subplot(gs[6:8, 0:2]) #Cross section  
        ax7=fig.add_subplot(gs[6:8, 2])  #Cross section hist

        #ax6=fig.add_subplot(gs[4:5, 0:2]) #Sigma
        #ax5=fig.add_subplot(gs[4:5, 2])  #hist sigma2     
        #ax15=fig.add_subplot(gs[5:7, 3:5])  #fft     

        ax11=fig.add_subplot(gs[0:2, 3:5]) #Sigma 95% 
        ax12=fig.add_subplot(gs[0:2, 5]) #hist Sigma 95% 
                
        #ax13=fig.add_subplot(gs[4, 3:5]) #Pol count 
        #ax14=fig.add_subplot(gs[4, 5])  #Pol hist

        #######
        ## ax0 postion 
        ax0.plot(t,x,'r',label='x [um]',alpha=0.4)
        ax0.plot(t,y,'g',label='y',alpha=0.4)
        ax0.plot(t,z,'b',label='z',alpha=0.4)
        ax0.legend()
        ax0.grid()

        ## ax1 Histogram over positions
        ax1.hist(x-np.mean(x), 20, facecolor='r', alpha=0.4); 
        ax1.hist(y-np.mean(y), 20, facecolor='g', alpha=0.4); 
        ax1.hist(z-np.mean(z), 20, facecolor='b', alpha=0.4)          
        ax1.set_xlabel('x-mean(x) [um]')

        ## ax2 plot 95% MLE std
        ax2.plot(t,Sx,'r',label='x std 95% MLE',alpha=0.4)
        ax2.plot(t,Sy,'g',label='y',alpha=0.4)
        ax2.plot(t,Sz,'b',label='z',alpha=0.4)
        ax2.grid(); ax2.legend()

        ## ax3 hist of 95% MLE std
        ax3.hist(Sx, 20, facecolor='r', alpha=0.4); 
        ax3.hist(Sy, 20, facecolor='g', alpha=0.4); 
        ax3.hist(Sz, 20, facecolor='b', alpha=0.4); 

        ## ax4 photon counts
        ax4.plot(t,n,'b',label='Photon counts')
        ax4.grid(); ax4.legend()
        ## ax5 photon hist
        ax5.hist(n,20, facecolor='b')

        ## ax6 cross section 
        ax6.plot(t,Cvalue,'b',label='Cvalue')
        ax6.grid(); ax6.legend()
        ## ax7 Cross section hist
        ax7.hist(Cvalue,20, facecolor='b')
        ax7.set_xlabel('[ms]')

        ## ax11 sum plot 95% MLE std
        ax11.plot(t,(Sx+Sy+Sz)/3,'r',label='xyz std 95% MLE',alpha=0.4)
        ax11.grid(); ax2.legend()

        ## ax12 sum hist of 95% MLE std
        ax12.hist((Sx+Sy+Sz)/3, 20, facecolor='r', alpha=0.4); 

        #########################################################
        if do_save_plot:
            file="T_" + str(TrajToGet) + '_' + FileEnd + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_' + FolderEnd +'/'
            Path(path).mkdir(parents=True, exist_ok=True)
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

    def plotProbMapProjections(ScanProject,LogPTot,xx,yy,zz,txyznM,txyzG,i,do_save_plot=1):
        GE=np.exp(LogPTot)
                
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = GridSpec(2, 3, figure=fig)
        ax0=fig.add_subplot(gs[0, 0])
        ax1=fig.add_subplot(gs[0, 1])
        ax2=fig.add_subplot(gs[0, 2])
        ax3=fig.add_subplot(gs[1, :])

        cs0=ax0.contourf(xx[:,:,0],yy[:,:,0],np.sum(GE,axis=2),alpha=1.0)
        ax0.plot(txyznM[1,i],txyznM[2,i],'r*')
        ax0.plot(txyzG[1],txyzG[2],'b*')
        fig.colorbar(cs0,ax=ax0)
        #cs0.set_clim(-56000,-40000)
        ax0.ticklabel_format(style='sci',scilimits=(0,0),axis='both')
        ax0.set_ylabel("y [um]")
        ax0.set_xlabel("x [um]")
                
        cs1=ax1.contourf(xx[0,:,:],zz[0,:,:],np.sum(GE,axis=0),alpha=1.0)
        ax1.plot(txyznM[1,i],txyznM[3,i],'r*')
        ax1.plot(txyzG[1],txyzG[3],'b*')
        fig.colorbar(cs1,ax=ax1)
        ax1.ticklabel_format(style='sci',scilimits=(0,0),axis='both')
        ax1.set_ylabel("z [um]")
        ax1.set_xlabel("x [um]")

        cs2=ax2.contourf(yy[:,0,:],zz[0,:,:],np.sum(GE,axis=1),alpha=1.0)
        ax2.plot(txyznM[2,i],txyznM[3,i],'r*')
        ax2.plot(txyzG[2],txyzG[3],'b*')
        fig.colorbar(cs2,ax=ax2)
        ax2.ticklabel_format(style='sci',scilimits=(0,0),axis='both')
        ax2.set_ylabel("z [um]")
        ax2.set_xlabel("y [um]")

        if i==0:
            ax3.plot([txyznM[0,:i+1],txyznM[0,:i+1]],[0,0],'-')
            ax3.plot([txyznM[0,:i+1],txyznM[0,:i+1]],[0,0],'-')
            ax3.plot([txyznM[0,:i+1],txyznM[0,:i+1]],[0,0],'-')
        else:
            ax3.plot(txyznM[0,:i+1],txyznM[1,:i+1]-txyznM[1,0],'-')
            ax3.plot(txyznM[0,:i+1],txyznM[2,:i+1]-txyznM[2,0],'-')
            ax3.plot(txyznM[0,:i+1],txyznM[3,:i+1]-txyznM[3,0],'-')

        ax3.legend(('x-x(0)','y-y(0)','z-z(0)'))
        ax3.set_ylabel("Pos [um]")
        ax3.set_xlabel("time [s]")
        ax3.set_title("Photon n: " + str(ni))

        if do_save_plot==1:
            file="ProjectionPlot_PCorr1_t" + str(TrajToGet) + "_" + str(i) + '.pdf'
            path=ScanProject.Files['AnalysisConfig'][:-18]                
            fig.savefig(path + file)
            plt.close()
        else:
            plt.show()

    def plotProbMap_Hist_Traj_sim(ScanProject,LogPTot,S,xx,yy,zz,txyznM,GroundTruth,ni,TrajToGet,i,TopHatPrior,do_save_plot=1):
        """
        This function need work and there are some strange data points that need to search for the root
        
        PT: probability map 3D 
        xyzpix: Groundtruth data 
        Gyxz:   Gaussian estimate ScanProject.TrajPosEstimation[TrajToGet]
        PTMyxz: MLE 
        Myxz: Laser shot positions 

        This function is a copy that was used under Simulator before but needs to be re orgenized
        """

        TPxy=np.max(TopHatPrior,axis=2)
        TPxy[TPxy<0]=np.nan
        TPxz=np.max(TopHatPrior,axis=0)
        TPxz[TPxz<0]=np.nan
        TPyz=np.max(TopHatPrior,axis=1)
        TPyz[TPyz<0]=np.nan


        PT    =np.exp(LogPTot)
        m = int(ScanProject.TrajectorySelectionRules['wndw'])

        VGxyz  =np.vstack((ScanProject.TrajPosEstimation[TrajToGet][1,:i+1],
                           ScanProject.TrajPosEstimation[TrajToGet][2,:i+1],
                           ScanProject.TrajPosEstimation[TrajToGet][-1,:i+1])) 
        VGxyz=np.transpose(VGxyz)
        Gxyz=VGxyz[-1,:]

        if np.sum(np.isnan(GroundTruth)):
            Vxyzpix=VGxyz
        else:
            Vxyzpix=GroundTruth[:m*i+1:m,:] #[um] using the last point to show
        xyzpix=Vxyzpix[-1,:]

        Vt=np.transpose(txyznM[0,:i+1])*1e3 #[ms]
        VPTMxyz=np.transpose(txyznM[1:4,:i+1])
        PTMxyz=VPTMxyz[-1,:]

        S[np.isnan(S)]=0
        S[np.where(S<np.max(S)*0.4)]=0
        Myxz=np.zeros((S.shape[3],3))
        for ii in range(S.shape[3]):
            indextosee=np.unravel_index(np.argmax(S[:,:,:,ii]),S[:,:,:,ii].shape)
            Myxz[ii,:]=[yy[indextosee],xx[indextosee],zz[indextosee]]   
        

        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = GridSpec(4, 3, figure=fig)
        ax0=fig.add_subplot(gs[0, 0])
        ax1=fig.add_subplot(gs[0, 1])
        ax2=fig.add_subplot(gs[0, 2])
        ax3=fig.add_subplot(gs[1, 0:2])
        ax4=fig.add_subplot(gs[2, 0:2])
        ax5=fig.add_subplot(gs[3, 0:2])
        ax6=fig.add_subplot(gs[1:4, 2])

        ax0.contourf(xx[:,:,0],yy[:,:,0],np.sum(PT,axis=2),alpha=1.0)
        ax0.contourf(xx[:,:,0],yy[:,:,0],TPxy,alpha=0.2)
        ax0.plot(xyzpix[0],xyzpix[1],'+r')
        ax0.plot(Gxyz[0],Gxyz[1],'*g')
        ax0.plot(PTMxyz[0],PTMxyz[1],'xm')
        #ax0.contourf(xx[:,:,0],yy[:,:,0],np.sum(np.sum(S,axis=2),axis=2),alpha=0.4)
        for ii in range(Myxz.shape[0]):
            #ax0.contourf(xx[:,:,0],yy[:,:,0],np.sum(S,axis=2)[:,:,ii],alpha=0.4)
            ax0.plot(Myxz[ii,1],Myxz[ii,0],'*b')
        ax0.set_xlabel("x [um]")
        ax0.set_ylabel("y [um]")
        ax0.legend(('GT','Gaussian','MLE'))
        
        ax1.contourf(zz[0,:,:],xx[0,:,:],np.sum(PT,axis=0),alpha=1.0)
        ax1.contourf(zz[0,:,:],xx[0,:,:],TPxz,alpha=0.2)
        ax1.plot(xyzpix[2],xyzpix[0],'+r')
        ax1.plot(Gxyz[2],Gxyz[0],'*g')
        ax1.plot(PTMxyz[2],PTMxyz[0],'xm')
        for ii in range(Myxz.shape[0]):
            #ax1.contourf(np.sum(S,axis=0)[:,:,ii],alpha=0.3)
            ax1.plot(Myxz[ii,2],Myxz[ii,1],'*b')
        ax1.set_xlabel("z [um]")
        ax1.set_ylabel("x [um]")

        ax2.contourf(zz[:,0,:],yy[:,0,:],np.sum(PT,axis=1),alpha=1.0)
        ax2.contourf(zz[:,0,:],yy[:,0,:],TPyz,alpha=0.2)
        ax2.plot(xyzpix[2],xyzpix[1],'+r')
        ax2.plot(Gxyz[2],Gxyz[1],'*g')
        ax2.plot(PTMxyz[2],PTMxyz[1],'xm')
        for ii in range(Myxz.shape[0]):
            #ax2.contourf(np.sum(S,axis=1)[:,:,ii],alpha=0.3)
            ax2.plot(Myxz[ii,2],Myxz[ii,0],'*b')
            
        ax2.set_xlabel("z [um]")
        ax2.set_ylabel("y [um]")

        ax3.plot(Vt,Vxyzpix[:,0],'-r'); ax4.plot(Vt,Vxyzpix[:,1],'-r'); ax5.plot(Vt,Vxyzpix[:,2],'-r')             
        ax3.plot(Vt,VGxyz[:,0],'-g'); ax4.plot(Vt,VGxyz[:,1],'-g'); ax5.plot(Vt,VGxyz[:,2],'-g')
        ax3.plot(Vt,VPTMxyz[:,0],'-m'); ax4.plot(Vt,VPTMxyz[:,1],'-m'); ax5.plot(Vt,VPTMxyz[:,2],'-m')
        ax3.set_ylabel("x [um]"); ax4.set_ylabel("y [um]"); ax5.set_ylabel("z [um]")
        ax3.grid(True); ax4.grid(True); ax5.grid(True)
        ax5.set_xlabel("t [ms]")
        
        DMLE=np.sqrt(np.sum((Vxyzpix-VPTMxyz)**2,axis=1)).flatten()
        DG  =np.sqrt(np.sum((Vxyzpix-VGxyz)**2,axis=1)).flatten()

        ax6.hist(DMLE, 20, facecolor='m', alpha=0.5)
        #ax6.hist(DG,   20, facecolor='g', alpha=0.5)
        ax6.set_xlabel("3D distance [nm]")
        #ax6.legend(('DMLE: ' + str(np.round(np.mean(DMLE),decimals=3)) + '+- ' + str(np.round(np.std(DMLE),decimals=3)),'DG: ' + str(np.round(np.mean(DG),decimals=3)) + '+- ' + str(np.round(np.std(DG),decimals=3))))

        #ax0.set_title("n_mean: " + str(np.round(n_mean, decimals=2)) )
        ax1.set_title("n: " + str(ni) + " =" + str(np.sum(ni)))
        ax2.set_title("DMLE: " + str(np.round(DMLE[i], decimals=3)) + " DG: " + str(np.round(DG[i], decimals=3)) )
        
        if do_save_plot:
            file="T_" + str(TrajToGet) +"_pos_" + str(i) + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

    def plotProbMap_Hist_Traj(ScanProject,LogPTot,S,xx,yy,zz,txyznM,n,TrajToGet,i,TopHatPrior,MemIndex,ZI,GroundTruth,do_save_plot=1):
        """        
        PT: probability map 3D 
        xyzpix: Groundtruth data 
        Gyxz:   Gaussian estimate ScanProject.TrajPosEstimation[TrajToGet]
        PTMyxz: MLE 
        Myxz: Laser shot positions 

        This function is a copy that was used under Simulator before but needs to be re orgenized
        """

        TPxy=np.max(TopHatPrior,axis=2)
        TPxy[TPxy<0]=np.nan
        TPxz=np.max(TopHatPrior,axis=0)
        TPxz[TPxz<0]=np.nan
        TPyz=np.max(TopHatPrior,axis=1)
        TPyz[TPyz<0]=np.nan

        PT    =np.exp(LogPTot)
        m = int(ScanProject.TrajectorySelectionRules['wndw'])
        ni=n[i*m:(i+1)*m]

            #time
        Vt=np.transpose(txyznM[0,:i+1])*1e3 #[ms]
            #MLE
        VPTMxyz=np.transpose(txyznM[1:4,:i+1])
        PTMxyz=VPTMxyz[-1,:]

            #Gaussian data
        VGxyz  =np.vstack((ScanProject.TrajPosEstimation[TrajToGet][1,:i+1],
                           ScanProject.TrajPosEstimation[TrajToGet][2,:i+1],
                           ScanProject.TrajPosEstimation[TrajToGet][-1,:i+1])) 
        VGxyz=np.transpose(VGxyz)
        VGxyz[:,2]=VGxyz[:,2]-VGxyz[0,2] + VPTMxyz[0,2]
        Gxyz=VGxyz[-1,:]

            #Groundtruth data 
        if np.prod(np.isnan(GroundTruth)):
            Vxyzpix=VGxyz
            xyzpix=Vxyzpix[-1,:] #[um] using the last point to show
        else:
            #Ground truth data is set to be the Gaussian data 
            Vxyzpix=GroundTruth[:(i+1)*m:m,:,0]*1e6 #[um]
            xyzpix=Vxyzpix[-1,:] #[um] using the last point to show


            #S matrix find peaks to mark out where the beams are
            #OBS! if the beams focus max are not with in the S matrix grid the result will be wrong 
        S[np.isnan(S)]=0
        S[np.where(S<np.max(S)*0.4)]=0
        Myxz=np.zeros((S.shape[3],3))
        for ii in range(S.shape[3]):
            indextosee=np.unravel_index(np.argmax(S[:,:,:,ii]),S[:,:,:,ii].shape)
            Myxz[ii,:]=[yy[indextosee],xx[indextosee],zz[indextosee]]   
        
        # Figure construction
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = GridSpec(5, 3, figure=fig)
        ax0=fig.add_subplot(gs[0, 0])
        ax1=fig.add_subplot(gs[0, 1])
        ax2=fig.add_subplot(gs[0, 2])
        ax3=fig.add_subplot(gs[1, 0:2])
        ax4=fig.add_subplot(gs[2, 0:2])
        ax5=fig.add_subplot(gs[3, 0:2])
        ax6=fig.add_subplot(gs[1, 2])
        ax7=fig.add_subplot(gs[2, 2])
        ax8=fig.add_subplot(gs[3, 2])
        ax9=fig.add_subplot(gs[4, 0:2])
        ax10=fig.add_subplot(gs[4, 2])

        #ax0 
        ax0.contourf(xx[:,:,0],yy[:,:,0],np.sum(PT,axis=2),alpha=1.0) #Likelihood projection to the xy plan
        ax0.contourf(xx[:,:,0],yy[:,:,0],TPxy,alpha=0.2)              #TopHat visualization  
        ax0.plot(xyzpix[0],xyzpix[1],'+r')                            #Groundtruth data 
        ax0.plot(Gxyz[0],Gxyz[1],'*g')                                #Gaussian data 
        ax0.plot(PTMxyz[0],PTMxyz[1],'xm')                            # MLE data
        for ii in range(Myxz.shape[0]):
            ax0.plot(Myxz[ii,1],Myxz[ii,0],'*b')                      #Beam Positions 

            #Configure ax0 plot  
        ax0.set_xlabel("x [um]")
        ax0.set_ylabel("y [um]")
        ax0.legend(('GT','Gaussian','MLE'))
        
        #ax1
        ax1.contourf(zz[0,:,:],xx[0,:,:],np.sum(PT,axis=0),alpha=1.0) #Likelihood projection to the zx plan
        ax1.contourf(zz[0,:,:],xx[0,:,:],TPxz,alpha=0.2)              #TopHat visualization  
        ax1.plot(xyzpix[2],xyzpix[0],'+r')                            #Groundtruth data 
        ax1.plot(Gxyz[2],Gxyz[0],'*g')                                #Gaussian data 
        ax1.plot(PTMxyz[2],PTMxyz[0],'xm')                            #MLE data
        for ii in range(Myxz.shape[0]):
            ax1.plot(Myxz[ii,2],Myxz[ii,1],'*b')                     #Beam Positions 

            #Configure ax1 plot 
        ax1.set_xlabel("z [um]")
        ax1.set_ylabel("x [um]")
        ax1.set_title("n: " + str(ni[np.argsort(MemIndex[i*m:(i+1)*m])]) + " =" + str(np.sum(ni)))
        
        #ax2
        ax2.contourf(zz[:,0,:],yy[:,0,:],np.sum(PT,axis=1),alpha=1.0)
        ax2.contourf(zz[:,0,:],yy[:,0,:],TPyz,alpha=0.2)
        ax2.plot(xyzpix[2],xyzpix[1],'+r')
        ax2.plot(Gxyz[2],Gxyz[1],'*g')
        ax2.plot(PTMxyz[2],PTMxyz[1],'xm')
        for ii in range(Myxz.shape[0]):
            #ax2.contourf(np.sum(S,axis=1)[:,:,ii],alpha=0.3)
            ax2.plot(Myxz[ii,2],Myxz[ii,0],'*b')
            
        ax2.set_xlabel("z [um]")
        ax2.set_ylabel("y [um]")

        #ax3-ax5
        ax3.plot(Vt,Vxyzpix[:,0],'-r',alpha=0.5); ax4.plot(Vt,Vxyzpix[:,1],'-r',alpha=0.5); ax5.plot(Vt,Vxyzpix[:,2],'-r',alpha=0.5)   #Groundtruth data            
        ax3.plot(Vt,VGxyz[:,0],'-g',alpha=0.5); ax4.plot(Vt,VGxyz[:,1],'-g',alpha=0.5); ax5.plot(Vt,VGxyz[:,2],'-g',alpha=0.5)         #Gaussian data 
        ax3.plot(Vt,VPTMxyz[:,0],'-m',alpha=0.5); ax4.plot(Vt,VPTMxyz[:,1],'-m',alpha=0.5); ax5.plot(Vt,VPTMxyz[:,2],'-m',alpha=0.5)   #MLE data
        ax3.set_ylabel("x [um]"); ax4.set_ylabel("y [um]"); ax5.set_ylabel("z [um]")
        ax3.grid(True); ax4.grid(True); ax5.grid(True)
        #ax5.set_xlabel("t [ms]")
        

        #ax6-ax8
        if np.prod(np.isnan(GroundTruth)):
            ax6.hist(VPTMxyz[:,0], 20, facecolor='m', alpha=0.5); ax7.hist(VPTMxyz[:,1], 20, facecolor='m', alpha=0.5); ax8.hist(VPTMxyz[:,2], 20, facecolor='m', alpha=0.5)
            ax6.hist(VGxyz[:,0],   20, facecolor='g', alpha=0.5); ax7.hist(VGxyz[:,1],   20, facecolor='g', alpha=0.5); ax8.hist(VGxyz[:,2],   20, facecolor='g', alpha=0.5)
            ax8.set_xlabel("Self distance [nm]")
        else: 
            ax6.hist(VPTMxyz[:,0]-Vxyzpix[:,0], 20, facecolor='m', alpha=0.5); ax7.hist(VPTMxyz[:,1]-Vxyzpix[:,1], 20, facecolor='m', alpha=0.5); ax8.hist(VPTMxyz[:,2]-Vxyzpix[:,2], 20, facecolor='m', alpha=0.5)
            ax6.hist(VGxyz[:,0]-Vxyzpix[:,0],   20, facecolor='g', alpha=0.5); ax7.hist(VGxyz[:,1]-Vxyzpix[:,1],   20, facecolor='g', alpha=0.5); ax8.hist(VGxyz[:,2]-Vxyzpix[:,2],   20, facecolor='g', alpha=0.5)
            ax8.set_xlabel("Groundtruth distance [nm]")
        #ax6.legend(('DMLE: ' + str(np.round(np.mean(DMLE),decimals=3)) + '+- ' + str(np.round(np.std(DMLE),decimals=3)),'DG: ' + str(np.round(np.mean(DG),decimals=3)) + '+- ' + str(np.round(np.std(DG),decimals=3))))
        #ax0.set_title("n_mean: " + str(np.round(n_mean, decimals=2)) )
        

        #ax9-ax10
        Elements=np.unique(MemIndex)
        wndw=int(m/len(Elements))
        wndwS=wndw #if 1 it is a moving average if equal to wndw it is like a fix window average
        color=cm.rainbow(np.linspace(0,1,len(Elements)))
        legendString=[]
        for MemUnique,c in zip(Elements,color):
            OfInterest=np.where(MemIndex==MemUnique)
            nT=n[OfInterest]
            nAV = np.convolve(np.ones(wndw),nT)
            if wndw==1:
                nAV=nAV[:i+1]
            else:
                nAV=nAV[wndw-1:-wndw+1:wndwS][:i+1]

            ax9.plot(Vt,((-1)**MemUnique)*nAV,'-',c=c,alpha=0.5) 
            ax10.hist(((-1)**MemUnique)*nAV, 20, facecolor=c, alpha=0.5)
            legendString.append('M' + str(int(MemUnique)) + ' ZI' + str(ZI[OfInterest][-1]) )

        ax9.set_xlabel('Time [ms]')
        ax9.set_ylabel('Photin #')
        ax9.legend(tuple(legendString))
        ax9.grid()
        
        if do_save_plot:
            file="T_" + str(TrajToGet) +"_pos_" + str(i) + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

    def plot_Hist_Traj(ScanProject,TrajToGet,do_save_plot=1):
        """
        This function is to plot trajectores
        """
        GroundTruth=np.nan
        txyzMSD=ScanProject.txyzMSD_Poisson_ProbMap[TrajToGet]
        txyznM=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]

        m = int(ScanProject.TrajectorySelectionRules['wndw'])
        ni=txyznM[4,:]

            #time
        Vt=(txyznM[0,:]-txyznM[0,0])*1e3 #[ms]
            #MLE
        VPTMxyz=np.transpose(txyznM[1:4,:])
            #Gaussian data
        VGxyz  =np.vstack((ScanProject.TrajPosEstimation[TrajToGet][1,:],
                           ScanProject.TrajPosEstimation[TrajToGet][2,:],
                           ScanProject.TrajPosEstimation[TrajToGet][-1,:])) 
        VGxyz=np.transpose(VGxyz)
        VGxyz[:,2]=VGxyz[:,2]-VGxyz[0,2] + VPTMxyz[0,2]

           #MSD data
        VPTMxyz_MSD=np.transpose(txyzMSD[1:4,:])

           #Groundtruth data 
        if np.prod(np.isnan(GroundTruth)):
            #Ground truth data is set to be the Gaussian data 
            Vxyzpix=VGxyz
        else:
            Vxyzpix=GroundTruth[::m,:,0]*1e6 #[um]

      
        # Figure construction 
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = GridSpec(7, 3, figure=fig)
        ax0=fig.add_subplot(gs[1, 0:2]) #MSD x
        ax1=fig.add_subplot(gs[3, 0:2]) #MSD y
        ax2=fig.add_subplot(gs[5, 0:2]) #MSD z

        ax3=fig.add_subplot(gs[0, 0:2]) #pos x
        ax4=fig.add_subplot(gs[2, 0:2]) #pos y
        ax5=fig.add_subplot(gs[4, 0:2]) #pos z

        ax6=fig.add_subplot(gs[0, 2])   #hist x
        ax7=fig.add_subplot(gs[2, 2])   #hist y
        ax8=fig.add_subplot(gs[4, 2])   #hist z

        ax9=fig.add_subplot(gs[6, 0:2]) #Photon count 
        ax10=fig.add_subplot(gs[6, 2])  #Photon hist


        #ax0-ax3
        ax0.plot(Vt,VPTMxyz_MSD[:,0],'-m',alpha=0.5); ax1.plot(Vt,VPTMxyz_MSD[:,1],'-m',alpha=0.5); ax2.plot(Vt,VPTMxyz_MSD[:,2],'-m',alpha=0.5)   #MSD MLE data
        ax0.set_ylabel("MSD x [um]"); ax1.set_ylabel("MSD y [um]"); ax2.set_ylabel("MSD z [um]")
        ax0.grid(True); ax1.grid(True); ax2.grid(True)
        ax0.set_title("Traj " + str(TrajToGet) )

        #ax3-ax5
        ax3.plot(Vt,Vxyzpix[:,0],'-r',alpha=0.5); ax4.plot(Vt,Vxyzpix[:,1],'-r',alpha=0.5); ax5.plot(Vt,Vxyzpix[:,2],'-r',alpha=0.5)   #Groundtruth data            
        ax3.plot(Vt,VGxyz[:,0],'-g',alpha=0.5);   ax4.plot(Vt,VGxyz[:,1],'-g',alpha=0.5);   ax5.plot(Vt,VGxyz[:,2],'-g',alpha=0.5)         #Gaussian data 
        ax3.plot(Vt,VPTMxyz[:,0],'-m',alpha=0.5); ax4.plot(Vt,VPTMxyz[:,1],'-m',alpha=0.5); ax5.plot(Vt,VPTMxyz[:,2],'-m',alpha=0.5)   #MLE data
        ax3.set_ylabel("x [um]"); ax4.set_ylabel("y [um]"); ax5.set_ylabel("z [um]")
        ax3.grid(True); ax4.grid(True); ax5.grid(True)
        #ax5.set_xlabel("t [ms]")
        

        #ax6-ax8
        if np.prod(np.isnan(GroundTruth)):
            ax6.hist(VPTMxyz[:,0], 20, facecolor='m', alpha=0.5); ax7.hist(VPTMxyz[:,1], 20, facecolor='m', alpha=0.5); ax8.hist(VPTMxyz[:,2], 20, facecolor='m', alpha=0.5)
            ax6.hist(VGxyz[:,0],   20, facecolor='g', alpha=0.5); ax7.hist(VGxyz[:,1],   20, facecolor='g', alpha=0.5); ax8.hist(VGxyz[:,2],   20, facecolor='g', alpha=0.5)
            ax8.set_xlabel("Self distance [nm]")
        else: 
            ax6.hist(VPTMxyz[:,0]-Vxyzpix[:,0], 20, facecolor='m', alpha=0.5); ax7.hist(VPTMxyz[:,1]-Vxyzpix[:,1], 20, facecolor='m', alpha=0.5); ax8.hist(VPTMxyz[:,2]-Vxyzpix[:,2], 20, facecolor='m', alpha=0.5)
            ax6.hist(VGxyz[:,0]-Vxyzpix[:,0],   20, facecolor='g', alpha=0.5); ax7.hist(VGxyz[:,1]-Vxyzpix[:,1],   20, facecolor='g', alpha=0.5); ax8.hist(VGxyz[:,2]-Vxyzpix[:,2],   20, facecolor='g', alpha=0.5)
            ax8.set_xlabel("Groundtruth distance [nm]")
        #ax6.legend(('DMLE: ' + str(np.round(np.mean(DMLE),decimals=3)) + '+- ' + str(np.round(np.std(DMLE),decimals=3)),'DG: ' + str(np.round(np.mean(DG),decimals=3)) + '+- ' + str(np.round(np.std(DG),decimals=3))))
        
        

        #ax9-ax10
        ax9.plot(Vt,ni,'-') 
        ax10.hist(ni, 20)        

        ax9.set_xlabel('Time [ms]')
        ax9.set_ylabel('Photons #')
        ax9.grid()
        
        if do_save_plot:
            file="T_" + str(TrajToGet) + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

    def plot_Hist_Traj_MDS(ScanProject,TrajToGet,do_save_plot=1):
        """
        This function is to plot trajectores
        """
        GroundTruth=np.nan
        txyznM=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]

            #MSD data 
        for steps in [20]:#range(int(len(ScanProject.txyzMSD_traj)/len(ScanProject.PoissonProbXYZMap_NumPFS))):
            txyzMSD=ScanProject.txyzMSD_traj[str(TrajToGet),str(steps)]
        txyzMSD[:,0]=(txyzMSD[:,0]-txyzMSD[0,0])*1e3 #[ms]
            #CVE MDS
        txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D'] #[mu^2/s]
        txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

        txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma'] #[mu]
        txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]

        m = int(ScanProject.TrajectorySelectionRules['wndw'])
        ni=txyznM[4,:]

            #time
        Vt=(txyznM[0,:]-txyznM[0,0])*1e3 #[ms]
            #MLE
        VPTMxyz=np.transpose(txyznM[1:4,:])
            #Gaussian data
        VGxyz  =np.vstack((ScanProject.TrajPosEstimation[TrajToGet][1,:],   #x
                           ScanProject.TrajPosEstimation[TrajToGet][2,:],   #y
                           ScanProject.TrajPosEstimation[TrajToGet][-1,:])) #z
        VGxyz=np.transpose(VGxyz)
        VGxyz[:,2]=VGxyz[:,2]-VGxyz[0,2] + VPTMxyz[0,2]
            #Polarization 
        Pol=ScanProject.TrajPosEstimation[TrajToGet][3,:]
           #MSD data
        #VPTMxyz_MSD=np.transpose(txyzMSD[1:4,:])

           #Groundtruth data 
        if np.prod(np.isnan(GroundTruth)):
            #Ground truth data is set to be the Gaussian data 
            Vxyzpix=VGxyz
        else:
            Vxyzpix=GroundTruth[::m,:,0]*1e6 #[um]

      
        # Figure construction
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = GridSpec(8, 3, figure=fig)
        ax0=fig.add_subplot(gs[1, 0:2]) #MSD x
        ax1=fig.add_subplot(gs[3, 0:2]) #MSD y
        ax2=fig.add_subplot(gs[5, 0:2]) #MSD z

        ax3=fig.add_subplot(gs[0, 0:2]) #pos x
        ax4=fig.add_subplot(gs[2, 0:2]) #pos y
        ax5=fig.add_subplot(gs[4, 0:2]) #pos z

        ax6=fig.add_subplot(gs[0, 2])   #hist x
        ax7=fig.add_subplot(gs[2, 2])   #hist y
        ax8=fig.add_subplot(gs[4, 2])   #hist z

        ax9=fig.add_subplot(gs[6, 0:2]) #Photon count 
        ax10=fig.add_subplot(gs[6, 2])  #Photon hist

        ax11=fig.add_subplot(gs[7, 0:2]) #Polarization 
        ax12=fig.add_subplot(gs[7, 2]) #Polarization hist

        ax13=fig.add_subplot(gs[1, 2]) #hist MSD x
        ax14=fig.add_subplot(gs[3, 2]) #hist MSD y
        ax15=fig.add_subplot(gs[5, 2]) #hist MSD z


        #ax0-ax3
        ax0.plot(txyzMSD[:,0],txyzMSD[:,1],'-m',alpha=0.5); ax1.plot(txyzMSD[:,0],txyzMSD[:,2],'-m',alpha=0.5); ax2.plot(txyzMSD[:,0],txyzMSD[:,3],'-m',alpha=0.5)   #MSD MLE data
        ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,1],'-g',alpha=0.5); ax1.plot(txyzCVE_D[:,0],txyzCVE_D[:,2],'-g',alpha=0.5); ax2.plot(txyzCVE_D[:,0],txyzCVE_D[:,3],'-g',alpha=0.5)   #D CVE MLE data
        ax0.plot(txyzCVE_Sigma[:,0],txyzCVE_Sigma[:,1],'-r',alpha=0.5); ax1.plot(txyzCVE_Sigma[:,0],txyzCVE_Sigma[:,2],'-r',alpha=0.5); ax2.plot(txyzCVE_Sigma[:,0],txyzCVE_Sigma[:,3],'-r',alpha=0.5)   #D CVE MLE data
        ax0.set_ylabel("Dx [um2/s]"); ax1.set_ylabel("Dy [um2/s]"); ax2.set_ylabel("Dz [um2/s]")
        ax0.grid(True); ax1.grid(True); ax2.grid(True)


        #ax3-ax5
        ax3.plot(Vt,Vxyzpix[:,0],'-r',alpha=0.5); ax4.plot(Vt,Vxyzpix[:,1],'-r',alpha=0.5); ax5.plot(Vt,Vxyzpix[:,2],'-r',alpha=0.5)   #Groundtruth data            
        ax3.plot(Vt,VGxyz[:,0],'-g',alpha=0.5);   ax4.plot(Vt,VGxyz[:,1],'-g',alpha=0.5);   ax5.plot(Vt,VGxyz[:,2],'-g',alpha=0.5)         #Gaussian data 
        ax3.plot(Vt,VPTMxyz[:,0],'-m',alpha=0.5); ax4.plot(Vt,VPTMxyz[:,1],'-m',alpha=0.5); ax5.plot(Vt,VPTMxyz[:,2],'-m',alpha=0.5)   #MLE data
        ax3.set_ylabel("x [um]"); ax4.set_ylabel("y [um]"); ax5.set_ylabel("z [um]")
        ax3.grid(True); ax4.grid(True); ax5.grid(True)
        #ax5.set_xlabel("t [ms]")
        

        #ax6-ax8
        if np.prod(np.isnan(GroundTruth)):
            if np.sum(np.isnan(VPTMxyz))==0:
                ax6.hist(VPTMxyz[:,0], 20, facecolor='m', alpha=0.5); ax7.hist(VPTMxyz[:,1], 20, facecolor='m', alpha=0.5); ax8.hist(VPTMxyz[:,2], 20, facecolor='m', alpha=0.5)
            if np.sum(np.isnan(VGxyz))==0:
                ax6.hist(VGxyz[:,0],   20, facecolor='g', alpha=0.5); ax7.hist(VGxyz[:,1],   20, facecolor='g', alpha=0.5); ax8.hist(VGxyz[:,2],   20, facecolor='g', alpha=0.5)
            ax8.set_xlabel("Self distance [nm]")
        else: 
            if np.sum(np.isnan(VPTMxyz))==0:
                    ax6.hist(VPTMxyz[:,0]-Vxyzpix[:,0], 20, facecolor='m', alpha=0.5); ax7.hist(VPTMxyz[:,1]-Vxyzpix[:,1], 20, facecolor='m', alpha=0.5); ax8.hist(VPTMxyz[:,2]-Vxyzpix[:,2], 20, facecolor='m', alpha=0.5)
            if np.sum(np.isnan(VGxyz))==0:
                    ax6.hist(VGxyz[:,0]-Vxyzpix[:,0],   20, facecolor='g', alpha=0.5); ax7.hist(VGxyz[:,1]-Vxyzpix[:,1],   20, facecolor='g', alpha=0.5); ax8.hist(VGxyz[:,2]-Vxyzpix[:,2],   20, facecolor='g', alpha=0.5)
            ax8.set_xlabel("Groundtruth distance [nm]")
        #ax6.legend(('DMLE: ' + str(np.round(np.mean(DMLE),decimals=3)) + '+- ' + str(np.round(np.std(DMLE),decimals=3)),'DG: ' + str(np.round(np.mean(DG),decimals=3)) + '+- ' + str(np.round(np.std(DG),decimals=3))))
        #ax0.set_title("n_mean: " + str(np.round(n_mean, decimals=2)) )
        

        #ax9-ax10 Photon counts
        ax9.plot(Vt,ni,'-') 
        ax10.hist(ni, 20)        
        ax9.set_ylabel('Photons #')
        ax9.grid()

        #ax11-ax12 Polarization counts
        ax11.plot(Vt,Pol,'-') 
        ax12.hist(Pol, 20)        

        ax11.set_xlabel('Time [ms]')
        ax11.set_ylabel('Polarization #')
        ax11.grid()
        

        #ax13-ax15
        ax13.hist(txyzMSD[:,1], 20, facecolor='m', alpha=0.5); ax14.hist(txyzMSD[:,2], 20, facecolor='m', alpha=0.5); ax15.hist(txyzMSD[:,3], 20, facecolor='m', alpha=0.5)
        ax13.hist(txyzCVE_D[:,1], 20, facecolor='g', alpha=0.5); ax14.hist(txyzCVE_D[:,2],   20, facecolor='g', alpha=0.5); ax15.hist(txyzCVE_D[:,3],   20, facecolor='g', alpha=0.5)
        ax15.set_xlabel("D [um^2/s]")

        #########################################################
        if do_save_plot:
            file="T_" + str(TrajToGet) + '_MSD_traj' + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

    def plot_Hist_Traj_MDS2(ScanProject,TrajToGet,do_save_plot=1):
        """
        This function is to plot trajectores
        """
        GroundTruth=np.nan
        txyznM=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]

            #Diffusion data 
        txyzMSD_D=np.zeros((len(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]),4))
        txyzMSD_D[:,0]=ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]
        txyzMSD_D[:,0]=(txyzMSD_D[:,0]-txyzMSD_D[0,0])*1e3 #[ms]
        txyzMSD_D[:,1:4]=np.transpose(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][1])

            #Localization error data 
        txyzMSD_Sigma=np.zeros((len(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]),4))
        txyzMSD_Sigma[:,0]=ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]
        txyzMSD_Sigma[:,0]=(txyzMSD_Sigma[:,0]-txyzMSD_Sigma[0,0])*1e3 #[ms]
        txyzMSD_Sigma[:,1:4]=np.transpose(np.sqrt(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][2]))

            #CVE MDS
        txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D'] #[mu^2/s]
        txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

        txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2'] #[mu]
        txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]

        m = int(ScanProject.TrajectorySelectionRules['wndw'])
        ni=txyznM[4,:]

            #time
        Vt=(txyznM[0,:]-txyznM[0,0])*1e3 #[ms]
            #MLE
        VPTMxyz=np.transpose(txyznM[1:4,:])
            #Gaussian data
        VGxyz  =np.vstack((ScanProject.TrajPosEstimation[TrajToGet][1,:],   #x
                           ScanProject.TrajPosEstimation[TrajToGet][2,:],   #y
                           ScanProject.TrajPosEstimation[TrajToGet][-1,:])) #z
        VGxyz=np.transpose(VGxyz)
        VGxyz[:,2]=VGxyz[:,2]-VGxyz[0,2] + VPTMxyz[0,2]
            #Polarization 
        Pol=ScanProject.TrajPosEstimation[TrajToGet][3,:]
           #MSD data
        #VPTMxyz_MSD=np.transpose(txyzMSD[1:4,:])

           #Groundtruth data 
        if np.prod(np.isnan(GroundTruth)):
            #Ground truth data is set to be the Gaussian data 
            Vxyzpix=VGxyz
        else:
            Vxyzpix=GroundTruth[::m,:,0]*1e6 #[um]

      
        # Figure construction
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = GridSpec(8, 3, figure=fig)
        ax0=fig.add_subplot(gs[1, 0:2]) #MSD x
        ax1=fig.add_subplot(gs[3, 0:2]) #MSD y
        ax2=fig.add_subplot(gs[5, 0:2]) #MSD z

        ax3=fig.add_subplot(gs[0, 0:2]) #pos x
        ax4=fig.add_subplot(gs[2, 0:2]) #pos y
        ax5=fig.add_subplot(gs[4, 0:2]) #pos z

        ax6=fig.add_subplot(gs[0, 2])   #hist x
        ax7=fig.add_subplot(gs[2, 2])   #hist y
        ax8=fig.add_subplot(gs[4, 2])   #hist z

        ax9=fig.add_subplot(gs[6, 0:2]) #Photon count 
        ax10=fig.add_subplot(gs[6, 2])  #Photon hist

        ax11=fig.add_subplot(gs[7, 0:2]) #Polarization 
        ax12=fig.add_subplot(gs[7, 2]) #Polarization hist

        ax13=fig.add_subplot(gs[1, 2]) #hist MSD x
        ax14=fig.add_subplot(gs[3, 2]) #hist MSD y
        ax15=fig.add_subplot(gs[5, 2]) #hist MSD z


        #ax0-ax3
        ax0.plot(txyzMSD_D[:,0],txyzMSD_D[:,1],'-b',alpha=0.5); ax1.plot(txyzMSD_D[:,0],txyzMSD_D[:,2],'-b',alpha=0.5); ax2.plot(txyzMSD_D[:,0],txyzMSD_D[:,3],'-b',alpha=0.5)   #MSD MLE data
        ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,1],'-g',alpha=0.5); ax1.plot(txyzCVE_D[:,0],txyzCVE_D[:,2],'-g',alpha=0.5); ax2.plot(txyzCVE_D[:,0],txyzCVE_D[:,3],'-g',alpha=0.5)   #D CVE MLE data
        ax0.plot(txyzCVE_Sigma[:,0],txyzCVE_Sigma[:,1],'-r',alpha=0.5); ax1.plot(txyzCVE_Sigma[:,0],txyzCVE_Sigma[:,2],'-r',alpha=0.5); ax2.plot(txyzCVE_Sigma[:,0],txyzCVE_Sigma[:,3],'-r',alpha=0.5)   #D CVE MLE data
        ax0.plot(txyzMSD_Sigma[:,0],txyzMSD_Sigma[:,1],'-m',alpha=0.5); ax1.plot(txyzMSD_Sigma[:,0],txyzMSD_Sigma[:,2],'-m',alpha=0.5); ax2.plot(txyzMSD_Sigma[:,0],txyzMSD_Sigma[:,3],'-m',alpha=0.5)   #D CVE MLE data
        ax0.set_ylabel("Dx [um2/s]"); ax1.set_ylabel("Dy [um2/s]"); ax2.set_ylabel("Dz [um2/s]")
        ax0.grid(True); ax1.grid(True); ax2.grid(True)


        #ax3-ax5
        ax3.plot(Vt,Vxyzpix[:,0],'-r',alpha=0.5); ax4.plot(Vt,Vxyzpix[:,1],'-r',alpha=0.5); ax5.plot(Vt,Vxyzpix[:,2],'-r',alpha=0.5)   #Groundtruth data            
        ax3.plot(Vt,VGxyz[:,0],'-g',alpha=0.5);   ax4.plot(Vt,VGxyz[:,1],'-g',alpha=0.5);   ax5.plot(Vt,VGxyz[:,2],'-g',alpha=0.5)         #Gaussian data 
        ax3.plot(Vt,VPTMxyz[:,0],'-m',alpha=0.5); ax4.plot(Vt,VPTMxyz[:,1],'-m',alpha=0.5); ax5.plot(Vt,VPTMxyz[:,2],'-m',alpha=0.5)   #MLE data
        ax3.set_ylabel("x [um]"); ax4.set_ylabel("y [um]"); ax5.set_ylabel("z [um]")
        ax3.grid(True); ax4.grid(True); ax5.grid(True)
        #ax5.set_xlabel("t [ms]")
        
        #ax6-ax8
        if np.prod(np.isnan(GroundTruth)):
            if np.sum(np.isnan(VPTMxyz))==0:
                ax6.hist(VPTMxyz[:,0], 20, facecolor='m', alpha=0.5); ax7.hist(VPTMxyz[:,1], 20, facecolor='m', alpha=0.5); ax8.hist(VPTMxyz[:,2], 20, facecolor='m', alpha=0.5)
            if np.sum(np.isnan(VGxyz))==0:
                ax6.hist(VGxyz[:,0],   20, facecolor='g', alpha=0.5); ax7.hist(VGxyz[:,1],   20, facecolor='g', alpha=0.5); ax8.hist(VGxyz[:,2],   20, facecolor='g', alpha=0.5)
            ax8.set_xlabel("Self distance [nm]")
        else: 
            if np.sum(np.isnan(VPTMxyz))==0:
                    ax6.hist(VPTMxyz[:,0]-Vxyzpix[:,0], 20, facecolor='m', alpha=0.5); ax7.hist(VPTMxyz[:,1]-Vxyzpix[:,1], 20, facecolor='m', alpha=0.5); ax8.hist(VPTMxyz[:,2]-Vxyzpix[:,2], 20, facecolor='m', alpha=0.5)
            if np.sum(np.isnan(VGxyz))==0:
                    ax6.hist(VGxyz[:,0]-Vxyzpix[:,0],   20, facecolor='g', alpha=0.5); ax7.hist(VGxyz[:,1]-Vxyzpix[:,1],   20, facecolor='g', alpha=0.5); ax8.hist(VGxyz[:,2]-Vxyzpix[:,2],   20, facecolor='g', alpha=0.5)
            ax8.set_xlabel("Groundtruth distance [nm]")
        #ax6.legend(('DMLE: ' + str(np.round(np.mean(DMLE),decimals=3)) + '+- ' + str(np.round(np.std(DMLE),decimals=3)),'DG: ' + str(np.round(np.mean(DG),decimals=3)) + '+- ' + str(np.round(np.std(DG),decimals=3))))
        #ax0.set_title("n_mean: " + str(np.round(n_mean, decimals=2)) )
        

        #ax9-ax10 Photon counts
        ax9.plot(Vt,ni,'-') 
        ax10.hist(ni, 20)        
        ax9.set_ylabel('Photons #')
        ax9.grid()

        #ax11-ax12 Polarization counts
        ax11.plot(Vt,Pol,'-') 
        ax12.hist(Pol, 20)        

        ax11.set_xlabel('Time [ms]')
        ax11.set_ylabel('Polarization #')
        ax11.grid()
        

        #ax13-ax15
        ax13.hist(txyzMSD_D[:,1], 20, facecolor='m', alpha=0.5); ax14.hist(txyzMSD_D[:,2], 20, facecolor='m', alpha=0.5); ax15.hist(txyzMSD_D[:,3], 20, facecolor='m', alpha=0.5)
        ax13.hist(txyzCVE_D[:,1], 20, facecolor='g', alpha=0.5); ax14.hist(txyzCVE_D[:,2],   20, facecolor='g', alpha=0.5); ax15.hist(txyzCVE_D[:,3],   20, facecolor='g', alpha=0.5)
        ax15.set_xlabel("D [um^2/s]")

        #########################################################
        if do_save_plot:
            file="T_" + str(TrajToGet) + '_MSD_traj' + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

    def plot_Hist_Traj_MDS3(ScanProject,TrajToGet,do_save_plot=1):
        """
        This function is to plot trajectores
        """
        GroundTruth=np.nan
        txyznM=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]

            #Diffusion data 
        txyzMSD_D=np.zeros((len(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]),4))
        txyzMSD_D[:,0]=ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]
        txyzMSD_D[:,0]=(txyzMSD_D[:,0]-txyzMSD_D[0,0])*1e3 #[ms]
        txyzMSD_D[:,1:4]=np.transpose(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][1])

            #Localization error data 
        txyzMSD_Sigma=np.zeros((len(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]),4))
        txyzMSD_Sigma[:,0]=ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]
        txyzMSD_Sigma[:,0]=(txyzMSD_Sigma[:,0]-txyzMSD_Sigma[0,0])*1e3 #[ms]
        txyzMSD_Sigma[:,1:4]=np.transpose(np.sqrt(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][2]))

            #CVE MDS
        txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D'] #[mu^2/s]
        txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

        txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2'] #[mu]
        txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]

        m = int(ScanProject.TrajectorySelectionRules['wndw'])
        ni=txyznM[4,:]

            #time
        Vt=(txyznM[0,:]-txyznM[0,0])*1e3 #[ms]
            #MLE
        VPTMxyz=np.transpose(txyznM[1:4,:])
        VPTMxyz=VPTMxyz-VPTMxyz[0,:]
            #Gaussian data
        VGxyz  =np.vstack((ScanProject.TrajPosEstimation[TrajToGet][1,:],   #x
                           ScanProject.TrajPosEstimation[TrajToGet][2,:],   #y
                           ScanProject.TrajPosEstimation[TrajToGet][-1,:])) #z
        VGxyz=np.transpose(VGxyz)
        VGxyz[:,2]=VGxyz[:,2]-VGxyz[0,2] + VPTMxyz[0,2]
            #Polarization 
        Pol=ScanProject.TrajPosEstimation[TrajToGet][3,:]
           #MSD data
        #VPTMxyz_MSD=np.transpose(txyzMSD[1:4,:])

           #Groundtruth data 
        if np.prod(np.isnan(GroundTruth)):
            #Ground truth data is set to be the Gaussian data 
            Vxyzpix=VGxyz
        else:
            Vxyzpix=GroundTruth[::m,:,0]*1e6 #[um]

      
        # Figure construction
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = GridSpec(8, 3, figure=fig)
        ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
        ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

        ax0=fig.add_subplot(gs[2:4, 0:2]) #MSD, CVE 
        ax1=fig.add_subplot(gs[4:6, 0:2]) #Sigma
        ax4=fig.add_subplot(gs[2:4, 2])  #hist MSD     
        ax5=fig.add_subplot(gs[4:6, 2])  #hist sigma2     

        ax9=fig.add_subplot(gs[6, 0:2]) #Photon count 
        ax10=fig.add_subplot(gs[6, 2])  #Photon hist

        ax11=fig.add_subplot(gs[7, 0:2]) #Polarization 
        ax12=fig.add_subplot(gs[7, 2]) #Polarization hist

        #ax0-ax1
        ax0.plot(txyzMSD_D[:,0],np.mean(txyzMSD_D[:,1:],axis=1),'-r',alpha=0.5, label='MSD');         #MSD MLE data
        ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:], axis=1),'-g',alpha=0.5, label='CVE');        #D CVE MLE data
        ax1.plot(txyzMSD_Sigma[:,0],np.mean(txyzMSD_Sigma[:,1:],axis=1),'-r',alpha=0.5, label='MSD'); #MSD MLE data
        ax1.plot(txyzCVE_Sigma[:,0],np.mean(txyzCVE_Sigma[:,1:],axis=1),'-g',alpha=0.5, label='CVE'); #D CVE MLE data
        
        ax0.set_ylabel("D [um2/s]"); 
        ax1.set_ylabel("sigma2 [um2]");
        ax0.grid(True); ax1.grid(True);
        ax0.legend();ax1.legend()

        #ax4-ax5
        ax4.hist(np.mean(txyzMSD_D[:,1:],axis=1), 20, facecolor='m', alpha=0.5); 
        ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), 20, facecolor='g', alpha=0.5); 
        ax4.set_xlabel("D [um^2/s]")

        ax5.hist(np.sqrt(np.mean(txyzCVE_Sigma[:,1:],axis=1))*1e3, 20, facecolor='g', alpha=0.5); 
        ax5.set_xlabel("Sigma [nm]")


        #ax3
        ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
        ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
        ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
        ax3.set_ylabel("[um]"); 
        ax3.grid(True);
        ax3.legend()
        
        #ax6
        if np.prod(np.isnan(GroundTruth)):
            if np.sum(np.isnan(VPTMxyz))==0:
                ax6.hist(VPTMxyz[:,0], 20, facecolor='r', alpha=0.4); 
                ax6.hist(VPTMxyz[:,1], 20, facecolor='g', alpha=0.4); 
                ax6.hist(VPTMxyz[:,2], 20, facecolor='b', alpha=0.4)          
                ax6.set_xlabel("[nm]")
        
        #ax9-ax10 Photon counts
        ax9.plot(Vt,ni,'-') 
        ax10.hist(ni, 20)        
        ax9.set_ylabel('Photons #')
        ax9.grid()

        #ax11-ax12 Polarization counts
        ax11.plot(Vt,Pol,'-') 
        ax12.hist(Pol, 20)        

        ax11.set_xlabel('Time [ms]')
        ax11.set_ylabel('Polarization #')
        ax11.grid()
        

        
        #########################################################
        if do_save_plot:
            file="T_" + str(TrajToGet) + '_MSD_traj' + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

        print('afd')

    def plot_Hist_Traj_MDS4(ScanProject,TrajToGet,do_save_plot=1):
        """
        This function is to plot trajectores
        """
        wndw=10
        GroundTruth=np.nan
        txyznM=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]*1.0

            #Diffusion data 
        txyzMSD_D=np.zeros((len(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]),4))
        txyzMSD_D[:,0]=ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]*1.0
        txyzMSD_D[:,0]=(txyzMSD_D[:,0]-txyzMSD_D[0,0])*1e3 #[ms]
        txyzMSD_D[:,1:4]=np.transpose(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][1]*1.0)

            #Localization error data 
        txyzMSD_Sigma=np.zeros((len(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]),4))
        txyzMSD_Sigma[:,0]=ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]*1.0
        txyzMSD_Sigma[:,0]=(txyzMSD_Sigma[:,0]-txyzMSD_Sigma[0,0])*1e3 #[ms]
        txyzMSD_Sigma[:,1:4]=np.transpose(np.sqrt(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][2]*1.0))

            #CVE MDS
        txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
        txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

        txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu]
        txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]

        m = int(ScanProject.TrajectorySelectionRules['wndw'])
        ni=txyznM[4,:]

            #time
        Vt=(txyznM[0,:]-txyznM[0,0])*1e3 #[ms]
            #MLE
        VPTMxyz=np.transpose(txyznM[1:4,:])
        VPTMxyz=VPTMxyz-VPTMxyz[0,:]
            #Running mean on the trajectory
        wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
        T_Mean= np.convolve(np.ones(wndw)/wndw,Vt)
        T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
        VPTMxyz_Mean=np.zeros((len(T_Mean),3))
        VPTMxyz_Var=np.zeros((len(T_Mean),3))
        for i in range(3):
            VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,VPTMxyz[:,i])
            VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

            VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(VPTMxyz[:,i])**2) #Var(X)=E[X^2]-E[X]^2
            VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2


            #Polarization 
        Pol=ScanProject.TrajPosEstimation[TrajToGet][3,:]*1.0
           #MSD data
        #VPTMxyz_MSD=np.transpose(txyzMSD[1:4,:])

      
        # Figure construction
        fig = plt.figure(constrained_layout=False)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = fig.add_gridspec(nrows=8, ncols=3, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.1, hspace=1.0)

        ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
        ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

        ax0=fig.add_subplot(gs[2:4, 0:2]) #MSD, CVE 
        ax1=fig.add_subplot(gs[4:6, 0:2]) #Sigma
        ax4=fig.add_subplot(gs[2:4, 2])  #hist D      
        ax5=fig.add_subplot(gs[4:6, 2])  #hist sigma2     

        ax9=fig.add_subplot(gs[6, 0:2]) #Photon count 
        ax10=fig.add_subplot(gs[6, 2])  #Photon hist

        ax11=fig.add_subplot(gs[7, 0:2]) #Polarization 
        ax12=fig.add_subplot(gs[7, 2]) #Polarization hist

        #ax0-ax1
        ax0.plot(txyzMSD_D[:,0],np.mean(txyzMSD_D[:,1:],axis=1),'-r',alpha=0.5, label='MSD');         #D MSD MLE data
        ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:], axis=1),'-g',alpha=0.5, label='CVE');        #D CVE MLE data
        #ax1.plot(txyzMSD_Sigma[:,0],np.mean(txyzMSD_Sigma[:,1:],axis=1),'-r',alpha=0.5, label='MSD'); #Sigma2 MSD MLE data
        ax1.plot(txyzCVE_Sigma[:,0],np.mean(txyzCVE_Sigma[:,1:],axis=1),'-g',alpha=0.5, label='CVE'); #Sigma2 CVE MLE data
        ax1.plot(T_Mean,np.mean(VPTMxyz_Var,axis=1),'-b',alpha=0.5, label='Var' + str(wndw));         #Var CVE MLE data
        
        ax0.set_ylabel("D [um2/s]"); 
        ax1.set_ylabel("sigma2 [um2]");
        ax0.grid(True); ax1.grid(True);
        #ax0.legend();#ax1.legend()

        #ax4 Histogram D CVE and MSD
        ax4.hist(np.mean(txyzMSD_D[:,1:],axis=1), 20, facecolor='r', alpha=0.5); 
        ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), 20, facecolor='g', alpha=0.5); 
        ax4.axvline(np.mean(txyzMSD_D[:,1:]), color='r',linestyle='-',label='MSD' + str(np.round(np.mean(txyzMSD_D[:,1:]),decimals=2)))
        ax4.axvline(np.mean(txyzCVE_D[:,1:]), color='g',linestyle='-',label='CVE:' + str(np.round(np.mean(txyzCVE_D[:,1:]),decimals=2)))
        ax4.set_xlabel("D [um^2/s]")
        ax4.legend()

        #ax5 Histogram Sigma Variance 
        ax5.hist(np.sqrt(np.mean(txyzCVE_Sigma[:,1:],axis=1))*1e3, 20, facecolor='g', alpha=0.5); 
        ax5.hist(np.sqrt(np.mean(VPTMxyz_Var,axis=1))*1e3, 20, facecolor='b', alpha=0.5); 
        ax5.axvline(np.sqrt(np.mean(VPTMxyz_Var))*1e3, color='b',linestyle='-',label='Var' + str(wndw) +':'+ str(np.round(np.sqrt(np.mean(VPTMxyz_Var))*1e3,decimals=2)))
        ax5.axvline(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3, color='g',linestyle='-',label='CVE:' + str(np.round(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3,decimals=2)))
        ax5.set_xlabel("Sigma [nm]")
        ax5.legend()

        #ax3
        ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
        ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
        ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
        ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.4,label='Mean:' + str(wndw));  #MLE data
        ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.4);  #MLE data
        ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.4)   #MLE data
        
        ax3.set_ylabel("[um]"); 
        ax3.grid(True);
        ax3.legend()
        
        #ax6
        if np.prod(np.isnan(GroundTruth)):
            if np.sum(np.isnan(VPTMxyz))==0:
                ax6.hist(np.sqrt(VPTMxyz_Var[:,0]), 20, facecolor='r', alpha=0.4); 
                ax6.hist(np.sqrt(VPTMxyz_Var[:,1]), 20, facecolor='g', alpha=0.4); 
                ax6.hist(np.sqrt(VPTMxyz_Var[:,2]), 20, facecolor='b', alpha=0.4)          
                ax6.set_xlabel('STD [um] over: ' + str(wndw))
        
        #ax9-ax10 Photon counts
        ax9.plot(Vt,ni,'-') 
        ax10.hist(ni, 20)        
        ax9.set_ylabel('Ph #')
        ax9.grid()

        #ax11-ax12 Polarization counts
        ax11.plot(Vt,Pol,'-') 
        ax12.hist(Pol, 20)        

        ax11.set_xlabel('Time [ms]')
        ax11.set_ylabel('Pol')
        ax11.grid()
        
        #########################################################
        if do_save_plot:
            file="T_" + str(TrajToGet) + '_MSD_traj' + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

    def plot_Hist_Traj_MDS5(ScanProject,TrajToGet,TrajDictionary,do_save_plot=0,FolderEnd=''):
        """
        This function is to plot trajectores
        """
        wndw=10
        txyznM=TrajDictionary[TrajToGet]*1.0

            #CVE D and sigma evaluation
        txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
        txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

        txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu]
        txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]

        ni=txyznM[4,:]

            #time
        Vt=(txyznM[0,:]-txyznM[0,0])*1e3 #[ms]
            #MLE
        VPTMxyz=np.transpose(txyznM[1:4,:])
        VPTMxyz=VPTMxyz-VPTMxyz[0,:]
            #Running mean on the trajectory
        wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
        T_Mean= np.convolve(np.ones(wndw)/wndw,Vt)
        T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
        VPTMxyz_Mean=np.zeros((len(T_Mean),3))
        VPTMxyz_Var=np.zeros((len(T_Mean),3))
        for i in range(3):
            VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,VPTMxyz[:,i])
            VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

            VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(VPTMxyz[:,i])**2) #Var(X)=E[X^2]-E[X]^2
            VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2


            #Polarization 
        #Pol=ScanProject.TrajPosEstimation[TrajToGet][3,:]*1.0
      
        # Figure construction
        fig = plt.figure(constrained_layout=False)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = fig.add_gridspec(nrows=7, ncols=3, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.1, hspace=1.0)

        ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
        ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

        ax0=fig.add_subplot(gs[2:4, 0:2]) #MSD, CVE 
        ax1=fig.add_subplot(gs[4:6, 0:2]) #Sigma
        ax4=fig.add_subplot(gs[2:4, 2])  #hist D      
        ax5=fig.add_subplot(gs[4:6, 2])  #hist sigma2     

        ax9=fig.add_subplot(gs[6, 0:2]) #Photon count 
        ax10=fig.add_subplot(gs[6, 2])  #Photon hist

        #ax11=fig.add_subplot(gs[7, 0:2]) #Polarization 
        #ax12=fig.add_subplot(gs[7, 2]) #Polarization hist

        #ax0-ax1
        ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:], axis=1),'-g',alpha=0.5, label='CVE');        #D CVE MLE data
        ax1.plot(txyzCVE_Sigma[:,0],np.mean(txyzCVE_Sigma[:,1:],axis=1),'-g',alpha=0.5, label='CVE'); #Sigma2 CVE MLE data
        ax1.plot(T_Mean,np.mean(VPTMxyz_Var,axis=1),'-b',alpha=0.5, label='Var' + str(wndw));         #Var CVE MLE data
        
        ax0.set_ylabel("D [um2/s]"); 
        ax1.set_ylabel("sigma2 [um2]");
        ax0.grid(True); ax1.grid(True);
        #ax0.legend();#ax1.legend()

        #ax4 Histogram D CVE and MSD
        ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), 20, facecolor='g', alpha=0.5); 
        ax4.axvline(np.mean(txyzCVE_D[:,1:]), color='g',linestyle='-',label='CVE:' + str(np.round(np.mean(txyzCVE_D[:,1:]),decimals=2)))
        ax4.set_xlabel("D [um^2/s]")
        ax4.legend()

        #ax5 Histogram Sigma Variance 
        ax5.hist(np.sqrt(np.mean(txyzCVE_Sigma[:,1:],axis=1))*1e3, 20, facecolor='g', alpha=0.5); 
        ax5.hist(np.sqrt(np.mean(VPTMxyz_Var,axis=1))*1e3, 20, facecolor='b', alpha=0.5); 
        ax5.axvline(np.sqrt(np.mean(VPTMxyz_Var))*1e3, color='b',linestyle='-',label='Var' + str(wndw) +':'+ str(np.round(np.sqrt(np.mean(VPTMxyz_Var))*1e3,decimals=2)))
        ax5.axvline(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3, color='g',linestyle='-',label='CVE:' + str(np.round(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3,decimals=2)))
        ax5.set_xlabel("Sigma [nm]")
        ax5.legend()

        #ax3
        ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
        ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
        ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
        ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.4,label='Mean:' + str(wndw));  #MLE data
        ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.4);  #MLE data
        ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.4)   #MLE data
        
        ax3.set_ylabel("[um]"); 
        ax3.grid(True);
        ax3.legend()
        
        #ax6
        if np.sum(np.isnan(VPTMxyz))==0:
            ax6.hist(np.sqrt(VPTMxyz_Var[:,0]), 20, facecolor='r', alpha=0.4); 
            ax6.hist(np.sqrt(VPTMxyz_Var[:,1]), 20, facecolor='g', alpha=0.4); 
            ax6.hist(np.sqrt(VPTMxyz_Var[:,2]), 20, facecolor='b', alpha=0.4)          
            ax6.set_xlabel('STD [um] over: ' + str(wndw))
        
        #ax9-ax10 Photon counts
        ax9.plot(Vt,ni,'-') 
        ax10.hist(ni, 20)        
        ax9.set_ylabel('Ph #')
        ax9.grid()

        #ax11-ax12 Polarization counts
        #ax11.plot(Vt,Pol,'-') 
        #ax12.hist(Pol, 20)        

        #ax11.set_xlabel('Time [ms]')
        #ax11.set_ylabel('Pol')
        #ax11.grid()
        
        #########################################################
        if do_save_plot:
            file="T_" + str(TrajToGet) + '_MSD_traj' + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_D_filtered' + FolderEnd + '/'
            Path(path).mkdir(parents=True, exist_ok=True)
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

    def plot_Hist_Traj_MDS6(ScanProject,TrajToGet,TrajDictionary,MovieStepSize=0,do_save_plot=0):
        """
        This function is to plot trajectores including cell image
        """
        xyImageBox=35 #pix
        CellMask=1
        
        if CellMask:
            mask=ScanProject.CamCellMask[:,:,0]>0
        CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]
        ImTot=ScanProject.ImgOverlay

        wndw=10
        txyznM=TrajDictionary[TrajToGet]*1.0

            #CVE D and sigma evaluation
        txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
        txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

        txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu]
        txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]

        ni=txyznM[4,:]

            #time
        Vt=(txyznM[0,:]-txyznM[0,0])*1e3 #[ms]
            #MLE
        VPTMxyz=np.transpose(txyznM[1:4,:])
        VPTMxyz=VPTMxyz-VPTMxyz[0,:]

            #Running mean on the trajectory
        wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
        T_Mean= np.convolve(np.ones(wndw)/wndw,Vt)
        T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
        VPTMxyz_Mean=np.zeros((len(T_Mean),3))
        VPTMxyz_Var=np.zeros((len(T_Mean),3))
        for i in range(3):
            VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,VPTMxyz[:,i])
            VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

            VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(VPTMxyz[:,i])**2) #Var(X)=E[X^2]-E[X]^2
            VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2

        #np.convolve(np.ones(wndw)/wndw,np.mean(txyzCVE_D[:,1:], axis=1))[wndw-1:-wndw+1:wndwS]
        

            #Create a D value colored traj plot
        x_pix=(VPTMxyz[:,0]+np.transpose(txyznM[1:4,:])[0,0])/CamRes
        y_pix=(VPTMxyz[:,1]+np.transpose(txyznM[1:4,:])[0,1])/CamRes
        points=np.array([x_pix, y_pix]).T.reshape(-1, 1, 2)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)
        
        D_norm=np.log(np.mean(txyzCVE_D[:,1:], axis=1)) #(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))/ np.max(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))
        D_norm[np.where(np.isnan(D_norm))]=0
        D_norm=1/(1+np.exp(-(D_norm)))
        t_D=np.concatenate((txyzCVE_D[:,0],[Vt[-1]]))
        D_traj=np.zeros(Vt.shape)
        j=0
        for i in range(1,len(t_D)):
            while Vt[j]<=t_D[i]:
                D_traj[j]=D_norm[i-1]
                j+=1
                if j>=len(Vt):
                    break

        colors = plt.cm.cool(D_traj)

            #Polarization 
        #Pol=ScanProject.TrajPosEstimation[TrajToGet][3,:]*1.0

        if MovieStepSize>0: FrameIndex=0
        else: FrameIndex=-1

        movieFlag=1
        while FrameIndex < len(Vt) and movieFlag==1:
        
            # Figure construction
            fig = plt.figure(constrained_layout=False)
            fig.set_size_inches(3*5, 2*5, forward=True)
            gs = fig.add_gridspec(nrows=7, ncols=5, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.2, hspace=1.0)

            ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
            ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

            ax0=fig.add_subplot(gs[2:4, 0:2]) #MSD, CVE 
            ax1=fig.add_subplot(gs[4:6, 0:2]) #Sigma
            ax4=fig.add_subplot(gs[2:4, 2])  #hist D      
            ax5=fig.add_subplot(gs[4:6, 2])  #hist sigma2     

            ax9=fig.add_subplot(gs[6, 0:2]) #Photon count 
            ax10=fig.add_subplot(gs[6, 2])  #Photon hist

            ax11=fig.add_subplot(gs[0:7, 3:5]) #Image 
        
            #ax0-ax1
            ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:], axis=1),'-g',alpha=0.5, label='CVE');        #D CVE MLE data
            #ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:], axis=1),'-r',alpha=0.5, label='CVE');        #D CVE MLE data
            if MovieStepSize>0: ax0.axvline(Vt[FrameIndex], color='r',linestyle='-')
            ax1.plot(txyzCVE_Sigma[:,0],np.mean(txyzCVE_Sigma[:,1:],axis=1),'-g',alpha=0.5, label='CVE'); #Sigma2 CVE MLE data
            ax1.plot(T_Mean,np.mean(VPTMxyz_Var,axis=1),'-b',alpha=0.5, label='Var' + str(wndw));         #Var CVE MLE data
            if MovieStepSize>0: ax1.axvline(Vt[FrameIndex], color='r',linestyle='-')

            ax0.set_ylabel("D [um2/s]"); 
            ax1.set_ylabel("sigma2 [um2]");
            ax0.grid(True); ax1.grid(True);
            #ax0.legend();#ax1.legend()

            #ax4 Histogram D CVE and MSD
            ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), 20, facecolor='g', alpha=0.5); 
            ax4.axvline(np.mean(txyzCVE_D[:,1:]), color='g',linestyle='-',label='CVE:' + str(np.round(np.mean(txyzCVE_D[:,1:]),decimals=2)))
            ax4.set_xlabel("D [um^2/s]")
            ax4.legend()

            #ax5 Histogram Sigma Variance 
            ax5.hist(np.sqrt(np.mean(txyzCVE_Sigma[:,1:],axis=1))*1e3, 20, facecolor='g', alpha=0.5); 
            ax5.hist(np.sqrt(np.mean(VPTMxyz_Var,axis=1))*1e3, 20, facecolor='b', alpha=0.5); 
            ax5.axvline(np.sqrt(np.mean(VPTMxyz_Var))*1e3, color='b',linestyle='-',label='Var' + str(wndw) +':'+ str(np.round(np.sqrt(np.mean(VPTMxyz_Var))*1e3,decimals=2)))
            ax5.axvline(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3, color='g',linestyle='-',label='CVE:' + str(np.round(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3,decimals=2)))
            ax5.set_xlabel("Sigma [nm]")
            ax5.legend()

            #ax3
            ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
            ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
            ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
            ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.4,label='Mean:' + str(wndw));  #MLE data
            ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.4);  #MLE data
            ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.4)   #MLE data
            if MovieStepSize>0: ax3.axvline(Vt[FrameIndex], color='r',linestyle='-')

            ax3.set_ylabel("[um]"); 
            ax3.grid(True);
            ax3.legend()
        
            #ax6
            if np.sum(np.isnan(VPTMxyz))==0:
                ax6.hist(np.sqrt(VPTMxyz_Var[:,0]), 20, facecolor='r', alpha=0.4); 
                ax6.hist(np.sqrt(VPTMxyz_Var[:,1]), 20, facecolor='g', alpha=0.4); 
                ax6.hist(np.sqrt(VPTMxyz_Var[:,2]), 20, facecolor='b', alpha=0.4)          
                ax6.set_xlabel('STD [um] over: ' + str(wndw))
        
            #ax9-ax10 Photon counts
            ax9.plot(Vt,ni,'-') 
            if MovieStepSize>0: ax9.axvline(Vt[FrameIndex], color='r',linestyle='-')
            ax10.hist(ni, 20)        
            ax9.set_ylabel('Ph #')
            ax9.grid()

            #Cell image
                #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
            ax11.imshow(ImTot[:,:,0],cmap='gray',alpha=1.0)
            #ax[0].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)
            #ax[0].imshow(ImTot[:,:,2],cmap='Reds',alpha=0.3)
            if CellMask:
                ax11.contour(mask,levels=0, linewidths=1, colors='blue', linestyles='solid')
                ax11.imshow(mask,cmap='Reds',alpha=0.3)

            if MovieStepSize>0: ls=LineCollection(segments[:FrameIndex], linewidths=1,colors=colors[:FrameIndex], linestyle='solid')
            else: ls=LineCollection(segments, linewidths=1,colors=colors, linestyle='solid')

            line=ax11.add_collection(ls)
            #fig.colorbar(line, ax=ax11)
            #ax11.plot(x,y,'-', c=color[0,:])

            xmean=np.mean(x_pix)
            ymean=np.mean(y_pix)
        
            ax11.set_xlim([xmean-xyImageBox,xmean+xyImageBox])
            ax11.set_ylim([ymean-xyImageBox,ymean+xyImageBox])
            ax11.set_xticks([]); ax11.set_yticks([])

            #########################################################
            if do_save_plot:
                if MovieStepSize>0:
                    file="MovieFrame_" + str(TrajToGet) + '_D_traj_' + str(FrameIndex)+ '.png'
                    path=ScanProject.Files['AnalysisConfig'][:-18] +'Movie_Traj_D_filtered/'
                    Path(path).mkdir(parents=True, exist_ok=True)
                else:
                    file="T_" + str(TrajToGet) + '_D_traj' + '.png'
                    path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_D_filtered/'
                    Path(path).mkdir(parents=True, exist_ok=True)
                fig.savefig(path + file)
                plt.close()
                print('Done!')
            else:
                plt.show()

            FrameIndex+=MovieStepSize
            if MovieStepSize<=0: break
            elif FrameIndex>=len(Vt):break

    def plot_Hist_Traj_MDS7(ScanProject,TrajToGet,TrajDictionary,HMM_prediction,HMM_Model,MovieStepSize=0,do_save_plot=0,FolderEnd='',fft=0):
            """
            This function is to plot 
                trajectores 
                Cell image
                HMM predicted traj
            """
            xyImageBox=35 #pix
            CellMask=1
        
            if CellMask:
                mask=ScanProject.CamCellMask[:,:,0]>0
            CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]
            try:
                ImTot=ScanProject.ImgOverlay
            except:
                print('No cell image found, using cell mask')
                ImTot=np.repeat(np.expand_dims(mask*1.0,axis=2),3,axis=2)

            wndw=10
            txyznM=TrajDictionary[TrajToGet]*1.0

                #CVE D and sigma evaluation
            txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
            txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

            txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu]
            txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]

            ni=txyznM[4,:]

                #time
            Vt=txyznM[0,:]              #[s]

                #Piezo stage position
            TStartEnd=[Vt[0],Vt[-1]] 
            StartEndPos=np.where(np.abs(np.diff((ScanProject.ScanMatrix[0,:]>TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<TStartEnd[1]))))[0]
            StartEndPos=np.expand_dims(np.append(np.array([0]),StartEndPos),axis=1)
            
            T_Raw,XP_Raw,YP_Raw,XE_Raw,YE_Raw,XO_Raw,YO_Raw,ZI_Raw,APD1_Raw,APD2_Raw,J_Raw,MemIndex_Raw,ZI_um_Raw,ZI_OffsetIndex_Raw =  _MS.get_traj_raw_data3(StartEndPos, 0, ScanProject.ScanMatrix, ScanProject)
            txyz_Traj_Piezo=np.vstack((T_Raw,XP_Raw,YP_Raw,ZI_um_Raw))
            txyz_Traj_Piezo=_MS.get_Scan_toImagCoordinates(ScanProject, txyz_Traj_Piezo)
            
            #ZeroPoint=np.array([XP_Raw[0],YP_Raw[0],ZI_um_Raw[0]])
                #MLE
            VPTMxyz_O=np.transpose(txyznM[1:4,:])
            
                #Start at zero time and at zero piezo pos
            VPTMxyz=VPTMxyz_O-txyz_Traj_Piezo[1:,0]#VPTMxyz_O[0,:]
            Vt=(Vt-Vt[0])*1e3 #[ms]

            txyz_Traj_Piezo=(txyz_Traj_Piezo.T-txyz_Traj_Piezo[:,0]).T
            txyz_Traj_Piezo[0,:]=txyz_Traj_Piezo[0,:]*1e3

                # Fourier spectra of the MLE signals
            if fft!=0:
                if 0:
                    wndw_fft=20
                    wndwS_fft=wndw_fft #if 1 it is a moving average, if equal to wndw it is a fix window average
                    T_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,Vt)
                    T_fft=T_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]

                    x_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,0])
                    x_fft=x_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    y_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,1])
                    y_fft=y_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    z_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,2])
                    z_fft=z_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    
                    fft_x_MLE, fx = _MS.fft_processing(x_fft,T_fft)
                    fft_y_MLE, fy = _MS.fft_processing(y_fft,T_fft)
                    fft_z_MLE, fz = _MS.fft_processing(z_fft,T_fft)
                else:
                    fft_x_MLE, fx = _MS.fft_processing(VPTMxyz[:,0],Vt)
                    fft_y_MLE, fy = _MS.fft_processing(VPTMxyz[:,1],Vt)
                    fft_z_MLE, fz = _MS.fft_processing(VPTMxyz[:,2],Vt)

                #Running mean on the trajectory
            wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            T_Mean= np.convolve(np.ones(wndw)/wndw,Vt)
            T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
            VPTMxyz_Mean=np.zeros((len(T_Mean),3))
            VPTMxyz_Var=np.zeros((len(T_Mean),3))
            for i in range(3):
                VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,VPTMxyz[:,i])
                VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

                VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(VPTMxyz[:,i])**2) #Var(X)=E[X^2]-E[X]^2
                VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2

            #np.convolve(np.ones(wndw)/wndw,np.mean(txyzCVE_D[:,1:], axis=1))[wndw-1:-wndw+1:wndwS]

                #Create a D value colored traj plot
            x_pix=(VPTMxyz_O[:,0])/CamRes
            y_pix=(VPTMxyz_O[:,1])/CamRes
            points=np.array([x_pix, y_pix]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
        
            D_norm=np.log(np.mean(txyzCVE_D[:,1:], axis=1)) #(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))/ np.max(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))
            D_norm[np.where(np.isnan(D_norm))]=0
            D_norm=1/(1+np.exp(-(D_norm)))
            t_D=np.concatenate((txyzCVE_D[:,0],[Vt[-1]]))
            D_traj=np.zeros(Vt.shape)
            j=0
            for i in range(1,len(t_D)):
                while Vt[j]<=t_D[i]:
                    D_traj[j]=D_norm[i-1]
                    j+=1
                    if j>=len(Vt):
                        break

            colors = plt.cm.cool(D_traj)

                #Polarization 
            wndwN=int(ScanProject.TrajectorySelectionRules['wndw'])
            OfInterest=np.where((ScanProject.ScanMatrix[0,:]>=TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<=TStartEnd[1]))[0]
            OfInterest=np.append(OfInterest,OfInterest[-1]+np.arange(wndwN)+1)
            N12=ScanProject.ScanMatrix[3:5,OfInterest]
            PolT=ScanProject.ScanMatrix[0,OfInterest]

            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1=np.convolve(np.ones(wndwN)/wndwN,N12[0,:])
            N1=N1[wndwN-1:-wndwN+1:wndwS]
            N2=np.convolve(np.ones(wndwN)/wndwN,N12[1,:])
            N2=N2[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]

            
            wndwN=5
            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1_mean=np.convolve(np.ones(wndwN)/wndwN,N1)
            N1_mean=N1_mean[wndwN-1:-wndwN+1:wndwS]
            N2_mean=np.convolve(np.ones(wndwN)/wndwN,N2)
            N2_mean=N2_mean[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]
            PolT=(PolT-PolT[0])*1e3 #ms
            Pol=(N1_mean-N2_mean)/(N1_mean+N2_mean)

        ####################################
            # Generate the inverse-Gamma distributions to plot
            dist_params=HMM_Model['Distributions']
            Occupancy_estimated=HMM_Model['Occupancy']
            DBias=HMM_Model['DBias']

            DFittParam=np.linspace(-DBias,10,num=200)
            invGamma={}
            invGammaAll=np.zeros(DFittParam.shape)
            for state,Occ in zip(dist_params,Occupancy_estimated):
                param=dist_params[state]
                invGamma[state]=stats.invgamma.pdf(DFittParam+DBias, a=param[0],scale=param[1])
                invGammaAll=invGammaAll + Occ*invGamma[state]


            NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
            invGammaAll=invGammaAll/NormFactor

        #####################################################################
            if MovieStepSize>0: FrameIndex=0
            else: FrameIndex=-1

            movieFlag=1
            while FrameIndex < len(Vt) and movieFlag==1:
        
                # Figure construction
                fig = plt.figure(constrained_layout=False)
                fig.set_size_inches(3*5, 2*5, forward=True)
                gs = fig.add_gridspec(nrows=7, ncols=6, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.2, hspace=1.0)

                ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
                ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

                ax0=fig.add_subplot(gs[2:4, 0:2]) #MSD, CVE 
                ax4=fig.add_subplot(gs[2:4, 2])  #hist D      
                if fft==0:
                    ax1=fig.add_subplot(gs[4:6, 0:2]) #Sigma               
                    ax5=fig.add_subplot(gs[4:6, 2])  #hist sigma2     
                else:
                    ax1=fig.add_subplot(gs[4:5, 0:2]) #Sigma
                    ax5=fig.add_subplot(gs[4:5, 2])  #hist sigma2     
                    ax15=fig.add_subplot(gs[5, 0:2])  #fft     

                ax9=fig.add_subplot(gs[6, 0:2]) #Photon count 
                ax10=fig.add_subplot(gs[6, 2])  #Photon hist

                ax11=fig.add_subplot(gs[2:6, 3:6]) #Image 
                ax12=fig.add_subplot(gs[0:2, 4:5]) #Info
                
                ax13=fig.add_subplot(gs[6, 3:5]) #Pol count 
                ax14=fig.add_subplot(gs[6, 5])  #Pol hist

        
                #ax0-ax1
                ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:], axis=1),'-g',alpha=0.5, label='CVE');        #D CVE MLE data
                #ax0.plot(txyzCVE_D[:,0],HMM_prediction,'-r',alpha=0.5, label='Resedual:' + str(np.round(np.sum((np.mean(txyzCVE_D[:,1:], axis=1)-HMM_prediction)**2),decimals=2)));        #D CVE MLE data
                if MovieStepSize>0: ax0.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax0.set_ylabel("D [um2/s]"); ax0.grid(True); 
                ax0.legend()

                ax1.plot(txyzCVE_Sigma[:,0],np.mean(txyzCVE_Sigma[:,1:],axis=1),'-g',alpha=0.5, label='CVE'); #Sigma2 CVE MLE data
                ax1.plot(T_Mean,np.mean(VPTMxyz_Var,axis=1),'-b',alpha=0.5, label='Var' + str(wndw));         #Var CVE MLE data
                if MovieStepSize>0: ax1.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax1.set_ylabel("sigma2 [um2]"); ax1.grid(True);
                ax1.legend()

                #ax4 Histogram D CVE and MSD
                ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), bins=DFittParam, facecolor='g', density=True); 
                #ax4.axvline(np.mean(txyzCVE_D[:,1:]), color='g',linestyle='-',label='CVE:' + str(np.round(np.mean(txyzCVE_D[:,1:]),decimals=2)))
                #for state,Occ in zip(invGamma,Occupancy_estimated):
                #    ax4.plot(DFittParam,Occ*invGamma[state]/NormFactor,label=state)
                #ax4.plot(DFittParam,invGammaAll,'-k',label='envelope')
                ax4.set_xlabel("D [um^2/s]")
                ax4.legend()

                #ax5 Histogram Sigma Variance 
                ax5.hist(np.sqrt(np.mean(txyzCVE_Sigma[:,1:],axis=1))*1e3, 20, facecolor='g', alpha=0.5); 
                ax5.hist(np.sqrt(np.mean(VPTMxyz_Var,axis=1))*1e3, 20, facecolor='b', alpha=0.5); 
                ax5.axvline(np.sqrt(np.mean(VPTMxyz_Var))*1e3, color='b',linestyle='-',label='Var' + str(wndw) +':'+ str(np.round(np.sqrt(np.mean(VPTMxyz_Var))*1e3,decimals=2)))
                ax5.axvline(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3, color='g',linestyle='-',label='CVE:' + str(np.round(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3,decimals=2)))
                ax5.set_xlabel("Sigma [nm]")
                ax5.legend()

                #ax3
                ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[1,:],'-m',alpha=0.4,label='Piezo');  #Piezo data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[2,:],'-m',alpha=0.4);                #Piezo data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[3,:],'-m',alpha=0.4);          #Piezo data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.4,label='Mean:' + str(wndw));  #MLE data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.4);  #MLE data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.4)   #MLE data
                if MovieStepSize>0: ax3.axvline(Vt[FrameIndex], color='r',linestyle='-')

                ax3.set_ylabel("[um]"); 
                ax3.grid(True);
                ax3.legend()
        
                #ax6
                if np.sum(np.isnan(VPTMxyz))==0:
                    ax6.hist(np.sqrt(VPTMxyz_Var[:,0]), 20, facecolor='r', alpha=0.4); 
                    ax6.hist(np.sqrt(VPTMxyz_Var[:,1]), 20, facecolor='g', alpha=0.4); 
                    ax6.hist(np.sqrt(VPTMxyz_Var[:,2]), 20, facecolor='b', alpha=0.4)          
                    ax6.set_xlabel('STD [um] over: ' + str(wndw))
        
                #ax9-ax10 Photon counts
                ax9.plot(Vt,ni,'-') 
                if MovieStepSize>0: ax9.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax10.hist(ni, 20)        
                ax9.set_ylabel('Ph #')
                ax9.grid()

                ax13.plot(PolT,Pol,'-b')
                #if MovieStepSize>0: ax13.axvline(PolT[FrameIndex], color='r',linestyle='-')
                ax13.grid(); ax13.set_ylim(-1,1)
                ax14.hist(Pol, bins=np.linspace(-1,1,40))        

                #fft ax15               
                if fft!=0:
                    ax15.semilogy(fx, fft_x_MLE,c='r', alpha=0.4, label='x-MLE')
                    ax15.semilogy(fy, fft_y_MLE,c='g', alpha=0.4, label='y-MLE')
                    ax15.semilogy(fz, fft_z_MLE,c='b', alpha=0.4, label='z-MLE')
                    ax15.set_xlabel('khz')
                    ax15.set_ylim((10**-5,10**-2))
                    ax15.grid(b=True, which='major', color='b', linestyle='-')
                    ax15.grid(b=True, which='minor', color='b', linestyle='--')
                #Cell image
                    #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
                ax11.imshow(ImTot[:,:,0],cmap='gray',alpha=1.0)
                #ax[0].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)
                #ax[0].imshow(ImTot[:,:,2],cmap='Reds',alpha=0.3)
                if CellMask:
                    ax11.contour(mask,levels=0, linewidths=1, colors='blue', linestyles='solid')
                    ax11.imshow(mask,cmap='Reds',alpha=0.3)

                if MovieStepSize>0: ls=LineCollection(segments[:FrameIndex], linewidths=1,colors=colors[:FrameIndex], linestyle='solid')
                else: ls=LineCollection(segments, linewidths=1,colors=colors, linestyle='solid')

                line=ax11.add_collection(ls)
                #fig.colorbar(line, ax=ax11)
                #ax11.plot(x,y,'-', c=color[0,:])

                xmean=np.mean(x_pix)
                ymean=np.mean(y_pix)
        
                ticks=np.arange(0,ImTot[:,:,0].shape[0],4)
                xticks=np.linspace(xmean-xyImageBox,xmean+xyImageBox,10)
                yticks=np.linspace(ymean-xyImageBox,ymean+xyImageBox,10)
                xtick_labels=np.round( (xticks-(xmean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]
                ytick_labels=np.round( (yticks-(ymean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]

                ax11.set_xticks(xticks); ax11.set_yticks(yticks)
                ax11.set_xticklabels(xtick_labels); ax11.set_yticklabels(ytick_labels)
                ax11.set_xlim([xmean-xyImageBox,xmean+xyImageBox])
                ax11.set_ylim([ymean-xyImageBox,ymean+xyImageBox])
                ax11.set_xlabel('um'); ax11.set_ylabel('um')

                #ax12 Info box            
                Values=np.zeros(HMM_Model['TransMat'].shape+np.array([3,0])) 
                Values[:HMM_Model['TransMat'].shape[0],:HMM_Model['TransMat'].shape[1]]=HMM_Model['TransMat']
                Values[HMM_Model['TransMat'].shape[0],:]=HMM_Model['D']
                Values[HMM_Model['TransMat'].shape[0]+1,:]=HMM_Model['Dwelltime']
                Values[HMM_Model['TransMat'].shape[0]+2,:]=HMM_Model['Occupancy']
                Values=np.round(Values,decimals=2)

                S=()
                for state in HMM_Model['Distributions']:
                    S+=(state,)
                StatesLables=list(S)
                S+=('D:', 'dwelltime:','POcc',)
                LablesAll=list(S)

                ax12.pcolor(Values*0, cmap=plt.cm.Greys)
                #ax12.imshow(Values)
                ax12.set_xticks(np.arange(len(StatesLables))+0.5, minor=False)
                ax12.set_yticks(np.arange(len(LablesAll))+0.5, minor=False)
                ax12.set_xticklabels(StatesLables, minor=False)
                ax12.set_yticklabels(LablesAll, minor=False)
                ax12.invert_yaxis()
                #ax12.xaxis.tick_top()

                for i in range(len(LablesAll)):
                    for j in range(len(StatesLables)):
                        text = ax12.text(j+0.5, i+0.5, Values[i, j],
                                       ha="center", va="center", color="k")                
                #ax12.axis('off')
                

                #########################################################
                if do_save_plot:
                    if MovieStepSize>0:
                        file="MovieFrame_" + str(TrajToGet) + '_D_traj_' + str(FrameIndex)+ '.png'
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Movie_Traj_D_filtered_HMM_' + str(len(StatesLables)) + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    else:
                        file="T_" + str(TrajToGet) + '_D_traj' + '.png'
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_D_filtered_HMM_' + str(len(StatesLables)) + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    fig.savefig(path + file)
                    plt.close()
                    print('Done!')
                else:
                    plt.show()

                FrameIndex+=MovieStepSize
                if MovieStepSize<=0: break
                elif FrameIndex>=len(Vt):break

    def plot_Hist_Traj_MDS8(ScanProject,TrajToGet,TrajDictionary,HMM_prediction,HMM_Model,MovieStepSize=0,do_save_plot=0,FolderEnd='',fft=0):
            """
            This function is to plot 
                trajectores 
                Cell image
                HMM predicted traj
                zCam z data
            """
            xyImageBox=35 #pix
            CellMask=1
        
            if CellMask:
                mask=ScanProject.CamCellMask[:,:,0]>0
            CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]
            try:
                ImTot=ScanProject.ImgOverlay
            except:
                print('No cell image found, using cell mask')
                ImTot=np.repeat(np.expand_dims(mask*1.0,axis=2),3,axis=2)

            wndw=10
            txyznM=TrajDictionary[TrajToGet]*1.0

                #CVE D and sigma evaluation
            txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
            txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

            txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu]
            txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]

            ni=txyznM[4,:]

                #time
            Vt=txyznM[0,:]              #[s]

                #Piezo stage position
            TStartEnd=[Vt[0],Vt[-1]] 
            StartEndPos=np.where(np.abs(np.diff((ScanProject.ScanMatrix[0,:]>TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<TStartEnd[1]))))[0]
            StartEndPos=np.expand_dims(np.append(np.array([0]),StartEndPos),axis=1)
            
            T_Raw,XP_Raw,YP_Raw,XE_Raw,YE_Raw,XO_Raw,YO_Raw,ZI_Raw,APD1_Raw,APD2_Raw,J_Raw,MemIndex_Raw,ZI_um_Raw,ZI_OffsetIndex_Raw, zCam_Raw =  _MS.get_traj_raw_data4(StartEndPos, 0, ScanProject.ScanMatrix, ScanProject)
            txyz_Traj_Piezo=np.vstack((T_Raw,XP_Raw,YP_Raw,ZI_um_Raw))
            txyz_Traj_Piezo=_MS.get_Scan_toImagCoordinates(ScanProject, txyz_Traj_Piezo)
            
            #ZeroPoint=np.array([XP_Raw[0],YP_Raw[0],ZI_um_Raw[0]])
                #MLE
            VPTMxyz_O=np.transpose(txyznM[1:4,:])
            
                #Start at zero time and at zero piezo pos
            VPTMxyz=VPTMxyz_O-txyz_Traj_Piezo[1:,0]#VPTMxyz_O[0,:]
            Vt=(Vt-Vt[0])*1e3 #[ms]

            txyz_Traj_Piezo=(txyz_Traj_Piezo.T-txyz_Traj_Piezo[:,0]).T
            txyz_Traj_Piezo[0,:]=txyz_Traj_Piezo[0,:]*1e3

                # Fourier spectra of the MLE signals
            if fft!=0:
                if 0:
                    wndw_fft=20
                    wndwS_fft=wndw_fft #if 1 it is a moving average, if equal to wndw it is a fix window average
                    T_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,Vt)
                    T_fft=T_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]

                    x_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,0])
                    x_fft=x_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    y_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,1])
                    y_fft=y_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    z_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,2])
                    z_fft=z_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    
                    fft_x_MLE, fx = _MS.fft_processing(x_fft,T_fft)
                    fft_y_MLE, fy = _MS.fft_processing(y_fft,T_fft)
                    fft_z_MLE, fz = _MS.fft_processing(z_fft,T_fft)
                else:
                    fft_x_MLE, fx = _MS.fft_processing(VPTMxyz[:,0],Vt)
                    fft_y_MLE, fy = _MS.fft_processing(VPTMxyz[:,1],Vt)
                    fft_z_MLE, fz = _MS.fft_processing(VPTMxyz[:,2],Vt)

                #Running mean on the trajectory
            wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            T_Mean= np.convolve(np.ones(wndw)/wndw,Vt)
            T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
            VPTMxyz_Mean=np.zeros((len(T_Mean),3))
            VPTMxyz_Var=np.zeros((len(T_Mean),3))
            for i in range(3):
                VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,VPTMxyz[:,i])
                VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

                VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(VPTMxyz[:,i])**2) #Var(X)=E[X^2]-E[X]^2
                VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2

            #np.convolve(np.ones(wndw)/wndw,np.mean(txyzCVE_D[:,1:], axis=1))[wndw-1:-wndw+1:wndwS]

                #Create a D value colored traj plot
            x_pix=(VPTMxyz_O[:,0])/CamRes
            y_pix=(VPTMxyz_O[:,1])/CamRes
            points=np.array([x_pix, y_pix]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
        
            D_norm=np.log(np.mean(txyzCVE_D[:,1:], axis=1)) #(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))/ np.max(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))
            D_norm[np.where(np.isnan(D_norm))]=0
            D_norm=1/(1+np.exp(-(D_norm)))
            t_D=np.concatenate((txyzCVE_D[:,0],[Vt[-1]]))
            D_traj=np.zeros(Vt.shape)
            j=0
            for i in range(1,len(t_D)):
                while Vt[j]<=t_D[i]:
                    D_traj[j]=D_norm[i-1]
                    j+=1
                    if j>=len(Vt):
                        break

            colors = plt.cm.cool(D_traj)

                #Polarization 
            wndwN=int(ScanProject.TrajectorySelectionRules['wndw'])
            OfInterest=np.where((ScanProject.ScanMatrix[0,:]>=TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<=TStartEnd[1]))[0]
            OfInterest=np.append(OfInterest,OfInterest[-1]+np.arange(wndwN)+1)
            N12=ScanProject.ScanMatrix[3:5,OfInterest]
            PolT=ScanProject.ScanMatrix[0,OfInterest]

            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1=np.convolve(np.ones(wndwN)/wndwN,N12[0,:])
            N1=N1[wndwN-1:-wndwN+1:wndwS]
            N2=np.convolve(np.ones(wndwN)/wndwN,N12[1,:])
            N2=N2[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]

            
            wndwN=5
            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1_mean=np.convolve(np.ones(wndwN)/wndwN,N1)
            N1_mean=N1_mean[wndwN-1:-wndwN+1:wndwS]
            N2_mean=np.convolve(np.ones(wndwN)/wndwN,N2)
            N2_mean=N2_mean[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]
            PolT=(PolT-PolT[0])*1e3 #ms
            Pol=(N1_mean-N2_mean)/(N1_mean+N2_mean)

        ####################################
            # Generate the inverse-Gamma distributions to plot
            dist_params=HMM_Model['Distributions']
            Occupancy_estimated=HMM_Model['Occupancy']
            DBias=HMM_Model['DBias']

            DFittParam=np.linspace(-DBias,10,num=200)
            invGamma={}
            invGammaAll=np.zeros(DFittParam.shape)
            for state,Occ in zip(dist_params,Occupancy_estimated):
                param=dist_params[state]
                invGamma[state]=stats.invgamma.pdf(DFittParam+DBias, a=param[0],scale=param[1])
                invGammaAll=invGammaAll + Occ*invGamma[state]


            NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
            invGammaAll=invGammaAll/NormFactor

        #####################################################################
            if MovieStepSize>0: FrameIndex=0
            else: FrameIndex=-1

            movieFlag=1
            while FrameIndex < len(Vt) and movieFlag==1:
        
                # Figure construction
                fig = plt.figure(constrained_layout=False)
                fig.set_size_inches(3*5, 2*5, forward=True)
                gs = fig.add_gridspec(nrows=7, ncols=6, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.2, hspace=1.0)

                ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
                ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

                ax0=fig.add_subplot(gs[2:4, 0:2]) #MSD, CVE 
                ax4=fig.add_subplot(gs[2:4, 2])  #hist D      
                if fft==0:
                    ax1=fig.add_subplot(gs[4:6, 0:2]) #Sigma               
                    ax5=fig.add_subplot(gs[4:6, 2])  #hist sigma2     
                else:
                    ax1=fig.add_subplot(gs[4:5, 0:2]) #Sigma
                    ax5=fig.add_subplot(gs[4:5, 2])  #hist sigma2     
                    ax15=fig.add_subplot(gs[5, 0:2])  #fft     

                ax9=fig.add_subplot(gs[6, 0:2]) #Photon count 
                ax10=fig.add_subplot(gs[6, 2])  #Photon hist

                ax11=fig.add_subplot(gs[2:6, 3:6]) #Image 
                ax12=fig.add_subplot(gs[0:2, 4:5]) #Info
                
                ax13=fig.add_subplot(gs[6, 3:5]) #Pol count 
                ax14=fig.add_subplot(gs[6, 5])  #Pol hist

        
                #ax0-ax1
                ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:], axis=1),'-g',alpha=0.5, label='CVE');        #D CVE MLE data
                #ax0.plot(txyzCVE_D[:,0],HMM_prediction,'-r',alpha=0.5, label='Resedual:' + str(np.round(np.sum((np.mean(txyzCVE_D[:,1:], axis=1)-HMM_prediction)**2),decimals=2)));        #D CVE MLE data
                if MovieStepSize>0: ax0.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax0.set_ylabel("D [um2/s]"); ax0.grid(True); 
                ax0.legend()

                ax1.plot(txyzCVE_Sigma[:,0],np.mean(txyzCVE_Sigma[:,1:],axis=1),'-g',alpha=0.5, label='CVE'); #Sigma2 CVE MLE data
                ax1.plot(T_Mean,np.mean(VPTMxyz_Var,axis=1),'-b',alpha=0.5, label='Var' + str(wndw));         #Var CVE MLE data
                if MovieStepSize>0: ax1.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax1.set_ylabel("sigma2 [um2]"); ax1.grid(True);
                ax1.legend()

                #ax4 Histogram D CVE and MSD
                ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), bins=DFittParam, facecolor='g', density=True); 
                #ax4.axvline(np.mean(txyzCVE_D[:,1:]), color='g',linestyle='-',label='CVE:' + str(np.round(np.mean(txyzCVE_D[:,1:]),decimals=2)))
                #for state,Occ in zip(invGamma,Occupancy_estimated):
                #    ax4.plot(DFittParam,Occ*invGamma[state]/NormFactor,label=state)
                #ax4.plot(DFittParam,invGammaAll,'-k',label='envelope')
                ax4.set_xlabel("D [um^2/s]")
                ax4.legend()

                #ax5 Histogram Sigma Variance 
                ax5.hist(np.sqrt(np.mean(txyzCVE_Sigma[:,1:],axis=1))*1e3, 20, facecolor='g', alpha=0.5); 
                ax5.hist(np.sqrt(np.mean(VPTMxyz_Var,axis=1))*1e3, 20, facecolor='b', alpha=0.5); 
                ax5.axvline(np.sqrt(np.mean(VPTMxyz_Var))*1e3, color='b',linestyle='-',label='Var' + str(wndw) +':'+ str(np.round(np.sqrt(np.mean(VPTMxyz_Var))*1e3,decimals=2)))
                ax5.axvline(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3, color='g',linestyle='-',label='CVE:' + str(np.round(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3,decimals=2)))
                ax5.set_xlabel("Sigma [nm]")
                ax5.legend()

                #ax3
                ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[1,:],'-m',alpha=0.4,label='Piezo');  #Piezo data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[2,:],'-m',alpha=0.4);                #Piezo data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[3,:],'-m',alpha=0.4);          #Piezo data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.4,label='Mean:' + str(wndw));  #MLE data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.4);  #MLE data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.4)   #MLE data
                ax3.plot((T_Raw-T_Raw[0])*1e3,zCam_Raw-zCam_Raw[0],'-b')
                if MovieStepSize>0: ax3.axvline(Vt[FrameIndex], color='r',linestyle='-')

                ax3.set_ylabel("[um]"); 
                ax3.grid(True);
                ax3.legend()
        
                #ax6
                if np.sum(np.isnan(VPTMxyz))==0:
                    ax6.hist(np.sqrt(VPTMxyz_Var[:,0]), 20, facecolor='r', alpha=0.4); 
                    ax6.hist(np.sqrt(VPTMxyz_Var[:,1]), 20, facecolor='g', alpha=0.4); 
                    ax6.hist(np.sqrt(VPTMxyz_Var[:,2]), 20, facecolor='b', alpha=0.4)          
                    ax6.set_xlabel('STD [um] over: ' + str(wndw))
        
                #ax9-ax10 Photon counts
                ax9.plot(Vt,ni,'-') 
                if MovieStepSize>0: ax9.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax10.hist(ni, 20)        
                ax9.set_ylabel('Ph #')
                ax9.grid()

                ax13.plot(PolT,Pol,'-b')
                #if MovieStepSize>0: ax13.axvline(PolT[FrameIndex], color='r',linestyle='-')
                ax13.grid(); ax13.set_ylim(-1,1)
                ax14.hist(Pol, bins=np.linspace(-1,1,40))        

                #fft ax15               
                if fft!=0:
                    ax15.semilogy(fx, fft_x_MLE,c='r', alpha=0.4, label='x-MLE')
                    ax15.semilogy(fy, fft_y_MLE,c='g', alpha=0.4, label='y-MLE')
                    ax15.semilogy(fz, fft_z_MLE,c='b', alpha=0.4, label='z-MLE')
                    ax15.set_xlabel('khz')
                    ax15.set_ylim((10**-5,10**-2))
                    ax15.grid(b=True, which='major', color='b', linestyle='-')
                    ax15.grid(b=True, which='minor', color='b', linestyle='--')
                #Cell image
                    #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
                ax11.imshow(ImTot[:,:,0],cmap='gray',alpha=1.0)
                #ax[0].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)
                #ax[0].imshow(ImTot[:,:,2],cmap='Reds',alpha=0.3)
                if CellMask:
                    ax11.contour(mask,levels=0, linewidths=1, colors='blue', linestyles='solid')
                    ax11.imshow(mask,cmap='Reds',alpha=0.3)

                if MovieStepSize>0: ls=LineCollection(segments[:FrameIndex], linewidths=1,colors=colors[:FrameIndex], linestyle='solid')
                else: ls=LineCollection(segments, linewidths=1,colors=colors, linestyle='solid')

                line=ax11.add_collection(ls)
                #fig.colorbar(line, ax=ax11)
                #ax11.plot(x,y,'-', c=color[0,:])

                xmean=np.mean(x_pix)
                ymean=np.mean(y_pix)
        
                ticks=np.arange(0,ImTot[:,:,0].shape[0],4)
                xticks=np.linspace(xmean-xyImageBox,xmean+xyImageBox,10)
                yticks=np.linspace(ymean-xyImageBox,ymean+xyImageBox,10)
                xtick_labels=np.round( (xticks-(xmean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]
                ytick_labels=np.round( (yticks-(ymean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]

                ax11.set_xticks(xticks); ax11.set_yticks(yticks)
                ax11.set_xticklabels(xtick_labels); ax11.set_yticklabels(ytick_labels)
                ax11.set_xlim([xmean-xyImageBox,xmean+xyImageBox])
                ax11.set_ylim([ymean-xyImageBox,ymean+xyImageBox])
                ax11.set_xlabel('um'); ax11.set_ylabel('um')

                #ax12 Info box            
                Values=np.zeros(HMM_Model['TransMat'].shape+np.array([3,0])) 
                Values[:HMM_Model['TransMat'].shape[0],:HMM_Model['TransMat'].shape[1]]=HMM_Model['TransMat']
                Values[HMM_Model['TransMat'].shape[0],:]=HMM_Model['D']
                Values[HMM_Model['TransMat'].shape[0]+1,:]=HMM_Model['Dwelltime']
                Values[HMM_Model['TransMat'].shape[0]+2,:]=HMM_Model['Occupancy']
                Values=np.round(Values,decimals=2)

                S=()
                for state in HMM_Model['Distributions']:
                    S+=(state,)
                StatesLables=list(S)
                S+=('D:', 'dwelltime:','POcc',)
                LablesAll=list(S)

                ax12.pcolor(Values*0, cmap=plt.cm.Greys)
                #ax12.imshow(Values)
                ax12.set_xticks(np.arange(len(StatesLables))+0.5, minor=False)
                ax12.set_yticks(np.arange(len(LablesAll))+0.5, minor=False)
                ax12.set_xticklabels(StatesLables, minor=False)
                ax12.set_yticklabels(LablesAll, minor=False)
                ax12.invert_yaxis()
                #ax12.xaxis.tick_top()

                for i in range(len(LablesAll)):
                    for j in range(len(StatesLables)):
                        text = ax12.text(j+0.5, i+0.5, Values[i, j],
                                        ha="center", va="center", color="k")                
                #ax12.axis('off')
                

                #########################################################
                if do_save_plot:
                    if MovieStepSize>0:
                        file="MovieFrame_" + str(TrajToGet) + '_D_traj_' + str(FrameIndex)+ '.png'
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Movie_Traj_D_filtered_HMM_' + str(len(StatesLables)) + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    else:
                        file="T_" + str(TrajToGet) + '_D_traj' + '.png'
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_D_filtered_HMM_' + str(len(StatesLables)) + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    fig.savefig(path + file)
                    plt.close()
                    print('Done!')
                else:
                    plt.show()

                FrameIndex+=MovieStepSize
                if MovieStepSize<=0: break
                elif FrameIndex>=len(Vt):break

    def plot_Hist_Traj_MDS9(ScanProject,TrajToGet,TrajDictionary,HMM_prediction,HMM_Model,MovieStepSize=0,do_save_plot=0,FolderEnd='',fft=0,fileType='.png'):
            """
            This function is to plot 
                trajectores 
                Cell image
                HMM predicted traj
                zCam z data
            """
            xyImageBox=35 #pix
            CellMask=1
        
            if CellMask:
                mask=ScanProject.CamCellMask[:,:,0]>0
            CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]
            try:
                ImTot=ScanProject.ImgOverlay
            except:
                print('No cell image found, using cell mask')
                ImTot=np.repeat(np.expand_dims(mask*1.0,axis=2),3,axis=2)

            wndw=10
            txyznM=TrajDictionary[TrajToGet]*1.0

                #CVE D and sigma evaluation
            txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
            txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

            txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu]
            txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]


                #time
            Vt=txyznM[0,:]              #[s]

                #Photon Counts
            ni=txyznM[4,:] #Raw
            DT=np.mean(np.diff(Vt))
            MLE_tail=ScanProject.TrajectorySelectionRules['MLETailFactor']
            window_size=20
            Ni=np.convolve((1-MLE_tail)**np.arange(window_size),ni)[0:-window_size+1]


                #Piezo stage position
            TStartEnd=[Vt[0],Vt[-1]] 
            StartEndPos=np.where(np.abs(np.diff((ScanProject.ScanMatrix[0,:]>TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<TStartEnd[1]))))[0]
            StartEndPos=np.expand_dims(np.append(np.array([0]),StartEndPos),axis=1)
            
            T_Raw,XP_Raw,YP_Raw,XE_Raw,YE_Raw,XO_Raw,YO_Raw,ZI_Raw,APD1_Raw,APD2_Raw,J_Raw,MemIndex_Raw,ZI_um_Raw,ZI_OffsetIndex_Raw, zCam_Raw =  _MS.get_traj_raw_data4(StartEndPos, 0, ScanProject.ScanMatrix, ScanProject)
            txyz_Traj_Piezo=np.vstack((T_Raw,XP_Raw,YP_Raw,ZI_um_Raw))
            txyz_Traj_Piezo=_MS.get_Scan_toImagCoordinates(ScanProject, txyz_Traj_Piezo)
            
            #ZeroPoint=np.array([XP_Raw[0],YP_Raw[0],ZI_um_Raw[0]])
                #MLE
            VPTMxyz_O=np.transpose(txyznM[1:4,:])
            
                #Start at zero time and at zero piezo pos
            VPTMxyz=VPTMxyz_O-txyz_Traj_Piezo[1:,0]#VPTMxyz_O[0,:]
            Vt=(Vt-Vt[0])*1e3 #[ms]

            txyz_Traj_Piezo=(txyz_Traj_Piezo.T-txyz_Traj_Piezo[:,0]).T
            txyz_Traj_Piezo[0,:]=txyz_Traj_Piezo[0,:]*1e3

                # Fourier spectra of the MLE signals
            if fft!=0:
                if 0:
                    wndw_fft=20
                    wndwS_fft=wndw_fft #if 1 it is a moving average, if equal to wndw it is a fix window average
                    T_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,Vt)
                    T_fft=T_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]

                    x_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,0])
                    x_fft=x_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    y_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,1])
                    y_fft=y_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    z_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,2])
                    z_fft=z_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    
                    fft_x_MLE, fx = _MS.fft_processing(x_fft,T_fft)
                    fft_y_MLE, fy = _MS.fft_processing(y_fft,T_fft)
                    fft_z_MLE, fz = _MS.fft_processing(z_fft,T_fft)
                else:
                    fft_x_MLE, fx = _MS.fft_processing(VPTMxyz[:,0],Vt)
                    fft_y_MLE, fy = _MS.fft_processing(VPTMxyz[:,1],Vt)
                    fft_z_MLE, fz = _MS.fft_processing(VPTMxyz[:,2],Vt)

                #Running mean on the trajectory
            wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            T_Mean= np.convolve(np.ones(wndw)/wndw,Vt)
            T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
            VPTMxyz_Mean=np.zeros((len(T_Mean),3))
            VPTMxyz_Var=np.zeros((len(T_Mean),3))
            for i in range(3):
                VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,VPTMxyz[:,i])
                VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

                VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(VPTMxyz[:,i])**2) #Var(X)=E[X^2]-E[X]^2
                VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2

            #np.convolve(np.ones(wndw)/wndw,np.mean(txyzCVE_D[:,1:], axis=1))[wndw-1:-wndw+1:wndwS]

                #Create a D value colored traj plot
            x_pix=(VPTMxyz_O[:,0])/CamRes
            y_pix=(VPTMxyz_O[:,1])/CamRes
            z_pix=(VPTMxyz_O[:,1])/CamRes
            #points=np.array([x_pix, y_pix, z_pix]).T.reshape(-1, 1, 3) 
            points=np.array([x_pix, y_pix]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
        
            D_norm=np.log(np.mean(txyzCVE_D[:,1:], axis=1)) #(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))/ np.max(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))
            D_norm[np.where(np.isnan(D_norm))]=0
            D_norm=1/(1+np.exp(-(D_norm)))
            t_D=np.concatenate((txyzCVE_D[:,0],[Vt[-1]]))
            D_traj=np.zeros(Vt.shape)
            j=0
            for i in range(1,len(t_D)):
                while Vt[j]<=t_D[i]:
                    D_traj[j]=D_norm[i-1]
                    j+=1
                    if j>=len(Vt):
                        break

            colors = plt.cm.cool(D_traj)

                #Polarization 
            wndwN=int(ScanProject.TrajectorySelectionRules['wndw'])
            OfInterest=np.where((ScanProject.ScanMatrix[0,:]>=TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<=TStartEnd[1]))[0]
            OfInterest=np.append(OfInterest,OfInterest[-1]+np.arange(wndwN)+1)
            N12=ScanProject.ScanMatrix[3:5,OfInterest]
            PolT=ScanProject.ScanMatrix[0,OfInterest]

            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1=np.convolve(np.ones(wndwN)/wndwN,N12[0,:])
            N1=N1[wndwN-1:-wndwN+1:wndwS]
            N2=np.convolve(np.ones(wndwN)/wndwN,N12[1,:])
            N2=N2[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]

            
            wndwN=5
            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1_mean=np.convolve(np.ones(wndwN)/wndwN,N1)
            N1_mean=N1_mean[wndwN-1:-wndwN+1:wndwS]
            N2_mean=np.convolve(np.ones(wndwN)/wndwN,N2)
            N2_mean=N2_mean[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]
            PolT=(PolT-PolT[0])*1e3 #ms
            Pol=(N1_mean-N2_mean)/(N1_mean+N2_mean)

        ####################################
            # Generate the inverse-Gamma distributions to plot
            dist_params=HMM_Model['Distributions']
            Occupancy_estimated=HMM_Model['Occupancy']
            DBias=HMM_Model['DBias']

            DFittParam=np.linspace(-DBias,10,num=200)
            invGamma={}
            invGammaAll=np.zeros(DFittParam.shape)
            for state,Occ in zip(dist_params,Occupancy_estimated):
                param=dist_params[state]
                invGamma[state]=stats.invgamma.pdf(DFittParam+DBias, a=param[0],scale=param[1])
                invGammaAll=invGammaAll + Occ*invGamma[state]


            NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
            invGammaAll=invGammaAll/NormFactor

        #####################################################################
            try:
                if MovieStepSize>0: FrameIndex=0
                else: FrameIndex=-1

                movieFlag=1
                while FrameIndex < len(Vt) and movieFlag==1:
        
                    # Figure construction
                    fig = plt.figure(constrained_layout=False)
                    fig.set_size_inches(3*5, 2*5, forward=True)
                    gs = fig.add_gridspec(nrows=7, ncols=6, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.2, hspace=1.0)

                    ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
                    ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

                    ax0=fig.add_subplot(gs[2:4, 0:2]) #MSD, CVE 
                    ax4=fig.add_subplot(gs[2:4, 2])  #hist D      
                    if fft==0:
                        ax1=fig.add_subplot(gs[4:6, 0:2]) #Sigma               
                        ax5=fig.add_subplot(gs[4:6, 2])  #hist sigma2     
                    else:
                        ax1=fig.add_subplot(gs[4:5, 0:2]) #Sigma
                        ax5=fig.add_subplot(gs[4:5, 2])  #hist sigma2     
                        ax15=fig.add_subplot(gs[5, 0:2])  #fft     

                    ax9=fig.add_subplot(gs[6, 0:2]) #Photon count 
                    ax10=fig.add_subplot(gs[6, 2])  #Photon hist

                    ax11=fig.add_subplot(gs[2:6, 3:6]) #Image 
                    ax12=fig.add_subplot(gs[0:2, 4:5]) #Info
                
                    ax13=fig.add_subplot(gs[6, 3:5]) #Pol count 
                    ax14=fig.add_subplot(gs[6, 5])  #Pol hist

        
                    #ax0-ax1
                    ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:4], axis=1),'-k',alpha=0.5, label='CVE xyz');        #D CVE MLE data
                    #ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:3], axis=1),'-k',alpha=0.5, label='CVE xy');        #D CVE MLE data
                    #ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,1],'-r',alpha=0.5, label='CVE x');        #D CVE MLE data
                    #ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,2],'-g',alpha=0.5, label='CVE y');        #D CVE MLE data
                    #ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,3],'-b',alpha=0.5, label='CVE z');        #D CVE MLE data
                    ax0.plot(txyzCVE_D[:,0],HMM_prediction,'-r',alpha=0.5, label='Resedual:' + str(np.round(np.sum((np.mean(txyzCVE_D[:,1:], axis=1)-HMM_prediction)**2),decimals=2)));        #D CVE MLE data
                    if MovieStepSize>0: ax0.axvline(Vt[FrameIndex], color='r',linestyle='-')
                    ax0.set_ylabel("D [um2/s]"); ax0.grid(True); 
                    ax0.legend()

                    ax1.plot(txyzCVE_Sigma[:,0],np.mean(txyzCVE_Sigma[:,1:],axis=1),'-g',alpha=0.5, label='CVE'); #Sigma2 CVE MLE data
                    ax1.plot(T_Mean,np.mean(VPTMxyz_Var,axis=1),'-b',alpha=0.5, label='Var' + str(wndw));         #Var CVE MLE data
                    if MovieStepSize>0: ax1.axvline(Vt[FrameIndex], color='r',linestyle='-')
                    ax1.set_ylabel("sigma2 [um2]"); ax1.grid(True);
                    ax1.legend()

                    #ax4 Histogram D CVE and MSD
                    ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), bins=DFittParam, facecolor='g', density=True); 
                    #ax4.axvline(np.mean(txyzCVE_D[:,1:]), color='g',linestyle='-',label='CVE:' + str(np.round(np.mean(txyzCVE_D[:,1:]),decimals=2)))
                    #for state,Occ in zip(invGamma,Occupancy_estimated):
                    #    ax4.plot(DFittParam,Occ*invGamma[state]/NormFactor,label=state)
                    #ax4.plot(DFittParam,invGammaAll,'-k',label='envelope')
                    ax4.set_xlabel("D [um^2/s]")
                    ax4.legend()

                    #ax5 Histogram Sigma Variance 
                    ax5.hist(np.sqrt(np.mean(txyzCVE_Sigma[:,1:],axis=1))*1e3, 20, facecolor='g', alpha=0.5); 
                    ax5.hist(np.sqrt(np.mean(VPTMxyz_Var,axis=1))*1e3, 20, facecolor='b', alpha=0.5); 
                    ax5.axvline(np.sqrt(np.mean(VPTMxyz_Var))*1e3, color='b',linestyle='-',label='Var' + str(wndw) +':'+ str(np.round(np.sqrt(np.mean(VPTMxyz_Var))*1e3,decimals=2)))
                    ax5.axvline(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3, color='g',linestyle='-',label='CVE:' + str(np.round(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3,decimals=2)))
                    ax5.set_xlabel("Sigma [nm]")
                    ax5.legend()

                    #ax3
                    ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
                    ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
                    ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
                    ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[1,:],'-m',alpha=0.4,label='Piezo');  #Piezo data
                    ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[2,:],'-m',alpha=0.4);                #Piezo data
                    ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[3,:],'-m',alpha=0.4);          #Piezo data
                    ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.4,label='Mean:' + str(wndw));  #MLE data
                    ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.4);  #MLE data
                    ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.4)   #MLE data
                    ax3.plot((T_Raw-T_Raw[0])*1e3,zCam_Raw-zCam_Raw[0],'-b')
                    if MovieStepSize>0: ax3.axvline(Vt[FrameIndex], color='r',linestyle='-')

                    ax3.set_ylabel("[um]"); 
                    ax3.grid(True);
                    ax3.legend()
        
                    #ax6
                    if np.sum(np.isnan(VPTMxyz))==0:
                        ax6.hist(np.sqrt(VPTMxyz_Var[:,0]), 20, facecolor='r', alpha=0.4); 
                        ax6.hist(np.sqrt(VPTMxyz_Var[:,1]), 20, facecolor='g', alpha=0.4); 
                        ax6.hist(np.sqrt(VPTMxyz_Var[:,2]), 20, facecolor='b', alpha=0.4)          
                        ax6.set_xlabel('STD [um] over: ' + str(wndw))
        
                    #ax9-ax10 Photon counts
                    ax9.plot(Vt,Ni,'-') 
                    if MovieStepSize>0: ax9.axvline(Vt[FrameIndex], color='r',linestyle='-')
                    ax10.hist(Ni, 20)        
                    ax9.set_ylabel('Ph #')
                    ax9.grid()

                    ax13.plot(PolT,Pol,'-b')
                    #if MovieStepSize>0: ax13.axvline(PolT[FrameIndex], color='r',linestyle='-')
                    ax13.grid(); ax13.set_ylim(-1,1)
                    ax14.hist(Pol, bins=np.linspace(-1,1,40))        

                    #fft ax15               
                    if fft!=0:
                        ax15.semilogy(fx, fft_x_MLE,c='r', alpha=0.4, label='x-MLE')
                        ax15.semilogy(fy, fft_y_MLE,c='g', alpha=0.4, label='y-MLE')
                        ax15.semilogy(fz, fft_z_MLE,c='b', alpha=0.4, label='z-MLE')
                        ax15.set_xlabel('khz')
                        ax15.set_ylim((10**-5,10**-2))
                        ax15.grid(b=True, which='major', color='b', linestyle='-')
                        ax15.grid(b=True, which='minor', color='b', linestyle='--')
                    #Cell image
                        #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
                    ax11.imshow(ImTot[:,:,0],cmap='gray',alpha=1.0)
                    #ax[0].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)
                    #ax[0].imshow(ImTot[:,:,2],cmap='Reds',alpha=0.3)
                    if CellMask:
                        ax11.contour(mask,levels=0, linewidths=1, colors='blue', linestyles='solid')
                        ax11.imshow(mask,cmap='Reds',alpha=0.3)

                    if MovieStepSize>0: ls=LineCollection(segments[:FrameIndex], linewidths=1,colors=colors[:FrameIndex], linestyle='solid')
                    else: ls=LineCollection(segments, linewidths=1,colors=colors, linestyle='solid')

                    line=ax11.add_collection(ls)
                    #fig.colorbar(line, ax=ax11)
                    #ax11.plot(x,y,'-', c=color[0,:])

                    xmean=np.mean(x_pix)
                    ymean=np.mean(y_pix)
        
                    ticks=np.arange(0,ImTot[:,:,0].shape[0],4)
                    xticks=np.linspace(xmean-xyImageBox,xmean+xyImageBox,10)
                    yticks=np.linspace(ymean-xyImageBox,ymean+xyImageBox,10)
                    xtick_labels=np.round( (xticks-(xmean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]
                    ytick_labels=np.round( (yticks-(ymean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]

                    ax11.set_xticks(xticks); ax11.set_yticks(yticks)
                    ax11.set_xticklabels(xtick_labels); ax11.set_yticklabels(ytick_labels)
                    ax11.set_xlim([xmean-xyImageBox,xmean+xyImageBox])
                    ax11.set_ylim([ymean-xyImageBox,ymean+xyImageBox])
                    ax11.set_xlabel('um'); ax11.set_ylabel('um')

                    #ax12 Info box            
                    Values=np.zeros(HMM_Model['TransMat'].shape+np.array([3,0])) 
                    Values[:HMM_Model['TransMat'].shape[0],:HMM_Model['TransMat'].shape[1]]=HMM_Model['TransMat']
                    Values[HMM_Model['TransMat'].shape[0],:]=HMM_Model['D']
                    Values[HMM_Model['TransMat'].shape[0]+1,:]=HMM_Model['Dwelltime']
                    Values[HMM_Model['TransMat'].shape[0]+2,:]=HMM_Model['Occupancy']
                    Values=np.round(Values,decimals=2)

                    S=()
                    for state in HMM_Model['Distributions']:
                        S+=(state,)
                    StatesLables=list(S)
                    S+=('D:', 'dwelltime:','POcc',)
                    LablesAll=list(S)

                    ax12.pcolor(Values*0, cmap=plt.cm.Greys)
                    #ax12.imshow(Values)
                    ax12.set_xticks(np.arange(len(StatesLables))+0.5, minor=False)
                    ax12.set_yticks(np.arange(len(LablesAll))+0.5, minor=False)
                    ax12.set_xticklabels(StatesLables, minor=False)
                    ax12.set_yticklabels(LablesAll, minor=False)
                    ax12.invert_yaxis()
                    #ax12.xaxis.tick_top()

                    for i in range(len(LablesAll)):
                        for j in range(len(StatesLables)):
                            text = ax12.text(j+0.5, i+0.5, Values[i, j],
                                            ha="center", va="center", color="k")                
                    #ax12.axis('off')
                

                    #########################################################
                    #fileType='.svg' #'.png' '.pdf'
                    if do_save_plot:
                        if MovieStepSize>0:
                            file="MovieFrame_" + str(TrajToGet) + '_D_traj_' + str(FrameIndex)+ fileType
                            path=ScanProject.Files['AnalysisConfig'][:-18] +'Movie_Traj_D_HMM_' + str(len(StatesLables)) + FolderEnd +'/'
                            Path(path).mkdir(parents=True, exist_ok=True)
                        else:
                            file="T_" + str(TrajToGet) + '_D_traj' + fileType
                            path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_D_HMM_' + str(len(StatesLables)) + FolderEnd +'/'
                            Path(path).mkdir(parents=True, exist_ok=True)
                        #fig.savefig(path + file)
                        fig.savefig(path + file, bbox_inches='tight')
                        plt.close()
                        print('Done!')
                    else:
                        plt.show()

                    FrameIndex+=MovieStepSize
                    if MovieStepSize<=0: break
                    elif FrameIndex>=len(Vt):break
            except:
                print('Something wrong in the plotting, skipping this one!')

    def plot_Hist_Traj_MDS9b(ScanProject,TrajToGet,TrajDictionary,HMM_prediction,HMM_Model,MovieStepSize=0,groundtruth=None,do_save_plot=0,FolderEnd='',fft=0,FileType='.png'):
            """
            This function is to plot 
                trajectores 
                Cell image
                HMM predicted traj
                HMM groundtruth traj
                zCam z data
            """
            
            xyImageBox=35 #pix
            CellMask=1
        
            if CellMask:
                mask=ScanProject.CamCellMask[:,:,0]>0
            CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]
            try:
                ImTot=ScanProject.ImgOverlay
            except:
                print('No cell image found, using cell mask')
                ImTot=np.repeat(np.expand_dims(mask*1.0,axis=2),3,axis=2)

            wndw=10
            txyznM=TrajDictionary[TrajToGet]*1.0

                #Groundtruth data 
            if np.sum(groundtruth==None):
                xyzgroundtruth=np.zeros((3,txyznM.shape[1]))
            else:
                if len(groundtruth.shape)==3:
                    xyzgroundtruthAll=groundtruth[:-1,:,:3]               
                    xyzgroundtruth=np.mean(xyzgroundtruthAll,axis=1)#Mean over blured out pixle
                    try:
                        DgroundtruthAll=ScanProject.GroundTruthParam['D_true'][(groundtruth[:,:,3].flatten()).astype('int')]
                        TgroundtruthAll=np.arange(DgroundtruthAll.shape[0])* ScanProject.GroundTruthParam['dt']*(1e3) #[ms]
                    except:
                        DgroundtruthAll=groundtruth[:,:,3].flatten()
                        TgroundtruthAll=np.arange(DgroundtruthAll.shape[0])* ScanProject.GroundTruthParam['dt']*(1e3) #[ms]
                else:
                    xyzgroundtruthAll=groundtruth[:,:-1].T               
                    xyzgroundtruth=xyzgroundtruthAll

                #CVE D and sigma evaluation
            txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
            txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu]


                #time
            Vt=txyznM[0,:]              #[s]

                #Photon Counts
            ni=txyznM[4,:] #Raw
            DT=np.mean(np.diff(Vt))
            MLE_tail=ScanProject.TrajectorySelectionRules['MLETailFactor']
            window_size=20
            Ni=np.convolve((1-MLE_tail)**np.arange(window_size),ni)[0:-window_size+1]


                #Piezo stage position
            TStartEnd=[Vt[0],Vt[-1]] 
            StartEndPos=np.where(np.abs(np.diff((ScanProject.ScanMatrix[0,:]>TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<TStartEnd[1]))))[0]
            StartEndPos=np.expand_dims(np.append(np.array([0]),StartEndPos),axis=1)
            
            T_Raw,XP_Raw,YP_Raw,XE_Raw,YE_Raw,XO_Raw,YO_Raw,ZI_Raw,APD1_Raw,APD2_Raw,J_Raw,MemIndex_Raw,ZI_um_Raw,ZI_OffsetIndex_Raw, zCam_Raw =  _MS.get_traj_raw_data4(StartEndPos, 0, ScanProject.ScanMatrix, ScanProject)
            txyz_Traj_Piezo=np.vstack((T_Raw,XP_Raw,YP_Raw,ZI_um_Raw))
            txyz_Traj_Piezo=_MS.get_Scan_toImagCoordinates(ScanProject, txyz_Traj_Piezo)
            
            #ZeroPoint=np.array([XP_Raw[0],YP_Raw[0],ZI_um_Raw[0]])
                #MLE
            VPTMxyz_O=np.transpose(txyznM[1:4,:])
            
                #Start at zero time and at zero piezo pos
            VPTMxyz=VPTMxyz_O-txyz_Traj_Piezo[1:,0]#VPTMxyz_O[0,:]
            txyzCVE_D[:,0]=(txyzCVE_D[:,0]-Vt[0])*1e3 #[ms]
            txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-Vt[0])*1e3 #[ms]
            Vt=(Vt-Vt[0])*1e3 #[ms]

            txyz_Traj_Piezo=(txyz_Traj_Piezo.T-txyz_Traj_Piezo[:,0]).T
            txyz_Traj_Piezo[0,:]=txyz_Traj_Piezo[0,:]*1e3

                # Fourier spectra of the MLE signals
            if fft!=0:
                if 0:
                    wndw_fft=20
                    wndwS_fft=wndw_fft #if 1 it is a moving average, if equal to wndw it is a fix window average
                    T_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,Vt)
                    T_fft=T_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]

                    x_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,0])
                    x_fft=x_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    y_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,1])
                    y_fft=y_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    z_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,2])
                    z_fft=z_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    
                    fft_x_MLE, fx = _MS.fft_processing(x_fft,T_fft)
                    fft_y_MLE, fy = _MS.fft_processing(y_fft,T_fft)
                    fft_z_MLE, fz = _MS.fft_processing(z_fft,T_fft)
                else:
                    fft_x_MLE, fx = _MS.fft_processing(VPTMxyz[:,0],Vt)
                    fft_y_MLE, fy = _MS.fft_processing(VPTMxyz[:,1],Vt)
                    fft_z_MLE, fz = _MS.fft_processing(VPTMxyz[:,2],Vt)

                #Running mean on the trajectory
            wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            T_Mean= np.convolve(np.ones(wndw)/wndw,Vt)
            T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
            VPTMxyz_Mean=np.zeros((len(T_Mean),3))
            VPTMxyz_Var=np.zeros((len(T_Mean),3))
            for i in range(3):
                VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,VPTMxyz[:,i])
                VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

                VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(VPTMxyz[:,i])**2) #Var(X)=E[X^2]-E[X]^2
                VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2

            #np.convolve(np.ones(wndw)/wndw,np.mean(txyzCVE_D[:,1:], axis=1))[wndw-1:-wndw+1:wndwS]

                #Create a D value colored traj plot
            x_pix=(VPTMxyz_O[:,0])/CamRes
            y_pix=(VPTMxyz_O[:,1])/CamRes
            z_pix=(VPTMxyz_O[:,1])/CamRes
            #points=np.array([x_pix, y_pix, z_pix]).T.reshape(-1, 1, 3) 
            points=np.array([x_pix, y_pix]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
        
            D_norm=np.log(np.mean(txyzCVE_D[:,1:], axis=1)) #(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))/ np.max(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))
            D_norm[np.where(np.isnan(D_norm))]=0
            D_norm=1/(1+np.exp(-(D_norm)))
            t_D=np.concatenate((txyzCVE_D[:,0],[Vt[-1]]))
            D_traj=np.zeros(Vt.shape)
            j=0
            for i in range(1,len(t_D)):
                while Vt[j]<=t_D[i]:
                    D_traj[j]=D_norm[i-1]
                    j+=1
                    if j>=len(Vt):
                        break

            colors = plt.cm.cool(D_traj)

                #Polarization 
            wndwN=int(ScanProject.TrajectorySelectionRules['wndw'])
            OfInterest=np.where((ScanProject.ScanMatrix[0,:]>=TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<=TStartEnd[1]))[0]
            OfInterest=np.append(OfInterest,OfInterest[-1]+np.arange(wndwN)+1)
            N12=ScanProject.ScanMatrix[3:5,OfInterest]
            PolT=ScanProject.ScanMatrix[0,OfInterest]

            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1=np.convolve(np.ones(wndwN)/wndwN,N12[0,:])
            N1=N1[wndwN-1:-wndwN+1:wndwS]
            N2=np.convolve(np.ones(wndwN)/wndwN,N12[1,:])
            N2=N2[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]

            
            wndwN=5
            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1_mean=np.convolve(np.ones(wndwN)/wndwN,N1)
            N1_mean=N1_mean[wndwN-1:-wndwN+1:wndwS]
            N2_mean=np.convolve(np.ones(wndwN)/wndwN,N2)
            N2_mean=N2_mean[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]
            PolT=(PolT-PolT[0])*1e3 #ms
            Pol=(N1_mean-N2_mean)/(N1_mean+N2_mean)

        ####################################
            # Generate the inverse-Gamma distributions to plot
            dist_params=HMM_Model['Distributions']
            Occupancy_estimated=HMM_Model['Occupancy']
            DBias=HMM_Model['DBias']

            DFittParam=np.linspace(-1,10,num=200)
            invGamma={}
            invGammaAll=np.zeros(DFittParam.shape)
            for state,Occ in zip(dist_params,Occupancy_estimated):
                param=dist_params[state]
                invGamma[state]=stats.invgamma.pdf(DFittParam+DBias, a=param[0],scale=param[1])
                invGammaAll=invGammaAll + Occ*invGamma[state]


            NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
            invGammaAll=invGammaAll/NormFactor

        #####################################################################
            if MovieStepSize>0: FrameIndex=0
            else: FrameIndex=-1

            movieFlag=1
            while FrameIndex < len(Vt) and movieFlag==1:
        
                # Figure construction
                fig = plt.figure(constrained_layout=False)
                fig.set_size_inches(3*5, 2*5, forward=True)
                gs = fig.add_gridspec(nrows=7, ncols=6, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.2, hspace=1.0)

                ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
                ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

                ax0=fig.add_subplot(gs[2:4, 0:2]) #MSD, CVE 
                ax4=fig.add_subplot(gs[2:4, 2])  #hist D      
                if fft==0:
                    ax1=fig.add_subplot(gs[4:6, 0:2]) #Sigma               
                    ax5=fig.add_subplot(gs[4:6, 2])  #hist sigma2     
                else:
                    ax1=fig.add_subplot(gs[4:5, 0:2]) #Sigma
                    ax5=fig.add_subplot(gs[4:5, 2])  #hist sigma2     
                    ax15=fig.add_subplot(gs[5, 0:2])  #fft     

                ax9=fig.add_subplot(gs[6, 0:2]) #Photon count 
                ax10=fig.add_subplot(gs[6, 2])  #Photon hist

                ax11=fig.add_subplot(gs[2:6, 3:6]) #Image 
                ax12=fig.add_subplot(gs[0:2, 4:5]) #Info
                
                ax13=fig.add_subplot(gs[6, 3:5]) #Pol count 
                ax14=fig.add_subplot(gs[6, 5])  #Pol hist

        
                #ax0-ax1
                ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:4], axis=1),'-k',alpha=0.5, label='CVE xyz');        #D CVE MLE data
                #ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:3], axis=1),'-k',alpha=0.5, label='CVE xy');        #D CVE MLE data
                #ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,1],'-r',alpha=0.5, label='CVE x');        #D CVE MLE data
                #ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,2],'-g',alpha=0.5, label='CVE y');        #D CVE MLE data
                #ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,3],'-b',alpha=0.5, label='CVE z');        #D CVE MLE data
                ax0.plot(TgroundtruthAll,DgroundtruthAll,'-g',alpha=0.5, label='Gtruth');        #D groundtruth
                
                ax0.plot(txyzCVE_D[:,0],HMM_prediction,'-r',alpha=0.5, label='Resedual:' + str(np.round(np.sum((np.mean(txyzCVE_D[:,1:], axis=1)-HMM_prediction)**2),decimals=2)));        #D CVE MLE data
                if MovieStepSize>0: ax0.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax0.set_ylabel("D [um2/s]"); ax0.grid(True); 
                ax0.legend()

                ax1.plot(txyzCVE_Sigma[:,0],np.mean(txyzCVE_Sigma[:,1:],axis=1),'-g',alpha=0.5, label='CVE'); #Sigma2 CVE MLE data
                ax1.plot(T_Mean,np.mean(VPTMxyz_Var,axis=1),'-b',alpha=0.5, label='Var' + str(wndw));         #Var CVE MLE data
                if MovieStepSize>0: ax1.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax1.set_ylabel("sigma2 [um2]"); ax1.grid(True);
                ax1.legend()

                #ax4 Histogram D CVE and MSD
                ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), bins=DFittParam, facecolor='g', density=True); 
                #ax4.axvline(np.mean(txyzCVE_D[:,1:]), color='g',linestyle='-',label='CVE:' + str(np.round(np.mean(txyzCVE_D[:,1:]),decimals=2)))
                #for state,Occ in zip(invGamma,Occupancy_estimated):
                #    ax4.plot(DFittParam,Occ*invGamma[state]/NormFactor,label=state)
                #ax4.plot(DFittParam,invGammaAll,'-k',label='envelope')
                ax4.set_xlabel("D [um^2/s]")
                ax4.legend()

                #ax5 Histogram Sigma Variance 
                ax5.hist((np.mean(txyzCVE_Sigma[:,1:],axis=1)/np.abs(np.mean(txyzCVE_Sigma[:,1:],axis=1)))*np.sqrt(np.abs(np.mean(txyzCVE_Sigma[:,1:],axis=1)))*1e3, 20, facecolor='g', alpha=0.5); 
                ax5.hist(np.sqrt(np.mean(VPTMxyz_Var,axis=1))*1e3, 20, facecolor='b', alpha=0.5); 
                ax5.axvline(np.sqrt(np.mean(VPTMxyz_Var))*1e3, color='b',linestyle='-',label='Var' + str(wndw) +':'+ str(np.round(np.sqrt(np.mean(VPTMxyz_Var))*1e3,decimals=2)))
                ax5.axvline((np.abs(np.mean(txyzCVE_Sigma[:,1:]))/np.mean(txyzCVE_Sigma[:,1:]))*np.sqrt(np.abs(np.mean(txyzCVE_Sigma[:,1:])))*1e3, color='g',linestyle='-',label='CVE:' + str(np.round(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3,decimals=2)))
                ax5.set_xlabel("Sigma [nm]")
                ax5.legend()

                #ax3
                ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[1,:],'-m',alpha=0.4,label='Piezo');  #Piezo data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[2,:],'-m',alpha=0.4);                #Piezo data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[3,:],'-m',alpha=0.4);          #Piezo data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.4,label='Mean:' + str(wndw));  #MLE data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.4);  #MLE data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.4)   #MLE data
                ax3.plot((T_Raw-T_Raw[0])*1e3,zCam_Raw-zCam_Raw[0],'-b')
                if MovieStepSize>0: ax3.axvline(Vt[FrameIndex], color='r',linestyle='-')

                ax3.set_ylabel("[um]"); 
                ax3.grid(True);
                ax3.legend()
        
                #ax6
                if np.sum(np.isnan(VPTMxyz))==0:
                    ax6.hist(np.sqrt(VPTMxyz_Var[:,0]), 20, facecolor='r', alpha=0.4); 
                    ax6.hist(np.sqrt(VPTMxyz_Var[:,1]), 20, facecolor='g', alpha=0.4); 
                    ax6.hist(np.sqrt(VPTMxyz_Var[:,2]), 20, facecolor='b', alpha=0.4)          
                    ax6.set_xlabel('STD [um] over: ' + str(wndw))
        
                #ax9-ax10 Photon counts
                ax9.plot(Vt,Ni,'-') 
                if MovieStepSize>0: ax9.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax10.hist(Ni, 20)        
                ax9.set_ylabel('Ph #')
                ax9.grid()

                ax13.plot(PolT,Pol,'-b')
                #if MovieStepSize>0: ax13.axvline(PolT[FrameIndex], color='r',linestyle='-')
                ax13.grid(); ax13.set_ylim(-1,1)
                ax14.hist(Pol, bins=np.linspace(-1,1,40))        

                #fft ax15               
                if fft!=0:
                    ax15.semilogy(fx, fft_x_MLE,c='r', alpha=0.4, label='x-MLE')
                    ax15.semilogy(fy, fft_y_MLE,c='g', alpha=0.4, label='y-MLE')
                    ax15.semilogy(fz, fft_z_MLE,c='b', alpha=0.4, label='z-MLE')
                    ax15.set_xlabel('khz')
                    ax15.set_ylim((10**-5,10**-2))
                    ax15.grid(b=True, which='major', color='b', linestyle='-')
                    ax15.grid(b=True, which='minor', color='b', linestyle='--')
                #Cell image
                    #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
                ax11.imshow(ImTot[:,:,0],cmap='gray',alpha=1.0)
                #ax[0].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)
                #ax[0].imshow(ImTot[:,:,2],cmap='Reds',alpha=0.3)
                if CellMask:
                    ax11.contour(mask,levels=0, linewidths=1, colors='blue', linestyles='solid')
                    ax11.imshow(mask,cmap='Reds',alpha=0.3)

                if MovieStepSize>0: ls=LineCollection(segments[:FrameIndex], linewidths=1,colors=colors[:FrameIndex], linestyle='solid')
                else: ls=LineCollection(segments, linewidths=1,colors=colors, linestyle='solid')

                line=ax11.add_collection(ls)
                #fig.colorbar(line, ax=ax11)
                #ax11.plot(x,y,'-', c=color[0,:])

                xmean=np.mean(x_pix)
                ymean=np.mean(y_pix)
        
                ticks=np.arange(0,ImTot[:,:,0].shape[0],4)
                xticks=np.linspace(xmean-xyImageBox,xmean+xyImageBox,10)
                yticks=np.linspace(ymean-xyImageBox,ymean+xyImageBox,10)
                xtick_labels=np.round( (xticks-(xmean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]
                ytick_labels=np.round( (yticks-(ymean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]

                ax11.set_xticks(xticks); ax11.set_yticks(yticks)
                ax11.set_xticklabels(xtick_labels); ax11.set_yticklabels(ytick_labels)
                ax11.set_xlim([xmean-xyImageBox,xmean+xyImageBox])
                ax11.set_ylim([ymean-xyImageBox,ymean+xyImageBox])
                ax11.set_xlabel('um'); ax11.set_ylabel('um')

                #ax12 Info box            
                Values=np.zeros(HMM_Model['TransMat'].shape+np.array([3,0])) 
                Values[:HMM_Model['TransMat'].shape[0],:HMM_Model['TransMat'].shape[1]]=HMM_Model['TransMat']
                Values[HMM_Model['TransMat'].shape[0],:]=HMM_Model['D']
                Values[HMM_Model['TransMat'].shape[0]+1,:]=HMM_Model['Dwelltime']
                Values[HMM_Model['TransMat'].shape[0]+2,:]=HMM_Model['Occupancy']
                Values=np.round(Values,decimals=2)

                S=()
                for state in HMM_Model['Distributions']:
                    S+=(state,)
                StatesLables=list(S)
                S+=('D:', 'dwelltime:','POcc',)
                LablesAll=list(S)

                ax12.pcolor(Values*0, cmap=plt.cm.Greys)
                #ax12.imshow(Values)
                ax12.set_xticks(np.arange(len(StatesLables))+0.5, minor=False)
                ax12.set_yticks(np.arange(len(LablesAll))+0.5, minor=False)
                ax12.set_xticklabels(StatesLables, minor=False)
                ax12.set_yticklabels(LablesAll, minor=False)
                ax12.invert_yaxis()
                #ax12.xaxis.tick_top()

                for i in range(len(LablesAll)):
                    for j in range(len(StatesLables)):
                        text = ax12.text(j+0.5, i+0.5, Values[i, j],
                                        ha="center", va="center", color="k")                
                #ax12.axis('off')
                

                #########################################################
                if do_save_plot:
                    if MovieStepSize>0:
                        file="MovieFrame_" + str(TrajToGet) + '_D_traj_' + str(FrameIndex)+ FileType
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Movie_Traj_D_filtered_HMM_' + str(len(StatesLables)) + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    else:
                        file="T_" + str(TrajToGet) + '_D_traj' + FileType
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_D_filtered_HMM_' + str(len(StatesLables)) + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    fig.savefig(path + file, bbox_inches='tight')
                    plt.close()
                    print('Done!')
                else:
                    plt.show()

                FrameIndex+=MovieStepSize
                if MovieStepSize<=0: break
                elif FrameIndex>=len(Vt):break

    def plot_Hist_Traj_MDS10(ScanProject,TrajToGet,TrajDictionary,HMM_prediction=None,HMM_Model=None,MovieStepSize=0,do_save_plot=0,FolderEnd='',fft=0):
            """
            This function is to plot 
                trajectores 
                Cell image
                HMM predicted traj
                zCam z data

                No info box for the HMM 
                Added traj offset to see each trajectroy better 
                Rearanging the plot positions 
            """
            TrajOffset=[0,0,0.5,1] #txyz
            xyImageBox=32 #pix 35
            CellMask=1
        
            if CellMask:
                mask=ScanProject.CamCellMask[:,:,0]>0
            CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]
            try:
                ImTot=ScanProject.ImgOverlay
            except:
                print('No cell image found, using cell mask')
                ImTot=np.repeat(np.expand_dims(mask*1.0,axis=2),3,axis=2)

            wndw=10
            txyznM=TrajDictionary[TrajToGet]*1.0

                #CVE D and sigma evaluation
            txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
            txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

            txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu]
            txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]

            ni=txyznM[4,:]

                #time
            Vt=txyznM[0,:]              #[s]

                #Piezo stage position
            TStartEnd=[Vt[0],Vt[-1]] 
            StartEndPos=np.where(np.abs(np.diff((ScanProject.ScanMatrix[0,:]>TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<TStartEnd[1]))))[0]
            StartEndPos=np.expand_dims(np.append(np.array([0]),StartEndPos),axis=1)
            
            T_Raw,XP_Raw,YP_Raw,XE_Raw,YE_Raw,XO_Raw,YO_Raw,ZI_Raw,APD1_Raw,APD2_Raw,J_Raw,MemIndex_Raw,ZI_um_Raw,ZI_OffsetIndex_Raw, zCam_Raw =  _MS.get_traj_raw_data4(StartEndPos, 0, ScanProject.ScanMatrix, ScanProject)
            txyz_Traj_Piezo=np.vstack((T_Raw,XP_Raw,YP_Raw,ZI_um_Raw))
            txyz_Traj_Piezo=_MS.get_Scan_toImagCoordinates(ScanProject, txyz_Traj_Piezo)
            
            #ZeroPoint=np.array([XP_Raw[0],YP_Raw[0],ZI_um_Raw[0]])
                #MLE
            VPTMxyz_O=np.transpose(txyznM[1:4,:])
            
                #Start at zero time and at zero piezo pos
            VPTMxyz=VPTMxyz_O-txyz_Traj_Piezo[1:,0]+TrajOffset[1:] #VPTMxyz_O[0,:]
            Vt=(Vt-Vt[0])*1e3 #[ms]

            txyz_Traj_Piezo=(txyz_Traj_Piezo.T-txyz_Traj_Piezo[:,0]+TrajOffset).T
            txyz_Traj_Piezo[0,:]=txyz_Traj_Piezo[0,:]*1e3

                # Fourier spectra of the MLE signals
            if fft!=0:
                if 0:
                    wndw_fft=20
                    wndwS_fft=wndw_fft #if 1 it is a moving average, if equal to wndw it is a fix window average
                    T_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,Vt)
                    T_fft=T_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]

                    x_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,0])
                    x_fft=x_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    y_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,1])
                    y_fft=y_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    z_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,2])
                    z_fft=z_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    
                    fft_x_MLE, fx = _MS.fft_processing(x_fft,T_fft)
                    fft_y_MLE, fy = _MS.fft_processing(y_fft,T_fft)
                    fft_z_MLE, fz = _MS.fft_processing(z_fft,T_fft)
                else:
                    fft_x_MLE, fx = _MS.fft_processing(VPTMxyz[:,0],Vt)
                    fft_y_MLE, fy = _MS.fft_processing(VPTMxyz[:,1],Vt)
                    fft_z_MLE, fz = _MS.fft_processing(VPTMxyz[:,2],Vt)

                #Running mean on the trajectory
            wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            T_Mean= np.convolve(np.ones(wndw)/wndw,Vt)
            T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
            VPTMxyz_Mean=np.zeros((len(T_Mean),3))
            VPTMxyz_Var=np.zeros((len(T_Mean),3))
            for i in range(3):
                VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,VPTMxyz[:,i])
                VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

                VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(VPTMxyz[:,i])**2) #Var(X)=E[X^2]-E[X]^2
                VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2

            #np.convolve(np.ones(wndw)/wndw,np.mean(txyzCVE_D[:,1:], axis=1))[wndw-1:-wndw+1:wndwS]

                #Create a D value colored traj plot
            x_pix=(VPTMxyz_O[:,0])/CamRes
            y_pix=(VPTMxyz_O[:,1])/CamRes
            z_pix=(VPTMxyz_O[:,1])/CamRes
            #points=np.array([x_pix, y_pix, z_pix]).T.reshape(-1, 1, 3) 
            points=np.array([x_pix, y_pix]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
        
            D_norm=np.log(np.mean(txyzCVE_D[:,1:], axis=1)) #(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))/ np.max(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))
            D_norm[np.where(np.isnan(D_norm))]=0
            D_norm=1/(1+np.exp(-(D_norm)))
            t_D=np.concatenate((txyzCVE_D[:,0],[Vt[-1]]))
            D_traj=np.zeros(Vt.shape)
            j=0
            for i in range(1,len(t_D)):
                while Vt[j]<=t_D[i]:
                    D_traj[j]=D_norm[i-1]
                    j+=1
                    if j>=len(Vt):
                        break

            colors = plt.cm.cool(D_traj)

                #Polarization 
            wndwN=int(ScanProject.TrajectorySelectionRules['wndw'])
            OfInterest=np.where((ScanProject.ScanMatrix[0,:]>=TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<=TStartEnd[1]))[0]
            OfInterest=np.append(OfInterest,OfInterest[-1]+np.arange(wndwN)+1)
            N12=ScanProject.ScanMatrix[3:5,OfInterest]
            PolT=ScanProject.ScanMatrix[0,OfInterest]

            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1=np.convolve(np.ones(wndwN)/wndwN,N12[0,:])
            N1=N1[wndwN-1:-wndwN+1:wndwS]
            N2=np.convolve(np.ones(wndwN)/wndwN,N12[1,:])
            N2=N2[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]

            
            wndwN=2
            wndwS=1  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1_mean=np.convolve(np.ones(wndwN)/wndwN,N1)
            N1_mean=N1_mean[wndwN-1:-wndwN+1:wndwS]
            N2_mean=np.convolve(np.ones(wndwN)/wndwN,N2)
            N2_mean=N2_mean[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]
            PolT=(PolT-PolT[0])*1e3 #ms
            Pol=(N1_mean-N2_mean)/(N1_mean+N2_mean)

        ####################################
            # Generate the inverse-Gamma distributions to plot
            if HMM_Model!=None:
                dist_params=HMM_Model['Distributions']
                Occupancy_estimated=HMM_Model['Occupancy']
                DBias=HMM_Model['DBias']
                DFittParam=np.linspace(-DBias,10,num=200)
                invGamma={}
                invGammaAll=np.zeros(DFittParam.shape)
                for state,Occ in zip(dist_params,Occupancy_estimated):
                    param=dist_params[state]
                    invGamma[state]=stats.invgamma.pdf(DFittParam+DBias, a=param[0],scale=param[1])
                    invGammaAll=invGammaAll + Occ*invGamma[state]
                NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
                invGammaAll=invGammaAll/NormFactor
            else:
                DFittParam=np.linspace(0,10,num=200)
                dist_params=0
                Occupancy_estimated=0
                DBias=0

        #####################################################################
            if MovieStepSize>0: FrameIndex=0
            else: FrameIndex=-1

            movieFlag=1
            while FrameIndex < len(Vt) and movieFlag==1:
        
                # Figure construction
                fig = plt.figure(constrained_layout=False)
                fig.set_size_inches(3*5, 2*5, forward=True)
                gs = fig.add_gridspec(nrows=7, ncols=6, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.2, hspace=1.0)

                ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
                ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

                ax0=fig.add_subplot(gs[2:4, 0:2]) #MSD, CVE 
                ax4=fig.add_subplot(gs[2:4, 2])  #hist D      
                if fft==0:
                    ax1=fig.add_subplot(gs[4:6, 0:2]) #Sigma               
                    ax5=fig.add_subplot(gs[4:6, 2])  #hist sigma2     
                else:
                    ax1=fig.add_subplot(gs[4:5, 0:2]) #Sigma
                    ax5=fig.add_subplot(gs[4:5, 2])  #hist sigma2     
                    ax15=fig.add_subplot(gs[5:7, 3:5])  #fft     

                ax9=fig.add_subplot(gs[5:7, 0:2]) #Photon count 
                ax10=fig.add_subplot(gs[5:7, 2])  #Photon hist

                ax11=fig.add_subplot(gs[0:4, 3:6]) #Image 
                
                ax13=fig.add_subplot(gs[4, 3:5]) #Pol count 
                ax14=fig.add_subplot(gs[4, 5])  #Pol hist

        
                #ax0-ax1
                #ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:4], axis=1),'-k',alpha=0.5, label='CVE xyz');        #D CVE MLE data
                ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:3], axis=1),'-k',alpha=0.5, label='CVE xy');        #D CVE MLE data
                #ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,1],'-r',alpha=0.5, label='CVE x');        #D CVE MLE data
                #ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,2],'-g',alpha=0.5, label='CVE y');        #D CVE MLE data
                ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,3],'-b',alpha=0.5, label='CVE z');        #D CVE MLE data
                #ax0.plot(txyzCVE_D[:,0],HMM_prediction,'-r',alpha=0.5, label='Resedual:' + str(np.round(np.sum((np.mean(txyzCVE_D[:,1:], axis=1)-HMM_prediction)**2),decimals=2)));        #D CVE MLE data
                if MovieStepSize>0: ax0.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax0.set_ylabel("D [um2/s]"); ax0.grid(True); 
                ax0.legend()

                ax1.plot(txyzCVE_Sigma[:,0],np.mean(txyzCVE_Sigma[:,1:],axis=1),'-g',alpha=0.5, label='CVE'); #Sigma2 CVE MLE data
                ax1.plot(T_Mean,np.mean(VPTMxyz_Var,axis=1),'-b',alpha=0.5, label='Var' + str(wndw));         #Var CVE MLE data
                if MovieStepSize>0: ax1.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax1.set_ylabel("sigma2 [um2]"); ax1.grid(True);
                ax1.legend()

                #ax4 Histogram D CVE and MSD
                ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), bins=DFittParam, facecolor='g', density=True); 
                #ax4.axvline(np.mean(txyzCVE_D[:,1:]), color='g',linestyle='-',label='CVE:' + str(np.round(np.mean(txyzCVE_D[:,1:]),decimals=2)))
                #for state,Occ in zip(invGamma,Occupancy_estimated):
                #    ax4.plot(DFittParam,Occ*invGamma[state]/NormFactor,label=state)
                #ax4.plot(DFittParam,invGammaAll,'-k',label='envelope')
                ax4.set_xlabel("D [um^2/s]")
                ax4.legend()

                #ax5 Histogram Sigma Variance 
                ax5.hist(np.sqrt(np.mean(txyzCVE_Sigma[:,1:],axis=1))*1e3, 20, facecolor='g', alpha=0.5); 
                ax5.hist(np.sqrt(np.mean(VPTMxyz_Var,axis=1))*1e3, 20, facecolor='b', alpha=0.5); 
                ax5.axvline(np.sqrt(np.mean(VPTMxyz_Var))*1e3, color='b',linestyle='-',label='Var' + str(wndw) +':'+ str(np.round(np.sqrt(np.mean(VPTMxyz_Var))*1e3,decimals=2)))
                ax5.axvline(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3, color='g',linestyle='-',label='CVE:' + str(np.round(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3,decimals=2)))
                ax5.set_xlabel("Sigma [nm]")
                ax5.legend()

                #ax3
                ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[1,:],'-m',alpha=0.4,label='Piezo');  #Piezo data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[2,:],'-m',alpha=0.4);                #Piezo data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[3,:],'-m',alpha=0.4);          #Piezo data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.4,label='Mean:' + str(wndw));  #MLE data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.4);  #MLE data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.4)   #MLE data
                ax3.plot((T_Raw-T_Raw[0])*1e3,zCam_Raw-zCam_Raw[0]+TrajOffset[3],'-b')
                if MovieStepSize>0: ax3.axvline(Vt[FrameIndex], color='r',linestyle='-')

                ax3.set_ylabel("[um]"); 
                ax3.grid(True);
                ax3.legend()
        
                #ax6
                if np.sum(np.isnan(VPTMxyz))==0:
                    ax6.hist(np.sqrt(VPTMxyz_Var[:,0]), 20, facecolor='r', alpha=0.4); 
                    ax6.hist(np.sqrt(VPTMxyz_Var[:,1]), 20, facecolor='g', alpha=0.4); 
                    ax6.hist(np.sqrt(VPTMxyz_Var[:,2]), 20, facecolor='b', alpha=0.4)          
                    ax6.set_xlabel('STD [um] over: ' + str(wndw))
        
                #ax9-ax10 Photon counts
                ax9.plot(Vt,ni,'-') 
                if MovieStepSize>0: ax9.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax10.hist(ni, 20)        
                ax9.set_ylabel('Ph #')
                ax9.set_xlabel('ms')
                ax9.grid()

                ax13.plot(PolT,Pol,'-b')
                #if MovieStepSize>0: ax13.axvline(PolT[FrameIndex], color='r',linestyle='-')
                ax13.grid(); ax13.set_ylim(-1,1)
                ax14.hist(Pol, bins=np.linspace(-1,1,40))        

                #fft ax15               
                if fft!=0:
                    ax15.semilogy(fx, fft_x_MLE,c='r', alpha=0.4, label='x-MLE')
                    ax15.semilogy(fy, fft_y_MLE,c='g', alpha=0.4, label='y-MLE')
                    ax15.semilogy(fz, fft_z_MLE,c='b', alpha=0.4, label='z-MLE')
                    ax15.set_xlabel('khz')
                    ax15.set_ylim((10**-5,10**-2))
                    #ax15.grid(b=True, which='major', color='b', linestyle='-')
                    #ax15.grid(b=True, which='minor', color='b', linestyle='--')
                #Cell image
                    #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
                ax11.imshow(ImTot[:,:,0],cmap='gray',alpha=1.0)
                #ax[0].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)
                #ax[0].imshow(ImTot[:,:,2],cmap='Reds',alpha=0.3)
                if CellMask:
                    ax11.contour(mask,levels=0, linewidths=1, colors='blue', linestyles='solid')
                    ax11.imshow(mask,cmap='Reds',alpha=0.3)

                if MovieStepSize>0: ls=LineCollection(segments[:FrameIndex], linewidths=1,colors=colors[:FrameIndex], linestyle='solid')
                else: ls=LineCollection(segments, linewidths=1,colors=colors, linestyle='solid')

                line=ax11.add_collection(ls)
                #fig.colorbar(line, ax=ax11)
                #ax11.plot(x,y,'-', c=color[0,:])

                xmean=np.mean(x_pix)
                ymean=np.mean(y_pix)
        
                ticks=np.arange(0,ImTot[:,:,0].shape[0],4)
                xticks=np.linspace(xmean-xyImageBox,xmean+xyImageBox,10)
                yticks=np.linspace(ymean-xyImageBox,ymean+xyImageBox,10)
                xtick_labels=np.round( (xticks-(xmean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]
                ytick_labels=np.round( (yticks-(ymean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]

                ax11.set_xticks(xticks); ax11.set_yticks(yticks)
                ax11.set_xticklabels(xtick_labels); ax11.set_yticklabels(ytick_labels)
                ax11.set_xlim([xmean-xyImageBox,xmean+xyImageBox])
                ax11.set_ylim([ymean-xyImageBox,ymean+xyImageBox])
                ax11.set_xlabel('um'); ax11.set_ylabel('um')
                

                #########################################################
                if do_save_plot:
                    if MovieStepSize>0:
                        file="MovieFrame_" + str(TrajToGet) + '_D_traj_' + str(FrameIndex)+ '.png'
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Movie_Traj_D_filtered_' + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    else:
                        file="T_" + str(TrajToGet) + '_D_traj' + '.png'
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_D_filtered_' + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    fig.savefig(path + file)
                    plt.close()
                    print('Done!')
                else:
                    plt.show()

                FrameIndex+=MovieStepSize
                if MovieStepSize<=0: break
                elif FrameIndex>=len(Vt):break

    def plot_Hist_Traj_MDS10b(ScanProject,TrajToGet,TrajDictionary,HMM_prediction=None,HMM_Model=None,MovieStepSize=0,do_save_plot=0,FolderEnd='',fft=0, PhotonHMM=None,photon_count_HMM_dic=None, fileType='.png'):
            """
            This function is to plot 
                trajectores 
                Cell image
                HMM predicted traj
                zCam z data
                HMM photon prediction

                No info box for the HMM 
                Added traj offset to see each trajectroy better 
                Rearanging the plot positions 
            """
            TrajOffset=[0,0,0.5,1] #txyz
            xyImageBox=32 #pix 35
            CellMask=1
        
            if CellMask:
                mask=ScanProject.CamCellMask[:,:,0]>0
            CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]
            try:
                ImTot=ScanProject.ImgOverlay
            except:
                print('No cell image found, using cell mask')
                ImTot=np.repeat(np.expand_dims(mask*1.0,axis=2),3,axis=2)


            if PhotonHMM!=None:
                PhotonHMMVector=PhotonHMM['HMM_prediction']*ScanProject.TrajectorySelectionRules['wndw']
            else:
                PhotonHMMVector=np.zeros((ScanProject.ScanMatrix[0,:]).shape)

            if photon_count_HMM_dic!=None:
                PhotonHMMParam=photon_count_HMM_dic['HMM_parameters']*ScanProject.TrajectorySelectionRules['wndw']
            else:
                PhotonHMMParam='Not found'


            wndw=10
            txyznM=TrajDictionary[TrajToGet]*1.0

                #CVE D and sigma evaluation
            txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
            txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

            txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu]
            txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]

            ni=txyznM[4,:]

                #time
            Vt=txyznM[0,:]              #[s]

                #Piezo stage position
            TStartEnd=[Vt[0],Vt[-1]] 
            StartEndPos=np.where(np.abs(np.diff((ScanProject.ScanMatrix[0,:]>TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<TStartEnd[1]))))[0]
            StartEndPos=np.expand_dims(np.append(np.array([0]),StartEndPos),axis=1)
            
            T_Raw,XP_Raw,YP_Raw,XE_Raw,YE_Raw,XO_Raw,YO_Raw,ZI_Raw,APD1_Raw,APD2_Raw,J_Raw,MemIndex_Raw,ZI_um_Raw,ZI_OffsetIndex_Raw, zCam_Raw =  _MS.get_traj_raw_data4(StartEndPos, 0, ScanProject.ScanMatrix, ScanProject)
            txyz_Traj_Piezo=np.vstack((T_Raw,XP_Raw,YP_Raw,ZI_um_Raw))
            txyz_Traj_Piezo=_MS.get_Scan_toImagCoordinates(ScanProject, txyz_Traj_Piezo)
            
                #Photon HMM prediction, use same time as piezo
            PhotonHMMVector=PhotonHMMVector[int(StartEndPos[1]):int(StartEndPos[2]+1)]

            #ZeroPoint=np.array([XP_Raw[0],YP_Raw[0],ZI_um_Raw[0]])
                #MLE
            VPTMxyz_O=np.transpose(txyznM[1:4,:])
            
                #Start at zero time and at zero piezo pos
            VPTMxyz=VPTMxyz_O-txyz_Traj_Piezo[1:,0]+TrajOffset[1:] #VPTMxyz_O[0,:]
            Vt=(Vt-Vt[0])*1e3 #[ms]

            txyz_Traj_Piezo=(txyz_Traj_Piezo.T-txyz_Traj_Piezo[:,0]+TrajOffset).T
            txyz_Traj_Piezo[0,:]=txyz_Traj_Piezo[0,:]*1e3

                # Fourier spectra of the MLE signals
            if fft!=0:
                if 0:
                    wndw_fft=20
                    wndwS_fft=wndw_fft #if 1 it is a moving average, if equal to wndw it is a fix window average
                    T_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,Vt)
                    T_fft=T_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]

                    x_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,0])
                    x_fft=x_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    y_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,1])
                    y_fft=y_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    z_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,2])
                    z_fft=z_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    
                    fft_x_MLE, fx = _MS.fft_processing(x_fft,T_fft)
                    fft_y_MLE, fy = _MS.fft_processing(y_fft,T_fft)
                    fft_z_MLE, fz = _MS.fft_processing(z_fft,T_fft)
                else:
                    fft_x_MLE, fx = _MS.fft_processing(VPTMxyz[:,0],Vt)
                    fft_y_MLE, fy = _MS.fft_processing(VPTMxyz[:,1],Vt)
                    fft_z_MLE, fz = _MS.fft_processing(VPTMxyz[:,2],Vt)

                #Running mean on the trajectory
            wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            T_Mean= np.convolve(np.ones(wndw)/wndw,Vt)
            T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
            VPTMxyz_Mean=np.zeros((len(T_Mean),3))
            VPTMxyz_Var=np.zeros((len(T_Mean),3))
            for i in range(3):
                VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,VPTMxyz[:,i])
                VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

                VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(VPTMxyz[:,i])**2) #Var(X)=E[X^2]-E[X]^2
                VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2

            #np.convolve(np.ones(wndw)/wndw,np.mean(txyzCVE_D[:,1:], axis=1))[wndw-1:-wndw+1:wndwS]

                #Create a D value colored traj plot
            x_pix=(VPTMxyz_O[:,0])/CamRes
            y_pix=(VPTMxyz_O[:,1])/CamRes
            z_pix=(VPTMxyz_O[:,1])/CamRes
            #points=np.array([x_pix, y_pix, z_pix]).T.reshape(-1, 1, 3) 
            points=np.array([x_pix, y_pix]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
        
            D_norm=np.log(np.mean(txyzCVE_D[:,1:], axis=1)) #(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))/ np.max(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))
            D_norm[np.where(np.isnan(D_norm))]=0
            D_norm=1/(1+np.exp(-(D_norm)))
            t_D=np.concatenate((txyzCVE_D[:,0],[Vt[-1]]))
            D_traj=np.zeros(Vt.shape)
            j=0
            for i in range(1,len(t_D)):
                while Vt[j]<=t_D[i]:
                    D_traj[j]=D_norm[i-1]
                    j+=1
                    if j>=len(Vt):
                        break

            colors = plt.cm.cool(D_traj)

                #Polarization 
            wndwN=int(ScanProject.TrajectorySelectionRules['wndw'])
            OfInterest=np.where((ScanProject.ScanMatrix[0,:]>=TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<=TStartEnd[1]))[0]
            OfInterest=np.append(OfInterest,OfInterest[-1]+np.arange(wndwN)+1)
            N12=ScanProject.ScanMatrix[3:5,OfInterest]
            PolT=ScanProject.ScanMatrix[0,OfInterest]

            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1=np.convolve(np.ones(wndwN)/wndwN,N12[0,:])
            N1=N1[wndwN-1:-wndwN+1:wndwS]
            N2=np.convolve(np.ones(wndwN)/wndwN,N12[1,:])
            N2=N2[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]

            
            wndwN=2
            wndwS=1  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1_mean=np.convolve(np.ones(wndwN)/wndwN,N1)
            N1_mean=N1_mean[wndwN-1:-wndwN+1:wndwS]
            N2_mean=np.convolve(np.ones(wndwN)/wndwN,N2)
            N2_mean=N2_mean[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]
            PolT=(PolT-PolT[0])*1e3 #ms
            Pol=(N1_mean-N2_mean)/(N1_mean+N2_mean)

        ####################################
            # Generate the inverse-Gamma distributions to plot
            if HMM_Model!=None:
                dist_params=HMM_Model['Distributions']
                Occupancy_estimated=HMM_Model['Occupancy']
                DBias=HMM_Model['DBias']
                DFittParam=np.linspace(-DBias,10,num=200)
                invGamma={}
                invGammaAll=np.zeros(DFittParam.shape)
                for state,Occ in zip(dist_params,Occupancy_estimated):
                    param=dist_params[state]
                    invGamma[state]=stats.invgamma.pdf(DFittParam+DBias, a=param[0],scale=param[1])
                    invGammaAll=invGammaAll + Occ*invGamma[state]
                NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
                invGammaAll=invGammaAll/NormFactor
            else:
                DFittParam=np.linspace(0,10,num=200)
                dist_params=0
                Occupancy_estimated=0
                DBias=0

        #####################################################################
            try:
                if MovieStepSize>0: FrameIndex=0
                else: FrameIndex=-1

                movieFlag=1
                while FrameIndex < len(Vt) and movieFlag==1:
        
                    # Figure construction
                    fig = plt.figure(constrained_layout=False)
                    fig.set_size_inches(3*5, 2*5, forward=True)
                    gs = fig.add_gridspec(nrows=7, ncols=6, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.2, hspace=1.0)

                    ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
                    ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

                    ax0=fig.add_subplot(gs[2:4, 0:2]) #MSD, CVE 
                    ax4=fig.add_subplot(gs[2:4, 2])  #hist D      
                    if fft==0:
                        ax1=fig.add_subplot(gs[4:6, 0:2]) #Sigma               
                        ax5=fig.add_subplot(gs[4:6, 2])  #hist sigma2     
                    else:
                        ax1=fig.add_subplot(gs[4:5, 0:2]) #Sigma
                        ax5=fig.add_subplot(gs[4:5, 2])  #hist sigma2     
                        ax15=fig.add_subplot(gs[5:7, 3:5])  #fft     

                    ax9=fig.add_subplot(gs[5:7, 0:2]) #Photon count 
                    ax10=fig.add_subplot(gs[5:7, 2])  #Photon hist

                    ax11=fig.add_subplot(gs[0:4, 3:6]) #Image 
                
                    ax13=fig.add_subplot(gs[4, 3:5]) #Pol count 
                    ax14=fig.add_subplot(gs[4, 5])  #Pol hist

        
                    #ax0-ax1
                    #ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:4], axis=1),'-k',alpha=0.5, label='CVE xyz');        #D CVE MLE data
                    ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:3], axis=1),'-k',alpha=0.5, label='CVE xy');        #D CVE MLE data
                    #ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,1],'-r',alpha=0.5, label='CVE x');        #D CVE MLE data
                    #ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,2],'-g',alpha=0.5, label='CVE y');        #D CVE MLE data
                    ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,3],'-b',alpha=0.5, label='CVE z');        #D CVE MLE data
                    #ax0.plot(txyzCVE_D[:,0],HMM_prediction,'-r',alpha=0.5, label='Resedual:' + str(np.round(np.sum((np.mean(txyzCVE_D[:,1:], axis=1)-HMM_prediction)**2),decimals=2)));        #D CVE MLE data
                    if MovieStepSize>0: ax0.axvline(Vt[FrameIndex], color='r',linestyle='-')
                    ax0.set_ylabel("D [um2/s]"); ax0.grid(True); 
                    ax0.legend()

                    ax1.plot(txyzCVE_Sigma[:,0],np.mean(txyzCVE_Sigma[:,1:],axis=1),'-g',alpha=0.5, label='CVE'); #Sigma2 CVE MLE data
                    ax1.plot(T_Mean,np.mean(VPTMxyz_Var,axis=1),'-b',alpha=0.5, label='Var' + str(wndw));         #Var CVE MLE data
                    if MovieStepSize>0: ax1.axvline(Vt[FrameIndex], color='r',linestyle='-')
                    ax1.set_ylabel("sigma2 [um2]"); ax1.grid(True);
                    ax1.legend()

                    #ax4 Histogram D CVE and MSD
                    ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), bins=DFittParam, facecolor='g', density=True); 
                    #ax4.axvline(np.mean(txyzCVE_D[:,1:]), color='g',linestyle='-',label='CVE:' + str(np.round(np.mean(txyzCVE_D[:,1:]),decimals=2)))
                    #for state,Occ in zip(invGamma,Occupancy_estimated):
                    #    ax4.plot(DFittParam,Occ*invGamma[state]/NormFactor,label=state)
                    #ax4.plot(DFittParam,invGammaAll,'-k',label='envelope')
                    ax4.set_xlabel("D [um^2/s]")
                    ax4.legend()

                    #ax5 Histogram Sigma Variance 
                    ax5.hist(np.sqrt(np.mean(txyzCVE_Sigma[:,1:],axis=1))*1e3, 20, facecolor='g', alpha=0.5); 
                    ax5.hist(np.sqrt(np.mean(VPTMxyz_Var,axis=1))*1e3, 20, facecolor='b', alpha=0.5); 
                    ax5.axvline(np.sqrt(np.mean(VPTMxyz_Var))*1e3, color='b',linestyle='-',label='Var' + str(wndw) +':'+ str(np.round(np.sqrt(np.mean(VPTMxyz_Var))*1e3,decimals=2)))
                    ax5.axvline(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3, color='g',linestyle='-',label='CVE:' + str(np.round(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3,decimals=2)))
                    ax5.set_xlabel("Sigma [nm]")
                    ax5.legend()

                    #ax3
                    ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
                    ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
                    ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
                    ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[1,:],'-m',alpha=0.4,label='Piezo');  #Piezo data
                    ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[2,:],'-m',alpha=0.4);                #Piezo data
                    ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[3,:],'-m',alpha=0.4);          #Piezo data
                    ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.4,label='Mean:' + str(wndw));  #MLE data
                    ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.4);  #MLE data
                    ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.4)   #MLE data
                    ax3.plot((T_Raw-T_Raw[0])*1e3,zCam_Raw-zCam_Raw[0]+TrajOffset[3],'-b')
                    if MovieStepSize>0: ax3.axvline(Vt[FrameIndex], color='r',linestyle='-')

                    ax3.set_ylabel("[um]"); 
                    ax3.grid(True);
                    ax3.legend()
        
                    #ax6
                    if np.sum(np.isnan(VPTMxyz))==0:
                        ax6.hist(np.sqrt(VPTMxyz_Var[:,0]), 20, facecolor='r', alpha=0.4); 
                        ax6.hist(np.sqrt(VPTMxyz_Var[:,1]), 20, facecolor='g', alpha=0.4); 
                        ax6.hist(np.sqrt(VPTMxyz_Var[:,2]), 20, facecolor='b', alpha=0.4)          
                        ax6.set_xlabel('STD [um] over: ' + str(wndw))
        
                    #ax9-ax10 Photon counts
                    ax9.plot(Vt,ni,'-') 
                    ax9.plot(txyz_Traj_Piezo[0,:],PhotonHMMVector,'-r',label='HMM: ' + str(np.round(PhotonHMMParam, decimals=0)) )
                    if MovieStepSize>0: ax9.axvline(Vt[FrameIndex], color='r',linestyle='-')
                    ax10.hist(ni, 20)        
                    ax9.set_ylabel('Ph #')
                    ax9.set_xlabel('ms')
                    ax9.grid()
                    ax9.legend()

                    ax13.plot(PolT,Pol,'-b')
                    #if MovieStepSize>0: ax13.axvline(PolT[FrameIndex], color='r',linestyle='-')
                    ax13.grid(); ax13.set_ylim(-1,1)
                    ax14.hist(Pol, bins=np.linspace(-1,1,40))        

                    #fft ax15               
                    if fft!=0:
                        ax15.semilogy(fx, fft_x_MLE,c='r', alpha=0.4, label='x-MLE')
                        ax15.semilogy(fy, fft_y_MLE,c='g', alpha=0.4, label='y-MLE')
                        ax15.semilogy(fz, fft_z_MLE,c='b', alpha=0.4, label='z-MLE')
                        ax15.set_xlabel('khz')
                        ax15.set_ylim((10**-5,10**-2))
                        #ax15.grid(b=True, which='major', color='b', linestyle='-')
                        #ax15.grid(b=True, which='minor', color='b', linestyle='--')
                    #Cell image
                        #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
                    ax11.imshow(ImTot[:,:,0],cmap='gray',alpha=1.0)
                    #ax[0].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)
                    #ax[0].imshow(ImTot[:,:,2],cmap='Reds',alpha=0.3)
                    if CellMask:
                        ax11.contour(mask,levels=0, linewidths=1, colors='blue', linestyles='solid')
                        ax11.imshow(mask,cmap='Reds',alpha=0.3)

                    if MovieStepSize>0: ls=LineCollection(segments[:FrameIndex], linewidths=1,colors=colors[:FrameIndex], linestyle='solid')
                    else: ls=LineCollection(segments, linewidths=1,colors=colors, linestyle='solid')

                    line=ax11.add_collection(ls)
                    #fig.colorbar(line, ax=ax11)
                    #ax11.plot(x,y,'-', c=color[0,:])

                    xmean=np.mean(x_pix)
                    ymean=np.mean(y_pix)
        
                    ticks=np.arange(0,ImTot[:,:,0].shape[0],4)
                    xticks=np.linspace(xmean-xyImageBox,xmean+xyImageBox,10)
                    yticks=np.linspace(ymean-xyImageBox,ymean+xyImageBox,10)
                    xtick_labels=np.round( (xticks-(xmean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]
                    ytick_labels=np.round( (yticks-(ymean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]

                    ax11.set_xticks(xticks); ax11.set_yticks(yticks)
                    ax11.set_xticklabels(xtick_labels); ax11.set_yticklabels(ytick_labels)
                    ax11.set_xlim([xmean-xyImageBox,xmean+xyImageBox])
                    ax11.set_ylim([ymean-xyImageBox,ymean+xyImageBox])
                    ax11.set_xlabel('um'); ax11.set_ylabel('um')
                

                    #########################################################
                    if do_save_plot:
                        if MovieStepSize>0:
                            file="MovieFrame_" + str(TrajToGet) + '_D_traj_' + str(FrameIndex)+ fileType
                            path=ScanProject.Files['AnalysisConfig'][:-18] +'Movie_Traj_D_' + FolderEnd +'/'
                            Path(path).mkdir(parents=True, exist_ok=True)
                        else:
                            file="T_" + str(TrajToGet) + '_D_traj' + fileType
                            path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_D_' + FolderEnd +'/'
                            Path(path).mkdir(parents=True, exist_ok=True)
                        #fig.savefig(path + file)
                        fig.savefig(path + file, bbox_inches='tight')
                        plt.close()
                        print('Done!')
                    else:
                        plt.show()

                    FrameIndex+=MovieStepSize
                    if MovieStepSize<=0: break
                    elif FrameIndex>=len(Vt):break
            except:
                print('Something wrong in the plotting, skipping this one!')


    def plot_Hist_Traj_MDS11(ScanProject,TrajToGet,TrajDictionary,HMM_prediction=None,HMM_Model=None,groundtruth=None,MovieStepSize=0,do_save_plot=0,FolderEnd='',fft=0, FileType='.png'):
            """
            This function is to plot 
                trajectores with groundtruth
                Cell image
                HMM predicted traj
                zCam z data

                No info box for the HMM 
                Added traj offset to see each trajectroy better 
                Rearanging the plot positions 
            """
            TrajOffset=[0,0,1.0,2.0] #txyz
            xyImageBox=32 #pix 35
            CellMask=1
        
            if CellMask:
                mask=ScanProject.CamCellMask[:,:,0]>0
            CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]
            try:
                ImTot=ScanProject.ImgOverlay
            except:
                print('No cell image found, using cell mask')
                ImTot=np.repeat(np.expand_dims(mask*1.0,axis=2),3,axis=2)

            wndw=10
            txyznM=TrajDictionary[TrajToGet]*1.0

                #Groundtruth data 
            if np.sum(groundtruth==None):
                xyzgroundtruth=np.zeros((3,txyznM.shape[1]))
            else:
                if len(groundtruth.shape)==3:
                    xyzgroundtruthAll=groundtruth[:-1,:,:3]               
                    xyzgroundtruth=np.mean(xyzgroundtruthAll,axis=1)#Mean over blured out pixle
                else:
                    xyzgroundtruthAll=groundtruth[:,:-1].T               
                    xyzgroundtruth=xyzgroundtruthAll


                
                

                #CVE D and sigma evaluation
            txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
            txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

            txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu^2]
            txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]
            txyzCVE_Sigma=np.abs(txyzCVE_Sigma) #Force all values to be positive


                #time
            Vt=txyznM[0,:]              #[s]

                #Photon Counts
            ni=txyznM[4,:] #Raw
            DT=np.mean(np.diff(Vt))
            MLE_tail=ScanProject.TrajectorySelectionRules['MLETailFactor']
            window_size=20
            Ni=np.convolve((1-MLE_tail)**np.arange(window_size),ni)[0:-window_size+1]


                #Piezo stage position
            TStartEnd=[Vt[0],Vt[-1]] 
            StartEndPos=np.where(np.abs(np.diff((ScanProject.ScanMatrix[0,:]>TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<TStartEnd[1]))))[0]
            StartEndPos=np.expand_dims(np.append(np.array([0]),StartEndPos),axis=1)
            
            T_Raw,XP_Raw,YP_Raw,XE_Raw,YE_Raw,XO_Raw,YO_Raw,ZI_Raw,APD1_Raw,APD2_Raw,J_Raw,MemIndex_Raw,ZI_um_Raw,ZI_OffsetIndex_Raw, zCam_Raw =  _MS.get_traj_raw_data4(StartEndPos, 0, ScanProject.ScanMatrix, ScanProject)
            txyz_Traj_Piezo=np.vstack((T_Raw,XP_Raw,YP_Raw,ZI_um_Raw))
            txyz_Traj_Piezo=_MS.get_Scan_toImagCoordinates(ScanProject, txyz_Traj_Piezo)
            
            #ZeroPoint=np.array([XP_Raw[0],YP_Raw[0],ZI_um_Raw[0]])
                #MLE
            VPTMxyz_O=np.transpose(txyznM[1:4,:])
            
                #Start at zero time and at zero piezo pos
            VPTMxyz=VPTMxyz_O-txyz_Traj_Piezo[1:,0]+TrajOffset[1:] #VPTMxyz_O[0,:]
            xyzgroundtruth=xyzgroundtruth-txyz_Traj_Piezo[1:,0]+TrajOffset[1:]
            xyzgroundtruthAll=xyzgroundtruthAll-txyz_Traj_Piezo[1:,0]+TrajOffset[1:]
            Vt=(Vt-Vt[0])*1e3 #[ms]

            txyz_Traj_Piezo=(txyz_Traj_Piezo.T-txyz_Traj_Piezo[:,0]+TrajOffset).T
            txyz_Traj_Piezo[0,:]=txyz_Traj_Piezo[0,:]*1e3

                # Fourier spectra of the MLE signals
            if fft!=0:
                if 0:
                    wndw_fft=20
                    wndwS_fft=wndw_fft #if 1 it is a moving average, if equal to wndw it is a fix window average
                    T_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,Vt)
                    T_fft=T_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]

                    x_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,0])
                    x_fft=x_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    y_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,1])
                    y_fft=y_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    z_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,2])
                    z_fft=z_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    
                    fft_x_MLE, fx = _MS.fft_processing(x_fft,T_fft)
                    fft_y_MLE, fy = _MS.fft_processing(y_fft,T_fft)
                    fft_z_MLE, fz = _MS.fft_processing(z_fft,T_fft)
                else:
                    fft_x_MLE, fx = _MS.fft_processing(VPTMxyz[:,0],Vt)
                    fft_y_MLE, fy = _MS.fft_processing(VPTMxyz[:,1],Vt)
                    fft_z_MLE, fz = _MS.fft_processing(VPTMxyz[:,2],Vt)

                #Running mean on the trajectory
            wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            T_Mean= np.convolve(np.ones(wndw)/wndw,Vt)
            T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
            VPTMxyz_Mean=np.zeros((len(T_Mean),3))
            VPTMxyz_Var=np.zeros((len(T_Mean),3))
            for i in range(3):
                VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,VPTMxyz[:,i])
                VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

                VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(VPTMxyz[:,i])**2) #Var(X)=E[X^2]-E[X]^2
                VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2

            #np.convolve(np.ones(wndw)/wndw,np.mean(txyzCVE_D[:,1:], axis=1))[wndw-1:-wndw+1:wndwS]

                #Create a D value colored traj plot
            x_pix=(VPTMxyz_O[:,0])/CamRes
            y_pix=(VPTMxyz_O[:,1])/CamRes
            z_pix=(VPTMxyz_O[:,1])/CamRes
            #points=np.array([x_pix, y_pix, z_pix]).T.reshape(-1, 1, 3) 
            points=np.array([x_pix, y_pix]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
        
            D_norm=np.log(np.mean(txyzCVE_D[:,1:], axis=1)) #(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))/ np.max(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))
            D_norm[np.where(np.isnan(D_norm))]=0
            D_norm=1/(1+np.exp(-(D_norm)))
            t_D=np.concatenate((txyzCVE_D[:,0],[Vt[-1]]))
            D_traj=np.zeros(Vt.shape)
            j=0
            for i in range(1,len(t_D)):
                while Vt[j]<=t_D[i]:
                    D_traj[j]=D_norm[i-1]
                    j+=1
                    if j>=len(Vt):
                        break

            colors = plt.cm.cool(D_traj)

                #Polarization 
            wndwN=int(ScanProject.TrajectorySelectionRules['wndw'])
            OfInterest=np.where((ScanProject.ScanMatrix[0,:]>=TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<=TStartEnd[1]))[0]
            OfInterest=np.append(OfInterest,OfInterest[-1]+np.arange(wndwN)+1)
            N12=ScanProject.ScanMatrix[3:5,OfInterest]
            PolT=ScanProject.ScanMatrix[0,OfInterest]

            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1=np.convolve(np.ones(wndwN)/wndwN,N12[0,:])
            N1=N1[wndwN-1:-wndwN+1:wndwS]
            N2=np.convolve(np.ones(wndwN)/wndwN,N12[1,:])
            N2=N2[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]

            
            wndwN=5
            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1_mean=np.convolve(np.ones(wndwN)/wndwN,N1)
            N1_mean=N1_mean[wndwN-1:-wndwN+1:wndwS]
            N2_mean=np.convolve(np.ones(wndwN)/wndwN,N2)
            N2_mean=N2_mean[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]
            PolT=(PolT-PolT[0])*1e3 #ms
            Pol=(N1_mean-N2_mean)/(N1_mean+N2_mean)

        ####################################
            # Generate the inverse-Gamma distributions to plot
            if HMM_Model!=None:
                dist_params=HMM_Model['Distributions']
                Occupancy_estimated=HMM_Model['Occupancy']
                DBias=HMM_Model['DBias']
                DFittParam=np.linspace(-1,10,num=200)
                invGamma={}
                invGammaAll=np.zeros(DFittParam.shape)
                for state,Occ in zip(dist_params,Occupancy_estimated):
                    param=dist_params[state]
                    invGamma[state]=stats.invgamma.pdf(DFittParam+DBias, a=param[0],scale=param[1])
                    invGammaAll=invGammaAll + Occ*invGamma[state]
                NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
                invGammaAll=invGammaAll/NormFactor
            else:
                DFittParam=np.linspace(0,10,num=200)
                dist_params=0
                Occupancy_estimated=0
                DBias=0

        #####################################################################
            if MovieStepSize>0: FrameIndex=0
            else: FrameIndex=-1

            movieFlag=1
            while FrameIndex < len(Vt) and movieFlag==1:
        
                # Figure construction
                fig = plt.figure(constrained_layout=False)
                fig.set_size_inches(3*5, 2*5, forward=True)
                gs = fig.add_gridspec(nrows=7, ncols=6, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.2, hspace=1.0)

                ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
                ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z

                ax0=fig.add_subplot(gs[2:4, 0:2]) #MSD, CVE 
                ax4=fig.add_subplot(gs[2:4, 2])  #hist D      
                if fft==0:
                    ax1=fig.add_subplot(gs[4:6, 0:2]) #Sigma               
                    ax5=fig.add_subplot(gs[4:6, 2])  #hist sigma2     
                else:
                    ax1=fig.add_subplot(gs[4:5, 0:2]) #Sigma
                    ax5=fig.add_subplot(gs[4:5, 2])  #hist sigma2     
                    ax15=fig.add_subplot(gs[5:7, 3:5])  #fft     

                ax9=fig.add_subplot(gs[5:7, 0:2]) #Photon count 
                ax10=fig.add_subplot(gs[5:7, 2])  #Photon hist

                ax11=fig.add_subplot(gs[0:4, 3:6]) #Image 
                
                ax13=fig.add_subplot(gs[4, 3:5]) #Pol count 
                ax14=fig.add_subplot(gs[4, 5])  #Pol hist

        
                #ax0 D CVE method
                #ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:4], axis=1),'-k',alpha=0.5, label='CVE xyz');        #D CVE MLE data
                ax0.plot(txyzCVE_D[:,0],np.mean(txyzCVE_D[:,1:3], axis=1),'-k',alpha=0.5, label='CVE xy');        #D CVE MLE data
                #ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,1],'-r',alpha=0.5, label='CVE x');        #D CVE MLE data
                #ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,2],'-g',alpha=0.5, label='CVE y');        #D CVE MLE data
                ax0.plot(txyzCVE_D[:,0],txyzCVE_D[:,3],'-b',alpha=0.5, label='CVE z');        #D CVE MLE data
                #ax0.plot(txyzCVE_D[:,0],HMM_prediction,'-r',alpha=0.5, label='Resedual:' + str(np.round(np.sum((np.mean(txyzCVE_D[:,1:], axis=1)-HMM_prediction)**2),decimals=2)));        #D CVE MLE data
                if MovieStepSize>0: ax0.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax0.set_ylabel("D [um2/s]"); ax0.grid(True); 
                ax0.legend()
                
                #ax4 Histogram D CVE and MSD
                ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), bins=DFittParam, facecolor='g', alpha=0.5 ,density=True); 
                #ax4.axvline(np.mean(txyzCVE_D[:,1:]), color='g',linestyle='-',label='CVE:' + str(np.round(np.mean(txyzCVE_D[:,1:]),decimals=2)))
                #for state,Occ in zip(invGamma,Occupancy_estimated):
                #    ax4.plot(DFittParam,Occ*invGamma[state]/NormFactor,label=state)
                #ax4.plot(DFittParam,invGammaAll,'-k',label='envelope')
                ax4.set_xlabel("D [um^2/s]")
                ax4.legend()


                #ax1  
                ax1.plot(txyzCVE_Sigma[:,0],np.sqrt(np.mean(txyzCVE_Sigma[:,1:],axis=1)),'-g',alpha=0.5, label='CVE'); #Sigma2 CVE MLE data
                ax1.plot(T_Mean,np.sqrt(np.mean(VPTMxyz_Var,axis=1)),'-b',alpha=0.5, label='Var' + str(wndw));         #Var CVE MLE data
                ax1.plot(Vt,np.sqrt(np.sum((VPTMxyz-xyzgroundtruth)**2,axis=1)),'-r',alpha=0.5, label='VarTrueMLE');         #Var CVE MLE data

                if MovieStepSize>0: ax1.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax1.set_ylabel("sigma [um]"); ax1.grid(True);
                ax1.legend()


                #ax5 Histogram Sigma Variance 
                ax5.hist(np.sqrt(np.mean(txyzCVE_Sigma[:,1:],axis=1))*1e3, 20, facecolor='g', alpha=0.5); 
                ax5.hist(np.sqrt(np.mean(VPTMxyz_Var,axis=1))*1e3, 20, facecolor='b', alpha=0.5); 
                ax5.hist(np.sqrt(np.sum((VPTMxyz-xyzgroundtruth)**2,axis=1))*1e3, 20, facecolor='r', alpha=0.5); 

                #ax5.axvline(np.sqrt(np.mean(VPTMxyz_Var))*1e3, color='b',linestyle='-',label='Var' + str(wndw) +':'+ str(np.round(np.sqrt(np.mean(VPTMxyz_Var))*1e3,decimals=2)))
                #ax5.axvline(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3, color='g',linestyle='-',label='CVE:' + str(np.round(np.sqrt(np.mean(txyzCVE_Sigma[:,1:]))*1e3,decimals=2)))
                ax5.set_xlabel("Sigma [nm]")
                ax5.legend()

                #ax3
                ax3.plot(Vt,xyzgroundtruth[:,0],'-c',alpha=0.4,label='xTrue'); #Groudntruth 
                ax3.plot(Vt,xyzgroundtruth[:,1],'-m',alpha=0.4,label='yTrue'); #Groudntruth 
                ax3.plot(Vt,xyzgroundtruth[:,2],'-y',alpha=0.4,label='zTrue'); #Groudntruth 
                #ax3.plot(Vt,xyzgroundtruthAll[:,:,0],'r.');
                #ax3.plot(Vt,xyzgroundtruthAll[:,:,1],'r.');
                #ax3.plot(Vt,xyzgroundtruthAll[:,:,2],'r.');
                ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[1,:],'-m',alpha=0.4,label='Piezo');  #Piezo data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[2,:],'-m',alpha=0.4);                #Piezo data
                ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[3,:],'-m',alpha=0.4);          #Piezo data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.4,label='Mean:' + str(wndw));  #MLE data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.4);  #MLE data
                ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.4)   #MLE data
                ax3.plot((T_Raw-T_Raw[0])*1e3,zCam_Raw-zCam_Raw[0]+TrajOffset[3],'-b')
                if MovieStepSize>0: ax3.axvline(Vt[FrameIndex], color='r',linestyle='-')

                ax3.set_ylabel("[um]"); 
                ax3.grid(True);
                ax3.legend()
        
                #ax6
                #Groundtruth data 
                if np.sum(groundtruth==None):
                    if np.sum(np.isnan(VPTMxyz))==0:
                        ax6.hist(np.sqrt(VPTMxyz_Var[:,0]), 20, facecolor='r', alpha=0.4); 
                        ax6.hist(np.sqrt(VPTMxyz_Var[:,1]), 20, facecolor='g', alpha=0.4); 
                        ax6.hist(np.sqrt(VPTMxyz_Var[:,2]), 20, facecolor='b', alpha=0.4)          
                        ax6.set_xlabel('STD [um] over: ' + str(wndw))                
                else:
                    ax6.hist(VPTMxyz[:,0]-xyzgroundtruth[:,0], 20, facecolor='r', alpha=0.4, label='xstd:' + str(np.round(np.std(VPTMxyz[:,0]-xyzgroundtruth[:,0]), decimals=3))); 
                    ax6.hist(VPTMxyz[:,1]-xyzgroundtruth[:,1], 20, facecolor='g', alpha=0.4,label='ystd:' + str(np.round(np.std(VPTMxyz[:,1]-xyzgroundtruth[:,1]), decimals=3))); 
                    ax6.hist(VPTMxyz[:,2]-xyzgroundtruth[:,2], 20, facecolor='b', alpha=0.4,label='zstd:' + str(np.round(np.std(VPTMxyz[:,2]-xyzgroundtruth[:,2]), decimals=3)))          
                    ax6.set_xlabel('MLE-Ture [um]')  
                    ax6.legend()

        
                #ax9-ax10 Photon counts
                ax9.plot(Vt,Ni,'-') 
                if MovieStepSize>0: ax9.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax10.hist(Ni, 20)        
                ax9.set_ylabel('Ph #')
                ax9.set_xlabel('ms')
                ax9.grid()

                ax13.plot(PolT,Pol,'-b')
                #if MovieStepSize>0: ax13.axvline(PolT[FrameIndex], color='r',linestyle='-')
                ax13.grid(); ax13.set_ylim(-1,1)
                ax14.hist(Pol, bins=np.linspace(-1,1,40))        

                #fft ax15               
                if fft!=0:
                    ax15.semilogy(fx, fft_x_MLE,c='r', alpha=0.4, label='x-MLE')
                    ax15.semilogy(fy, fft_y_MLE,c='g', alpha=0.4, label='y-MLE')
                    ax15.semilogy(fz, fft_z_MLE,c='b', alpha=0.4, label='z-MLE')
                    ax15.set_xlabel('khz')
                    ax15.set_ylim((10**-5,10**-2))
                    #ax15.grid(b=True, which='major', color='b', linestyle='-')
                    #ax15.grid(b=True, which='minor', color='b', linestyle='--')
                #Cell image
                    #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
                ax11.imshow(ImTot[:,:,0],cmap='gray',alpha=1.0)
                #ax[0].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)
                #ax[0].imshow(ImTot[:,:,2],cmap='Reds',alpha=0.3)
                if CellMask:
                    ax11.contour(mask,levels=0, linewidths=1, colors='blue', linestyles='solid')
                    ax11.imshow(mask,cmap='Reds',alpha=0.3)

                if MovieStepSize>0: ls=LineCollection(segments[:FrameIndex], linewidths=1,colors=colors[:FrameIndex], linestyle='solid')
                else: ls=LineCollection(segments, linewidths=1,colors=colors, linestyle='solid')

                line=ax11.add_collection(ls)
                #fig.colorbar(line, ax=ax11)
                #ax11.plot(x,y,'-', c=color[0,:])

                xmean=np.mean(x_pix)
                ymean=np.mean(y_pix)
        
                ticks=np.arange(0,ImTot[:,:,0].shape[0],4)
                xticks=np.linspace(xmean-xyImageBox,xmean+xyImageBox,10)
                yticks=np.linspace(ymean-xyImageBox,ymean+xyImageBox,10)
                xtick_labels=np.round( (xticks-(xmean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]
                ytick_labels=np.round( (yticks-(ymean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]

                ax11.set_xticks(xticks); ax11.set_yticks(yticks)
                ax11.set_xticklabels(xtick_labels); ax11.set_yticklabels(ytick_labels)
                ax11.set_xlim([xmean-xyImageBox,xmean+xyImageBox])
                ax11.set_ylim([ymean-xyImageBox,ymean+xyImageBox])
                ax11.set_xlabel('um'); ax11.set_ylabel('um')
                

                #########################################################
                if do_save_plot:
                    if MovieStepSize>0:
                        file="MovieFrame_" + str(TrajToGet) + '_D_traj_' + str(FrameIndex)+ FileType
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Movie_Traj_D_filtered_' + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    else:
                        file="T_" + str(TrajToGet) + '_D_traj' + FileType
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_D_filtered_' + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    fig.savefig(path + file, bbox_inches='tight')
                    plt.close()
                    print('Done!')
                else:
                    plt.show()

                FrameIndex+=MovieStepSize
                if MovieStepSize<=0: break
                elif FrameIndex>=len(Vt):break

    def plot_ArticleFig_1(ScanProject,TrajToGet,TrajDictionary,resVsPrate_Dic, wndw=10,HMM_prediction=None,HMM_Model=None,groundtruth=None,MovieStepSize=0,do_save_plot=0,FolderEnd='',fft=0,FileType='.png'):
            """
            This function is to plot 
                trajectores with groundtruth
                Cell image
                HMM predicted traj
                zCam z data

                No info box for the HMM 
                Added traj offset to see each trajectroy better 
                Rearanging the plot positions 
            """
            TrajOffset=[0,0,1.0,2.0] #txyz
            xyImageBox=32 #pix 35
            CellMask=1
            #wndw=10        

            if CellMask:
                mask=ScanProject.CamCellMask[:,:,0]>0
            CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]
            try:
                ImTot=ScanProject.ImgOverlay
            except:
                print('No cell image found, using cell mask')
                ImTot=np.repeat(np.expand_dims(mask*1.0,axis=2),3,axis=2)

            txyznM=TrajDictionary[TrajToGet]*1.0

                #Groundtruth data 
            if np.sum(groundtruth==None):
                xyzgroundtruth=np.zeros((3,txyznM.shape[1]))
            else:
                if len(groundtruth.shape)==3:
                    xyzgroundtruthAll=groundtruth[:-1,:,:3]               
                    xyzgroundtruth=np.mean(xyzgroundtruthAll,axis=1)#Mean over blured out pixle
                else:
                    xyzgroundtruthAll=groundtruth[:,:-1].T               
                    xyzgroundtruth=xyzgroundtruthAll
 

                #CVE D and sigma evaluation
            txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
            txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

            txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu^2]
            txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]
            txyzCVE_Sigma=np.abs(txyzCVE_Sigma) #Force all values to be positive

                #time
            Vt=txyznM[0,:]              #[s]

                #Photon Counts
            ni=txyznM[4,:] #Raw
            DT=np.mean(np.diff(Vt))
            MLE_tail=ScanProject.TrajectorySelectionRules['MLETailFactor']
            window_size=20
            Ni=np.convolve((1-MLE_tail)**np.arange(window_size),ni)[0:-window_size+1]


                #Piezo stage position
            TStartEnd=[Vt[0],Vt[-1]] 
            StartEndPos=np.where(np.abs(np.diff((ScanProject.ScanMatrix[0,:]>TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<TStartEnd[1]))))[0]
            StartEndPos=np.expand_dims(np.append(np.array([0]),StartEndPos),axis=1)
            
            T_Raw,XP_Raw,YP_Raw,XE_Raw,YE_Raw,XO_Raw,YO_Raw,ZI_Raw,APD1_Raw,APD2_Raw,J_Raw,MemIndex_Raw,ZI_um_Raw,ZI_OffsetIndex_Raw, zCam_Raw =  _MS.get_traj_raw_data4(StartEndPos, 0, ScanProject.ScanMatrix, ScanProject)
            txyz_Traj_Piezo=np.vstack((T_Raw,XP_Raw,YP_Raw,ZI_um_Raw))
            txyz_Traj_Piezo=_MS.get_Scan_toImagCoordinates(ScanProject, txyz_Traj_Piezo)
            
            #ZeroPoint=np.array([XP_Raw[0],YP_Raw[0],ZI_um_Raw[0]])
                #MLE
            VPTMxyz_O=np.transpose(txyznM[1:4,:])
            
                #Start at zero time and at zero piezo pos
            VPTMxyz=VPTMxyz_O-txyz_Traj_Piezo[1:,0]+TrajOffset[1:] #VPTMxyz_O[0,:]
            if np.sum(groundtruth==None)==0:
                xyzgroundtruth=xyzgroundtruth-txyz_Traj_Piezo[1:,0]+TrajOffset[1:]
                xyzgroundtruthAll=xyzgroundtruthAll-txyz_Traj_Piezo[1:,0]+TrajOffset[1:]
            Vt=(Vt-Vt[0])*1e3 #[ms]

            txyz_Traj_Piezo=(txyz_Traj_Piezo.T-txyz_Traj_Piezo[:,0]+TrajOffset).T
            txyz_Traj_Piezo[0,:]=txyz_Traj_Piezo[0,:]*1e3

                # Fourier spectra of the MLE signals
            if fft!=0:
                if 0:
                    wndw_fft=20
                    wndwS_fft=wndw_fft #if 1 it is a moving average, if equal to wndw it is a fix window average
                    T_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,Vt)
                    T_fft=T_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]

                    x_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,0])
                    x_fft=x_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    y_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,1])
                    y_fft=y_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    z_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,2])
                    z_fft=z_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    
                    fft_x_MLE, fx = _MS.fft_processing(x_fft,T_fft)
                    fft_y_MLE, fy = _MS.fft_processing(y_fft,T_fft)
                    fft_z_MLE, fz = _MS.fft_processing(z_fft,T_fft)
                else:
                    fft_x_MLE, fx = _MS.fft_processing(VPTMxyz[:,0],Vt)
                    fft_y_MLE, fy = _MS.fft_processing(VPTMxyz[:,1],Vt)
                    fft_z_MLE, fz = _MS.fft_processing(VPTMxyz[:,2],Vt)

                #Running mean on the trajectory
            wndwS=1 #wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            T_Mean= np.convolve(np.ones(wndw)/wndw,np.concatenate(( np.ones(wndw-1)*Vt[0],Vt)))
            T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
            VPTMxyz_Mean=np.zeros((len(T_Mean),3))
            VPTMxyz_Var=np.zeros((len(T_Mean),3))
            for i in range(3):
                VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,np.concatenate(( np.ones(wndw-1)*VPTMxyz[0,i],VPTMxyz[:,i])))
                VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

                VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(np.concatenate(( np.ones(wndw-1)*VPTMxyz[0,i],VPTMxyz[:,i])))**2) #Var(X)=E[X^2]-E[X]^2
                VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2
            
            #np.convolve(np.ones(wndw)/wndw,np.mean(txyzCVE_D[:,1:], axis=1))[wndw-1:-wndw+1:wndwS]

                #Create a D value colored traj plot
            x_pix=(VPTMxyz_O[:,0])/CamRes
            y_pix=(VPTMxyz_O[:,1])/CamRes
            z_pix=(VPTMxyz_O[:,1])/CamRes
            #points=np.array([x_pix, y_pix, z_pix]).T.reshape(-1, 1, 3) 
            points=np.array([x_pix, y_pix]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
        
            D_norm=np.log(np.mean(txyzCVE_D[:,1:], axis=1)) #(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))/ np.max(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))
            D_norm[np.where(np.isnan(D_norm))]=0
            D_norm=1/(1+np.exp(-(D_norm)))
            t_D=np.concatenate((txyzCVE_D[:,0],[Vt[-1]]))
            D_traj=np.zeros(Vt.shape)
            j=0
            for i in range(1,len(t_D)):
                while Vt[j]<=t_D[i]:
                    D_traj[j]=D_norm[i-1]
                    j+=1
                    if j>=len(Vt):
                        break

            colors = plt.cm.cool(D_traj)

                #Polarization 
            wndwN=int(ScanProject.TrajectorySelectionRules['wndw'])
            OfInterest=np.where((ScanProject.ScanMatrix[0,:]>=TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<=TStartEnd[1]))[0]
            OfInterest=np.append(OfInterest,OfInterest[-1]+np.arange(wndwN)+1)
            N12=ScanProject.ScanMatrix[3:5,OfInterest]
            PolT=ScanProject.ScanMatrix[0,OfInterest]

            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1=np.convolve(np.ones(wndwN)/wndwN,N12[0,:])
            N1=N1[wndwN-1:-wndwN+1:wndwS]
            N2=np.convolve(np.ones(wndwN)/wndwN,N12[1,:])
            N2=N2[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]

            
            wndwN=5
            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1_mean=np.convolve(np.ones(wndwN)/wndwN,N1)
            N1_mean=N1_mean[wndwN-1:-wndwN+1:wndwS]
            N2_mean=np.convolve(np.ones(wndwN)/wndwN,N2)
            N2_mean=N2_mean[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]
            PolT=(PolT-PolT[0])*1e3 #ms
            Pol=(N1_mean-N2_mean)/(N1_mean+N2_mean)

        ####################################
            # Generate the inverse-Gamma distributions to plot
            if HMM_Model!=None:
                dist_params=HMM_Model['Distributions']
                Occupancy_estimated=HMM_Model['Occupancy']
                DBias=HMM_Model['DBias']
                DFittParam=np.linspace(-DBias,10,num=200)
                invGamma={}
                invGammaAll=np.zeros(DFittParam.shape)
                for state,Occ in zip(dist_params,Occupancy_estimated):
                    param=dist_params[state]
                    invGamma[state]=stats.invgamma.pdf(DFittParam+DBias, a=param[0],scale=param[1])
                    invGammaAll=invGammaAll + Occ*invGamma[state]
                NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
                invGammaAll=invGammaAll/NormFactor
            else:
                DFittParam=np.linspace(0,10,num=200)
                dist_params=0
                Occupancy_estimated=0
                DBias=0

        #####################################################################
            if MovieStepSize>0: FrameIndex=0
            else: FrameIndex=-1

            movieFlag=1
            while FrameIndex < len(Vt) and movieFlag==1:
        
                # Figure construction
                fig = plt.figure(constrained_layout=False)
                figsize_cm=np.array([9.8,8.3])*3
                fig.set_size_inches(figsize_cm[0]/2.54, figsize_cm[1]/2.54, forward=True)
                gs = fig.add_gridspec(nrows=5, ncols=3, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.15, hspace=0.4)

                ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
                ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z   

                ax9=fig.add_subplot(gs[2, 0:2]) #Photon count 
                ax10=fig.add_subplot(gs[2, 2])  #Photon hist

                ax4=fig.add_subplot(gs[3:5, 0])  #hist D   

                ax11=fig.add_subplot(gs[3:5, 1],projection='3d') #Image 
                ax12=fig.add_subplot(gs[3:5, 2]) #res vs photon rate 
                

                        
                #ax4 Histogram D CVE and MSD
                #ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), bins=DFittParam, facecolor='g', alpha=0.5 ,density=True); 
                ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), bins=DFittParam, facecolor='g', density=True); 
                ax4.axvline(np.mean(txyzCVE_D[:,1:]), color='k',linestyle='-',label='mean:' + str(np.round(np.mean(txyzCVE_D[:,1:]),decimals=2)))
                #for state,Occ in zip(invGamma,Occupancy_estimated):
                #    ax4.plot(DFittParam,Occ*invGamma[state]/NormFactor,label=state)
                #ax4.plot(DFittParam,invGammaAll,'-k',label='envelope')
                ax4.set_yticklabels([])
                ax4.set_xlabel("[um^2/s]")
                ax4.legend()


                #ax3
                if np.sum(groundtruth==None):
                    ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.7,label='xMean:' + str(wndw));  #MLE data
                    ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.7);  #MLE data
                    ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.7)   #MLE data
                else:
                    ax3.plot(Vt,xyzgroundtruth[:,0],'-k',alpha=0.7,label='xTrue'); #Groudntruth 
                    ax3.plot(Vt,xyzgroundtruth[:,1],'-k',alpha=0.7,label='yTrue'); #Groudntruth 
                    ax3.plot(Vt,xyzgroundtruth[:,2],'-k',alpha=0.7,label='zTrue'); #Groudntruth 
                #ax3.plot(Vt,xyzgroundtruthAll[:,:,0],'r.');
                #ax3.plot(Vt,xyzgroundtruthAll[:,:,1],'r.');
                #ax3.plot(Vt,xyzgroundtruthAll[:,:,2],'r.');
                ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
                #ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[1,:],'-m',alpha=0.4,label='Piezo');  #Piezo data
                #ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[2,:],'-m',alpha=0.4);                #Piezo data
                #ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[3,:],'-m',alpha=0.4);          #Piezo data
                #ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.4,label='Mean:' + str(wndw));  #MLE data
                #ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.4);  #MLE data
                #ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.4)   #MLE data
                #ax3.plot((T_Raw-T_Raw[0])*1e3,zCam_Raw-zCam_Raw[0]+TrajOffset[3],'-b')
                if MovieStepSize>0: ax3.axvline(Vt[FrameIndex], color='r',linestyle='-')

                ax3.set_ylabel("[um]"); 
                ax3.grid(True);
                #ax3.legend()
        
                #ax6
                #Groundtruth data 
                if np.sum(groundtruth==None):
                    if np.sum(np.isnan(VPTMxyz))==0:
                        ax6.hist(VPTMxyz[:,2]-VPTMxyz_Mean[:,2], 20, facecolor='b', alpha=0.4,label='zstd:' + str(np.round(np.std(VPTMxyz[:,2]-VPTMxyz_Mean[:,2]), decimals=3)))          
                        ax6.hist(VPTMxyz[:,0]-VPTMxyz_Mean[:,0], 20, facecolor='r', alpha=0.4, label='xstd:' + str(np.round(np.std(VPTMxyz[:,0]-VPTMxyz_Mean[:,0]), decimals=3))); 
                        ax6.hist(VPTMxyz[:,1]-VPTMxyz_Mean[:,1], 20, facecolor='g', alpha=0.4,label='ystd:' + str(np.round(np.std(VPTMxyz[:,1]-VPTMxyz_Mean[:,1]), decimals=3))); 
                        ax6.set_xlabel('MLE-mean [um]')  

                        #ax6.hist(np.sqrt(VPTMxyz_Var[:,0]), 20, facecolor='r', alpha=0.4); 
                        #ax6.hist(np.sqrt(VPTMxyz_Var[:,1]), 20, facecolor='g', alpha=0.4); 
                        #ax6.hist(np.sqrt(VPTMxyz_Var[:,2]), 20, facecolor='b', alpha=0.4)          
                        #ax6.set_xlabel('STD [um] over: ' + str(wndw))                
                else:
                    ax6.hist(VPTMxyz[:,0]-xyzgroundtruth[:,0], 20, facecolor='r', alpha=0.4, label='xstd:' + str(np.round(np.std(VPTMxyz[:,0]-xyzgroundtruth[:,0]), decimals=3))); 
                    ax6.hist(VPTMxyz[:,1]-xyzgroundtruth[:,1], 20, facecolor='g', alpha=0.4,label='ystd:' + str(np.round(np.std(VPTMxyz[:,1]-xyzgroundtruth[:,1]), decimals=3))); 
                    ax6.hist(VPTMxyz[:,2]-xyzgroundtruth[:,2], 20, facecolor='b', alpha=0.4,label='zstd:' + str(np.round(np.std(VPTMxyz[:,2]-xyzgroundtruth[:,2]), decimals=3)))          
                    ax6.set_xlabel('MLE-Ture [um]')  
                ax6.legend()
                ax6.set_yticklabels([])

        
                #ax9-ax10 Photon counts
                ax9.plot(Vt,Ni,'-') 
                if MovieStepSize>0: ax9.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax9.set_ylabel('Ph #')
                ax9.set_xlabel('ms')
                ax9.grid()

                ax10.hist(Ni, 20) 
                ax10.set_yticklabels([])
                ax10.set_xlabel('[Ph #]')  

                #3D image
                if 0:
                        #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
                    if MovieStepSize>0: ls=LineCollection(segments[:FrameIndex], linewidths=1,colors=colors[:FrameIndex], linestyle='solid')
                    else: ls=LineCollection(segments, linewidths=1,colors=colors, linestyle='solid')

                    line=ax11.add_collection(ls)
                    #fig.colorbar(line, ax=ax11)
                    #ax11.plot(x,y,'-', c=color[0,:])

                    xmean=np.mean(x_pix)
                    ymean=np.mean(y_pix)
        
                    ticks=np.arange(0,ImTot[:,:,0].shape[0],4)
                    xticks=np.linspace(xmean-xyImageBox,xmean+xyImageBox,10)
                    yticks=np.linspace(ymean-xyImageBox,ymean+xyImageBox,10)
                    xtick_labels=np.round( (xticks-(xmean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]
                    ytick_labels=np.round( (yticks-(ymean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]

                    ax11.set_xticks(xticks); ax11.set_yticks(yticks)
                    ax11.set_xticklabels(xtick_labels); ax11.set_yticklabels(ytick_labels)
                    ax11.set_xlim([xmean-xyImageBox,xmean+xyImageBox])
                    ax11.set_ylim([ymean-xyImageBox,ymean+xyImageBox])
                    #ax11.set_xlabel('um'); ax11.set_ylabel('um')
                    ax11.set_xticklabels([]);ax11.set_yticklabels([])


                ax11.plot3D(VPTMxyz[:,0],VPTMxyz[:,1], VPTMxyz[:,2],'-k',alpha=0.4);
                ax11.plot3D(VPTMxyz[:,0],VPTMxyz[:,1], VPTMxyz[:,2],'ob',alpha=0.4,markersize=1);
                #scatterPlot=ax11.scatter(VPTMxyz[:,0],VPTMxyz[:,1], VPTMxyz[:,2],s=1, c=Sxyz,cmap=cm.jet);   
                ax11.set_xlabel('x [um]');ax11.set_ylabel('y [um]');ax11.set_zlabel('z [um]')
                ax11.view_init(elev=35., azim=-70)
                # Hide axes ticks
                #ax11.set_xticks([-1,1])
                #ax11.set_yticks([-1,1])
                #ax11.set_zticks([-2,0])
                #ax11.xaxis._axinfo['juggled'] = (0,0,0)
                #ax11.yaxis._axinfo['juggled'] = (1,1,1)
                ax11.zaxis._axinfo['juggled'] = (-2,2,2)


                # Res vs photon rate 
                ax12.plot(resVsPrate_Dic[TrajToGet][:,3]/1e3, resVsPrate_Dic[TrajToGet][:,0]*1e3,'-or',alpha=0.4,label='x')
                ax12.plot(resVsPrate_Dic[TrajToGet][:,3]/1e3, resVsPrate_Dic[TrajToGet][:,1]*1e3,'-og',alpha=0.4,label='y')
                ax12.plot(resVsPrate_Dic[TrajToGet][:,3]/1e3, resVsPrate_Dic[TrajToGet][:,2]*1e3,'-ob',alpha=0.4,label='z')
                ax122=ax12.twinx()
                ax122.plot(resVsPrate_Dic[TrajToGet][:,3]/1e3, resVsPrate_Dic[TrajToGet][:,4]*1e3,'-*k',alpha=0.4,label='')
                ax12.set_ylabel("Tracking error [nm]"); 
                ax122.set_ylabel("Time res [ms]"); 
                ax12.set_xlabel("Photon rate [kHz]");
                ax12.grid(True);
                ax12.legend()

                

                #########################################################

                if do_save_plot:
                    if MovieStepSize>0:
                        file="MovieFrame_" + str(TrajToGet) + '_D_traj_' + str(FrameIndex)+ FileType
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Movie_Art_Traj_D_filtered_' + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    else:
                        file="Art_" + str(TrajToGet) + '_D_traj' + FileType
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Art_D_filtered_' + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    fig.savefig(path + file, bbox_inches='tight')
                    plt.close()
                    print('Done!')
                else:
                    plt.show()

                FrameIndex+=MovieStepSize
                if MovieStepSize<=0: break
                elif FrameIndex>=len(Vt):break

    def plot_ArticleFig_1b(ScanProject,TrajToGet,TrajDictionary,resVsPrate_Dic, wndw=10,HMM_prediction=None,HMM_Model=None,groundtruth=None,MovieStepSize=0,do_save_plot=0,FolderEnd='',fft=0,FileType='.png'):
            """
            This function is to plot 
                trajectores with groundtruth
                Cell image
                HMM predicted traj
                zCam z data

                No info box for the HMM 
                Added traj offset to see each trajectroy better 
                Rearanging the plot positions 
            """
            TrajOffset=[0,0,1.0,2.0] #txyz
            xyImageBox=32 #pix 35
            CellMask=1
            #wndw=10        

            if CellMask:
                mask=ScanProject.CamCellMask[:,:,0]>0
            CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]
            try:
                ImTot=ScanProject.ImgOverlay
            except:
                print('No cell image found, using cell mask')
                ImTot=np.repeat(np.expand_dims(mask*1.0,axis=2),3,axis=2)

            txyznM=TrajDictionary[TrajToGet]*1.0

                #Groundtruth data 
            if np.sum(groundtruth==None):
                xyzgroundtruth=np.zeros((3,txyznM.shape[1]))
            else:
                if len(groundtruth.shape)==3:
                    xyzgroundtruthAll=groundtruth[:-1,:,:3]               
                    xyzgroundtruth=np.mean(xyzgroundtruthAll,axis=1)#Mean over blured out pixle
                else:
                    xyzgroundtruthAll=groundtruth[:,:-1].T               
                    xyzgroundtruth=xyzgroundtruthAll
 

                #CVE D and sigma evaluation
            txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
            txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

            txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu^2]
            txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]
            txyzCVE_Sigma=np.abs(txyzCVE_Sigma) #Force all values to be positive

                #time
            Vt=txyznM[0,:]              #[s]

                #Photon Counts
            ni=txyznM[4,:] #Raw
            DT=np.mean(np.diff(Vt))
            MLE_tail=ScanProject.TrajectorySelectionRules['MLETailFactor']
            window_size=20
            Ni=np.convolve((1-MLE_tail)**np.arange(window_size),ni)[0:-window_size+1]


                #Piezo stage position
            TStartEnd=[Vt[0],Vt[-1]] 
            StartEndPos=np.where(np.abs(np.diff((ScanProject.ScanMatrix[0,:]>TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<TStartEnd[1]))))[0]
            StartEndPos=np.expand_dims(np.append(np.array([0]),StartEndPos),axis=1)
            
            T_Raw,XP_Raw,YP_Raw,XE_Raw,YE_Raw,XO_Raw,YO_Raw,ZI_Raw,APD1_Raw,APD2_Raw,J_Raw,MemIndex_Raw,ZI_um_Raw,ZI_OffsetIndex_Raw, zCam_Raw =  _MS.get_traj_raw_data4(StartEndPos, 0, ScanProject.ScanMatrix, ScanProject)
            txyz_Traj_Piezo=np.vstack((T_Raw,XP_Raw,YP_Raw,ZI_um_Raw))
            txyz_Traj_Piezo=_MS.get_Scan_toImagCoordinates(ScanProject, txyz_Traj_Piezo)
            
            #ZeroPoint=np.array([XP_Raw[0],YP_Raw[0],ZI_um_Raw[0]])
                #MLE
            VPTMxyz_O=np.transpose(txyznM[1:4,:])
            
                #Start at zero time and at zero piezo pos
            VPTMxyz=VPTMxyz_O-txyz_Traj_Piezo[1:,0]+TrajOffset[1:] #VPTMxyz_O[0,:]
            if np.sum(groundtruth==None)==0:
                xyzgroundtruth=xyzgroundtruth-txyz_Traj_Piezo[1:,0]+TrajOffset[1:]
                xyzgroundtruthAll=xyzgroundtruthAll-txyz_Traj_Piezo[1:,0]+TrajOffset[1:]
            Vt=(Vt-Vt[0])*1e3 #[ms]

            txyz_Traj_Piezo=(txyz_Traj_Piezo.T-txyz_Traj_Piezo[:,0]+TrajOffset).T
            txyz_Traj_Piezo[0,:]=txyz_Traj_Piezo[0,:]*1e3

                # Fourier spectra of the MLE signals
            if fft!=0:
                if 0:
                    wndw_fft=20
                    wndwS_fft=wndw_fft #if 1 it is a moving average, if equal to wndw it is a fix window average
                    T_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,Vt)
                    T_fft=T_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]

                    x_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,0])
                    x_fft=x_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    y_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,1])
                    y_fft=y_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    z_fft= np.convolve(np.ones(wndw_fft)/wndw_fft,VPTMxyz[:,2])
                    z_fft=z_fft[wndw_fft-1:-wndw_fft+1:wndwS_fft]
                    
                    fft_x_MLE, fx = _MS.fft_processing(x_fft,T_fft)
                    fft_y_MLE, fy = _MS.fft_processing(y_fft,T_fft)
                    fft_z_MLE, fz = _MS.fft_processing(z_fft,T_fft)
                else:
                    fft_x_MLE, fx = _MS.fft_processing(VPTMxyz[:,0],Vt)
                    fft_y_MLE, fy = _MS.fft_processing(VPTMxyz[:,1],Vt)
                    fft_z_MLE, fz = _MS.fft_processing(VPTMxyz[:,2],Vt)

                #Running mean on the trajectory
            wndwS=1 #wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            T_Mean= np.convolve(np.ones(wndw)/wndw,np.concatenate(( np.ones(wndw-1)*Vt[0],Vt)))
            T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
            VPTMxyz_Mean=np.zeros((len(T_Mean),3))
            VPTMxyz_Var=np.zeros((len(T_Mean),3))
            for i in range(3):
                VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,np.concatenate(( np.ones(wndw-1)*VPTMxyz[0,i],VPTMxyz[:,i])))
                VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

                VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(np.concatenate(( np.ones(wndw-1)*VPTMxyz[0,i],VPTMxyz[:,i])))**2) #Var(X)=E[X^2]-E[X]^2
                VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2
            
            #np.convolve(np.ones(wndw)/wndw,np.mean(txyzCVE_D[:,1:], axis=1))[wndw-1:-wndw+1:wndwS]

                #Create a D value colored traj plot
            x_pix=(VPTMxyz_O[:,0])/CamRes
            y_pix=(VPTMxyz_O[:,1])/CamRes
            z_pix=(VPTMxyz_O[:,1])/CamRes
            #points=np.array([x_pix, y_pix, z_pix]).T.reshape(-1, 1, 3) 
            points=np.array([x_pix, y_pix]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
        
            D_norm=np.log(np.mean(txyzCVE_D[:,1:], axis=1)) #(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))/ np.max(np.mean(txyzCVE_D[:,1:], axis=1)-np.min(np.mean(txyzCVE_D[:,1:], axis=1)))
            D_norm[np.where(np.isnan(D_norm))]=0
            D_norm=1/(1+np.exp(-(D_norm)))
            t_D=np.concatenate((txyzCVE_D[:,0],[Vt[-1]]))
            D_traj=np.zeros(Vt.shape)
            j=0
            for i in range(1,len(t_D)):
                while Vt[j]<=t_D[i]:
                    D_traj[j]=D_norm[i-1]
                    j+=1
                    if j>=len(Vt):
                        break

            colors = plt.cm.cool(D_traj)

                #Polarization 
            wndwN=int(ScanProject.TrajectorySelectionRules['wndw'])
            OfInterest=np.where((ScanProject.ScanMatrix[0,:]>=TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<=TStartEnd[1]))[0]
            OfInterest=np.append(OfInterest,OfInterest[-1]+np.arange(wndwN)+1)
            N12=ScanProject.ScanMatrix[3:5,OfInterest]
            PolT=ScanProject.ScanMatrix[0,OfInterest]

            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1=np.convolve(np.ones(wndwN)/wndwN,N12[0,:])
            N1=N1[wndwN-1:-wndwN+1:wndwS]
            N2=np.convolve(np.ones(wndwN)/wndwN,N12[1,:])
            N2=N2[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]

            
            wndwN=5
            wndwS=wndwN  #if 1 it is a moving average, if equal to wndw it is a fix window average
            N1_mean=np.convolve(np.ones(wndwN)/wndwN,N1)
            N1_mean=N1_mean[wndwN-1:-wndwN+1:wndwS]
            N2_mean=np.convolve(np.ones(wndwN)/wndwN,N2)
            N2_mean=N2_mean[wndwN-1:-wndwN+1:wndwS]
            PolT=np.convolve(np.ones(wndwN)/wndwN,PolT)
            PolT=PolT[wndwN-1:-wndwN+1:wndwS]
            PolT=(PolT-PolT[0])*1e3 #ms
            Pol=(N1_mean-N2_mean)/(N1_mean+N2_mean)

        ####################################
            # Generate the inverse-Gamma distributions to plot
            if HMM_Model!=None:
                dist_params=HMM_Model['Distributions']
                Occupancy_estimated=HMM_Model['Occupancy']
                DBias=HMM_Model['DBias']
                DFittParam=np.linspace(-DBias,10,num=200)
                invGamma={}
                invGammaAll=np.zeros(DFittParam.shape)
                for state,Occ in zip(dist_params,Occupancy_estimated):
                    param=dist_params[state]
                    invGamma[state]=stats.invgamma.pdf(DFittParam+DBias, a=param[0],scale=param[1])
                    invGammaAll=invGammaAll + Occ*invGamma[state]
                NormFactor=np.sum(invGammaAll*np.diff(DFittParam)[0])
                invGammaAll=invGammaAll/NormFactor
            else:
                DFittParam=np.linspace(0,10,num=200)
                dist_params=0
                Occupancy_estimated=0
                DBias=0

        #####################################################################
            if MovieStepSize>0: FrameIndex=0
            else: FrameIndex=-1

            movieFlag=1
            while FrameIndex < len(Vt) and movieFlag==1:
        
                # Figure construction
                fig = plt.figure(constrained_layout=False)
                figsize_cm=np.array([9.8,8.3])*3
                fig.set_size_inches(figsize_cm[0]/2.54, figsize_cm[1]/2.54, forward=True)
                gs = fig.add_gridspec(nrows=5, ncols=3, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.15, hspace=0.4)

                ax3=fig.add_subplot(gs[0:2, 0:2]) #pos
                ax6=fig.add_subplot(gs[0:2, 2])   #hist x,y,z   

                ax9=fig.add_subplot(gs[2, 0:2]) #Photon count 
                ax10=fig.add_subplot(gs[2, 2])  #Photon hist

                #ax4=fig.add_subplot(gs[3:5, 0])  #hist D   

                ax11=fig.add_subplot(gs[3:5, 2],projection='3d') #Image 
                ax12=fig.add_subplot(gs[3:5, 0:2]) #res vs photon rate 
                

                        
                #ax4 Histogram D CVE and MSD
                #ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), bins=DFittParam, facecolor='g', alpha=0.5 ,density=True); 
                ##ax4.hist(np.mean(txyzCVE_D[:,1:],axis=1), bins=DFittParam, facecolor='g', density=True); 
                ##ax4.axvline(np.mean(txyzCVE_D[:,1:]), color='k',linestyle='-',label='mean:' + str(np.round(np.mean(txyzCVE_D[:,1:]),decimals=2)))
                #for state,Occ in zip(invGamma,Occupancy_estimated):
                #    ax4.plot(DFittParam,Occ*invGamma[state]/NormFactor,label=state)
                #ax4.plot(DFittParam,invGammaAll,'-k',label='envelope')
                ##ax4.set_yticklabels([])
                ##ax4.set_xlabel("[um^2/s]")
                ##ax4.legend()


                #ax3
                if np.sum(groundtruth==None):
                    ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.7,label='xMean:' + str(wndw));  #MLE data
                    ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.7);  #MLE data
                    ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.7)   #MLE data
                else:
                    ax3.plot(Vt,xyzgroundtruth[:,0],'-k',alpha=0.7,label='xTrue'); #Groudntruth 
                    ax3.plot(Vt,xyzgroundtruth[:,1],'-k',alpha=0.7,label='yTrue'); #Groudntruth 
                    ax3.plot(Vt,xyzgroundtruth[:,2],'-k',alpha=0.7,label='zTrue'); #Groudntruth 
                #ax3.plot(Vt,xyzgroundtruthAll[:,:,0],'r.');
                #ax3.plot(Vt,xyzgroundtruthAll[:,:,1],'r.');
                #ax3.plot(Vt,xyzgroundtruthAll[:,:,2],'r.');
                ax3.plot(Vt,VPTMxyz[:,0],'-r',alpha=0.4,label='x');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,1],'-g',alpha=0.4,label='y');  #MLE data
                ax3.plot(Vt,VPTMxyz[:,2],'-b',alpha=0.4,label='z')   #MLE data
                #ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[1,:],'-m',alpha=0.4,label='Piezo');  #Piezo data
                #ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[2,:],'-m',alpha=0.4);                #Piezo data
                #ax3.plot(txyz_Traj_Piezo[0,:],txyz_Traj_Piezo[3,:],'-m',alpha=0.4);          #Piezo data
                #ax3.plot(T_Mean,VPTMxyz_Mean[:,0],'-k',alpha=0.4,label='Mean:' + str(wndw));  #MLE data
                #ax3.plot(T_Mean,VPTMxyz_Mean[:,1],'-k',alpha=0.4);  #MLE data
                #ax3.plot(T_Mean,VPTMxyz_Mean[:,2],'-k',alpha=0.4)   #MLE data
                #ax3.plot((T_Raw-T_Raw[0])*1e3,zCam_Raw-zCam_Raw[0]+TrajOffset[3],'-b')
                if MovieStepSize>0: ax3.axvline(Vt[FrameIndex], color='r',linestyle='-')

                ax3.xaxis.set_major_locator(MultipleLocator(500))
                ax3.yaxis.set_major_locator(MultipleLocator(0.5))
                ax3.set_ylabel("[um]"); 
                ax3.grid(True);
                #ax3.legend()
        
                #ax6
                #Groundtruth data 
                if np.sum(groundtruth==None):
                    if np.sum(np.isnan(VPTMxyz))==0:
                        ax6.hist(VPTMxyz[:,2][:-(int(wndw/2)-1)]-VPTMxyz_Mean[:,2][int(wndw/2)-1:], 20, facecolor='b', alpha=0.4,label='zstd:' + str(np.round(np.std(VPTMxyz[:,2][:- (int(wndw/2)-1)] -VPTMxyz_Mean[:,2][int(wndw/2)-1:]), decimals=3)))          
                        ax6.hist(VPTMxyz[:,0][:-(int(wndw/2)-1)]-VPTMxyz_Mean[:,0][int(wndw/2)-1:], 20, facecolor='r', alpha=0.4, label='xstd:' + str(np.round(np.std(VPTMxyz[:,0][:-(int(wndw/2)-1)]-VPTMxyz_Mean[:,0][int(wndw/2)-1:]), decimals=3))); 
                        ax6.hist(VPTMxyz[:,1][:-(int(wndw/2)-1)]-VPTMxyz_Mean[:,1][int(wndw/2)-1:], 20, facecolor='g', alpha=0.4,label='ystd:' + str(np.round(np.std(VPTMxyz[:,1][:- (int(wndw/2)-1)]-VPTMxyz_Mean[:,1][int(wndw/2)-1:]), decimals=3))); 
                        ax6.set_xlabel('MLE-mean [um]')  

                        #ax6.hist(np.sqrt(VPTMxyz_Var[:,0]), 20, facecolor='r', alpha=0.4); 
                        #ax6.hist(np.sqrt(VPTMxyz_Var[:,1]), 20, facecolor='g', alpha=0.4); 
                        #ax6.hist(np.sqrt(VPTMxyz_Var[:,2]), 20, facecolor='b', alpha=0.4)          
                        #ax6.set_xlabel('STD [um] over: ' + str(wndw))                
                else:
                    ax6.hist(VPTMxyz[:,0]-xyzgroundtruth[:,0], 20, facecolor='r', alpha=0.4, label='xstd:' + str(np.round(np.std(VPTMxyz[:,0]-xyzgroundtruth[:,0]), decimals=3))); 
                    ax6.hist(VPTMxyz[:,1]-xyzgroundtruth[:,1], 20, facecolor='g', alpha=0.4,label='ystd:' + str(np.round(np.std(VPTMxyz[:,1]-xyzgroundtruth[:,1]), decimals=3))); 
                    ax6.hist(VPTMxyz[:,2]-xyzgroundtruth[:,2], 20, facecolor='b', alpha=0.4,label='zstd:' + str(np.round(np.std(VPTMxyz[:,2]-xyzgroundtruth[:,2]), decimals=3)))          
                    ax6.set_xlabel('MLE-Ture [um]')  
                ax6.legend()
                ax6.set_yticklabels([])

        
                #ax9-ax10 Photon counts
                ax9.plot(Vt,Ni,'-') 
                if MovieStepSize>0: ax9.axvline(Vt[FrameIndex], color='r',linestyle='-')
                ax9.xaxis.set_major_locator(MultipleLocator(500))
                ax9.set_ylabel('Ph #')
                ax9.set_xlabel('ms')
                ax9.grid()

                ax10.hist(Ni, 20) 
                ax10.set_yticklabels([])
                ax10.set_xlabel('[Ph #]')  

                #3D image
                if 0:
                        #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
                    if MovieStepSize>0: ls=LineCollection(segments[:FrameIndex], linewidths=1,colors=colors[:FrameIndex], linestyle='solid')
                    else: ls=LineCollection(segments, linewidths=1,colors=colors, linestyle='solid')

                    line=ax11.add_collection(ls)
                    #fig.colorbar(line, ax=ax11)
                    #ax11.plot(x,y,'-', c=color[0,:])

                    xmean=np.mean(x_pix)
                    ymean=np.mean(y_pix)
        
                    ticks=np.arange(0,ImTot[:,:,0].shape[0],4)
                    xticks=np.linspace(xmean-xyImageBox,xmean+xyImageBox,10)
                    yticks=np.linspace(ymean-xyImageBox,ymean+xyImageBox,10)
                    xtick_labels=np.round( (xticks-(xmean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]
                    ytick_labels=np.round( (yticks-(ymean-xyImageBox))*CamRes,decimals=2).astype(str) #[um]

                    ax11.set_xticks(xticks); ax11.set_yticks(yticks)
                    ax11.set_xticklabels(xtick_labels); ax11.set_yticklabels(ytick_labels)
                    ax11.set_xlim([xmean-xyImageBox,xmean+xyImageBox])
                    ax11.set_ylim([ymean-xyImageBox,ymean+xyImageBox])
                    #ax11.set_xlabel('um'); ax11.set_ylabel('um')
                    ax11.set_xticklabels([]);ax11.set_yticklabels([])


                ax11.plot3D(VPTMxyz[:,0],VPTMxyz[:,1], VPTMxyz[:,2],'-k',alpha=0.4);
                ax11.plot3D(VPTMxyz[:,0],VPTMxyz[:,1], VPTMxyz[:,2],'ob',alpha=0.4,markersize=1);
                #scatterPlot=ax11.scatter(VPTMxyz[:,0],VPTMxyz[:,1], VPTMxyz[:,2],s=1, c=Sxyz,cmap=cm.jet);   
                ax11.set_xlabel('x [um]');ax11.set_ylabel('y [um]');ax11.set_zlabel('z [um]')
                ax11.view_init(elev=35., azim=-70)
                # Hide axes ticks
                #ax11.set_xticks([-1,1])
                #ax11.set_yticks([-1,1])
                #ax11.set_zticks([-2,0])
                #ax11.xaxis._axinfo['juggled'] = (0,0,0)
                #ax11.yaxis._axinfo['juggled'] = (1,1,1)
                #ax11.zaxis._axinfo['juggled'] = (-2,2,2)


                # Res vs photon rate 
                ax12.plot(resVsPrate_Dic[TrajToGet][:,3]/1e3, resVsPrate_Dic[TrajToGet][:,0]*1e3,'-or',alpha=0.4,label='x')
                ax12.plot(resVsPrate_Dic[TrajToGet][:,3]/1e3, resVsPrate_Dic[TrajToGet][:,1]*1e3,'-og',alpha=0.4,label='y')
                ax12.plot(resVsPrate_Dic[TrajToGet][:,3]/1e3, resVsPrate_Dic[TrajToGet][:,2]*1e3,'-ob',alpha=0.4,label='z')
                ax122=ax12.twinx()
                ax122.plot(resVsPrate_Dic[TrajToGet][:,3]/1e3, resVsPrate_Dic[TrajToGet][:,4]*1e3,'-*k',alpha=0.4,label='')
                ax12.set_xticks(np.arange(0,500,50))
                ax12.set_xlim(0,460)
                ax122.set_xticks(np.arange(0,500,50))
                ax122.set_xlim(0,460)
                ax12.set_yticks(np.arange(0,130,10))
                ax12.set_ylabel("Tracking error [nm]"); 
                ax122.set_ylabel("Time res [ms]"); 
                ax12.set_xlabel("Photon rate [kHz]");
                ax12.grid(True);
                ax12.legend()

                

                #########################################################

                if do_save_plot:
                    if MovieStepSize>0:
                        file="MovieFrame_" + str(TrajToGet) + '_D_traj_' + str(FrameIndex)+ FileType
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Movie_Art_Traj_D_filtered_' + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    else:
                        file="Art_" + str(TrajToGet) + '_D_traj' + FileType
                        path=ScanProject.Files['AnalysisConfig'][:-18] +'Art_D_filtered_' + FolderEnd +'/'
                        Path(path).mkdir(parents=True, exist_ok=True)
                    fig.savefig(path + file, bbox_inches='tight')
                    plt.close()
                    print('Done!')
                else:
                    plt.show()

                FrameIndex+=MovieStepSize
                if MovieStepSize<=0: break
                elif FrameIndex>=len(Vt):break

    def get_res(ScanProject,TrajToGet,TrajDictionary,wndw=10, groundtruth=None):

            TrajOffset=[0,0,1.0,2.0] #txyz
            txyznM=TrajDictionary[TrajToGet]*1.0

                #Groundtruth data 
            if np.sum(groundtruth==None):
                xyzgroundtruth=np.zeros((3,txyznM.shape[1]))
            else:
                if len(groundtruth.shape)==3:
                    xyzgroundtruthAll=groundtruth[:-1,:,:3]               
                    xyzgroundtruth=np.mean(xyzgroundtruthAll,axis=1)#Mean over blured out pixle
                else:
                    xyzgroundtruthAll=groundtruth[:,:-1].T               
                    xyzgroundtruth=xyzgroundtruthAll
                

                #CVE D and sigma evaluation
            txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
            txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]

            txyzCVE_Sigma=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu^2]
            txyzCVE_Sigma[:,0]=(txyzCVE_Sigma[:,0]-txyzCVE_Sigma[0,0])*1e3 #[ms]
            txyzCVE_Sigma=np.abs(txyzCVE_Sigma) #Force all values to be positive

            ni=txyznM[4,:]

                #time
            Vt=txyznM[0,:]              #[s]


            # Photon rate
            DT=np.mean(np.diff(Vt))
            MLE_tail=ScanProject.TrajectorySelectionRules['MLETailFactor']
            window_size=20
            Ni=np.convolve((1-MLE_tail)**np.arange(window_size),ni)[0:-window_size+1]
            Prate = np.mean(Ni)/DT

            # Time res
            if MLE_tail<1:
                Thalf=DT * (1+np.log(0.5)/np.log(1-MLE_tail))
            else:
                Thalf=DT
            

                #Piezo stage position
            TStartEnd=[Vt[0],Vt[-1]] 
            StartEndPos=np.where(np.abs(np.diff((ScanProject.ScanMatrix[0,:]>TStartEnd[0])*(ScanProject.ScanMatrix[0,:]<TStartEnd[1]))))[0]
            StartEndPos=np.expand_dims(np.append(np.array([0]),StartEndPos),axis=1)
            
            T_Raw,XP_Raw,YP_Raw,XE_Raw,YE_Raw,XO_Raw,YO_Raw,ZI_Raw,APD1_Raw,APD2_Raw,J_Raw,MemIndex_Raw,ZI_um_Raw,ZI_OffsetIndex_Raw, zCam_Raw =  _MS.get_traj_raw_data4(StartEndPos, 0, ScanProject.ScanMatrix, ScanProject)
            txyz_Traj_Piezo=np.vstack((T_Raw,XP_Raw,YP_Raw,ZI_um_Raw))
            txyz_Traj_Piezo=_MS.get_Scan_toImagCoordinates(ScanProject, txyz_Traj_Piezo)
            
            #ZeroPoint=np.array([XP_Raw[0],YP_Raw[0],ZI_um_Raw[0]])
                #MLE
            VPTMxyz_O=np.transpose(txyznM[1:4,:])
            
                #Start at zero time and at zero piezo pos
            VPTMxyz=VPTMxyz_O-txyz_Traj_Piezo[1:,0]+TrajOffset[1:] #VPTMxyz_O[0,:]
            if  np.sum(groundtruth==None)==0:
                xyzgroundtruth=xyzgroundtruth-txyz_Traj_Piezo[1:,0]+TrajOffset[1:]
                xyzgroundtruthAll=xyzgroundtruthAll-txyz_Traj_Piezo[1:,0]+TrajOffset[1:]
            Vt=(Vt-Vt[0])*1e3 #[ms]

            txyz_Traj_Piezo=(txyz_Traj_Piezo.T-txyz_Traj_Piezo[:,0]+TrajOffset).T
            txyz_Traj_Piezo[0,:]=txyz_Traj_Piezo[0,:]*1e3

                #Running mean on the trajectory
            wndwS=1 #wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            T_Mean= np.convolve(np.ones(wndw)/wndw,np.concatenate(( np.ones(wndw-1)*Vt[0],Vt)))
            T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
            VPTMxyz_Mean=np.zeros((len(T_Mean),3))
            VPTMxyz_Var=np.zeros((len(T_Mean),3))
            for i in range(3):
                VPTMxyz_temp  = np.convolve(np.ones(wndw)/wndw,np.concatenate(( np.ones(wndw-1)*VPTMxyz[0,i],VPTMxyz[:,i])))
                VPTMxyz_Mean[:,i]  = VPTMxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

                VPTMxyz_Var_temp  = np.convolve(np.ones(wndw)/wndw,(np.concatenate(( np.ones(wndw-1)*VPTMxyz[0,i],VPTMxyz[:,i])))**2) #Var(X)=E[X^2]-E[X]^2
                VPTMxyz_Var[:,i]  = VPTMxyz_Var_temp[wndw-1:-wndw+1:wndwS] - (VPTMxyz_Mean[:,i])**2
            
            #Groundtruth data 
            if np.sum(groundtruth==None):
                if np.sum(np.isnan(VPTMxyz))==0:
                    #res=np.std(VPTMxyz-VPTMxyz_Mean,axis=0)
                    res=np.std(VPTMxyz[:-(int(wndw/2)-1),:] -VPTMxyz_Mean[int(wndw/2)-1:,:],axis=0)
            else:
                res=np.std(VPTMxyz-xyzgroundtruth,axis=0)
                    
            return np.hstack((res, Prate,Thalf))
    
    def plot_Diffusion_Summery(ScanProject,TrajDictionary,do_save_plot=0,FileEnd=''):
        """
        This function is to plot trajectores
        """
        D_all=np.array([])
        for TrajToGet in range(len(TrajDictionary)):
                if ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D'].shape[0]!=0:
                    #txyznM=TrajDictionary[TrajToGet]*1.0
                        #CVE D and sigma evaluation
                    txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
                    D_all=np.append(D_all,np.mean(txyzCVE_D[:,1:],axis=1))


        #

        #log-normal disrtibution 
        #mu=np.log(np.median(D_all))
        #sigma2=2*(np.log(np.mean(D_all))-mu)
        #PDF_logNormal=(1/(D*np.sqrt(sigma2*2*np.pi)))*np.exp( -((np.log(D-mu))**2)/(2*sigma2) )

        #fit to invGamma distribution 
        
        
        invfit_alpha, invfit_loc, invfit_scale=stats.invgamma.fit(D_all)#this fitting handles offsets thus no need to add bias to D
        DFittParam=np.linspace(invfit_loc,20,num=400)
        PDF_invGamma=stats.invgamma.pdf(DFittParam, a=invfit_alpha,loc=invfit_loc, scale=invfit_scale)
        DFittParam=DFittParam
        invfit_beta=invfit_scale 
        if invfit_alpha<=1:
            print('Alpha is smaller then 1: ' +str(invfit_alpha))
        elif invfit_alpha<=2:
            print('Alpha is smaller then 2: ' +str(invfit_alpha))

        invGammaMean=(invfit_beta/(invfit_alpha-1))+invfit_loc
        invGammaMode=(invfit_beta/(invfit_alpha+1))+invfit_loc
        invGammaSTD=np.sqrt(invfit_beta**2/((invfit_alpha-1)**2 *(invfit_alpha-2)))

        # Figure construction
        fig = plt.figure(constrained_layout=False)
        fig.set_size_inches(3*3, 2*3, forward=True)
        gs = fig.add_gridspec(nrows=2, ncols=2, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.1, hspace=1.0)

        ax0=fig.add_subplot(gs[0:2, 0:2]) #pos

        #ax4 Histogram D CVE and MSD
        ax0.hist(D_all,bins=DFittParam, density=True); 
        #ax0.plot(D,PDF_logNormal,'-r')
        ax0.plot(DFittParam,PDF_invGamma,'-k',label='invGamma')
        ax0.axvline(np.mean(D_all), color='g',linestyle='-',label='CVE mean:' + str(np.round(np.mean(D_all),decimals=2)) + ' +-' + str(np.round(np.std(D_all),decimals=2)))
        ax0.axvline(invGammaMean, color='g',linestyle='--',label='mean:' + str(np.round(invGammaMean,decimals=2))+ ' +-' + str(np.round(invGammaSTD,decimals=2)) )
        #ax0.axvline(np.median(D_all), color='r',linestyle='-',label='CVE median:' + str(np.round(np.median(D_all),decimals=2)))
        #ax0.axvline(stats.mode(D_all)[0][0], color='m',linestyle='-',label='CVE mode:' + str(np.round(stats.mode(D_all)[0][0],decimals=2)))
        ax0.axvline(invGammaMode, color='m',linestyle='--',label='mode:' + str(np.round(invGammaMode,decimals=2)))
        
        ax0.set_xlabel("D [um^2/s]")
        ax0.legend()

        
        #########################################################
        if do_save_plot:
            file="D_all"  + '_CVE_traj' + FileEnd + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18] #+'Traj_D_filtered/'
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()
   
    def plot_Hist_Traj_MDS_All(ScanProject,do_save_plot=0):
        """
        This function is to plot trajectores
        """
        D_CVE=np.zeros(len(ScanProject.txyzMSD_traj))
        Sigma_CVE=np.zeros(len(ScanProject.txyzMSD_traj))
        D_MSD=np.zeros(len(ScanProject.txyzMSD_traj))
        R_Var_Mean=np.zeros(len(ScanProject.txyzMSD_traj))
        R_Dic={}
        index=0
        for TrajToGet in ScanProject.txyzMSD_traj: 
            TrajToGet=int(TrajToGet[0])
            wndw=10
            txyznM=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]*1.0

                #time
            Vt=(txyznM[0,:]-txyznM[0,0])*1e3 #[ms]
                #Position data MLE
            VPTMxyz=np.transpose(txyznM[1:4,:])
            VPTMxyz=VPTMxyz-VPTMxyz[0,:]
            R=np.sqrt(np.sum(VPTMxyz**2,axis=1))
            R_Dic[index,'R']=R
            R_Dic[index,'Vt']=Vt

                #Running mean on the trajectory
            wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
            T_Mean= np.convolve(np.ones(wndw)/wndw,Vt)
            T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
            R_Mean=np.zeros((len(T_Mean),3))
            R_Var=np.zeros((len(T_Mean),3))
            
            R_Mean  = np.convolve(np.ones(wndw)/wndw,R)
            R_Mean  = R_Mean[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

            R_Var  = np.convolve(np.ones(wndw)/wndw, R**2) #Var(X)=E[X^2]-E[X]^2
            R_Var  = R_Var[wndw-1:-wndw+1:wndwS] - R_Mean**2
            R_Dic[index,'R_Mean']=R_Mean
            R_Dic[index,'R_Var']=R_Var
            R_Dic[index,'T_Mean']=T_Mean
            R_Var_Mean[index]=np.mean(R_Var)

                #Diffusion data MSD D
            txyzMSD_D=np.zeros((len(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]),4))
            txyzMSD_D[:,0]=ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][0]*1.0
            txyzMSD_D[:,0]=(txyzMSD_D[:,0]-txyzMSD_D[0,0])*1e3 #[ms]
            txyzMSD_D[:,1:4]=np.transpose(ScanProject.txyzMSD_traj[str(TrajToGet),'TDS2'][1])*1.0
            D_MSD[index]=np.mean(txyzMSD_D[:,1:])

                #Diffusion CVE D
            txyzCVE_D=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'D']*1.0 #[mu^2/s]
            txyzCVE_D[:,0]=(txyzCVE_D[:,0]-txyzCVE_D[0,0])*1e3 #[ms]
            D_CVE[index]=np.mean(txyzCVE_D[:,1:])

                #Localization error data CVE varinace 
            txyzCVE_Sigma2=ScanProject.txyzCVEMSD_traj[str(TrajToGet),'Sigma2']*1.0 #[mu]
            txyzCVE_Sigma2[:,0]=(txyzCVE_Sigma2[:,0]-txyzCVE_Sigma2[0,0])*1e3 #[ms]
            Sigma_CVE[index]=np.sqrt(np.mean(txyzCVE_Sigma2[:,1:]))

                #Update index
            index+=1

      
        # Figure construction
        fig = plt.figure(constrained_layout=False)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = fig.add_gridspec(nrows=4, ncols=4, left=0.05, right=0.99, top=0.99, bottom=0.07, wspace=0.1, hspace=1.0)

        ax0=fig.add_subplot(gs[0:2,:]) #pos

        ax1=fig.add_subplot(gs[2:4, 0:2])  #hist D      
        ax2=fig.add_subplot(gs[2:4, 2:4])  #hist sigma2     


        for index in range(len(ScanProject.txyzMSD_traj)):
                    #ax0
            ax0.plot(R_Dic[index,'Vt'],R_Dic[index,'R'],'-b',alpha=0.4);  #MLE data
            ax0.plot(R_Dic[index,'T_Mean'],R_Dic[index,'R_Mean'],'-k',alpha=0.4);  #MLE data

        ax0.set_ylabel("[um]"); 
        ax0.set_xlabel("[ms]"); 
        ax0.grid(True);
        ax0.legend(('R','R_Mean'))

        #ax1 Histogram D CVE and MSD
        ax1.hist(D_CVE, 20, facecolor='g', alpha=0.5); 
        ax1.hist(D_MSD, 20, facecolor='r', alpha=0.5); 
        ax1.axvline(np.mean(D_MSD), color='r',linestyle='-',label='MSD:' + str(np.round(np.mean(D_MSD),decimals=2)))
        ax1.axvline(np.mean(D_CVE), color='g',linestyle='-',label='CVE:' + str(np.round(np.mean(D_CVE),decimals=2)))
        ax1.set_xlabel("D [um^2/s]")
        ax1.legend()

        #ax2 Histogram Sigma Variance 
        ax2.hist(Sigma_CVE*1e3, 20, facecolor='g', alpha=0.5); 
        ax2.hist(np.sqrt(R_Var_Mean)*1e3, 20, facecolor='b', alpha=0.5); 
        ax2.axvline(np.sqrt(np.mean(R_Var_Mean))*1e3, color='b',linestyle='-',label='Var:' + str(np.round(np.sqrt(np.mean(R_Var_Mean))*1e3,decimals=2)))
        ax2.axvline(np.mean(Sigma_CVE)*1e3, color='g',linestyle='-',label='CVE:' + str(np.round(np.mean(Sigma_CVE)*1e3,decimals=2)))
        ax2.set_xlabel("Sigma [nm]")
        ax2.legend()

        
        #########################################################
        if do_save_plot:
            file="T_" + 'D_All_traj' + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show() 

    def plot_Hist_Traj_Mean_STD(ScanProject,TrajToGet,wndw=200,do_save_plot=1):
        """
        This function is to plot trajectores
        """
        GroundTruth=np.nan
        txyznM=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]

        m = int(ScanProject.TrajectorySelectionRules['wndw'])
        ni=txyznM[4,:]

            #time
        Vt=(txyznM[0,:]-txyznM[0,0])*1e3 #[ms]        
            #MLE
        VPTMxyz=np.transpose(txyznM[1:4,:])
            #Running mean on the trajectory
        wndwS=wndw #if 1 it is a moving average, if equal to wndw it is a fix window average
        T_Mean= np.convolve(np.ones(wndw)/wndw,Vt)
        T_Mean=T_Mean[wndw-1:-wndw+1:wndwS]
        
        VGxyz=np.zeros((len(T_Mean),3))
        VGxyz_STD=np.zeros((len(T_Mean),3))
        for i in range(3):
            VGxyz_temp  = np.convolve(np.ones(wndw)/wndw,VPTMxyz[:,i])
            VGxyz[:,i]  = VGxyz_temp[wndw-1:-wndw+1:wndwS]                      #Mean E[X]

            VGxyz_STD_temp  = np.convolve(np.ones(wndw)/wndw,(VPTMxyz[:,i])**2) #Var(X)=E[X^2]-E[X]^2
            VGxyz_STD[:,i]  = VGxyz_STD_temp[wndw-1:-wndw+1:wndwS] - (VGxyz[:,i])**2

        VGxyz_STD=np.sqrt(VGxyz_STD)

            #Polarization 
        Pol=ScanProject.TrajPosEstimation[TrajToGet][3,:]
           #MSD data

           #Groundtruth data 
        if np.prod(np.isnan(GroundTruth)):
            #Ground truth data is set to be the Gaussian data 
            Vxyzpix=VGxyz
        else:
            Vxyzpix=GroundTruth[::m,:,0]*1e6 #[um]

      
        # Figure construction
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(5*3, 3*5, forward=True)
        gs = GridSpec(8, 3, figure=fig)
        ax0=fig.add_subplot(gs[1, 0:2]) #MSD x
        ax1=fig.add_subplot(gs[3, 0:2]) #MSD y
        ax2=fig.add_subplot(gs[5, 0:2]) #MSD z

        ax3=fig.add_subplot(gs[0, 0:2]) #pos x
        ax4=fig.add_subplot(gs[2, 0:2]) #pos y
        ax5=fig.add_subplot(gs[4, 0:2]) #pos z

        ax6=fig.add_subplot(gs[0, 2])   #hist x
        ax7=fig.add_subplot(gs[2, 2])   #hist y
        ax8=fig.add_subplot(gs[4, 2])   #hist z

        ax9=fig.add_subplot(gs[6, 0:2]) #Photon count 
        ax10=fig.add_subplot(gs[6, 2])  #Photon hist

        ax11=fig.add_subplot(gs[7, 0:2]) #Polarization 
        ax12=fig.add_subplot(gs[7, 2]) #Polarization hist

        ax13=fig.add_subplot(gs[1, 2]) #hist MSD x
        ax14=fig.add_subplot(gs[3, 2]) #hist MSD y
        ax15=fig.add_subplot(gs[5, 2]) #hist MSD z


        #ax0-ax3
        
        ax0.plot(T_Mean,VGxyz_STD[:,0],'-m',alpha=0.5); ax1.plot(T_Mean,VGxyz_STD[:,1],'-m',alpha=0.5); ax2.plot(T_Mean,VGxyz_STD[:,2],'-m',alpha=0.5)   #STD MLE data
        ax0.set_ylabel("STD [um]"); ax1.set_ylabel("STD [um]"); ax2.set_ylabel("STD [um]")
        ax0.grid(True); ax1.grid(True); ax2.grid(True)


        #ax3-ax5
        ax3.plot(T_Mean,VGxyz[:,0],'-g',alpha=0.5);   ax4.plot(T_Mean,VGxyz[:,1],'-g',alpha=0.5);   ax5.plot(T_Mean,VGxyz[:,2],'-g',alpha=0.5)         #Gaussian data 
        ax3.plot(Vt,VPTMxyz[:,0],'-m',alpha=0.5); ax4.plot(Vt,VPTMxyz[:,1],'-m',alpha=0.5); ax5.plot(Vt,VPTMxyz[:,2],'-m',alpha=0.5)   #MLE data
        ax3.set_ylabel("x [um]"); ax4.set_ylabel("y [um]"); ax5.set_ylabel("z [um]")
        ax3.grid(True); ax4.grid(True); ax5.grid(True)
        #ax5.set_xlabel("t [ms]")
        

        #ax6-ax8
        if np.prod(np.isnan(GroundTruth)):
            if np.sum(np.isnan(VPTMxyz))==0:
                ax6.hist(VPTMxyz[:,0], 20, facecolor='m', alpha=0.5); ax7.hist(VPTMxyz[:,1], 20, facecolor='m', alpha=0.5); ax8.hist(VPTMxyz[:,2], 20, facecolor='m', alpha=0.5)
            if np.sum(np.isnan(VGxyz))==0:
                ax6.hist(VGxyz[:,0],   20, facecolor='g', alpha=0.5); ax7.hist(VGxyz[:,1],   20, facecolor='g', alpha=0.5); ax8.hist(VGxyz[:,2],   20, facecolor='g', alpha=0.5)
            ax8.set_xlabel("Self distance [nm]")
        else: 
            if np.sum(np.isnan(VPTMxyz))==0:
                    ax6.hist(VPTMxyz[:,0]-Vxyzpix[:,0], 20, facecolor='m', alpha=0.5); ax7.hist(VPTMxyz[:,1]-Vxyzpix[:,1], 20, facecolor='m', alpha=0.5); ax8.hist(VPTMxyz[:,2]-Vxyzpix[:,2], 20, facecolor='m', alpha=0.5)
            if np.sum(np.isnan(VGxyz))==0:
                    ax6.hist(VGxyz[:,0]-Vxyzpix[:,0],   20, facecolor='g', alpha=0.5); ax7.hist(VGxyz[:,1]-Vxyzpix[:,1],   20, facecolor='g', alpha=0.5); ax8.hist(VGxyz[:,2]-Vxyzpix[:,2],   20, facecolor='g', alpha=0.5)
            ax8.set_xlabel("Groundtruth distance [nm]")
        #ax6.legend(('DMLE: ' + str(np.round(np.mean(DMLE),decimals=3)) + '+- ' + str(np.round(np.std(DMLE),decimals=3)),'DG: ' + str(np.round(np.mean(DG),decimals=3)) + '+- ' + str(np.round(np.std(DG),decimals=3))))
        

        #ax9-ax10 Photon counts
        ax9.plot(Vt,ni,'-') 
        ax10.hist(ni, 20)        
        ax9.set_ylabel('Photons #')
        ax9.grid()

        #ax11-ax12 Polarization counts
        ax11.plot(Vt,Pol,'-') 
        ax12.hist(Pol, 20)        

        ax11.set_xlabel('Time [ms]')
        ax11.set_ylabel('Polarization #')
        ax11.grid()
        

        #ax13-ax15
        ax13.hist(VGxyz_STD[:,0], 20, facecolor='m', alpha=0.5); ax14.hist(VGxyz_STD[:,1], 20, facecolor='m', alpha=0.5); ax15.hist(VGxyz_STD[:,2], 20, facecolor='m', alpha=0.5)
        ax15.set_xlabel("[um]")


        ax0.set_title("Traj: " + str(TrajToGet) + ' Mean window size: ' + str(wndw))
        #########################################################
        if do_save_plot:
            file="T_" + str(TrajToGet) + '_Mean_STD_traj' + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

    def plotPhotonCount(ScanProject,TrajToGet,do_save_plot=1):
        """        
        Plots the photons counts
        """
        wndw=1
        wndwS=wndw #if 1 it is a moving average (do not work in this code) if equal to wndw it is like a fix window average

        T,XP,YP,XE,YE,XO,YO,ZI,APD1,APD2,J,MemIndex,ZI_um,ZI_OffsetIndex =  _MS.get_traj_raw_data3(ScanProject.Trajs, TrajToGet, ScanProject.ScanMatrix, ScanProject)
        #plt.plot(T-T[0],ZI_um); plt.plot(T-T[0],XP);plt.plot(T-T[0],YP);plt.grid();plt.show()

        Elements=np.unique(MemIndex).astype('int')
        MLength=int(len(Elements)*np.floor(len(T)/len(Elements)))
        SubLength=int(np.floor(len(T)/len(Elements)))

        T=(T[:MLength]-T[0])*1e3 #[ms]
        n=APD1[:MLength]+APD2[:MLength]
        ZI=ZI_OffsetIndex[:MLength]
        MemIndex=MemIndex[:MLength]       
        
        n_all=np.zeros((len(Elements),int(SubLength/wndwS)))
        t_all=np.zeros((len(Elements),int(SubLength/wndwS)))
        for i in range(len(Elements)):
            OfInterest=np.where(MemIndex==Elements[i])
            n_temp = np.convolve(np.ones(wndw),n[OfInterest])
            t_temp = np.convolve(np.ones(wndw)/wndw,T[OfInterest])

            if wndw!=1:
                n_all[i,:]=n_temp[wndw-1:-wndw+1:wndwS]
                t_all[i,:]=t_temp[wndw-1:-wndw+1:wndwS]
            else:
                n_all[i,:]=n_temp
                t_all[i,:]=t_temp

        
        #Calculate ratios 
        P_all=np.zeros(n_all.shape)
        for i in range(n_all.shape[0]):
            P_all[i,:]=n_all[i,:]/np.sum(n_all,axis=0) 
        
        #Separate +- counts 
        for i in range(len(Elements)):
            n_all[i,:]=n_all[i,:]*(-1)**Elements[i]

        # Figure construction
        amount=6
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = GridSpec(amount, 1, figure=fig)
        ax={}
        for i in range(amount):
            ax[i]=fig.add_subplot(gs[i, 0])
 
        color=cm.rainbow(np.linspace(0,1,len(Elements)))
        pL=['x','px','y','py','z','pz']
        R=(np.array(range(int(amount/2)))*2).astype('int')
        for p in R:
            ax[p].plot(t_all[p,:],n_all[p,:],'-',c=color[p,:],alpha=0.5,label='M' + str(int(Elements[p])) ) 
            ax[p].plot(t_all[p+1,:],n_all[p+1,:],'-',c=color[p+1,:],alpha=0.5,label='M' + str(int(Elements[p+1]))) 

            ax[p+1].plot(t_all[p,:],P_all[p,:],'-',c=color[p,:],alpha=0.5,label='M' + str(int(Elements[p])) ) 
            ax[p+1].plot(t_all[p+1,:],P_all[p+1,:],'-',c=color[p+1,:],alpha=0.5,label='M' + str(int(Elements[p+1]))) 

            ax[p].grid();    ax[p+1].grid()
            ax[p].set_ylabel(pL[p] + ' #'); ax[p+1].set_ylabel(pL[p+1])
            ax[p].legend(); ax[p+1].legend()
        
        ax[amount-1].set_xlabel('Time [ms]')        
        
        if do_save_plot:
            file="T_" + str(TrajToGet) + 'PhotonCount' + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Saving: ' + file)
        else:
            plt.show()

    def plotPhotonCount2(ScanProject,TrajToGet,Trajs,wndw=10,do_save_plot=1):
        """        
        Plots the photons counts per pattern position
        """
        wndwS=wndw #if 1 it is a moving average (do not work for this part) if equal to wndw it is like a fix window average
        
        T,XP,YP,XE,YE,XO,YO,ZI,APD1,APD2,J,MemIndex,ZI_um,ZI_OffsetIndex,zcam =  _MS.get_traj_raw_data4(ScanProject.Trajs, TrajToGet, ScanProject.ScanMatrix, ScanProject)
        #plt.plot(T-T[0],ZI_um); plt.plot(T-T[0],XP);plt.plot(T-T[0],YP);plt.grid();plt.show()


        Elements=np.unique(MemIndex).astype('int')      #Find what type of memory elements are in the data
        SubLength=int(np.floor(len(T)/len(Elements)))   #Find largest amount of section given one memory cycle 
        MLength=int(len(Elements)*SubLength)            #Find longest section length that includes as many memory cycles as possible          

        T=(T[:MLength]-T[0])*1e3 #[ms]
        n=APD1[:MLength]+APD2[:MLength]
        ZI=ZI_OffsetIndex[:MLength]
        MemIndex=MemIndex[:MLength]   
        XP=XP[:MLength]
        YP=YP[:MLength]
        XE=XE[:MLength]
        YE=YE[:MLength]
        ZI_um=ZI_um[:MLength]
        zcam=zcam[:MLength]

        n_all=np.zeros((len(Elements),int(SubLength/wndwS)))
        t_all=np.zeros((len(Elements),int(SubLength/wndwS)))
        XP_all=np.zeros((len(Elements),int(SubLength/wndwS)))
        YP_all=np.zeros((len(Elements),int(SubLength/wndwS)))
        ZP_all=np.zeros((len(Elements),int(SubLength/wndwS)))
        XE_all=np.zeros((len(Elements),int(SubLength/wndwS)))
        YE_all=np.zeros((len(Elements),int(SubLength/wndwS)))
        ZC_all=np.zeros((len(Elements),int(SubLength/wndwS)))
        for i in range(len(Elements)):
            OfInterest=np.where(MemIndex==Elements[i])
            n_temp = np.convolve(np.ones(wndw),n[OfInterest])
            t_temp = np.convolve(np.ones(wndw)/wndw,T[OfInterest])
            XP_temp = np.convolve(np.ones(wndw)/wndw,XP[OfInterest])
            YP_temp = np.convolve(np.ones(wndw)/wndw,YP[OfInterest])
            ZP_temp = np.convolve(np.ones(wndw)/wndw,ZI_um[OfInterest])
            XE_temp = np.convolve(np.ones(wndw)/wndw,XE[OfInterest])
            YE_temp = np.convolve(np.ones(wndw)/wndw,YE[OfInterest])
            ZC_temp = np.convolve(np.ones(wndw)/wndw,zcam[OfInterest])

            if wndw!=1:
                n_all[i,:]=n_temp[wndw-1:-wndw+1:wndwS]
                t_all[i,:]=t_temp[wndw-1:-wndw+1:wndwS]
                XP_all[i,:]=XP_temp[wndw-1:-wndw+1:wndwS]
                YP_all[i,:]=YP_temp[wndw-1:-wndw+1:wndwS]
                ZP_all[i,:]=ZP_temp[wndw-1:-wndw+1:wndwS]
                XE_all[i,:]=XE_temp[wndw-1:-wndw+1:wndwS]
                YE_all[i,:]=YE_temp[wndw-1:-wndw+1:wndwS]
                ZC_all[i,:]=ZC_temp[wndw-1:-wndw+1:wndwS]
            else:
                n_all[i,:]=n_temp
                t_all[i,:]=t_temp
                XP_all[i,:]=XP_temp
                YP_all[i,:]=YP_temp
                ZP_all[i,:]=ZP_temp
                XE_all[i,:]=XE_temp
                YE_all[i,:]=YE_temp
                ZC_all[i,:]=ZC_temp
        
        if 1:
            for i in range(13):
                a=0+i;b=14+i;c=28+i;
                if i==0:
                    n_all[a,:]=n_all[a,:]*0

                fig, ax = plt.subplots()
                ax.plot(t_all[a,:],n_all[a,:],'-r',label=str(a));
                ax.plot(t_all[b,:],n_all[b,:],'-g',label=str(b));
                ax.plot(t_all[c,:],n_all[c,:],'-b',label=str(c));
                plt.legend()

                do_save_plot=1
                if do_save_plot:
                    file="T_" + str(TrajToGet) + 'PhotonCount_mem' + str(a) + '_' + str(b) + '_' + str(c) + '.png'
                    path=ScanProject.Files['AnalysisConfig'][:-18] + 'PhotonCounts' + '/'
                    Path(path).mkdir(parents=True, exist_ok=True)
                    fig.savefig(path + file)
                    plt.close()
                    print('Saving: ' + file)
                else:
                    plt.show()
        if 0:
            #Calculate ratios 
            P_all=np.zeros(n_all.shape)
            for i in range(n_all.shape[0]):
                P_all[i,:]=n_all[i,:]/np.sum(n_all,axis=0) 
        
            #Separate +- counts 
            for i in range(len(Elements)):
                n_all[i,:]=n_all[i,:]*(-1)**Elements[i]

            #Simple pos estimataion (obs up to a scale factor and only for the EOD part)
            Posisch_xyz=np.zeros((3,P_all.shape[1]))
            Posisch_xyz[0,:]=(P_all[0,:]-P_all[1,:])
            Posisch_xyz[1,:]=(P_all[2,:]-P_all[3,:])
            Posisch_xyz[2,:]=(P_all[4,:]-P_all[5,:])

            ##Test
            txyznM=Trajs[TrajToGet]; txyznM[0,:]=txyznM[0,:]*1e3 #ms
            Time=np.convolve(np.ones(wndw)/wndw,(txyznM[0,:]-txyznM[0,0])*1e3)[wndw-1:-wndw+1:wndwS]
            Y=np.convolve(np.ones(wndw)/wndw,txyznM[2,:])[wndw-1:-wndw+1:wndwS]
            X=np.convolve(np.ones(wndw)/wndw,txyznM[1,:])[wndw-1:-wndw+1:wndwS]

            ##
            if 1:
                fig, axs = plt.subplots(3, 1)        
                axs[0].plot(txyznM[0,:]-txyznM[0,0],txyznM[1,:]-np.mean(txyznM[1,:]),'gray');
                axs[0].plot(t_all[0,:],(np.mean(XP_all,axis=0)-np.mean(np.mean(XP_all,axis=0))+np.mean(XE_all,axis=0))+0.5*((n_all[2,:]+n_all[3,:])/np.sum(np.abs(n_all),axis=0)),'r');        
                axs[0].plot(t_all[0,:],(np.mean(XP_all,axis=0)-np.mean(np.mean(XP_all,axis=0))+np.mean(XE_all,axis=0))+0.25*((n_all[2,:]+n_all[3,:])/np.sum(np.abs(n_all),axis=0)),'b');        
                axs[0].plot(t_all[0,:],(np.mean(XP_all,axis=0)-np.mean(np.mean(XP_all,axis=0))+np.mean(XE_all,axis=0)),'m');
                axs[0].set_ylabel('x'); axs[0].grid()

                axs[1].plot(txyznM[0,:]-txyznM[0,0],txyznM[2,:]-np.mean(txyznM[2,:]),'gray');
                axs[1].plot(t_all[0,:],(np.mean(YP_all,axis=0)-np.mean(np.mean(YP_all,axis=0))+np.mean(YE_all,axis=0))+0.5*((n_all[0,:]+n_all[1,:])/np.sum(np.abs(n_all),axis=0)),'r');        
                axs[1].plot(t_all[0,:],(np.mean(YP_all,axis=0)-np.mean(np.mean(YP_all,axis=0))+np.mean(YE_all,axis=0))+0.2*((n_all[0,:]+n_all[1,:])/np.sum(np.abs(n_all),axis=0)),'b');        
                axs[1].plot(t_all[0,:],(np.mean(YP_all,axis=0)-np.mean(np.mean(YP_all,axis=0))+np.mean(YE_all,axis=0)),'m');
                axs[1].set_ylabel('y'); axs[1].grid()

                axs[2].plot(txyznM[0,:]-txyznM[0,0],txyznM[3,:]-np.mean(txyznM[3,:]),'gray');
                axs[2].plot(t_all[0,:],1.0*(np.mean(ZP_all,axis=0)-np.mean(np.mean(ZP_all,axis=0)))-0.5*((n_all[4,:]+n_all[5,:])/np.sum(np.abs(n_all),axis=0)),'r');        
                axs[2].plot(t_all[0,:],1.0*(np.mean(ZP_all,axis=0)-np.mean(np.mean(ZP_all,axis=0)))-0.15*((n_all[4,:]+n_all[5,:])/np.sum(np.abs(n_all),axis=0)),'b');        
                axs[2].plot(t_all[0,:],1.0*(np.mean(ZP_all,axis=0)-np.mean(np.mean(ZP_all,axis=0))),'m');
                axs[2].plot(t_all[0,:],1.0*(np.mean(ZP_all,axis=0)-np.mean(np.mean(ZP_all,axis=0)))[0]+(np.mean(ZC_all,axis=0)-np.mean(ZC_all,axis=0)[0]),'k');
                axs[2].set_ylabel('z'); axs[2].grid()

                axs[2].set_xlabel('ms')
                axs[0].set_title("Traj:" + str(TrajToGet))
                plt.show()
       

            # Figure construction
            if 1:
                amount=6
                fig = plt.figure(constrained_layout=True)
                fig.set_size_inches(3*5, 2*5, forward=True)
                fig.suptitle('Trajectory #' + str(TrajToGet), fontsize=16)
                gs = GridSpec(amount, 1, figure=fig)
                ax={}
                for i in range(amount):
                    ax[i]=fig.add_subplot(gs[i, 0])
 
                color=cm.rainbow(np.linspace(0,1,len(Elements)))
                pL=['x','px','y','py','z','pz']
                R=(np.array(range(int(amount/2)))*2).astype('int')
                for p in R:
                    ax[p].plot(t_all[p,:],n_all[p,:],'-',c=color[p,:],alpha=0.5,label='M' + str(int(Elements[p])) ) 
                    ax[p].plot(t_all[p+1,:],n_all[p+1,:],'-',c=color[p+1,:],alpha=0.5,label='M' + str(int(Elements[p+1]))) 

                    ax[p+1].plot(t_all[p,:],P_all[p,:],'-',c=color[p,:],alpha=0.5,label='M' + str(int(Elements[p])) ) 
                    ax[p+1].plot(t_all[p+1,:],P_all[p+1,:],'-',c=color[p+1,:],alpha=0.5,label='M' + str(int(Elements[p+1]))) 
                    ax[p+1].plot(t_all[p,:],Posisch_xyz[int(p/2),:],'-',c='k',alpha=0.5) 

                    ax[p].grid();    ax[p+1].grid()
                    ax[p].set_ylabel(pL[p] + ' #'); ax[p+1].set_ylabel(pL[p+1])
                    ax[p].legend(); ax[p+1].legend()
        
                ax[amount-1].set_xlabel('Time [ms]')        
        
                if do_save_plot:
                    file="T_" + str(TrajToGet) + 'PhotonCount' + '.png'
                    path=ScanProject.Files['AnalysisConfig'][:-18] + 'PhotonCounts' + '/'
                    Path(path).mkdir(parents=True, exist_ok=True)
                    fig.savefig(path + file)
                    plt.close()
                    print('Saving: ' + file)
                else:
                    plt.show()

    def plotPiezostages(ScanProject,TrajToGet,do_save_plot=1):
        """        
        Plots the photons counts
        """
        txyznM=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]
        xyzG=np.vstack((ScanProject.TrajPosEstimation[TrajToGet][1,:],
                    ScanProject.TrajPosEstimation[TrajToGet][2,:],
                    ScanProject.TrajPosEstimation[TrajToGet][-1,:])) 


        T,XP,YP,XE,YE,XO,YO,ZI,APD1,APD2,J,MemIndex,ZI_um,ZI_OffsetIndex =  _MS.get_traj_raw_data3(ScanProject.Trajs, TrajToGet, ScanProject.ScanMatrix, ScanProject)
        T=(T-T[0])*1e3 #[ms]

        # Figure construction
        amount=3
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = GridSpec(amount, 1, figure=fig)
        ax={}
        for i in range(amount):
            ax[i]=fig.add_subplot(gs[i, 0])
 
        color=cm.rainbow(np.linspace(0,1,3))
        ax[0].plot(T,XP-XP[0],'-',c='b',alpha=0.5,label='x-piezo')
        ax[0].plot(T,YP-YP[0],'-',c='g',alpha=0.5,label='y-piezo')
        ax[0].plot(T,ZI_um-ZI_um[0],'-',c='r',alpha=0.5,label='z-piezo') 
        ax[0].set_ylabel('um')
        ax[0].legend() 
        ax[0].grid()

        ax[1].plot(T,XE,'-',c='b',alpha=0.5,label='x-EOD')
        ax[1].plot(T,YE,'-',c='g',alpha=0.5,label='y-EOD')
        ax[1].set_ylabel('um')
        ax[1].legend() 
        ax[1].grid()

        ax[2].plot((txyznM[0,:]-txyznM[0,0])*1e3,txyznM[1,:]-txyznM[1,0],'-',c='b',alpha=0.5,label='x-MLE')
        ax[2].plot((txyznM[0,:]-txyznM[0,0])*1e3,txyznM[2,:]-txyznM[2,0],'-',c='g',alpha=0.5,label='y-MLE')
        ax[2].plot((txyznM[0,:]-txyznM[0,0])*1e3,txyznM[3,:]-txyznM[3,0],'-',c='r',alpha=0.5,label='z-MLE')

        ax[2].set_ylabel('um')
        ax[2].legend() 
        ax[2].grid()

        ax[amount-1].set_xlabel('Time [ms]')      

        
        if do_save_plot:
            file="T_" + str(TrajToGet) + '_Piezo' + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

        print('Done')

    def plotPiezostagesFFT(ScanProject,TrajToGet,trim=np.array([0,0]),do_save_plot=1):
        """        
        Plots the photons counts
        """
        #MLE data
        txyznM=ScanProject.PoissonProbXYZMap_NumPFS[TrajToGet]*1.0
        txyznM[0,:]=(txyznM[0,:]-txyznM[0,0])*1e3 #[ms]

        #Raw data
        T,XP,YP,_,_,_,_,_,_,_,_,_,ZI_um,_ =  _MS.get_traj_raw_data3(ScanProject.Trajs, TrajToGet, ScanProject.ScanMatrix, ScanProject)
        T=(T-T[0])*1e3 #[ms]

        #Trim
        if np.sum(trim)>0:
            valid=np.where((T>=trim[0])*(T<=trim[1]))
            T=T[valid]
            XP=XP[valid]
            YP=YP[valid]
            ZI_um=ZI_um[valid]

            valid=np.where((txyznM[0,:]>=trim[0])*(txyznM[0,:]<=trim[1]))
            txyznM=txyznM[:,valid[0]]



        # Fourier spectra of the MLE signals
        fft_x_MLE, fx = _MS.fft_processing(txyznM[1,:],txyznM[0,:])
        fft_y_MLE, fy = _MS.fft_processing(txyznM[2,:],txyznM[0,:])
        fft_z_MLE, fz = _MS.fft_processing(txyznM[3,:],txyznM[0,:])


        # Figure construction
        amount=3
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(3*5, 2*5, forward=True)
        gs = GridSpec(amount, 1, figure=fig)
        ax={}
        for i in range(amount):
            ax[i]=fig.add_subplot(gs[i, 0])
 
        color=cm.rainbow(np.linspace(0,1,3))
        ax[0].plot(T,XP-XP[0],'-',c='b',alpha=0.5,label='x-piezo')
        ax[0].plot(T,YP-YP[0],'-',c='g',alpha=0.5,label='y-piezo')
        ax[0].plot(T,ZI_um-ZI_um[0],'-',c='r',alpha=0.5,label='z-piezo') 
        ax[0].set_ylabel('um')
        ax[0].legend() 
        ax[0].grid()

        ax[1].plot(txyznM[0,:],txyznM[1,:]-txyznM[1,0]+0.2,'-',c='b',alpha=1.0,label='x-MLE')
        ax[1].plot(txyznM[0,:],txyznM[2,:]-txyznM[2,0],'-',c='g',alpha=1.0,label='y-MLE')
        ax[1].plot(txyznM[0,:],txyznM[3,:]-txyznM[3,0]-0.2,'-',c='r',alpha=1.0,label='z-MLE')

        ax[1].set_ylabel('um')
        ax[1].legend() 
        ax[1].grid()
        ax[1].set_xlabel('Time [ms]')      

        ax[2].semilogy(fx, fft_x_MLE,c='b', alpha=1.0, label='x-MLE')
        ax[2].semilogy(fy, fft_y_MLE,c='g', alpha=1.0, label='y-MLE')
        ax[2].semilogy(fz, fft_z_MLE,c='r', alpha=1.0, label='z-MLE')
        ax[2].set_xlabel('khz')
        ax[2].legend() 
        ax[2].grid()

        
        if do_save_plot:
            file="T_" + str(TrajToGet) + '_Piezo' + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

        print('Done')

    def plot_ovelayTrajOnCells(ScanProject,do_save_plot=1):
        #Plots the images: 
        #   Bright filed image
        #   Fluorecense image
        #   Scan image
        CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]

        ImTot=ScanProject.ImgOverlay

        # Get the transformation and get the Affine transform invers see: https://en.wikipedia.org/wiki/Affine_transformation
        Tran=ScanProject.ImgOverlayTransform
        Tran[0:2,0:2]=np.linalg.inv(Tran[0:2,0:2])
        Tran[0:2,2]=-np.matmul(Tran[0:2,0:2],Tran[0:2,2])
        
        Traj_xyz_Campix={}
        for i in range(len(ScanProject.TrajPosEstimation)):
            Traj_xyz=ScanProject.TrajPosEstimation[i][[1,2,10],:]
            Traj_xy1_Campix=Traj_xyz/CamRes
            Traj_xy1_Campix[2,:]=1      #Remove z and change it to a 1 for the Affine transromation matrix
            Traj_xy1_Campix= np.matmul(Tran,Traj_xy1_Campix) #Apply transformation 

            Traj_xyz_Campix[i]=Traj_xy1_Campix
            Traj_xyz_Campix[i][2,:]=Traj_xyz[2,:]/CamRes


        #Build figure
        amount=1
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(amount*20, 20, forward=True)
        gs = GridSpec(1, amount, figure=fig)
        ax={}
        for i in range(amount):
            ax[i]=fig.add_subplot(gs[0,i])
        

            #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
        ax[0].imshow(ImTot[:,:,0],cmap='gray',alpha=1.0)
        ax[0].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)
        ax[0].imshow(ImTot[:,:,2],cmap='Reds',alpha=0.3)
        for i in range(len(Traj_xyz_Campix)):
            ax[0].plot(Traj_xyz_Campix[i][0,:],Traj_xyz_Campix[i][1,:],'-r')
            ax[0].text(Traj_xyz_Campix[i][0,0],Traj_xyz_Campix[i][1,0],str(i))
        ax[0].set_title('Amount of tracks: '+ str((len(Traj_xyz_Campix))) )

        if do_save_plot:
            file="Traj_Overlay_" + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

    def plot_ovelayTrajOnCells2(ScanProject,TrajDictionary,CellMask=False,do_save_plot=1):
        #Plots the images: 
        #   Bright filed image
        #   Fluorecense image
        #   Scan image
        if CellMask:
            mask=ScanProject.CamCellMask[:,:,0]>0

        CamRes=ScanProject.SetupSpecifics['Px']/(ScanProject.SetupSpecifics['Mag']*ScanProject.SetupSpecifics['Mag2'])*1e6 #[um]

        ImTot=ScanProject.ImgOverlay
        
        #Build figure
        amount=1
        fig = plt.figure(constrained_layout=True)
        fig.set_size_inches(amount*20, 20, forward=True)
        gs = GridSpec(1, amount, figure=fig)
        ax={}
        for i in range(amount):
            ax[i]=fig.add_subplot(gs[0,i])
        

            #Combined  # https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
        ax[0].imshow(ImTot[:,:,0],cmap='gray',alpha=1.0)
        #ax[0].imshow(ImTot[:,:,1],cmap='jet',alpha=0.3)
        #ax[0].imshow(ImTot[:,:,2],cmap='Reds',alpha=0.3)
        if CellMask:
            ax[0].contour(mask,levels=0, linewidths=1, colors='blue', linestyles='solid')
            ax[0].imshow(mask,cmap='Reds',alpha=0.3)

        color = plt.cm.hsv(np.linspace(0,1,len(TrajDictionary)))
        for i in range(len(TrajDictionary)):
            ax[0].plot(TrajDictionary[i][1,:]/CamRes,TrajDictionary[i][2,:]/CamRes,'-', c=color[i,:])
            ax[0].text(TrajDictionary[i][1,0]/CamRes,TrajDictionary[i][2,0]/CamRes,str(i))
        ax[0].set_title('Amount of tracks: '+ str((len(TrajDictionary))) )

        if do_save_plot:
            file="Traj_filtered_Overlay_" + '.png'
            path=ScanProject.Files['AnalysisConfig'][:-18]
            fig.savefig(path + file)
            plt.close()
            print('Done!')
        else:
            plt.show()

    def PlotCRB(ScanProject,P,xx,yy,zz,TopHatPrior,do_save_plot=1):
        TopHat=np.zeros(TopHatPrior.shape)
        TopHat[np.where(TopHatPrior<0)]=np.nan
        SRB_step=np.zeros((1,20))
        lambda_N=np.zeros(20)
        Hx,Hy,Hz=_MS.SobelFeldman3DKernal()
        for step in range(10):
            lambda_n=(1.0 + 4*step)*np.array([1,1,1,1,1,1]) #(110,105,70)
            SqCRBx=_MS.Get_CRB(P,xx,Hx,lambda_n)
            SqCRBy=_MS.Get_CRB(P,yy,Hy,lambda_n)
            SqCRBz=_MS.Get_CRB(P,zz,Hz,lambda_n)
            SqCRB=(SqCRBx+SqCRBy+SqCRBz)/3
            lambda_N[step]=np.sum(lambda_n)

            for i in range(SqCRB.shape[2]):
                fig = plt.figure(constrained_layout=True)
                ax0=fig.add_subplot()
                ax0.contourf(xx[:,:,i],yy[:,:,i],TopHatPrior[:,:,i],alpha=1.0)              #TopHat visualization  
                cs0=ax0.contourf(xx[:,:,i],yy[:,:,i],SqCRB[:,:,i],levels=np.linspace(0,200,50) ) 
                fig.colorbar(cs0,ax=ax0)
                ax0.set_title("Photon used:" + str(np.round(lambda_n,decimals=2)))
                ax0.set_xlabel('[um]')
                ax0.set_ylabel('[um]')
                if do_save_plot:
                    file="CRB" + "_pos_" + str(i) + '_step_' + str(step) + '.png'
                    path=ScanProject.Files['AnalysisConfig'][:-18]
                    fig.savefig(path + file)
                    plt.close()
                else:
                    plt.show()

    def waveFile(ScanProject,TrajToGet,TrajDictionary,do_save=1, FolderEnd=''):
        """
        Convert trajectory to wave files
        Each coordinate is output in its own file
        """
        
        txyznM=TrajDictionary[TrajToGet]*1.0

        Component=['x','y','z']
        for index in range(3):
            Traj=txyznM[index+1,:]
            scaled = np.int16(Traj/np.max(np.abs(Traj)) * 32767)
            #########################################################
            if do_save:            
                file="T_" + str(TrajToGet) + '_traj_' + Component[index] + '.wav'
                path=ScanProject.Files['AnalysisConfig'][:-18] +'Traj_Sound' + FolderEnd +'/'
                Path(path).mkdir(parents=True, exist_ok=True)
                _waveWrite(path + file, 4762, scaled) #4762, 44100
                print('Done!')

    def PlotVectorfield(ScanProject,traj, groundtruth):
        """
        Using traj and groundtruth perpoint to visualize how each point has moved and where
        traj - has the format t,x,y,z,n 
        groundtruth - has the format x,y,z
        """
        from mayavi import mlab
        xyz_ture=groundtruth*1e3 #[nm]
        xyz_Est=np.append(traj[1:4,:],np.expand_dims(groundtruth[:,-1],1),1)*1e3 #[nm] For some resone we do not evaluate the last point, thus we add it from the groundtruth
        
        #3D Matrix form 
        #Vector postions
        xx=np.reshape(xyz_ture[0,:],ScanProject.Sim_par['Npoints_xyz'])
        yy=np.reshape(xyz_ture[1,:],ScanProject.Sim_par['Npoints_xyz'])
        zz=np.reshape(xyz_ture[2,:],ScanProject.Sim_par['Npoints_xyz'])
        #Vector direction and magnitude 
        uu= np.reshape(xyz_Est[0,:]-xyz_ture[0,:],ScanProject.Sim_par['Npoints_xyz'])
        vv= np.reshape(xyz_Est[1,:]-xyz_ture[1,:],ScanProject.Sim_par['Npoints_xyz'])
        ww= np.reshape(xyz_Est[2,:]-xyz_ture[2,:],ScanProject.Sim_par['Npoints_xyz'])

        #1D list Form
        #Vector postions
        x=xyz_ture[0,:]
        y=xyz_ture[1,:]
        z=xyz_ture[2,:]
        #Vector direction and magnitude 
        u= xyz_Est[0,:]-xyz_ture[0,:]
        v= xyz_Est[1,:]-xyz_ture[1,:]
        w= xyz_Est[2,:]-xyz_ture[2,:]

        #Tophat 
        #STxyz=ScanProject.TrajectorySelectionRules['STxyz']*1e3 #[nm]
        #xyzCenterGuess=ScanProject.TrajectorySelectionRules['TopHatoffset']*1e3 #[nm] axis name [y,x,z]
        STxyz=np.array([0.4, 0.4, 0.5])*1e3 #[nm]
        xyzCenterGuess=np.array([0, 0, 0])*1e3 #[nm]
        
        xyzCenterGuess[2]=-xyzCenterGuess[2] #This since we flipp z in the reconstruction

        TopHatPrior=np.ones(xx.shape)
        #TopHatPrior[np.where(((xx-xyzCenterGuess[0])**2 / (STxyz[0]**2) + (yy-xyzCenterGuess[1])**2 / (STxyz[1]**2) + (zz-xyzCenterGuess[2])**2 / (STxyz[2]**2))>1.0)]=0  
        #TopHatPrior[np.where( np.invert((np.abs(xx-xyzCenterGuess[0])<STxyz[0]) * (np.abs(yy-xyzCenterGuess[1])<STxyz[1]) * (np.abs(zz-xyzCenterGuess[2])<STxyz[2])) )]=0  

        #Weed out acording to distance
        D=np.sqrt(uu**2+vv**2+ww**2)
        Bound=[0,1000]#[nm]
        NotOfInterest=np.where( (((D>=Bound[0])*(D<=Bound[1]))*TopHatPrior==False) )
        uu[NotOfInterest]=np.nan;vv[NotOfInterest]=np.nan;ww[NotOfInterest]=np.nan


        if 1:
            #Ploting 
            mlab.figure(fgcolor=(0., 0., 0.), bgcolor=(1, 1, 1))
            src_vector = mlab.pipeline.vector_field(uu,vv,ww)
            src_tophat=mlab.pipeline.scalar_field(TopHatPrior)
        
            #iso=mlab.pipeline.iso_surface(src_tophat, vmin=0.9, vmax=1.0,opacity=0.1)
            vec=mlab.pipeline.vectors(src_vector, mask_points=1, scale_factor=1/4) 
            mlab.colorbar(object=vec,nb_labels=20, orientation='vertical')
            #mlab.axes()
            #mlab.orientation_axes()
            mlab.outline()
            mlab.show()

            #mlab.quiver3d(x,y,z,u,v,w,mask_points=1,scale_factor=1.0); mlab.outline(); mlab.colorbar(nb_labels=20, orientation='vertical'); mlab.axes();mlab.orientation_axes();mlab.show()

        if 1:
            fig, axs = plt.subplots(3, 2)
            axs[0,0].plot(xx.flatten(),'-g')
            axs[0,0].plot((uu+xx).flatten(),'-r')
            axs[0,1].hist(uu.flatten(),bins=50)

            axs[1,0].plot(yy.flatten(),'-g')
            axs[1,0].plot((vv+yy).flatten(),'-r')
            axs[1,1].hist(vv.flatten(),bins=50)

            axs[2,0].plot(zz.flatten(),'-g')
            axs[2,0].plot((ww+zz).flatten(),'-r')
            axs[2,1].hist(ww.flatten(),bins=50)

            plt.show()
        print('324')
